This is messed up
Package: foo3
Version: 0.0.0.9000
Title: What the Package Does (one line, title case)
Description: What the package does (one paragraph).
Authors@R: person("Jim", "Hester", email = "james.f.hester@gmail.com", role = c("aut", "cre"))
License: GPL-3
Encoding: UTF-8
LazyData: true
ByteCompile: true
RoxygenNote: 6.0.1

Package: pkgdependstest
Title: Title
Version: 1.0.0
Author: Gábor Csárdi
Maintainer: Gábor Csárdi <csardi.gabor@gmail.com>
Description: Test package.
License: MIT + file LICENSE
LazyData: true
VignetteBuilder: knitr
Suggests:
  knitr,
  markdown,
  rmarkdown
Encoding: UTF-8

Package: badbuild
Title: Title
Version: 1.0.0
Author: Gábor Csárdi
Maintainer: Gábor Csárdi <csardi.gabor@gmail.com>
Description: Test package.
License: MIT + file LICENSE
LazyData: true
Encoding: UTF-8

# Generated by roxygen2: do not edit by hand

S3method("[",pkg_resolution_result)
S3method("[",pkg_scan_deps)
S3method("[",pkg_solution_result)
S3method("[",pkg_sysreqs_check_result)
S3method("[",pkgplan_downloads)
S3method(format,package_build_error)
S3method(format,package_packaging_error)
S3method(format,package_uncompress_error)
S3method(format,pkg_name_check)
S3method(format,pkg_name_check_basics)
S3method(format,pkg_name_check_sentiment)
S3method(format,pkg_name_check_urban)
S3method(format,pkg_name_check_wikipedia)
S3method(format,pkg_name_check_wiktionary)
S3method(format,pkg_scan_deps)
S3method(format,pkg_solution_failures)
S3method(format,pkg_solution_result)
S3method(format,pkg_sysreqs_check_result)
S3method(format,pkgplan_lp_problem)
S3method(format,remote_ref)
S3method(print,package_build_error)
S3method(print,package_packaging_error)
S3method(print,package_uncompress_error)
S3method(print,pkg_name_check)
S3method(print,pkg_name_check_basics)
S3method(print,pkg_name_check_sentiment)
S3method(print,pkg_name_check_urban)
S3method(print,pkg_name_check_wikipedia)
S3method(print,pkg_name_check_wiktionary)
S3method(print,pkg_scan_deps)
S3method(print,pkg_solution_failures)
S3method(print,pkg_solution_result)
S3method(print,pkg_sysreqs_check_result)
S3method(print,pkginstall_result)
S3method(print,pkgplan_lp_problem)
S3method(print,snapshot)
S3method(snapshot,pkg_resolution_result)
export(as_pkg_dependencies)
export(current_config)
export(current_r_platform)
export(default_platforms)
export(ghr)
export(ghrepo)
export(install_package_plan)
export(is_valid_package_name)
export(lib_status)
export(new_pkg_deps)
export(new_pkg_download_proposal)
export(new_pkg_installation_plan)
export(new_pkg_installation_proposal)
export(parse_pkg_ref)
export(parse_pkg_refs)
export(pkg_build)
export(pkg_dep_types)
export(pkg_dep_types_hard)
export(pkg_dep_types_soft)
export(pkg_deps)
export(pkg_download_proposal)
export(pkg_installation_plan)
export(pkg_installation_proposal)
export(pkg_name_check)
export(pkg_rx)
export(repo)
export(scan_deps)
export(sysreqs_check_installed)
export(sysreqs_db_list)
export(sysreqs_db_match)
export(sysreqs_db_update)
export(sysreqs_fix_installed)
export(sysreqs_install_plan)
export(sysreqs_is_supported)
export(sysreqs_list_system_packages)
export(sysreqs_platforms)
importFrom(stats,na.omit)
importFrom(utils,modifyList)
importFrom(utils,untar)
useDynLib(pkgdepends, .registration = TRUE, .fixes = "c_")

pkgdepends
================

<!-- README.md is generated from README.Rmd. Please edit that file -->

> Package Dependency Resolution, Downloads and Installation

<!-- badges: start -->

[![lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![R build
status](https://github.com/r-lib/pkgdepends/workflows/R-CMD-check/badge.svg)](https://github.com/r-lib/pkgdepends/actions)
[![R-CMD-check](https://github.com/r-lib/pkgdepends/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/pkgdepends/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/r-lib/pkgdepends/graph/badge.svg)](https://app.codecov.io/gh/r-lib/pkgdepends)
<!-- badges: end -->

pkgdepends is a toolkit for package dependencies, downloads and
installations, to be used in other packages. If you are looking for a
package manager, see [pak](https://github.com/r-lib/pak).

# Features

- Look up package dependencies recursively.
- Visualize package dependencies.
- Download packages and their dependencies.
- Install downloaded packages.
- Includes a dependency solver to find a consistent set of dependencies.
- Supports CRAN and Bioconductor packages automatically.
- Supports packages on GitHub and GitLab.
- Supports packages in git repositories.
- Supports package bundles or files on the web.
- Supports local package file and trees.
- Supports the `Remotes` entry in the `DESCRIPTION` file.
- Caches metadata and downloaded packages via
  [pkgcache](https://github.com/r-lib/pkgcache)
- Performs all downloads and HTTP queries concurrently.
- Builds and installs packages in parallel.

# Install

Install the package with:

``` r
install.packages("pkgdepends")
```

If you need the development version, install it with

``` r
pak::pak("r-lib/pkgdepends")
```

# Usage

``` r
library(pkgdepends)
```

## Package references

A package reference (ref) specifies a location from which an R package
can be obtained from. Examples:

    devtools
    cran::devtools
    bioc::Biobase
    r-lib/pkgdepends
    https://github.com/r-lib/pkgdepends
    local::~/works/shiny

See [“Package
references”](https://r-lib.github.io/pkgdepends/reference/pkg_refs.html)
for details.

## Package dependencies

Dependencies of the development version of the cli package:

``` r
pd <- new_pkg_deps("r-lib/pkgcache")
pd$solve()
pd$draw()
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="/Users/gaborcsardi/works/pkgdepends/man/figures/README/deps-dark.svg">
<img src="/Users/gaborcsardi/works/pkgdepends/man/figures/README/deps.svg" width="100%" />
</picture>

See the
[`pkg_deps`](https://r-lib.github.io/pkgdepends/reference/pkg_deps.html)
class for details.

## Package downloads

Downloading all dependencies of a package:

``` r
pdl <- new_pkg_download_proposal("r-lib/cli")
pdl$resolve()
pdl$download()
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="/Users/gaborcsardi/works/pkgdepends/man/figures/README/download-dark.svg">
<img src="/Users/gaborcsardi/works/pkgdepends/man/figures/README/download.svg" width="100%" />
</picture>

See the
[`pkg_download_proposal`](https://r-lib.github.io/pkgdepends/reference/pkg_download_proposal.html)
class for details.

## Package installation

Installing or updating a set of package:

``` r
lib <- tempfile()
dir.create(lib)
pdi <- new_pkg_installation_proposal(
  "r-lib/cli",
  config = list(library = lib)
)
pdi$solve()
pdi$download()
pdi$install()
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="/Users/gaborcsardi/works/pkgdepends/man/figures/README/install-dark.svg">
<img src="/Users/gaborcsardi/works/pkgdepends/man/figures/README/install.svg" width="100%" />
</picture>

## Dependency resolution

[`pkg_deps`](https://r-lib.github.io/pkgdepends/reference/pkg_deps.html),
[`pkg_download_proposal`](https://r-lib.github.io/pkgdepends/reference/pkg_download_proposal.html)
and
[`pkg_installation_proposal`](https://r-lib.github.io/pkgdepends/reference/pkg_installation_proposal.html)
all resolve their dependencies recursively, to obtain information about
all packages needed for the specified [package
references](https://r-lib.github.io/pkgdepends/reference/pkg_refs.html).
See [“Dependency
resolution”](https://r-lib.github.io/pkgdepends/reference/pkg_resolution.html)
for details.

## The dependency solver

The dependency solver takes the resolution information, and works out
the exact versions of each package that must be installed, such that
version and other requirements are satisfied. See [“The dependency
solver”](https://r-lib.github.io/pkgdepends/reference/pkg_solution.html)
for details.

## Installation plans

[`pkg_installation_proposal`](https://r-lib.github.io/pkgdepends/reference/pkg_installation_proposal.html)
can create installation plans, and then also install them. It is also
possible to import installation plans that were created by other tools.
See [“Installation
plans”](https://r-lib.github.io/pkgdepends/reference/install_plans.html)
for details.

## Configuration

The details of
[`pkg_deps`](https://r-lib.github.io/pkgdepends/reference/pkg_deps.html),
[`pkg_download_proposal`](https://r-lib.github.io/pkgdepends/reference/pkg_download_proposal.html)
and
[`pkg_installation_proposal`](https://r-lib.github.io/pkgdepends/reference/pkg_installation_proposal.html)
can be tuned with a list of configuration options. See
[“Configuration”](https://r-lib.github.io/pkgdepends/reference/pkg_config.html)
for details.

# Related

- [pak](https://github.com/r-lib/pak) – R package manager
- [pkgcache](https://github.com/r-lib/pkgcache) – Metadata and package
  cache
- [devtools](https://github.com/r-lib/devtools) – Tools for R package
  developers

## Code of Conduct

Please note that the pkgdepends project is released with a [Contributor
Code of
Conduct](https://r-lib.github.io/pkgdepends/dev/CODE_OF_CONDUCT.html).
By contributing to this project, you agree to abide by its terms.

# License

MIT (c) RStudio

parse_remote_deps <- function(specs, config, ...) {
  parsed_specs <- re_match(specs, type_deps_rx())
  parsed_specs$ref <- parsed_specs$.text
  cn <- setdiff(colnames(parsed_specs), c(".match", ".text"))
  parsed_specs <- parsed_specs[, cn]
  parsed_specs$type <- "deps"
  lapply(
    seq_len(nrow(parsed_specs)),
    function(i) as.list(parsed_specs[i, ])
  )
}

resolve_remote_deps <- function(
  remote,
  direct,
  config,
  cache,
  dependencies,
  ...
) {
  in_pkg <- is_package_root(remote$path)
  if (in_pkg || !is.dir(remote$path)) {
    ret <- resolve_remote_local(
      remote,
      direct,
      config,
      cache,
      dependencies,
      ...
    )
  } else {
    ret <- resolve_remote_local_autodeps(
      remote,
      direct,
      config,
      cache,
      dependencies,
      ...
    )
  }

  # We need to do some extra work for the case when a dependency
  # depends on the ref itself. E.g. when in pak::local_install_dev_deps()
  # a soft dependency depends on the local package.
  # - Because of this, we add a "-deps" postfix to the package name
  #   of the deps:: ref, so it not taken as the solution for the package
  #   by the solver.
  # - We also add a local:: ref for the package, because maybe the local
  #   package is not available from CRAN, or the upgrade policy is used
  #   and the local version is more recent. (This is typically the case
  #   while working on a package, and also on the CI.)

  # Workaround to be able to put this in a data frame
  ret$metadata <- list(ret$metadata)

  # Need to pass this to te new result
  unknown <- ret$unknown_deps
  ret$unknown_deps <- NULL

  # Create the local:: ref and update it properly
  local <- ret
  local$type <- "local"
  local$ref <- sub("^deps::", "local::", local$ref)
  local$direct <- FALSE
  local$remote[[1]] <- parse_pkg_ref(remote$ref)

  # Now make sure deps:: is not used by the solver.
  # The installer will treat deps:: specially and just ignore it, anyway.
  ret$package <- paste0(ret$package, "-deps")
  ret$sources <- list(character())

  # rbind.data.frame is not good with the list columns, TODO
  ret2 <- rbind(as_data_frame(ret), as_data_frame(local))
  attr(ret2, "unknown_deps") <- unknown

  ret2
}

resolve_remote_local_autodeps <- function(
  remote,
  direct,
  config,
  cache,
  dependencies,
  ...
) {
  proc <- cli::cli_process_start(
    "Scanning dependencies in {.path {remote$path}}"
  )
  deps <- scan_deps(remote$path, root = remote$path)
  cli::cli_process_done(proc)
  cli::cli_verbatim(paste(c(format(deps), ""), collapse = "\n"))
  cli::cli_verbatim(" ")
  tmpdesc <- tempfile()
  on.exit(unlink(tmpdesc), add = TRUE)
  dsc <- desc::desc("!new")
  hard <- deps$package[deps$type == "prod"]
  soft <- deps$package[deps$type != "prod"]
  dsc$set(
    Package = "localprojectautoscan",
    Version = "1.0.0",
    Title = "Local Project",
    License = "Unknown"
  )
  for (p in hard) {
    dsc$set_dep(p, type = "Depends")
  }
  for (s in soft) {
    dsc$set_dep(p, type = "Suggests")
  }
  dsc$write(tmpdesc)
  resolve_from_description(
    tmpdesc,
    paste0("file://", normalizePath(tmpdesc)),
    remote,
    direct,
    config,
    cache,
    dependencies[[2 - direct]]
  )
}

download_remote_deps <- function(
  resolution,
  target,
  target_tree,
  config,
  cache,
  which,
  on_progress
) {
  ## Nothing to do here
  "Had"
}

satisfy_remote_deps <- function(resolution, candidate, config, ...) {
  ## TODO: we can probably do better than this
  FALSE
}

installedok_remote_deps <- function(installed, solution, config, ...) {
  FALSE
}

## ----------------------------------------------------------------------
## Internal functions

type_deps_rx <- function() {
  paste0(
    "^",
    "(?:deps::)",
    "(?<path>.*)",
    "$"
  )
}

test_that("parse_remote_deps", {
  expect_snapshot(parse_remote_deps("deps::/foo/bar"))
  expect_snapshot(parse_remote_deps("deps::foo/bar"))
  expect_snapshot(parse_remote_deps("deps::~foo/bar"))
})

test_that("resolve_remote_deps", {
  setup_fake_apps()
  tmp <- withr::local_tempdir()
  withr::local_dir(tmp)
  writeLines(
    c(
      "Package: mypackage",
      "Version: 1.0.0",
      "Depends: pkg2",
      "Imports: pkg3",
      "Suggests: needspak"
    ),
    "DESCRIPTION"
  )

  prop <- suppressMessages(new_pkg_installation_proposal(
    "deps::.",
    config = list(library = tempfile())
  ))
  suppressMessages(prop$resolve())
  expect_snapshot(
    prop$get_resolution()[, c("ref", "package", "version")]
  )
  prop$solve()
  expect_snapshot(
    prop$get_solution()
  )
})

test_that("download_remote_deps", {
  expect_equal(download_remote_deps(), "Had")
})

test_that("satisfy_remote_seps", {
  expect_false(satisfy_remote_deps())
})

test_that("installedok_remote_deps", {
  expect_false(installedok_remote_deps())
})

test_that("type_deps_rx", {
  expect_true(grepl(type_deps_rx(), "deps::/foo/bar", perl = TRUE))
  expect_true(grepl(type_deps_rx(), "deps::foo/bar", perl = TRUE))
  expect_true(grepl(type_deps_rx(), "deps::~/foo/bar", perl = TRUE))
})

#' Scan R code for dependent packages
#'
#' Scan all R files of a project or directory for packages used within
#' them. It parses R code to find `library(package)`, `package::func()`,
#' and similar calls that imply package dependencies. See details below.
#'
#' # Detected dependencies
#'
#' `scan_deps()` detects package dependencies from these R expressions:
#' * `library()`, `require()`, `loadNamespace()` and `requireNamespace`
#'   calls.
#' * `::` and `:::` operators.
#' * Any of the calls in this list in R code from R markdown or quarto
#'   `R` and `Rscript` (case insensitive) code blocks or inline R code.
#' * A dependency on the methods package is inferred from finding
#'   `setClass()` and/or `setGeneric()` calls.
#' * `xfun::pkg_attach()` and `xfun::pkg_attach2()` calls.
#' * `pacman::p_load()` calls.
#' * `modules::import()` and `modules::module()` calls.
#' * `import::from()`, `import::here()` and `import::into()` calls.
#' * `box::use()` calls.
#' * `targets::tar_option_set(packages = ...)` calls.
#' * Any of the calls in this list in R code from `glue::glue()` strings.
#' * A dependency on the svglite package is inferred from
#'   `ggplot2::ggsave()` calls saving `.svg` files.
#' * Dependencies from `parsnip::set_engine()` calls, the default engine
#'   to package mapping is:
#'   - `"glm"` -> stats,
#'   - `"glmnet"` -> glmnet,
#'   - `"keras"` -> keras,
#'   - `"kknn"` -> kknn,
#'   - `"nnet"` -> nnet,
#'   - `"rpart"` -> rpart,
#'   - `"spark"` -> sparklyr,
#'   - `"stan"` -> rstanarm.
#'   You can override the default mapping by setting the
#'   `renv.parsnip.engines` option to a named list.
#' * A dependency on the xml2 package is inferred from using the
#'   "Junit" reporter (`JunitReporter`) from the testthat package.
#' * A dependency on the ragg package is inferred from setting the default
#'   knitr device (`dev` option) to `"ragg_png"`.
#' * A dependency on the hexbin package is inferred from using
#'   `ggplot2::geom_hex()`.
#' * A custom symbol name to package name mapping can be defined in the
#'   `renv.dependencies.database` option. This must be a named list of
#'   named lists, where the outer names are package names, the inner names
#'   are function or object names, and the values are package names. E.g.
#'   ```
#'   options(renv.dependencies.database = list(
#'     ggplot2 = list(geom_hex = "hexbin"),
#'     testthat = list(JunitReporter = "xml2")
#'   ))
#'   ```
#'
#' # Dependency types
#'
#' `scan_deps()` classifies package dependencies into three groups, based
#' on which files they were found:
#' * Production dependencies: `"prod"`.
#' * Test dependencies: `"test"`.
#' * Development dependencies: `"dev"`.
#'
#' @param path Files and/or directories to scan. Defaults to the current
#'   project, detected by finding the first parent directory of the current
#'   working directory, that contains a file or directory called
#'   `r cli::format_inline("{.or {pkgdepends:::project_root_anchors}}")`.
#'   (Note that this is different from `renv::dependencies()`, which only
#'   scans the current working directory by default!)
#'
#'   If `path` is not `NULL`, then only the specified files and directories
#'   are scanned, the directories recursively. In this case the `root`
#'   argument is used as the project root, to find `.gitignore` and
#'   `.renvignore` files. All entries of `path` must be within the `root`,
#'   the project root.
#' @param root The root directory of the project. It is used to find the
#'   `.gitignore` and `.renvignore` files. By default the same algorithm
#'   is used to detect this as for `path`. If `path` is specified and it is
#'   not within the detected or specified `root`, `scan_path()` throws an
#'   error.
#' @return Data frame with columns:
#'   * `path`: Path to the file in which the dependencies was found.
#'   * `package`: Detected package dependency. Typically a package name,
#'     but it can also be a package reference, e.g. a package from GitHub.
#'   * `type`: Dependency type. It is `"prod"`, `"test"` or `"dev"`. See
#'     'Dependency types' below.
#'   * `code`: The piece of code the dependency was extracted from.
#'   * `start_row`: Start row of the code the dependency was extracted
#'     from.
#'   * `start_column`: Start column of the code the dependency was
#'     extracted from.
#'   * `start_byte`: Start byte of the code the dependency was extracted
#'     from.
#'
#' Note the data frame may contain the same package multiple times, if it
#' was detected multiple times, e.g. multiple `library()` calls load the
#' same package.
#'
#' @export

scan_deps <- function(path = NULL, root = NULL) {
  if (is.null(path)) {
    path <- find_project_root()
    root <- root %||% path
  } else {
    root <- root %||% find_project_root()
  }
  assert_that(
    is_string(root),
    is_character(path)
  )
  if (!file.exists(root)) {
    throw(pkg_error("Project root {.path {root}} does not exist."))
  }
  if (any(bad <- !file.exists(path))) {
    throw(pkg_error("Path{?s} do{?es/} not exist: {.path {path[bad]}}."))
  }
  check_inside_dir(root, path)
  paths <- c(
    dir(path, pattern = scan_deps_pattern(), recursive = TRUE),
    if (root %in% path) {
      dir(root, pattern = scan_deps_pattern_root(), recursive = FALSE)
    }
  )
  full_paths <- normalizePath(file.path(path, paths), winslash = "/")
  deps_list <- lapply(full_paths, scan_path_deps)
  deps <- do.call("rbind", c(list(scan_deps_df()), deps_list))
  # write back the relative paths
  deps$path <- paths[match(deps$path, full_paths)]
  deps$type <- get_dep_type_from_path(deps$path, deps$type)
  class(deps) <- c("pkg_scan_deps", class(deps))
  deps
}

scan_deps_pattern <- function() {
  ptrns <- c(
    "[.]R$",
    "[.]r$",
    "[.]Rmd$",
    "[.]rmd$",
    "[.]rmarkdown",
    "[.]Rmarkdown",
    "[.][Rr]nw",
    "[.]qmd$",
    "[.]Rproj$",
    "^_bookdown[.]yml$",
    "^_quarto[.]yml$",
    NULL
  )
  paste0("(", paste0(collapse = "|", ptrns), ")")
}

scan_deps_pattern_root <- function() {
  ptrns <- c(
    "^DESCRIPTION$",
    "^NAMESPACE$",
    "^_pkgdown.yml$",
    "^renv[.]lock$",
    "^rsconnect$",
    NULL
  )
  paste0("(", paste0(collapse = "|", ptrns), ")")
}

scan_deps_file_type_use_basename <- function() {
  # for these we don't use the extension but the basename
  # to decide which parser to use
  c(
    "DESCRIPTION",
    "NAMESPACE",
    "_bookdown.yml",
    "_pkgdown.yml",
    "_quarto.yml",
    "renv.lock",
    NULL
  )
}

scan_deps_file_type <- function(paths) {
  ext <- tolower(file_ext(paths))
  bsn <- basename(paths)
  ifelse(
    bsn %in% scan_deps_file_type_use_basename() | is.na(ext),
    bsn,
    ext
  )
}

# -------------------------------------------------------------------------

# needs to increase as the deps discovry code changes, otherwise we don't
# apply the new discovery code
deps_cache_version <- 2L

get_deps_cache_path <- function(hash = NULL) {
  root <- file.path(get_user_cache_dir()$root, "deps", deps_cache_version)
  if (is.null(hash)) {
    root
  } else {
    file.path(root, substr(hash, 1, 2), hash)
  }
}

clear_deps_cache <- function() {
  unlink(dirname(get_deps_cache_path()), recursive = TRUE)
}

re_r_dep <- function() {
  db <- renv_dependencies_database()
  fns <- as.character(unlist(lapply(db, names)))
  paste0(
    collapse = "|",
    c(
      "library",
      "require",
      "loadNamespace",
      "::",
      "setClass",
      "setGeneric",
      "pkg_attach",
      "p_load",
      "module",
      "import",
      "box::",
      "tar_option_set",
      "glue",
      "ggsave",
      "set_engine",
      "opts_chunk",
      "geom_hex",
      "JunitReporter",
      fns
    )
  )
}

scan_path_deps <- function(path) {
  code <- if (file.info(path)$isdir) {
    NULL
  } else {
    readBin(path, "raw", file.size(path))
  }

  # check if already known, set path
  should_cache <- scan_path_should_cache(path)
  if (should_cache) {
    hash <- cli::hash_raw_xxhash(code)
    cache <- get_deps_cache_path(hash)
    if (file.exists(cache)) {
      deps <- readRDS(cache)
      if (!is.null(deps) && nrow(deps) > 0) {
        deps$path <- path
        deps$type <- get_dep_type_from_path(deps$path, deps$type)
      }
      return(deps)
    }
  }

  # scan it if it is worth it, based on a quick check
  maybe_has_deps <- scan_deps_file_type(path) != "r" ||
    length(grepRaw(re_r_dep(), code)) > 0
  deps <- if (maybe_has_deps) {
    scan_path_deps_do(code, path)
  }

  # save it to the cache, but anonimize it first. If no deps, save NULL
  if (should_cache) {
    deps_no_path <- deps
    if (!is.null(deps_no_path) && nrow(deps_no_path) > 0) {
      deps_no_path$path <- ""
    }
    dir.create(dirname(cache), showWarnings = FALSE, recursive = TRUE)
    saveRDS(deps_no_path, cache)
  }

  deps
}

scan_deps_df <- function(
  path = character(),
  ref = package,
  package = character(),
  version = "*",
  type = get_dep_type_from_path(path),
  code = character(),
  start_row = 1L,
  start_column = 1L,
  start_byte = 1L
) {
  data_frame(
    path = path,
    ref = ref,
    package = package,
    version = version,
    type = type,
    code = code,
    start_row = start_row,
    start_column = start_column,
    start_byte = start_byte
  )
}

scan_path_should_cache <- function(paths) {
  # we don't want to cache the ones that depend on the file
  # name, because caching is content-based.
  !basename(paths) %in%
    c(
      "_bookdown.yml",
      "_pkgdown.yml",
      "_quarto.yml",
      "renv.lock",
      "rsconnect",
      NULL
    )
}

scan_path_deps_do <- function(code, path) {
  ext <- scan_deps_file_type(path)
  switch(
    ext,
    ".r" = scan_path_deps_do_r(code, path),
    ".qmd" = ,
    ".rmarkdown" = ,
    ".rmd" = scan_path_deps_do_rmd(code, path),
    ".rnw" = scan_path_deps_do_rnw(code, path),
    ".ipynb" = scan_path_deps_do_ipynb(code, path),
    "DESCRIPTION" = scan_path_deps_do_dsc(code, path),
    "NAMESPACE" = scan_path_deps_do_namespace(code, path),
    "_bookdown.yml" = scan_path_deps_do_bookdown(code, path),
    "_pkgdown.yml" = scan_path_deps_do_pkgdown(code, path),
    "_quarto.yml" = scan_path_deps_do_quarto(code, path),
    "renv.lock" = scan_path_deps_do_renv_lock(code, path),
    "rsconnect" = scan_path_deps_do_rsconnect(code, path),
    ".rproj" = scan_path_deps_do_rproj(code, path),
    stop("Cannot parse ", ext, " file for dependencies, internal error")
  )
}

# -------------------------------------------------------------------------

scan_path_deps_do_r <- function(code, path, ranges = NULL) {
  hits <- code_query(code, q_deps(), ranges = ranges)
  mct <- hits$matched_captures

  # q_library_0 hits are generic ones, only use them if they are not hit
  gen_pat <- hits$patterns$id[hits$patterns$name == "q_library_0"]
  gen_hits <- mct[mct$pattern %in% gen_pat, ]

  # for these patterns we need to work from the function names
  fn_patterns <- "methods"
  fn_pat <- hits$patterns$id[hits$patterns$name %in% fn_patterns]
  fn_hits <- mct[mct$pattern %in% fn_pat, ]

  # junit reporter needs xml2
  jr_patterns <- "junit_reporter"
  jr_pat <- hits$patterns$id[hits$patterns$name %in% jr_patterns]
  jr_hits <- mct[mct$pattern %in% jr_pat, ]

  # knit ragg_png device needs ragg
  ragg_patterns <- "knitr_dev"
  ragg_pat <- hits$patterns$id[hits$patterns$name %in% ragg_patterns]
  ragg_hits <- mct[mct$pattern %in% ragg_pat, ]

  # database that matches symbols to packages
  db_patterns <- "database"
  db_pat <- hits$patterns$id[hits$patterns$name %in% db_patterns]
  db_hits <- mct[mct$pattern %in% db_pat, ]

  pkg_hits <- mct[
    !mct$pattern %in% c(gen_pat, fn_pat, jr_pat, ragg_pat, db_hits),
  ]
  rbind(
    if (nrow(pkg_hits) > 0) scan_path_deps_do_pkg_hits(pkg_hits, path),
    if (nrow(fn_hits) > 0) scan_path_deps_do_fn_hits(fn_hits, path),
    if (nrow(gen_hits) > 0) scan_path_deps_do_gen_hits(gen_hits, path),
    if (nrow(jr_hits) > 0) scan_path_deps_do_jr_hits(jr_hits, path),
    if (nrow(ragg_hits) > 0) scan_pat_deps_do_ragg_hits(ragg_hits, path),
    if (nrow(db_hits) > 0) scan_pat_deps_do_db_hits(db_hits, path)
  )
}

scan_path_deps_do_pkg_hits <- function(hits, path) {
  pkg <- hits$code[hits$name == "pkg-name"]
  scan_deps_df(
    path = path,
    package = pkg,
    code = hits$code[hits$name == "dep-code"],
    start_row = hits$start_row[hits$name == "dep-code"],
    start_column = hits$start_column[hits$name == "dep-code"],
    start_byte = hits$start_byte[hits$name == "dep-code"]
  )
}

scan_path_deps_do_fn_hits <- function(hits, path) {
  fn_pkg_map <- c(setClass = "methods", setGeneric = "methods")
  fn_names <- hits$code[hits$name == "fn-name"]
  scan_deps_df(
    path = path,
    package = fn_pkg_map[fn_names],
    code = hits$code[hits$name == "dep-code"],
    start_row = hits$start_row[hits$name == "dep-code"],
    start_column = hits$start_column[hits$name == "dep-code"],
    start_byte = hits$start_byte[hits$name == "dep-code"]
  )
}

scan_path_deps_do_gen_hits <- function(hits, path) {
  code <- hits$code[hits$name == "dep-code"]
  fn <- hits$code[hits$name == "fn-name"]
  ns <- vcapply(
    unique(hits$match),
    function(m) {
      hits[hits$match == m & hits$name == "ns-name", ]$code %|0|%
        NA_character_
    }
  )
  pkgs <- lapply(seq_along(code), function(i) {
    safe_parse_pkg_from_call(ns[i], fn[i], code[i])
  })
  pkgs_count <- lengths(pkgs)
  scan_deps_df(
    path = path,
    package = unlist(pkgs),
    code = rep(code, pkgs_count),
    start_row = rep(hits$start_row[hits$name == "dep-code"], pkgs_count),
    start_column = rep(hits$start_column[hits$name == "dep-code"], pkgs_count),
    start_byte = rep(hits$start_byte[hits$name == "dep-code"], pkgs_count)
  )
}

scan_path_deps_do_jr_hits <- function(hits, path) {
  code <- hits$code[hits$name == "dep-code"]
  scan_deps_df(
    path = path,
    package = "xml2",
    code = code,
    start_row = hits$start_row[hits$name == "dep-code"],
    start_column = hits$start_column[hits$name == "dep-code"],
    start_byte = hits$start_byte[hits$name == "dep-code"]
  )
}

scan_pat_deps_do_ragg_hits <- function(hits, path) {
  wcodes <- which(hits$name == "dep-code")
  for (wc in wcodes) {
    expr <- parse(text = hits$code[wc], keep.source = FALSE)
    matched <- match.call(
      function(...) {},
      expr,
      expand.dots = FALSE
    )
    args <- matched[["..."]]
    if ("dev" %in% names(args) && args[["dev"]] == "ragg_png") {
      return(scan_deps_df(
        path = path,
        package = "ragg",
        code = hits$code[wc],
        start_row = hits$start_row[wc],
        start_column = hits$start_column[wc],
        start_byte = hits$start_byte[wc]
      ))
    }
  }
  NULL
}

scan_pat_deps_do_db_hits <- function(hits, path) {
  db <- renv_dependencies_database()
  fns <- unlist(lapply(db, names))
  map <- unlist(unname(db), recursive = FALSE)
  pkgs <- unlist(map[hits$code])
  scan_deps_df(
    path = path,
    package = pkgs,
    code = hits$code,
    start_row = hits$start_row,
    start_column = hits$start_column,
    start_byte = hits$start_byte
  )
}

# nocov start
prot_xfun_pkg_attach <- function(..., install, message) {}
prot_xfun_pkg_attach2 <- function(...) {}
prot_pacman_p_load <- function(
  ...,
  char,
  install,
  update,
  character.only
) {}
prot_modules_import <- function(
  from,
  ...,
  attach = TRUE,
  where = parent.frame()
) {}
prot_modules_module <- function(
  expr = {},
  topEncl = NULL,
  envir = parent.frame()
) {}
prot_import_from <- function(.from, ..., .character_only = FALSE) {}
prot_import_here <- function(.from, ..., .character_only = FALSE) {}
prot_import_into <- function(
  .into,
  ...,
  .from,
  .library = NULL,
  .directory = ".",
  .all = NULL,
  .except = character(),
  .chdir = TRUE,
  .character_only = FALSE,
  .S3 = FALSE
) {}
prot_box_use <- function(...) {}
prot_targets_tar_option_set <- function(
  tidy_eval = NULL,
  packages = NULL,
  ...
) {}
prot_glue_glue <- function(
  ...,
  .sep = "",
  .envir = parent.frame(),
  .open = "{",
  .close = "}"
) {}
prot_ggplot2_ggsave <- function(filename, ...) {}
prot_parsnip_set_engine <- function(object, engine, ...) {}
prot_r6_r6class <- function(
  classname = NULL,
  public = list(),
  private = NULL,
  active = NULL,
  inherit = NULL,
  ...
) {}
prot_testthat_test_package <- function(package, reporter = NULL, ...) {}
prot_testthat_test_dir <- function(
  path,
  filter = NULL,
  reporter = NULL,
  ...
) {}
prot_testthat_test_file <- function(path, reporter = NULL, ...) {}
# nocov end

safe_parse_pkg_from_call <- function(ns, fn, code) {
  tryCatch(
    parse_pkg_from_call(ns, fn, code),
    error = function(...) NULL
  )
}

parse_pkg_from_call_match <- function(fn, code) {
  expr <- parse(text = code, keep.source = FALSE)
  fun <- switch(
    fn,
    "library" = base::library,
    "require" = base::require,
    "loadNamespace" = base::loadNamespace,
    "requireNamespace" = base::requireNamespace,
    "pkg_attach" = prot_xfun_pkg_attach,
    "pkg_attach2" = prot_xfun_pkg_attach2,
    "p_load" = prot_pacman_p_load,
    "import" = prot_modules_import,
    "module" = prot_modules_module,
    "from" = prot_import_from,
    "here" = prot_import_here,
    "into" = prot_import_into,
    "use" = prot_box_use,
    "tar_option_set" = prot_targets_tar_option_set,
    "glue" = prot_glue_glue,
    "ggsave" = prot_ggplot2_ggsave,
    "set_engine" = prot_parsnip_set_engine,
    "R6Class" = prot_r6_r6class,
    "test_package" = prot_testthat_test_package,
    "test_dir" = prot_testthat_test_dir,
    "test_file" = prot_testthat_test_file
  )
  match.call(fun, expr, expand.dots = FALSE)
}

parse_pkg_from_call <- function(ns, fn, code) {
  matched <- parse_pkg_from_call_match(fn, code)
  switch(
    fn,
    "library" = ,
    "require" = parse_pkg_from_call_library(ns, fn, matched),
    "loadNamespace" = ,
    "requireNamespace" = parse_pkg_from_call_loadnamespace(ns, fn, matched),
    "pkg_attach" = ,
    "pkg_attach2" = parse_pkg_from_call_xfun(ns, fn, matched),
    "p_load" = parse_pkg_from_call_pacman(ns, fn, matched),
    "import" = parse_pkg_from_call_modules_import(ns, fn, matched),
    "module" = parse_pkg_from_call_modules_module(ns, fn, matched),
    "from" = ,
    "here" = ,
    "into" = parse_pkg_from_call_import(ns, fn, matched),
    "use" = parse_pkg_from_call_box(ns, fn, matched),
    "tar_option_set" = parse_pkg_from_call_targets(ns, fn, matched),
    "glue" = parse_pkg_from_call_glue(ns, fn, matched),
    "ggsave" = parse_pkg_from_call_ggplot2(ns, fn, matched),
    "set_engine" = parse_pkg_from_call_parsnip(ns, fn, matched),
    "R6Class" = parse_pkg_from_call_testthat_r6class(ns, fn, matched),
    "test_package" = ,
    "test_dir" = ,
    "test_file" = parse_pkg_from_call_testthat_test(ns, fn, matched)
  )
}

parse_pkg_from_call_library <- function(ns, fn, matched) {
  if (!is.na(ns) && ns != "base") {
    return(NULL)
  }
  pkg <- matched[["package"]]
  if (is.character(pkg) && length(pkg) == 1) {
    return(pkg)
  }
  if (
    is.symbol(pkg) &&
      identical(matched[["character.only"]] %||% FALSE, FALSE)
  ) {
    return(as.character(pkg))
  }
  NULL
}

parse_pkg_from_call_loadnamespace <- function(ns, fn, matched) {
  if (!is.na(ns) && ns != "base") {
    return(NULL)
  }
  pkg <- matched[["package"]]
  if (is.character(pkg) && length(pkg) == 1) {
    return(pkg)
  }
  NULL
}

parse_pkg_from_call_xfun <- function(ns, fn, matched) {
  if (!is.na(ns) && ns != "xfun") {
    return(NULL)
  }
  pkgs <- unlist(lapply(
    matched[["..."]],
    function(x) if (is.character(x)) x
  ))
  if (length(pkgs) > 0) {
    return(pkgs)
  }
  NULL
}

parse_pkg_from_call_pacman <- function(ns, fn, matched) {
  if (!is.na(ns) && ns != "pacman") {
    return(NULL)
  }
  # list of characters and symbols
  pkgs <- as.list(matched[["..."]])

  # character vector or scalar
  char <- matched[["char"]]
  if (length(char) > 0 && char[[1]] == quote(c)) {
    pkgs <- c(pkgs, as.list(char[-1]))
  } else if (is.character(char)) {
    pkgs <- c(pkgs, as.list(char))
  }
  if (matched[["character.only"]] %||% FALSE) {
    pkgs <- pkgs[vlapply(pkgs, function(x) is.character(x))]
  } else {
    pkgs <- pkgs[vlapply(pkgs, function(x) is.symbol(x) || is.character(x))]
  }
  pkgs <- vcapply(pkgs, as.character)
  if (length(pkgs) > 0) {
    return(pkgs)
  }
  NULL
}

parse_pkg_from_call_modules_import <- function(ns, fn, matched) {
  if (!is.na(ns) && ns != "modules") {
    return(NULL)
  }
  pkgs <- as.character(matched[["from"]])
  if (length(pkgs) > 0) {
    return(pkgs)
  }
  NULL
}

parse_pkg_from_call_modules_module <- function(ns, fn, matched) {
  if (!is.na(ns) && ns != "modules") {
    return(NULL)
  }
  expr <- as.character(matched[["expr"]])
  hits <- code_query(expr, q_module_import())[["matched_captures"]]
  code <- hits$code[hits$name == "dep-code"]
  pkgs <- lapply(seq_along(code), function(i) {
    safe_parse_pkg_from_call(ns, "import", code[i])
  })
  if (length(pkgs) > 0) {
    return(unlist(pkgs))
  }
  NULL
}

parse_pkg_from_call_import <- function(ns, fn, matched) {
  if (!is.na(ns) && ns != "import") {
    return(NULL)
  }
  from <- matched[[".from"]]
  if (is.symbol(from)) {
    if (!identical(matched[[".character_only"]], TRUE)) {
      from <- as.character(from)
    }
  }
  if (!is.character(from) || length(from) != 1) {
    return(NULL)
  }

  # '.from' can also be an R script; if it appears to be a path, then ignore it
  # https://github.com/rstudio/renv/issues/1743
  if (grepl("\\.[rR]$", from, perl = TRUE) && grepl("[/\\]", from)) {
    return(NULL)
  }

  from
}

parse_pkg_from_call_box <- function(ns, fn, matched) {
  if (!is.na(ns) && ns != "box") {
    return(NULL)
  }
  args <- as.list(matched[["..."]])
  pkgs <- na.omit(vcapply(args, function(arg) {
    if (!is.symbol(arg) && identical(arg[[1]], quote(`/`))) {
      return(NA_character_)
    }
    name <- if (is.symbol(arg) && !identical(arg, quote(expr = ))) {
      as.character(arg)
    } else if (
      identical(arg[[1]], quote(`[`)) &&
        length(arg) > 1L &&
        is.symbol(arg[[2L]])
    ) {
      as.character(arg[[2L]])
    }
    if (is.null(name) || name == "." || name == "..") {
      return(NA_character_)
    }
    name
  }))

  if (length(pkgs) > 0) {
    return(pkgs)
  }
  NULL
}

parse_pkg_from_call_targets <- function(ns, fn, matched) {
  if (!is.na(ns) && ns != "targets") {
    return(NULL)
  }
  pkgs <- matched[["packages"]]
  pkgs <- dependencies_eval(pkgs)
  if (is.character(pkgs) && length(pkgs) > 0) {
    return(pkgs)
  }
  NULL
}

# from renv:::renv_dependencies_eval
dependencies_eval <- function(expr) {
  syms <- c(
    "list",
    "c",
    "T",
    "F",
    "{",
    "(",
    "[",
    "[[",
    "::",
    ":::",
    "$",
    "@",
    ":",
    "+",
    "-",
    "*",
    "/",
    "<",
    ">",
    "<=",
    ">=",
    "==",
    "!=",
    "!",
    "&",
    "&&",
    "|",
    "||"
  )
  vals <- mget(syms, envir = baseenv())
  envir <- list2env(vals, parent = emptyenv())
  eval(expr, envir = envir)
}

parse_pkg_from_call_glue <- function(ns, fn, matched) {
  if (!is.na(ns) && ns != "glue") {
    return(NULL)
  }
  args <- as.list(matched[["..."]])
  nm <- names(args) %||% rep.int("", length(args))
  str <- args[!nzchar(nm) & vlapply(args, is.character)]
  code <- character()
  for (s in str) {
    asNamespace("cli")$glue(
      s,
      .open = matched[[".open"]] %||% "{",
      .close = matched[[".close"]] %||% "}",
      .transformer = function(x, envir) {
        code <<- c(code, x)
      }
    )
  }

  pkgs <- unlist(lapply(
    code,
    function(x) scan_path_deps_do_r(x, path = "")[["package"]]
  ))
  if (length(pkgs) > 0) {
    return(pkgs)
  }
  NULL
}

parse_pkg_from_call_ggplot2 <- function(ns, fn, matched) {
  if (!is.na(ns) && ns != "ggplot2") {
    return(NULL)
  }
  fn <- matched[["filename"]]
  if (!is.character(fn)) {
    return(NULL)
  }
  # check for attempts to save to '.svg', and assume svglite is
  # required in this scenario.
  if (any(endsWith(fn, ".svg"))) {
    return("svglite")
  }
  NULL
}

parse_pkg_from_call_parsnip <- function(ns, fn, matched) {
  if (!is.na(ns) && ns != "parsnip") {
    return(NULL)
  }
  engine <- matched[["engine"]]
  if (!is.character(engine) || length(engine) != 1L) {
    return(NULL)
  }

  map <- getOption(
    "renv.parsnip.engines",
    default = list(
      glm = "stats",
      glmnet = "glmnet",
      keras = "keras",
      kknn = "kknn",
      nnet = "nnet",
      rpart = "rpart",
      spark = "sparklyr",
      stan = "rstanarm"
    )
  )

  pkgs <- if (is.function(map)) {
    map(engine)
  } else {
    map[[engine]]
  }

  if (length(pkgs) > 0) {
    return(pkgs)
  }
  NULL
}

parse_pkg_from_call_testthat_r6class <- function(ns, fn, matched) {
  if (!is.na(ns) && ns != "R6") {
    return(NULL)
  }
  inherit <- matched[["inherit"]]
  if (
    identical(inherit, quote(JunitReporter)) ||
      identical(inherit, quote(testthat::JunitReporter))
  ) {
    return("xml2")
  }
  NULL
}

parse_pkg_from_call_testthat_test <- function(ns, fn, matched) {
  if (!is.na(ns) && ns != "testthat") {
    return(NULL)
  }
  reporter <- matched[["reporter"]]
  if (
    identical(reporter, "Junit") ||
      identical(reporter, "junit") ||
      identical(reporter, quote(JunitReporter)) ||
      identical(reporter, quote(JunitReporter))
  ) {
    return("xml2")
  }
  NULL
}

# -------------------------------------------------------------------------

scan_path_deps_do_rmd <- function(code, path) {
  hits <- code_query(code, language = "markdown", query = q_deps_rmd())
  inl_pat <- hits$patterns$id[hits$patterns$name == "inline"]
  inl_hits <- hits$matched_captures[
    hits$matched_captures$pattern %in% inl_pat,
  ]
  hdr_pat <- hits$patterns$id[hits$patterns$name == "header"]
  hdr_hits <- hits$matched_captures[
    hits$matched_captures$pattern %in% hdr_pat,
  ]
  blk_hits <- hits$matched_captures[
    !hits$matched_captures$pattern %in% c(inl_pat, hdr_pat),
  ]
  rbind(
    if (nrow(inl_hits)) scan_path_deps_do_inline_hits(code, inl_hits, path),
    if (nrow(blk_hits)) scan_path_deps_do_block_hits(code, blk_hits, path),
    if (nrow(hdr_hits)) scan_path_deps_do_header_hits(code, hdr_hits, path),
    if (nrow(blk_hits)) scan_path_deps_do_rmarkdown(code, blk_hits, path)
  )
}

range_cols <- c(
  "start_row",
  "start_column",
  "end_row",
  "end_column",
  "start_byte",
  "end_byte"
)

scan_path_deps_do_inline_hits <- function(code, inl_hits, path) {
  wcnd <- which(inl_hits$name == "inline")
  wcnd <- wcnd[grepl("`", inl_hits$code[wcnd], fixed = TRUE)]
  wcnd <- wcnd[grepl(re_r_dep(), inl_hits$code[wcnd])]
  if (length(wcnd) == 0) {
    return(NULL)
  }

  inl_ranges <- inl_hits[wcnd, range_cols]
  r_hits <- code_query(
    code,
    language = "markdown-inline",
    ranges = inl_ranges,
    query = q_deps_rmd_inline()
  )
  cpt <- r_hits$matched_captures
  pre_drop <- nchar(cpt$code[cpt$name == "csd1"])
  post_drop <- nchar(cpt$code[cpt$name == "csd2"])
  r_code <- omit_pre_post(cpt$code[cpt$name == "code"], pre_drop, post_drop)
  wcnd2 <- substr(r_code, 1, 2) == "r " & grepl(re_r_dep(), r_code)
  if (!any(wcnd2)) {
    return(NULL)
  }
  # need to adjust the ranges for the _ASCII_ (!) delimiters
  r_ranges <- cpt[cpt$name == "code", ][wcnd2, range_cols]
  r_ranges$start_byte <- r_ranges$start_byte + pre_drop[wcnd2] + 2L # 'r '
  r_ranges$start_column <- r_ranges$start_column + pre_drop[wcnd2] + 2L
  r_ranges$end_byte <- r_ranges$end_byte - post_drop[wcnd2]
  scan_path_deps_do_r(code, path = path, ranges = r_ranges)
}

scan_path_deps_do_block_hits <- function(code, blk_hits, path) {
  wcnd <- which(blk_hits$name == "content")
  wcnd <- wcnd[grepl(re_r_dep(), blk_hits$code[wcnd])]
  if (length(wcnd) == 0) {
    return(NULL)
  }

  r_ranges <- blk_hits[wcnd, range_cols]
  scan_path_deps_do_r(code, path = path, ranges = r_ranges)
}

scan_path_deps_do_rmarkdown <- function(code, blk_hits, path) {
  blk_hits <- blk_hits[blk_hits$name == "language", ]
  rchk <- tolower(blk_hits$code) %in% c("r", "rscript")
  if (any(rchk)) {
    # only add the first chunk, adding all R chunks seems too much
    rchk <- which(rchk)[1]
    scan_deps_df(
      path = path,
      package = "rmarkdown",
      code = blk_hits$code[rchk],
      start_row = blk_hits$start_row[rchk],
      start_column = blk_hits$start_column[rchk],
      start_byte = blk_hits$start_byte[rchk]
    )
  }
}

# Crossref: https://github.com/r-lib/pkgdepends/issues/399
# This is pretty difficult, unfortunately, but could not come up with a
# simpler solution.
# * We use tree-sitter to parse and search the YAML, so that we can have
#   coordinates, and also the search is much simpler than when using a YAML
#   parser.
# * The tree-sitter parser cannot scan the actual values of the scalars, so
#   we use a YAML parser for that (libyaml). (No, it is not better to scan
#   them manually, they are quite involved.)
# * Scanning the scalars is a transformation, not just a subsetting, so
#   we lose the correct coordinates for the things (e.g. R code) _within_
#   the values. We still have the coordinates for the values, though.
# * We don't handle references correctly, because the tree-sitter parser
#   does not help with that. For that we'd need to parse the whole YAML
#   with libyaml. Maybe we'll do that in the future.

scan_path_deps_do_header_hits <- function(code, hdr_hits, path) {
  hits <- code_query(
    code,
    language = "yaml",
    query = q_deps_yaml_header(),
    ranges = hdr_hits[, range_cols]
  )

  shiny_pat <- hits$patterns$id[hits$patterns$name == "shiny"]
  shiny_hits <- hits$matched_captures[
    hits$matched_captures$pattern %in% shiny_pat,
  ]
  pkgstr_pat <- hits$patterns$id[hits$patterns$name == "pkgstring"]
  pkgstr_hits <- hits$matched_captures[
    hits$matched_captures$pattern %in% pkgstr_pat,
  ]
  bslib_pat <- hits$patterns$id[hits$patterns$name == "bslib"]
  bslib_hits <- hits$matched_captures[
    hits$matched_captures$pattern %in% bslib_pat,
  ]
  tag_pat <- hits$patterns$id[hits$patterns$name == "tag"]
  tag_hits <- hits$matched_captures[
    hits$matched_captures$pattern %in% tag_pat,
  ]

  rbind(
    if (nrow(shiny_hits)) {
      scan_path_deps_do_header_shiny_hits(code, shiny_hits, path)
    },
    if (nrow(pkgstr_hits)) {
      scan_path_deps_do_header_pkgstr_hits(code, pkgstr_hits, path)
    },
    if (nrow(bslib_hits)) {
      scan_path_deps_do_header_bslib_hits(code, bslib_hits, path)
    },
    if (nrow(tag_hits)) {
      scan_path_deps_do_header_tag_hits(code, tag_hits, path)
    }
  )
}

scan_path_deps_do_header_shiny_hits <- function(code, hits, path) {
  hits <- hits[hits$name == "value", ]
  vals <- yaml_parse_scalar(hits$code)
  shiny <- vals == "shiny"
  scan_deps_df(
    path = path,
    package = "shiny",
    code = hits$code[shiny],
    start_row = hits$start_row[shiny],
    start_column = hits$start_column[shiny],
    start_byte = hits$start_byte[shiny]
  )
}

scan_path_deps_do_header_pkgstr_hits <- function(code, hits, path) {
  vals <- yaml_parse_scalar(hits$code)
  pkg <- vapply(vals, FUN.VALUE = character(1), function(x) {
    tryCatch(
      {
        expr <- parse(text = x, keep.source = FALSE)[[1]]
        if (
          length(expr) == 3 &&
            is.call(expr) &&
            (identical(expr[[1]], quote(`::`)) ||
              identical(expr[[1]], quote(`:::`)))
        ) {
          as.character(expr[[2]])
        } else {
          NA_character_
        }
      },
      error = function(...) NA_character_
    )
  })
  if (all(is.na(pkg))) {
    return(NULL)
  }
  hits <- hits[!is.na(pkg), ]
  pkg <- na.omit(pkg)
  scan_deps_df(
    path = path,
    package = pkg,
    code = hits$code,
    start_row = hits$start_row,
    start_column = hits$start_column,
    start_byte = hits$start_byte
  )
}

scan_path_deps_do_header_bslib_hits <- function(code, hits, path) {
  scan_deps_df(
    path = path,
    package = "bslib",
    code = hits$code[hits$name == "code"],
    start_row = hits$start_row[hits$name == "code"],
    start_column = hits$start_column[hits$name == "code"],
    start_byte = hits$start_byte[hits$name == "code"]
  )
}

scan_path_deps_do_header_tag_hits <- function(code, hits, path) {
  hits <- hits[hits[["name"]] == "code", ]
  vals <- yaml_parse_scalar(hits$code)
  res <- lapply(seq_along(vals), function(vi) {
    r1 <- scan_path_deps_do_r(vals[vi], path = path)
    # need to replace the positions with the ones from the YAML file
    # we cannot use ranges here, because the R code is a a transformation
    # of the text in the YAML file, not merely a subset. So we just mark
    # the beginning of the R code in the YAML, instead of the real position
    # of the `library()` etc. calls.
    r1[["start_row"]][] <- hits[vi, "start_row"]
    r1[["start_column"]][] <- hits[vi, "start_column"]
    r1[["start_byte"]][] <- hits[vi, "start_byte"]
    r1
  })
  do.call(rbind, res)
}

yaml_parse_scalar <- function(x) {
  vcapply(x, function(x) .Call(c_yaml_parse_scalar, x), USE.NAMES = FALSE)
}

# -------------------------------------------------------------------------

scan_path_deps_do_dsc <- function(code, path) {
  if (is.raw(code)) {
    code <- rawToChar(code)
  }
  dsc <- desc::desc(text = code)
  deps <- resolve_ref_deps(
    dsc$get_deps(),
    dsc$get("Remotes")[[1]],
    dsc$get(extra_config_fields(dsc$fields()))
  )
  deps <- deps[deps$package != "R", ]
  version <- ifelse(deps$op == "", "*", paste0(deps$op, deps$version))
  scan_deps_df(
    path = path,
    ref = deps$ref,
    package = deps$package,
    version = version,
    type = get_dep_type_from_description_field(deps$type),
    code = deps$ref
  )
}

# -------------------------------------------------------------------------

scan_path_deps_do_namespace <- function(code, path) {
  tmp <- tempfile()
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  mkdirp(file.path(tmp, "pkg"))
  if (is.raw(code)) {
    writeBin(code, file.path(tmp, "pkg", "NAMESPACE"))
  } else {
    writeLines(code, file.path(tmp, "pkg", "NAMESPACE"))
  }
  info <- parseNamespaceFile(package = "pkg", package.lib = tmp)
  pkg <- unique(vcapply(info$imports, "[[", 1))
  scan_deps_df(
    path = path,
    package = pkg,
    type = "prod",
    code = pkg
  )
}

# -------------------------------------------------------------------------

scan_path_deps_do_bookdown <- function(code, path) {
  scan_deps_df(
    path = path,
    package = "bookdown",
    code = NA_character_
  )
}

scan_path_deps_do_pkgdown <- function(code, path) {
  scan_deps_df(
    path = path,
    package = "pkgdown",
    code = NA_character_
  )
}

scan_path_deps_do_quarto <- function(code, path) {
  # renv does not include anything for quarto
  # Do we want a 'dev' dependency for the quarto package?
  # Maybe that's too opinionated?
}

scan_path_deps_do_renv_lock <- function(code, path) {
  scan_deps_df(
    path = path,
    package = "renv",
    code = NA_character_
  )
}

scan_path_deps_do_rsconnect <- function(code, path) {
  scan_deps_df(
    path = path,
    package = "rsconnect",
    code = NA_character_
  )
}

# -------------------------------------------------------------------------

scan_path_deps_do_rproj <- function(code, path) {
  con <- if (is.raw(code)) rawConnection(code) else textConnection(code)
  on.exit(close(con), add = TRUE)
  if ("yes" %in% tolower(read.dcf(con, fields = "PackageUseDevtools"))) {
    scan_deps_df(
      path = path,
      package = c("devtools", "roxygen2"),
      type = "dev",
      code = "PackageUseDevtools: Yes"
    )
  }
}

# -------------------------------------------------------------------------

# knitr::all_patterns$rnw
re_rnw <- list(
  chunk.begin = "^\\s*<<(.*)>>=.*$",
  chunk.end = "^\\s*@\\s*(%+.*|)$",
  inline.code = "\\\\Sexpr\\{([^}]+)\\}",
  inline.comment = "^\\s*%.*",
  ref.chunk = "^\\s*<<(.+)>>\\s*$",
  header.begin = "(^|\n)\\s*\\\\documentclass[^}]+\\}",
  document.begin = "\\s*\\\\begin\\{document\\}"
)

scan_path_deps_do_rnw <- function(code, path) {
  if (is.raw(code)) {
    code <- rawToChar(code)
    Encoding(code) <- "UTF-8"
  }
  code <- unlist(strsplit(code, "\n", fixed = TRUE))
  chunks <- scan_path_deps_do_rnw_chunks(code)

  do.call(
    "rbind",
    lapply(chunks, function(c) scan_path_deps_do_r(c$code, path))
  )
}

# along renv_dependencies_discover_chunks_ignore
scan_path_deps_rnw_chunk_is_ignored <- function(chunk) {
  # renv.ignore = TRUE
  if (is_truthy(chunk$params[["renv.ignore"]])) {
    return(TRUE)
  }

  # engine is not R / Rscript
  engine <- chunk$params[["engine"]] %||% "R"
  if (!is.character(engine) || !tolower(engine) %in% c("r", "rscript")) {
    return(TRUE)
  }

  # eval = FALSE
  eval <- chunk$params[["eval"]]
  if (!is.null(eval) && !is_truthy(eval)) {
    return(TRUE)
  }

  # skip learnr exercises
  if (is_truthy(chunk$params[["exercise"]])) {
    return(TRUE)
  }

  # skip chunks whose labels end in '-display'
  if (endsWith(chunk$params[["label"]] %||% "", "-display")) {
    return(TRUE)
  }

  FALSE
}

scan_path_deps_do_rnw_del_placeholders <- function(chunk) {
  mch <- gregexpr("<<[^>]+>>", chunk$code)
  repl <- lapply(
    mch,
    function(x) strrep(" ", pmax(0, attr(x, "match.length")))
  )
  regmatches(chunk$code, mch) <- repl
  chunk
}

scan_path_deps_do_rnw_chunks <- function(code) {
  ranges <- scan_path_deps_do_rnw_ranges(code)
  from <- viapply(ranges, "[[", 1L)
  to <- viapply(ranges, "[[", 2L)
  chunks <- .mapply(
    function(from, to) {
      cheader <- code[from]
      ccode <- code[(from + 1):to]
      if (ccode[[length(ccode)]] == "@") {
        ccode <- ccode[-length(ccode)]
      }
      scan_path_deps_do_rnw_parse_chunk(cheader, ccode)
    },
    list(from, to),
    NULL
  )

  # some chunks are ignored
  ignored <- vlapply(chunks, scan_path_deps_rnw_chunk_is_ignored)
  chunks <- chunks[!ignored]

  # remove reused chunk placeholders
  chunks <- lapply(chunks, scan_path_deps_do_rnw_del_placeholders)

  chunks
}

# along xfun::csv_options
scan_path_deps_do_rnw_parse_chunk_header <- function(header) {
  opts <- sub(re_rnw$chunk.begin, "\\1", header)
  # Note: this is not a "real" eval, because we are in an 'alist'
  tryCatch(
    res <- eval(parse(
      text = paste("alist(", xfun_quote_label(opts), ")"),
      keep.source = FALSE
    )),
    error = function(e) {
      stop(
        "Invalid syntax for chunk options: ",
        opts,
        "\n",
        conditionMessage(e)
      )
    }
  )

  idx <- which(names(res) == "")
  j <- NULL
  for (i in idx) {
    if (identical(res[[i]], alist(,)[[1]])) {
      j <- c(j, i)
    }
  }
  if (length(j)) {
    res[j] <- NULL
  }
  idx <- if (is.null(names(res)) && length(res) == 1L) {
    1L
  } else {
    which(names(res) == "")
  }
  if ((n <- length(idx)) > 1L || (length(res) > 1L && is.null(names(res)))) {
    stop(
      "Invalid chunk options: ",
      res,
      "\n\nAll options must be of the form 'tag=value' except for the chunk label."
    )
  }
  if (is.null(res$label)) {
    if (n == 0L) {
      res$label <- ""
    } else {
      names(res)[idx] <- "label"
    }
  }
  if (!is.character(res$label)) {
    res$label <- gsub(" ", "", as.character(as.expression(res$label)))
  }
  if (res$label == "") {
    res$label <- NULL
  }
  res
}

xfun_quote_label <- function(x) {
  x <- gsub("^\\s*,?", "", x)
  if (grepl("^\\s*[^'\"](,|\\s*$)", x)) {
    x <- gsub("^\\s*([^'\"])(,|\\s*$)", "'\\1'\\2", x)
  } else if (grepl("^\\s*[^'\"](,|[^=]*(,|\\s*$))", x)) {
    x <- gsub("^\\s*([^'\"][^=]*)(,|\\s*$)", "'\\1'\\2", x)
  }
  x
}

scan_path_deps_do_rnw_parse_chunk <- function(header, code) {
  params <- tryCatch(
    scan_path_deps_do_rnw_parse_chunk_header(header),
    error = function(...) list(a = 1)[0]
  )
  list(params = params, code = code)
}

scan_path_deps_do_rnw_ranges <- function(code) {
  beg <- grep(re_rnw$chunk.begin, code)
  end <- c(grep(re_rnw$chunk.end, code), length(code) + 1L)
  lapply(seq_along(beg), function(bx) {
    # for every start we find its end. The end is the next end marker,
    # except when there is another begin marker before
    bmx <- beg[bx]
    emx <- end[end > bmx][1]
    if (bx < length(beg) && beg[bx + 1L] < emx) {
      emx <- beg[bx + 1L] - 1L
    }
    c(bmx, emx)
  })
}

# -------------------------------------------------------------------------

scan_path_deps_do_ipynb <- function(code, path) {
  ipynb <- jsonlite::fromJSON(code, simplifyVector = FALSE)
  if (!identical(ipynb$metadata$kernelspec$language, "R")) {
    return(NULL)
  }
  ir <- if (identical(ipynb$metadata$kernelspec$name, "ir")) {
    scan_deps_df(
      path = path,
      package = "IRkernel",
      code = NA_character_
    )
  }

  deps <- lapply(ipynb$cells, function(cell) {
    if (identical(cell$cell_type, "code")) {
      c1 <- paste(unlist(cell$source), collapse = "")
      scan_path_deps_do_r(c1, path)
    }
  })

  adeps <- drop_nulls(c(list(ir), deps))

  do.call("rbind", adeps)
}

test_that("scan_deps", {
  local_reproducible_output(width = 500)
  withr::local_envvar(R_PKG_CACHE_DIR = tmp <- tempfile())
  on.exit(unlink(tmp), add = TRUE)

  project <- test_path("fixtures/scan/project-1")
  expect_snapshot(variant = .Platform$OS.type, {
    scan_deps(project, root = project)[]
  })
  expect_snapshot(variant = .Platform$OS.type, {
    scan_deps(project, root = project)
  })
})

test_that("scan_deps errors", {
  mkdirp(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  withr::local_dir(tmp)
  writeLines(c("Package: foo", "Version: 1.0.0"), "DESCRIPTION")
  mkdirp(c("foo", "bar", "foobar"))
  expect_snapshot(error = TRUE, {
    # invalid types
    scan_deps(1:10)
    scan_deps(root = mtcars)
    # root does not exist
    scan_deps(root = "sdfssdfsdf")
    # path does not exist
    scan_deps("sdfssdfsdf")
    scan_deps(c("sdfssdfsdf-1", "sdfssdfsdf-2"))
    # path are not inside root
    scan_deps("foo", "bar")
    scan_deps(c("foo", "foobar"), "bar")
  })
})

test_that("get_deps_cache_path", {
  local_reproducible_output(width = 500)
  withr::local_envvar(R_PKG_CACHE_DIR = tmp <- tempfile())
  on.exit(unlink(tmp), add = TRUE)

  expect_snapshot(transform = transform_tempdir, {
    writeLines(get_deps_cache_path())
    writeLines(get_deps_cache_path("badcafe"))
  })
})

test_that("clear_deps_cache", {
  local_reproducible_output(width = 500)
  withr::local_envvar(R_PKG_CACHE_DIR = tmp <- tempfile())
  on.exit(unlink(tmp), add = TRUE)

  cval <- get_deps_cache_path("badcafe")
  mkdirp(dirname(cval))
  file.create(cval)

  expect_snapshot({
    dir(tmp, recursive = TRUE)
  })

  clear_deps_cache()
  expect_snapshot({
    dir(tmp, recursive = TRUE)
  })
})

test_that("re_r_dep", {
  expect_snapshot({
    re_r_dep()
  })
})

test_that("scan_path_deps", {
  local_reproducible_output(width = 500)
  withr::local_envvar(R_PKG_CACHE_DIR = tmp <- tempfile())
  on.exit(unlink(tmp), add = TRUE)

  rfile <- test_path("fixtures/scan/project-1/R/code.R")
  expect_snapshot(variant = .Platform$OS.type, {
    scan_path_deps(rfile)
  })

  # now from the cache
  fake(scan_path_deps, "re_r_dep", function(...) stop("no"))
  expect_snapshot(variant = .Platform$OS.type, {
    scan_path_deps(rfile)
  })
})

test_that("scan_deps_df", {
  expect_snapshot({
    scan_deps_df()
  })
})

test_that("scan_path_deps_do", {
  rfile <- test_path("fixtures/scan/project-1/R/code.R")
  nsfile <- test_path("fixtures/scan/NAMESPACE")
  expect_snapshot({
    scan_path_deps_do(readLines(rfile), basename(rfile))
    scan_path_deps_do(readLines(nsfile), basename(nsfile))
  })

  expect_snapshot(error = TRUE, {
    scan_path_deps_do("code", "foo.unknown")
  })
})

test_that("scan_path_deps_do_r", {
  local_reproducible_output(width = 500)
  rfile <- test_path("fixtures/scan/project-1/R/code.R")
  expect_snapshot({
    scan_path_deps_do_r(readLines(rfile), rfile)
  })
})

# test_that("scan_path_deps_do_pkg_hits", { })

test_that("scan_path_deps_do_fn_hits", {
  local_reproducible_output(width = 500)
  rfile <- test_path("fixtures/scan/methods.R")
  expect_snapshot({
    scan_path_deps_do_r(readLines(rfile), rfile)
  })
})

# test_that("scan_path_deps_do_gen_hits", { })

test_that("scan_path_deps_do_jr_hits", {
  local_reproducible_output(width = 500)
  rfile <- test_path("fixtures/scan/junit.R")
  expect_snapshot({
    scan_path_deps_do_r(readLines(rfile), rfile)
  })
})

test_that("scan_pat_deps_do_ragg_hits", {
  local_reproducible_output(width = 500)
  rfile <- test_path("fixtures/scan/knitr.Rmd")
  expect_snapshot({
    scan_path_deps_do_rmd(readLines(rfile), rfile)
  })
  rfile <- test_path("fixtures/scan/noragg.Rmd")
  expect_snapshot({
    scan_path_deps_do_rmd(readLines(rfile), rfile)
  })
})

# test_that("scan_pat_deps_do_db_hits", { })

test_that("safe_parse_pkg_from_call", {
  # error
  expect_null(
    safe_parse_pkg_from_call(NA_character_, "library", "library(error")
  )
  expect_snapshot(
    safe_parse_pkg_from_call(NA_character_, "library", "library(qwe)")
  )
})

# test_that("parse_pkg_from_call_match", { })

test_that("parse_pkg_from_call", {
  expect_snapshot({
    parse_pkg_from_call(NA_character_, "library", "library(qwe)")
    parse_pkg_from_call("base", "loadNamespace", "loadNamespace('q1')")
    parse_pkg_from_call(
      "base",
      "requireNamespace",
      "requireNamespace('q1')"
    )
    parse_pkg_from_call(
      NA_character_,
      "pkg_attach",
      "pkg_attach('foobar')"
    )
    parse_pkg_from_call(
      NA_character_,
      "pkg_attach2",
      "pkg_attach2('foobar')"
    )
    parse_pkg_from_call("pacman", "p_load", "p_load('p1')")
    parse_pkg_from_call(NA_character_, "import", "import(x1)")
    parse_pkg_from_call(NA_character_, "module", "module({import('x2')})")
    parse_pkg_from_call("import", "from", "import::from(dplyr)")
    parse_pkg_from_call(
      "import",
      "into",
      "import::into('operators', .from = 'dplyr')"
    )
    parse_pkg_from_call("import", "here", "import::here('dplyr')")
    parse_pkg_from_call("box", "use", "box::use(dplyr[filter, select])")
    parse_pkg_from_call(
      "targets",
      "tar_option_set",
      "tar_option_set(packages = c('p1', 'p2'))"
    )
    parse_pkg_from_call(
      "glue",
      "glue",
      "glue::glue('blah {library(x5)} blah')"
    )
    parse_pkg_from_call(
      NA_character_,
      "ggsave",
      "ggsave(filename = 'foo.svg')"
    )
    parse_pkg_from_call(
      NA_character_,
      "set_engine",
      "set_engine(engine = 'spark')"
    )
    parse_pkg_from_call(
      "R6",
      "R6Class",
      "R6::R6Class('foobar', inherit = JunitReporter)"
    )
    parse_pkg_from_call(
      "testthat",
      "test_dir",
      "testthat::test_dir(reporter = 'junit')"
    )
  })
})

test_that("parse_pkg_from_call_library", {
  ppcl <- function(fn, code, ns = NA_character_) {
    matched <- parse_pkg_from_call_match(fn, code)
    parse_pkg_from_call_library(ns, fn, matched)
  }
  expect_null(
    ppcl("library", "library(qqq)", ns = "other")
  )
  expect_null(
    ppcl("library", "library(qqq, character.only = TRUE)")
  )
  expect_null(
    ppcl("require", "require(qqq)", ns = "other")
  )
  expect_null(
    ppcl("require", "require(qqq, character.only = TRUE)")
  )
  expect_snapshot({
    ppcl("library", "library(qqq)")
    ppcl("library", "library('qqq')")
    ppcl("library", "library(qqq)", ns = "base")
    ppcl("require", "require(qqq)")
    ppcl("require", "require('qqq')")
    ppcl("require", "require('qqq')", ns = "base")
  })
})

test_that("parse_pkg_from_call_loadnamespace", {
  ppcln <- function(fn, code, ns = NA_character_) {
    matched <- parse_pkg_from_call_match(fn, code)
    parse_pkg_from_call_loadnamespace(ns, fn, matched)
  }
  expect_null(
    ppcln("loadNamespace", "loadNamespace('www')", ns = "other")
  )
  expect_null(
    ppcln("loadNamespace", "loadNamespace(www)")
  )
  expect_null(
    ppcln("loadNamespace", "loadNamespace(c('one', 'two'))")
  )
  expect_null(
    ppcln("loadNamespace", "loadNamespace(123)")
  )
  expect_equal(
    ppcln("loadNamespace", "loadNamespace('eee')"),
    "eee"
  )
  expect_equal(
    ppcln("loadNamespace", "loadNamespace('eee')", ns = "base"),
    "eee"
  )

  expect_null(
    ppcln("requireNamespace", "requireNamespace('www')", ns = "other")
  )
  expect_null(
    ppcln("requireNamespace", "requireNamespace(www)")
  )
  expect_null(
    ppcln("requireNamespace", "requireNamespace(c('one', 'two'))")
  )
  expect_null(
    ppcln("requireNamespace", "requireNamespace(123)")
  )
  expect_equal(
    ppcln("requireNamespace", "requireNamespace('eee')"),
    "eee"
  )
  expect_equal(
    ppcln("requireNamespace", "requireNamespace('eee')", ns = "base"),
    "eee"
  )
})

test_that("parse_pkg_from_call_xfun", {
  ppcx <- function(fn, code, ns = NA_character_) {
    matched <- parse_pkg_from_call_match(fn, code)
    parse_pkg_from_call_xfun(ns, fn, matched)
  }
  expect_null(
    ppcx("pkg_attach", "pkg_attach('qwe')", ns = "nope")
  )
  expect_null(
    ppcx("pkg_attach", "pkg_attach()")
  )
  expect_equal(
    ppcx("pkg_attach", "pkg_attach('p1', 'p2', 'p3')"),
    c("p1", "p2", "p3")
  )
  expect_equal(
    ppcx("pkg_attach2", "pkg_attach2('p1', 'p2', 'p3')"),
    c("p1", "p2", "p3")
  )
})

test_that("parse_pkg_from_call_pacman", {
  ppcp <- function(code, ns = NA_character_) {
    matched <- parse_pkg_from_call_match("p_load", code)
    parse_pkg_from_call_pacman(ns, fn, matched)
  }
  expect_null(ppcp("p_load()", ns = "foo"))
  expect_null(ppcp("p_load(xx, character.only = TRUE)"))
  expect_equal(ppcp("p_load(p1, 'p2', p3)"), c("p1", "p2", "p3"))
  expect_equal(ppcp("p_load(char = 'pp')"), 'pp')
  expect_equal(ppcp("p_load(char = c('p1', 'p2'))"), c("p1", "p2"))
})

test_that("parse_pkg_from_call_modules_import", {
  ppcmi <- function(code, ns = NA_character_) {
    matched <- parse_pkg_from_call_match("import", code)
    parse_pkg_from_call_modules_import(ns, fn, matched)
  }
  expect_null(ppcmi("import('pp')", ns = "foo"))
  expect_null(ppcmi("import(NULL)"))
  expect_equal(ppcmi("import('pp')"), 'pp')
  expect_equal(ppcmi("import(pp)"), 'pp')
})

test_that("parse_pkg_from_call_modules_module", {
  ppcmm <- function(code, ns = NA_character_) {
    matched <- parse_pkg_from_call_match("module", code)
    parse_pkg_from_call_modules_module(ns, fn, matched)
  }
  expect_null(ppcmm("module(x)", ns = "foo"))
  expect_null(ppcmm('module({})'))
  expect_equal(
    ppcmm(
      "module({
      # other expressions, mixed with import()
      pkg::fun()
      blah + blah
      import(p1)
      baaaaah
      import('p2')
    })"
    ),
    c('p1', 'p2')
  )
})

test_that("parse_pkg_from_call_import", {
  ppci <- function(fn, code, ns = NA_character_) {
    matched <- parse_pkg_from_call_match(fn, code)
    parse_pkg_from_call_import(ns, fn, matched)
  }
  expect_null(ppci("from", "import::from(foo)", ns = "xx"))
  expect_equal(ppci("from", "import::from(foo)"), "foo")
  expect_equal(ppci("from", "import::from('foo')"), "foo")
  expect_equal(ppci("here", "import::here(foo)"), "foo")
  expect_equal(ppci("here", "import::here('foo')"), "foo")
  expect_equal(ppci("into", "import::into(.from = foo)"), "foo")
  expect_equal(ppci("into", "import::into(.from = 'foo')"), "foo")
  expect_null(ppci("from", "import::from(xx, .character_only = TRUE)"))
  expect_null(ppci("from", "import::from('./path.R')"))
})

test_that("parse_pkg_from_call_box", {
  ppcb <- function(code, ns = NA_character_) {
    matched <- parse_pkg_from_call_match("use", code)
    parse_pkg_from_call_box(ns, fn, matched)
  }
  expect_null(ppcb("box::use(pkg)", ns = 'not'))
  expect_null(ppcb("box::use(foo/bar)"))
  expect_null(ppcb("box::use(.[ff])"))
  expect_null(ppcb("box::use(..[ff])"))
  expect_equal(ppcb("box::use(pkg)"), "pkg")
  expect_equal(ppcb("box::use(pkg[f1, f2])"), c("pkg"))
  expect_equal(ppcb("box::use(pkg0, pkg[f1, f2])"), c("pkg0", "pkg"))
})

test_that("parse_pkg_from_call_targets", {
  ppct <- function(code, ns = NA_character_) {
    matched <- parse_pkg_from_call_match("tar_option_set", code)
    parse_pkg_from_call_targets(ns, fn, matched)
  }
  expect_null(ppct("tar_option_set(packages = 'pp')", ns = 'not'))
  expect_null(ppct("tar_option_set()"))
  expect_equal(
    ppct("tar_option_set(packages = { 1:10; c('p1', 'p2') })"),
    c("p1", "p2")
  )
})

test_that("dependencies_eval", {
  expect_snapshot({
    dependencies_eval(quote({
      1:10
      c(10:1)[1:3]
    }))
  })
})

test_that("parse_pkg_from_call_glue", {
  ppcg <- function(code, ns = NA_character_) {
    matched <- parse_pkg_from_call_match("glue", code)
    parse_pkg_from_call_glue(ns, fn, matched)
  }
  expect_null(ppcg("glue('{library(xx)}')", ns = "nope"))
  expect_null(ppcg("glue('no code at all')"))
  expect_equal(
    ppcg("glue('some {library(p1)} code {p2::f()}')"),
    c("p1", "p2")
  )
})

test_that("parse_pkg_from_call_ggplot2", {
  ppcgg <- function(code, ns = NA_character_) {
    matched <- parse_pkg_from_call_match("ggsave", code)
    parse_pkg_from_call_ggplot2(ns, fn, matched)
  }
  expect_null(ppcgg("ggsave(filename = 'foo.svg')", ns = 'not'))
  expect_null(ppcgg("ggsave(filename = 'foo.png')"))
  expect_null(ppcgg("ggsave(filename = var)"))
  expect_equal(ppcgg("ggsave(filename = 'foo.svg')"), "svglite")
})

test_that("parse_pkg_from_call_parsnip", {
  ppcp <- function(code, ns = NA_character_) {
    matched <- parse_pkg_from_call_match("set_engine", code)
    parse_pkg_from_call_parsnip(ns, fn, matched)
  }
  expect_null(ppcp("set_engine(engine = 'keras')", ns = "nope"))
  expect_null(ppcp("set_engine()"))
  withr::local_options(renv.parsnip.engines = NULL)
  expect_equal(ppcp("set_engine(engine = 'glm')"), "stats")
  withr::local_options(renv.parsnip.engines = list(foo = "bar"))
  expect_equal(ppcp("set_engine(engine = 'foo')"), "bar")
  withr::local_options(renv.parsnip.engines = function(x) "eng")
  expect_equal(ppcp("set_engine(engine = 'foo')"), "eng")
  withr::local_options(renv.parsnip.engines = function(x) NULL)
  expect_null(ppcp("set_engine(engine = 'foo')"))
})

test_that("parse_pkg_from_call_testthat_r6class", {
  ppcttr6 <- function(code, ns = NA_character_) {
    matched <- parse_pkg_from_call_match("R6Class", code)
    parse_pkg_from_call_testthat_r6class(ns, fn, matched)
  }
  expect_null(ppcttr6("R6Class(inherit = JunitReporter)", ns = 'not'))
  expect_null(ppcttr6("R6Class(inherit = someother)"))
  expect_equal(ppcttr6("R6Class(inherit = JunitReporter)"), "xml2")
  expect_equal(
    ppcttr6("R6Class(inherit = testthat::JunitReporter)"),
    "xml2"
  )
})

test_that("parse_pkg_from_call_testthat_test", {
  ppcttt <- function(fn, code, ns = NA_character_) {
    matched <- parse_pkg_from_call_match(fn, code)
    parse_pkg_from_call_testthat_test(ns, fn, matched)
  }
  expect_null(
    ppcttt("test_dir", "test_dir(reporter = 'junit')", ns = "other")
  )
  expect_null(
    ppcttt("test_dir", "test_dir(reporter = 'other')")
  )
  expect_equal(
    ppcttt("test_dir", "test_dir(reporter = 'junit')"),
    "xml2"
  )
})

test_that("scan_path_deps_do_rmd", {
  local_reproducible_output(width = 500)
  path <- test_path("fixtures/scan/chunk-errors.Rmd")
  expect_snapshot({
    scan_path_deps_do_rmd(readLines(path), "chunk-errors.Rmd")
  })
})

test_that("scan_path_deps_do_rmd #2", {
  local_reproducible_output(width = 500)
  path <- test_path("fixtures/scan/inline-chunks.Rmd")
  expect_snapshot({
    scan_path_deps_do_rmd(readLines(path), "inline-chunks.Rmd")
  })
})

test_that("scan_path_deps_do_rmd #3", {
  local_reproducible_output(width = 500)
  path <- test_path("fixtures/scan/nothing.Rmd")
  expect_snapshot({
    scan_path_deps_do_rmd(readLines(path), "nothing.Rmd")
  })
})

# test_that("scan_path_deps_do_inline_hits", { })
# test_that("scan_path_deps_do_block_hits", { })

test_that("scan_path_deps_do_header_hits", {
  local_reproducible_output(width = 500)
  path <- test_path("fixtures/scan/header.Rmd")
  expect_snapshot({
    scan_path_deps_do_rmd(readLines(path), basename(path))
  })
})

test_that("scan_path_deps_do_header_shiny_hits", {
  local_reproducible_output(width = 500)
  path <- test_path("fixtures/scan/header-shiny.Rmd")
  expect_snapshot({
    scan_path_deps_do_rmd(readLines(path), basename(path))
  })
  path <- test_path("fixtures/scan/header-shiny2.Rmd")
  expect_snapshot({
    scan_path_deps_do_rmd(readLines(path), basename(path))
  })
})

test_that("scan_path_deps_do_header_bslib_hits", {
  local_reproducible_output(width = 500)
  path <- test_path("fixtures/scan/header-bslib.Rmd")
  expect_snapshot({
    scan_path_deps_do_rmd(readLines(path), basename(path))
  })
})

test_that("scan_path_deps_do_dsc", {
  local_reproducible_output(width = 500)
  path <- test_path("fixtures/scan/DESCRIPTION")
  expect_snapshot({
    print(scan_path_deps_do_dsc(readLines(path), basename(path)), n = Inf)
  })
})

test_that("scan_path_deps_do_namespace", {
  local_reproducible_output(width = 500)
  path <- test_path("fixtures/scan/NAMESPACE")
  expect_snapshot({
    print(
      scan_path_deps_do_namespace(
        readBin(path, "raw", 10000),
        path
      ),
      n = Inf
    )
  })
})

test_that("scan_path_deps_do_{bookdown,pkgdown,quarto,renv_lock,rsconnect}", {
  local_reproducible_output(width = 500)
  withr::local_dir(test_path("fixtures/scan/project-2"))
  expect_snapshot({
    scan_deps()[]
  })
})

test_that(".Rproj file", {
  local_reproducible_output(width = 500)
  project <- test_path("fixtures/scan/project-3")
  expect_snapshot({
    scan_deps(project, root = project)[]
  })
})

test_that("scan_path_deps_do_rnw_ranges", {
  path <- test_path("fixtures/scan/test.Rnw")
  code <- readLines(path)
  expect_snapshot({
    scan_path_deps_do_rnw_ranges(code)
  })
})

test_that("scan_path_deps_do_rnw_parse_chunk_header", {
  expect_snapshot({
    scan_path_deps_do_rnw_parse_chunk_header("")
    scan_path_deps_do_rnw_parse_chunk_header("name")
    scan_path_deps_do_rnw_parse_chunk_header(
      "name, foo = 1, bar = TRUE, this = that"
    )
  })
})

test_that(".Rnw file", {
  local_reproducible_output(width = 500)
  path <- test_path("fixtures/scan/test.Rnw")
  expect_snapshot({
    scan_path_deps_do_rnw(
      readLines(path),
      basename(path)
    )
    scan_path_deps_do(
      readLines(path),
      basename(path)
    )
  })
})

test_that("Ignored chunks in .Rnw file", {
  local_reproducible_output(width = 500)
  path <- test_path("fixtures/scan/ignore-test.Rnw")
  expect_snapshot({
    scan_path_deps_do(
      readBin(path, "raw", file.size(path)),
      basename(path)
    )
  })
})

test_that("IPython notebook", {
  local_reproducible_output(width = 500)
  path <- test_path("fixtures/scan/notebook.ipynb")
  expect_snapshot({
    scan_path_deps_do(readLines(path), basename(path))
  })
})

## ------------------------------------------------------------------------
## API

parse_remote_bioc <- function(specs, config, ...) {
  ## BioC is the same as CRAN, except for cran:: -> bioc::
  parsed_specs <- re_match(specs, standard_rx("bioc"))
  parsed_specs$ref <- parsed_specs$.text
  cn <- setdiff(colnames(parsed_specs), c(".match", ".text"))
  parsed_specs <- parsed_specs[, cn]
  eq <- parsed_specs$atleast == "" & parsed_specs$version != ""
  parsed_specs$atleast[eq] <- "=="
  parsed_specs$type <- "bioc"
  lapply(
    seq_len(nrow(parsed_specs)),
    function(i) as.list(parsed_specs[i, ])
  )
}

resolve_remote_bioc <- function(
  remote,
  direct,
  config,
  cache,
  dependencies,
  progress_bar,
  ...
) {
  force(remote)
  force(direct)
  force(dependencies)
  versions <- if ("type" %in% names(remote)) {
    remote$version
  } else {
    vcapply(remote, "[[", "version") # nocov
  }

  if (all(versions %in% c("", "current"))) {
    resolve_from_metadata(remote, direct, config, cache, dependencies)
  } else {
    type_cran_resolve_version(remote, direct, config, cache, dependencies) # nocov
  }
}

download_remote_bioc <- function(
  resolution,
  target,
  target_tree,
  config,
  cache,
  which,
  on_progress
) {
  # we need to check if there is a new binary build. for source packages
  # we can always use the cache
  download_ping_if_not_source(resolution, target, config, cache, on_progress)
}

satisfy_remote_bioc <- function(resolution, candidate, config, ...) {
  ## 1. candidate must be a bioc, standard or installed ref
  if (!candidate$type %in% c("bioc", "standard", "installed")) {
    return(structure(
      FALSE,
      reason = "Type must be 'bioc', 'standard' or 'installed'"
    ))
  }

  ## 2. installed refs must be from bioc
  if (candidate$type == "installed") {
    repotype <- candidate$extra[[1]]$repotype
    if (!identical(repotype, "bioc")) {
      return(structure(FALSE, reason = "Installed package not from BioC"))
    }
  }

  ## 3. package names must match
  if (resolution$package != candidate$package) {
    return(structure(FALSE, reason = "Package names differ"))
  }

  ## 4. installed package must not be older for direct refs
  if (resolution$direct) {
    if (
      candidate$type == "installed" &&
        package_version(resolution$version) > candidate$version
    ) {
      return(structure(FALSE, reason = "Direct ref needs update"))
    }
  }

  ## 5. version requirements must be satisfied. Otherwise good.
  if (resolution$remote[[1]]$version == "") {
    return(TRUE)
  }

  if (
    !version_satisfies(
      candidate$version,
      resolution$remote[[1]]$atleast,
      resolution$remote[[1]]$version
    )
  ) {
    return(structure(FALSE, reason = "Insufficient version"))
  }

  TRUE
}

installedok_remote_bioc <- function(installed, solution, config, ...) {
  if (solution$platform != "source") {
    # Binary packages are simple. We need to match `$build` to make sure
    # that we are installing the same build.
    identical(installed$package, solution$package) &&
      identical(installed$version, solution$version) &&
      installedok_remote_standard_platform(installed, solution, config, ...) &&
      identical(installed[["built"]], solution[["built"]])
  } else {
    # Source packages are different, because only the installed one has a
    # 'Built' field.
    identical(installed$package, solution$package) &&
      identical(installed$version, solution$version) &&
      identical(installed[["repotype"]], "bioc")
  }
}

test_that("parse_remote", {
  pr <- parse_pkg_refs("bioc::Biobase")[[1]]
  expect_equal(pr$package, "Biobase")
  expect_equal(pr$atleast, "")
  expect_equal(pr$version, "")
  expect_equal(pr$ref, "bioc::Biobase")
  expect_equal(pr$type, "bioc")
  expect_true("remote_ref_bioc" %in% class(pr))
  expect_true("remote_ref" %in% class(pr))
})

test_that("resolve_remote", {
  setup_fake_apps()

  conf <- current_config()
  cache <- list(
    package = NULL,
    metadata = pkgcache::get_cranlike_metadata_cache()
  )

  res <- suppressMessages(synchronise(resolve_remote_bioc(
    parse_pkg_ref("bioc::Biobase"),
    TRUE,
    conf,
    cache,
    dependencies = FALSE
  )))

  res$md5sum <- "<md5sum>"

  # If this fails then we need to update bioc-config.yaml in pkgcache
  expect_snapshot(
    as.list(res),
    transform = function(x) transform_local_port(transform_bioc_version(x)),
    variant = if (getRversion() < "3.6.0") "old-r" else "new-r"
  )
})

test_that("failed resolution", {
  setup_fake_apps()

  conf <- current_config()
  cache <- list(
    package = NULL,
    metadata = pkgcache::get_cranlike_metadata_cache()
  )

  res <- suppressMessages(synchronise(resolve_remote_bioc(
    parse_pkg_ref("bioc::nosuchpackageinbiocforsure"),
    TRUE,
    conf,
    cache,
    dependencies = FALSE
  )))

  expect_true(all(res$status == "FAILED"))

  expect_snapshot(
    as.list(res),
    transform = function(x) transform_local_port(transform_bioc_version(x)),
    variant = paste0("pillar-", packageVersion("pillar"))
  )
})

test_that("download_remote", {
  setup_fake_apps()

  dir.create(tmp <- tempfile())
  dir.create(tmp2 <- tempfile())
  on.exit(unlink(c(tmp, tmp2), recursive = TRUE), add = TRUE)

  conf <- current_config()
  conf$platforms <- "source"
  conf$cache_dir <- tmp
  conf$package_cache_dir <- tmp2
  cache <- list(
    package = pkgcache::package_cache$new(conf$package_cache_dir),
    metadata = pkgcache::get_cranlike_metadata_cache()
  )

  res <- suppressMessages(synchronise(resolve_remote_bioc(
    parse_pkg_ref("bioc::Biobase"),
    TRUE,
    conf,
    cache,
    dependencies = FALSE
  )))

  target <- file.path(conf$cache_dir, res$target[1])
  tree <- paste0(target, "-t")
  dl <- synchronise(download_remote_bioc(
    res[1, ],
    target,
    tree,
    conf,
    cache,
    on_progress = NULL
  ))

  expect_equal(dl, "Got")
  expect_true(file.exists(target))

  unlink(target)
  dl2 <- synchronise(download_remote_bioc(
    res[1, ],
    target,
    tree,
    conf,
    cache,
    on_progress = NULL
  ))
  expect_true(dl2 %in% c("Had", "Current"))
  expect_true(file.exists(target))
})

test_that("satisfies_remote", {
  res <- make_fake_resolution(`bioc::eisa@>=1.0.0` = list())

  ## GitHub type is never good
  bad1 <- make_fake_resolution(`github::r-lib/eisa` = list())
  expect_false(ans <- satisfy_remote_bioc(res, bad1))
  expect_match(attr(ans, "reason"), "Type must be")

  ## Missing DESCRIPTION for installed type
  bad2 <- make_fake_resolution(`installed::foobar` = list())
  expect_false(ans <- satisfy_remote_bioc(res, bad2))
  expect_match(attr(ans, "reason"), "not from BioC")

  ## installed, but not from BioC
  fake_desc <- desc::desc("!new")
  bad3 <- make_fake_resolution(
    `installed::foobar` = list(
      extra = list(list(description = fake_desc))
    )
  )
  expect_false(ans <- satisfy_remote_bioc(res, bad3))
  expect_match(attr(ans, "reason"), "not from BioC")

  ## BioC type, but package name does not match
  bad4 <- make_fake_resolution(`bioc::eisa2` = list())
  expect_false(ans <- satisfy_remote_bioc(res, bad4))
  expect_match(attr(ans, "reason"), "names differ")

  ## installed type, but package name does not match
  bad5 <- make_fake_resolution(
    `installed::foobar` = list(
      package = "eisa2",
      extra = list(list(repotype = "bioc"))
    )
  )
  expect_false(ans <- satisfy_remote_bioc(res, bad5))
  expect_match(attr(ans, "reason"), "names differ")

  ## BioC type, but version is not good enough
  bad6 <- make_fake_resolution(`bioc::eisa` = list(version = "0.0.1"))
  expect_false(ans <- satisfy_remote_bioc(res, bad6))
  expect_match(attr(ans, "reason"), "Insufficient version")

  ## Same version, BioC
  ok1 <- make_fake_resolution(`bioc::eisa` = list())
  expect_true(satisfy_remote_bioc(res, ok1))

  ## Newer version, BioC
  ok2 <- make_fake_resolution(`bioc::eisa` = list(version = "2.0.0"))
  expect_true(satisfy_remote_bioc(res, ok2))

  ## Same version, installed
  ok3 <- make_fake_resolution(
    `installed::foobar` = list(
      package = "eisa",
      extra = list(list(repotype = "bioc"))
    )
  )
  expect_true(satisfy_remote_bioc(res, ok3))

  ## Newer version, installed
  ok4 <- make_fake_resolution(
    `installed::foobar` = list(
      package = "eisa",
      version = "2.0.0",
      extra = list(list(repotype = "bioc"))
    )
  )
  expect_true(satisfy_remote_bioc(res, ok4))
})

test_that("satisfies_remote 2", {
  res <- make_fake_resolution(`bioc::eisa@>=1.0.0` = list(direct = TRUE))

  bad1 <- make_fake_resolution(
    `installed::foobar` = list(
      package = "eisa",
      version = "0.0.1",
      extra = list(list(repotype = "bioc"))
    )
  )
  expect_false(satisfy_remote_bioc(res, bad1))

  res2 <- make_fake_resolution(`bioc::eisa` = list())

  ok1 <- make_fake_resolution(
    `installed::foobar` = list(
      package = "eisa",
      version = "0.0.1",
      extra = list(list(repotype = "bioc"))
    )
  )
  expect_true(satisfy_remote_bioc(res2, ok1))
})

test_that("installedok", {
  # Bioc binary
  sol <- list(
    repotype = "bioc",
    package = "package",
    version = "1.0.0",
    platform = "aarch64-apple-darwin20",
    built = "R 4.2.0; aarch64-apple-darwin20; 2022-10-08 08:37:45 UTC; unix"
  )
  expect_true(installedok_remote_standard(
    list(
      repotype = "cran",
      package = "package",
      version = "1.0.0",
      platform = "aarch64-apple-darwin20",
      built = "R 4.2.0; aarch64-apple-darwin20; 2022-10-08 08:37:45 UTC; unix"
    ),
    sol
  ))
  expect_false(installedok_remote_standard(
    list(
      package = "package",
      version = "1.0.0",
      platform = "aarch64-apple-darwin20",
      built = "R 4.2.0; aarch64-apple-darwin20; 2022-11-08 08:37:45 UTC; unix"
    ),
    sol
  ))

  # Bioc source
  sol <- list(
    repotype = "bioc",
    package = "package",
    version = "1.0.0",
    platform = "source"
  )
  expect_true(installedok_remote_standard(
    list(
      package = "package",
      version = "1.0.0",
      repotype = "bioc"
    ),
    sol
  ))
  expect_false(installedok_remote_standard(
    list(
      package = "package",
      version = "1.0.0"
    ),
    sol
  ))
})

#' Package downloads
#'
#' The [`pkg_download_proposal`] and [`pkg_installation_proposal`] classes
#' both have download methods, to downloads package files into a
#' configured directory (see ['Configuration'][pkgdepends-config]).
#'
#' They return a `pkg_download_result` object, which is a data frame,
#' that adds extra columns to [`pkg_resolution_result`] (for
#' [`pkg_download_proposal`]) or [`pkg_solution_result`]
#' (for [`pkg_installation_proposal`]):
#'
#' ```{r child = {options(rx_downloads = TRUE); "tools/doc/resolution-result.Rmd" }}
#' ```
#' `r { options(rx_downloads = TRUE); doc_share_rmd("tools/doc/resolution-result.Rmd", "inst/docs/download-result.rds")}`
#'
#' @name pkg_downloads
#' @aliases pkg_download_result
NULL

pkgplan_download_resolution <- function(self, private) {
  if (is.null(private$resolution)) {
    self$resolve()
  }
  if (private$dirty) {
    throw(pkg_error(
      "Package list has changed, you need to call the {.code $resolve()}
       method again?"
    ))
  }
  on.exit(private$done_progress_bar(), add = TRUE)
  on.exit(cli::ansi_show_cursor(), add = TRUE)
  cli::ansi_hide_cursor()
  synchronise(self$async_download_resolution())
}

pkgplan_async_download_resolution <- function(self, private) {
  self
  private
  if (is.null(private$resolution)) {
    self$resolve()
  }
  if (private$dirty) {
    throw(pkg_error(
      "Package list has changed, you need to call the {.code $resolve()}
       method again?"
    ))
  }

  pkgplan_async_download_internal(
    self,
    private,
    private$resolution$result,
    "resolution"
  )$then(function(value) {
    private$downloads <- value
    self$get_resolution_download()
  })
}

pkgplan_download_solution <- function(self, private) {
  if (is.null(private$solution)) {
    self$solve()
  }
  if (private$dirty) {
    throw(pkg_error(
      "Package list has changed, you need to call the {.code $resolve()}
       method again?"
    ))
  }
  on.exit(private$done_progress_bar(), add = TRUE)
  on.exit(cli::ansi_show_cursor(), add = TRUE)
  cli::ansi_hide_cursor()
  synchronise(self$async_download_solution())
}

pkgplan_async_download_solution <- function(self, private) {
  if (is.null(private$solution)) {
    self$solve()
  }
  if (private$dirty) {
    throw(pkg_error(
      "Package list has changed, you need to call the {.code $resolve()}
       method again?"
    ))
  }

  pkgplan_async_download_internal(
    self,
    private,
    private$solution$result$data,
    "solution"
  )$then(function(value) {
    private$solution_downloads <- value
    self$get_solution_download()
  })
}

pkgplan_stop_for_solution_download_error <- function(self, private) {
  dl <- self$get_solution_download()
  if (any(bad <- tolower(dl$download_status) == "failed")) {
    msgs <- vcapply(
      which(bad),
      function(i) {
        urls <- format_items(dl$sources[[i]])
        sprintf("Failed to download %s from %s.", dl$package[i], urls)
      }
    )
    msg <- paste(msgs, collapse = "\n")
    err <- structure(
      list(message = msg, call = NULL, errors = dl$download_errors[bad]),
      class = c("error", "condition")
    )
    stop(err)
  }
}

pkgplan_stop_for_resolution_download_error <- function(self, private) {
  dl <- self$get_resolution_download()
  if (any(bad <- tolower(dl$download_status) == "failed")) {
    msgs <- vcapply(
      which(bad),
      function(i) {
        urls <- format_items(dl$sources[[i]])
        sprintf("Failed to download %s from %s.", dl$package[i], urls)
      }
    )
    msg <- paste(msgs, collapse = "\n")
    err <- structure(
      list(message = msg, call = NULL),
      class = c("error", "condition")
    )
    stop(err)
  }
}

pkgplan_async_download_internal <- function(self, private, what, which) {
  if (any(what$status != "OK")) {
    stop("Resolution has errors, cannot start downloading")
  }
  start <- Sys.time()
  private$progress_bar <- private$create_progress_bar(what)

  dl <- lapply(seq_len(nrow(what)), function(idx) {
    force(idx)
    private$download_res(
      what[idx, ],
      which = which,
      on_progress = function(data) private$update_progress_bar(idx, "got", data)
    )$then(function(x) {
      private$update_progress_bar(idx, "done", x)
      x
    })$catch(error = function(x) private$update_progress_bar(idx, "error", x))
  })

  when_all(.list = dl)$then(function(dls) {
    what$fulltarget <- vcapply(dls, "[[", "fulltarget")
    what$fulltarget_tree <- vcapply(dls, "[[", "fulltarget_tree")
    what$download_status <- vcapply(dls, "[[", "download_status")
    what$download_error <- lapply(dls, function(x) x$download_error[[1]])
    what$file_size <- vdapply(dls, "[[", "file_size")
    what$used_cached_binary <- vlapply(dls, "[[", "used_cached_binary")
    class(what) <- c("pkgplan_downloads", class(what))
    attr(what, "metadata")$download_start <- start
    attr(what, "metadata")$download_end <- Sys.time()
    what
  })$finally(function() private$done_progress_bar())
}

pkgplan_download_res <- function(self, private, res, which, on_progress) {
  force(private)
  download_remote(
    res,
    config = private$config,
    cache = private$cache,
    which = which,
    on_progress = on_progress
  )
}

download_remote <- function(
  res,
  config,
  cache,
  which,
  on_progress = NULL,
  remote_types = NULL
) {
  remote_types <- c(default_remote_types(), remote_types)
  dl <- remote_types[[res$type]]$download %||% type_default_download
  target <- file.path(config$get("cache_dir"), res$target)
  target_tree <- file.path(config$get("cache_dir"), paste0(res$target, "-t"))
  mkdirp(dirname(target))
  async(dl)(
    res,
    target,
    target_tree,
    config,
    cache = cache,
    which = which,
    on_progress = on_progress
  )$then(function(s) {
    if (
      length(res$sources[[1]]) &&
        !file.exists(target) &&
        !file.exists(target_tree)
    ) {
      stop("Failed to download ", res$type, " package ", res$package)
    }

    dlres <- res
    if (!grepl("^Had", s) && !identical(s, "Got") && !identical(s, "Current")) {
      s <- "Got"
    }
    if (grepl("^Had-binary-", s)) {
      dlres$used_cached_binary <- TRUE
      s <- "Had"
    } else {
      dlres$used_cached_binary <- FALSE
    }
    dlres$fulltarget <- target
    dlres$fulltarget_tree <- target_tree
    dlres$download_status <- s
    dlres$download_error <- list(NULL)
    dlres$file_size <- file.size(target)
    dlres
  })$catch(error = function(err) {
    dlres <- res
    dlres$used_cached_binary <- NA
    dlres$fulltarget <- target
    dlres$fulltarget_tree <- target_tree
    dlres$download_status <- "Failed"
    dlres$download_error <- list(err)
    dlres$file_size <- NA_integer_
    dlres
  })
}

download_ping_if_not_source <- function(
  resolution,
  target,
  config,
  cache,
  on_progress
) {
  resolution
  target
  config
  cache
  on_progress
  mkdirp(dirname(target))

  if (is_true_param(resolution$params[[1]], "nocache")) {
    ## If the cache is ignored, then we just download it and put it
    ## at the right place
    download_one_of(
      resolution$sources[[1]],
      target,
      on_progress = on_progress
    )$then(function() "Got")
  } else if (resolution$platform == "source") {
    ## If it is a source package, then the package name, version number
    ## and package type must match. If there is such a package in the cache
    ## we just take it
    cache$package$async_copy_or_add(
      target,
      resolution$sources[[1]],
      path = resolution$target,
      package = resolution$package,
      version = resolution$version,
      platform = resolution$platform,
      on_progress = on_progress,
      http_headers = default_download_headers(resolution$sources[[1]])
    )$then(function(.) attr(., "action"))
  } else {
    ## If not a source package, then we try to update it, in case there is
    ## a newly built binary
    cache$package$async_update_or_add(
      target,
      resolution$sources[[1]],
      path = resolution$target,
      package = resolution$package,
      version = resolution$version,
      platform = resolution$platform,
      on_progress = on_progress,
      http_headers = default_download_headers(resolution$sources[[1]])
    )$then(function(.) attr(., "action"))
  }
}

download_ping_if_no_sha <- function(
  resolution,
  target,
  config,
  cache,
  on_progress
) {
  resolution
  target
  config
  cache
  on_progress
  mkdirp(dirname(target))

  # If the cache is ignored, then we download it, always
  if (is_true_param(resolution$params[[1]], "nocache")) {
    return(
      download_one_of(
        resolution$sources[[1]],
        target,
        on_progress = on_progress
      )$then(function() "Got")
    )
  }

  # If we wanted to _install_ a source package, then look for a
  # locally built binary in the cache
  rver <- config$get("r-versions")
  if (
    resolution$platform == "source" &&
      config$get("goal") == "install" &&
      !is_true_param(resolution$params[[1]], "source") &&
      length(rver) == 1
  ) {
    ## Try to find a binary in the cache
    cplt <- current_r_platform()
    bin <- cache$package$copy_to(
      target,
      package = resolution$package,
      version = resolution$version,
      platform = cplt,
      built = TRUE,
      rversion = rver
    )
    if (nrow(bin)) {
      return(async_constant(paste0("Had-binary-", cplt)))
    }
  }

  if (!"sha256" %in% names(resolution) || is.na(resolution$sha256)) {
    ## Otherwise we need to ping or download a package
    ## If we don't know the hash of the CRAN package, then just download
    ## it. This happens if there is some discrepancy between the package
    ## data and the metadata.
    cache$package$async_update_or_add(
      target,
      resolution$sources[[1]],
      path = resolution$target,
      package = resolution$package,
      version = resolution$version,
      platform = resolution$platform,
      on_progress = on_progress,
      http_headers = default_download_headers(resolution$sources[[1]])
    )$then(function(.) attr(., "action"))
  } else {
    ## There is a sha hash in the metadata, so we can search for that
    ## in the package cache.
    cache$package$async_copy_or_add(
      target,
      resolution$sources[[1]],
      path = resolution$target,
      package = resolution$package,
      version = resolution$version,
      platform = resolution$platform,
      sha256 = resolution$sha256,
      on_progress = on_progress,
      http_headers = default_download_headers(resolution$sources[[1]])
    )$then(function(.) attr(., "action"))
  }
}

pkgplan_get_resolution_download <- function(self, private) {
  if (is.null(private$downloads)) {
    stop("No downloads")
  }
  private$downloads
}

pkgplan_get_solution_download <- function(self, private) {
  if (is.null(private$solution_downloads)) {
    stop("No downloads")
  }
  private$solution_downloads
}

#' @export

`[.pkgplan_downloads` <- function(x, i, j, drop = FALSE) {
  class(x) <- setdiff(class(x), "pkgplan_downloads")
  NextMethod("[")
}

type_default_download <- function(
  resolution,
  target,
  config,
  cache,
  on_progress
) {
  ## TODO
  stop("Not implemented yet")
}

default_download_headers <- function(url) {
  if (any(grepl("^https://ghcr.io", url))) {
    c("Authorization" = "Bearer QQ==")
  }
}

cat(
  sep = "\n",
  paste0(
    "\\renewcommand{\\eval}{\\Sexpr[stage=render,results=rd]{",
    Sys.getenv("R_PACKAGE_NAME"),
    ":::#1}}"
  ),
  paste0(
    "\\renewcommand{\\evalatinstall}{\\Sexpr[stage=install,results=rd]{",
    Sys.getenv("R_PACKAGE_NAME"),
    ":::#1}}"
  ),
  # The 'top' variants do not work in R < 4.0.0, because a top-level \Sexpr
  # is not allowed, even if it is put in by an Rd macro. Once we do not
  # support R 3.6.x, we can use the 'top' variants.
  # See also the comments in the include_docs() function.
  paste0(
    "\\renewcommand{\\evaltop}{\\Sexpr[stage=render,results=rd]{",
    Sys.getenv("R_PACKAGE_NAME"),
    ":::#1}}"
  ),
  paste0(
    "\\renewcommand{\\evalatinstalltop}{\\Sexpr[stage=install,results=rd]{",
    Sys.getenv("R_PACKAGE_NAME"),
    ":::#1}}"
  ),
  file = "man/macros/eval2.Rd"
)

#' @param refs Package names or references. See
#'   ['Package references'][pkg_refs] for the syntax.
#' @param config Configuration options, a named list. See
#'  ['Configuration'][pkgdepends-config]. If it does not include `library`, then
#'  `.libPaths()[1]` is added as `library`.
#' @param ... Additional arguments, passed to
#'   [`pkg_installation_proposal$new()`](#method-new).
#' @return `new_pkg_installation_proposal()` returns a new
#'   `pkg_installation_proposal` object.
#'
#' @export
#' @rdname pkg_installation_proposal

new_pkg_installation_proposal <- function(refs, config = list(), ...) {
  config$library <- config$library %||% .libPaths()[[1]]
  config$library <- path_norm(config$library)
  pkg_installation_proposal$new(refs, config = config, ...)
}

#' R6 class for package download and installation.
#'
#' @description
#' Download and install R packages, with their dependencies, from various
#' sources.
#'
#' @details
#' `new_pkg_installation_proposal()` creates a new object from the
#' `pkg_installation_proposal` class. The advantage of
#' `new_pkg_installation_proposal()` compared to using the
#' [pkg_installation_proposal] constructor directly is that it avoids
#' making pkgdepends a build time dependency.
#'
#' Typical workflow to install a set of packages:
#'
#' 1. Create a `pkg_installation_proposal` object with
#'    `new_pkg_installation_proposal()`.
#' 1. Resolve all possible dependencies with
#'    [`pkg_installation_proposal$resolve()`](#method-resolve).
#' 1. Solve the package dependencies, to get an installation plan, with
#'    [`pkg_installation_proposal$solve()`](#method-solve).
#' 1. Download all files with
#'    [`pkg_installation_proposal$download()`](#method-download).
#' 1. Install the downloaded files with
#'    [`pkg_installation_proposal$install()`](#methods-install).
#'
#' @param refs Package names or references. See
#'   ['Package references'][pkg_refs] for the syntax.
#'
#' @export
#' @examples
#' \dontrun{
#' pdi <- new_pkg_installation_proposal(
#'   "pak",
#'   config = list(library = tempfile())
#' )
#' pdi
#'
#' pdi$resolve()
#' pdi
#'
#' pdi$solve()
#' pdi
#'
#' pdi$download()
#' pdi
#' }

pkg_installation_proposal <- R6::R6Class(
  "pkg_installation_proposal",
  public = list(
    #' @description
    #' Create a new `pkg_installation_proposal` object. Consider using
    #' `new_pkg_installation_proposal()` instead of calling the constructor
    #' directly.
    #'
    #' The returned object can be used to look up (recursive) dependencies
    #' of R packages from various sources, and then download and install
    #' the package files.
    #'
    #' @param config Configuration options, a named list. See
    #'   ['Configuration'][pkgdepends-config]. It needs to include the package
    #'   library to install to, in `library`.
    #' @param policy Solution policy. See ['The dependency
    #'   solver'][pkg_solution].
    #' @param remote_types Custom remote ref types, this is for advanced
    #'   use, and experimental currently.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' pdi <- new_pkg_installation_proposal(
    #'   "r-lib/pkgdepends",
    #'   config = list(library = tempfile()))
    #' pdi

    initialize = function(
      refs,
      config = list(),
      policy = c("lazy", "upgrade"),
      remote_types = NULL
    ) {
      config$goal <- "install"
      policy <- match.arg(policy)
      assert_that(is_character(config$library))
      private$library <- config$library[1]
      private$policy <- policy
      private$plan <- pkg_plan$new(
        refs,
        config = config,
        library = config$library,
        remote_types = remote_types
      )
    },

    #' @description
    #' The package refs that were used to create the
    #' `pkg_installation_proposal` object.
    #'
    #' @return
    #' A character vector of package refs that were used to create the
    #' `pkg_installation_proposal` object.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' pdi <- new_pkg_installation_proposal("r-lib/pkgdepends")
    #' pdi$get_refs()

    get_refs = function() private$plan$get_refs(),

    #' @description
    #' Configuration options for the `pkg_installation_proposal` object. See
    #' ['Configuration'][pkgdepends-config] for details.
    #'
    #' @return
    #' Named list. See ['Configuration'][pkgdepends-config] for the configuration
    #' options.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' pdi <- new_pkg_installation_proposal(
    #'   "pak",
    #'   config = list(library = tempfile())
    #' )
    #' pdi$get_config()

    get_config = function() private$plan$get_config(),

    #' @description
    #' Resolve the dependencies of the specified package references. This
    #' usually means downloading metadata from CRAN and Bioconductor,
    #' unless already cached, and also from GitHub if GitHub refs were
    #' included, either directly or indirectly. See
    #' ['Dependency resolution'][pkg_resolution] for details.
    #'
    #' @return
    #' The `pkg_installation_proposal` object, invisibly.
    #'
    #' @examples
    #' \dontrun{
    #' pdi <- new_pkg_installation_proposal(
    #'   "pak",
    #'   config = list(library = tempfile())
    #' )
    #'
    #' pdi$resolve()
    #' pdi$get_resolution()
    #' }

    resolve = function() {
      private$plan$resolve()
      invisible(self)
    },

    #' @description
    #' The same as [`resolve()`](#method-resolve), but asynchronous. This
    #' method is for advanced use.
    #'
    #' @return
    #' A deferred value.

    async_resolve = function() private$plan$async_resolve(),

    #' @description
    #' Query the result of the dependency resolution. This method can be
    #' called after [`resolve()`](#method-resolve) has completed.
    #'
    #' @return
    #' A [pkg_resolution_result] object, which is also a data frame. See
    #' ['Dependency resolution'][pkg_resolution] for its columns.
    #'
    #' @examples
    #' \dontrun{
    #' pdi <- new_pkg_installation_proposal(
    #'   "r-lib/pkgdepends",
    #'   config = list(library = tempfile())
    #' )
    #' pdi$resolve()
    #' pdi$get_resolution()
    #' }

    get_resolution = function() private$plan$get_resolution(),

    #' @description
    #' Returns the current policy of the dependency solver.
    #' See ['The dependency solver'][pkg_solution] for details.
    #'
    #' @return
    #' A character vector of length one.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' pdi <- new_pkg_installation_proposal(
    #'   "r-lib/pkgdepends",
    #'   config = list(library = tempfile())
    #' )
    #' pdi$get_solve_policy()
    #' pdi$set_solve_policy("upgrade")
    #' pdi$get_solve_policy()

    get_solve_policy = function() private$policy,

    #' @description
    #' Set the current policy of the dependency solver.
    #' If the object already contains a solution and the new policy is
    #' different than the old policy, then the solution is deleted.
    #' See ['The dependency solver'][pkg_solution] for details.
    #'
    #' @param policy Policy to set.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' pdi <- new_pkg_installation_proposal(
    #'   "r-lib/pkgdepends",
    #'   config = list(library = tempfile())
    #' )
    #' pdi$get_solve_policy()
    #' pdi$set_solve_policy("upgrade")
    #' pdi$get_solve_policy()

    set_solve_policy = function(policy = c("lazy", "upgrade")) {
      policy <- match.arg(policy)
      if (private$policy != policy) {
        private$plan$delete_solution()
        private$policy <- policy
      }
    },

    #' @description
    #' Solve the package dependencies. Out of the resolved dependencies,
    #' it works out a set of packages, that can be installed together to
    #' create a functional installation. The set includes all directly
    #' specified packages, and all required (or suggested, depending on
    #' the configuration) packages as well. It includes every package at
    #' most once. See ['The dependency solver'][pkg_solution] for details.
    #'
    #' @return
    #' The `pkg_installation_proposal` object itself, invisibly.
    #'
    #' @examples
    #' \dontrun{
    #' pdi <- new_pkg_installation_proposal(
    #'   "r-lib/pkgdepends",
    #'   config = list(library = tempfile())
    #' )
    #' pdi$resolve()
    #' pdi$solve()
    #' pdi$get_solution()
    #' }

    solve = function() {
      private$plan$solve(policy = private$policy)
      invisible(self)
    },

    #' @description
    #' Returns the solution of the package dependencies.
    #'
    #' @return
    #' A [pkg_solution_result] object, which is a list. See
    #' [pkg_solution_result] for details.
    #'
    #' @examples
    #' \dontrun{
    #' pdi <- new_pkg_installation_proposal(
    #'   "r-lib/pkgdepends",
    #'   config = list(library = tempfile())
    #' )
    #' pdi$resolve()
    #' pdi$solve()
    #' pdi$get_solution()
    #' }

    get_solution = function() private$plan$get_solution(),

    #' @description
    #' Show the solution on the screen.
    #'
    #' @param key Whether to show the key to the package list annotation.
    #' @return
    #' A [pkg_solution_result] object, which is a list. See
    #' [pkg_solution_result] for details.
    #'
    #' @examples
    #' \dontrun{
    #' pdi <- new_pkg_installation_proposal(
    #'   "r-lib/pkgdepends",
    #'   config = list(library = tempfile())
    #' )
    #' pdi$resolve()
    #' pdi$solve()
    #' pdi$get_solution()
    #' pdi$show_solution()
    #' }

    show_solution = function(key = FALSE) private$plan$show_solution(key),

    #' @description
    #' Query and categorize system requirements.

    get_sysreqs = function() private$plan$get_sysreqs(),

    #' @description
    #' Show system requirements for the packages in the solution.

    show_sysreqs = function() private$plan$show_sysreqs(),

    #' @description
    #' Error if the dependency solver failed to find a consistent set of
    #' packages that can be installed together.
    #'
    #' @examples
    #' \dontrun{
    #' # This is an error, because the packages conflict:
    #' pdi <- new_pkg_installation_proposal(
    #'   c("r-lib/pak", "cran::pak"),
    #'   config = list(library = tempfile())
    #' )
    #' pdi$resolve()
    #' pdi$solve()
    #' pdi
    #' # This fails:
    #' # pdi$stop_for_solution_error()
    #' }

    stop_for_solution_error = function() {
      private$plan$stop_for_solve_error()
    },

    #' @description
    #' Create a lock file that contains the information to perform
    #' the installation later, possibly in another R session.
    #'
    #' @details
    #' Note, since the URLs of CRAN and most CRAN-like repositories change
    #' over time, in practice you cannot perform the plan of the lock file
    #' _much_ later. For example, binary packages of older package version
    #' are removed, and won't be found.
    #'
    #' Similarly, for `url::` remote types, the URL might hold an updated
    #' version of the package, compared to when the lock file was created.
    #' Should this happen, pkgdepends prints a warning, but it will try
    #' to continue the installation. The installation might fail if the
    #' updated package has different (e.g. new) dependencies.
    #'
    #' Currently the intended use case of lock files in on CI systems, to
    #' facilitate caching. The (hash of the) lock file provides a good key
    #' for caching systems.
    #'
    #' @param path Name of the lock file. The default is `pkg.lock` in the
    #'   current working directory.
    #' @param version Only version 1 is supported currently.

    create_lockfile = function(path = "pkg.lock", version = 1) {
      private$plan$export_install_plan(plan_file = path, version = version)
    },

    #' @description
    #' Draw a tree of package dependencies. It returns a `tree` object, see
    #' [cli::tree()]. Printing this object prints the dependency tree to the
    #' screen.
    #'
    #' @return
    #' A `tree` object from the cli package, see [cli::tree()].
    #'
    #' @examples
    #' \dontrun{
    #' pdi <- new_pkg_installation_proposal(
    #'   "pak",
    #'   config = list(library = tempfile())
    #' )
    #' pdi$resolve()
    #' pdi$solve()
    #' pdi$draw()
    #' }

    draw = function() private$plan$draw_solution_tree(),

    #' @description
    #' Download all packages that are part of the solution. It uses the
    #' package cache in the pkgcache package by default, to avoid downloads
    #' if possible.
    #'
    #' @return
    #' The `pkg_installation_proposal` object itself, invisibly.
    #'
    #' @examples
    #' \dontrun{
    #' pdi <- new_pkg_installation_proposal(
    #'   c("r-lib/pak", "cran::pak"),
    #'   config = list(library = tempfile())
    #' )
    #' pdi$resolve()
    #' pdi$solve()
    #' pdi$download()
    #' pdi$get_downloads()
    #' }

    download = function() invisible(private$plan$download_solution()),

    #' @description
    #' The same as [`download()`](#method-download), but asynchronous.
    #' This method is for advanced use.
    #'
    #' @return
    #' A deferred value.

    async_download = function() private$plan$async_download_solution(),

    #' @description
    #' Returns the summary of the package downloads.
    #'
    #' @return
    #' A [pkg_download_result] object, which is a list. See
    #' [pkg_download_result] for details.
    #'
    #' @examples
    #' \dontrun{
    #' pdi <- new_pkg_installation_proposal(
    #'   c("r-lib/pak", "cran::pak"),
    #'   config = list(library = tempfile())
    #' )
    #' pdi$resolve()
    #' pdi$solve()
    #' pdi$download()
    #' pdi$get_downloads()
    #' }

    get_downloads = function() private$plan$get_solution_download(),

    #' @description
    #' Throw and error if the some of the downloads have failed for the
    #' most recent
    #' [`pkg_installation_proposal$download()`](#method-download) call.

    stop_for_download_error = function() {
      private$plan$stop_for_solution_download_error()
    },

    #' @description
    #' Install the downloaded packages. It calls [install_package_plan()].
    #'
    #' @return
    #' The return value of [install_package_plan()].

    install = function() {
      plan <- private$plan$get_install_plan()
      nw <- get_num_workers()
      cache <- get_private(private$plan)$cache$package
      install_package_plan(
        plan,
        lib = private$library,
        num_workers = nw,
        cache = cache
      )
    },

    #' @description
    #' Install system requirements. It does nothing if system requirements
    #' are turned off.

    install_sysreqs = function() {
      config <- get_private(private$plan)$config
      if (!config$get("sysreqs")) {
        return()
      }
      srq <- self$get_solution()$data$sysreqs_packages
      if (is.null(srq)) {
        return(invisible())
      } # nocov
      cmds <- sysreqs2_scripts(
        srq,
        sysreqs_platform = config$get("sysreqs_platform"),
        missing = !config$get("sysreqs_update")
      )
      sysreqs_install(cmds, config)
    },

    #' Create an installation plan for the downloaded packages.
    #'
    #' @return
    #' An installation plan, see ['Installation plans'][install_plans] for
    #' the format.
    #'
    #' @examples
    #' \dontrun{
    #' pdi <- new_pkg_installation_proposal(
    #'   "pak",
    #'   config = list(library = tempfile())
    #' )
    #' pdi$resolve()
    #' pdi$solve()
    #' pdi$download()
    #' pdi$get_install_plan()
    #' }

    get_install_plan = function() private$plan$get_install_plan(),

    #' @description
    #' Format a `pkg_installation_proposal` object, typically for printing.
    #'
    #' @param ... not used currently.
    #'
    #' @return
    #' A character vector, each element should be a line in the printout.

    format = function(...) {
      refs <- private$plan$get_refs()

      has_dls <- private$plan$has_solution_downloads()
      dls <- if (has_dls) private$plan$get_solution_download()
      dls_err <- has_dls && any(dls$status == "Failed")

      has_sol <- private$plan$has_solution()
      sol <- if (has_sol) private$plan$get_solution()
      sol_err <- has_sol && sol$status != "OK"

      has_sys <- has_sol && !is.null(sol$sysreqs)

      c(
        "<pkg_installation_proposal>",
        "+ refs:",
        paste0("  - ", refs),
        paste0("+ solution policy: ", private$policy),
        if (has_sol) "+ has solution",
        if (sol_err) "x has solution errors",
        if (has_dls) "+ has downloads",
        if (dls_err) "x has download errors",
        if (!has_sol) "(use `$solve()` to solve dependencies)",
        if (has_sol && !sol_err && !has_dls) {
          "(use `$download()` to download packages)"
        },
        if (has_sol) "(use `$show_solution()` to see the packages to install",
        if (has_sol) "(use `$get_solution()` to see the full solution results)",
        if (has_sol && !sol_err) "(use `$draw()` to draw the dependency tree)",
        if (has_sol) "(use `$create_lockfile()` to write a lock file)",
        if (has_dls) "(use `$get_downloads()` to get download data)",
        if (has_dls) "(use `$get_install_plan()` to get the installation plan)",
        if (has_sys) "(use `$install_sysreqs()` to install system packages)",
        if (has_dls) "(use `$install()` to install the packages)"
      )
    },

    #' @description
    #' Prints a `pkg_installation_proposal` object to the screen.
    #'
    #' The printout includes:
    #'
    #' * The package refs.
    #' * The policy of the dependency solver.
    #' * Whether the object has the solved dependencies.
    #' * Whether the solution had errors.
    #' * Whether the object has downloads.
    #' * Whether the downloads had errors.
    #' * Advice on which methods to call next.
    #'
    #' See the example below.
    #'
    #' @param ... not used currently.
    #' @return
    #' The `pkg_installation_proposal` object itself, invisibly.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method print
    #' pdi <- new_pkg_installation_proposal(
    #'   "pak",
    #'   config = list(library = tempfile())
    #' )
    #' pdi
    #'
    #' pdi$resolve()
    #' pdi
    #'
    #' pdi$solve()
    #' pdi
    #'
    #' pdi$download()
    #' pdi

    print = function(...) cat(self$format(...), sep = "\n")
  ),

  private = list(
    plan = NULL,
    policy = NULL,
    library = NULL
  )
)

test_that("dependencies", {
  pkgs <- "
    Package: pkg
    Version: 1.0.0
    Imports: pi
    Depends: pd
    Suggests: ps
    Enhances: pe
    LinkingTo: pl

    Package: pi
    Depends: pid
    Imports: pii
    Suggests: pis
    Enhances: pie
    LinkingTo: pil

    Package: pd
    Depends: pdd
    Imports: pdi
    Suggests: pds
    Enhances: pde
    LinkingTo: pdl

    Package: ps
    Depends: psd
    Imports: psi
    Suggests: pss
    Enhances: pse
    LinkingTo: psl

    Package: pe
    Depends: ped
    Imports: pei
    Suggests: pes
    Enhances: pee
    LinkingTo: pel

    Package: pl
    Depends: pld
    Imports: pli
    Suggests: pls
    Enhances: ple
    LinkingTo: pll"

  oth <- apply(
    expand.grid("p", c("i", "d", "s", "e", "l"), c("i", "d", "s", "e", "l")),
    1,
    paste,
    collapse = ""
  )
  apkgs <- paste0(c(pkgs, sprintf("    Package: %s", oth)), collapse = "\n\n")
  repo <- dcf(apkgs)

  setup_fake_apps(cran_repo = repo)

  p <- suppressMessages(new_pkg_installation_proposal(
    "pkg",
    config = list(dependencies = NA, library = tempfile())
  ))
  suppressMessages(p$resolve())
  suppressMessages(p$solve())
  expect_snapshot(p$draw(), transform = transform_bytes)

  p <- suppressMessages(new_pkg_installation_proposal(
    "pkg",
    config = list(dependencies = TRUE, library = tempfile())
  ))
  suppressMessages(p$resolve())
  suppressMessages(p$solve())
  expect_snapshot(p$draw(), transform = transform_bytes)

  # We can restrict dependencies. We need to test this more, to
  # make sure that the install plan gets the right dependencies.
  p <- suppressMessages(new_pkg_installation_proposal(
    "pkg",
    config = list(dependencies = "LinkingTo", library = tempfile())
  ))
  suppressMessages(p$resolve())
  suppressMessages(p$solve())
  expect_snapshot(p$draw(), transform = transform_bytes)

  suppressMessages(p$download())
  plan <- p$get_install_plan()
  expect_snapshot({
    plan$package
    plan$dependencies
  })
})

test_that("self dependencies are OK", {
  lib <- tempfile()
  lock <- tempfile()
  on.exit(unlink(c(lib, lock), recursive = TRUE), add = TRUE)

  repo <- dcf(
    "
    Package: pkg
    Suggests: pkg
  "
  )

  setup_fake_apps(cran_repo = repo)

  p <- suppressMessages(new_pkg_installation_proposal(
    "pkg",
    config = list(
      dependencies = TRUE,
      library = lib
    )
  ))
  suppressMessages(p$resolve())
  suppressMessages(p$solve())
  sol <- p$get_solution()$data
  sol <- sol[order(sol$ref), ]
  expect_equal(sol$package, "pkg")
  p$create_lockfile(lock)

  plan <- new_pkg_installation_plan(lockfile = lock)
  expect_equal(
    plan$get_solution()$data$dependencies,
    list(character())
  )
})

test_that("circular soft-dependencies are OK", {
  lib <- tempfile()
  lock <- tempfile()
  on.exit(unlink(c(lib, lock), recursive = TRUE), add = TRUE)

  repo <- dcf(
    "
    Package: pkg1
    Suggests: pkg2, pkg3

    Package: pkg2
    Suggests: pkg1

    Package: pkg3
  "
  )

  setup_fake_apps(cran_repo = repo)

  p <- suppressMessages(new_pkg_installation_proposal(
    c("pkg1", "pkg2"),
    config = list(
      dependencies = TRUE,
      library = lib
    )
  ))
  suppressMessages(p$resolve())
  suppressMessages(p$solve())
  sol <- p$get_solution()$data
  sol <- sol[order(sol$ref), ]
  expect_equal(sort(sol$package), c("pkg1", "pkg2", "pkg3"))
  p$create_lockfile(lock)

  plan <- new_pkg_installation_plan(lockfile = lock)
  expect_equal(
    plan$get_solution()$data$dependencies,
    list(character(), character(), character())
  )
})

test_that("url remote basics", {
  setup_fake_apps()

  url <- paste0(
    fake_cran$url(),
    "/src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz"
  )
  r <- suppressMessages(new_pkg_installation_proposal(
    paste0("url::", url),
    config = list(library = tempfile())
  ))

  suppressMessages(r$resolve())
  suppressMessages(r$solve())
  suppressMessages(r$download())

  expect_snapshot(
    r$get_solution(),
    transform = transform_local_port
  )

  dl <- r$get_downloads()
  expect_null(dl$download_error[[1]])
  expect_true(file.exists(dl$fulltarget))
})

test_that("url to an tree (not a source package), from a zip file", {
  setup_fake_gh_app()
  url <- paste0(
    fake_gh$url(),
    "repos/r-lib/pak/zipball/",
    "111ef906acb58fe406370f7bc0a72cac55dbbb231ea687494c25742ca521255a"
  )
  r <- suppressMessages(new_pkg_installation_proposal(
    paste0("url::", url),
    config = list(library = tempfile())
  ))

  suppressMessages(r$resolve())
  suppressMessages(r$solve())
  suppressMessages(r$download())

  expect_snapshot(
    r$get_solution(),
    transform = transform_local_port
  )

  dl <- r$get_downloads()
  expect_null(dl$download_error[[1]])
  expect_true(file.exists(dl$fulltarget_tree))
})

test_that("url remote and install plan", {
  # We test this separately, because it is slightly different.
  # We need to download the file both at resolution and also at
  # downloads.
  setup_fake_apps()

  url <- paste0(
    fake_cran$url(),
    "/src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz"
  )
  r <- suppressMessages(new_pkg_installation_proposal(
    paste0("url::", url),
    config = list(library = tempfile())
  ))
  lockfile <- tempfile()
  on.exit(unlink(lockfile), add = TRUE)
  r$resolve()
  r$solve()
  r$create_lockfile(lockfile)

  tmpdir <- type_url_tempdir(r$get_resolution()$remote[[1]], r$get_config())
  rimraf(c(tmpdir$archive, tmpdir$extract, tmpdir$ok))

  plan <- new_pkg_installation_plan(lockfile)
  suppressMessages(plan$download())
  expect_null(plan$get_downloads()$download_error[[1]])
  file.exists(plan$get_downloads()$fulltarget)
})

test_that("satisfy", {
  # different package name is bad (although not really possible)
  expect_snapshot(satisfy_remote_url(
    list(package = "package"),
    list(package = "other")
  ))

  # installed, reinstall requested?
  expect_snapshot(satisfy_remote_url(
    list(
      package = "package",
      params = list(c(reinstall = ""))
    ),
    list(
      type = "installed",
      package = "package"
    )
  ))

  # installed, no etag
  expect_snapshot(satisfy_remote_url(
    list(
      package = "package",
      metadata = list(list(RemoteEtag = "myetag"))
    ),
    list(
      type = "installed",
      package = "package"
    )
  ))

  # installed, etag mismatch
  expect_snapshot(satisfy_remote_url(
    list(
      package = "package",
      metadata = list(list(RemoteEtag = "myetag"))
    ),
    list(
      type = "installed",
      package = "package",
      extra = list(list(remoteetag = "otheretag"))
    )
  ))

  # installed, etag match
  expect_snapshot(satisfy_remote_url(
    list(
      package = "package",
      metadata = list(list(RemoteEtag = "myetag"))
    ),
    list(
      type = "installed",
      package = "package",
      extra = list(list(remoteetag = "myetag"))
    )
  ))

  # url, url mismatch
  expect_snapshot(satisfy_remote_url(
    list(
      package = "package",
      ref = "https://myurl"
    ),
    list(
      type = "url",
      package = "package",
      ref = "https://anotherurl"
    )
  ))

  # url, url match
  expect_snapshot(satisfy_remote_url(
    list(
      package = "package",
      ref = "https://myurl"
    ),
    list(
      type = "url",
      package = "package",
      ref = "https://myurl"
    )
  ))

  # other type of remote
  expect_snapshot(satisfy_remote_url(
    list(
      package = "package",
      ref = "https://myurl"
    ),
    list(
      type = "local",
      package = "package"
    )
  ))
})

test_that("nocache", {
  setup_fake_apps()

  url <- paste0(
    fake_cran$url(),
    "/src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz"
  )
  r <- suppressMessages(new_pkg_installation_proposal(
    paste0("url::", url, "?nocache"),
    config = list(library = tempfile())
  ))

  suppressMessages(r$resolve())
  suppressMessages(r$solve())
  suppressMessages(r$download())

  expect_snapshot(
    r$get_solution(),
    transform = transform_local_port
  )

  dl <- r$get_downloads()
  expect_null(dl$download_error[[1]])
  expect_true(file.exists(dl$fulltarget))
})

test_that("bad archive file (multiple directories)", {
  mkdirp(tmp <- tempfile())
  withr::local_dir(tmp)
  mkdirp("pkg1")
  mkdirp("pkg2")
  expect_snapshot(
    error = TRUE,
    get_pkg_dir_from_archive_dir(tmp),
    transform = transform_tempdir
  )
})

test_that("installedok", {
  expect_true(installedok_remote_url(
    list(
      package = "package",
      version = "1.0.0",
      remotetype = "url",
      remoteetag = "myetag"
    ),
    list(
      package = "package",
      version = "1.0.0",
      metadata = list(list("RemoteEtag" = "myetag"))
    )
  ))

  expect_false(installedok_remote_url(
    list(
      package = "package",
      version = "1.0.0",
      remotetype = "url",
      remoteetag = "myetag"
    ),
    list(
      package = "package",
      version = "1.0.0",
      metadata = list(list("RemoteEtag" = "otheretag"))
    )
  ))
})

test_that("failure in non-needed package is ignored", {
  setup_fake_gh_app()
  setup_fake_apps()
  pkgcache::pkg_cache_delete_files()

  lib <- withr::local_tempdir()
  p <- new_pkg_installation_proposal(
    c("needspak", "r-lib/pak"),
    config = list(library = lib)
  )
  suppressWarnings(p$solve())
  expect_snapshot(p$draw(), transform = transform_bytes)
})

test_that("local dep needs name", {
  setup_fake_apps()

  withr::local_dir(tmp <- withr::local_tempdir())

  mkdirp("pkg1")
  file.create("pkg1/NAMESPACE")
  writeLines(
    c(
      "Package: pkg1",
      "Version: 1.0.0",
      "License: MIT",
      "Imports: pkg2",
      "Remotes: local::./pkg2"
    ),
    "pkg1/DESCRIPTION"
  )

  mkdirp("pkg2")
  file.create("pkg2/NAMESPACE")
  writeLines(
    c(
      "Package: pkg2",
      "Version: 1.0.0",
      "License: MIT"
    ),
    "pkg2/DESCRIPTION"
  )

  mkdirp("lib")
  config <- list(library = file.path(tmp, "lib"))
  p <- new_pkg_installation_proposal("./pkg1", config = config)
  p$solve()
  expect_snapshot(
    error = TRUE,
    p$stop_for_solution_error(),
    transform = transform_no_srcref
  )

  desc::desc_set(Remotes = "pkg2=local::./pkg2", file = "pkg1")
  p <- new_pkg_installation_proposal("./pkg1", config = config)
  p$solve()
  expect_snapshot(p$get_solution())
})

test_that("url dep needs name", {
  setup_fake_apps()

  withr::local_dir(tmp <- withr::local_tempdir())

  mkdirp("pkg2")
  file.create("pkg2/NAMESPACE")

  url <- paste0(
    fake_cran$url(),
    "/src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz"
  )
  writeLines(
    c(
      "Package: pkg2",
      "Version: 1.0.0",
      "License: MIT",
      "Imports: pkg1",
      paste0("Remotes: url::", url)
    ),
    "pkg2/DESCRIPTION"
  )

  mkdirp("lib")
  config <- list(library = file.path(tmp, "lib"))
  p <- new_pkg_installation_proposal("./pkg2", config = config)
  p$solve()
  expect_snapshot(
    error = TRUE,
    p$stop_for_solution_error(),
    transform = function(x) transform_no_srcref(transform_local_port(x))
  )

  desc::desc_set(Remotes = paste0("pkg1=url::", url), file = "pkg2")
  p <- new_pkg_installation_proposal("./pkg2", config = config)
  p$solve()
  expect_snapshot(
    p$get_solution(),
    transform = transform_local_port
  )
})

test_that("resolve_remote", {
  pkgcache::pkg_cache_delete_files()
  setup_fake_apps()

  config <- list(sysreqs_platform = "unknown")
  prop <- new_pkg_installation_proposal("pkg1", config = config)
  suppressMessages(prop$resolve())
  res <- prop$get_resolution()

  expect_snapshot(
    {
      snapshot(res, extra = "all")
    },
    transform = fix_port
  )
})

test_that("resolve_remote, multiple", {
  pkgcache::pkg_cache_delete_files()
  setup_fake_apps()

  config <- list(sysreqs_platform = "unknown")
  prop <- new_pkg_installation_proposal(
    c("cran::pkg3", "pkg1"),
    config = config
  )
  suppressMessages(prop$resolve())
  res <- prop$get_resolution()

  expect_snapshot(
    {
      snapshot(res, extra = "all")
    },
    transform = fix_port
  )
})

test_that("dependencies", {
  pkgcache::pkg_cache_delete_files()
  setup_fake_apps()

  config <- list(sysreqs_platform = "unknown")
  prop <- new_pkg_installation_proposal("pkg3", config = config)
  suppressMessages(prop$resolve())
  res <- prop$get_resolution()

  expect_snapshot(
    {
      snapshot(res, extra = "all")
    },
    transform = fix_port
  )
})

test_that("failed resolution", {
  pkgcache::pkg_cache_delete_files()
  setup_fake_apps()

  config <- list(sysreqs_platform = "unknown")
  prop <- new_pkg_installation_proposal(
    "cran::xxyyzzqwertyqwerty",
    config = config
  )
  suppressMessages(prop$resolve())
  res <- prop$get_resolution()

  expect_snapshot(
    {
      snapshot(res, extra = "all")
    },
    transform = fix_port
  )
})

test_that("failed resolution, multiple", {
  pkgcache::pkg_cache_delete_files()
  setup_fake_apps()

  config <- list(sysreqs_platform = "unknown")
  prop <- new_pkg_installation_proposal(
    c("cran::pkg1", "cran::xxyyzzqwertyqwerty"),
    config = config
  )
  suppressMessages(prop$resolve())
  res <- prop$get_resolution()

  expect_snapshot(
    {
      snapshot(res, extra = "all")
    },
    transform = fix_port
  )
})

test_that("resolve current version", {
  pkgcache::pkg_cache_delete_files()
  setup_fake_apps()

  config <- list(sysreqs_platform = "unknown")
  prop <- new_pkg_installation_proposal(
    "cran::pkg1@current",
    config = config
  )
  suppressMessages(prop$resolve())
  res <- prop$get_resolution()

  expect_snapshot(
    {
      snapshot(res, extra = "all")
    },
    transform = fix_port
  )
})

test_that("resolve an old version", {
  pkgcache::pkg_cache_delete_files()
  setup_fake_apps()

  config <- list(sysreqs_platform = "unknown")
  prop <- new_pkg_installation_proposal("pkg1@0.9.0", config = config)
  suppressMessages(prop$resolve())
  res <- prop$get_resolution()

  expect_snapshot(
    {
      snapshot(res, extra = "all")
    },
    transform = function(x) {
      transform_etag(transform_hash(transform_no_srcref(fix_port(x))))
    }
  )

  config <- list(sysreqs_platform = "unknown")
  prop <- new_pkg_installation_proposal("pkg1@1.0.0", config = config)
  suppressMessages(prop$resolve())
  res <- prop$get_resolution()

  expect_snapshot(
    {
      snapshot(res, extra = "all")
    },
    transform = function(x) transform_no_srcref(fix_port(x))
  )
})

test_that("resolve a version range", {
  pkgcache::pkg_cache_delete_files()
  setup_fake_apps()

  config <- list(sysreqs_platform = "unknown")
  prop <- new_pkg_installation_proposal("pkg1@>=0.9.0", config = config)
  suppressMessages(prop$resolve())
  res <- prop$get_resolution()

  expect_snapshot(
    {
      snapshot(res, extra = "all")
    },
    transform = function(x) transform_no_srcref(fix_port(x))
  )
})

test_that("download_remote", {
  pkgcache::pkg_cache_delete_files()
  setup_fake_apps()

  # this the cache for getOption(repos)
  pkgcache::pkg_cache_delete_files()

  conf <- list(library = lib <- tempfile())
  prop <- suppressMessages(
    new_pkg_installation_proposal("cran::pkg1", config = conf)
  )
  suppressMessages(prop$resolve())
  suppressMessages(prop$solve())
  dl <- suppressMessages(prop$download())

  expect_true(file.exists(dl$fulltarget))
  expect_false(file.exists(dl$fulltarget_tree))
  expect_equal(dl$download_status, "Got")
  expect_equal(dl$file_size, file.size(dl$fulltarget))
  expect_equal(dl$download_error, list(NULL))

  cache <- pkgcache::pkg_cache_summary()
  expect_equal(cache$files, 1)
  expect_equal(cache$size, dl$file_size)

  # second time it will be cached
  prop <- suppressMessages(
    new_pkg_installation_proposal("cran::pkg1", config = conf)
  )
  suppressMessages(prop$resolve())
  suppressMessages(prop$solve())
  dl2 <- suppressMessages(prop$download())

  expect_true(file.exists(dl2$fulltarget))
  expect_false(file.exists(dl2$fulltarget_tree))
  expect_equal(dl2$download_status, "Had")
  expect_equal(dl2$file_size, file.size(dl2$fulltarget))
  expect_equal(dl2$download_error, list(NULL))

  pkgcache::pkg_cache_delete_files()
})

test_that("versioned cran", {
  pkgcache::pkg_cache_delete_files()
  setup_fake_apps()

  dsc <- desc::desc("!new")
  dsc$set_dep("pkg3", "Imports")
  dsc$set_dep("pkg1", "Imports")
  dsc$set_remotes(c("pkg3@0.9.9", "pkg1@1.0.0"))
  dir.create(tmp <- tempfile())
  dsc_path <- file.path(tmp, "local", "DESCRIPTION")
  dir.create(dirname(dsc_path))
  dir.create(lib <- tempfile())
  on.exit(unlink(c(tmp, lib), recursive = TRUE), add = TRUE)
  dsc$write(dsc_path)
  prop <- pkgdepends::new_pkg_deps(
    paste0("local::", dirname(dsc_path)),
    config = list(library = lib)
  )
  suppressMessages(prop$solve())

  expect_snapshot(
    prop$draw(),
    transform = function(x) {
      sub("local::.*$", "local::<path>", transform_bytes(x))
    }
  )
})

test_that("reinstall, standard", {
  pkgcache::pkg_cache_delete_files()
  setup_fake_apps()

  lib <- withr::local_tempdir("pkgdepends-")
  config <- list(library = lib)
  p <- new_pkg_installation_proposal("pkg3", config = config)
  suppressMessages(p$solve())
  suppressMessages(p$download())
  suppressMessages(p$install())

  p <- new_pkg_installation_proposal("pkg1", config = config)
  suppressMessages(p$solve())
  expect_snapshot(
    {
      "pkg1 is already installed"
      p$get_solution()
    },
    transform = transform_installed_in_temp
  )

  p <- new_pkg_installation_proposal("pkg3?reinstall", config = config)
  suppressMessages(p$solve())
  expect_snapshot(
    {
      "request a reinstall"
      p$get_solution()
    },
    transform = transform_installed_in_temp
  )

  p <- new_pkg_installation_proposal(
    c("pkg3", "pkg1=?reinstall"),
    config = config
  )
  suppressMessages(p$solve())
  expect_snapshot(
    {
      "request a reinstall of a dependency"
      p$get_solution()
    },
    transform = transform_installed_in_temp
  )

  p <- new_pkg_installation_proposal(
    c("pkg3", "pkg1?reinstall"),
    config = config
  )
  suppressMessages(p$solve())
  expect_snapshot(
    {
      "one reinstall, one not"
      p$get_solution()
    },
    transform = transform_installed_in_temp
  )
})

test_that("reinstsll from URL", {
  pkgcache::pkg_cache_delete_files()
  setup_fake_apps()

  lib <- withr::local_tempdir("pkgdepends-")
  config <- list(library = lib)
  p <- new_pkg_installation_proposal("pkg3", config = config)
  suppressMessages(p$solve())
  suppressMessages(p$download())
  suppressMessages(p$install())

  url <- paste0(
    fake_cran$url(),
    "/src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz"
  )
  p <- new_pkg_installation_proposal(
    paste0("url::", url, "?reinstall"),
    config = config
  )
  suppressMessages(p$solve())
  expect_snapshot(
    {
      "reinstall from direct URL"
      p$get_solution()
    },
    transform = function(x) transform_local_port(transform_installed_in_temp(x))
  )

  p <- new_pkg_installation_proposal(
    c("pkg1=?reinstall", paste0("url::", url)),
    config = config
  )
  suppressMessages(p$solve())
  expect_snapshot(
    {
      "reinstall from URL, extra parameters"
      p$get_solution()
    },
    transform = function(x) transform_local_port(transform_installed_in_temp(x))
  )
})

test_that("source", {
  repo <- dcf(
    "
    Package: pkg
    Imports: pkg2

    Package: pkg2
  "
  )

  setup_fake_apps(
    cran_repo = repo,
    cran_options = list(platforms = c("windows", "source"))
  )

  p <- suppressMessages(new_pkg_installation_proposal(
    "pkg?source",
    config = list(
      dependencies = TRUE,
      library = tempfile(),
      platforms = c("x86_64-w64-mingw32", "source")
    )
  ))
  suppressMessages(p$resolve())
  suppressMessages(p$solve())
  sol <- p$get_solution()$data
  sol <- sol[order(sol$ref), ]
  expect_snapshot(sol[, c("package", "platform")])
})

test_that("source for dependency", {
  repo <- dcf(
    "
    Package: pkg
    Imports: pkg2

    Package: pkg2
  "
  )

  setup_fake_apps(
    cran_repo = repo,
    cran_options = list(platforms = c("windows", "source"))
  )

  p <- suppressMessages(new_pkg_installation_proposal(
    c("pkg", "pkg2=?source"),
    config = list(
      dependencies = TRUE,
      library = tempfile(),
      platforms = c("x86_64-w64-mingw32", "source")
    )
  ))
  suppressMessages(p$resolve())
  suppressMessages(p$solve())
  sol <- p$get_solution()$data
  sol <- sol[order(sol$ref), ]
  expect_snapshot(sol[, c("package", "platform")])
})

test_that("?ignore is ignored", {
  lib <- tempfile()
  lock <- tempfile()
  on.exit(unlink(c(lib, lock), recursive = TRUE), add = TRUE)

  repo <- dcf(
    "
    Package: pkg
    Suggests: pkg2

    Package: pkg2

    Package: pkg3
  "
  )

  setup_fake_apps(cran_repo = repo)

  p <- suppressMessages(new_pkg_installation_proposal(
    c("pkg", "pkg2=?ignore"),
    config = list(
      dependencies = TRUE,
      library = lib
    )
  ))
  suppressMessages(p$resolve())
  suppressMessages(p$solve())
  sol <- p$get_solution()$data
  sol <- sol[order(sol$ref), ]
  expect_equal(sol$package, "pkg")
  p$create_lockfile(lock)

  plan <- new_pkg_installation_plan(lockfile = lock)
  expect_equal(
    plan$get_solution()$data$dependencies,
    list(character())
  )
})

test_that("?ignore-before-r is ignored", {
  lib <- tempfile()
  lock <- tempfile()
  on.exit(unlink(c(lib, lock), recursive = TRUE), add = TRUE)

  repo <- dcf(
    "
    Package: pkg
    Suggests: pkg2

    Package: pkg2

    Package: pkg3
  "
  )

  setup_fake_apps(cran_repo = repo)

  p <- suppressMessages(new_pkg_installation_proposal(
    c("pkg", "pkg2=?ignore-before-r=100.0"),
    config = list(
      dependencies = TRUE,
      library = lib
    )
  ))
  suppressMessages(p$resolve())
  suppressMessages(p$solve())
  sol <- p$get_solution()$data
  sol <- sol[order(sol$ref), ]
  expect_equal(sol$package, "pkg")
  p$create_lockfile(lock)

  plan <- new_pkg_installation_plan(lockfile = lock)
  expect_equal(
    plan$get_solution()$data$dependencies,
    list(character())
  )
})

test_that("installed preferred over download", {
  setup_fake_apps()

  lib <- tempfile()
  on.exit(unlink(lib, recursive = TRUE), add = TRUE)

  p <- suppressMessages(new_pkg_installation_proposal(
    "pkg2",
    config = list(dependencies = TRUE, library = lib)
  ))
  suppressMessages(p$solve())
  suppressMessages(p$download())
  suppressMessages(p$install())

  p2 <- suppressMessages(new_pkg_installation_proposal(
    "pkg2",
    config = list(dependencies = TRUE, library = lib)
  ))
  suppressMessages(p2$solve())
  sol <- p2$get_solution()$data
  sol <- sol[order(sol$ref), ]
  expect_snapshot(sol[, c("type", "package", "version")])

  # multiple libraries
  lib2 <- tempfile()
  on.exit(unlink(lib2, recursive = TRUE), add = TRUE)

  p3 <- suppressMessages(new_pkg_installation_proposal(
    "pkg2",
    config = list(dependencies = TRUE, library = c(lib2, lib))
  ))
  suppressMessages(p3$solve())
  sol <- p3$get_solution()$data
  sol <- sol[order(sol$ref), ]
  expect_snapshot(sol[, c("type", "package", "version")])

  p4 <- suppressMessages(new_pkg_installation_proposal(
    "pkg2",
    config = list(dependencies = TRUE, library = c(lib2))
  ))
  suppressMessages(p4$solve())
  sol <- p4$get_solution()$data
  sol <- sol[order(sol$ref), ]
  expect_snapshot(sol[, c("type", "package", "version")])
})

test_that("build_package", {
  args <- NULL
  mockery::stub(
    build_package,
    "pkgbuild::build",
    function(...) args <<- list(...)
  )
  build_package(tmp <- tempfile())
  expect_equal(args$path, tmp)
})

test_that("vignettes can be turned on and off", {
  setup_fake_apps()
  local_cli_config()
  dir.create(tmplib <- tempfile())
  on.exit(rimraf(tmplib), add = TRUE)
  pkgdir <- test_path("fixtures", "packages", "vignettes")
  inst <- new_pkg_installation_proposal(
    paste0("local::", pkgdir, "?nocache"),
    config = list(`build-vignettes` = FALSE, library = tmplib)
  )
  expect_snapshot({
    inst$solve()
    inst$download()
    inst$install()
  })

  expect_false("doc" %in% dir(file.path(tmplib, "pkgdependstest")))
  rimraf(tmplib, "pkgdependstest")

  if (Sys.which("pandoc") == "") {
    skip("Needs pandoc")
  }

  inst2 <- new_pkg_installation_proposal(
    paste0("local::", pkgdir, "?nocache"),
    config = list(`build-vignettes` = TRUE, library = tmplib)
  )

  expect_snapshot({
    inst2$solve()
    inst2$download()
    inst2$install()
  })

  expect_true("doc" %in% dir(file.path(tmplib, "pkgdependstest")))
})

cli::test_that_cli("draw_solution_tree", {
  cran_app_pkgs <- dcf(
    "
    Package: pkgconfig
    Version: 2.0.3
    Suggests: covr, testthat, disposables (>= 1.0.3)
    Imports: utils

    Package: dplyr
    Version: 1.0.10
    Imports: generics, glue (>= 1.3.2), lifecycle (>= 1.0.1), magrittr (>= 1.5), methods, R6, rlang (>= 1.0.2), tibble (>= 2.1.3), tidyselect (>= 1.1.1), utils, vctrs (>= 0.4.1), pillar (>= 1.5.1)
    Suggests: bench, broom, callr, covr, DBI, dbplyr (>= 1.4.3), ggplot2, knitr, Lahman, lobstr, microbenchmark, nycflights13, purrr, rmarkdown, RMySQL, RPostgreSQL, RSQLite, testthat (>= 3.1.1), tidyr, withr

    Package: R6
    Version: 2.5.1
    Suggests: testthat, pryr

    Package: cli
    Version: 3.4.1
    Imports: utils
    Suggests: callr, covr, digest, glue (>= 1.6.0), grDevices, htmltools, htmlwidgets, knitr, methods, mockery, processx, ps (>= 1.3.4.9000), rlang (>= 1.0.2.9003), rmarkdown, rprojroot, rstudioapi, testthat, tibble, whoami, withr

    Package: fansi
    Version: 1.0.3
    Imports: grDevices, utils
    Suggests: unitizer, knitr, rmarkdown

    Package: generics
    Version: 0.1.3
    Imports: methods
    Suggests: covr, pkgload, testthat (>= 3.0.0), tibble, withr

    Package: glue
    Version: 1.6.2
    Imports: methods
    Suggests: covr, crayon, DBI, dplyr, forcats, ggplot2, knitr, magrittr, microbenchmark, R.utils, rmarkdown, rprintf, RSQLite, stringr, testthat (>= 3.0.0), vctrs (>= 0.3.0), waldo (>= 0.3.0), withr

    Package: lifecycle
    Version: 1.0.3
    Imports: cli (>= 3.4.0), glue, rlang (>= 1.0.6)
    Suggests: covr, crayon, knitr, lintr, rmarkdown, testthat (>= 3.0.1), tibble, tidyverse, tools, vctrs, withr

    Package: magrittr
    Version: 2.0.3
    Suggests: covr, knitr, rlang, rmarkdown, testthat

    Package: pillar
    Version: 1.8.1
    Imports: cli (>= 2.3.0), fansi, glue, lifecycle, rlang (>= 1.0.2), utf8 (>= 1.1.0), utils, vctrs (>= 0.3.8)
    Suggests: bit64, debugme, DiagrammeR, dplyr, formattable, ggplot2, knitr, lubridate, nanotime, nycflights13, palmerpenguins,  rmarkdown, scales, stringi, survival, testthat (>= 3.1.1), tibble, units (>= 0.7.2), vdiffr, withr

    Package: rlang
    Version: 1.0.6
    Imports: utils
    Suggests: cli (>= 3.1.0), covr, crayon, fs, glue, knitr, magrittr, methods, pillar, rmarkdown, stats, testthat (>= 3.0.0), tibble, usethis, vctrs (>= 0.2.3), withr

    Package: tibble
    Version: 3.1.8
    Imports: fansi (>= 0.4.0), lifecycle (>= 1.0.0), magrittr, methods, pillar (>= 1.7.0), pkgconfig, rlang (>= 1.0.2), utils, vctrs (>= 0.3.8)
    Suggests: bench, bit64, blob, brio, callr, cli, covr, crayon (>= 1.3.4), DiagrammeR, dplyr, evaluate, formattable, ggplot2, hms, htmltools, knitr, lubridate, mockr, nycflights13, pkgbuild, pkgload, purrr, rmarkdown, stringi, testthat (>= 3.0.2), tidyr, withr

    Package: tidyselect
    Version: 1.2.0
    Imports: cli (>= 3.3.0), glue (>= 1.3.0), lifecycle (>= 1.0.3), rlang (>= 1.0.4), vctrs (>= 0.4.1), withr
    Suggests: covr, crayon, dplyr, knitr, magrittr, rmarkdown, stringr, testthat (>= 3.1.1), tibble (>= 2.1.3)

    Package: utf8
    Version: 1.2.2
    Suggests: cli, covr, knitr, rlang, rmarkdown, testthat (>= 3.0.0), withr

    Package: vctrs
    Version: 0.4.2
    Imports: cli (>= 3.2.0), glue, rlang (>= 1.0.2)
    Suggests: bit64, covr, crayon, dplyr (>= 0.8.5), generics, knitr, pillar (>= 1.4.4), pkgdown (>= 2.0.1), rmarkdown, testthat (>= 3.0.0), tibble (>= 3.1.3), withr, xml2, waldo (>= 0.2.0), zeallot

    Package: withr
    Version: 2.5.0
    Imports: graphics, grDevices, stats
    Suggests: callr, covr, DBI, knitr, lattice, methods, rlang, rmarkdown (>= 2.12), RSQLite, testthat (>= 3.0.0)
  "
  )

  fake_cran <- webfakes::local_app_process(
    cran_app(cran_app_pkgs),
    opts = webfakes::server_opts(num_threads = 3)
  )
  withr::local_options(
    repos = c(CRAN = fake_cran$url()),
    pkg.cran_metadata_url = fake_cran$url(),
    pkg.emoji = FALSE
  )
  withr::local_envvar(
    R_PKG_CRAN_METADATA_URL = fake_cran$url()
  )

  dir.create(lib <- tempfile())
  on.exit(unlink(lib, recursive = TRUE), add = TRUE)

  r <- pkg_plan$new(
    c("pkgconfig", "dplyr"),
    library = lib,
    config = list(use_bioconductor = FALSE)
  )
  suppressMessages(r$resolve())
  suppressMessages(r$solve())

  expect_snapshot(
    r$draw_solution_tree(),
    transform = transform_bytes
  )
})

test_that("tree from lockfile", {
  setup_fake_apps()
  lib <- withr::local_tempdir()

  # create a lockfile first
  plan <- suppressMessages(new_pkg_installation_proposal(
    "pkg3",
    config = list(library = lib)
  ))
  suppressMessages(plan$resolve())
  plan$solve()
  lockfile <- tempfile(fileext = ".lock")
  on.exit(unlink(lockfile), add = TRUE)
  plan$create_lockfile(lockfile)

  plan <- pkgdepends::new_pkg_installation_plan(
    lockfile,
    config = list(library = lib)
  )
  expect_snapshot(plan$draw(), transform = transform_bytes)
})

test_that("update", {
  setup_fake_apps()
  pkgcache::pkg_cache_delete_files()
  lib <- withr::local_tempdir()

  url <- paste0(
    fake_cran$url(),
    "/src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz"
  )

  config <- list(library = lib)
  plan <- suppressMessages(new_pkg_installation_proposal(
    paste0("url::", url),
    config = config
  ))
  suppressMessages(plan$solve())
  suppressMessages(plan$download())
  suppressMessages(plan$install())

  plan2 <- new_pkg_installation_proposal("pkg1", config = config)
  plan2$solve()
  expect_snapshot(plan2$draw(), transform = transform_bytes)

  plan3 <- new_pkg_installation_proposal("any::pkg1", config = config)
  plan3$solve()
  expect_snapshot(plan3$draw(), transform = transform_bytes)
})

test_that("has_emoji", {
  mockery::stub(has_emoji, "cli::is_utf8_output", FALSE)
  expect_false(has_emoji())

  mockery::stub(has_emoji, "cli::is_utf8_output", TRUE)
  withr::local_options(pkg.emoji = TRUE)
  expect_true(has_emoji())

  withr::local_options(pkg.emoji = FALSE)
  expect_false(has_emoji())

  withr::local_options(pkg.emoji = NULL)
  mockery::stub(has_emoji, "Sys.info", c(sysname = "Linux"))
  expect_false(has_emoji())

  mockery::stub(has_emoji, "Sys.info", c(sysname = "Darwin"))
  expect_true(has_emoji())
})

test_that("no emoji", {
  mockery::stub(emoji, "has_emoji", FALSE)
  mockery::stub(emoji, "emo_builder", "\U1F477")
  expect_snapshot({
    emoji("rocket")
    emoji("sparkles")
    emoji("hand")
    emoji("dl")
    emoji("builder")
    emoji("wrench")
    emoji("pkg")
    emoji("pkgs")
    emoji("foobar")
  })
})

# The rest is for UTF-8 systems only

test_that("emoji", {
  if (!l10n_info()$"UTF-8") {
    skip("Not UTF-8")
  }
  mockery::stub(emoji, "has_emoji", TRUE)
  mockery::stub(emoji, "emo_builder", "\U1F477")
  expect_snapshot({
    emoji("rocket")
    emoji("sparkles")
    emoji("hand")
    emoji("dl")
    emoji("builder")
    emoji("wrench")
    emoji("pkg")
    emoji("pkgs")
    emoji("foobar")
  })
})

test_that("emo_builder", {
  if (!l10n_info()$"UTF-8") {
    skip("Not UTF-8")
  }

  # make it deterministic
  mockery::stub(
    emo_builder,
    "sample",
    function(x, size, replace = FALSE, ...) {
      rep_len(utils::tail(x, size), size)
    }
  )

  mockery::stub(emo_builder, "rstudio_detect", list(type = "rstudio_terminal"))
  expect_snapshot(emo_builder(5))

  mockery::stub(emo_builder, "rstudio_detect", list(type = "rstudio_console"))
  expect_snapshot(emo_builder(5))

  mockery::stub(emo_builder, "rstudio_detect", list(type = "not_rstudio"))
  expect_snapshot(emo_builder(5))
})

test_that("parse_remote_git", {
  expect_snapshot(parse_remote_git("git::https://github.com/r-lib/cli"))
  expect_snapshot(parse_remote_git("git::https://github.com/r-lib/cli.git"))
  expect_snapshot(parse_remote_git("git::https://github.com/r-lib/cli@branch"))
  expect_snapshot(parse_remote_git(
    "git::https://github.com/r-lib/cli.git@branch"
  ))

  expect_snapshot(parse_remote_git("pkg=git::https://github.com/r-lib/cli"))
  expect_snapshot(parse_remote_git("pkg=git::https://github.com/r-lib/cli.git"))
  expect_snapshot(parse_remote_git(
    "pkg=git::https://github.com/r-lib/cli@branch"
  ))
  expect_snapshot(parse_remote_git(
    "pkg=git::https://github.com/r-lib/cli.git@branch"
  ))

  expect_snapshot(parse_pkg_refs(c(
    "git::https://github.com/cran/falsy.git",
    "git::https://github.com/cran/falsy2.git"
  )))
})

test_that("resolve_remote_git", {
  skip_on_cran()
  setup_fake_apps()
  prop <- suppressMessages(new_pkg_installation_proposal(
    "git::https://github.com/r-lib/cli@v3.6.0",
    config = list(library = tempfile(), sysreqs_platform = "unknown")
  ))
  suppressMessages(prop$resolve())
  res <- prop$get_resolution()
  attr(res, "metadata")$resolution_start <- NULL
  attr(res, "metadata")$resolution_end <- NULL
  expect_snapshot(
    as.list(res),
    variant = paste0("pillar-", packageVersion("pillar"))
  )
})

test_that("download_remote_git", {
  skip_on_cran()
  setup_fake_apps()
  prop <- new_pkg_installation_proposal(
    "git::https://github.com/r-lib/cli@v3.6.0"
  )
  suppressMessages(prop$solve())
  suppressMessages(prop$download())
  expect_snapshot(
    dir(file.path(prop$get_downloads()$fulltarget_tree, "cli"))
  )
  expect_equal(
    desc::desc_get(
      "Version",
      file.path(prop$get_downloads()$fulltarget_tree, "cli")
    ),
    c(Version = "3.6.0")
  )
  expect_equal(prop$get_downloads()$download_status, "Got")
})

test_that("satisfy_remote_git", {
  expect_equal(
    satisfy_remote_git(list(package = "foo"), list(package = "bar")),
    structure(FALSE, reason = "Package names differ")
  )

  expect_equal(
    satisfy_remote_git(
      list(package = "foo", params = list(c("reinstall" = ""))),
      list(type = "installed", package = "foo")
    ),
    structure(FALSE, reason = "Re-install requested")
  )

  expect_equal(
    satisfy_remote_git(
      list(
        package = "foo",
        extra = list(list(remotesha = "badcafe"))
      ),
      list(
        type = "installed",
        package = "foo",
        extra = list(list(remotesha = "badbadcafe"))
      )
    ),
    structure(FALSE, reason = "Installed package sha mismatch")
  )

  expect_true(
    satisfy_remote_git(
      list(
        package = "foo",
        extra = list(list(remotesha = "badcafe"))
      ),
      list(
        type = "installed",
        package = "foo",
        extra = list(list(remotesha = "badcafe"))
      )
    )
  )

  expect_true(
    satisfy_remote_git(
      list(package = "foo"),
      list(type = "local", package = "foo")
    )
  )

  expect_equal(
    satisfy_remote_git(
      list(
        package = "foo",
        extra = list(list(remotesha = "badcafe"))
      ),
      list(
        type = "github",
        package = "foo",
        extra = list(list(remotesha = "badbadcafe"))
      )
    ),
    structure(FALSE, reason = "Candidate package sha mismatch")
  )

  expect_true(
    satisfy_remote_git(
      list(
        package = "foo",
        extra = list(list(remotesha = "badcafe"))
      ),
      list(
        type = "github",
        package = "foo",
        extra = list(list(remotesha = "badcafe"))
      )
    )
  )
})

test_that("installedok_remote_git", {
  expect_true(
    installedok_remote_git(
      list(package = "foo", version = "1.0.0", remotesha = "badcafe"),
      list(
        package = "foo",
        version = "1.0.0",
        metadata = list(c("RemoteSha" = "badcafe"))
      )
    )
  )

  expect_false(
    installedok_remote_git(
      list(package = "foo", version = "1.0.0", remotesha = "badcafe"),
      list(
        package = "foo",
        version = "1.0.1",
        metadata = list(c("RemoteSha" = "badbadcafe"))
      )
    )
  )

  expect_false(
    installedok_remote_git(
      list(package = "foo", version = "1.0.0", remotesha = "badcafe"),
      list(
        package = "foo",
        version = "1.0.0",
        metadata = list(c("RemoteSha" = "badbadcafe"))
      )
    )
  )
})

test_that("type_git_get_data", {
  # TODO
  expect_true(TRUE)
})

test_that("type_git_make_resolution", {
  # TODO
  expect_true(TRUE)
})

test_that("can package a tree", {
  skip_on_cran()
  local_cli_config()

  tmp <- withr::local_tempdir()
  file.copy(test_path("fixtures", "foo"), tmp, recursive = TRUE)

  dir.create(lib <- tempfile())
  on.exit(unlink(lib, recursive = TRUE), add = TRUE)

  plan <- data_frame(
    ref = paste0("local::./foo"),
    type = "local",
    direct = TRUE,
    status = "OK",
    package = "foo",
    version = "0.0.0.9000",
    binary = FALSE,
    packaged = FALSE,
    dependencies = list(character()),
    file = tmp,
    needscompilation = FALSE,
    install_args = ""
  )

  expect_snapshot(install_package_plan(plan, lib = lib))

  expect_true(file.exists(file.path(lib, "foo")))
  expect_true(file.exists(file.path(lib, "foo", "DESCRIPTION")))
  expect_true(file.exists(file.path(lib, "foo", "NAMESPACE")))
})

test_that("can package a compressed tree", {
  skip_on_cran()
  local_cli_config()

  tmp <- withr::local_tempdir()
  pkgzip <- file.path(tmp, "foo-t.zip")
  zip::zipr(pkgzip, test_path("fixtures", "foo"))

  dir.create(lib <- tempfile())
  on.exit(unlink(lib, recursive = TRUE), add = TRUE)

  plan <- data.frame(
    stringsAsFactors = FALSE,
    ref = paste0("local::./foo"),
    type = "local",
    direct = TRUE,
    status = "OK",
    package = "foo",
    version = "0.0.0.9000",
    binary = FALSE,
    packaged = FALSE,
    dependencies = I(list(character())),
    file = pkgzip,
    vignettes = FALSE,
    needscompilation = FALSE,
    metadata = I(list(character()))
  )

  expect_snapshot(install_package_plan(plan, lib = lib))

  expect_true(file.exists(file.path(lib, "foo")))
  expect_true(file.exists(file.path(lib, "foo", "DESCRIPTION")))
  expect_true(file.exists(file.path(lib, "foo", "NAMESPACE")))
})


test_that("can package a source package", {
  skip_on_cran()
  local_cli_config()

  tmp <- withr::local_tempdir()
  pkg <- source_test_package("foo")

  dir.create(lib <- tempfile())
  on.exit(unlink(lib, recursive = TRUE), add = TRUE)

  plan <- data.frame(
    stringsAsFactors = FALSE,
    ref = paste0("local::./foo"),
    type = "local",
    direct = TRUE,
    status = "OK",
    package = "foo",
    version = "0.0.0.9000",
    binary = FALSE,
    dependencies = I(list(character())),
    file = pkg,
    vignettes = FALSE,
    needscompilation = FALSE,
    metadata = I(list(character()))
  )

  expect_snapshot(install_package_plan(plan, lib = lib))

  expect_true(file.exists(file.path(lib, "foo")))
  expect_true(file.exists(file.path(lib, "foo", "DESCRIPTION")))
  expect_true(file.exists(file.path(lib, "foo", "NAMESPACE")))
  expect_true(file.exists(file.path(lib, "foo", "installed-file")))
})

test_that("add_recursive_dependencies", {
  plan <- data_frame(
    package = c("p1", "p2", "p3"),
    type = "cran",
    binary = TRUE,
    dependencies = list("p2", "p3", character()),
    file = NA_character_,
    needscompilation = FALSE
  )

  expect_equal(add_recursive_dependencies(plan), plan)

  plan$binary <- FALSE
  expect_equal(add_recursive_dependencies(plan), plan)

  plan$binary <- c(FALSE, TRUE, FALSE)
  expect_snapshot(add_recursive_dependencies(plan))
  expect_snapshot(add_recursive_dependencies(plan)$dependencies)
})

test_that("ignore-build-errors parameter", {
  setup_fake_apps()
  local_cli_config()
  dir.create(tmplib <- tempfile())
  on.exit(rimraf(tmplib), add = TRUE)
  pkgdir1 <- test_path("fixtures", "packages", "badbuild")
  pkgdir2 <- test_path("fixtures", "packages", "goodbuild")
  inst <- new_pkg_installation_proposal(
    paste0("local::", c(pkgdir1, pkgdir2), "?nocache"),
    config = list(library = tmplib, platforms = "source")
  )
  expect_snapshot(
    {
      suppressMessages(inst$solve())
      suppressMessages(inst$download())
      inst$install()
    },
    error = TRUE
  )

  inst <- new_pkg_installation_proposal(
    paste0("local::", c(pkgdir1, pkgdir2), "?nocache&ignore-build-errors"),
    config = list(library = tmplib, platforms = "source")
  )
  expect_snapshot({
    suppressMessages(inst$solve())
    suppressMessages(inst$download())
    inst$install()
  })
})

test_that("install package from GH", {
  setup_fake_apps()
  setup_fake_gh_app()
  pkgcache::pkg_cache_delete_files()

  lib <- withr::local_tempdir()
  config <- list(library = lib)
  pkg <- new_pkg_installation_proposal("r-lib/crayon", config = config)
  suppressMessages(pkg$solve())
  suppressMessages(pkg$download())
  suppressMessages(pkg$install())
  expect_true(file.exists(file.path(lib, "crayon")))

  # check cache state, must have a tree, a source and a binary package
  check_cache <- function() {
    cache <- pkgcache::pkg_cache_list()
    expect_equal(nrow(cache), 3L)
    expect_true(any(cache$package == "crayon" & !cache$built))
    expect_true(any(
      cache$package == "crayon" & cache$built & cache$platform == "source"
    ))
    expect_true(any(
      cache$package == "crayon" & cache$built & cache$platform != "source"
    ))
  }
  check_cache()

  # install from cache, binary is selected
  remove.packages("crayon", lib = lib)
  pkg <- new_pkg_installation_proposal("r-lib/crayon", config = config)
  suppressMessages(pkg$solve())
  suppressMessages(pkg$download())
  expect_equal(
    pkg$get_downloads()$download_status,
    paste("Had", current_r_platform())
  )
  suppressMessages(pkg$install())
  expect_true(file.exists(file.path(lib, "crayon")))

  # cache is updated with the binary
  check_cache()

  # install from cache, no binary, source package is selected
  remove.packages("crayon", lib = lib)
  pkgcache::pkg_cache_delete_files(
    built = TRUE,
    platform = current_r_platform()
  )
  pkg <- new_pkg_installation_proposal("r-lib/crayon", config = config)
  suppressMessages(pkg$solve())
  suppressMessages(pkg$download())
  expect_equal(pkg$get_downloads()$download_status, "Had")
  suppressMessages(pkg$install())
  expect_true(file.exists(file.path(lib, "crayon")))

  # install from cache, no binary, no source, tree is selected
  remove.packages("crayon", lib = lib)
  pkgcache::pkg_cache_delete_files(built = TRUE)
  pkg <- new_pkg_installation_proposal("r-lib/crayon", config = config)
  suppressMessages(pkg$solve())
  suppressMessages(pkg$download())
  expect_equal(pkg$get_downloads()$download_status, "Had")
  suppressMessages(pkg$install())
  expect_true(file.exists(file.path(lib, "crayon")))

  # cache is updated with the source and binary
  check_cache()
})

test_that("install package from GH, in subdir", {
  setup_fake_apps()
  setup_fake_gh_app()
  pkgcache::pkg_cache_delete_files()

  lib <- withr::local_tempdir()
  config <- list(library = lib)
  pkg <- new_pkg_installation_proposal("wesm/feather/R", config = config)
  suppressMessages(pkg$solve())
  suppressMessages(pkg$download())
  suppressMessages(pkg$install())
  expect_true(file.exists(file.path(lib, "feather")))

  # check cache state, must have a tree, a source and a binary package
  check_cache <- function() {
    cache <- pkgcache::pkg_cache_list()
    expect_equal(nrow(cache), 3L)
    expect_true(any(cache$package == "feather" & !cache$built))
    expect_true(any(
      cache$package == "feather" & cache$built & cache$platform == "source"
    ))
    expect_true(any(
      cache$package == "feather" & cache$built & cache$platform != "source"
    ))
  }
  check_cache()

  # install from cache, binary is selected
  remove.packages("feather", lib = lib)
  pkg <- new_pkg_installation_proposal("wesm/feather/R", config = config)
  suppressMessages(pkg$solve())
  suppressMessages(pkg$download())
  expect_equal(
    pkg$get_downloads()$download_status,
    paste("Had", current_r_platform())
  )
  suppressMessages(pkg$install())
  expect_true(file.exists(file.path(lib, "feather")))

  # cache is updated with the binary
  check_cache()

  # install from cache, no binary, source package is selected
  remove.packages("feather", lib = lib)
  pkgcache::pkg_cache_delete_files(
    built = TRUE,
    platform = current_r_platform()
  )
  pkg <- new_pkg_installation_proposal("wesm/feather/R", config = config)
  suppressMessages(pkg$solve())
  suppressMessages(pkg$download())
  expect_equal(pkg$get_downloads()$download_status, "Had")
  suppressMessages(pkg$install())
  expect_true(file.exists(file.path(lib, "feather")))

  # install from cache, no binary, no source, tree is selected
  remove.packages("feather", lib = lib)
  pkgcache::pkg_cache_delete_files(built = TRUE)
  pkg <- new_pkg_installation_proposal("wesm/feather/R", config = config)
  suppressMessages(pkg$solve())
  suppressMessages(pkg$download())
  expect_equal(pkg$get_downloads()$download_status, "Had")
  suppressMessages(pkg$install())
  expect_true(file.exists(file.path(lib, "feather")))

  # cache is updated with the source and binary
  check_cache()
})

test_that("binary preferred over source", {
  repo <- dcf(
    "
    Package: pkg
    Imports: pkg2

    Package: pkg2
  "
  )

  setup_fake_apps(
    cran_repo = repo,
    cran_options = list(platforms = c("windows", "source"))
  )

  p <- suppressMessages(new_pkg_installation_proposal(
    "pkg",
    config = list(
      dependencies = TRUE,
      library = tempfile(),
      platforms = c("x86_64-w64-mingw32", "source")
    )
  ))
  suppressMessages(p$resolve())
  suppressMessages(p$solve())
  sol <- p$get_solution()$data
  sol <- sol[order(sol$ref), ]
  expect_snapshot(sol[, c("package", "platform")])
})

test_that("but source is used if that version is required", {
  # we still need this to have a Bioc repo
  setup_fake_apps()
  repo1 <- dcf(
    "
    Package: pkg
    Imports: pkg2 (>= 2.0.0)

    Package: pkg2
    Version: 1.0.0
  "
  )

  repo2 <- dcf(
    "
   Package: pkg2
   Version: 2.0.0
  "
  )

  fake1 <- webfakes::new_app_process(
    cran_app(repo1, options = list(platforms = c("windows", "source")))
  )
  fake2 <- webfakes::new_app_process(
    cran_app(repo2, options = list(platforms = "source"))
  )
  withr::local_options(repos = c(CRAN = fake1$url(), X = fake2$url()))

  p <- suppressMessages(new_pkg_installation_proposal(
    "pkg",
    config = list(
      dependencies = TRUE,
      library = tempfile(),
      platforms = c("x86_64-w64-mingw32", "source")
    )
  ))
  suppressMessages(p$resolve())
  suppressMessages(p$solve())
  sol <- p$get_solution()$data
  sol <- sol[order(sol$ref), ]
  expect_snapshot(sol[, c("package", "version", "platform")])
})

test_that("can opt in into including LinkingTo", {
  repo <- dcf(
    "
    Package: pkg
    LinkingTo: pkg2

    Package: pkg2
  "
  )

  setup_fake_apps(
    cran_repo = repo,
    cran_options = list(platforms = c("windows", "source"))
  )

  do <- function() {
    p <- suppressMessages(new_pkg_installation_proposal(
      "pkg",
      config = list(
        dependencies = TRUE,
        library = tempfile(),
        platforms = c("x86_64-w64-mingw32", "source")
      )
    ))
    suppressMessages(p$resolve())
    suppressMessages(p$solve())
    sol <- p$get_solution()$data
    sol <- sol[order(sol$ref), ]
    expect_snapshot(sol[, c("package", "platform")])
  }

  withr::local_options(pkg.include_linkingto = TRUE)
  do()

  withr::local_options(pkg.include_linkingto = FALSE)
  do()
})

test_that("parse_remote_any", {
  setup_fake_apps()
  pkgcache::pkg_cache_delete_files()
  lib <- withr::local_tempdir()

  # create a lockfile first
  plan <- suppressMessages(new_pkg_installation_proposal(
    "any::pkg3",
    config = list(library = lib)
  ))
  suppressMessages(plan$resolve())
  plan$solve()
  lockfile <- tempfile(fileext = ".lock")
  on.exit(unlink(lockfile), add = TRUE)
  plan$create_lockfile(lockfile)

  plan <- suppressMessages(new_pkg_installation_proposal(
    "any::pkg3",
    config = list(library = lib)
  ))
  suppressMessages(plan$resolve())
  plan$solve()
  expect_snapshot(plan$draw(), transform = transform_bytes)
  suppressMessages(plan$download())
  suppressMessages(plan$install())

  # do it again, should keep installed package
  plan <- suppressMessages(new_pkg_installation_proposal(
    "any::pkg3",
    config = list(library = lib)
  ))
  suppressMessages(plan$resolve())
  plan$solve()
  expect_snapshot(plan$draw(), transform = transform_bytes)
  suppressMessages(plan$download())
  suppressMessages(plan$install())

  # Now check the lockfile, it should not install anything
  plan <- pkgdepends::new_pkg_installation_plan(
    lockfile,
    config = list(library = lib)
  )

  suppressMessages(plan$update())
  suppressMessages(plan$download())
  suppressMessages(plan$install())
})

test_that("dep_types", {
  setup_fake_apps()
  pkgcache::pkg_cache_delete_files()
  lib <- withr::local_tempdir()

  plan <- suppressMessages(new_pkg_installation_proposal(
    c("any::pkg3", "pkg3"),
    config = list(library = lib, dependencies = TRUE)
  ))
  suppressMessages(plan$resolve())
  plan$solve()
  res <- plan$get_resolution()
  res <- res[order(res$ref), ]
  expect_snapshot(as.list(res[c("ref", "dep_types")]))
  suppressMessages(plan$download())
  suppressMessages(plan$install())

  plan <- suppressMessages(new_pkg_installation_proposal(
    c("any::pkg3", "pkg3"),
    config = list(library = lib, dependencies = TRUE)
  ))
  suppressMessages(plan$resolve())
  plan$solve()

  res <- plan$get_resolution()
  res <- res[order(res$ref), ]
  res$ref <- sub("installed::.*pkg", "installed::<path>/pkg", res$ref)
  expect_snapshot(as.list(res[c("ref", "dep_types")]))
})

test_that("new_pkg_installation_plan", {
  setup_fake_apps()
  pkgcache::pkg_cache_delete_files()
  lib <- withr::local_tempdir()
  lockfile <- tempfile()
  on.exit(unlink(lockfile), add = TRUE)

  config <- list(library = lib)
  prop <- pst(new_pkg_installation_proposal("pkg3", config = config))
  pst(prop$solve())

  prop$create_lockfile(lockfile)

  plan <- new_pkg_installation_plan(lockfile, config = config)
  expect_snapshot(plan)

  plan$update()

  expect_snapshot(error = TRUE, {
    plan$resolve()
    plan$async_resolve()
  })
  expect_equal(plan$get_solve_policy(), NA_character_)
  expect_snapshot(error = TRUE, {
    plan$set_solve_policy()
    plan$solve()
  })
})

test_that("install package from GH, in subdir", {
  setup_fake_apps()
  setup_fake_gh_app()
  pkgcache::pkg_cache_delete_files()

  lib <- withr::local_tempdir()
  config <- list(library = lib)
  pkg <- new_pkg_installation_proposal("wesm/feather/R", config = config)
  suppressMessages(pkg$solve())
  lock <- file.path(withr::local_tempdir(), "pkg.lock")
  suppressMessages(pkg$create_lockfile(lock))

  plan <- new_pkg_installation_plan(lock, config = config)
  suppressMessages(plan$download())
  suppressMessages(plan$install())
  expect_true(file.exists(file.path(lib, "feather")))

  # check cache state, must have a tree, a source and a binary package
  check_cache <- function() {
    cache <- pkgcache::pkg_cache_list()
    expect_equal(nrow(cache), 3L)
    expect_true(any(cache$package == "feather" & !cache$built))
    expect_true(any(
      cache$package == "feather" & cache$built & cache$platform == "source"
    ))
    expect_true(any(
      cache$package == "feather" & cache$built & cache$platform != "source"
    ))
  }
  check_cache()

  # install from cache, binary is selected
  remove.packages("feather", lib = lib)
  plan <- new_pkg_installation_plan(lock, config = config)
  suppressMessages(plan$download())
  expect_equal(
    plan$get_downloads()$download_status,
    paste("Had", current_r_platform())
  )
  suppressMessages(plan$install())
  expect_true(file.exists(file.path(lib, "feather")))

  # cache is updated with the binary
  check_cache()

  # install from cache, no binary, source package is selected
  remove.packages("feather", lib = lib)
  pkgcache::pkg_cache_delete_files(
    built = TRUE,
    platform = current_r_platform()
  )
  plan <- new_pkg_installation_plan(lock, config = config)
  suppressMessages(plan$download())
  expect_equal(plan$get_downloads()$download_status, "Had")
  suppressMessages(plan$install())
  expect_true(file.exists(file.path(lib, "feather")))

  # install from cache, no binary, no source, tree is selected
  remove.packages("feather", lib = lib)
  pkgcache::pkg_cache_delete_files(built = TRUE)
  plan <- new_pkg_installation_plan(lock, config = config)
  suppressMessages(plan$download())
  expect_equal(plan$get_downloads()$download_status, "Had")
  suppressMessages(plan$install())
  expect_true(file.exists(file.path(lib, "feather")))

  # cache is updated with the source and binary
  check_cache()
})

test_that("sysreqs", {
  # Not the best test, because there are no system requirements to update...
  setup_fake_apps()
  pkgcache::pkg_cache_delete_files()
  lib <- withr::local_tempdir()
  lockfile <- tempfile()
  on.exit(unlink(lockfile), add = TRUE)

  config <- list(library = lib)
  prop <- pst(new_pkg_installation_proposal("pkg3", config = config))
  pst(prop$solve())

  prop$create_lockfile(lockfile)

  plan <- new_pkg_installation_plan(lockfile, config = config)
  expect_snapshot(plan)

  plan$update()
  plan$update_sysreqs()

  expect_snapshot(error = TRUE, {
    plan$resolve()
    plan$async_resolve()
  })
  expect_equal(plan$get_solve_policy(), NA_character_)
  expect_snapshot(error = TRUE, {
    plan$set_solve_policy()
    plan$solve()
  })
})

test_that("install_sysreqs", {
  setup_fake_apps()
  dir.create(lib <- tempfile())
  on.exit(unlink(lib, recursive = TRUE), add = TRUE)

  config <- list(
    library = lib,
    sysreqs = TRUE,
    sysreqs_platform = "aarch64-unknown-linux-gnu-ubuntu-22.04",
    sysreqs_lookup_system = FALSE,
    sysreqs_dry_run = TRUE,
    sysreqs_sudo = FALSE
  )
  prop <- new_pkg_installation_proposal("curl", config = config)
  suppressMessages(prop$solve())
  lock <- tempfile(fileext = ".lock")
  on.exit(unlink(lock), add = TRUE)
  prop$create_lockfile(path = lock)

  plan <- new_pkg_installation_plan(lock)
  expect_snapshot(
    {
      plan$show_solution()
      plan$show_sysreqs()
    },
    transform = transform_bytes
  )
})

test_that("update_sysreqs", {
  setup_fake_apps()
  dir.create(lib <- tempfile())
  on.exit(unlink(lib, recursive = TRUE), add = TRUE)

  config <- list(
    library = lib,
    sysreqs = TRUE,
    sysreqs_platform = "aarch64-unknown-linux-gnu-ubuntu-22.04",
    sysreqs_lookup_system = FALSE,
    sysreqs_dry_run = TRUE,
    sysreqs_sudo = FALSE
  )
  prop <- new_pkg_installation_proposal("curl", config = config)
  suppressMessages(prop$solve())
  lock <- tempfile(fileext = ".lock")
  on.exit(unlink(lock), add = TRUE)
  prop$create_lockfile(path = lock)

  plan <- new_pkg_installation_plan(lock, config = config)
  iplan <- get_private(plan)$plan

  # inject system package data
  fake <- data_frame(
    status = "ii",
    package = c("libssl-dev", "libssl3"),
    version = "3.0.2-0ubuntu1.10",
    provides = list(list())
  )
  old <- iplan$.__enclos_env__$private$solution$result$data$sysreqs_packages
  new <- sysreqs_update_state(old, fake)
  iplan$.__enclos_env__$private$solution$result$data$sysreqs_packages <- new
  expect_snapshot({
    plan$show_sysreqs()
  })
})

test_that("update_sysreqs with old lock file", {
  setup_fake_apps()
  dir.create(lib <- tempfile())
  on.exit(unlink(lib, recursive = TRUE), add = TRUE)

  # does nothing if no sysreqs_packages
  config <- list(
    library = lib,
    sysreqs = FALSE,
    sysreqs_platform = "unknown",
    sysreqs_lookup_system = FALSE
  )
  prop <- new_pkg_installation_proposal("curl", config = config)
  suppressMessages(prop$solve())
  lock <- tempfile(fileext = ".lock")
  on.exit(unlink(lock), add = TRUE)
  prop$create_lockfile(path = lock)

  config$sysreqs <- TRUE
  plan <- new_pkg_installation_plan(lock, config = config)
  iplan <- get_private(plan)$plan

  pkgplan_update_sysreqs(iplan, get_private(iplan))
  expect_snapshot({
    plan$show_sysreqs()
  })
})

test_that("new_pkg_installation_proposal", {
  setup_fake_apps()
  pkgcache::pkg_cache_delete_files()

  prop <- suppressMessages(new_pkg_installation_proposal(
    "pkg3",
    config = list(library = tempfile())
  ))

  expect_snapshot(prop)
  expect_snapshot(prop$get_refs())
  expect_snapshot(sort(prop$get_config()$list()))

  suppressMessages(prop$resolve())
  expect_snapshot(
    prop$get_resolution()[, c("ref", "type", "directpkg", "package", "error")]
  )

  expect_snapshot(prop$get_solve_policy())
  prop$set_solve_policy("upgrade")
  expect_snapshot(prop$get_solve_policy())

  suppressMessages(prop$solve())
  expect_silent(prop$stop_for_solution_error())
  expect_snapshot(prop$get_solution())
  expect_snapshot(
    prop$show_solution(),
    transform = transform_bytes
  )

  expect_snapshot(
    prop$draw(),
    transform = transform_bytes
  )
})

test_that("async_resolve", {
  setup_fake_apps()
  pkgcache::pkg_cache_delete_files()

  prop <- suppressMessages(new_pkg_installation_proposal(
    "pkg3",
    config = list(library = tempfile())
  ))

  expect_snapshot(
    suppressMessages(synchronize(
      prop$async_resolve()$then(function() "done")
    ))
  )
  expect_snapshot(
    prop$get_resolution()[, c("ref", "type", "directpkg", "package", "error")]
  )
})

test_that("create_lockfile", {
  setup_fake_apps()
  pkgcache::pkg_cache_delete_files()

  prop <- suppressMessages(new_pkg_installation_proposal(
    "pkg3",
    config = list(library = tempfile())
  ))

  suppressMessages(prop$solve())
  tmp <- tempfile(fileext = ".lock")
  on.exit(unlink(tmp), add = TRUE)

  suppressMessages(prop$create_lockfile(tmp))
  expect_error(lock <- jsonlite::fromJSON(tmp), NA)

  lock$packages$filesize <- 100

  # Newer windows adds a --no-multiarch install parameter, by default
  vrt <- if (.Platform$OS.type == "windows" && getRversion() >= "4.2.0") {
    "windows"
  } else {
    "unix"
  }
  expect_snapshot(
    lock$packages,
    transform = function(x) transform_local_port(transform_sha(x)),
    variant = vrt
  )
})

test_that("download", {
  setup_fake_apps()
  pkgcache::pkg_cache_delete_files()

  prop <- suppressMessages(new_pkg_installation_proposal(
    "pkg3",
    config = list(library = tempfile())
  ))

  suppressMessages(prop$resolve())
  suppressMessages(prop$solve())
  suppressMessages(prop$download())
  expect_snapshot(
    prop$get_downloads()[, c("target", "error")]
  )
  expect_silent(prop$stop_for_download_error())
})

test_that("async_download", {
  setup_fake_apps()
  pkgcache::pkg_cache_delete_files()

  prop <- suppressMessages(new_pkg_installation_proposal(
    "pkg3",
    config = list(library = tempfile())
  ))

  suppressMessages(prop$resolve())
  suppressMessages(prop$solve())
  suppressMessages(synchronize(prop$async_download()))
  expect_snapshot(
    prop$get_downloads()[, c("target", "error")]
  )
  expect_silent(prop$stop_for_download_error())
})

test_that("install", {
  setup_fake_apps()
  pkgcache::pkg_cache_delete_files()

  lib <- withr::local_tempdir()
  prop <- suppressMessages(new_pkg_installation_proposal(
    "pkg3",
    config = list(library = lib)
  ))

  suppressMessages(prop$resolve())
  suppressMessages(prop$solve())
  suppressMessages(prop$download())
  suppressMessages(prop$install())

  expect_snapshot(
    pkgcache::parse_installed(lib)[, c("Package", "RemoteRepos")],
    transform = transform_local_port
  )
})

test_that("get_install_plan", {
  setup_fake_apps()
  pkgcache::pkg_cache_delete_files()

  lib <- withr::local_tempdir()
  prop <- suppressMessages(new_pkg_installation_proposal(
    "pkg3",
    config = list(library = lib)
  ))

  suppressMessages(prop$resolve())
  suppressMessages(prop$solve())
  suppressMessages(prop$download())
  expect_snapshot(
    prop$get_install_plan()[, c("package", "direct", "dependencies")]
  )
  expect_snapshot(
    prop$get_install_plan()[["dependencies"]]
  )
})

test_that("show_sysreqs", {
  setup_fake_apps()
  dir.create(lib <- tempfile())
  on.exit(unlink(lib, recursive = TRUE), add = TRUE)

  config <- list(
    library = lib,
    sysreqs = TRUE,
    sysreqs_platform = "aarch64-unknown-linux-gnu-ubuntu-22.04",
    sysreqs_lookup_system = FALSE
  )
  prop <- new_pkg_installation_proposal("curl", config = config)
  suppressMessages(prop$solve())
  expect_snapshot({
    prop$show_sysreqs()
  })
})

test_that("install_sysreqs", {
  setup_fake_apps()
  withr::local_envvar(CI = NA_character_)
  dir.create(lib <- tempfile())
  on.exit(unlink(lib, recursive = TRUE), add = TRUE)

  config <- list(
    library = lib,
    sysreqs = TRUE,
    sysreqs_platform = "aarch64-unknown-linux-gnu-ubuntu-22.04",
    sysreqs_lookup_system = FALSE,
    sysreqs_dry_run = TRUE,
    sysreqs_sudo = FALSE
  )
  prop <- new_pkg_installation_proposal("curl", config = config)
  suppressMessages(prop$solve())
  expect_snapshot({
    prop$install_sysreqs()
  })

  config$sysreqs <- FALSE
  prop <- new_pkg_installation_proposal("curl", config = config)
  suppressMessages(prop$solve())
  expect_snapshot({
    prop$install_sysreqs()
  })

  config$sysreqs <- TRUE
  prop <- new_pkg_installation_proposal("pkg1", config = config)
  suppressMessages(prop$solve())
  expect_snapshot({
    prop$install_sysreqs()
  })
})

test_that("resolve", {
  skip_on_cran()
  tmp <- tempfile()
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  p <- suppressMessages(new_pkg_installation_proposal(
    "gitlab::gaborcsardi/cli",
    config = list(library = tmp)
  ))
  suppressMessages(p$resolve())
  res <- p$get_resolution()
  expect_snapshot({
    res$error
    res$package
    res$version
    res$metadata[[1]]
  })

  # group + branch
  p <- suppressMessages(new_pkg_installation_proposal(
    "gitlab::r-hub/filelock@cran-1-0-2",
    config = list(library = tmp)
  ))
  suppressMessages(p$resolve())
  res <- p$get_resolution()
  expect_snapshot({
    res$error
    res$package
    res$version
    res$metadata[[1]]
  })

  # tag
  p <- suppressMessages(new_pkg_installation_proposal(
    "gitlab::r-hub/filelock@v1.0.2",
    config = list(library = tmp)
  ))
  suppressMessages(p$resolve())
  res <- p$get_resolution()
  expect_snapshot({
    res$error
    res$package
    res$version
    res$metadata[[1]]
  })

  # subdirectory
  p <- suppressMessages(new_pkg_installation_proposal(
    "gitlab::gaborcsardi/feather/-/R",
    config = list(library = tmp, dependencies = FALSE)
  ))
  suppressMessages(p$resolve())
  res <- p$get_resolution()
  expect_snapshot({
    res$error
    res$package
    res$version
    res$metadata[[1]]
  })
})

test_that("download", {
  skip_on_cran()
  tmp <- tempfile()
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  # subdirectory
  p <- suppressMessages(new_pkg_installation_proposal(
    "gitlab::gaborcsardi/feather/-/R",
    config = list(library = tmp, dependencies = FALSE)
  ))
  suppressMessages(p$resolve())
  res <- p$get_resolution()
  suppressMessages(p$solve())
  suppressMessages(p$download())
  dl <- p$get_downloads()
  expect_true(dir.exists(dl$fulltarget_tree))
  expect_snapshot(
    dir(file.path(dl$fulltarget_tree, "feather", "R"))
  )
})

test_that("satisfy", {
  expect_true(
    satisfy_remote_gitlab(
      list(
        package = "foo",
        extra = list(list(remotesha = "badcafe"))
      ),
      list(
        type = "gitlab",
        package = "foo",
        extra = list(list(remotesha = "badcafe"))
      )
    )
  )
})

test_that("installedok", {
  expect_true(
    installedok_remote_gitlab(
      list(package = "foo", version = "1.0.0", remotesha = "badcafe"),
      list(
        package = "foo",
        version = "1.0.0",
        metadata = list(c("RemoteSha" = "badcafe"))
      )
    )
  )
})

test_that("R version dependencies", {
  setup_fake_gh_app()
  setup_fake_apps()

  lib <- withr::local_tempdir()
  p1 <- new_pkg_installation_proposal(
    "futurama",
    config = list(library = lib)
  )
  suppressMessages(p1$solve())
  expect_snapshot(
    error = TRUE,
    p1$stop_for_solution_error()
  )

  p2 <- new_pkg_installation_proposal(
    "needsfuturama",
    config = list(library = lib)
  )
  suppressMessages(p2$solve())
  expect_snapshot(
    error = TRUE,
    p2$stop_for_solution_error()
  )
})

test_that("conflicting dependencies", {
  pkgs <- make_fake_resolution(
    `aa` = list(
      direct = TRUE,
      deps = list(make_fake_deps(Imports = "cc", Remotes = "cran::cc"))
    ),
    `bb` = list(
      direct = TRUE,
      deps = list(make_fake_deps(Imports = "cc", Remotes = "cc/cc"))
    ),
    `cran::cc` = list(),
    `cc/cc` = list(extra = list(list(remotesha = "badcafe")))
  )
  dsc <- describe_fake_error(pkgs)
  expect_equal(dsc$ref, c("aa", "cran::cc"))
  expect_equal(dsc$type, c("standard", "cran"))
  expect_equal(dsc$direct, c(TRUE, FALSE))
  expect_equal(dsc$status, c("OK", "OK"))
  expect_equal(dsc$package, c("aa", "cc"))
  expect_equal(dsc$failure_type, c("dep-failed", "conflict"))
  expect_snapshot(dsc)
})

test_that("conflicting dependencies downstream", {
  pkgs <- make_fake_resolution(
    `a0` = list(direct = TRUE, deps = list(make_fake_deps(Imports = "aa, bb"))),
    `aa` = list(
      direct = TRUE,
      deps = list(make_fake_deps(Imports = "cc", Remotes = "cran::cc"))
    ),
    `bb` = list(
      direct = TRUE,
      deps = list(make_fake_deps(Imports = "cc", Remotes = "cc/cc"))
    ),
    `cran::cc` = list(),
    `cc/cc` = list(extra = list(list(sha = "badcafe")))
  )
  dsc <- describe_fake_error(pkgs)
  expect_equal(dsc$ref, c("a0", "bb", "cc/cc"))
  expect_equal(dsc$type, c("standard", "standard", "github"))
  expect_equal(dsc$direct, c(TRUE, TRUE, FALSE))
  expect_equal(dsc$status, c("OK", "OK", "OK"))
  expect_equal(dsc$package, c("a0", "bb", "cc"))
  expect_equal(dsc$failure_type, c("dep-failed", "dep-failed", "conflict"))
  expect_snapshot(dsc)
})

test_that("conflicting dependencies and installed packages", {
  gh <- list(
    users = list(
      "cran" = list(
        repos = list(
          dm = list(
            commits = list(
              list(
                sha = "96dac7a1cffea51709fa3c53c3424d4047ec279286583b8e1e46c701e98d0954",
                branch = "main",
                tag = "1.0.0",
                files = list(
                  DESCRIPTION = "Package: dm\nVersion: 1.0.0\nImports: tidyr\n",
                  NAMESPACE = ""
                )
              )
            )
          ),
          dplyr = list(
            commits = list(
              list(
                sha = "b259bcc4a82d28effdc59d3d51d4976b32ce29056ef9da48eeb9424e6eaffa16",
                branch = "main",
                tag = "1.0.9",
                files = list(
                  DESCRIPTION = "Package: dplyr\nVersion: 1.0.9\n"
                )
              )
            )
          )
        )
      )
    )
  )

  cran <- dcf(
    "
    Package: tidyr
    Version: 1.0.0
    Imports: dplyr (>= 1.0.10)

    Package: dplyr
    Version: 1.1.1

    Package: foo
    Imports: bar

    Package: bar
  "
  )

  fake_gh <- webfakes::local_app_process(gh_app(gh))
  withr::local_envvar(R_PKG_GITHUB_API_URL = fake_gh$url())
  setup_fake_apps(cran_repo = cran)

  # -----------------------------------------------------------------------

  lib <- withr::local_tempdir()

  p <- new_pkg_installation_proposal(
    c("cran/dm@1.0.0", "cran/dplyr@1.0.9"),
    config = list(library = lib)
  )
  suppressMessages(p$resolve())
  suppressMessages(p$solve())
  expect_snapshot(p$get_solution()$failures)

  # -----------------------------------------------------------------------
  # fake an installed package

  dir.create(file.path(lib, "tidyr"))
  writeLines(
    c(
      "Package: tidyr",
      "Version: 1.0.0",
      "Imports: dplyr (>= 1.0.10)",
      "Repository: CRAN",
      # need this on older windows, because pak matches archs on that
      if (is_windows()) {
        paste0(
          "Built: R ",
          getRversion(),
          "; ",
          "i386+x86_64-w64-mingw32; ",
          "2023-02-06 08:15:41 UTC; ",
          "windows"
        )
      }
    ),
    file.path(lib, "tidyr", "DESCRIPTION")
  )
  p <- new_pkg_installation_proposal(
    c("cran/dm@1.0.0", "cran/dplyr@1.0.9"),
    config = list(library = lib)
  )
  suppressMessages(p$resolve())
  suppressMessages(p$solve())
  expect_snapshot(
    p$get_solution()$failures,
    transform = function(x) {
      sub("installed::.*tidyr", "installed::<path>/tidyr", x)
    }
  )

  # -----------------------------------------------------------------------
  # cannot install anything, accoridng to the solver...

  local <- withr::local_tempdir()
  writeLines(
    c(
      "Package: local123",
      "Version: 1.0.0",
      "Imports: dm, dplyr",
      "Remotes: cran/dm@1.0.0, cran/dplyr@1.0.9"
    ),
    file.path(local, "DESCRIPTION")
  )

  # First without installed packages
  p <- new_pkg_installation_proposal(
    paste0("local::", local),
    config = list(library = lib)
  )
  suppressMessages(p$resolve())
  suppressMessages(p$solve())
  expect_snapshot(
    p$get_solution()$failures,
    transform = function(x) {
      sub(paste0("local::.*", basename(local)), "local::<path>/<pkg>", x)
    }
  )

  # -----------------------------------------------------------------------
  # cannot install anything for some direct refs

  p2 <- new_pkg_installation_proposal(
    c(paste0("local::", local), "foo"),
    config = list(library = lib)
  )
  suppressMessages(p2$resolve())
  suppressMessages(p2$solve())
  expect_snapshot(
    p2$get_solution()$failures,
    transform = function(x) {
      sub(paste0("local::.*", basename(local)), "local::<path>/<pkg>", x)
    }
  )
})

# -------------------------------------------------------------------------
#' Check if an R package name is available.
#'
#' Additionally, look up the candidate name in a number of
#' dictionaries, to make sure that it does not have a negative
#' meaning.
#'
#' ## Valid package name check
#'
#' Check the validity of `name` as a package name. See 'Writing R
#' Extensions' for the allowed package names. Also checked against a list
#' of names that are known to cause problems.
#'
#' ## CRAN checks
#'
#' Check `name` against the names of all past and current packages on
#' CRAN, including base and recommended packages.
#'
#' ## Bioconductor checks
#'
#' Check `name` against all past and current Bioconductor packages.
#'
#' ## Profanity check
#'
#' Check `name` with <https://www.purgomalum.com/service/containsprofanity>
#' to make sure it is not a profanity.
#'
#' ## Dictionaries
#'
#' See the `dictionaries` argument.
#'
#' @param name Package name candidate.
#' @param dictionaries Character vector, the dictionaries to query.
#'   Available dictionaries:
#'     * `wikipedia`
#'     * `wiktionary`,
#'     * `sentiment` (<https://github.com/fnielsen/afinn>),
#'     * `urban` (Urban Dictionary).
#'   If `NULL` (by default), the Urban Dictionary is omitted, as it
#'   is often offensive.
#' @return `pkg_name_check` object with a custom print method.
#'
#' @export
#' @examplesIf pkgdepends:::is_online()
#' pkg_name_check("cli")

pkg_name_check <- function(name, dictionaries = NULL) {
  synchronise(async_pkg_name_check(name, dictionaries))
}

async_pkg_name_check <- function(name, dictionaries = NULL) {
  assert_that(
    is_string(name),
    is.null(dictionaries) || is_character(dictionaries)
  )

  default_dictionaries <- c(
    "wikipedia",
    "wiktionary",
    "sentiment"
  )
  dicts <- dictionaries %||% default_dictionaries

  basics <- async_pnc_basics(name)
  result <- when_all(
    basics = basics,
    wikipedia = if ("wikipedia" %in% dicts) async_wikipedia_get(name),
    wiktionary = if ("wiktionary" %in% dicts) async_wiktionary_get(name),
    sentiment = if ("sentiment" %in% dicts) async_sentiment_get(name),
    urban = if ("urban" %in% dicts) async_urban_get(name)
  )$then(function(res) add_class(res, "pkg_name_check"))
}

#' @export

format.pkg_name_check <- function(x, limit = 6, ...) {
  x <- x[!vlapply(x, is.null)]
  unlist(lapply(x, format, limit = limit, ...))
}

#' @export

print.pkg_name_check <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

async_pnc_basics <- function(name) {
  when_all(
    name = name,
    valid = pnc_valid(name),
    base = pnc_base(name),
    crandb = async_pnc_crandb(name),
    bioc = async_pnc_bioc(name),
    profanity = async_profanity_get(name)
  )$then(function(res) add_class(res, "pkg_name_check_basics"))
}

# -------------------------------------------------------------------------

forbidden_package_names <- function() {
  c("pkg", "description")
}

pnc_valid <- function(name) {
  ans <- TRUE
  rx <- paste0("^", pkg_rx()$pkg_name, "$")
  if (!grepl(rx, name)) {
    ans <- FALSE
  }
  # This is not needed currently. But just in case character ranges will
  # accept non-ascii characters on some platforms, we keep it.
  if (ans && any(charToRaw(name) > 127)) {
    ans <- FALSE
  }
  if (ans && name %in% forbidden_package_names()) {
    ans <- FALSE
  }
  add_class(ans, "pkg_name_check_valid")
}

# -------------------------------------------------------------------------

pnc_base <- function(name) {
  bs <- base_packages()
  mch <- match(tolower(name), tolower(bs))
  add_class(
    list(base = is.na(mch), package = if (!is.na(mch)) bs[mch]),
    "pkg_name_check_base"
  )
}

# -------------------------------------------------------------------------
# TODO: this is not included currently. Do we need it?

async_cranlike_check <- function(name) {
  name

  repos <- c(
    CRAN = "https://cran.r-project.org",
    CRANextra = "https://www.stats.ox.ac.uk/pub/RWin"
  )

  meta <- pkgcache::cranlike_metadata_cache$new(
    platforms = "source",
    bioc = current_config()$get("use_bioconductor"),
    cran_mirror = "https://cran.r-project.org",
    repos = repos,
    update_after = as.difftime(5, units = "mins")
  )

  meta$async_check_update()$then(function(data) {
    mch <- match(tolower(name), tolower(data$pkgs$package))
    ret <- list(cran = TRUE, bioc = TRUE, package = NA_character_)
    if (!is.na(mch)) {
      ret$package <- data$pkgs$package[mch]
      type <- data$pkgs$type[mch]
      if (type == "cran") {
        ret$cran <- FALSE
      }
      if (type == "bioc") ret$bioc <- FALSE
    }
    ret
  })$then(function(res) add_class(res, "pkg_name_check_cranlike"))
}

# -------------------------------------------------------------------------

async_pnc_crandb <- function(name) {
  async_pnc_crandb_query(name)$then(pnc_crandb_process)
}

async_pnc_crandb_query <- function(name) {
  base <- Sys.getenv(
    "PKG_NAME_CHECK_CRANDB_URL",
    "https://crandb.r-pkg.org:2053/cran/_design/app/_view/alllower"
  )
  url <- paste0(base, "?key=%22", tolower(name), "%22")
  http_get(url)
}

pnc_crandb_process <- function(response) {
  http_stop_for_status(response)
  ans <- jsonlite::fromJSON(rawToChar(response$content), simplifyVector = FALSE)
  ret <- list(
    crandb = length(ans$rows) == 0,
    package = if (length(ans$rows) > 0) vcapply(ans$rows, "[[", "value")
  )
  add_class(ret, "pkg_name_check_crandb")
}

# -------------------------------------------------------------------------

#' @export

print.pkg_name_check_basics <- function(x, ...) {
  writeLines(format(x, ...))
  invisible(x)
}

#' @export

format.pkg_name_check_basics <- function(x, ...) {
  cw <- cli::console_width()
  stars <- cli::col_yellow(cli::symbol$en_dash, "*", cli::symbol$en_dash)
  title <- paste(stars, cli::style_bold(cli::col_blue(x$name)), stars)
  tbox <- cli::boxx(
    cli::ansi_align(title, width = cw - 4, align = "center"),
    padding = c(0, 1, 0, 1),
    border_style = "double",
    border_col = cli::col_yellow
  )

  ys <- cli::col_green(cli::symbol$tick)
  no <- cli::col_red(cli::symbol$cross)
  mark <- function(x) if (x) ys else no
  str <- c(
    paste(mark(x$valid), " valid name"),
    paste(mark(x$base$base && x$crandb$crandb), " CRAN"),
    paste(mark(x$bioc$bioc), " Bioconductor"),
    if ("profanity" %in% names(x)) {
      paste0(
        mark(!x$profanity),
        if (!x$profanity) "  not a" else " ",
        " profanity"
      )
    }
  )
  cols <- cli::ansi_columns(str, width = cw - 4, max_cols = length(str))
  bbox <- cli::boxx(cols, padding = c(0, 1, 0, 1), border_col = cli::col_silver)

  c(tbox, bbox)
}

# -------------------------------------------------------------------------

async_wikipedia_get <- function(terms) {
  async_wikipedia_get_query(terms)$then(http_stop_for_status)$then(
    function(resp) wikipedia_get_process(terms, resp)
  )$then(function(res) add_class(res, "pkg_name_check_wikipedia"))
}

async_wikipedia_get_query <- function(terms) {
  url <- Sys.getenv(
    "PKG_NAME_CHECK_WIKIPEDIA_URL",
    "https://en.wikipedia.org/w/api.php"
  )
  data <- make_wikipedia_data(terms)
  http_post(url, data = data)
}

make_wikipedia_data <- function(terms, intro = TRUE) {
  terms <- enc2utf8(as.character(terms))
  params <- c(
    action = "query",
    format = "json",
    prop = "extracts",
    titles = utils::URLencode(paste(terms, collapse = "|")),
    redirects = 1,
    exintro = if (intro) 1L,
    explaintext = 1
  )
  pstr <- paste0(names(params), "=", params, collapse = "&")
  charToRaw(pstr)
}

wikipedia_get_process <- function(
  terms,
  resp,
  base_url = "https://en.wikipedia.org/wiki/"
) {
  obj <- jsonlite::fromJSON(
    rawToChar(resp$content),
    simplifyVector = FALSE
  )

  map_norm <- structure(
    vcapply(obj$query$normalized, "[[", "to"),
    names = vcapply(obj$query$normalized, "[[", "from")
  )
  nterms <- ifelse(is.na(map_norm[terms]), terms, map_norm[terms])

  map_redir <- structure(
    vcapply(obj$query$redirects, "[[", "to"),
    names = vcapply(obj$query$redirects, "[[", "from")
  )
  rterms <- ifelse(is.na(map_redir[nterms]), nterms, map_redir[nterms])

  map_text <- structure(
    vcapply(obj$query$pages, function(x) x$extract %||% NA_character_),
    names = vcapply(obj$query$pages, "[[", "title")
  )
  text <- map_text[rterms]

  url <- paste0(base_url, gsub(" ", "_", rterms))

  data_frame(
    term = terms,
    normalized = unname(nterms),
    redirected = unname(rterms),
    title = unname(rterms),
    text = unname(text),
    url = ifelse(is.na(text), NA_character_, url)
  )
}

#' @export

print.pkg_name_check_wikipedia <- function(x, ...) {
  writeLines(format(x, ...))
  invisible(x)
}

#' @export

format.pkg_name_check_wikipedia <- function(x, limit = 6, ...) {
  x <- x[1, , drop = FALSE]

  if (is.na(x$text)) {
    x$text <- "No definition found"
    x$url <- ""
  }

  txt <- paste(
    cli::style_underline(x$redirected),
    if (x$redirected != x$normalized) paste0("(from ", x$normalized, ")"),
    clean_wikipedia_text(x$text)
  )

  cw <- cli::console_width()
  wrp <- cli::ansi_strwrap(txt, width = cw - 4)
  hdr <- cli::col_green("Wikipedia")
  ftr <- cli::col_blue(x$url)
  alg <- cli::ansi_align(wrp, width = cw - 4)
  if (length(alg) > limit) {
    alg <- c(alg[1:limit], cli::symbol$ellipsis)
  }
  cli::boxx(
    alg,
    padding = c(0, 1, 0, 1),
    header = hdr,
    footer = ftr,
    border_col = cli::col_silver
  )
}

clean_wikipedia_text <- function(x) {
  x <- trimws(x)
  x <- sub(
    "may refer to:$",
    "may refer to multiple articles, see link.",
    x
  )
  x <- gsub(" ()", "", x, fixed = TRUE)
  x <- gsub("\n", "\n\n", x, fixed = TRUE)
}

# -------------------------------------------------------------------------

async_wiktionary_get <- function(terms) {
  async_wiktionary_get_query(terms)$then(
    function(resp) wiktionary_get_process(terms, resp)
  )
}

async_wiktionary_get_query <- function(terms) {
  url <- Sys.getenv(
    "PKG_NAME_CHECK_WIKTIONARY_URL",
    "https://en.wiktionary.org/w/api.php"
  )
  data <- make_wikipedia_data(terms, intro = FALSE)
  http_post(url, data = data)
}

wiktionary_get_process <- function(terms, resp) {
  http_stop_for_status(resp)

  res <- wikipedia_get_process(
    terms,
    resp,
    base_url = "https://en.wiktionary.org/wiki/"
  )[, c("term", "text", "url")]

  add_class(res, "pkg_name_check_wiktionary")
}

#' @export

print.pkg_name_check_wiktionary <- function(x, ...) {
  writeLines(format(x, ...))
  invisible(x)
}

#' @export

format.pkg_name_check_wiktionary <- function(x, limit = 6, ...) {
  x <- x[1, , drop = FALSE]

  if (is.na(x$text)) {
    x$text <- "No definition found"
    x$url <- ""
  }

  txt <- paste(cli::style_underline(x$term), clean_wiktionary_text(x$text))

  cw <- cli::console_width()
  wrp <- cli::ansi_strwrap(txt, width = cw - 4)
  wrp <- wrp[cli::ansi_strip(wrp) != ""]
  if (length(wrp) > limit) {
    wrp <- c(wrp[1:limit], cli::symbol$ellipsis)
  }
  hdr <- cli::col_green("Wiktionary")
  ftr <- cli::col_blue(x$url)
  alg <- cli::ansi_align(wrp, width = cw - 4)
  cli::boxx(
    alg,
    padding = c(0, 1, 0, 1),
    header = hdr,
    footer = ftr,
    border_col = cli::col_silver
  )
}

clean_wiktionary_text <- function(x) {
  langs <- strsplit(x, "\n== ", fixed = TRUE)[[1]][-1]
  eng <- grep("^English", langs, value = TRUE)
  if (length(eng) == 0) {
    return("No English definition found")
  }
  eng2 <- sub("^English ==", "", eng)
  # remove pronunciation
  eng3 <- sub("\n=== Pronunciation ===\n+[^=]*\n+===", "\n===", eng2)
  # remove empty subsections
  eng4 <- gsub("\n==== [^=]+ ====\n", "\n\n", eng3)
  eng5 <- gsub(
    "\n=== ([^=]+) ===\n",
    paste0("\n\n", cli::style_underline("\\1:"), " "),
    eng4
  )
  eng6 <- cli::ansi_trimws(eng5)
  eng7 <- gsub("\n[\n]+", "\n\n", eng6)
  eng8 <- gsub("([^\n])\n([^\n])", "\\1 \\2", eng7)
  eng8
}

# -------------------------------------------------------------------------

async_profanity_get <- function(term) {
  async_profanity_get_query(term)$then(http_stop_for_status)$catch(
    error = function(err) list(content = charToRaw("NA"))
  )$then(function(resp) profanity_get_process(term, resp))
}

async_profanity_get_query <- function(term) {
  base <- Sys.getenv(
    "PKG_NAME_CHECK_PROFANITY_URL",
    "https://www.purgomalum.com/service/containsprofanity"
  )
  url <- paste0(base, "?text=", utils::URLencode(term))
  http_get(url)
}

profanity_get_process <- function(term, resp) {
  txt <- as.logical(rawToChar(resp$content))
  add_class(txt, "pkg_name_check_profanity")
}

# -------------------------------------------------------------------------

sentiment_get_has_data <- function() !is.null(pkgd_data$sentiment)

async_sentiment_get <- function(term) {
  start <- if (!sentiment_get_has_data()) {
    async_sentiment_get_data()
  } else {
    async_constant(pkgd_data$sentiment)
  }

  start$then(function(stm) structure(stm[term], names = term))$then(
    function(res) add_class(res, "pkg_name_check_sentiment")
  )
}

async_sentiment_get_data <- function() {
  url <- Sys.getenv(
    "PKG_NAME_CHECK_SENTIMENT_URL",
    "https://raw.githubusercontent.com/fnielsen/afinn/master/afinn/data/AFINN-en-165.txt"
  )
  http_get(url)$then(http_stop_for_status)$catch(
    error = function(err) list(content = raw())
  )$then(function(resp) {
    chr <- rawToChar(resp$content)
    lns <- strsplit(chr, "\n", fixed = TRUE)[[1]]
    cls <- strsplit(lns, "\t", fixed = TRUE)
    pkgd_data$sentiment <- structure(
      as.integer(vapply(cls, "[[", character(1), 2)),
      names = vapply(cls, "[[", character(1), 1)
    )
  })
}

sentiment_string <- function(x) {
  emo <- c(
    "-5" = "\U0001F62D\U0001F62D",
    "-4" = "\U0001F62D",
    "-3" = "\U0001F61E\U0001F61E",
    "-2" = "\U0001F61E",
    "-1" = "\U0001F610",
    "0" = "\U0001F610",
    "1" = "\U0001F642",
    "2" = "\U0001F642\U0001F642",
    "3" = "\U0001F606",
    "4" = "\U0001F606\U0001F606",
    "5" = "\U0001F970"
  )

  asc <- c(
    "-5" = ";(;(",
    "-4" = ";(",
    "-3" = ":(:(",
    "-2" = ":(",
    "-1" = ":|",
    "0" = ":|",
    "1" = ":)",
    "2" = ":):)",
    "3" = ":D",
    "4" = ":D:D",
    "5" = ""
  )
  if (has_emoji()) emo[as.character(x)] else asc[as.character(x)]
}

#' @export

print.pkg_name_check_sentiment <- function(x, ...) {
  writeLines(format(x, ...))
  invisible(x)
}

#' @export

format.pkg_name_check_sentiment <- function(x, ...) {
  if (is.na(x)) {
    x <- 0
  }
  str <- sentiment_string(x)
  txt <- paste0("Sentiment: ", str, cli::col_silver(paste0(" (", x, ")")))
  cw <- cli::console_width()
  alg <- cli::ansi_align(txt, width = cw - 4)
  cli::boxx(alg, padding = c(0, 1, 0, 1), border_col = cli::col_silver)
}

# -------------------------------------------------------------------------

async_urban_get <- function(term) {
  async_urban_get_query(term)$then(function(resp) urban_get_process(term, resp))
}

async_urban_get_query <- function(term) {
  base <- Sys.getenv(
    "PKG_NAME_CHECK_URBAN_URL",
    "http://api.urbandictionary.com/v0/"
  )
  url <- paste0(base, "define?term=", utils::URLencode(term))
  http_get(url)$then(http_stop_for_status)$catch(
    error = function(err) list(content = charToRaw('{"list":[]}'))
  )
}

urban_get_process <- function(term, resp) {
  obj <- jsonlite::fromJSON(
    rawToChar(resp$content),
    simplifyVector = FALSE
  )$list
  tbl <- data_frame(
    definition = vcapply(obj, "[[", "definition"),
    permalink = vcapply(obj, "[[", "permalink"),
    thumbs_up = viapply(obj, "[[", "thumbs_up"),
    word = vcapply(obj, "[[", "word"),
    written_on = vcapply(obj, "[[", "written_on"),
    example = vcapply(obj, "[[", "example"),
    thumbs_down = viapply(obj, "[[", "thumbs_down")
  )
  add_class(tbl, "pkg_name_check_urban")
}

#' @export

print.pkg_name_check_urban <- function(x, ...) {
  writeLines(format(x, ...))
  invisible(x)
}

#' @export

format.pkg_name_check_urban <- function(x, limit = 6, ...) {
  if (nrow(x) == 0) {
    txt <- "No definition found."
    ftr <- ""
  } else {
    x <- x[1, , drop = FALSE]
    txt <- clean_urban(x$definition)
    ftr <- cli::col_blue(x$permalink)
  }
  cw <- cli::console_width()
  wrp <- cli::ansi_strwrap(txt, width = cw - 4)
  hdr <- cli::col_green("Urban dictionary")
  alg <- cli::ansi_align(wrp, width = cw - 4)
  if (length(alg) > limit) {
    alg <- c(alg[1:limit], cli::symbol$ellipsis)
  }
  cli::boxx(
    alg,
    padding = c(0, 1, 0, 1),
    header = hdr,
    footer = ftr,
    border_col = cli::col_silver
  )
}

clean_urban <- function(x) {
  gsub("[\r\n]", " ", x)
}

# -------------------------------------------------------------------------

async_pnc_bioc <- function(name) {
  # A removed package? Although maybe this is OK?
  removed <- pnc_bioc_removed()
  mch <- match(tolower(name), tolower(removed))
  if (!is.na(mch)) {
    return(pnc_bioc_false(removed[mch]))
  }

  # An annotatation package?
  ann <- pnc_bioc_old_annotation()
  mch <- match(tolower(name), tolower(ann))
  if (!is.na(mch)) {
    return(pnc_bioc_false(ann[mch]))
  }

  # Need to query
  async_pnc_bioc_web(name)
}

pnc_bioc_false <- function(pkg) {
  async_constant(list(bioc = FALSE, package = pkg))
}

async_pnc_bioc_web <- function(name) {
  name
  async_pnc_bioc_query(name)$then(
    function(response) pnc_bioc_process(name, response)
  )
}

async_pnc_bioc_query <- function(name) {
  base1 <- Sys.getenv(
    "PKG_NAME_CHECK_BIOC_URL",
    "https://git.bioconductor.org/info?packages/"
  )
  url1 <- paste0(base1, substr(tolower(name), 1, 1))
  url2 <- paste0(base1, substr(toupper(name), 1, 1))
  url3 <- Sys.getenv("PKG_NAME_CHECK_BIOC_ANN_URL")
  if (url3 == "") {
    url3 <- paste0(
      pkgcache::bioc_repos(pkgcache::bioc_devel_version())[["BioCann"]],
      "/src/contrib/PACKAGES.gz"
    )
  }
  when_all(http_get(url1), http_get(url2), http_get(url3))
}

pnc_bioc_process <- function(name, response) {
  pkgs <- unique(c(
    pnc_bioc_parse(response[[1]]),
    pnc_bioc_parse(response[[2]]),
    pnc_bioc_parse_pgz(response[[3]])
  ))
  mch <- match(tolower(name), tolower(pkgs))
  add_class(
    list(bioc = is.na(mch), package = if (!is.na(mch)) pkgs[mch]),
    "pkg_name_check_bioc"
  )
}

pnc_bioc_parse <- function(response) {
  http_stop_for_status(response)
  cnt <- strsplit(rawToChar(response$content), "\n")[[1]]
  mch <- re_match(cnt, "packages/(?<package>[a-z0-9A-Z.]+)")
  na.omit(mch$package)
}

pnc_bioc_parse_pgz <- function(response) {
  rc <- rawConnection(response$content)
  on.exit(close(rc), add = TRUE)
  pkgs <- read.dcf(gzcon(rc))
  pkgs[, "Package"]
}

# These are the packages that are not in the Bioconductor git repo

pnc_bioc_removed <- function() {
  c(
    # removed in 3.5
    "betr",
    "encoDnaseI",
    "ggtut",
    # removed in 3.4
    "AffyTiling",
    "cellHTS",
    "DASiR",
    "DAVIDQuery",
    "GenoView",
    "inSilicoDb",
    "inSilicoMerging",
    "jmosaic",
    "metaX",
    "MMDiff",
    "neaGUI",
    "Rolexa",
    "RWebServices",
    "SJava",
    "SomaticCA",
    "spade",
    # removed in 3.1
    "asmn",
    "COPDSexualDimorphism",
    "DNaseR",
    "flowFlowJo",
    "flowPhyto",
    # removed in 3.0
    "RMAPPER",
    "virtualArray",
    # removed in 2.14
    "Agi4x44PreProcess",
    "maDB",
    "pgUtils",
    # removed in 2.13
    "dualKS",
    "externalVector",
    "GeneGroupAnalysis",
    "iFlow",
    "KEGGSOAP",
    "xmapcore",
    # removed in 2.12
    "cosmo",
    "cosmoGUI",
    "gene2pathway",
    # removed in 2.11
    "PatientGeneSets",
    # removed in 2.10
    "edd",
    "GeneRfold",
    "ontoTools",
    "GeneR",
    "RMAGEML",
    "RTooks4TB",
    # Packages removed with Bioconductor 2.9 release
    "GeneSpring",
    "GeneTraffic",
    "makePlatformDesign",
    "Rdbi",
    "RdbiPgSQL",
    "rflowcyt",
    "Rredland",
    "Ruuid",
    "simulatorAPMS",
    "snpMatrix",
    # Packages removed with Bioconductor 2.8 release
    "exonmap",
    "biocDatasets",
    # Packages removed with Bioconductor 2.7 release
    "matchprobes",
    "GeneticsBase",
    "fbat",
    # Packages removed with Bioconductor 2.6 release
    "keggorth",
    # Packages removed with Bioconductor 2.5 release
    "stam",
    "Rintact",
    # Packages removed with Bioconductor 2.4 release
    "AnnBuilder",
    "RSNPer",
    # Packages removed with Bioconductor 2.3 release
    "SemSim",
    "WidgetInvoke",
    # Packages removed with Bioconductor 2.2 release
    "SAGElyzer",
    "GeneTS",
    "arrayMagic",
    "chromoViz",
    "iSPlot",
    "iSNetword",
    # Packages removed with Bioconductor 2.1 release
    "bim",
    "arrayQCplot",
    # Packages removed with Bioconductor 2.0 release
    "gff3Plotter",
    "mmgmos",
    "applera",
    # Packages removed with Bioconductor 1.9 release
    "reposTools",
    "exprExternal",
    "SNAData",
    "rfcdmin",
    "goCluster",
    "GenomeBase",
    "y2hStat",
    # Packages removed with Bioconductor 1.8 release
    "msbase",
    "ideogram",
    # Packages removed with Bioconductor 1.7 release
    "mscalib"
  )
}

# nocov start
function() {
  bv <- setdiff(
    as.character(pkgcache::bioc_version_map()$bioc_version),
    c("1.6", "1.7")
  )
  pkgs <- lapply(bv, function(v) {
    repos <- pkgcache::bioc_repos(v)["BioCann"]
    rownames(utils::available.packages(repos = repos))
  })
  packages <- sort(unique(unlist(pkgs)))
  saveRDS(
    packages,
    file = "inst/exdata/biocpackages.rds",
    version = 2,
    compress = "xz"
  )
}
# nocov end

pnc_bioc_old_annotation <- function() {
  readRDS(system.file("exdata", "biocpackages.rds", package = "pkgdepends"))
}

test_that("pkg_name_check", {
  mockery::stub(
    pkg_name_check,
    "async_pkg_name_check",
    function(...) async_constant("boo")
  )
  expect_equal(pkg_name_check("foobar"), "boo")
})

test_that("async_pkg_name_check", {
  expect_snapshot(error = TRUE, {
    sy(async_pkg_name_check(11))
    sy(async_pkg_name_check(c("a", "b")))
    sy(async_pkg_name_check(NA_character_))
    sy(async_pkg_name_check("x", 11))
    sy(async_pkg_name_check("x", NA_character_))
  })

  calls <- 0L
  fun <- function(name) calls <<- calls + 1L
  mockery::stub(async_pkg_name_check, "async_pnc_basics", fun)
  mockery::stub(async_pkg_name_check, "async_wikipedia_get", fun)
  mockery::stub(async_pkg_name_check, "async_wiktionary_get", fun)
  mockery::stub(async_pkg_name_check, "async_sentiment_get", fun)
  mockery::stub(async_pkg_name_check, "async_urban_get", fun)

  ans <- sy(async_pkg_name_check("foo"))
  expect_s3_class(ans, "pkg_name_check")
  expect_equal(sort(unlist(ans, use.names = FALSE)), 1:4)
})

# format.pkg_name_check via print

test_that("print.pkg_name_check", {
  local_edition(3)
  local_reproducible_output()
  ans <- fixture$get({
    list(
      pkg_name_check("tools"),
      pkg_name_check("tools", dictionaries = "urban")
    )
  })
  expect_snapshot(print(ans[[1]]))
  expect_snapshot(print(ans[[2]]))
})

test_that("async_pnc_basics", {
  mockery::stub(async_pnc_basics, "async_pnc_crandb", "crandb")
  mockery::stub(async_pnc_basics, "async_pnc_bioc", "bioc")
  mockery::stub(async_pnc_basics, "async_profanity_get", "profanity")
  ans <- sy(async_pnc_basics("name"))
  expect_s3_class(ans, "pkg_name_check_basics")
  expect_equal(
    names(ans),
    c("name", "valid", "base", "crandb", "bioc", "profanity")
  )
})

test_that("forbidden_package_names", {
  expect_true(is.character(forbidden_package_names()))
})

test_that("pnc_valid", {
  good <- c("A3", "aa", "a.3", "foo.bar.foobar")
  bad <- c("3aaa", "a.", "aaaa.", "aaa-bbb", "description", "ff\u00e1f")
  for (c in good) {
    expect_true(pnc_valid(c))
  }
  for (c in bad) {
    expect_false(pnc_valid(c))
  }

  mockery::stub(pnc_valid, "grepl", TRUE)
  expect_false(pnc_valid("ff\u00e1f"))
})

test_that("pnc_base", {
  local_edition(3)
  local_reproducible_output()
  expect_snapshot(pnc_base("base"))
  expect_snapshot(pnc_base("Base"))
  expect_snapshot(pnc_base("TOOLS"))
  expect_snapshot(pnc_base("definitely-not"))
})

test_that("async_cranlike_check", {
  ## TODO: need a CRAN dummy for this
})

test_that("format.pkg_name_basics", {
  local_edition(3)
  local_reproducible_output()
  bss <- fixture$get({
    list(
      sy(async_pnc_basics("pwr")),
      sy(async_pnc_basics(paste0("s", "h", "i", "t")))
    )
  })
  expect_snapshot({
    writeLines(format(bss[[1]]))
    writeLines(format(bss[[2]]))
  })
})

test_that("crandb_check", {
  if (Sys.getenv("PKG_NAME_CHECK_REAL") == "") {
    withr::local_envvar(PKG_NAME_CHECK_CRANDB_URL = check_app$url("/crandb"))
  }

  res <- synchronize(when_all(
    good = async_pnc_crandb("no-such-package"),
    bad = async_pnc_crandb("dbi"),
    bad2 = async_pnc_crandb("DBI"),
    bad3 = async_pnc_crandb("Dbi")
  ))

  expect_true(res$good$crandb)
  expect_false(res$bad$crandb)
  expect_false(res$bad2$crandb)
  expect_false(res$bad3$crandb)

  expect_null(res$good$package)
  expect_equal(res$bad$package, "DBI")
  expect_equal(res$bad2$package, "DBI")
  expect_equal(res$bad2$package, "DBI")

  lapply(res, expect_s3_class, "pkg_name_check_crandb")
})

test_that("async_wikipedia_get", {
  withr::local_envvar(
    PKG_NAME_CHECK_WIKIPEDIA_URL = check_app$url("/wikipedia")
  )
  ret <- synchronize(async_wikipedia_get("foobar"))
  expect_equal(ret$term, "foobar")
  expect_equal(ret$normalized, "Foobar")
})

test_that("async_wikipedia_get_query", {
  local_edition(3)
  local_reproducible_output()
  withr::local_envvar(PKG_NAME_CHECK_WIKIPEDIA_URL = check_app$url("/echo"))
  ret <- synchronize(async_wikipedia_get_query("foobar"))
  expect_snapshot(show_request(ret))
})

# make_wikipedia_data
# wikipedia_get_process

test_that("format.pkg_name_check_wikipedia", {
  local_edition(3)
  local_reproducible_output()
  wpd <- fixture$get({
    list(
      sy(async_wikipedia_get("cli")),
      sy(async_wikipedia_get("surely-not-this")),
      sy(async_wikipedia_get("GNU R"))
    )
  })
  expect_snapshot({
    writeLines(format(wpd[[1]]))
    writeLines(format(wpd[[2]]))
    writeLines(format(wpd[[3]]))
  })
})

# clean_wikipedia_text

test_that("async_wiktionary_get", {
  local_edition(3)
  local_reproducible_output()
  resp <- fixture$get({
    sy(async_wiktionary_get_query("foobar"))
  })
  mockery::stub(
    async_wiktionary_get,
    "async_wiktionary_get_query",
    function(...) async_constant(resp)
  )
  ans <- sy(async_wiktionary_get("foobar"))
  expect_s3_class(ans, "pkg_name_check_wiktionary")
  expect_s3_class(ans, "tbl")
  expect_snapshot(print(ans))
})

test_that("async_wiktionary_get_query", {
  local_edition(3)
  local_reproducible_output()
  withr::local_envvar(PKG_NAME_CHECK_WIKTIONARY_URL = check_app$url("/echo"))
  ret <- synchronize(async_wiktionary_get_query("foobar"))
  expect_snapshot(show_request(ret))
})

test_that("wiktionary_get_process", {
  resp <- fixture$get({
    sy(async_wiktionary_get_query("foobar"))
  })
  ans <- wiktionary_get_process("foobar", resp)
  expect_s3_class(ans, "pkg_name_check_wiktionary")
  expect_s3_class(ans, "tbl")
  expect_match(ans$text, "== English ==")
})

test_that("format.pkg_name_check_wiktionary", {
  local_edition(3)
  local_reproducible_output()
  resp <- fixture$get({
    sy(async_wiktionary_get_query("foobar"))
  })
  resp2 <- fixture$get({
    sy(async_wiktionary_get_query("not-at-all-sdfsdfsdf"))
  })
  ans <- wiktionary_get_process("foobar", resp)
  ans2 <- wiktionary_get_process("not-at-all-sdfsdfsdf", resp2)
  expect_snapshot({
    writeLines(format(ans))
    writeLines(format(ans2))
  })
})

# clean_wiktionary_text

test_that("async_profanity_get", {
  resp <- fixture$get({
    sy(async_profanity_get_query("nope"))
  })
  mockery::stub(
    async_profanity_get,
    "async_profanity_get_query",
    function(...) async_constant(resp)
  )
  ans <- sy(async_profanity_get("nope"))
  expect_s3_class(ans, "pkg_name_check_profanity")
  expect_false(ans)
})

test_that("async_profanity_get_query", {
  local_edition(3)
  local_reproducible_output()
  withr::local_envvar(PKG_NAME_CHECK_PROFANITY_URL = check_app$url("/echo"))
  ret <- synchronize(async_profanity_get_query("foobar"))
  expect_snapshot(show_request(ret))
})

test_that("profanity_get_process", {
  resp <- fixture$get({
    sy(async_profanity_get_query("nope"))
  })
  ans <- profanity_get_process("nope", resp)
  expect_s3_class(ans, "pkg_name_check_profanity")
  expect_false(ans)
})

sentiment_ex <- function() c(abuse = -3, irony = -1, xo = 3, xoxoxo = 4)

test_that("async_sentiment_get", {
  expect_true(is.logical(sentiment_get_has_data()))

  mockery::stub(async_sentiment_get, "sentiment_get_has_data", FALSE)
  mockery::stub(
    async_sentiment_get,
    "async_sentiment_get_data",
    function(...) async_constant(sentiment_ex())
  )

  ans <- sy(async_sentiment_get("xo"))
  expect_equal(
    ans,
    structure(c(xo = 3), class = c("pkg_name_check_sentiment", "numeric"))
  )

  mockery::stub(async_sentiment_get, "sentiment_get_has_data", TRUE)
  mockery::stub(
    async_sentiment_get,
    "async_constant",
    function(...) async_constant(sentiment_ex())
  )

  ans <- sy(async_sentiment_get("xo"))
})

test_that("async_sentiment_get_data", {
  withr::local_envvar(
    PKG_NAME_CHECK_SENTIMENT_URL = check_app$url("/sentiment")
  )
  ans <- sy(async_sentiment_get_data())
  pkgd_data$sentiment <- NULL
  expect_equal(ans, sentiment_ex())
})

test_that("sentiment_string", {
  local_reproducible_output()
  expect_true(is_ascii(sentiment_string(1)))
  mockery::stub(sentiment_string, "has_emoji", TRUE)
  expect_false(is_ascii(sentiment_string(1)))
})

test_that("format.pkg_name_check_sentiment", {
  local_edition(3)
  local_reproducible_output()
  mockery::stub(async_sentiment_get, "sentiment_get_has_data", FALSE)
  withr::local_envvar(
    PKG_NAME_CHECK_SENTIMENT_URL = check_app$url("/sentiment")
  )
  ans1 <- sy(async_sentiment_get("xo"))
  ans2 <- sy(async_sentiment_get("something"))
  pkgd_data$sentiment <- NULL
  expect_snapshot({
    writeLines(format(ans1))
    writeLines(format(ans2))
  })
})

test_that("async_urban_get", {
  local_edition(3)
  local_reproducible_output()
  mockery::stub(async_urban_get, "async_urban_get_query", function(...) {
    async_constant(fixture$get({
      sy(async_urban_get_query("tool"))
    }))
  })
  ans <- sy(async_urban_get("tool"))
  expect_s3_class(ans, "pkg_name_check_urban")
  expect_s3_class(ans, "tbl")
  expect_snapshot(ans)
})

test_that("async_urban_get_query", {
  local_edition(3)
  local_reproducible_output()
  withr::local_envvar(PKG_NAME_CHECK_URBAN_URL = check_app$url("/echo/"))
  ret <- synchronize(async_urban_get_query("foobar"))
  expect_snapshot(show_request(ret))
})

test_that("urban_get_process", {
  local_edition(3)
  local_reproducible_output()
  resp <- fixture$get({
    sy(async_urban_get_query("tool"))
  })
  ans <- urban_get_process("tool", resp)
  expect_s3_class(ans, "pkg_name_check_urban")
  expect_s3_class(ans, "tbl")
  expect_snapshot(ans)
})

test_that("format.pkg_name_check_urban", {
  local_edition(3)
  local_reproducible_output()
  ans <- fixture$get({
    sy(async_urban_get("tool"))
  })
  ans2 <- fixture$get({
    sy(async_urban_get("not-this-one-asdfsf"))
  })
  expect_snapshot(writeLines(format(ans)))
  expect_snapshot(writeLines(format(ans2)))

  ans <- ans[1, ]
  ans$definition[1] <- paste(rep(ans$definition[1], 10), collapse = "\n")
  expect_snapshot(writeLines(format(ans)))
})

test_that("clean_urban", {
  expect_equal(clean_urban("foo \r\n\r\n bar"), "foo      bar")
  expect_equal(clean_urban("foo \r\n\n bar"), "foo     bar")
  expect_equal(clean_urban("foo \n\n bar"), "foo    bar")
})

test_that("async_pnc_bioc", {
  expect_equal(sy(async_pnc_bioc("bim")), sy(pnc_bioc_false("bim")))
  expect_equal(sy(async_pnc_bioc("yeast2")), sy(pnc_bioc_false("yeast2")))

  mockery::stub(
    async_pnc_bioc,
    "async_pnc_bioc_web",
    function(...) async_constant(pnc_bioc_false("Biobase"))
  )
  expect_equal(sy(async_pnc_bioc("Biobase")), sy(pnc_bioc_false("Biobase")))
})

# "pnc_bioc_false"

test_that("async_pnc_bioc_web", {
  local_edition(3)
  local_reproducible_output()
  response <- fixture$get({
    sy(async_pnc_bioc_query("all"))
  })
  mockery::stub(
    async_pnc_bioc_web,
    "async_pnc_bioc_query",
    function(...) async_constant(response)
  )
  ans <- sy(async_pnc_bioc_web("all"))
  expect_snapshot(ans)
})

test_that("pnc_bioc_query", {
  local_edition(3)
  local_reproducible_output()
  withr::local_envvar(
    PKG_NAME_CHECK_BIOC_URL = check_app$url("/bioc/")
  )
  mockery::stub(
    async_pnc_bioc_query,
    "pkgcache::bioc_repos",
    list(BioCann = check_app$url("/biocann/"))
  )
  ans <- pnc_bioc_process("all", sy(async_pnc_bioc_query("all")))
  ans2 <- pnc_bioc_process("Agcdf", sy(async_pnc_bioc_query("Agcdf")))
  ans3 <- pnc_bioc_process("axx", sy(async_pnc_bioc_query("axx")))
  expect_snapshot(ans)
  expect_snapshot(ans2)
  expect_snapshot(ans3)
})

test_that("pnc_bioc_process", {
  response <- fixture$get({
    sy(async_pnc_bioc_query("all"))
  })
  ans <- pnc_bioc_process("all", response)
  expect_s3_class(ans, "pkg_name_check_bioc")
  expect_false(ans$bioc)
  expect_equal(ans$package, "ALL")
})

test_that("pnc_bioc_parse", {
  local_edition(3)
  local_reproducible_output()
  withr::local_envvar(
    PKG_NAME_CHECK_BIOC_URL = check_app$url("/bioc/"),
    PKG_NAME_CHECK_BIOC_ANN_URL = check_app$url(
      "/biocann/src/contrib/PACKAGES.gz"
    )
  )
  pkg1 <- pnc_bioc_parse(sy(async_pnc_bioc_query("all"))[[1]])
  pkg2 <- pnc_bioc_parse(sy(async_pnc_bioc_query("all"))[[2]])
  expect_snapshot(pkg1)
  expect_snapshot(pkg2)
})

test_that("pnc_bioc_parse_pgz", {
  local_edition(3)
  local_reproducible_output()
  withr::local_envvar(
    PKG_NAME_CHECK_BIOC_URL = check_app$url("/bioc/"),
    PKG_NAME_CHECK_BIOC_ANN_URL = check_app$url(
      "/biocann/src/contrib/PACKAGES.gz"
    )
  )
  pkg <- pnc_bioc_parse_pgz(sy(async_pnc_bioc_query("all"))[[3]])
  expect_snapshot(pkg)
})

test_that("pnc_bioc_removed", {
  expect_true(is.character(pnc_bioc_removed()))
})

test_that("pnc_bioc_old_annotation", {
  expect_true(is.character(pnc_bioc_old_annotation()))
})

pkgplan_draw_solution_tree <- function(self, private, pkgs, annotate) {
  assert_that(is.null(pkgs) || is_character(pkgs))

  self$stop_for_solve_error()
  sol <- self$get_solution()$data
  sol <- sol[order(sol$package), ]
  pkgs <- pkgs %||% sol$package[sol$directpkg]

  data <- sol[, "package", drop = FALSE]

  if ("dependencies" %in% names(sol)) {
    data$deps <- sol$dependencies
  } else {
    deps <- lapply(seq_len(nrow(sol)), function(i) {
      d <- sol$deps[[i]]
      d[tolower(d$type) %in% tolower(sol$dep_types[[i]]), ]
    })
    deps <- lapply(deps, "[[", "package")
    data$deps <- deps
  }
  data$deps <- lapply(data$deps, intersect, data$package)

  ann_version <- function() {
    v <- cli::col_silver(sol$version)
    upd <- sol$lib_status == "update"
    if (any(upd)) {
      v[upd] <- cli::col_green(paste(
        sol$old_version[upd],
        "->",
        sol$version[upd]
      ))
    }
    nupd <- sol$lib_status == "no-update"
    if (any(nupd)) {
      v[nupd] <- cli::col_green(paste(
        sol$version[nupd],
        "<",
        sol$new_version[nupd]
      ))
    }
    v
  }

  data$label <- ifelse(
    sol$type %in% c("cran", "bioc", "standard", "installed"),
    data$package,
    sol$ref
  )
  data$trimmed <- data$label
  data$label <- paste(data$label, ann_version())
  data$label[sol$directpkg] <-
    cli::style_italic(cli::style_bold(cli::col_cyan(data$label[sol$directpkg])))

  if (annotate) {
    builder <- emoji("builder")
    ann <- get_tree_annotation(sol)
    data$label <- paste(
      data$label,
      annotate_tree(sol, ann, builder = builder)
    )
  }

  trees <- unlist(lapply(
    pkgs,
    function(p) c(cli::tree(data, root = p, trim = TRUE), "")
  ))

  if (annotate && any(unlist(ann))) {
    key <- paste0(
      if (any(ann$new)) paste(" |", cli::col_green(emoji("sparkles")), "new"),
      if (any(ann$upd)) paste(" |", cli::col_green(emoji("rocket")), "update"),
      if (any(ann$noupd)) {
        paste(" |", cli::col_green(emoji("hand")), "outdated")
      },
      if (any(ann$dl)) paste(" |", cli::col_green(emoji("dl")), "download"),
      if (any(ann$build)) paste(" |", cli::col_green(builder), "build"),
      if (any(ann$compile)) {
        paste(" |", cli::col_green(emoji("wrench")), "compile")
      }
    )
    key <- paste0("Key: ", sub("^ [|]", "", key))
    trees <- c(trees, key)
  }

  class(trees) <- c("cli_tree", "tree", "character")
  trees
}

has_emoji <- function() {
  if (!cli::is_utf8_output()) {
    return(FALSE)
  }
  if (isTRUE(opt <- getOption("pkg.emoji"))) {
    return(TRUE)
  }
  if (identical(opt, FALSE)) {
    return(FALSE)
  }
  if (Sys.info()[["sysname"]] != "Darwin") {
    return(FALSE)
  }
  TRUE
}

get_tree_annotation <- function(sol) {
  ann <- data.frame(
    stringsAsFactors = FALSE,
    new = sol$lib_status == "new",
    upd = sol$lib_status == "update",
    noupd = sol$lib_status == "no-update",
    dl = !is.na(sol$cache_status) & sol$cache_status == "miss"
  )
  ann$build <- (ann$new | ann$upd) & sol$platform == "source"
  ann$compile <- (ann$new | ann$upd) &
    !is.na(sol$needscompilation) &
    sol$needscompilation

  ann
}

annotate_tree <- function(sol, ann, builder = NULL) {
  builder <- builder %||% emoji("builder")
  cli::col_green(paste0(
    ifelse(ann$new, emoji("sparkles"), ""),
    ifelse(ann$upd, emoji("rocket"), ""),
    ifelse(ann$noupd, emoji("hand"), ""),
    ifelse(ann$build, builder, ""),
    ifelse(ann$compile, emoji("wrench"), ""),
    ifelse(ann$dl, paste(emoji("dl"), format_file_size(sol$filesize)), "")
  ))
}

format_file_size <- function(x) {
  bts <- str_trim(format_bytes$pretty_bytes(x))
  cli::col_silver(
    ifelse(is.na(x), "(unknown size)", paste0("(", bts, ")"))
  )
}

emoji <- function(what, alt = NULL) {
  emo <- has_emoji()
  switch(
    what,
    "rocket" = if (emo) "\U1F680" else alt %||% "[upd]",
    "sparkles" = if (emo) "\u2728" else alt %||% "[new]",
    "hand" = if (emo) "\u270B" else alt %||% "[old]",
    "dl" = if (emo) " \u2B07" else alt %||% "[dl]",
    "builder" = if (emo) emo_builder() else alt %||% "[bld]",
    "wrench" = if (emo) "\U1F527" else alt %||% "[cmp]",
    "pkg" = if (emo) "\U1F4E6" else alt %||% "pkg",
    "pkgs" = if (emo) "\U1F4E6" else alt %||% "pkgs",
    ""
  )
}

emo_builder <- function(n = 1) {
  base <- "\U1F477"
  tone <- c(
    "\U1F477\U1F3FB",
    "\U1F477\U1F3FC",
    "\U1F477\U1F3FD",
    "\U1F477\U1F3FE",
    "\U1F477\U1F3FF"
  )
  gend <- c(
    "\U1F477\u200D\u2640\uFE0F",
    "\U1F477\u200D\u2642\uFE0F",
    "\U1F477\U1F3FB\u200D\u2640\uFE0F",
    "\U1F477\U1F3FB\u200D\u2642\uFE0F",
    "\U1F477\U1F3FC\u200D\u2640\uFE0F",
    "\U1F477\U1F3FC\u200D\U2642\uFE0F",
    "\U1F477\U1F3FD\u200D\u2640\uFE0F",
    "\U1F477\U1F3FD\u200D\u2642\uFE0F",
    "\U1F477\U1F3FE\u200D\u2640\uFE0F",
    "\U1F477\U1F3FE\u200D\u2642\uFE0F",
    "\U1F477\U1F3FF\u200D\u2640\uFE0F",
    "\U1F477\U1F3FF\u200D\u2642\uFE0F"
  )

  rstudio_type <- rstudio_detect()$type
  if (rstudio_type == "rstudio_terminal") {
    ppl <- base
  } else if (rstudio_type == "rstudio_console") {
    ppl <- c(base, tone)
  } else {
    ppl <- c(base, tone, gend)
  }

  sample(ppl, n, replace = TRUE)
}

mkdirp <- function(dir, msg = NULL) {
  s <- vlapply(dir, dir.create, recursive = TRUE, showWarnings = FALSE)
  if (any(s) && !is.null(msg)) {
    cli::cli_alert_info(c(msg, ": {.path {dir[s]}}"))
  }
  invisible(s)
}

test_that("make_unzip_process", {
  zipfile <- system.file(package = "pkgdepends", "tools", "xxx.zip")
  mkdirp(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  px <- make_unzip_process(zipfile, exdir = tmp)
  px$wait(5000)
  px$kill()

  expect_equal(px$get_exit_status(), 0)
  expect_true(file.exists(file.path(tmp, "xxx")))
})

describe("verify_extracted_package", {
  skip_on_cran()
  tmp <- tempfile()
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  run <- function(pkgfile) {
    unlink(tmp, recursive = TRUE)
    mkdirp(tmp)
    utils::untar(pkgfile, exdir = tmp)
    verify_extracted_package(pkgfile, tmp)
  }

  it("errors if archive doesn't contain a DESCRIPTION file", {
    f1 <- local_binary_package("test1")
    expect_error(
      run(f1),
      "'.*test1[.]tgz' is not a valid R package, it is an empty archive",
      class = "install_input_error"
    )
  })

  it("errors if archive DESCRIPTION is not in the root directory", {
    f2 <- local_binary_package("test2", "foo/DESCRIPTION" = character())
    expect_snapshot(error = TRUE, run(f2), transform = transform_tempdir)
  })

  it("can handle multiple DESCRIPTION files", {
    f3 <- local_binary_package(
      "test3",
      "DESCRIPTION" = c("Package: test3", "Built: 2017-01-01"),
      "tests/testthat/DESCRIPTION" = character(),
      "Meta/package.rds" = character()
    )
    expect_s3_class(run(f3)$desc, "description")

    f4 <- local_binary_package(
      "test4",
      "pkgdir/DESCRIPTION" = c("Package: test4", "Built: 2017-01-01"),
      "Meta/package.rds" = character()
    )
    expect_snapshot(error = TRUE, run(f4), transform = transform_tempdir)
  })

  it("fails if the binary does not contain package.rds", {
    f5 <- local_binary_package("test5", "DESCRIPTION" = character())
    expect_snapshot(error = TRUE, run(f5), transform = transform_tempdir)
  })

  it("fails if the DESCRIPTION file is empty", {
    f6 <- local_binary_package(
      "test6",
      "DESCRIPTION" = character(),
      "Meta/package.rds" = character()
    )
    expect_snapshot(error = TRUE, run(f6), transform = transform_tempdir)
  })

  it("fails if the DESCRIPTION file has no 'Built' entry", {
    f7 <- local_binary_package(
      "test7",
      "DESCRIPTION" = c("Package: test7"),
      "Meta/package.rds" = character()
    )
    expect_error(
      run(f7),
      "'.*test7[.]tgz' is not a valid binary, no 'Built' entry in 'test7/DESCRIPTION'",
      class = "install_input_error"
    )
  })
})

test_that("verify_extrancted_package errors", {
  pkg_dir <- file.path("fixtures", "packages")

  expect_error(
    verify_extracted_package("bad1", file.path(pkg_dir, "bad1")),
    "single directory",
    class = "install_input_error"
  )

  expect_error(
    verify_extracted_package("bad2", file.path(pkg_dir, "bad2")),
    "invalid",
    class = "install_input_error"
  )

  expect_error(
    verify_extracted_package("bad3", file.path(pkg_dir, "bad3")),
    "Package",
    class = "install_input_error"
  )

  expect_error(
    verify_extracted_package("bad4", file.path(pkg_dir, "bad4")),
    "package name mismatch",
    class = "install_input_error"
  )
})

parse_remote_gitlab <- function(specs, config, ...) {
  pds <- re_match(specs, gitlab_rx())
  pds$ref <- pds$.text
  pds$protocol[pds$protocol == ""] <- "https"
  pds$host[pds$host == ""] <- "gitlab.com"
  pds$path <- paste0("/", pds$projectpath, "/", pds$project)
  pds$dotgit <- ""
  pds$commitish[pds$commitish == ""] <- "HEAD"
  pds$url <- paste0(pds$protocol, "://", pds$host, pds$path, ".git")
  cn <- setdiff(colnames(pds), c(".match", ".text"))
  pds <- pds[, cn]
  pds$type <- "gitlab"
  pds$package <- ifelse(nzchar(pds$package), pds$package, pds$project)
  lapply(
    seq_len(nrow(pds)),
    function(i) as.list(pds[i, ])
  )
}

resolve_remote_gitlab <- function(
  remote,
  direct,
  config,
  cache,
  dependencies,
  ...
) {
  resolve_remote_git(remote, direct, config, cache, dependencies, ...)$then(
    function(res) {
      res$metadata["RemoteHost"] <- remote$host
      res$metadata["RemoteRepo"] <- remote$project
      res$metadata["RemoteUsername"] <- remote$projectpath
      res$metadata["RemoteType"] <- "gitlab"
      if (!is.null(remote$subdir) && remote$subdir != "") {
        res$metadata["RemoteSubdir"] <- remote$subdir
      }
      res
    }
  )
}

download_remote_gitlab <- function(
  resolution,
  target,
  target_tree,
  config,
  cache,
  which,
  on_progress
) {
  download_remote_git(
    resolution,
    target,
    target_tree,
    config,
    cache,
    which,
    on_progress
  )
}

satisfy_remote_gitlab <- function(resolution, candidate, config, ...) {
  satisfy_remote_git(resolution, candidate, config, ...)
}

installedok_remote_gitlab <- function(installed, solution, config, ...) {
  installedok_remote_git(installed, solution, config, ...)
}

# source: https://docs.gitlab.com/ee/user/reserved_names.html#limitations-on-usernames-project-and-group-names
gitlab_slug_rx <- function() {
  "[a-zA-Z0-9][-._a-zA-Z0-9]*[a-zA-Z0-9]"
}

gitlab_project_rx <- function() {
  paste0("(?<project>", gitlab_slug_rx(), ")")
}

gitlab_project_path_rx <- function() {
  paste0("(?<projectpath>", gitlab_slug_rx(), "(?:/", gitlab_slug_rx(), ")*)")
}

gitlab_rx <- function() {
  paste0(
    "^",
    ## Optional package name
    "(?:(?<package>",
    package_name_rx(),
    ")=)?",
    "gitlab::",
    ## Optional protocol::host
    "(?:(?<protocol>[^/]*)://(?<host>[^/]+)/)?",
    gitlab_project_path_rx(),
    "/",
    gitlab_project_rx(),
    ## Optional subdirectory, prefixed with /-, ie project/-/sub/dir
    "(?:/-",
    github_subdir_rx(),
    ")?",
    "(?:",
    github_commitish_rx(),
    ")?",
    "$"
  )
}

#' @export

format.package_packaging_error <- function(x, ...) {
  format_error_with_stdout(x, ...)
}

#' @export

print.package_packaging_error <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
}

#' @export

format.package_uncompress_error <- function(x, ...) {
  out <- conditionMessage(x)
  if (!is.null(x$data$stdout)) {
    stdout <- last_stdout_lines(x$stdout, "", prefix = "O> ")[-(1:2)]
    out <- c(out, "", "Standard output:", stdout)
  }
  out
}

#' @export

print.package_uncompress_error <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
}

#' @export

format.package_build_error <- function(x, ...) {
  format_error_with_stdout(x, ...)
}

#' @export

print.package_build_error <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
}

is_loaded <- function(package) {
  package %in% loadedNamespaces()
}

create_temp_dir <- function(..., tmpdir = tempdir()) {
  f <- tempfile(tmpdir = tmpdir, ...)
  dir.create(f)
  f
}

library_cache <- function(lib) {
  lib_cache <- file.path(lib, "_cache")
  dir.create(lib_cache, recursive = TRUE, showWarnings = FALSE)
  lib_cache
}

lock_cache <- function(cache, pkg_name, lock = TRUE) {
  use_lock <- !identical(lock, FALSE)
  my_lock <- NULL
  if (use_lock) {
    lockfile <- file.path(cache, sprintf("%s.lock", pkg_name))
    # TODO: timeout and fail?
    my_lock <- filelock::lock(lockfile)
  }
  my_lock
}

unlock <- function(lock) {
  if (is.null(lock)) {
    return()
  }
  filelock::unlock(lock)
}


sysname <- function() {
  res <- tolower(Sys.info()[["sysname"]])
  map <- c(darwin = "mac", "sunos" = "solaris")[res]
  res[!is.na(map)] <- map
  res
}

rep_list <- function(n, expr) {
  lapply(integer(n), eval.parent(substitute(function(...) expr)))
}

cut_into_lines <- function(x) {
  x <- do.call(paste0, as.list(x))
  x <- gsub("\r\n", "\n", x, fixed = TRUE)
  x <- strsplit(x, "\n", fixed = TRUE)[[1]]
  if (length(x)) x else ""
}

fixtures <- list(
  "resolution-simple.rds" = function() {
    r <- pkg_plan$new("pkgconfig", lib = tempfile())
    r$resolve()
  },

  "resolution-progress.rds" = function() {
    r <- pkg_plan$new("progress", lib = tempfile())
    r$resolve()
  },

  "resolution-installed.rds" = function() {
    dir.create(tmp <- tempfile())
    on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
    install.packages("pkgconfig", lib = tmp)
    r <- pkg_plan$new("pkgconfig", library = tmp)
    r$resolve()
  },

  "resolution-gh-vs-cran.rds" = function() {
    dir.create(tmp <- tempfile())
    on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
    r <- pkg_plan$new(
      c("cran::pkgconfig", "github::r-lib/pkgconfig"),
      lib = tmp
    )
    r$resolve()
  },

  "solution-crayon.rds" = function() {
    dir.create(tmp <- tempfile())
    on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
    r <- pkg_plan$new("crayon", lib = tempfile())
    r$resolve()
    r$solve()
  },

  "solution-igraph.rds" = function() {
    dir.create(tmp <- tempfile())
    on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
    r <- pkg_plan$new("igraph", lib = tempfile())
    r$resolve()
    r$solve()
  }
)

fixture_dir <- function() {
  testthat::test_path("fixtures")
}

get_fixture <- function(file) {
  file.path(fixture_dir(), file)
}

read_fixture <- function(file) {
  readRDS(get_fixture(file))
}

update_fixtures <- function(files = NULL) {
  if (is.null(files)) {
    files <- names(fixtures)
  }
  fdir <- fixture_dir()
  for (f in files) {
    output <- file.path(fdir, f)
    cat(output, sep = "\n")
    saveRDS(fixtures[[f]](), file = output, version = 2)
  }
}

test_that("is_valid_package, etc", {
  expect_true(is_valid_package(get_fixture("foobar_1.0.0.tar.gz")))
  expect_true(is_valid_package(get_fixture("foobar_1.0.0.zip")))
  expect_false(is_valid_package(get_fixture("foobar_1.0.1.tar.gz")))
  expect_false(is_valid_package(get_fixture("foobar_1.0.2.tar.gz")))
  expect_false(is_valid_package(get_fixture("foobar_1.0.1.zip")))
  expect_false(is_valid_package(get_fixture("foobar_1.0.2.zip")))
})

pkg_plan <- R6::R6Class(
  "pkg_plan",
  public = list(
    initialize = function(
      refs,
      config = list(),
      library = NULL,
      remote_types = NULL,
      lockfile = NULL
    ) {
      pkgplan_init(
        self,
        private,
        refs,
        config,
        library,
        remote_types,
        lockfile
      )
    },

    get_refs = function() private$refs,
    has_resolution = function() !is.null(private$resolution$result),
    has_clean_resolution = function() {
      self$has_resolution() && (all(private$resolution$result$status == "OK"))
    },
    has_resolution_downloads = function() !is.null(private$downloads),
    has_solution_downloads = function() !is.null(private$solution_downloads),
    has_solution = function() !is.null(private$solution),
    get_config = function() private$config,

    async_resolve = function() pkgplan_async_resolve(self, private),
    resolve = function() pkgplan_resolve(self, private),
    get_resolution = function() pkgplan_get_resolution(self, private),

    async_download_resolution = function() {
      pkgplan_async_download_resolution(self, private)
    },
    download_resolution = function() pkgplan_download_resolution(self, private),
    get_resolution_download = function() {
      pkgplan_get_resolution_download(self, private)
    },

    solve = function(policy = c("lazy", "upgrade")) {
      pkgplan_solve(self, private, match.arg(policy))
    },
    delete_solution = function() private$solution <- NULL,
    stop_for_solve_error = function() {
      pkgplan_stop_for_solve_error(self, private)
    },
    get_solution = function() pkgplan_get_solution(self, private),
    show_solution = function(key = FALSE) {
      pkgplan_show_solution(self, private, key)
    },
    get_sysreqs = function() pkgplan_get_sysreqs(self, private),
    show_sysreqs = function() pkgplan_show_sysreqs(self, private),
    get_install_plan = function() {
      pkgplan_install_plan(self, private, downloads = TRUE)
    },
    export_install_plan = function(plan_file = "pkg.lock", version = 2) {
      pkgplan_export_install_plan(self, private, plan_file, version)
    },
    draw_solution_tree = function(pkgs = NULL, annotate = TRUE) {
      pkgplan_draw_solution_tree(self, private, pkgs, annotate)
    },

    async_download_solution = function() {
      pkgplan_async_download_solution(self, private)
    },
    download_solution = function() pkgplan_download_solution(self, private),
    get_solution_download = function() {
      pkgplan_get_solution_download(self, private)
    },
    stop_for_solution_download_error = function() {
      pkgplan_stop_for_solution_download_error(self, private)
    },
    stop_for_resolution_download_error = function() {
      pkgplan_stop_for_resolution_download_error(self, private)
    },

    update = function() pkgplan_update(self, private),
    update_sysreqs = function() pkgplan_update_sysreqs(self, private),

    print = function(...) pkgplan_print(self, private, ...)
  ),

  private = list(
    refs = NULL,
    dirty = FALSE,
    remotes = list(),
    cache = NULL,
    resolution = NULL,
    solution = NULL,
    downloads = NULL,
    solution_downloads = NULL,
    download_cache = NULL,
    config = NULL,
    progress_bar = NULL,
    progress_bar_timer = NULL,
    remote_types = NULL,
    system_packages = NULL,
    sysreqs = NULL,

    download_res = function(res, which, on_progress = NULL) {
      pkgplan_download_res(self, private, res, which, on_progress)
    },
    subset_resolution = function(which) {
      pkgplan__subset_resolution(self, private, which)
    },
    create_lp_problem = function(pkgs, policy) {
      pkgplan__create_lp_problem(self, private, pkgs, policy)
    },
    solve_lp_problem = function(problem) {
      pkgplan__solve_lp_problem(self, private, problem)
    },

    create_progress_bar = function(what) {
      bar <- pkgplan__create_progress_bar(what)
      pkgplan__init_progress_bar(bar)
      bar
    },
    update_progress_bar = function(idx, event, data) {
      pkgplan__update_progress_bar(private$progress_bar, idx, event, data)
    },
    done_progress_bar = function() {
      if (!is.null(private$progress_bar)) {
        pkgplan__done_progress_bar(private$progress_bar)
        private$progress_bar <- NULL
      }
    }
  )
)

#' @importFrom utils modifyList

pkgplan_init <- function(
  self,
  private,
  refs,
  config,
  library,
  remote_types,
  lockfile
) {
  if (!is.null(lockfile)) {
    return(pkgplan_init_lockfile(
      self,
      private,
      lockfile,
      config,
      library,
      remote_types
    ))
  }

  assert_that(
    is_character(refs),
    is.null(library) || is_non_empty_character(library)
  )

  private$refs <- refs
  private$remotes <- parse_pkg_refs(refs)
  private$config <- current_config()$update(config)
  private$config$set("library", library)
  private$remote_types <- remote_types %||% default_remote_types()

  if (!is.null(library)) {
    mkdirp(library[1], msg = "Creating library directory")
    library[1] <- normalizePath(library[1])
  }
  mkdirp(private$download_cache <- private$config$get("cache_dir"))

  installed <- make_installed_cache(library)
  if (!is.null(installed)) {
    installed <- merge_installed_caches(
      installed,
      make_installed_cache(.Library, priority = "recommended")
    )
  }

  private$cache <- list(
    metadata = pkgcache::cranlike_metadata_cache$new(
      replica_path = private$config$get("metadata_cache_dir"),
      platforms = private$config$get("platforms"),
      r_version = private$config$get("r_versions"),
      cran_mirror = private$config$get("cran_mirror"),
      update_after = private$config$get("metadata_update_after"),
      bioc = private$config$get("use_bioconductor")
    ),
    package = pkgcache::package_cache$new(private$config$get(
      "package_cache_dir"
    )),
    installed = installed
  )

  private$dirty <- TRUE
  invisible(self)
}

pkgplan_init_lockfile <- function(
  self,
  private,
  lockfile,
  config,
  library,
  remote_types
) {
  assert_that(
    is_path(lockfile),
    is_optional_path(library)
  )

  private$config <- current_config()$update(config)
  private$config$set("library", library)
  private$remote_types <- remote_types %||% default_remote_types()

  if (!is.null(library)) {
    mkdirp(library[1], msg = "Creating library directory")
    library[1] <- normalizePath(library[1])
  }
  mkdirp(private$download_cache <- private$config$get("cache_dir"))

  raw <- jsonlite::fromJSON(readLines(lockfile), simplifyVector = FALSE)
  if (raw$lockfile_version != 1) {
    throw(pkg_error(
      "This version of {pak_or_pkgdepends()} (version {pakx_version()})
       does not support lockfile version {raw$lockfile_version}.",
      i = "This lockfile was probably created by a newer version."
    ))
  }

  pkgs <- raw$packages
  refs <- vcapply(pkgs, "[[", "ref")

  sysreqs_packages <- lapply(pkgs, function(x) {
    sq <- x[["sysreqs_packages"]]
    sq <- lapply(sq, function(sq1) {
      sq1[["sysreq"]] <- unlist(sq1[["sysreq"]])
      sq1[["packages"]] <- unlist(sq1[["packages"]])
      sq1
    })
    sq
  })

  soldata <- data_frame(
    ref = refs,
    type = vcapply(pkgs, "[[", "type"),
    direct = vlapply(pkgs, "[[", "direct"),
    directpkg = vlapply(pkgs, "[[", "directpkg"),
    status = "OK",
    package = vcapply(pkgs, "[[", "package"),
    version = vcapply(pkgs, "[[", "version"),
    license = vcapply(pkgs, function(x) x$license %||% NA_character_),
    needscompilation = vlapply(pkgs, function(x) x$needscompilation %||% NA),
    sha256 = vcapply(pkgs, function(x) x$sha256 %||% NA_character_),
    filesize = viapply(pkgs, function(x) x$filesize %||% NA_integer_),
    built = vcapply(pkgs, function(x) x$built %||% NA_character_),
    platform = vcapply(pkgs, "[[", "platform"),
    rversion = vcapply(pkgs, "[[", "rversion"),
    target = vcapply(pkgs, "[[", "target"),
    dependencies = lapply(
      pkgs,
      function(x) unlist(x$dependencies) %||% character()
    ),
    sources = lapply(pkgs, function(x) unlist(x$sources)),
    metadata = lapply(pkgs, function(x) unlist(x$metadata)),
    dep_types = lapply(pkgs, function(x) unlist(x$dep_types)),
    remote = parse_pkg_refs(refs),
    cache_status = "miss",
    lib_status = "new",
    old_version = NA_character_,
    new_version = vcapply(pkgs, "[[", "version"),
    extra = list(list()),
    install_args = lapply(
      pkgs,
      function(x) unlist(x$install_args) %||% character()
    ),
    repotype = vcapply(pkgs, function(x) x$repotype %||% NA_character_),
    params = lapply(pkgs, function(x) unlist(x$params)),
    sysreqs = vcapply(pkgs, function(x) x[["sysreqs"]] %||% NA_character_),
    sysreqs_packages = sysreqs_packages
  )

  private$refs <- refs[soldata$direct]
  private$remotes <- soldata$remote
  private$dirty <- FALSE

  private$cache <- list(
    metadata = NULL,
    package = pkgcache::package_cache$new(private$config$get(
      "package_cache_dir"
    )),
    installed = NULL
  )

  sysreqs <- raw$sysreqs
  if (!is.null(sysreqs)) {
    sysreqs$packages <- unlist(sysreqs$packages)
    sysreqs["pre_install"] <- list(unlist(sysreqs$pre_install))
    sysreqs["post_install"] <- list(unlist(sysreqs$post_install))
    sysreqs["install_scripts"] <- list(unlist(sysreqs$install_scripts))
  }

  private$resolution <- list(result = soldata)
  private$solution <- list(
    result = structure(
      class = c("pkg_solution_result", "list"),
      list(
        status = "OK",
        data = soldata,
        problem = list(pkgs = soldata),
        solution = NULL,
        sysreqs = sysreqs
      )
    )
  )

  invisible(self)
}

pkgplan_get_total_files <- function(self, private) {
  nrow(private$resolution$result)
}

pkgplan_print <- function(self, private, ...) {
  cat("<pkg_plan>\n")

  ## refs
  refs <- self$get_refs()
  cat(
    strwrap(
      paste0("- refs: ", paste(backtick(refs), collapse = ", ")),
      indent = 0,
      exdent = 4
    ),
    sep = "\n"
  )

  ## library
  if (!is.null(private$config$get("library"))) {
    cat(
      "- library:",
      paste(backtick(private$config$get("library")), collapse = ", "),
      "\n"
    )
  }

  ## resolution
  if (self$has_resolution()) {
    if (self$has_clean_resolution()) {
      cat("- has resolution\n")
    } else {
      cat("- has resolution, with errors\n")
    }
  }

  ## solution
  if (!is.null(private$solution$result)) {
    if (private$solution$result$status == "OK") {
      cat("- has solution\n")
    } else {
      cat("- has solution, with errors\n")
    }
  }

  invisible(self)
}

pkgplan_update <- function(self, private) {
  libstat <- make_installed_cache(private$config$get("library"))$pkgs
  for (i in seq_len(nrow(private$solution$result$data))) {
    pkg <- private$solution$result$data$package[i]
    if (!pkg %in% libstat$package) {
      next
    }

    installed <- libstat[match(pkg, libstat$package), ]
    solution <- private$solution$result$data[i, ]

    par <- solution$params[[1]]
    if (is_true_param(par, "reinstall") || is_true_param(par, "nocache")) {
      next
    }

    if (!installedok_remote(installed, solution)) {
      next
    }

    private$solution$result$data$ref[i] <- installed$ref
    private$solution$result$data$type[i] <- "installed"
    private$solution$result$data$sources[[i]] <- character()
    private$solution$result$data$remote[[i]] <- parse_pkg_ref(installed$ref)
    private$solution$result$data$cache_status[i] <- NA
    private$solution$result$data$lib_status[i] <- "current"
    private$solution$result$data$old_version[[i]] <- installed$version
    private$solution$result$data$new_version[[i]] <- NA
  }
}

pkgplan_update_sysreqs <- function(self, private) {
  if (!private$config$get("sysreqs")) {
    return(invisible())
  }
  # Stop here is no sysreqs at all, no need to look up system packages
  sys <- self$get_solution()$data$sysreqs_packages
  if (length(unlist(sys)) == 0) {
    return(invisible())
  }

  private$solution$result$data$sysreqs_packages <-
    sysreqs_update_state(sys)
}

test_that("parse_remote", {
  path <- get_fixture("foobar_1.0.0.tar.gz")
  ref <- paste0("local::", path)
  pr <- parse_pkg_ref(ref)
  expect_true(is.list(pr))
  expect_equal(pr$path, path)
  expect_equal(pr$ref, ref)
  expect_equal(pr$type, "local")
})

test_that("resolve_remote", {
  setup_fake_apps()

  conf <- current_config()
  cache <- list(
    package = NULL,
    metadata = pkgcache::get_cranlike_metadata_cache()
  )

  ## Absolute path
  path <- get_fixture("foobar_1.0.0.tar.gz")
  ref <- paste0("local::", path)
  res <- synchronise(resolve_remote_local(
    parse_pkg_ref(ref),
    TRUE,
    conf,
    cache,
    dependencies = FALSE
  ))

  res$metadata <- as.list(res$metadata)
  # seems really hard to remove the path
  res$sources[[1]] <- basename(res$sources[[1]])
  expect_snapshot(res)

  ## Relative path?
  fix_dir <- fixture_dir()
  ref2 <- paste0("local::", "foobar_1.0.0.tar.gz")
  withr::with_dir(
    fix_dir,
    res <- synchronise(resolve_remote_local(
      parse_pkg_ref(ref2),
      TRUE,
      conf,
      cache,
      dependencies = FALSE
    ))
  )

  res$metadata <- as.list(res$metadata)
  # seems really hard to remove the path
  res$sources[[1]] <- basename(res$sources[[1]])
  expect_snapshot(res)
})

test_that("resolution error", {
  conf <- current_config()
  cache <- list(
    package = NULL,
    metadata = pkgcache::get_cranlike_metadata_cache()
  )

  path <- get_fixture("foobar_10.0.0.tar.gz")
  ref <- paste0("local::", path)
  err <- tryCatch(
    synchronise(
      resolve_remote_local(
        parse_pkg_ref(ref),
        TRUE,
        conf,
        cache,
        dependencies = FALSE
      )
    ),
    error = function(x) x
  )

  expect_s3_class(err, "error")
})

test_that("download_remote", {
  setup_fake_apps()

  dir.create(tmp <- tempfile())
  dir.create(tmp2 <- tempfile())
  on.exit(unlink(c(tmp, tmp2), recursive = TRUE), add = TRUE)

  conf <- current_config()
  conf$platforms <- "macos"
  conf$cache_dir <- tmp
  conf$package_cache_dir <- tmp2
  cache <- list(
    package = pkgcache::package_cache$new(conf$package_cache_dir),
    metadata = pkgcache::get_cranlike_metadata_cache()
  )

  ## Absolute path
  path <- get_fixture("foobar_1.0.0.tar.gz")
  ref <- paste0("local::", path)

  rem <- pkg_plan$new(ref)
  suppressMessages(rem$resolve())
  res <- rem$get_resolution()

  target <- file.path(conf$cache_dir, res$target[1])
  tree <- paste0(target, "-tree")
  mkdirp(dirname(target))
  download <- function(res) {
    download_remote_local(res, target, tree, conf, cache, on_progress = NULL)
  }
  dl1 <- synchronise(download(res[1, ]))
  expect_equal(dl1, "Got")
  expect_true(file.exists(target))

  ## Relative path?
  fix_dir <- fixture_dir()
  ref2 <- paste0("local::", "foobar_1.0.0.tar.gz")

  withr::with_dir(fix_dir, {
    rem <- pkg_plan$new(ref2)
    rem$resolve()
    res <- rem$get_resolution()
  })

  target <- file.path(conf$cache_dir, res$target[1])
  tree <- paste0(target, "-tree")
  mkdirp(dirname(target))
  dl1 <- download_remote_local(
    res,
    target,
    tree,
    conf,
    cache,
    on_progress = NULL
  )

  expect_equal(dl1, "Got")
  expect_true(file.exists(target))
})

test_that("download_remote directory", {
  setup_fake_apps()

  dir.create(tmp <- tempfile())
  dir.create(tmp2 <- tempfile())
  dir.create(tmp3 <- tempfile())
  on.exit(unlink(c(tmp, tmp2, tmp3), recursive = TRUE), add = TRUE)

  conf <- current_config()
  conf$platforms <- "source"
  conf$cache_dir <- tmp
  conf$package_cache_dir <- tmp2
  cache <- list(
    package = pkgcache::package_cache$new(conf$package_cache_dir),
    metadata = pkgcache::get_cranlike_metadata_cache()
  )

  path <- get_fixture("foobar_1.0.0.tar.gz")
  untar(path, exdir = tmp3)
  path <- file.path(tmp3, "foobar")
  ref <- paste0("local::", path)

  rem <- pkg_plan$new(ref)
  suppressMessages(rem$resolve())
  res <- rem$get_resolution()

  target <- file.path(conf$cache_dir, res$target[1])
  tree <- paste0(target, "-tree")
  mkdirp(dirname(target))
  download <- function(res) {
    download_remote_local(res, target, tree, conf, cache, on_progress = NULL)
  }
  dl1 <- synchronise(download(res[1, ]))
  expect_equal(dl1, "Got")
  expect_true(file.exists(tree))
  expect_true(is.dir(tree))
})

test_that("download_remote error", {
  setup_fake_apps()

  dir.create(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  dir.create(tmp2 <- tempfile())
  on.exit(unlink(tmp2, recursive = TRUE), add = TRUE)

  path <- get_fixture("foobar_1.0.0.tar.gz")
  file.copy(path, tmp2)
  ref <- paste0("local::", path2 <- file.path(tmp2, basename(path)), "?nocache")
  r <- pkg_plan$new(
    ref,
    config = list(dependencies = FALSE, cache_dir = tmp)
  )
  r$resolve()
  unlink(path2)
  expect_snapshot(invisible(r$download_resolution()))
  dl <- r$get_resolution_download()

  expect_false(file.exists(dl$fulltarget))
  expect_s3_class(dl, "pkgplan_downloads")
  expect_true(all(dl$ref == sub("?nocache", "", ref, fixed = TRUE)))
  expect_true(all(dl$type == "local"))
  expect_true(all(dl$direct))
  expect_true(all(dl$status == "OK"))
  expect_true(all(dl$package == "foobar"))
  expect_true(all(dl$download_status == "Failed"))
})

test_that("satisfy", {
  expect_snapshot({
    "different package name, FALSE"
    satisfy_remote_local(
      list(package = "foo"),
      list(package = "bar")
    )

    "different type is bad"
    satisfy_remote_local(
      list(
        package = "foo",
        remote = list(list(package = NA_character_)),
        path = "/tmp/foo"
      ),
      list(
        type = "notgood",
        package = "foo",
        path = "/tmp/foo"
      )
    )

    "any package name is good, good"
    satisfy_remote_local(
      list(
        package = "foo",
        remote = list(list(package = NA_character_, path = "/tmp/foo"))
      ),
      list(
        type = "local",
        package = "foo",
        remote = list(list(path = "/tmp/foo"))
      )
    )

    "requested a different package, bad"
    satisfy_remote_local(
      list(
        package = "foo",
        remote = list(list(package = "bar", path = "/tmp/foo"))
      ),
      list(
        type = "local",
        package = "foo",
        remote = list(list(package = "bar", path = "/tmp/foo"))
      )
    )

    "different paths are bad"
    satisfy_remote_local(
      list(
        package = "foo",
        remote = list(list(package = "foo", path = "/tmp/foo"))
      ),
      list(
        type = "local",
        package = "foo",
        remote = list(list(package = "foo", path = "/tmp/foo2"))
      )
    )
  })
})

test_that("installedok", {
  ## Always FALSE, independently of arguments
  expect_false(installedok_remote_local())
})

test_that("parse_pkg_refs, github", {
  cases <- list(
    list("user/repo"),
    list("github::user/repo"),
    list("pkg=user/repo", package = "pkg"),
    list("pkg=github::user/repo", package = "pkg"),
    list("user/repo/subdir", subdir = "subdir"),
    list("user/repo@badcafe", commitish = "badcafe"),
    list("user/repo#123", pull = "123"),
    list("user/repo@*release", release = "*release"),
    list("github::user/repo/subdir", subdir = "subdir"),
    list("github::user/repo@badcafe", commitish = "badcafe"),
    list("github::user/repo#123", pull = "123"),
    list("github::user/repo@*release", release = "*release"),
    list("pkg=user/repo/subdir", package = "pkg", subdir = "subdir"),
    list("pkg=user/repo@badcafe", package = "pkg", commitish = "badcafe"),
    list("pkg=user/repo#123", package = "pkg", pull = "123"),
    list("pkg=user/repo@*release", package = "pkg", release = "*release"),

    # github url cases
    list("git@github.com:user/repo.git"),
    list("git@github.ubc.ca:user/repo.git"),
    list("https://github.com/user/repo"),
    list("https://github.ubc.ca/user/repo"),
    list(
      "https://github.com/user/repo/tree/i-am-a-branch",
      commitish = "i-am-a-branch"
    ),
    list("https://github.com/user/repo/commit/1234567", commitish = "1234567"),
    list("https://github.com/user/repo/pull/108", pull = "108"),
    list(
      "https://github.com/user/repo/releases/tag/1.0.0",
      commitish = "1.0.0"
    ),
    list("https://github.com/user/repo/releases/latest", release = "latest"),
    list("https://github.com/user/repo/releases/latest", release = "latest"),
    list("https://github.com/foo/bar", username = "foo", repo = "bar"),
    list("git@github.com:foo/bar.git", username = "foo", repo = "bar"),

    # Username and repo can have hyphens in them
    list(
      "git@github.com:foo-bar/baz-qux.git",
      username = "foo-bar",
      repo = "baz-qux"
    )
  )

  for (case in cases) {
    expect_equal_named_lists(
      p <- parse_pkg_refs(case[[1]])[[1]],
      utils::modifyList(
        list(
          package = case$repo %||% "repo",
          username = "user",
          repo = "repo",
          subdir = "",
          commitish = "",
          pull = "",
          release = "",
          ref = case[[1]],
          type = "github",
          params = character()
        ),
        case[-1]
      )
    )
    expect_s3_class(p, c("remote_ref_github", "remote_ref"))
  }
})

test_that("github url regexes", {
  cases <- list(
    list("https://github.com/u/repo.git", c(username = "u", repo = "repo")),
    list("https://github.com/u/re.po", c(username = "u", repo = "re.po")),
    list("https://github.com/u/re.po.git", c(username = "u", repo = "re.po"))
  )
  for (c in cases) {
    m <- re_match(c[[1]], github_url_rx())
    for (n in names(c[[2]])) {
      expect_equal(c[[2]][[n]], m[[n]])
    }
  }
})

test_that("resolve_remote", {
  setup_fake_apps()
  setup_fake_gh_app()

  refs <- c(
    "gaborcsardi/secret",
    "gaborcsardi/secret@x",
    "r-lib/crayon",
    "github::r-lib/crayon",
    "crayon=r-lib/crayon",
    "crayon=github::r-lib/crayon",
    "wesm/feather/R",
    "r-lib/crayon@b5221ab0246050",
    "r-lib/crayon#79",
    # TODO: fix "r-lib/testthat@*release",
    "r-lib/pkgconfig#7"
  )

  # set library to avoid installed packages
  suppressMessages(
    plan <- new_pkg_installation_proposal(
      refs,
      config = list(library = tempfile())
    )
  )
  suppressMessages(plan$resolve())

  res <- plan$get_resolution()
  res <- res[order(res$ref), ]

  res[["sources"]] <- sub("^http://127.0.0.1:[0-9]+/", "", unlist(res$sources))
  expect_snapshot(res[, c("ref", "sources")])

  expect_snapshot(
    res$metadata,
    transform = transform_local_port
  )
})

test_that("failed resolution", {
  setup_fake_gh_app()

  nonrepo <- paste0(basename(tempfile()), "/", basename(tempfile()))
  r <- pkg_plan$new(nonrepo)
  suppressMessages(r$resolve())
  res <- r$get_resolution()

  expect_true(all(res$status == "FAILED"))

  ## Existing repo, no R package there

  r <- pkg_plan$new("github::r-lib/crayon/R")
  suppressMessages(r$resolve())
  res <- r$get_resolution()

  expect_true(all(res$status == "FAILED"))
})

test_that("download_remote", {
  setup_fake_gh_app()

  old_cache <- Sys.getenv("R_PKG_CACHE_DIR")
  on.exit(Sys.setenv(R_PKG_CACHE_DIR = old_cache))
  Sys.setenv(R_PKG_CACHE_DIR = cache <- tempfile())
  on.exit(unlink(cache, recursive = TRUE), add = TRUE)

  dir.create(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  sha <- "b5221ab024605019800ddea474f7a0981a4d53f719f5af2b1af627b34e0760b2"
  ref <- paste0("github::r-lib/crayon@", sha)
  r <- pkg_plan$new(
    ref,
    config = list(dependencies = FALSE, cache_dir = tmp)
  )

  ## -----------------------------------------------------
  ## We get the tree zip first
  expect_error(suppressMessages(r$resolve()), NA)
  expect_error(suppressMessages(r$download_resolution()), NA)
  dl <- r$get_resolution_download()

  expect_s3_class(dl, "pkgplan_downloads")
  expect_true(dl$ref == ref)
  expect_true(dl$type == "github")
  expect_true(dl$direct)
  expect_true(dl$status == "OK")
  expect_true(dl$package == "crayon")
  expect_true(dl$download_status == "Got")

  ## We indeed downloaded a tree
  expect_false(file.exists(dl$fulltarget))
  expect_true(file.exists(dl$fulltarget_tree))

  ## It was added to the cache
  expect_equal(pkgcache::pkg_cache_list()$sha256, sha)

  ## -----------------------------------------------------
  ## Now the tree is in the cache, so it should come from there
  r <- pkg_plan$new(
    ref,
    config = list(dependencies = FALSE, cache_dir = tmp)
  )

  expect_error(suppressMessages(r$resolve()), NA)
  expect_error(suppressMessages(r$download_resolution()), NA)
  dl <- r$get_resolution_download()

  expect_s3_class(dl, "pkgplan_downloads")
  expect_true(dl$ref == ref)
  expect_true(dl$type == "github")
  expect_true(dl$direct)
  expect_true(dl$status == "OK")
  expect_true(dl$package == "crayon")
  expect_true(dl$download_status == "Had")

  expect_false(file.exists(dl$fulltarget))
  expect_true(file.exists(dl$fulltarget_tree))

  ## Still in the cache
  expect_equal(pkgcache::pkg_cache_list()$sha256, sha)

  ## -----------------------------------------------------
  ## Put a (fake) built version in the cache, to similate a cache hit
  file.copy(dl$fulltarget_tree, dl$fulltarget)
  pkgcache::pkg_cache_add_file(
    NULL,
    dl$fulltarget,
    dl$target,
    package = "crayon",
    sha = sha,
    built = TRUE,
    vignettes = TRUE
  )
  unlink(c(dl$fulltarget, dl$fulltarget_tree))

  r <- pkg_plan$new(
    ref,
    config = list(dependencies = FALSE, cache_dir = tmp)
  )

  expect_error(suppressMessages(r$resolve()), NA)
  expect_error(suppressMessages(r$download_resolution()), NA)
  dl <- r$get_resolution_download()

  expect_s3_class(dl, "pkgplan_downloads")
  expect_true(dl$ref == ref)
  expect_true(dl$type == "github")
  expect_true(dl$direct)
  expect_true(dl$status == "OK")
  expect_true(dl$package == "crayon")
  expect_true(dl$download_status == "Had")

  expect_true(file.exists(dl$fulltarget))
  expect_false(file.exists(dl$fulltarget_tree))
})

test_that("satisfies_remote", {
  res <- make_fake_resolution(
    `github::r-lib/crayon` = list(
      extra = list(list(remotesha = "badcafe")),
      package = "crayon",
      version = "1.0.0"
    )
  )

  ## Different package name
  bad1 <- make_fake_resolution(
    `github::r-lib/crayon` = list(
      package = "crayon2"
    )
  )
  expect_false(ans <- satisfy_remote_github(res, bad1))
  expect_match(attr(ans, "reason"), "names differ")

  ## Installed ref without sha
  fake_desc <- desc::desc("!new")
  bad2 <- make_fake_resolution(
    `installed::foobar` = list(
      extra = list(list(description = fake_desc)),
      package = "crayon"
    )
  )
  expect_false(ans <- satisfy_remote_github(res, bad2))
  expect_match(attr(ans, "reason"), "Installed package sha mismatch")

  ## Installed ref with different sha
  fake_desc <- desc::desc("!new")
  fake_desc$set("RemoteSha" = "notsobad")
  bad3 <- make_fake_resolution(
    `installed::foobar` = list(
      extra = list(list(description = fake_desc)),
      package = "crayon"
    )
  )
  expect_false(ans <- satisfy_remote_github(res, bad3))
  expect_match(attr(ans, "reason"), "Installed package sha mismatch")

  ## Other package, different sha
  bad4 <- make_fake_resolution(
    `bioc::bar` = list(
      package = "crayon",
      extra = list(list(remotesha = "notsobad"))
    )
  )
  expect_false(ans <- satisfy_remote_github(res, bad4))
  expect_match(attr(ans, "reason"), "Candidate package sha mismatch")

  ## Corrent sha, GitHub
  ok1 <- make_fake_resolution(
    `installed::foo` = list(
      extra = list(list(remotesha = "badcafe")),
      package = "crayon",
      version = "1.0.0"
    )
  )
  expect_true(satisfy_remote_github(res, ok1))

  ## Corrent sha, another type
  ok2 <- make_fake_resolution(
    `local::bar` = list(
      package = "crayon",
      extra = list(list(remotesha = "badcafe"))
    )
  )
  expect_true(ans <- satisfy_remote_github(res, ok2))

  ## Local packages are OK, even w/o sha
  ## https://github.com/r-lib/pkgdepends/issues/229
  ok3 <- make_fake_resolution(
    `local::bar` = list(
      package = "crayon"
    )
  )
  expect_true(ans <- satisfy_remote_github(res, ok3))
})

test_that("satisfies_remote 2", {
  ## installed is not OK, if reinstall is requested
  res <- make_fake_resolution(
    `github::r-lib/crayon?reinstall` = list(
      extra = list(list(remotesha = "badcafe")),
      package = "crayon",
      version = "1.0.0",
      params = c(reinstall = "")
    )
  )
  bad1 <- make_fake_resolution(
    `installed::foo` = list(
      extra = list(list(remotesha = "badcafe")),
      package = "crayon",
      version = "1.0.0"
    )
  )
  expect_false(ans <- satisfy_remote_github(res, bad1))
})

test_that("satisfies_remote 3", {
  # other refs are OK, as long as the sha matches
  ## installed is not OK, if reinstall is requested
  res <- make_fake_resolution(
    `github::r-lib/crayon` = list(
      extra = list(list(remotesha = "badcafe")),
      package = "crayon",
      version = "1.0.0",
      params = c(reinstall = "")
    )
  )

  ok1 <- make_fake_resolution(
    `url::foo` = list(
      extra = list(list(remotesha = "badcafe")),
      package = "crayon",
      version = "1.0.0"
    )
  )
  expect_true(ans <- satisfy_remote_github(res, ok1))

  bad1 <- make_fake_resolution(
    `url::foo` = list(
      extra = list(list(remotesha = "baddddd")),
      package = "crayon",
      version = "1.0.0"
    )
  )
  expect_false(ans <- satisfy_remote_github(res, bad1))
})

test_that("installedok", {
  expect_true(installedok_remote_github(
    list(
      package = "foo",
      version = "1.2.3",
      remotesha = "badcafe"
    ),
    list(
      package = "foo",
      version = "1.2.3",
      metadata = list(list(RemoteSha = "badcafe"))
    )
  ))

  expect_false(installedok_remote_github(
    list(
      package = "foo",
      version = "1.2.3",
      remotesha = "badcafe"
    ),
    list(
      package = "foo",
      version = "1.2.3"
    )
  ))
})

test_that("hard example, because easch package has many candidates", {
  skip("unfinished")

  lib <- test_temp_dir()
  install.packages(
    c("dplyr", "rcmdcheck", "qgraph", "usethis", "ggplot2", "knitr"),
    lib = lib
  )

  r <- pkg_plan$new(
    c(
      "tidyverse/dplyr",
      "r-lib/rcmdcheck",
      "cran/qgraph",
      "r-lib/usethis",
      "tidyverse/ggplot2",
      "yihui/knitr",
      "r-lib/rray"
    ),
    lib = lib
  )
  r$resolve()

  pkgs <- r$get_resolution()
  prob <- r$.__enclos_env__$private$create_lp_problem(pkgs, "upgrade")
  sol <- pkgplan_i_solve_lp_problem(prob)

  expect_true(sol$objval < 1e6)
})

fixture <- local({
  get_test_data <- function(expr) {
    code <- enc2utf8(paste0(deparse(expr, backtick = TRUE), collapse = "\n"))
    hash <- cli::hash_md5(code)
    snap <- getOption("testthat.snapshotter")
    list(
      root = testthat::test_path(),
      expr = expr,
      code = code,
      hash = hash,
      file = snap$file,
      test = snap$test,
      seen = snap$test_file_seen
    )
  }

  mkdirp <- function(dir) {
    s <- vlapply(dir, dir.create, recursive = TRUE, showWarnings = FALSE)
    invisible(s)
  }

  find_test_file <- function(ctx, root) {
    ptn <- paste0("^test[-_]", ctx, "[.][Rr]")
    file <- grep(ptn, dir(root), value = TRUE)[1]
    as.character(file)
  }

  context_name <- function(filename) {
    # Remove test- prefix
    filename <- sub("^test[-_]", "", filename)
    # Remove terminal extension
    filename <- sub("[.][Rr]$", "", filename)
    filename
  }

  rstudio_tickle <- function() {
    if (!base::requireNamespace("rstudioapi", quietly = TRUE)) {
      return()
    }

    if (!asNamespace("rstudioapi")$hasFun("executeCommand")) {
      return()
    }

    asNamespace("rstudioapi")$executeCommand("vcsRefresh")
    asNamespace("rstudioapi")$executeCommand("refreshFiles")
  }

  read_lines <- function(path, n = -1L, encoding = "UTF-8") {
    base::readLines(path, n = n, encoding = encoding, warn = FALSE)
  }

  write_lines <- function(text, path) {
    base::writeLines(enc2utf8(text), path, useBytes = TRUE)
  }

  fix_to_md <- function(data) {
    h2 <- paste0("# ", names(data), "\n\n")
    code_block <- function(x) paste0(indent_add(x), collapse = "\n\n---\n\n")
    data <- vapply(data, code_block, character(1))

    paste0(h2, data, "\n\n", collapse = "")
  }

  indent_add <- function(x, prefix = "    ") {
    paste0(prefix, gsub("\n", paste0("\n", prefix), x, fixed = TRUE))
  }

  indent_del <- function(x, prefix = "    ") {
    x <- gsub(paste0("^", prefix), "", x)
    x <- gsub(paste0("\n", prefix), "\n", x)
    x
  }

  compact <- function(x) {
    x[lengths(x) > 0]
  }

  new_fixture_reporter <- function(data) {
    FixtureReporter <- R6::R6Class(
      "FixtureReporter",
      inherit = testthat::Reporter,
      public = list(
        fix_dir = "_fixtures",
        file = NULL,
        test = NULL,
        cur_fixs = NULL,
        files_seen = NULL,
        hashes_seen = NULL,

        initialize = function(files_seen = character()) {
          self$files_seen <- files_seen
        },

        start_file = function(path, test = NULL) {
          self$files_seen <- c(self$files_seen, path)
          self$file <- context_name(path)
          self$cur_fixs <- list()
          if (!is.null(test)) {
            self$start_test(NULL, test)
          }
        },

        start_test = function(context, test) {
          self$test <- test

          if (length(self$cur_fixs[[test]]) > 0) {
            warning("Duplicate test, discarding previous fixtures")
          }
          self$cur_fixs[[test]] <- list()
        },

        add_fixture = function(hash, code) {
          value <- paste(
            collapse = "\n",
            c(
              "Hash",
              paste0("  ", hash),
              "Code",
              paste0("  ", strsplit(code, "\n", fixed = TRUE)[[1]])
            )
          )
          self$hashes_seen <- c(self$hashes_seen, hash)
          self$cur_fixs <- self$fix_append(self$cur_fixs, value)
        },

        end_file = function() {
          dir.create(self$fix_dir, showWarnings = FALSE)
          self$fixs_write(self$cur_fixs)
        },

        end_reporter = function() {
          tests <- testthat::find_test_scripts(".", full.names = FALSE)
          if (all(tests %in% self$files_seen)) {
            rds <- dir(self$fix_dir, pattern = "[.]rds$")
            seen <- paste0(self$hashes_seen, ".rds")
            unused <- setdiff(rds, seen)
            if (length(unused) > 0) {
              message("Deleting unused fixtures: ", unused)
              unlink(file.path(self$fix_dir, unused))
            }
          }
          if (length(dir(self$fix_dir)) == 0) {
            unlink(self$fix_dir, recursive = TRUE)
          }
          rstudio_tickle()
        },

        fix_append = function(data, fix) {
          data[[self$test]] <- c(data[[self$test]], list(fix))
          data
        },

        # File management ----------------------------------------------------------

        fixs_write = function(data, suffix = "") {
          data <- compact(data)
          if (length(data) > 0) {
            out <- fix_to_md(data)
            # trim off last line since write_lines() adds one
            out <- gsub("\n$", "", out)
            write_lines(out, self$fix_path(suffix))
          } else {
            self$fixs_delete(suffix)
          }
        },

        fixs_delete = function(suffix = "") {
          unlink(self$fix_path(suffix))
        },

        fix_path = function(suffix = "") {
          file.path(self$fix_dir, paste0(self$file, suffix, ".md"))
        }
      )
    )

    fixer <- FixtureReporter$new(data$seen)

    fixer
  }

  setup_reporter <- function(data) {
    rep <- testthat::get_reporter()
    if (!isTRUE(rep$capabilities$fixture_setup)) {
      rep$capabilities$fixture_setup <- TRUE
      if (class(rep)[1] != "MultiReporter") {
        return()
      }
      fixer <- new_fixture_reporter(data)
      file <- find_test_file(data$file, data$root)
      if (is.na(file)) {
        stop("Cannot determine test file for fixtures")
      }
      fixer$start_file(file, data$test)
      rep$reporters <- c(rep$reporters, list(fixer))
    }
    wfix <- sapply(rep$reporters, inherits, "FixtureReporter")
    if (sum(wfix) != 1) {
      stop("Must be exactly one FixtureReporter")
    }
    fixer <- rep$reporters[[which(wfix)]]
    fixer$add_fixture(data$hash, data$code)
  }

  get <- function(expr, envir = parent.frame()) {
    expr <- substitute(expr)
    data <- get_test_data(expr)
    if (!is.null(data$file)) {
      setup_reporter(data)
    }
    fxfile <- paste0(data$hash, ".rds")
    fxpath <- file.path(data$root, "_fixtures", fxfile)
    upd <- isTRUE(getOption("fixtures.update"))
    upd1 <- isTRUE(getOption("fixtures.update1"))
    if (upd || upd1) {
      if (upd1) {
        options(fixtures.update1 = NULL)
      }
      value <- eval(expr, envir = envir)
      mkdirp(dirname(fxpath))
      saveRDS(value, fxpath, version = 2L)
      value
    } else {
      tryCatch(
        suppressWarnings(readRDS(fxpath)),
        error = function(err) {
          stop("Cannot read fixture file `", fxpath, "`")
        }
      )
    }
  }

  update <- function(files = NULL, ...) {
    opt <- options(fixtures.update = TRUE)
    on.exit(options(opt), add = TRUE)
    testthat::test_local(filter = files, ...)
  }

  update_next <- function() {
    options(fixtures.update1 = TRUE)
  }

  snap_from_md <- function(lines) {
    lines <- gsub("\r", "", lines, fixed = TRUE)

    h2 <- grepl("^# ", lines)
    tests_group <- cumsum(h2)
    tests <- split(lines[!h2], tests_group[!h2])
    names(tests) <- gsub("^# ", "", lines[h2])

    split_tests <- function(lines) {
      sep <- grepl("^-{3, }", lines)
      case_group <- cumsum(sep)

      # Remove first line and last line, separator, line above and line below
      sep_loc <- which(sep)
      drop <- c(1, sep_loc, sep_loc + 1, sep_loc - 1, length(lines))

      cases <- unname(split(lines[-drop], case_group[-drop]))
      code_unblock <- function(x) paste0(indent_del(x), collapse = "\n")
      vapply(cases, code_unblock, character(1))
    }

    lapply(tests, split_tests)
  }

  read_hashes <- function(path) {
    lines <- readLines(path)
    fixes <- snap_from_md(lines)
    hashes <- sub("^Hash\n[ ]+([a-f0-9]+)\n.*$", "\\1", unlist(fixes))
    unique(sort(unname(hashes)))
  }

  read_all_hashes <- function(root, reports) {
    unique(sort(unlist(lapply(file.path(root, reports), read_hashes))))
  }

  cleanup <- function(force = FALSE) {
    root <- testthat::test_path()
    fixroot <- file.path(root, "_fixtures")
    reports <- dir(fixroot, pattern = "[.]md$")
    tests <- testthat::find_test_scripts(root, full.names = FALSE)
    tests_used <- vcapply(sub("[.]md$", "", reports), find_test_file, root)
    old <- names(tests_used)[which(is.na(tests_used))]
    old <- setdiff(old, "_other")
    if (length(old) > 0) {
      old <- paste0(old, ".md")
      message("Unused test fixture reports: ", paste(old, collapse = ", "))
      reports <- setdiff(reports, old)
    }

    allfix <- dir(fixroot, pattern = "[.]rds$")
    usedfix <- paste0(read_all_hashes(fixroot, reports), ".rds")

    old2 <- setdiff(allfix, usedfix)
    if (length(old2) > 0) {
      message("Unused test fixture files: ", paste(old2, collapse = ", "))
    }

    if (length(old) > 0 || length(old2) > 0) {
      if (!force) {
        message("Call with `force = TRUE` to remove these files")
      } else {
        message("Removing them.")
        unlink(file.path(fixroot, old))
        unlink(file.path(fixroot, old2))
      }
    }
  }

  list(
    get = get,
    update = update,
    update_next = update_next,
    cleanup = cleanup,
    .internal = environment()
  )
})

test_that("current_config runs without error", {
  expect_no_error(current_config())
})

test_that("current_config runs without error when difftime config set", {
  withr::local_envvar(PKG_SYSREQS_DB_UPDATE_TIMEOUT = "30")
  expect_error(
    current_config()$get("sysreqs_db_update_timeout"),
    "Invalid time interval specification"
  )
  withr::local_envvar(PKG_SYSREQS_DB_UPDATE_TIMEOUT = "30s")
  expect_equal(
    current_config()$get("sysreqs_db_update_timeout"),
    as.difftime(30, units = 'secs'),
    ignore_attr = TRUE
  )
  withr::local_envvar(PKG_SYSREQS_DB_UPDATE_TIMEOUT = "30h")
  expect_equal(
    current_config()$get("sysreqs_db_update_timeout"),
    as.difftime(30, units = "hours"),
    ignore_attr = TRUE
  )
})
stopifnot(Sys.getenv('PKG_SYSREQS_DB_UPDATE_TIMEOUT') == '')

# nocov start

#' @description
#' pkgdepends is a toolkit for package dependencies, downloads and
#' installations, to be used in other packages. If you are looking for a
#' package manager, see [pak](https://github.com/r-lib/pak).
#'
#' @useDynLib pkgdepends, .registration = TRUE, .fixes = "c_"
#' @includeRmd tools/doc/README-body.Rmd
"_PACKAGE"

fix_check <- function() {
  R6::R6Class
  processx::process
  invisible()
}

# nocov end

#' Create a tar background process
#'
#' Use an external tar program, if there is a working one, otherwise use
#' the internal implementation.
#'
#' When using the internal implementation, we need to start another R
#' process.
#'
#' @param tarfile Tar file.
#' @param files Files or regular expressions to set what to extract. if
#'   `NULL` then everything is extracted.
#' @param exdir Where to extract the archive. It must exist.
#' @param restore_times Whether to restore file modification times.
#' @param post_process Function to call after the extraction.
#' @param stdout Standard output of process.
#' @param stderr Standard error of process.
#' @param ... Extra arguments for the process constructor.
#' @return The [processx::process] object.
#' @noRd

make_untar_process <- function(
  tarfile,
  files = NULL,
  exdir = ".",
  restore_times = TRUE,
  post_process = NULL,
  stdout = "|",
  stderr = "2>&1",
  ...
) {
  internal <- need_internal_tar()
  if (internal) {
    r_untar_process$new(
      tarfile,
      files,
      exdir,
      restore_times,
      post_process = post_process,
      stdout = stdout,
      stderr = stderr,
      ...
    )
  } else {
    external_untar_process$new(
      tarfile,
      files,
      exdir,
      restore_times,
      post_process = post_process,
      stdout = stdout,
      stderr = stderr,
      ...
    )
  }
}

#' Check if we need to use R's internal tar implementation
#'
#' This is slow, because we need to start an R child process, and the
#' implementation is also very slow. So it is better to use an external tar
#' program, if we can. We test this by trying to uncompress a `.tar.gz`
#' archive using the external program. The name of the tar program is
#' taken from the `TAR` environment variable, if this is unset then `tar`
#' is used.
#'
#' @return Whether we need to use the internal tar implementation.
#' @noRd

need_internal_tar <- local({
  internal <- NULL
  function() {
    if (!is.null(internal)) {
      return(internal)
    }

    mkdirp(tmp <- tempfile())
    on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
    tarfile <- system.file(package = .packageName, "tools", "pkg_1.0.0.tgz")

    tryCatch(
      p <- external_untar_process$new(tarfile, exdir = tmp),
      error = function(e) {
        internal <<- TRUE
      }
    )
    if (!is.null(internal)) {
      return(internal)
    }

    p$wait(timeout = 2000)
    p$kill()
    internal <<- p$get_exit_status() != 0 ||
      !file.exists(file.path(tmp, "pkg", "DESCRIPTION"))
    internal
  }
})

#' R6 class for an external un-tar process
#'
#' @description
#' Uses the system's `tar` program, in a background process.
#'
#' @noRd

external_untar_process <- R6::R6Class(
  "external_untar_process",
  inherit = processx::process,

  public = list(
    #' @details
    #' Start running the background process that extracts the file.
    #'
    #' @param tarfile Path to the `.tar` or `.tar.gz`, etc. file to
    #' uncompress.
    #' @param files List of files to uncompress, see [utils::untar()].
    #' @param exdir Directory to extract the files to.
    #' @param restore_times Whether to restore modification files.
    #' @param tar Name of the external `tar` program. Defaults to
    #' `TAR` environment variable, or `tar` if unset.
    #' @param stdout Standard output of process.
    #' @param stderr Standard error of process.
    #' @param post_process Function to call, once the extraction is
    #' done, or `NULL`
    #' @param ... Extra arguments for the process constructor.
    #' @return New `r_untar_process` object.

    initialize = function(
      tarfile,
      files = NULL,
      exdir = ".",
      restore_times = TRUE,
      tar = Sys.getenv("TAR", "tar"),
      stdout = "|",
      stderr = "2>&1",
      post_process = NULL,
      ...
    ) {
      private$options <- list(
        tarfile = normalizePath(tarfile),
        files = files,
        exdir = exdir,
        restore_times = restore_times,
        tar = tar
      )

      private$options$args <- eup_get_args(private$options)
      super$initialize(
        tar,
        private$options$args,
        post_process = post_process,
        stdout = stdout,
        stderr = stderr,
        ...
      )
      invisible(self)
    }
  ),

  private = list(
    options = NULL
  )
)

#' R6 class for an R un-tar process
#'
#' @description
#' Uses [utils::untar()], in a background process.
#'
#' @noRd

r_untar_process <- R6::R6Class(
  "r_untar_process",
  inherit = callr::r_process,

  public = list(
    #' @details
    #' Start running the background R process that extracts the file.
    #'
    #' @param tarfile Path to the `.tar` or `.tar.gz`, etc. file to
    #' uncompress.
    #' @param files List of files to uncompress, see [utils::untar()].
    #' @param exdir Directory to extract the files to.
    #' @param restore_times Whether to restore modification files.
    #' @param post_process Function to call, once the extraction is
    #' done, or `NULL`
    #' @param stdout Standard output of process.
    #' @param stderr Standard error of process.
    #' @param ... Extra arguments for the process cosntructor.
    #' @return New `r_untar_process` object.

    initialize = function(
      tarfile,
      files = NULL,
      exdir = ".",
      restore_times = TRUE,
      post_process = NULL,
      stdout = "|",
      stderr = "2>&1",
      ...
    ) {
      options <- list(
        tarfile = normalizePath(tarfile),
        files = files,
        exdir = exdir,
        restore_times = restore_times,
        tar = tar,
        post_process = post_process
      )

      process_options <- callr::r_process_options(
        stdout = stdout,
        stderr = stderr
      )
      process_options$func <- function(options) {
        # nocov start
        ret <- utils::untar(
          tarfile = options$tarfile,
          files = options$files,
          list = FALSE,
          exdir = options$exdir,
          compressed = NA,
          restore_times = options$restore_times,
          tar = "internal"
        )

        if (!is.null(options$post_process)) options$post_process() else ret
        # nocov end
      }
      process_options$args <- list(options = options)
      super$initialize(process_options)
    }
  ),

  private = list(
    options = NULL
  )
)

eup_get_args <- function(options) {
  c(
    "-x",
    "-f",
    path_norm(options$tarfile),
    "-C",
    path_norm(options$exdir),
    # do not restore ownership, this is problematic on some mounts, e.f. sshfs
    "-o",
    get_untar_decompress_arg(options$tarfile),
    if (!options$restore_times) "-m",
    options$files
  )
}

get_untar_decompress_arg <- function(tarfile) {
  type <- detect_package_archive_type(tarfile)
  switch(
    type,
    "gzip" = "-z",
    "bzip2" = "-j",
    "xz" = "-J",
    "zip" = throw(pkg_error(
      "{.path {tarfile}} is not a tar file, it looks like a zip file"
    )),
    "unknown" = character()
  )
}

detect_package_archive_type <- function(file) {
  buf <- readBin(file, what = "raw", n = 6)
  if (is_gzip(buf)) {
    "gzip"
  } else if (is_zip(buf)) {
    "zip"
  } else if (is_bzip2(buf)) {
    "bzip2"
  } else if (is_xz(buf)) {
    "xz"
  } else {
    "unknown"
  }
}

is_gzip <- function(buf) {
  if (!is.raw(buf)) {
    buf <- readBin(buf, what = "raw", n = 3)
  }
  length(buf) >= 3 &&
    buf[1] == 0x1f &&
    buf[2] == 0x8b &&
    buf[3] == 0x08
}

is_bzip2 <- function(buf) {
  if (!is.raw(buf)) {
    buf <- readBin(buf, what = "raw", n = 3)
  }
  length(buf) >= 3 &&
    buf[1] == 0x42 &&
    buf[2] == 0x5a &&
    buf[3] == 0x68
}

is_xz <- function(buf) {
  if (!is.raw(buf)) {
    buf <- readBin(buf, what = "raw", n = 6)
  }
  length(buf) >= 6 &&
    buf[1] == 0xFD &&
    buf[2] == 0x37 &&
    buf[3] == 0x7A &&
    buf[4] == 0x58 &&
    buf[5] == 0x5A &&
    buf[6] == 0x00
}

is_zip <- function(buf) {
  if (!is.raw(buf)) {
    buf <- readBin(buf, what = "raw", n = 4)
  }
  length(buf) >= 4 &&
    buf[1] == 0x50 &&
    buf[2] == 0x4b &&
    (buf[3] == 0x03 || buf[3] == 0x05 || buf[5] == 0x07) &&
    (buf[4] == 0x04 || buf[4] == 0x06 || buf[4] == 0x08)
}

make_uncompress_process <- function(archive, exdir = ".", ...) {
  type <- detect_package_archive_type(archive)

  if (type == "unknown") {
    throw(pkg_error(
      "Cannot extract {.path {archive}}, unknown archive type.",
      .class = "install_input_error"
    ))
  }

  if (type == "zip") {
    make_unzip_process(archive, exdir = exdir)
  } else {
    make_untar_process(archive, exdir = exdir)
  }
}

# This is similar but uses the async framework

run_uncompress_process <- function(archive, exdir = ".", ...) {
  type <- detect_package_archive_type(archive)
  if (type == "unknown") {
    throw(pkg_error(
      "Cannot extract {.path {archive}}, unknown archive type.",
      .class = "install_input_error"
    ))
  }

  stdout <- tempfile()
  if (type == "zip") {
    external_process(
      make_unzip_process,
      zipfile = archive,
      exdir = exdir,
      stdout = stdout,
      stderr = stdout
    )
  } else {
    external_process(
      make_untar_process,
      tarfile = archive,
      exdir = exdir,
      stdout = stdout,
      stderr = stdout
    )
  }
}

test_that("is_gzip, is_bzip2, is_xz, iz_zip", {
  cases <- list(
    list("is_gzip", "xxx.gz", 3),
    list("is_bzip2", "xxx.bz2", 3),
    list("is_xz", "xxx.xz", 6),
    list("is_zip", "xxx.zip", 4)
  )

  lapply(cases, function(case) {
    fun <- get(case[[1]])
    arch <- system.file(package = "pkgdepends", "tools", case[[2]])
    expect_true(fun(arch))

    buf <- readBin(arch, what = "raw", n = case[[3]])
    expect_true(fun(buf))
    expect_false(fun(utils::head(buf, -1)))

    others <- setdiff(c("is_gzip", "is_bzip2", "is_xz", "is_zip"), case[[1]])
    for (ofun in others) {
      expect_false(get(ofun)(arch))
      expect_false(get(ofun)(buf))
    }
  })
})

test_that("detect_package_archive_type", {
  cases <- list(
    list("gzip", "xxx.gz"),
    list("bzip2", "xxx.bz2"),
    list("xz", "xxx.xz"),
    list("zip", "xxx.zip"),
    list("unknown", "xxx")
  )

  lapply(cases, function(case) {
    arch <- system.file(package = "pkgdepends", "tools", case[[2]])
    expect_equal(detect_package_archive_type(arch), case[[1]])
  })
})

test_that("get_untar_decompress_arg", {
  cases <- list(
    list("-z", "xxx.gz"),
    list("-j", "xxx.bz2"),
    list("-J", "xxx.xz"),
    list(character(), "xxx")
  )

  lapply(cases, function(case) {
    arch <- system.file(package = "pkgdepends", "tools", case[[2]])
    expect_identical(get_untar_decompress_arg(arch), case[[1]])
  })

  zip <- system.file(package = "pkgdepends", "tools", "xxx.zip")
  expect_snapshot(
    error = TRUE,
    get_untar_decompress_arg(zip),
    transform = function(x) {
      sub("'.*xxx[.]zip'", "'<path>/xxx.zip'", x)
    }
  )
})

test_that("eup_get_args", {
  opts <- list(
    tarfile = system.file(package = "pkgdepends", "tools", "pkg_1.0.0.tgz"),
    files = NULL,
    exdir = "exdir",
    restore_times = TRUE,
    tar = "tar"
  )

  expect_equal(
    eup_get_args(opts),
    c(
      "-x",
      "-f",
      path_norm(opts$tarfile),
      "-C",
      path_norm(opts$exdir),
      "-o",
      "-z"
    )
  )

  ## No need to ungzip
  opts$tarfile <- system.file(package = "pkgdepends", "tools", "xxx")
  expect_equal(
    eup_get_args(opts),
    c("-x", "-f", path_norm(opts$tarfile), "-C", path_norm(opts$exdir), "-o")
  )

  ## Files are specified
  opts$files <- c("this", "that")
  expect_equal(
    eup_get_args(opts),
    c(
      "-x",
      "-f",
      path_norm(opts$tarfile),
      "-C",
      path_norm(opts$exdir),
      "-o",
      opts$files
    )
  )

  ## Do not restore times
  opts$restore_times <- FALSE
  expect_equal(
    eup_get_args(opts),
    c(
      "-x",
      "-f",
      path_norm(opts$tarfile),
      "-C",
      path_norm(opts$exdir),
      "-o",
      "-m",
      opts$files
    )
  )
})

test_that("external_untar_process", {
  if (need_internal_tar()) {
    skip("external R does not work")
  }

  tarfile <- system.file(package = "pkgdepends", "tools", "pkg_1.0.0.tgz")
  mkdirp(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  px <- external_untar_process$new(tarfile, exdir = tmp)
  px$wait(5000)
  px$kill()

  expect_equal(px$get_exit_status(), 0)
  expect_true(file.exists(file.path(tmp, "pkg", "DESCRIPTION")))
})

test_that("r_untar_process", {
  tarfile <- system.file(package = "pkgdepends", "tools", "pkg_1.0.0.tgz")
  mkdirp(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  px <- r_untar_process$new(tarfile, exdir = tmp)
  px$wait(5000)
  px$kill()

  expect_equal(px$get_exit_status(), 0)
  expect_true(file.exists(file.path(tmp, "pkg", "DESCRIPTION")))
})

test_that("make_untar_process", {
  tarfile <- system.file(package = "pkgdepends", "tools", "pkg_1.0.0.tgz")
  mkdirp(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  px <- make_untar_process(tarfile, exdir = tmp)
  px$wait(5000)
  px$kill()

  expect_equal(px$get_exit_status(), 0)
  expect_true(file.exists(file.path(tmp, "pkg", "DESCRIPTION")))
})

test_that("make_untar_process, internal tar", {
  mockery::stub(make_untar_process, "need_internal_tar", TRUE)

  tarfile <- system.file(package = "pkgdepends", "tools", "pkg_1.0.0.tgz")
  mkdirp(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  px <- make_untar_process(tarfile, exdir = tmp)
  px$wait(5000)
  px$kill()

  expect_equal(px$get_exit_status(), 0)
  expect_true(file.exists(file.path(tmp, "pkg", "DESCRIPTION")))
})

### -----------------------------------------------------------------------
### API

parse_remote_github <- function(specs, config, ...) {
  pds <- re_match(specs, github_rx())
  if (any(unk <- is.na(pds$.match))) {
    pds[unk] <- re_match(specs[unk], github_url_rx())
    pds[unk, "subdir"] <- ""
  }

  pds$ref <- pds$.text
  cn <- setdiff(colnames(pds), c(".match", ".text"))
  pds <- pds[, cn]
  pds$type <- "github"
  pds$package <- ifelse(nzchar(pds$package), pds$package, pds$repo)
  lapply(
    seq_len(nrow(pds)),
    function(i) as.list(pds[i, ])
  )
}

resolve_remote_github <- function(
  remote,
  direct,
  config,
  cache,
  dependencies,
  ...
) {
  force(direct)
  force(dependencies)
  ## Get the DESCRIPTION data, and the SHA we need
  type_github_get_data(remote)$then(function(resp) {
    data <- list(
      desc = resp$description,
      sha = resp$sha,
      remote = remote,
      direct = direct,
      dependencies = dependencies[[2 - direct]]
    )
    type_github_make_resolution(data)
  })
}

download_remote_github <- function(
  resolution,
  target,
  target_tree,
  config,
  cache,
  which,
  on_progress
) {
  ## A GitHub package needs to be built, from the downloaded repo
  ## If we are downloading a solution, then we skip building the vignettes,
  ## because these will be built later by pkginstall.
  ##
  ## We cache both the downloaded repo snapshot and the built package in
  ## the package cache. So this is how we go:
  ##
  ## 1. If there is a locally built bianry in the cache (including
  ##    vignettes if they are requested), then we use that.
  ## 2. If there is a built package in the cache (including vignettes
  ##    if they are needed), then we use that.
  ## 3. If there is a repo snapshot in the cache, we build an R package
  ##    from it. (Add also add it to the cache.)
  ## 4. Otherwise we download the repo, add it to the cache, build the
  ##    R package, and add that to the cache as well.

  package <- resolution$package
  sha <- resolution$extra[[1]][["remotesha"]] %||% NA_character_
  need_vignettes <- which == "resolution"
  nocache <- is_true_param(resolution$params[[1]], "nocache")
  source <- is_true_param(resolution$params[[1]], "source")

  # in case there is a leftover package/tree here
  unlink(c(target, target_tree), recursive = TRUE)

  ## 1. Check for a binary package

  if (!nocache && !source) {
    ptfm <- current_r_platform()
    hit <- cache$package$copy_to(
      target,
      package = package,
      sha256 = sha,
      built = TRUE,
      platform = ptfm,
      rversion = current_r_version(),
      .list = c(if (need_vignettes) c(vignettes = TRUE))
    )

    if (nrow(hit)) {
      return(paste("Had", ptfm)) # TODO: untested currently
    }
  }

  ## 2. Check if we have a built package in the cache. We do not check the
  ## ref or the type, so the package could have been built from a local
  ## ref or from another repo. As long as the sha is the same, we are
  ## fine. If we don't require vignetted, then a package with or without
  ## vignettes is fine.

  if (!nocache) {
    hit <- cache$package$copy_to(
      target,
      package = package,
      sha256 = sha,
      built = TRUE,
      .list = c(if (need_vignettes) c(vignettes = TRUE))
    )
    if (nrow(hit)) {
      "!DEBUG found GH `resolution$ref`@`sha` in the cache"
      return("Had")
    }
  }

  ## 3. Check if we have a repo snapshot in the cache.

  rel_target <- resolution$target
  if (!nocache) {
    subdir <- resolution$remote[[1]]$subdir
    hit <- cache$package$copy_to(
      target_tree,
      package = package,
      sha256 = sha,
      built = FALSE
    )
    if (nrow(hit)) {
      "!DEBUG found GH zip for `resolution$ref`@`sha` in the cache"
      return("Had")
    }
  }

  ## 4. Need to download the repo

  "!DEBUG Need to download GH package `resolution$ref`@`sha`"
  urls <- resolution$sources[[1]]
  rel_zip <- paste0(rel_target, "-t")
  type_github_download_repo(
    urls,
    target_tree,
    rel_zip,
    sha,
    package,
    cache,
    on_progress,
    nocache
  )$then(function() {
    "!DEBUG Building package `resolution$package`"
    return("Got")
  })
}

type_github_download_repo <- function(
  urls,
  repo_zip,
  rel_zip,
  sha,
  package,
  cache,
  on_progress,
  nocache
) {
  urls
  repo_zip
  sha
  package
  cache
  on_progress
  nocache
  ## TODO: progress
  headers <- type_github_get_headers()
  download_file(
    urls,
    repo_zip,
    on_progress = on_progress,
    headers = headers
  )$then(function() {
    if (!nocache) {
      cache$package$add(
        repo_zip,
        rel_zip,
        package = package,
        sha = sha,
        built = FALSE
      )
    }
    "Got"
  })
}

## ----------------------------------------------------------------------

satisfy_remote_github <- function(resolution, candidate, config, ...) {
  ## 1. package name must match
  if (resolution$package != candidate$package) {
    return(structure(FALSE, reason = "Package names differ"))
  }

  ## 1. installed ref is good, if it has the same sha
  if (candidate$type == "installed") {
    want_reinst <- is_true_param(resolution$params[[1]], "reinstall")
    if (want_reinst) {
      return(structure(FALSE, reason = "Re-install requested"))
    }
    sha1 <- candidate$extra[[1]][["remotesha"]] %||% NA_character_
    sha2 <- resolution$extra[[1]][["remotesha"]] %||% NA_character_
    ok <- is_string(sha1) && is_string(sha2) && same_sha(sha1, sha2)
    if (!ok) {
      return(structure(FALSE, reason = "Installed package sha mismatch"))
    } else {
      return(TRUE)
    }
  }

  ## 2. local packages satisfy a GH remote
  ## See https://github.com/r-lib/pkgdepends/issues/229
  if (candidate$type == "local") {
    return(TRUE)
  }

  ## 3. other refs are also good, as long as they have the same sha
  sha1 <- (if (is.list(candidate$extra[[1]])) {
    candidate$extra[[1]][["remotesha"]]
  }) %||%
    NA_character_
  sha2 <- resolution$extra[[1]][["remotesha"]] %||% NA_character_
  ok <- is_string(sha1) && is_string(sha2) && same_sha(sha1, sha2)
  if (!ok) {
    return(structure(FALSE, reason = "Candidate package sha mismatch"))
  } else {
    return(TRUE)
  }
}

installedok_remote_github <- function(installed, solution, config, ...) {
  identical(installed$package, solution$package) &&
    identical(installed$version, solution$version) &&
    identical(installed[["remotesha"]], solution$metadata[[1]][["RemoteSha"]])
}

## ----------------------------------------------------------------------
## Internal functions

type_github_builtin_token <- function() {
  pats <- c(
    paste0("3687d8b", "b0556b7c3", "72ba1681d", "e5e689b", "3ec61279"),
    paste0("8ffecf5", "13a136f3d", "23bfe46c4", "2d67b3c", "966baf7b")
  )
  once_per_session(cli::cli_alert_warning(c(
    "Using bundled GitHub PAT. ",
    "Please add your own PAT using {.code gitcreds::gitcreds_set()}."
  )))
  sample(pats, 1)
}

type_github_get_headers <- function() {
  headers <- c(
    "Accept" = "application/vnd.github.v3+json",
    "Content-Type" = "application/json; charset=utf-8"
  )

  token <- NA_character_
  if (Sys.getenv("CI", "") != "") {
    token <- Sys.getenv("CI_GITHUB_TOKEN", NA_character_)
  }
  if (is.na(token)) {
    token <- tryCatch(
      gitcreds_get()$password,
      error = function(e) NA_character_
    )
  }
  if (is.na(token)) {
    token <- type_github_builtin_token()
  }
  headers <- c(headers, c("Authorization" = paste("token", token)))

  headers <- c(headers, c("User-Agent" = "r-lib/pak"))
  headers
}

type_github_get_data <- function(rem) {
  dx <- if (!is.null(rem$pull) && rem$pull != "") {
    type_github_get_data_pull(rem)
  } else if (!is.null(rem$release) && rem$release != "") {
    type_github_get_data_release(rem)
  } else {
    type_github_get_data_ref(rem)
  }

  dx$then(function(data) {
    chain_error(
      dsc <- desc::desc(text = data$desc),
      new_github_baddesc_error(rem, call)
    )
    list(sha = data$sha, description = dsc)
  })
}

type_github_get_data_ref <- function(rem) {
  user <- rem$username
  repo <- rem$repo
  ref <- rem$commitish %|z|% "HEAD"
  subdir <- rem$subdir %&z&% paste0(utils::URLencode(rem$subdir), "/")

  query <- sprintf(
    "{
    repository(owner: \"%s\", name: \"%s\") {
      description: object(expression: \"%s:%sDESCRIPTION\") {
        ... on Blob {
          isBinary
          text
        }
      }
      sha: object(expression: \"%s\") {
        oid
      }
    }
  }",
    user,
    repo,
    ref,
    subdir,
    ref
  )

  github_query(query)$then(function(resp) {
    check_github_response_ref(resp$response, resp$obj, rem, call. = call)
  })$then(function(obj) {
    list(
      sha = obj[[c("data", "repository", "sha", "oid")]],
      desc = obj[[c("data", "repository", "description", "text")]]
    )
  })
}

check_github_response_ref <- function(resp, obj, rem, call.) {
  if (!is.null(obj$errors)) {
    throw(new_github_query_error(rem, resp, obj, call.))
  }
  if (isTRUE(obj[[c("data", "repository", "description", "isBinary")]])) {
    throw(new_github_baddesc_error(rem, call.))
  }
  if (is.null(obj[[c("data", "repository", "sha")]])) {
    throw(new_github_noref_error(rem, call.))
  }
  if (is.null(obj[[c("data", "repository", "description", "text")]])) {
    throw(new_github_no_package_error(rem, call.))
  }
  obj
}

type_github_get_data_pull <- function(rem) {
  call <- sys.call(-1)
  user <- rem$username
  repo <- rem$repo
  pull <- rem$pull
  subdir <- rem$subdir %&z&% paste0(utils::URLencode(rem$subdir), "/")

  query <- sprintf(
    "{
    repository(owner: \"%s\", name: \"%s\") {
      pullRequest(number: %s) {
        headRefOid
        headRef {
          target {
            ... on Commit {
              file(path: \"%sDESCRIPTION\") {
                object {
                  ... on Blob {
                    isBinary
                    text
                  }
                }
              }
            }
          }
        }
      }
    }
  }",
    user,
    repo,
    pull,
    subdir
  )

  github_query(query)$then(function(resp) {
    check_github_response_pull(resp$response, resp$obj, rem, call. = call)
  })$then(function(obj) {
    ref <- obj[[c("data", "repository", "pullRequest", "headRefOid")]]
    txt <- obj[[c(
      "data",
      "repository",
      "pullRequest",
      "headRef",
      "target",
      "file",
      "object",
      "text"
    )]]
    list(sha = ref, desc = txt)
  })
}

check_github_response_pull <- function(resp, obj, rem, call.) {
  if (!is.null(obj$errors)) {
    throw(new_github_query_error(rem, resp, obj, call.))
  }
  # No full coverage here, because unless something goes super wrong,
  # these cases almost never happen.
  if (!is.null(obj$errors)) {
    throw(new_github_query_error(rem, resp, obj, call.)) # nocov
  }

  if (
    isTRUE(obj[[c(
      "data",
      "repository",
      "pullRequest",
      "headRef",
      "target",
      "file",
      "object",
      "isBinary"
    )]])
  ) {
    throw(new_github_baddesc_error(rem, call.)) # nocov
  }
  if (
    is.null(obj[[c(
      "data",
      "repository",
      "pullRequest",
      "headRef",
      "target",
      "file",
      "object"
    )]])
  ) {
    throw(new_github_no_package_error(rem, call.))
  }
  obj
}

type_github_get_data_release <- function(rem) {
  call <- sys.call(-1)
  user <- rem$username
  repo <- rem$repo
  ref <- NULL
  subdir <- rem$subdir %&z&% paste0(utils::URLencode(rem$subdir), "/")

  query <- sprintf(
    "{
    repository(owner: \"%s\", name:\"%s\") {
      latestRelease {
        tagName
        tagCommit {
          oid,
          file(path: \"%sDESCRIPTION\") {
            object {
              ... on Blob {
                isBinary
                text
              }
            }
          }
        }
      }
    }
  }",
    user,
    repo,
    subdir
  )

  github_query(query)$then(function(resp) {
    check_github_response_release(resp$response, resp$obj, rem, call. = call)
  })$then(function(obj) {
    ref <- obj[[c("data", "repository", "latestRelease", "tagCommit", "oid")]]
    txt <- obj[[c(
      "data",
      "repository",
      "latestRelease",
      "tagCommit",
      "file",
      "object",
      "text"
    )]]
    list(sha = ref, desc = txt)
  })
}

check_github_response_release <- function(resp, obj, rem, call.) {
  if (!is.null(obj$errors)) {
    throw(new_github_query_error(rem, resp, obj, call.))
  }
  if (is.null(obj[[c("data", "repository", "latestRelease")]])) {
    throw(new_github_no_release_error(rem, call.))
  }
  if (
    isTRUE(obj[[c(
      "data",
      "repository",
      "latestRelease",
      "tagCommit",
      "file",
      "object",
      "isBinary"
    )]])
  ) {
    throw(new_github_baddesc_error(rem, call.))
  }
  if (
    is.null(obj[[c(
      "data",
      "repository",
      "latestRelease",
      "tagCommit",
      "file",
      "object"
    )]])
  ) {
    throw(new_github_no_package_error(rem, call.))
  }
  obj
}

type_github_make_resolution <- function(data) {
  deps <- resolve_ref_deps(
    data$desc$get_deps(),
    data$desc$get("Remotes"),
    data$desc$get(extra_config_fields(data$desc$fields()))
  )

  url <- Sys.getenv("R_PKG_GITHUB_API_URL", "https://api.github.com")
  proto <- sub(":.*$", "", url)
  host <- sub("^[^:]*://", "", url)
  sha <- data$sha
  sha7 <- substr(sha, 1, 7)
  username <- data$remote$username
  repo <- data$remote$repo
  subdir <- data$remote$subdir %|z|% NULL
  commitish <- data$remote$commitish %|z|% NULL
  pull <- data$remote$pull %|z|% NULL
  release <- data$remote$release %|z|% NULL
  package <- data$desc$get_field("Package")
  version <- data$desc$get_field("Version")
  dependencies <- data$dependencies
  unknown <- deps$ref[deps$type %in% dependencies]
  unknown <- setdiff(unknown, c(base_packages(), "R"))

  meta <- c(
    RemoteType = "github",
    RemoteHost = host,
    RemoteRepo = repo,
    RemoteUsername = username,
    RemotePkgRef = data$remote$ref,
    RemoteRef = if (is.null(pull)) commitish %||% "HEAD" else NULL,
    RemotePull = pull,
    RemoteSha = sha,
    RemoteSubdir = subdir,
    GithubRepo = repo,
    GithubUsername = username,
    GithubRef = if (is.null(pull)) commitish %||% "HEAD" else NULL,
    GitHubPull = pull,
    GithubSHA1 = sha,
    GithubSubdir = subdir
  )

  list(
    ref = data$remote$ref,
    type = data$remote$type,
    direct = data$direct,
    status = "OK",
    package = package,
    version = version,
    license = data$desc$get_field("License", NA_character_),
    sources = sprintf(
      "%s://%s/repos/%s/%s/zipball/%s",
      proto,
      host,
      username,
      repo,
      sha
    ),
    target = sprintf("src/contrib/%s_%s_%s.tar.gz", package, version, sha7),
    remote = list(data$remote),
    deps = list(deps),
    unknown_deps = unknown,
    extra = list(list(remotesha = sha)),
    metadata = meta,
    params = data$remote$params,
    sysreqs = data$desc$get_field("SystemRequirements", "")
  )
}

github_query <- function(
  query,
  url = paste0(
    Sys.getenv("R_PKG_GITHUB_API_URL", "https://api.github.com"),
    "/graphql"
  ),
  headers = character(),
  ...
) {
  query
  url
  headers
  list(...)

  headers <- c(headers, type_github_get_headers())
  data <- tojson$write_str(
    list(query = query),
    opts = list(auto_unbox = TRUE)
  )
  resp <- NULL
  obj <- NULL

  http_post(url, data = data, headers = headers, ...)$catch(
    error = function(err) throw(new_github_offline_error())
  )$then(function(res) {
    resp <<- res
    json <- rawToChar(res$content %||% raw())
    obj <<- if (nzchar(json)) jsonlite::fromJSON(json, simplifyVector = FALSE)
    res
  })$then(http_stop_for_status)$catch(async_http_error = function(err) {
    throw(new_github_http_error(resp, obj), parent = err)
  })$then(function(res) {
    list(response = resp, obj = obj)
  })
}

# -----------------------------------------------------------------------
# Errors

new_github_error <- function(..., call. = NULL) {
  cond <- new_error(..., call. = call.)
  class(cond) <- c("github_error", class(cond))
  cond
}

# No internet?

new_github_offline_error <- function(call. = NULL) {
  new_github_error("Cannot query GitHub, are you offline?", call. = call.)
}

# HTTP error

new_github_http_error <- function(response, obj, call. = NULL) {
  if (
    response$status_code == 401 &&
      nzchar(obj$message) &&
      grepl("Bad credentials", obj$message)
  ) {
    return(new_github_badpat_error(call. = call.))
  }
  new_github_error("GitHub HTTP error", call. = call.)
}

# Error in a query

new_github_query_error <- function(rem, response, obj, call. = NULL) {
  if ("RATE_LIMITED" %in% vcapply(obj$errors, "[[", "type")) {
    return(new_github_ratelimited_error(response, obj, call. = NULL))

    # we don't actually get this response currently
  } else if (
    grepl("Could not resolve to a User", vcapply(obj$errors, "[[", "message"))
  ) {
    return(new_github_nouser_error(rem, obj, call. = call.)) # nocov
  } else if (
    grepl(
      "Could not resolve to a Repository",
      vcapply(obj$errors, "[[", "message")
    )
  ) {
    return(new_github_norepo_error(rem, obj, call. = call.))
  } else if (
    grepl(
      "Could not resolve to a PullRequest",
      vcapply(obj$errors, "[[", "message")
    )
  ) {
    return(new_github_nopr_error(rem, obj, call. = call.))
  }

  # Otherwise some generic code
  ghmsgs <- sub("\\.?$", ".", vcapply(obj$errors, "[[", "message")) # nocov
  msg <- paste0("GitHub error: ", paste0(ghmsgs, collapse = ", ")) # nocov
  new_github_error(msg, call. = call.) # nocov
}

# No such user/org

new_github_nouser_error <- function(rem, obj, call. = NULL) {
  # nocov
  new_github_error(
    # nocov
    "Can't find GitHub user {rem$username}.", # nocov
    call. = call. # nocov
  ) # nocov
}

# No such repo

new_github_norepo_error <- function(rem, obj, call. = NULL) {
  new_github_error(
    sprintf("Can't find GitHub repo %s/%s.", rem$username, rem$repo),
    call. = call.
  )
}
# Not an R package?

new_github_no_package_error <- function(rem, call. = NULL) {
  subdir <- rem$subdir %&z&% paste0(" in directory '", rem$subdir, "'")
  msg <- sprintf(
    "Can't find R package in GitHub repo %s/%s%s",
    rem$username,
    rem$repo,
    subdir
  )
  new_github_error(msg, call. = call.)
}

# Invalid GitHub PAT

new_github_badpat_error <- function(call. = NULL) {
  new_github_error(paste0(
    "Bad GitHub credentials, ",
    "make sure that your GitHub token is valid."
  ))
}

# DESCRIPTION does not parse

new_github_baddesc_error <- function(rem, call. = NULL) {
  subdir <- rem$subdir %&z&% paste0(", in directory `", rem$subdir, "`")
  msg <- sprintf(
    "Can't parse DESCRIPTION file in GitHub repo %s/%s%s",
    rem$username,
    rem$repo,
    subdir
  )
  new_github_error(msg, call. = call.)
}

# No such PR

new_github_nopr_error <- function(rem, obj, call. = NULL) {
  msg <- sprintf(
    "Can't find PR #%s in GitHub repo %s/%s",
    rem$pull,
    rem$username,
    rem$repo
  )
  new_github_error(msg, call. = call.)
}

# No releases
new_github_no_release_error <- function(rem, obj, call. = NULL) {
  msg <- sprintf(
    "Can't find any release in GitHub repo %s/%s.",
    rem$username,
    rem$repo
  )
  new_github_error(msg, call. = call.)
}

# No such branch/tag/ref

new_github_noref_error <- function(rem, call. = NULL) {
  ref <- rem$commitish %|z|% "HEAD"
  msg <- sprintf(
    "Can't find reference @%s in GitHub repo %s/%s.",
    ref,
    rem$username,
    rem$repo
  )
  new_github_error(msg, call. = call.)
}

# Rate limited

new_github_ratelimited_error <- function(response, obj, call. = NULL) {
  headers <- curl::parse_headers_list(response$headers)
  ghmsgs <- sub("\\.?$", ".", vcapply(obj$errors, "[[", "message"))
  msg <- paste0("GitHub error: ", paste0(ghmsgs, collapse = ", "))
  if ("x-ratelimit-reset" %in% names(headers)) {
    reset <- format(
      .POSIXct(headers$`x-ratelimit-reset`, tz = "UTC"),
      usetz = TRUE
    )
    msg <- paste0(msg, " Rate limit will reset at ", reset, ".")
  }
  new_github_error(msg, call. = call.)
}

test_that("type_github_get_data, sha, description", {
  setup_fake_gh_app()

  expect_snapshot(
    synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak")))
  )
  expect_snapshot(
    synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak@HEAD")))
  )
  expect_snapshot(
    synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak@v0.1.2")))
  )
  expect_snapshot(
    synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak@*release")))
  )
  expect_snapshot(
    synchronise(type_github_get_data(parse_pkg_ref(
      "r-lib/pak@e65de1e9630d"
    )))
  )
  expect_snapshot(
    synchronise(type_github_get_data(parse_pkg_ref(
      "r-lib/pak@e65de1e9630dbfcaf1044718b742bf806486b107"
    )))
  )
  expect_snapshot(
    synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak#90")))
  )
  expect_snapshot(
    synchronise(type_github_get_data(parse_pkg_ref(
      "wesm/feather/R@ec40c1eae1ac83b86fc41bb2f5cd916152d19015"
    )))
  )
})

test_that("type_github_get_data, no such user", {
  setup_fake_gh_app()
  expect_snapshot(
    error = TRUE,
    synchronise(type_github_get_data(parse_pkg_ref("r-lib-xxx-xxx/pak")))
  )
  expect_snapshot(
    error = TRUE,
    synchronise(type_github_get_data(parse_pkg_ref("r-lib-xxx-xxx/pak#90")))
  )
})

test_that("type_github_get_data, no such repo", {
  setup_fake_gh_app()
  expect_snapshot(
    error = TRUE,
    synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak-xxx-xxx")))
  )
  expect_snapshot(
    error = TRUE,
    synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak-xxx-xxx#90")))
  )
})

test_that("github_query, invalid PAT", {
  setup_fake_gh_app()

  withr::local_envvar(c(
    GITHUB_PAT_GITHUB_COM = "invalid",
    GITHUB_TOKEN = "invalid",
    GITHUB_PAT = "invalid",
    CI_GITHUB_TOKEN = "invalid"
  ))

  expect_snapshot(
    error = TRUE,
    transform = transform_no_srcref,
    synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak-xxx-xxx")))
  )
  expect_snapshot(
    error = TRUE,
    transform = transform_no_srcref,
    synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak-xxx-xxx#90")))
  )
})

test_that("github_query, no internet", {
  withr::local_envvar(R_PKG_GITHUB_API_URL = "https://127.0.0.1:80")
  expect_snapshot(
    error = TRUE,
    synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak")))
  )
})

test_that("github_query, rate limited", {
  rem <- parse_pkg_ref("r-lib/pak")
  res <- readRDS(test_path("fixtures/gh-rate-limit-response.rds"))
  json <- rawToChar(res$content)
  obj <- jsonlite::fromJSON(json, simplifyVector = FALSE)
  expect_error(
    throw(new_github_query_error(rem, res, obj, NULL)),
    "API rate limit exceeded. Rate limit will reset at",
    class = "github_error"
  )
})

test_that("github_query, access denied", {
  setup_fake_gh_app()
  withr::local_envvar(
    c(
      GITHUB_PAT_GITHUB_COM = NA_character_,
      GITHUB_TOKEN = NA_character_,
      GITHUB_PAT = NA_character_
    )
  )

  expect_snapshot(
    error = TRUE,
    synchronise(type_github_get_data(parse_pkg_ref("gaborcsardi/secret-test")))
  )
})

test_that("cannot find R package on GitHub, no DESCRIPTION", {
  setup_fake_gh_app()

  expect_snapshot(
    error = TRUE,
    synchronise(type_github_get_data(parse_pkg_ref("tidyverse/tidyverse.org")))
  )

  expect_snapshot(
    error = TRUE,
    synchronise(type_github_get_data(parse_pkg_ref("r-lib/cranyon/R")))
  )

  expect_snapshot(
    error = TRUE,
    synchronise(type_github_get_data(parse_pkg_ref("r-lib/crayon/R#79")))
  )
})

test_that("cannot parse DESCRIPTION on GH", {
  setup_fake_gh_app()

  expect_snapshot(
    error = TRUE,
    synchronise(type_github_get_data(parse_pkg_ref("r-lib/bad@main"))),
    transform = transform_no_srcref
  )

  expect_snapshot(
    error = TRUE,
    synchronise(type_github_get_data(parse_pkg_ref("r-lib/bad#100"))),
    transform = transform_no_srcref
  )

  # binary description
  expect_snapshot(
    error = TRUE,
    synchronize(type_github_get_data(parse_pkg_ref("r-lib/bad/bin@main"))),
    transform = transform_no_srcref
  )
})

test_that("http error", {
  withr::local_envvar(
    R_PKG_GITHUB_API_URL = paste0(fake_gh$url(), "/404")
  )

  expect_snapshot(
    error = TRUE,
    transform = transform_no_srcref,
    synchronise(type_github_get_data(parse_pkg_ref("foo/bar")))
  )
})

test_that("no such PR error", {
  setup_fake_gh_app()
  expect_snapshot(
    error = TRUE,
    synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak#89"))),
    transform = transform_no_srcref
  )
})

test_that("no such ref error", {
  setup_fake_gh_app()
  expect_snapshot(
    error = TRUE,
    synchronise(type_github_get_data(parse_pkg_ref(
      "r-lib/pak@bad-ref-no-no-no"
    ))),
    transform = transform_no_srcref
  )
})

test_that("no release error", {
  setup_fake_gh_app()
  expect_snapshot(
    error = TRUE,
    synchronise(type_github_get_data(parse_pkg_ref("r-lib/bad@*release"))),
    transform = transform_no_srcref
  )
})

test_that("builtin token messages once per session", {
  once_per_session(reset = TRUE)
  mockery::stub(type_github_builtin_token, "sample", "builtin-token")
  expect_snapshot(type_github_builtin_token())
  expect_snapshot(type_github_builtin_token())
})

test_that("CI specific token is picked up if set", {
  withr::local_envvar(
    "CI_GITHUB_TOKEN" = "ci-token",
    "CI" = "true"
  )
  expect_snapshot(type_github_get_headers())
})

test_that("builtin token is used if no other token is available", {
  withr::local_envvar(
    "CI" = NA_character_,
    GITHUB_PAT = NA_character_,
    GITHUB_TOKEN = NA_character_,
    GITHUB_PAT_GITHUB_COM = "FAIL"
  )
  mockery::stub(
    type_github_get_headers,
    "type_github_builtin_token",
    "builtin-token"
  )
  expect_snapshot(type_github_get_headers())
})

alert <- function(type, msg, ..., .envir = parent.frame()) {
  switch(
    type,
    success = cli::cli_alert_success(msg, ..., .envir = .envir),
    info = cli::cli_alert_info(msg, ..., .envir = .envir),
    warning = cli::cli_alert_warning(msg, ..., .envir = .envir),
    danger = cli::cli_alert_danger(msg, ..., .envir = .envir)
  )
}

create_progress_bar <- function(state) {
  bar <- new.env(parent = emptyenv())
  bar$spinner <- cli::get_spinner()
  bar$spinner_state <- 1L
  if (should_show_progress_bar()) {
    bar$status <- cli::cli_status("Installing...", .auto_close = FALSE)
  }

  bar$simple <- is_older_rstudio()

  bar
}

update_progress_bar <- function(state, tick = 0) {
  if (is.null(state$progress$status)) {
    return()
  }

  plan <- state$plan
  total <- nrow(plan)
  installed <- sum(plan$install_done)
  built <- sum(plan$build_done)

  building <- sum(buildingl <- !plan$build_done & !is.na(plan$worker_id))
  installing <- sum(!buildingl & !is.na(plan$worker_id))

  chars <- progress_chars()

  xbar <- make_install_bar(installed / total, built / total, width = 15)
  xbuilt <- make_progress_block(state, chars$build, built, total, building)
  xinst <- make_progress_block(state, chars$inst, installed, total, installing)
  xmsg <- make_install_trailing_progress_msg(state)

  w <- cli::console_width()
  v1 <- paste0(xbuilt, " | ", xinst, " | ")
  v2 <- paste0(xbar, " | ", v1)
  v3 <- paste0(v2, xmsg)
  if (cli::ansi_nchar(v3, type = "width") <= w) {
    st <- v3
  } else if (cli::ansi_nchar(v2, type = "width") <= w) {
    st <- v2
  } else {
    st <- v1
  }

  if (state$progress$simple) {
    st <- cli::ansi_strip(st)
  }
  cli::cli_status_update(state$progress$status, st)
}

## p1 <= p2 must hold

make_install_bar <- function(p1, p2, width) {
  width <- width - 2L

  w1 <- if (isTRUE(all.equal(p1, 1))) width else trunc(width * p1)
  w2 <- if (isTRUE(all.equal(p2, 1))) width - w1 else trunc(width * (p2 - p1))

  chars <- progress_chars()
  p1chars <- rep(chars$fill, w1)
  p2chars <- rep(chars$half, w2)
  xchars <- rep(" ", max(width - w1 - w2, 0))
  bar <- paste(
    c(chars$lpar, p1chars, p2chars, xchars, chars$rpar),
    collapse = ""
  )

  if (is_older_rstudio()) bar else cli::col_green(bar)
}

make_progress_block <- function(state, sym, done, total, prog) {
  prgs <- state$progress
  spin <- prgs$spinner$frames[[prgs$spinner_state]]
  prgs$spinner_state <- prgs$spinner_state %% length(prgs$spinner$frames) + 1L
  paste0(
    sym,
    "  ",
    done,
    "/",
    total,
    if (prog) paste0(" ", spin, " ", prog) else "    "
  )
}

done_progress_bar <- function(state) {
  if (!is.null(state$progress$status)) {
    cli::cli_status_clear(state$progress$status)
  }
}

make_install_trailing_progress_msg <- function(state) {
  working <- !is.na(state$plan$worker_id)
  installing <- state$plan$build_done & working
  building <- !state$plan$build_done & working

  building_pkgs <- paste(state$plan$package[building], collapse = ", ")
  installing_pkgs <- paste(state$plan$package[installing], collapse = ", ")

  paste0(
    if (any(building)) paste0("building ", building_pkgs),
    if (any(building) && any(installing)) ", ",
    if (any(installing)) paste0("installing ", installing_pkgs)
  )
}

parse_remote_git <- function(specs, config, ...) {
  pds <- re_match(specs, git_rx())
  pds$ref <- pds$.text
  cn <- setdiff(colnames(pds), c(".match", ".text"))
  pds <- pds[, cn]
  pds$type <- "git"
  pds$dotgit <- ifelse(grepl("[.]git$", pds$repo), ".git", "")
  pds$repo <- sub("[.]git$", "", pds$repo)
  pds$package <- ifelse(nzchar(pds$package), pds$package, pds$repo)
  pds$url <- paste0(
    pds$protocol,
    "://",
    pds$host,
    pds$path,
    pds$repo,
    pds$dotgit
  )
  pds$commitish[pds$commitish == ""] <- "HEAD"
  lapply(
    seq_len(nrow(pds)),
    function(i) as.list(pds[i, ])
  )
}

resolve_remote_git <- function(
  remote,
  direct,
  config,
  cache,
  dependencies,
  ...
) {
  type_git_get_data(remote)$then(function(resp) {
    data <- list(
      desc = resp$description,
      sha = resp$sha,
      remote = remote,
      direct = direct,
      dependencies = dependencies[[2 - direct]]
    )
    type_git_make_resolution(data)
  })
}

download_remote_git <- function(
  resolution,
  target,
  target_tree,
  config,
  cache,
  which,
  on_progress
) {
  # This is similar to github

  package <- resolution$package
  sha <- resolution$extra[[1]][["remotesha"]] %||% NA_character_
  need_vignettes <- which == "resolution"
  nocache <- is_true_param(resolution$params[[1]], "nocache")
  source <- is_true_param(resolution$params[[1]], "source")

  # in case there is a leftover tree here
  unlink(target_tree, recursive = TRUE)

  ## 1. Check for a binary package

  if (!nocache && !source) {
    ptfm <- current_r_platform()
    hit <- cache$package$copy_to(
      target,
      package = package,
      sha256 = sha,
      built = TRUE,
      platform = ptfm,
      rversion = current_r_version(),
      .list = c(if (need_vignettes) c(vignettes = TRUE))
    )

    if (nrow(hit)) {
      return(paste("Had", ptfm)) # TODO: untested currently
    }
  }

  ## 2. Check if we have a built package in the cache. We do not check the
  ## ref or the type, so the package could have been built from a local
  ## ref or from another repo. As long as the sha is the same, we are
  ## fine. If we don't require vignetted, then a package with or without
  ## vignettes is fine.

  if (!nocache) {
    hit <- cache$package$copy_to(
      target,
      package = package,
      sha256 = sha,
      built = TRUE,
      .list = c(if (need_vignettes) c(vignettes = TRUE))
    )
    if (nrow(hit)) {
      return("Had")
    }
  }

  ## 3. Check if we have a repo snapshot in the cache.

  rel_target <- resolution$target
  subdir <- resolution$remote[[1]]$subdir
  if (!nocache) {
    hit <- cache$package$copy_to(
      target_tree,
      package = package,
      sha256 = sha,
      built = FALSE
    )
    if (nrow(hit)) {
      return("Had")
    }
  }

  ## 4. Need to download the repo

  url <- resolution$remote[[1]]$url
  sha <- resolution$metadata[[1]][["RemoteSha"]]
  pkgdir <- file.path(target_tree, resolution$package)
  mkdirp(pkgdir)
  p <- async_git_download_repo(url, ref = sha, output = pkgdir)

  # submodules?
  submodules <- config$get("git-submodules")
  if (submodules) {
    p <- p$then(function(x) async_update_git_submodules_r(pkgdir, subdir))
  }

  p$then(function() {
    "Got"
  })
}

satisfy_remote_git <- function(resolution, candidate, config, ...) {
  ## 1. package name must match
  if (resolution$package != candidate$package) {
    return(structure(FALSE, reason = "Package names differ"))
  }

  ## 2. installed ref is good, if it has the same sha
  if (candidate$type == "installed") {
    want_reinst <- is_true_param(resolution$params[[1]], "reinstall")
    if (want_reinst) {
      return(structure(FALSE, reason = "Re-install requested"))
    }
    sha1 <- candidate$extra[[1]][["remotesha"]] %||% NA_character_
    sha2 <- resolution$extra[[1]][["remotesha"]] %||% NA_character_
    ok <- is_string(sha1) && is_string(sha2) && same_sha(sha1, sha2)
    if (!ok) {
      return(structure(FALSE, reason = "Installed package sha mismatch"))
    } else {
      return(TRUE)
    }
  }

  ## 3. local packages satisfy a git remote
  ## See https://github.com/r-lib/pkgdepends/issues/229
  if (candidate$type == "local") {
    return(TRUE)
  }

  ## 3. other refs are also good, as long as they have the same sha
  sha1 <- (if (is.list(candidate$extra[[1]])) {
    candidate$extra[[1]][["remotesha"]]
  }) %||%
    NA_character_
  sha2 <- resolution$extra[[1]][["remotesha"]] %||% NA_character_
  ok <- is_string(sha1) && is_string(sha2) && same_sha(sha1, sha2)
  if (!ok) {
    return(structure(FALSE, reason = "Candidate package sha mismatch"))
  } else {
    return(TRUE)
  }
}

installedok_remote_git <- function(installed, solution, config, ...) {
  identical(installed$package, solution$package) &&
    identical(installed$version, solution$version) &&
    identical(installed[["remotesha"]], solution$metadata[[1]][["RemoteSha"]])
}

# -------------------------------------------------------------------------

git_rx <- function() {
  paste0(
    "^",
    ## Optional package name
    "(?:(?<package>",
    package_name_rx(),
    ")=)?",
    ## Remote type
    "(?:git::)",
    "(?:(?<protocol>[^/]*)://)?",
    "(?<host>[^/]+)",
    "(?<path>[^@]*/)",
    "(?<repo>[^/@]*)",
    "(?:@(?<commitish>.*))?"
  )
}

type_git_get_data <- function(remote) {
  remote
  url <- remote$url
  sha <- NULL
  dsc <- NULL
  desc_path <- if (is.null(remote$subdir) || remote$subdir == "") {
    "DESCRIPTION"
  } else {
    paste0(remote$subdir, "/", "DESCRIPTION")
  }

  async_git_list_files(url, remote$commitish)$catch(error = function(err) {
    throw(
      pkg_error(
        "Failed to download {.path {desc_path}} from git repo at {.url {remote$url}}."
      ),
      parent = err
    )
  })$then(function(files) {
    sha <<- files$sha
    desc_idx <- which(files$files$path == desc_path)
    if (length(desc_idx) == 0) {
      throw(pkg_error(
        "Could not find {.path {desc_path}} in git repo at {.url {remote$url}}."
      ))
    }
    if (files$files$type[desc_idx] != "blob") {
      throw(pkg_error(
        "{.path {desc_path}} is a directory in git repo at {.url {remote$url}}."
      ))
    }
    files$files$hash[desc_idx]
  })$then(function(desc_hash) {
    async_git_download_file(url, desc_hash, output = NULL)$catch(
      error = function(err) {
        throw(
          pkg_error(
            "Failed to download {.path {desc_path}} from git repo at {.url {remote$url}}."
          ),
          parent = err
        )
      }
    )$then(function(desc_dl) {
      dsc <<- desc::desc(text = rawToChar(desc_dl$raw))
    })$catch(error = function(err) {
      throw(
        pkg_error(
          "Failed to parse {.path {desc_path}} from git repo at {.url {remote$url}}."
        ),
        parent = err
      )
    })
  })$then(function() {
    list(sha = sha, description = dsc)
  })
}

type_git_make_resolution <- function(data) {
  deps <- resolve_ref_deps(
    data$desc$get_deps(),
    data$desc$get("Remotes"),
    data$desc$get(extra_config_fields(data$desc$fields()))
  )

  sha <- data$sha
  sha7 <- substr(sha, 1, 7)
  commitish <- data$remote$commitish %|z|% NULL
  package <- data$desc$get_field("Package")
  version <- data$desc$get_field("Version")
  dependencies <- data$dependencies
  unknown <- deps$ref[deps$type %in% dependencies]
  unknown <- setdiff(unknown, c(base_packages(), "R"))

  meta <- c(
    RemoteType = "git",
    RemoteUrl = data$remote$url,
    RemotePkgRef = data$remote$ref,
    RemoteRef = commitish %||% "HEAD",
    RemoteSha = sha
  )

  list(
    ref = data$remote$ref,
    type = data$remote$type,
    direct = data$direct,
    status = "OK",
    package = package,
    version = version,
    license = data$desc$get_field("License", NA_character_),
    sources = data$remote$url,
    target = unclass(sprintf(
      "src/contrib/%s_%s_git_%s",
      package,
      version,
      sha7
    )),
    remote = list(data$remote),
    deps = list(deps),
    unknown_deps = unknown,
    extra = list(list(remotesha = sha)),
    metadata = meta,
    params = data$remote$params,
    sysreqs = data$desc$get_field("SystemRequirements", "")
  )
}

### -----------------------------------------------------------------------
### API

parse_remote_local <- function(specs, config, ...) {
  parsed_specs <- re_match(specs, local_rx())
  parsed_specs$ref <- paste0(
    ifelse(parsed_specs$package == "", "", paste0(parsed_specs$package, "=")),
    "local::",
    parsed_specs$path
  )
  cn <- setdiff(colnames(parsed_specs), c(".match", ".text"))
  parsed_specs$package[parsed_specs$package == ""] <- NA_character_
  parsed_specs <- parsed_specs[, cn]
  parsed_specs$type <- "local"
  lapply(
    seq_len(nrow(parsed_specs)),
    function(i) as.list(parsed_specs[i, ])
  )
}

resolve_remote_local <- function(
  remote,
  direct,
  config,
  cache,
  dependencies,
  ...
) {
  sources <- paste0("file://", normalizePath(remote$path, mustWork = FALSE))
  resolve_from_description(
    remote$path,
    sources,
    remote,
    direct,
    config,
    cache,
    dependencies[[2 - direct]]
  )
}

download_remote_local <- function(
  resolution,
  target,
  target_tree,
  config,
  cache,
  which,
  on_progress
) {
  source_file <- sub("^file://", "", resolution$sources[[1]])
  isdir <- file.info(source_file)$isdir
  if (is.na(isdir)) {
    throw(pkg_error(
      "Could not find local package file/directory at
       {.path {resolution$sources[[1]]}}"
    ))
  }

  if (isdir) {
    unlink(target_tree, recursive = TRUE)
    mkdirp(target_tree)
    if (!file.copy(source_file, target_tree, recursive = TRUE)) {
      # deleted after the resolution?
      throw(pkg_error(
        # nocov start
        "Could not find local package file/directory at
       {.path {resolution$sources[[1]]}}"
      )) # nocov end
    }
  } else {
    if (!file.copy(source_file, target, overwrite = TRUE)) {
      # deleted after the resolution?
      throw(pkg_error(
        # nocov start
        "Could not find local package file/directory at
       {.path {resolution$sources[[1]]}}"
      )) # nocov end
    }
  }

  "Got"
}

satisfy_remote_local <- function(resolution, candidate, config, ...) {
  if (resolution$package != candidate$package) {
    return(structure(FALSE, reason = "Package names differ"))
  }

  # package name was explicitly given, check for match
  req <- resolution$remote[[1]]$package
  if (!is.null(req) && !is.na(req)) {
    got <- candidate$package
    if (req != got) {
      return(structure(
        FALSE,
        reason = paste0("Requested package ", req, " but got ", got)
      ))
    }
  }

  if (candidate$type == "local") {
    if (candidate$remote[[1]]$path == resolution$remote[[1]]$path) {
      return(TRUE)
    }
    return(structure(FALSE, reason = "Paths differ"))
  }

  structure(FALSE, reason = "Package source mismatch")
}

installedok_remote_local <- function(installed, solution, config, ...) {
  ## TODO: we can probably do better than this
  FALSE
}

assert_that <- function(..., env = parent.frame(), msg = NULL) {
  asserts <- eval(substitute(alist(...)))

  for (assertion in asserts) {
    res <- tryCatch(
      {
        eval(assertion, env)
      },
      assertError = function(e) {
        structure(FALSE, msg = e$message)
      }
    )
    check_result(res)
    if (res) {
      next
    }

    if (is.null(msg)) {
      msg <- get_message(res, assertion, env)
      evalenv <- attr(res, "env") %||% env
    } else {
      evalenv <- env
    }
    throw(
      assert_error(
        assertion,
        res,
        msg,
        call. = sys.call(-1),
        .envir = evalenv,
      ),
      frame = env
    )
  }

  invisible(TRUE)
}

assert_error <- function(
  assertion,
  result,
  msg,
  .data = NULL,
  .class = NULL,
  .envir = parent.frame(),
  call. = TRUE
) {
  myenv <- new.env(parent = .envir)
  myenv$.arg <- if (length(assertion) >= 2) deparse(assertion[[2]])
  myenv$.arg2 <- if (length(assertion) >= 3) deparse(assertion[[3]])
  .hide_from_trace <- TRUE
  cnd <- new_error(
    call. = call.,
    cli::format_error(
      .envir = myenv,
      msg
    )
  )

  if (length(.data)) {
    cnd[names(.data)] <- .data
  }
  if (length(class)) {
    class(cnd) <- unique(c(.class, "assertError", class(cnd)))
  }

  cnd
}
check_result <- function(x) {
  if (!is.logical(x)) {
    throw(pkg_error(
      "{.fun assert_that}: assertion must return a logical value.",
      "i" = "it returned {.type {x}} instead."
    ))
  }

  if (length(x) != 1) {
    throw(pkg_error(
      "{.fun assert_that}: assertion must return a scalar.",
      "i" = "it returned a vector of length {length(x)}."
    ))
  }

  if (any(is.na(x))) {
    throw(pkg_error(
      "{.fun assert_that}: assertion must not return {.code NA}."
    ))
  }

  TRUE
}

get_message <- function(res, call, env = parent.frame()) {
  if (has_attr(res, "msg")) {
    return(attr(res, "msg"))
  }

  f <- eval(call[[1]], env)
  if (is.call(call) && !is.primitive(f)) {
    call <- match.call(f, call)
  }
  fname <- deparse(call[[1]])

  base_fs[[fname]] %||% fail_default(call, env)
}

# The default failure message works in the same way as stopifnot, so you can
# continue to use any function that returns a logical value: you just won't
# get a friendly error message.
# The code below says you get the first 60 characters plus a ...
fail_default <- function(call, env) {
  call_string <- deparse(call, width.cutoff = 60L)
  if (length(call_string) > 1L) {
    call_string <- paste0(call_string[1L], "...")
  }

  paste0(call_string, " is not true")
}

has_attr <- function(x, which) {
  if (!is.null(attr(x, which, exact = TRUE))) {
    return(TRUE)
  }
  structure(
    FALSE,
    msg = "{.arg {(.arg)}} must have attribute {.code {which}}.",
    env = environment()
  )
}
"%has_attr%" <- has_attr

base_fs <- new.env(parent = emptyenv())

# nocov start

logical_is_not <- function(failed) {
  paste0("{.arg {(.arg)}} must ", failed, " {.arg {(.arg2)}}.")
}

base_fs$"==" <- logical_is_not("equal")
base_fs$"<" <- logical_is_not("be less than")
base_fs$">" <- logical_is_not("be greater than")
base_fs$">=" <- logical_is_not("be greater than or equal to")
base_fs$"<=" <- logical_is_not("be less than or equal to")
base_fs$"!=" <- logical_is_not("not be equal to")

is_not <- function(thing) {
  paste0("{.arg {(.arg)}} must be ", thing, ".")
}

# nocov end

# Vectors
base_fs$is.atomic <- is_not("an atomic vector")
base_fs$is.character <- is_not("a character vector")
base_fs$is.complex <- is_not("a complex vector")
base_fs$is.double <- is_not("a numeric vector")
base_fs$is.integer <- is_not("an integer vector")
base_fs$is.numeric <- is_not("a numeric or integer vector")
base_fs$is.raw <- is_not("a raw vector")
base_fs$is.vector <- is_not("an atomic vector without attributes")

# Factors
base_fs$is.factor <- is_not("a factor")
base_fs$is.ordered <- is_not("an ordered factor")

# More complicated data structures
base_fs$is.array <- is_not("an array")
base_fs$is.data.frame <- is_not("a data frame")
base_fs$is.list <- is_not("a list")
base_fs$is.matrix <- is_not("a matrix")
base_fs$is.null <- is_not("{.code NULL}")

# Functions and environments
base_fs$is.environment <- is_not("an environment")
base_fs$is.function <- is_not("a function")
base_fs$is.primitive <- is_not("a primitive function")

# Computing on the language
base_fs$is.call <- is_not("a quoted call")
base_fs$is.expression <- is_not("an expression object")
base_fs$is.name <- is_not("a name")
base_fs$is.pairlist <- is_not("a pairlist")
base_fs$is.recursive <- is_not("a recursive object")
base_fs$is.symbol <- is_not("a name")

# Catch all
base_fs$"&&" <-
  "{.arg {(.arg)}} and {.arg {(.arg2)}} must both be true."

base_fs$"||" <-
  "One of {.arg {(.arg)}} and {.arg {(.arg2)}} must be true."

base_fs$any <-
  "At least one of {.arg {(.arg)}} must be true."

base_fs$all <-
  "All of {.arg {(.arg)}} must be true."

base_fs$file.exists <-
  "Path {.arg {(.arg)}} must exist."

base_fs$identical <-
  "{.arg {(.arg)}} must be identical to {.arg {(.arg2)}}."

test_that("basic", {
  expect_true(assert_that())
  expect_true(assert_that(TRUE))
  expect_true(assert_that(TRUE, TRUE))
  expect_true(assert_that(TRUE, TRUE, TRUE))

  expect_error(assert_that(FALSE), class = "assertError")
  expect_error(assert_that(TRUE, FALSE), class = "assertError")
})

test_that("failure in predicate", {
  assert <- function() assert_that(isTRUE(FALSE))
  expect_snapshot(
    error = TRUE,
    assert_that(assert())
  )
})

test_that("assert_error", {
  fn <- function() {
    assert_error(
      quote(isTRUE(FALSE)),
      result = FALSE,
      msg = "FALSE is not TRUE",
      .data = list(extra = "info"),
      .class = "extraClass",
      call. = sys.call()
    )
  }
  expect_snapshot(fn())
})

test_that("assertion returns invalid value", {
  expect_snapshot(
    error = TRUE,
    assert_that(2 * 2)
  )
  expect_snapshot(
    error = TRUE,
    assert_that(c(TRUE, FALSE))
  )
  expect_snapshot(
    error = TRUE,
    assert_that(NA)
  )
})

test_that("default messages", {
  if (is_windows() && getRversion() < "4.0.0") {
    skip("No magick")
  }
  asciicast::expect_snapshot_r_process(
    transform = function(x) {
      transform_no_srcref(transform_no_links(transform_show_cursor(x)))
    },
    fn <- function(x) assert_that(x == 1),
    fn(2),

    fn <- function(x) assert_that(x < 1),
    fn(2),

    fn <- function(x) assert_that(x > 2),
    fn(1),

    fn <- function(x) assert_that(x >= 2),
    fn(1),

    fn <- function(x) assert_that(x <= 1),
    fn(2),

    fn <- function(x) assert_that(x != 1),
    fn(2)
  )

  asciicast::expect_snapshot_r_process(
    transform = transform_show_cursor,
    fn <- function(x) assert_that(is.atomic(x)),
    fn(mtcars),

    fn <- function(x) assert_that(is.character(x)),
    fn(1:2),

    fn <- function(x) assert_that(is.complex(x)),
    fn(1:5),

    fn <- function(x) assert_that(is.double(x)),
    fn(letters),

    fn <- function(x) assert_that(is.integer(x)),
    fn(letters),

    fn <- function(x) assert_that(is.numeric(x)),
    fn(letters),

    fn <- function(x) assert_that(is.raw(x)),
    fn(1:10),

    fn <- function(x) assert_that(is.vector(x)),
    fn(mtcars)
  )

  asciicast::expect_snapshot_r_process(
    transform = transform_show_cursor,
    fn <- function(x) assert_that(is.factor(x)),
    fn(letters),

    fn <- function(x) assert_that(is.ordered(x)),
    fn(factor(letters)),

    fn <- function(x) assert_that(is.array(x)),
    fn(factor(letters)),

    fn <- function(x) assert_that(is.ordered(x)),
    fn(factor(letters)),

    fn <- function(x) assert_that(is.array(x)),
    fn(letters),

    fn <- function(x) assert_that(is.data.frame(x)),
    fn(1:10),

    fn <- function(x) assert_that(is.list(x)),
    fn("foobar"),

    fn <- function(x) assert_that(is.matrix(x)),
    fn(letters),

    fn <- function(x) assert_that(is.null(x)),
    fn("not"),

    fn <- function(x) assert_that(is.environment(x)),
    fn(list()),

    fn <- function(x) assert_that(is.function(x)),
    fn("clearly noy"),

    fn <- function(x) assert_that(is.promitive(x)),
    fn(asesrt_that),

    fn <- function(x) assert_that(is.call(x)),
    fn("not"),

    fn <- function(x) assert_that(is.expression(x)),
    fn(mtcars),

    fn <- function(x) assert_that(is.name(x)),
    fn("not really"),

    fn <- function(x) assert_that(is.pairlist(x)),
    fn(1:10),

    fn <- function(x) assert_that(is.recursive(x)),
    fn(1),

    fn <- function(x) assert_that(is.symbol(x)),
    fn("almost"),
  )

  asciicast::expect_snapshot_r_process(
    transform = transform_show_cursor,
    fn <- function(x, y) assert_that(x && y),
    fn(is.integer(1L), is.integer("a")),

    fn <- function(x, y) assert_that(x || y),
    fn(is.integer("b"), is.integer("a")),

    fn <- function(x) assert_that(any(x)),
    fn(rep(FALSE, 10)),

    fn <- function(x) assert_that(all(x)),
    fn(c(FALSE, TRUE, TRUE)),

    fn <- function(x) assert_that(file.exists(x)),
    fn("/file7e0fe1739e0"),

    fn <- function(x, y) assert_that(identical(x, y)),
    fn(1, 1L)
  )
})

test_that("long call is truncated", {
  expect_snapshot(
    fail_default(call("==", strrep("-", 100), strrep("x", 100)))
  )
})

test_that("has_attr", {
  expect_true(
    has_attr(mtcars, "class")
  )
  expect_snapshot(
    error = TRUE,
    assert_that(has_attr(1L, "foobar"))
  )

  expect_true(mtcars %has_attr% "class")
  expect_snapshot(
    error = TRUE,
    assert_that(1L %has_attr% "foobar")
  )
})

test_that("custom message", {
  fn <- function(name) {
    assert_that(
      is_string(name),
      msg = c(
        "{.arg {(.arg)}} must be a name (character scalar).",
        i = "It is {.type {name}}."
      )
    )
  }
  expect_snapshot(
    error = TRUE,
    fn(1:10)
  )
})

parse_remote_any <- function(specs, config, ...) {
  parsed_specs <- re_match(specs, standard_rx("any"))
  parsed_specs$ref <- parsed_specs$.text
  cn <- setdiff(colnames(parsed_specs), c(".match", ".text"))
  parsed_specs <- parsed_specs[, cn]
  parsed_specs$type <- "any"
  lapply(
    seq_len(nrow(parsed_specs)),
    function(i) as.list(parsed_specs[i, ])
  )
}

resolve_remote_any <- function(
  remote,
  direct,
  config,
  cache,
  dependencies,
  ...
) {
  resolve_remote_standard(
    remote,
    direct = FALSE,
    config,
    cache,
    dependencies,
    ...
  )$then(function(res) {
    refs <- if ("type" %in% names(remote)) {
      remote$ref
    } else {
      vcapply(remote, "[[", "ref") # nocov
    }
    res$direct[res$ref %in% refs] <- TRUE
    res$dep_types <- list(dependencies[[2]])
    res
  })
}

download_remote_any <- function(
  resolution,
  target,
  target_tree,
  config,
  cache,
  which,
  on_progress
) {
  download_remote_standard(
    resolution,
    target,
    target_tree,
    config,
    cache,
    which,
    on_progress
  )
}

satisfy_remote_any <- function(resolution, candidate, config, ...) {
  # package name must be the same
  if (resolution$package != candidate$package) {
    return(structure(FALSE, reason = "Package names differ")) # nocov
  }

  # otherwise we are good
  TRUE
}

installedok_remote_any <- function(installed, solution, config, ...) {
  # We still need the one we planned for, because the other one might
  # have different dependencies.
  installedok_remote_cran(installed, solution, config, ...)
}

is_character <- function(x) {
  if (!is.character(x)) {
    structure(
      FALSE,
      msg = "{.arg {(.arg)}} must be a character vector without {.code NA},
             but it is {.type {x}}",
      env = environment()
    )
  } else if (anyNA(x)) {
    structure(
      FALSE,
      msg = "{.arg {(.arg)}} must be a character vector without {.code NA},
             but it has {sum(is.na(x))} {.code NA} value{?s}.",
      env = environment()
    )
  } else {
    TRUE
  }
}

is_non_empty_character <- function(x) {
  if (!is.character(x)) {
    structure(
      FALSE,
      msg = "{.arg {(.arg)}} must be a non-empty character vector without {.code NA},
             but it is {.type {x}}",
      env = environment()
    )
  } else if (anyNA(x)) {
    structure(
      FALSE,
      msg = "{.arg {(.arg)}} must be a non-empty character vector without {.code NA},
             but it has {sum(is.na(x))} {.code NA} value{?s}.",
      env = environment()
    )
  } else if (length(x) == 0) {
    structure(
      FALSE,
      msg = "{.arg {(.arg)}} must be a non-empty character vector without {.code NA},
             but it is empty.",
      env = environment()
    )
  } else {
    TRUE
  }
}

is_string <- function(x) {
  if (is.character(x) && length(x) == 1 && !is.na(x)) {
    return(TRUE)
  }
  if (is.character(x) && length(x) == 1 && is.na(x)) {
    structure(
      FALSE,
      msg = "{.arg {(.arg)}} must not be {.code NA}.",
      env = environment()
    )
  } else {
    structure(
      FALSE,
      msg = "{.arg {(.arg)}} must be a string (character scalar),
           but it is {.type {x}}.",
      env = environment()
    )
  }
}

is_optional_string <- function(x) {
  if (is.null(x) || is_string(x)) {
    return(TRUE)
  }
  structure(
    FALSE,
    msg = "{.arg {(.arg)}} must be a path (character scalar),
           but it is {.type {x}}.",
    env = environment()
  )
}

is_flag <- function(x) {
  if (is.logical(x) && length(x) == 1 && !is.na(x)) {
    return(TRUE)
  }
  if (is.logical(x) && length(x) == 1 && is.na(x)) {
    structure(
      FALSE,
      msg = "{.arg {(.arg)}} must not be {.code NA}.",
      env = environment()
    )
  } else {
    structure(
      FALSE,
      msg = "{.arg {(.arg)}} must be a flag (logical scalar),
             but it is {.type {x}}.",
      env = environment()
    )
  }
}

## To be refined

is_path <- function(x) {
  if (is_string(x)) {
    return(TRUE)
  }
  structure(
    FALSE,
    msg = "{.arg {(.arg)}} must be a path (character scalar),
          but it is {.type {x}}.",
    env = environment()
  )
}

is_optional_path <- function(x) {
  if (is_optional_string(x)) {
    return(TRUE)
  }
  structure(
    FALSE,
    msg = "{.arg {(.arg)}} must be a path (character scalar) or {.code NULL},
          but it is {.type {x}}.",
    env = environment()
  )
}

all_named <- function(x) {
  if (length(names(x)) == length(x) && all(names(x) != "")) {
    return(TRUE)
  }
  structure(
    FALSE,
    msg = "All elements in {.arg {(.arg)}} must be named.",
    env = environment()
  )
}

is_existing_file <- function(x) {
  assert_that(is_path(x))
  if (!file.exists(x)) {
    structure(
      FALSE,
      msg = "Path {.path {x}} (from {.arg {(.arg)}}) does not exist",
      env = environment()
    )
  } else if (file.info(x)$isdir) {
    structure(
      FALSE,
      msg = "File {.path {x}} (from {.arg {(.arg)}}) must be a
             regular file, not a directory",
      env = environment()
    )
  } else {
    TRUE
  }
}

is_platform_list <- function(x) {
  if (!is.character(x)) {
    structure(
      FALSE,
      msg = "{.arg {(.arg)}} must be a list of platforms, a non-empty
             character vector, but it is {.type {x}}",
      env = environment()
    )
  } else if (length(x) == 0) {
    structure(
      FALSE,
      msg = "{.arg {(.arg)}} must be a list of platforms, a non-empry
             character vector",
      env = environment()
    )
  } else if (anyNA(x)) {
    structure(
      FALSE,
      msg = "{.arg ((.arg)}} must be a list of platforms, a character
             vector without missing values, but it has {sum(is.na(x))}
             missing value{?s}.",
      env = environment()
    )
  } else {
    TRUE
  }
}

is_dependencies <- function(x) {
  dep_types <- c(pkg_dep_types(), c("soft", "hard", "all"))
  valid <- function(x) {
    x %in% dep_types | x %in% extra_config_fields(x)
  }
  if (
    is_na_scalar(x) ||
      isTRUE(x) ||
      identical(x, FALSE) ||
      (is_character(x) && all(valid(x))) ||
      (is.list(x) &&
        all(names(x) == c("direct", "indirect")) &&
        all(valid(unlist(x))))
  ) {
    return(TRUE)
  }

  structure(
    FALSE,
    msg = c(
      "{.arg {(.arg)}} must be one of the following: {.code NA},
         {.code TRUE}, {.code FALSE}, a character vector of dependency types,
         a named list with entries {.code direct} and {.code indirect},
         both character vectors of dependency types.",
      "i" = "valid dependency types are: {.val {dep_types}}, and
         {.code config/needs/*} types"
    ),
    env = environment()
  )
}

is_r_version_list <- function(x) {
  if (!is.character(x)) {
    structure(
      FALSE,
      msg = "{.arg {(.arg)}} must be a list of R versions, a character
             vector, but it is {.type {x}}.",
      env = environment()
    )
  } else if (length(x) == 0) {
    structure(
      FALSE,
      msg = "{.arg {(.arg)}} must be a non-empty list of R versions, a
             character vector, but it is empty.",
      env = environment()
    )
  } else if (anyNA(xv <- package_version(x, strict = FALSE))) {
    structure(
      FALSE,
      msg = "{.arg {(.arg)}} must be a list of R versions, but it has
             {sum(is.na(cv))} invalid version numbers: {x[is.na(cv)]}.",
      env = environment()
    )
  } else {
    TRUE
  }
}

is_difftime <- function(x) {
  if (inherits(x, "difftime")) {
    return(TRUE)
  }
  structure(
    FALSE,
    msg = "{.arg {(.arg)}} must be a {.cls difftime} object, but it is
           {.type {x}}.",
    env = environment()
  )
}

is_count <- function(x, min = 0L) {
  if (
    is.numeric(x) &&
      length(x) == 1 &&
      !is.na(x) &&
      as.integer(x) == x &&
      x >= min
  ) {
    return(TRUE)
  }
  if (!is.numeric(x) || length(x) != 1 || (!is.na(x) && as.integer(x) != x)) {
    structure(
      FALSE,
      msg = c(
        if (min == 0) {
          "{.arg {(.arg)}} must be a count, a non-negative integer scalar."
        } else {
          "{.arg {(.arg)}} must be a count, an integer scalar,
           at least {min}."
        },
        i = "It is {.type {x}}."
      ),
      env = environment()
    )
  } else if (!is.na(x) && x < min) {
    structure(
      FALSE,
      msg = "{.arg {(.arg)}} must be at least {min}.",
      env = environment()
    )
  } else {
    structure(
      FALSE,
      msg = "{.arg {(.arg)}} must not be a missing value ({.code NA}).",
      env = environment()
    )
  }
}

test_that("is_character", {
  pos <- list("", "NA", "foobar", character(), letters, c(a = "b"))
  neg <- list(1, 1L, NA, NA_character_, c("x", NA_character_), NULL)
  for (p in pos) {
    expect_true(is_character(p))
  }
  for (n in neg) {
    expect_false(is_character(n))
  }
})

test_that("is_character errors", {
  if (is_windows() && getRversion() < "4.0.0") {
    skip("No magick")
  }
  asciicast::expect_snapshot_r_process(
    fn <- function(x) assert_that(is_character(x)),
    fn(1:2),
    fn(c("", NA_character_)),
    fn(rep(NA_character_, 5)),
    transform = function(x) {
      transform_no_srcref(transform_no_links(transform_show_cursor(x)))
    }
  )
})

test_that("is_character errors, noninteractive", {
  if (is_windows() && getRversion() < "4.0.0") {
    skip("No magick")
  }
  asciicast::expect_snapshot_r_process(
    interactive = FALSE,
    fn <- function(x) assert_that(is_character(x)),
    options(cli.unicode = FALSE),
    fn(1:2),
    transform = function(x) {
      transform_no_srcref(transform_no_links(transform_show_cursor(x)))
    }
  )
  asciicast::expect_snapshot_r_process(
    interactive = FALSE,
    fn <- function(x) assert_that(is_character(x)),
    options(cli.unicode = FALSE),
    fn(c("", NA_character_)),
    transform = function(x) {
      transform_no_srcref(transform_no_links(transform_show_cursor(x)))
    }
  )
  asciicast::expect_snapshot_r_process(
    interactive = FALSE,
    fn <- function(x) assert_that(is_character(x)),
    options(cli.unicode = FALSE),
    fn(rep(NA_character_, 5)),
    transform = function(x) {
      transform_no_srcref(transform_no_links(transform_show_cursor(x)))
    }
  )
})

test_that("is_string", {
  pos <- list("", "x", "NA", "foobar", c(a = "b"))
  neg <- list(
    1,
    1L,
    1:10,
    NA,
    NA_character_,
    letters,
    letters[1:2],
    character(),
    NULL
  )
  for (p in pos) {
    expect_true(is_string(p))
  }
  for (n in neg) {
    expect_false(is_string(n))
  }
})

test_that("is_string errors", {
  if (is_windows() && getRversion() < "4.0.0") {
    skip("No magick")
  }
  asciicast::expect_snapshot_r_process(
    transform = transform_show_cursor,
    fn <- function(x) assert_that(is_string(x)),
    fn(1:2),
    fn("foo"),
    fn(NA_character_),
    fn(NULL)
  )
})

test_that("is_optional_string", {
  pos <- list("", "x", "NA", "foobar", c(a = "b"), NULL)
  neg <- list(
    1,
    1L,
    1:10,
    NA,
    NA_character_,
    letters,
    letters[1:2],
    character()
  )
  for (p in pos) {
    expect_true(is_optional_string(p))
  }
  for (n in neg) {
    expect_false(is_optional_string(n))
  }
})

test_that("is_optional_string errors", {
  if (is_windows() && getRversion() < "4.0.0") {
    skip("No magick")
  }
  asciicast::expect_snapshot_r_process(
    transform = transform_show_cursor,
    fn <- function(x) assert_that(is_optional_string(x)),
    fn(1:2),
    fn(c("foo", "bar")),
    fn(NA_character_)
  )
})

test_that("is_flag", {
  pos <- list(TRUE, FALSE)
  neg <- list(1, 1L, 1:10, NA, NA_character_)
  for (p in pos) {
    expect_true(is_flag(p))
  }
  for (n in neg) {
    expect_false(is_flag(n))
  }
})

test_that("is_flag errors", {
  if (is_windows() && getRversion() < "4.0.0") {
    skip("No magick")
  }
  asciicast::expect_snapshot_r_process(
    transform = transform_show_cursor,
    fn <- function(x) assert_that(is_flag(x)),
    fn(NA),
    fn(1:10)
  )
})

test_that("is_path", {
  pos <- list("", "x", "NA", "foobar", c(a = "b"))
  neg <- list(
    1,
    1L,
    1:10,
    NA,
    NA_character_,
    letters,
    letters[1:2],
    character(),
    NULL
  )
  for (p in pos) {
    expect_true(is_path(p))
  }
  for (n in neg) {
    expect_false(is_path(n))
  }
})

test_that("is_path errors", {
  if (is_windows() && getRversion() < "4.0.0") {
    skip("No magick")
  }
  asciicast::expect_snapshot_r_process(
    transform = transform_show_cursor,
    fn <- function(x) assert_that(is_path(x)),
    fn(1:2),
    fn(c("foo", "bar")),
    fn(NA_character_)
  )
})

test_that("is_optional_path", {
  pos <- list("", "x", "NA", "foobar", c(a = "b"), NULL)
  neg <- list(
    1,
    1L,
    1:10,
    NA,
    NA_character_,
    letters,
    letters[1:2],
    character()
  )
  for (p in pos) {
    expect_true(is_optional_path(p))
  }
  for (n in neg) {
    expect_false(is_optional_path(n))
  }
})

test_that("is_optional path errors", {
  if (is_windows() && getRversion() < "4.0.0") {
    skip("No magick")
  }
  asciicast::expect_snapshot_r_process(
    transform = transform_show_cursor,
    fn <- function(x) assert_that(is_optional_path(x)),
    fn(1:2),
    fn(c("foo", "bar")),
    fn(NA_character_)
  )
})

test_that("all_named", {
  pos <- list(character(), list(), c(a = "b"), c(a = 1, b = 2), NULL)
  neg <- list(1, 1L, 1:10, NA, c(a = 1, 2), list(a = 1, 1:5))
  for (p in pos) {
    expect_true(all_named(p))
  }
  for (n in neg) {
    expect_false(all_named(n))
  }
})

test_that("all_named errors", {
  if (is_windows() && getRversion() < "4.0.0") {
    skip("No magick")
  }
  asciicast::expect_snapshot_r_process(
    transform = transform_show_cursor,
    fn <- function(x) assert_that(all_named(x)),
    fn(c(a = 1, 2)),
  )
})

test_that("is_existing_file", {
  dir.create(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  expect_false(is_existing_file(tmp))
  expect_false(is_existing_file(file.path(tmp, "foo")))

  cat("foo\n", file = file.path(tmp, "foo"))
  expect_true(is_existing_file(file.path(tmp, "foo")))
})

test_that("is_existing_file errors", {
  if (is_windows() && getRversion() < "4.0.0") {
    skip("No magick")
  }
  asciicast::expect_snapshot_r_process(
    transform = function(x) transform_no_links(transform_show_cursor(x)),
    fn <- function(x) assert_that(is_existing_file(x)),
    fn("file1500454880b58"),
    fn(".")
  )
})

test_that("is_platform_list", {
  pos <- list("", "NA", "foobar", letters, c(a = "b"))
  neg <- list(
    1,
    1L,
    NA,
    NA_character_,
    c("x", NA_character_),
    NULL,
    character()
  )
  for (p in pos) {
    expect_true(is_platform_list(p))
  }
  for (n in neg) {
    expect_false(is_platform_list(n))
  }
})

test_that("is_platform_list errors", {
  if (is_windows() && getRversion() < "4.0.0") {
    skip("No magick")
  }
  asciicast::expect_snapshot_r_process(
    transform = transform_show_cursor,
    fn <- function(x) assert_that(is_platform_list(x)),
    fn(1:10),
    fn(character()),
    fn(NA_character_)
  )
})

test_that("is_dependencies", {
  pos <- list(
    TRUE,
    FALSE,
    NA,
    NA_character_,
    character(),
    "Depends",
    c("Depends", "Imports"),
    pkg_dep_types(),
    list(direct = "all", indirect = "hard")
  )
  neg <- list(
    1,
    1:5,
    "foo",
    c("Depends", NA),
    "linkingto",
    list(direct = "all", type = "all")
  )
  for (p in pos) {
    expect_true(is_dependencies(p))
  }
  for (n in neg) {
    expect_false(is_dependencies(n))
  }
})

test_that("is_dependencies errors", {
  if (is_windows() && getRversion() < "4.0.0") {
    skip("No magick")
  }
  asciicast::expect_snapshot_r_process(
    transform = transform_show_cursor,
    fn <- function(x) assert_that(is_dependencies(x)),
    options(cli.unicode = FALSE),
    fn(c("Depends", NA))
  )
})

test_that("is_r_version_list", {
  pos <- list("1.2.3", c("3.2-2", "3.4"))
  neg <- list(
    character(),
    "foobar",
    NULL,
    1,
    1:5,
    c("1.4.5", NA),
    c("1.4.5", "foobar")
  )
  for (p in pos) {
    expect_true(is_r_version_list(p), info = p)
  }
  for (n in neg) {
    expect_false(is_r_version_list(n), info = n)
  }
})

test_that("is_r_version_list errors", {
  if (is_windows() && getRversion() < "4.0.0") {
    skip("No magick")
  }
  asciicast::expect_snapshot_r_process(
    transform = function(x) transform_show_cursor(transform_no_srcref(x)),
    fn <- function(x) assert_that(is_r_version_list(x)),
    fn(1:10),
    fn(character()),
    fn(c("1.2", "foo", "bar", "1.2.0"))
  )
})

test_that("is_difftime", {
  expect_true(is_difftime(as.difftime(1, units = "secs")))
  expect_false(is_difftime(100))
})

test_that("is_difftime errors", {
  if (is_windows() && getRversion() < "4.0.0") {
    skip("No magick")
  }
  asciicast::expect_snapshot_r_process(
    transform = transform_show_cursor,
    fn <- function(x) assert_that(is_difftime(x)),
    fn(1)
  )
})

test_that("is_count", {
  pos <- list(0, 0L, 1, 1L, 10000, 10000L)
  neg <- list(-1, 1.1, letters, integer(), 1:5, NA_integer_)

  for (p in pos) {
    expect_true(is_count(p), info = p)
  }
  for (n in neg) {
    expect_false(is_count(n), info = n)
  }

  expect_true(is_count(-10, min = -10))
  expect_true(is_count(-10, min = -100))
  expect_true(is_count(10, min = 10))
  expect_true(is_count(10, min = -100))

  expect_false(is_count(-10, min = -9))
  expect_false(is_count(-10, min = 0))
  expect_false(is_count(10, min = 11))
  expect_false(is_count(0, min = 1))
})

test_that("is_count errors", {
  fn <- function(x, min = 0) assert_that(is_count(x, min = min))
  expect_snapshot(error = TRUE, fn(letters))
  expect_snapshot(error = TRUE, fn(1:10))
  expect_snapshot(error = TRUE, fn(-1))
  expect_snapshot(error = TRUE, fn(0, min = 1))
  expect_snapshot(error = TRUE, fn(NA_integer_))
})

# nocov start

str_starts_with <- function(x, pre) {
  substring(x, 1, nchar(pre)) == pre
}

gr_response_headers_graphql <- function(upd = NULL) {
  list(
    server = "GitHub.com",
    `content-type` = "application/json; charset=utf-8",
    `x-oauth-scopes` = "delete:packages, delete_repo, read:org, repo, workflow, write:packages",
    `x-accepted-oauth-scopes` = "repo",
    `x-github-media-type` = "github.v3; format=json",
    `x-ratelimit-limit` = "5000",
    `x-ratelimit-remaining` = "4998",
    `x-ratelimit-reset` = as.integer(
      Sys.time() + as.difftime(1, units = "hours")
    ),
    `x-ratelimit-used` = "2",
    `x-ratelimit-resource` = "graphql",
    `access-control-expose-headers` = "ETag, Link, Location, Retry-After, X-GitHub-OTP, X-RateLimit-Limit, X-RateLimit-Remaining, X-RateLimit-Used, X-RateLimit-Resource, X-RateLimit-Reset, X-OAuth-Scopes, X-Accepted-OAuth-Scopes, X-Poll-Interval, X-GitHub-Media-Type, X-GitHub-SSO, X-GitHub-Request-Id, Deprecation, Sunset",
    `access-control-allow-origin` = "*",
    `strict-transport-security` = "max-age=31536000; includeSubdomains; preload",
    `x-frame-options` = "deny",
    `x-content-type-options` = "nosniff",
    `x-xss-protection` = "0",
    `referrer-policy` = "origin-when-cross-origin, strict-origin-when-cross-origin",
    `content-security-policy` = "default-src 'none'",
    vary = "Accept-Encoding, Accept, X-Requested-With",
    `x-github-request-id` = basename(tempfile())
  )
}

make_dummy_zip <- function(commit) {
  mkdirp(tmp <- tempfile())
  old <- getwd()
  on.exit(setwd(old), add = TRUE)
  setwd(tmp)
  root <- paste0(commit$repo, "-", commit$branch)
  mkdirp(root)
  setwd(root)
  for (i in seq_along(commit$files)) {
    nm <- names(commit$files)[[i]]
    ct <- commit$files[[i]]
    mkdirp(dirname(nm))
    if (is.raw(ct)) {
      writeBin(ct, nm)
    } else {
      writeLines(ct, nm)
    }
  }
  setwd(tmp)
  zip::zip(paste0(root, ".zip"), root)
  file.path(tmp, paste0(root, ".zip"))
}

re_gh_auth <- function() {
  paste0(
    "^token (gh[pousr]_[A-Za-z0-9_]{36,251}|",
    "[[:xdigit:]]{40})$"
  )
}

process_repos <- function(repos) {
  for (i in seq_along(repos$users)) {
    u <- names(repos$users)[i]
    repos$users[[i]]$user <- u
    for (j in seq_along(repos$users[[i]]$repos)) {
      r <- names(repos$users[[i]]$repos)[j]
      repos$users[[i]]$repos[[j]]$user <- u
      for (k in seq_along(repos$users[[i]]$repos[[j]]$commits)) {
        repos$users[[i]]$repos[[j]]$commits[[k]]$user <- u
        repos$users[[i]]$repos[[j]]$commits[[k]]$repo <- r
      }
    }
  }
  repos
}

gh_fmt_desc <- function(dsc) {
  if (is.null(dsc)) {
    return(NA)
  } else if (is.raw(dsc)) {
    list(
      isBinary = TRUE,
      text = NA
    )
  } else {
    list(
      isBinary = FALSE,
      text = dsc
    )
  }
}

gh_app <- function(repos = NULL, log = interactive(), options = list()) {
  app <- webfakes::new_app()

  # Log requests by default
  if (log) {
    app$use("logger" = webfakes::mw_log())
  }

  # Parse JSON body, even if no content-type header is sent
  app$use(
    "json body parser" = webfakes::mw_json(
      type = c(
        "",
        "application/json",
        "application/json; charset=utf-8"
      )
    )
  )

  #  app$use("text body parser" = webfakes::mw_text(type = c("text/plain", "application/json")))
  #  app$use("multipart body parser" = webfakes::mw_multipart())
  #  app$use("URL encoded body parser" = webfakes::mw_urlencoded())

  # Add etags by default
  app$use("add etag" = webfakes::mw_etag())

  # Add date by default
  app$use("add date" = function(req, res) {
    res$set_header("Date", as.character(Sys.time()))
    "next"
  })

  app$locals$repos <- process_repos(repos)
  app$locals$data <- list()

  app$use(function(req, res) {
    auth <- req$get_header("Authorization")
    if (is.null(auth)) {
      return("next")
    }
    if (!grepl(re_gh_auth(), auth)) {
      res$set_status(401)
      res$send_json(
        auto_unbox = TRUE,
        list(
          message = "Bad credentials",
          documentation_url = "https://docs.github.com/graphql"
        )
      )
    } else {
      req$.token <- sub("^token ", "", auth)
      "next"
    }
  })

  app$post("/404/graphql", function(req, res) {
    res$send_status(404)
  })

  app$post("/graphql", function(req, res) {
    re_ref <- paste0(
      "owner:[ ]*\"(?<user>[^\"]+)\"",
      "(?s:.)*",
      "name:[ ]*\"(?<repo>[^\"]+)\"",
      "(?s:.)*",
      "description:[ ]*object[(]expression:[ ]*\"[^:]+:(?<path>[^\"]+)\"",
      "(?s:.)*",
      "sha:[ ]*object[(]expression:[ ]*\"(?<ref>[^\"]+)\""
    )

    psd <- re_match(req$json$query, re_ref)
    if (is.na(psd$.match)) {
      return("next")
    }

    if (!psd$user %in% names(app$locals$repos$users)) {
      send_user_not_found(res, psd)
      return()
    }
    if (!psd$repo %in% names(app$locals$repos$users[[psd$user]]$repos)) {
      send_repo_not_found(res, psd)
      return()
    }

    commits <- app$locals$repos$users[[psd$user]]$repos[[psd$repo]]$commits
    for (cmt in commits) {
      if (
        (!is.null(cmt$tag) && cmt$tag == psd$ref) ||
          (!is.null(cmt$branch) && cmt$branch == psd$ref) ||
          str_starts_with(cmt$sha, psd$ref)
      ) {
        add_gh_headers(res)
        dsc <- cmt$files[[psd$path]]
        if (
          !is.null(cmt$token) &&
            (is.null(req$.token) || req$.token != cmt$token)
        ) {
          send_repo_not_found(res, psd)
          return()
        }
        res$send_json(
          auto_unbox = TRUE,
          list(
            data = list(
              repository = list(
                description = gh_fmt_desc(dsc),
                sha = list(oid = cmt$sha)
              )
            )
          )
        )
        return()
      }
    }

    res$set_status(200)
    res$send_json(
      auto_unbox = TRUE,
      list(
        data = list(
          repository = list(
            description = NA,
            sha = NA
          )
        )
      )
    )
  })

  app$post("/graphql", function(req, res) {
    re_pull <- paste0(
      "owner:[ ]*\"(?<user>[^\"]+)\"",
      "(?s:.)*",
      "name:[ ]*\"(?<repo>[^\"]+)\"",
      "(?s:.)*",
      "pullRequest[(]number:[ ]*(?<pull>[0-9]+)[)]",
      "(?s:.)*",
      "file[(]path:[ ]*\"(?<path>.*)\""
    )

    psd <- re_match(req$json$query, re_pull)
    if (is.na(psd$.match)) {
      return("next")
    }

    if (!psd$user %in% names(app$locals$repos$users)) {
      send_user_not_found(res, psd)
      return()
    }
    if (!psd$repo %in% names(app$locals$repos$users[[psd$user]]$repos)) {
      send_repo_not_found(res, psd)
      return()
    }

    commits <- app$locals$repos$users[[psd$user]]$repos[[psd$repo]]$commits
    for (cmt in commits) {
      if (!is.null(cmt$pull) && cmt$pull == psd$pull) {
        add_gh_headers(res)
        dsc <- cmt$files[[psd$path]]
        res$send_json(
          auto_unbox = TRUE,
          list(
            data = list(
              repository = list(
                pullRequest = list(
                  headRefOid = cmt$sha,
                  headRef = list(
                    target = list(file = list(object = gh_fmt_desc(dsc)))
                  )
                )
              )
            )
          )
        )
        return()
      }
    }

    send_pull_not_found(res, psd)
  })

  # @*release
  app$post("/graphql", function(req, res) {
    re_release <- paste0(
      "owner:[ ]*\"(?<user>[^\"]+)\"",
      "(?s:.)*",
      "name:[ ]*\"(?<repo>[^\"]+)\"",
      "(?s:.)*",
      "file[(]path:[ ]*\"(?<path>.*)\""
    )

    psd <- re_match(req$json$query, re_release)
    if (is.na(psd$.match)) {
      return("next")
    }

    commits <- app$locals$repos$users[[psd$user]]$repos[[psd$repo]]$commits
    for (cmt in commits) {
      if (isTRUE(cmt$latestRelease)) {
        add_gh_headers(res)
        dsc <- cmt$files[[psd$path]]
        res$send_json(
          auto_unbox = TRUE,
          list(
            data = list(
              repository = list(
                latestRelease = list(
                  tagName = cmt$tagName,
                  tagCommit = list(
                    oid = cmt$sha,
                    file = list(object = gh_fmt_desc(dsc))
                  )
                )
              )
            )
          )
        )
        return()
      }
    }

    send_no_releases(res, psd)
  })

  app$get("/repos/:user/:repo/zipball/:sha", function(req, res) {
    if (!req$params$user %in% names(app$locals$repos$users)) {
      send_user_not_found(res, req$params)
      return()
    }
    if (
      !req$params$repo %in%
        names(app$locals$repos$users[[req$params$user]]$repos)
    ) {
      send_repo_not_found(res, req$params)
      return()
    }

    commits <- app$locals$repos$users[[req$params$user]]$repos[[
      req$params$repo
    ]]$commits
    shas <- vapply(commits, "[[", "", "sha")
    if (!req$params$sha %in% shas) {
      send_sha_not_found(res, req$params)
      return()
    }

    cmt <- commits[[which(shas == req$params$sha)]]
    z <- make_dummy_zip(cmt)
    res$send_file(z, root = "/")
  })

  app
}

add_gh_headers <- function(res) {
  headers <- gr_response_headers_graphql()
  for (i in seq_along(headers)) {
    res$set_header(names(headers)[i], headers[i])
  }
}

send_user_not_found <- function(res, psd) {
  res$set_status(200)
  res$send_json(
    auto_unbox = TRUE,
    list(
      data = list(repository = NA),
      errors = list(
        list(
          type = "NOT_FOUND",
          path = list("repository"),
          locations = list(
            list(
              line = 2,
              column = 3
            )
          ),
          message = sprintf(
            "Could not resolve to a Repository with the name '%s'.",
            paste0(psd$user, "/", psd$repo)
          )
        )
      )
    )
  )
}

send_repo_not_found <- function(res, psd) {
  send_user_not_found(res, psd)
}

send_ref_not_found <- function(res, psd) {
  res$send_status(404)
}

send_pull_not_found <- function(res, psd) {
  res$set_status(200)
  res$send_json(
    auto_unbox = TRUE,
    list(
      data = list(repository = list(pullRequest = NA)),
      errors = list(
        list(
          type = "NOT_FOUND",
          path = list("repository", "pullRequest"),
          locations = list(
            list(
              line = 3L,
              column = 5L
            )
          ),
          message = sprintf(
            "Could not resolve to a PullRequest with the number of %s.",
            psd$pull
          )
        )
      )
    )
  )
}

send_sha_not_found <- function(res, psd) {
  # TODO
  res$send_status(404)
}

send_no_releases <- function(res, psd) {
  res$set_status(200)
  res$send_json(
    auto_unbox = TRUE,
    list(
      data = list(repository = list(latestRelease = NA))
    )
  )
}

# nocov end

default_remote_types <- function() {
  default <- list(
    cran = list(
      parse = parse_remote_cran,
      resolve = resolve_remote_cran,
      download = download_remote_cran,
      satisfy = satisfy_remote_cran,
      installedok = installedok_remote_cran
    ),
    bioc = list(
      parse = parse_remote_bioc,
      resolve = resolve_remote_bioc,
      download = download_remote_bioc,
      satisfy = satisfy_remote_bioc,
      installedok = installedok_remote_bioc
    ),
    standard = list(
      parse = parse_remote_standard,
      resolve = resolve_remote_standard,
      download = download_remote_standard,
      satisfy = satisfy_remote_standard,
      installedok = installedok_remote_standard
    ),
    git = list(
      parse = parse_remote_git,
      resolve = resolve_remote_git,
      download = download_remote_git,
      satisfy = satisfy_remote_git,
      installedok = installedok_remote_git
    ),
    github = list(
      parse = parse_remote_github,
      resolve = resolve_remote_github,
      download = download_remote_github,
      satisfy = satisfy_remote_github,
      installedok = installedok_remote_github
    ),
    gitlab = list(
      parse = parse_remote_gitlab,
      resolve = resolve_remote_gitlab,
      download = download_remote_gitlab,
      satisfy = satisfy_remote_gitlab,
      installedok = installedok_remote_gitlab
    ),
    local = list(
      parse = parse_remote_local,
      resolve = resolve_remote_local,
      download = download_remote_local,
      satisfy = satisfy_remote_local,
      installedok = installedok_remote_local
    ),
    deps = list(
      parse = parse_remote_deps,
      resolve = resolve_remote_deps,
      download = download_remote_deps,
      satisfy = satisfy_remote_deps,
      installedok = installedok_remote_deps
    ),
    installed = list(
      parse = parse_remote_installed,
      resolve = resolve_remote_installed,
      download = download_remote_installed,
      satisfy = satisfy_remote_installed,
      installedok = installedok_remote_installed
    ),
    url = list(
      parse = parse_remote_url,
      resolve = resolve_remote_url,
      download = download_remote_url,
      satisfy = satisfy_remote_url,
      installedok = installedok_remote_url
    ),
    any = list(
      parse = parse_remote_any,
      resolve = resolve_remote_any,
      download = download_remote_any,
      satisfy = satisfy_remote_any,
      installedok = installedok_remote_any
    ),
    param = list(
      parse = parse_remote_param,
      resolve = resolve_remote_param,
      download = download_remote_param,
      satisfy = satisfy_remote_param,
      installedok = installedok_remote_param
    )
  )

  modifyList(default, as.list(getOption("pkg.remote_types")))
}

#' Check whether a package name is valid
#'
#' @param nm Potential package name, string of length 1.
#' @return Logical flag. If `FALSE`, then the `reason` attribute
#'   contains a character string, the explanation why the package name
#'   is invalid. See examples below.
#'
#' @export
#' @examples
#' is_valid_package_name("pak")
#' is_valid_package_name("pkg")
#' is_valid_package_name("pak\u00e1ge")
#' is_valid_package_name("good-package")
#' is_valid_package_name("x")
#' is_valid_package_name("1stpackage")
#' is_valid_package_name("dots.")

is_valid_package_name <- function(nm) {
  assert_that(is_string(nm))

  # CRAN did not like these names
  forbidden_names <- c("pkg", "description")

  # package_name_rx is in parse-remotes.R
  if (
    grepl(paste0("^", package_name_rx(), "$"), nm) &&
      !tolower(nm) %in% forbidden_names
  ) {
    return(TRUE)
  }

  # From WRE:
  # The mandatory ‘Package’ field gives the name of the package.
  # This should contain only (ASCII) letters, numbers and dot, have at
  # least two characters and start with a letter and not end in a dot.

  why <- if (tolower(nm) %in% forbidden_names) {
    "Package name forbidden by CRAN."
  } else if (!is_ascii(nm)) {
    "It can only contain ASCII characters."
  } else if (grepl("[^a-zA-Z0-9.]", nm)) {
    "It can only contain letters, numbers and dot."
  } else if (nchar(nm) < 2) {
    "It must have at least two characters."
  } else if (!grepl("^[a-zA-Z]", nm)) {
    "It must start with a letter."
  } else if (grepl("[.]$", nm)) {
    "It must not end with a dot."
  } else {
    "Package name is invalid." # nocov
  }

  structure(FALSE, reason = why)
}

is_ascii <- function(x) {
  all(charToRaw(x) < 128)
}

test_that("is_valid_package_name", {
  expect_snapshot({
    is_valid_package_name("foo")
    is_valid_package_name("pkg")
    is_valid_package_name("f\u00e1f")
    is_valid_package_name("foo-bar")
    is_valid_package_name("x")
    is_valid_package_name("1xyz")
    is_valid_package_name("dotted.")
  })
})

tojson <- local({
  `%||%` <- function(l, r) if (is.null(l)) r else l

  map2 <- function(x, y, fn, ...) {
    mapply(fn, x, y, ..., SIMPLIFY = FALSE)
  }

  filter <- function(v, fn) {
    keep <- vapply(v, fn, logical(1))
    v[keep]
  }

  # FIXME: is this escaping the right things?
  jq <- function(x) {
    encodeString(x, quote = "\"", justify = "none")
  }

  # 1. add a "key": at the begining of each element, unless is.null(key)
  # 2. add a comma after each elelemt, except the last one
  # Each element can be a character vector, so `key` is added to the first
  # element of the character vectors, and comma to the last ones.
  comma <- function(x, opts, key = NULL) {
    len <- length(x)
    stopifnot(len >= 1)

    if (!is.null(key)) {
      nokey <- is.na(key) | key == ""
      key[nokey] <- seq_along(x)[nokey]
      x <- map2(jq(key), x, function(k, el) {
        el[1] <- paste0(k, if (opts$pretty) ": " else ":", el[1])
        el
      })
    }

    # No commas needed for scalars
    if (len == 1) {
      return(x)
    }

    x2 <- lapply(x, function(el) {
      el[length(el)] <- paste0(el[length(el)], ",")
      el
    })

    # Keep the last list element as is
    x2[[len]] <- x[[len]]
    x2
  }

  j_null <- function(x, opts) {
    "{}"
  }

  # Data frames are done row-wise.
  # Atomic columns are unboxed. Atomic NA values are omitted.
  # List columns remove the extra wrapping list.
  j_df <- function(x, opts) {
    sub <- unlist(comma(
      lapply(seq_len(nrow(x)), function(i) {
        row <- as.list(x[i, ])
        row <- filter(row, function(v) !(is.atomic(v) && is.na(v)))
        row[] <- lapply(row, function(v) {
          if (is.atomic(v)) {
            unbox(v)
          } else if (is.list(v)) {
            v[[1]]
          } else {
            v
          }
        })
        j_list(row, opts)
      })
    ))
    if (opts$pretty) {
      c("[", paste0("  ", sub), "]")
    } else {
      paste0(c("[", sub, "]"), collapse = "")
    }
  }

  # Returns a character vector. Named lists are dictionaries, unnnamed
  # ones are lists. Missing dictionary keys are filled in.
  # Keys do _NOT_ need to be unique.
  j_list <- function(x, opts) {
    if (length(x) == 0L) {
      if (is.null(names(x))) "[]" else "{}"
    } else if (is.null(names(x))) {
      sub <- unlist(comma(lapply(x, j, opts), opts))
      if (opts$pretty) {
        c("[", paste0("  ", sub), "]")
      } else {
        paste(c("[", sub, "]"), collapse = "")
      }
    } else {
      sub <- unlist(comma(lapply(x, j, opts), opts, names(x)))
      if (opts$pretty) {
        c("{", paste0("  ", sub), "}")
      } else {
        paste(c("{", sub, "}"), collapse = "")
      }
    }
  }

  # Atomic vectors are converted to lists, even if they have names.
  # The names are lost. Pretty formatting keeps a vector in one line
  # currently. NA is converted to null.
  j_atomic <- function(x, opts) {
    if (!typeof(x) %in% c("logical", "integer", "double", "character")) {
      stop("Cannot convert atomic ", typeof(x), " vectors to JSON.")
    }
    len <- length(x)

    if (len == 0) {
      return("[]")
    }

    unbox <- (opts$auto_unbox && len == 1) || "unbox" %in% class(x)

    if (is.character(x)) {
      x <- jq(enc2utf8(x))
    }

    if (is.logical(x)) {
      # tolower() keeps NAs, we'll sub them later
      x <- tolower(x)
    }

    if (unbox) {
      if (is.na(x) || x == "NA") "null" else paste0(x)
    } else {
      x[is.na(x) | x == "NA"] <- "null"
      sep <- if (opts$pretty) " " else ""
      paste0("[", paste(comma(x), collapse = sep), "]")
    }
  }

  j <- function(x, opts) {
    if (is.null(x)) {
      j_null(x, opts)
    } else if (is.data.frame(x)) {
      j_df(x, opts)
    } else if (is.list(x)) {
      j_list(x, opts)
    } else if (is.atomic(x)) {
      j_atomic(x, opts)
    } else {
      stop("Cannot convert type ", typeof(x), " to JSON.")
    }
  }

  write_str <- function(x, opts = NULL) {
    paste0(write_lines(x, opts), collapse = "\n")
  }

  write_file <- function(x, file, opts = NULL) {
    writeLines(write_lines(x, opts), file)
  }

  write_lines <- function(x, opts = NULL) {
    opts <- list(
      auto_unbox = opts$auto_unbox %||% FALSE,
      pretty = opts$pretty %||% FALSE
    )
    j(x, opts)
  }

  unbox <- function(x) {
    if (!is.atomic(x)) {
      stop("Can only unbox atomic scalar, not ", typeof(x), ".")
    }
    if (length(x) != 1) {
      stop("Cannot unbox vector of length ", length(x), ".")
    }
    class(x) <- c("unbox", class(x))
    x
  }

  list(
    .envir = environment(),
    write_str = write_str,
    write_file = write_file,
    write_lines = write_lines,
    unbox = unbox
  )
})

test_that("tojson is standalone", {
  stenv <- tojson$.envir
  objs <- ls(stenv, all.names = TRUE)
  funs <- Filter(function(x) is.function(stenv[[x]]), objs)
  funobjs <- mget(funs, stenv)

  orig <- parent.env(stenv)
  parent.env(stenv) <- baseenv()
  on.exit(parent.env(stenv) <- orig, add = TRUE)

  expect_message(
    mapply(
      codetools::checkUsage,
      funobjs,
      funs,
      MoreArgs = list(report = message)
    ),
    NA
  )
})

test_that("write_str vectors", {
  # empty vectors
  expect_snapshot({
    tojson$write_str(NULL)
    tojson$write_str(list())
    tojson$write_str(integer())
    tojson$write_str(double())
    tojson$write_str(character())
    tojson$write_str(logical())
  })

  # cannot convert some atomic vectors
  expect_snapshot(error = TRUE, {
    tojson$write_str(complex())
    tojson$write_str(raw())
  })

  # scalars
  expect_snapshot({
    tojson$write_str(list(1))
    tojson$write_str(1L)
    tojson$write_str(1.0)
    tojson$write_str("foo")
    tojson$write_str(TRUE)
  })

  # vectors
  expect_snapshot({
    tojson$write_str(list(1, 2))
    tojson$write_str(1:2)
    tojson$write_str(c(1.0, 2.0))
    tojson$write_str(c("foo", "bar"))
    tojson$write_str(c(TRUE, FALSE))
  })

  # NA in vectors
  expect_snapshot({
    tojson$write_str(NA_integer_)
    tojson$write_str(NA_real_)
    tojson$write_str(NA_character_)
    tojson$write_str(NA)
  })

  expect_snapshot({
    tojson$write_str(c(1L, NA_integer_))
    tojson$write_str(c(1.0, NA_real_))
    tojson$write_str(c("foo", NA_character_))
    tojson$write_str(c(TRUE, NA))
  })

  # names are ignored
  expect_snapshot({
    tojson$write_str(c(a = 1L, b = 2L))
    tojson$write_str(c(a = 1.0, b = 2.0))
    tojson$write_str(c(a = "foo", b = "bar"))
    tojson$write_str(c(a = TRUE, b = FALSE))
  })
})

test_that("write_str character encoding and escaping", {
  utf8 <- "G\u00e1bor"
  expect_snapshot({
    cat(tojson$write_str("foo\"\\bar"))
    charToRaw(tojson$write_str(iconv(utf8, "UTF-8", "latin1")))
  })
})

test_that("lists", {
  # embedded lists
  expect_snapshot({
    cat(tojson$write_str(list(list(1, 2, 3), list(4, 5, 6), list(7, 8, 9))))
  })

  # named lists
  expect_snapshot({
    cat(tojson$write_str(list(a = 1, b = 2)))
  })

  # nested named lists
  expect_snapshot({
    cat(tojson$write_str(list(
      a = list(a1 = 1, a2 = 2),
      b = list(b1 = 3, b2 = 4)
    )))
  })

  # fill in names
  expect_snapshot({
    cat(tojson$write_str(list(
      a = list(1, a2 = 2),
      list(b1 = 3, 4)
    )))
  })
})

test_that("write_file", {
  tmp <- tempfile()
  on.exit(unlink(tmp), add = TRUE)

  tojson$write_file(mtcars, tmp)
  lns <- tojson$write_lines(mtcars)
  expect_equal(readLines(tmp), lns)
})

test_that("auto-unbox", {
  # scalars
  expect_snapshot({
    tojson$write_str(list(1), opts = list(auto_unbox = TRUE))
    tojson$write_str(1L, opts = list(auto_unbox = TRUE))
    tojson$write_str(1.0, opts = list(auto_unbox = TRUE))
    tojson$write_str("foo", opts = list(auto_unbox = TRUE))
    tojson$write_str(TRUE, opts = list(auto_unbox = TRUE))
  })

  # NA in vectors
  expect_snapshot({
    tojson$write_str(NA_integer_, opts = list(auto_unbox = TRUE))
    tojson$write_str(NA_real_, opts = list(auto_unbox = TRUE))
    tojson$write_str(NA_character_, opts = list(auto_unbox = TRUE))
    tojson$write_str(NA, opts = list(auto_unbox = TRUE))
  })

  # embedded lists
  expect_snapshot({
    cat(tojson$write_str(
      list(list(1, 2, 3), list(4, 5, 6), list(7, 8, 9)),
      opts = list(auto_unbox = TRUE)
    ))
  })

  # named lists
  expect_snapshot({
    cat(tojson$write_str(
      list(a = 1, b = 2),
      opts = list(auto_unbox = TRUE)
    ))
  })

  # nested named lists
  expect_snapshot({
    cat(tojson$write_str(
      list(
        a = list(a1 = 1, a2 = 2),
        b = list(b1 = 3, b2 = 4)
      ),
      opts = list(auto_unbox = TRUE)
    ))
  })

  # mixing scalars and vectors
  expect_snapshot({
    cat(tojson$write_str(
      list(list(1, 2:3, 4), list(4:5, 6, 7:8), list(9:10, 11, 12)),
      opts = list(auto_unbox = TRUE)
    ))
  })

  expect_snapshot({
    cat(tojson$write_str(
      list(a = 1:3, b = 4),
      opts = list(auto_unbox = TRUE)
    ))
  })

  expect_snapshot({
    cat(tojson$write_str(
      list(
        a = list(a1 = 1, a2 = 2:3),
        b = list(b1 = 4:6, b2 = 7)
      ),
      opts = list(auto_unbox = TRUE)
    ))
  })
})

test_that("unbox", {
  # scalars
  expect_snapshot({
    tojson$write_str(list(tojson$unbox(1)))
    tojson$write_str(tojson$unbox(1L))
    tojson$write_str(tojson$unbox(1.0))
    tojson$write_str(tojson$unbox("foo"))
    tojson$write_str(tojson$unbox(TRUE))
  })

  # NA in vectors
  expect_snapshot({
    tojson$write_str(tojson$unbox(NA_integer_))
    tojson$write_str(tojson$unbox(NA_real_))
    tojson$write_str(tojson$unbox(NA_character_))
    tojson$write_str(tojson$unbox(NA))
  })

  # embedded lists
  expect_snapshot({
    cat(tojson$write_str(
      list(
        list(1, tojson$unbox(2), 3),
        list(tojson$unbox(4), 5, 6),
        list(7, 8, 9)
      )
    ))
  })

  # named lists
  expect_snapshot({
    cat(tojson$write_str(
      list(a = 1, b = tojson$unbox(2))
    ))
  })

  # nested named lists
  expect_snapshot({
    cat(tojson$write_str(
      list(
        a = list(a1 = tojson$unbox(1), a2 = 2),
        b = list(b1 = 3, b2 = tojson$unbox(4))
      )
    ))
  })

  expect_snapshot(error = TRUE, {
    tojson$unbox(list())
    tojson$unbox(1:2)
    tojson$unbox(double(2))
    tojson$unbox(character(2))
    tojson$unbox(logical(2))
  })
})

test_that("pretty", {
  # vectors
  expect_snapshot({
    tojson$write_str(1:5, opts = list(pretty = TRUE))
    tojson$write_str(1:5 / 2, opts = list(pretty = TRUE))
    tojson$write_str(letters[1:5], opts = list(pretty = TRUE))
    tojson$write_str(1:5 %% 2 == 0, opts = list(pretty = TRUE))
  })

  # lists
  expect_snapshot({
    cat(tojson$write_str(
      list(
        list(1, tojson$unbox(2), 3),
        list(tojson$unbox(4), 5, 6),
        list(7, 8, 9)
      ),
      opts = list(pretty = TRUE)
    ))
  })

  # named lists
  expect_snapshot({
    cat(tojson$write_str(
      list(a = 1, b = tojson$unbox(2)),
      opts = list(pretty = TRUE)
    ))
  })

  # nested named lists
  expect_snapshot({
    cat(tojson$write_str(
      list(
        a = list(a1 = tojson$unbox(1), a2 = 2),
        b = list(b1 = 3, b2 = tojson$unbox(4))
      ),
      opts = list(pretty = TRUE)
    ))
  })
})

test_that("data frames", {
  # easy
  df <- mtcars[1:3, 1:3]
  expect_snapshot({
    tojson$write_str(df)
    cat(tojson$write_str(df, opts = list(pretty = TRUE)))
  })

  # NAs
  df[1, 1] <- df[2, 3] <- NA
  expect_snapshot({
    tojson$write_str(df)
    cat(tojson$write_str(df, opts = list(pretty = TRUE)))
  })

  # list columns
  df$list <- I(list(1:2, 5, letters[1:3]))
  expect_snapshot({
    tojson$write_str(df)
    cat(tojson$write_str(df, opts = list(pretty = TRUE)))
  })
})

#' @details
#' `current_r_platform()` detects the platform of the current R version.
#'
#' By default pkgdepends works with source packages and binary packages for
#' the current platform. You can change this, see
#' ['Configuration'][pkgdepends-config].
#'
#' The following platform names can be configured and returned by
#' `current_r_platform()` and `default_platforms()`:
#' * `"source"` for source packages,
#' * A platform string like `R.version$platform`, but on Linux the name
#'   and version of the distribution are also included. Examples:
#'   - `x86_64-apple-darwin17.0`: macOS High Sierra.
#'   - `aarch64-apple-darwin20`: macOS Big Sur on arm64.
#'   - `x86_64-w64-mingw32`: 64 bit Windows.
#'   - `i386-w64-mingw32`: 32 bit Windows.
#'   - `i386+x86_64-w64-mingw32`: 64 bit + 32 bit Windows.
#'   - `i386-pc-solaris2.10`: 32 bit Solaris. (Some broken 64 Solaris
#'     builds might have the same platform string, unfortunately.)
#'   - `x86_64-pc-linux-gnu-debian-10`: Debian Linux 10 on x86_64.
#'   - `x86_64-pc-linux-musl-alpine-3.14.1`: Alpine Linux.
#'   - `x86_64-pc-linux-gnu-unknown`: Unknown Linux Distribution on x86_64.
#'   - `s390x-ibm-linux-gnu-ubuntu-20.04`: Ubuntu Linux 20.04 on S390x.
#'   - `amd64-portbld-freebsd12.1`: FreeBSD 12.1 on x86_64.
#'
#' In addition, the following platform names can be used to configure
#' pkgdepends:
#' * `"macos"` for macOS binaries that are appropriate for the R versions
#'   pkgdepends is working with (defaulting to the version of the current
#'   session), as defined by CRAN binaries. E.g. on R 3.5.0 macOS binaries
#'   are built for macOS El Capitan.
#' * `"windows"` for Windows binaries for the default CRAN architecture.
#'   This is currently Windows Vista for all supported R versions, but it
#'   might change in the future. The actual binary packages in the
#'   repository might support both 32 bit and 64 builds, or only one of
#'   them. In practice 32-bit only packages are very rare. CRAN builds
#'   before and including R 4.1 have both architectures, from R 4.2 they
#'   are 64 bit only. `"windows"` is an alias to `i386+x86_64-w64-mingw32`
#'   currently.
#'
#' @return
#' `current_r_platform()` returns a string, the name of the current
#' platform.
#'
#' @export
#' @rdname default_platforms

current_r_platform <- function() {
  pkgcache::current_r_platform()
}

#' R platforms
#'
#'`default_platfoms()` returns the default platforms for the current R
#' session. These typically consist of the detected platform of the current
#' R session, and `"source"`, for source packages.
#'
#' @return
#' `default_platforms()` returns a character vector of platform names.
#'
#' @family platform functions
#' @export
#' @examples
#' current_r_platform()
#' default_platforms()

default_platforms <- function() unique(c(current_r_platform(), "source"))

# Is `cand` an OK platform for `exp`? This is pretty straightforward,
# except for windows.
#
# NOTE: RSPM delivers binaries as source packages, so we
#       do not need to handle RSPM specially, source packages
#       will be always accepted, unless the user explicitly
#       opts out from them.

platform_is_ok <- function(cand, exp, exp_archs = NULL) {
  if (cand %in% c("*", "source") && "source" %in% exp) {
    return(TRUE)
  }

  # This is an installed linux package, probably, prefix is OK, e.g.
  # cand = x86_64-pc-linux-gnu vs exp = x86_64-pc-linux-gnu-ubuntu-18.04
  if (grepl("-linux-", cand) && any(substr(exp, 1, nchar(cand)) == cand)) {
    return(TRUE)
  }

  if (cand %in% c("i386+x86_64-w64-mingw32", "x86_64+i386-w64-mingw32")) {
    # This is a multi-arch binary, that is OK, if binaries are allowed
    any(c("x86_64-w64-mingw32", "i386-w64-mingw32") %in% exp)
  } else if (cand == "x86_64-w64-mingw32") {
    # This is an x64 only binary. If we are on x64 and we "prefer-x64"
    # then it is OK. Otherwise we would prefer a multi-arch binary or
    # a source package.
    "x86_64-w64-mingw32" %in% exp && exp_archs == "prefer-x64"
  } else if (cand == "i386-w64-mingw32") {
    # This is an i386 only binary. This is not OK currently.
    # (We do not allow an i386-only installation.)
    FALSE
  } else if (cand == "") {
    # PPM with binaries gives an empty string if the package doesn't require
    # compilation. So this is OK.
    TRUE
  } else {
    # Otherwise it is just a match
    cand %in% exp
  }
}

repo <- local({
  # List packages in local repo -------------------------------------------

  repo_list <- function(..., path = ".") {
    pkgs <- suppressWarnings(
      pkgcache::parse_packages(file.path(path, "PACKAGES"))
    )
    if (nrow(pkgs) == 0) {
      pkgs$Package <- character()
      pkgs$Version <- character()
      pkgs$Depends <- character()
      pkgs$Imports <- character()
      pkgs$Suggests <- character()
      pkgs$Enhances <- character()
      pkgs$LinkingTo <- character()
      pkgs$License <- character()
      pkgs$File <- character()
      pkgs$DownloadURL <- character()
      pkgs$OS <- character()
      pkgs$Arch <- character()
      pkgs$Built <- character()
      pkgs$Filesize <- character()
      pkgs$SHA256 <- character()
      pkgs$RVersion <- character()
      pkgs$Platform <- character()
      pkgs$GraphicsAPIVersion <- character()
      pkgs$InternalsId <- character()
      pkgs$SystemRequirements <- character()
    }

    pkgs
  }

  # Delete package from local repo ----------------------------------------

  repo_delete <- function(package, ..., path = ".") {
    pkgs <- repo_list(path = path)
    idx <- find_in_data_frame(pkgs, package = package, ...)
    if (length(idx)) {
      pkgs <- pkgs[-idx, , drop = FALSE]
    }
    PACKAGES <- file.path(path, "PACKAGES")
    write_dcf(pkgs, PACKAGES)
  }

  # Add package to local repo ---------------------------------------------

  repo_add <- function(file, ..., path = ".", release_org = "cran") {
    pkgs <- repo_list(path = path)
    pkg_data <- get_package_data(file, release_org = release_org)
    pkgs <- rbind_expand(pkgs, pkg_data)
    PACKAGES <- file.path(path, "PACKAGES")
    write_dcf(pkgs, PACKAGES)
  }

  # Update package in local repo ------------------------------------------

  repo_update <- function(file, ..., path = ".", release_org = "cran") {
    pkgs <- repo_list(path = path)
    pkg_data <- get_package_data(file, release_org = release_org)

    idx <- which(
      pkgs$Package == pkg_data$Package &
        pkgs$RVersion == pkg_data$RVersion &
        (is.na(pkg_data$OS) | is.na(pkgs$OS) | pkgs$OS == pkg_data$OS) &
        (is.na(pkg_data$Arch) | is.na(pkgs$Arch) | pkgs$Arch == pkg_data$Arch)
    )
    if (length(idx)) {
      pkgs <- pkgs[-idx, , drop = FALSE]
    }

    pkgs <- rbind_expand(pkgs, pkg_data)
    PACKAGES <- file.path(path, "PACKAGES")
    write_dcf(pkgs, PACKAGES)
  }

  # List packages in repo on GH -------------------------------------------

  repo_list_gh <- function(repo, subdir) {
    dir.create(tmp <- tempfile())
    on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
    get_packages(repo, subdir, tmp)
    repo_list(path = tmp)
  }

  # Update packages in GH repo --------------------------------------------

  repo_update_gh <- function(repo, subdir, files, release_org = "cran") {
    files <- normalizePath(files)

    oldwd <- getwd()
    workdir <- tempfile()
    on.exit(
      {
        setwd(oldwd)
        unlink(workdir, recursive = TRUE)
      },
      add = TRUE
    )
    dir.create(workdir)
    setwd(workdir)

    git("clone", paste0("https://github.com/", repo))
    prepo <- parse_slug(repo)
    setwd(prepo$repo)
    setwd(subdir)

    git("config", "user.email", "csardi.gabor@gmail.com")
    git("config", "user.name", "Gabor Csardi")
    git("config", "credential.helper", "cache")
    gitcreds::gitcreds_approve(list(
      url = "https://github.com",
      username = "PersonalAccessToken",
      password = Sys.getenv("GITHUB_PAT")
    ))

    repeat {
      git("pull")
      for (file in files) {
        repo_update(file, release_org = release_org)
      }
      tryCatch(
        {
          git_push()
          break
        },
        error = function(err) {
          if (
            !grepl("(non-fast-forward|fetch first|cannot lock ref)", err$stderr)
          ) {
            stop(err)
          }
          git("reset", "HEAD^")
          git("checkout", "--", ".")
        }
      )
    }

    invisible()
  }

  # -----------------------------------------------------------------------
  # Internal functions

  get_desc <- function(path) {
    pkg <- sub("_.*$", "", basename(path))
    dir.create(tmp <- tempfile())
    on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
    untar(path, files = paste0(pkg, "/DESCRIPTION"), exdir = tmp)
    desc::desc(file.path(tmp, pkg))
  }

  get_package_data <- function(path, release_org = "cran") {
    desc <- get_desc(path)
    pkgname <- desc$get_field("Package")
    pkgver <- desc$get_field("Version")
    built <- desc$get_built()
    rminor <- get_minor_r_version(built$R)
    os <- canonize_os(built$Platform)
    arch <- canonize_arch(built$Platform)
    deps <- unname(desc$get("Depends"))
    pkg <- data_frame(
      Package = pkgname,
      Version = pkgver,
      Depends = paste0(
        if (!is.na(deps)) paste0(deps, ", "),
        "R (>= ",
        rminor,
        "), R (< ",
        rminor,
        ".99)"
      ),
      Imports = unname(desc$get("Imports")),
      Suggests = unname(desc$get("Suggests")),
      Enhances = unname(desc$get("Enhances")),
      LinkingTo = unname(desc$get("LinkingTo")),
      License = unname(desc$get("License")),
      File = basename(path),
      DownloadURL = paste0(
        "https://github.com/",
        release_org,
        "/",
        pkgname,
        "/releases/download/",
        pkgver,
        "/",
        basename(path)
      ),
      OS = os,
      Arch = arch,
      Built = desc$get_field("Built"),
      Filesize = file.size(path),
      SHA256 = cli::hash_file_sha256(path),
      RVersion = rminor,
      Platform = unname(desc$get("RemoteBuildPlatform")),
      GraphicsAPIVersion = unname(desc$get("GraphicsAPIVersion")),
      InternalsId = unname(desc$get("InternalsId")),
      SystemRequirements = unname(desc$get("SystemRequirements"))
    )

    pkg
  }

  write_dcf <- function(meta, PACKAGES, quiet = TRUE) {
    if (!quiet) {
      cat("Writing ", PACKAGES, "\n")
    }
    meta <- as.matrix(meta)
    write.dcf(meta, PACKAGES, width = 200)
    con <- gzfile(paste0(PACKAGES, ".gz"), "wt")
    write.dcf(meta, con, width = 200)
    close(con)
    saveRDS(meta, paste0(PACKAGES, ".rds"), compress = "xz", version = 2)
    invisible()
  }

  # case insensitive!

  find_in_data_frame <- function(df, ..., .list = NULL) {
    cols <- drop_nulls(c(list(...), .list))
    idx <- seq_len(nrow(df))
    names(df) <- tolower(names(df))
    for (i in seq_along(cols)) {
      if (length(idx) == 0) {
        break
      }
      n <- tolower(names(cols)[i])
      idx <- idx[df[[n]][idx] %in% cols[[i]]]
    }

    idx
  }

  canonize_arch <- function(platform) {
    if (platform == "") {
      return(NA_character_)
    }
    arch <- strsplit(platform, "-", fixed = TRUE)[[1]][1]
    c("aarch64" = "arm64", "x86_64" = "amd64", "s390x" = "s390x")[[arch]]
  }

  canonize_os <- function(platform) {
    if (platform == "") {
      return(NA_character_)
    }
    os <- strsplit(platform, "-", fixed = TRUE)[[1]][3]
    if (substr(os, 1, 6) == "darwin") {
      os <- "darwin"
    }
    if (substr(os, 1, 5) == "mingw") {
      os <- "windows"
    }
    if (substr(os, 1, 7) == "solaris") {
      os <- "solaris"
    }
    os
  }

  git <- function(
    ...,
    echo_cmd = TRUE,
    echo = TRUE,
    dry_run = FALSE,
    stderr_to_stdout = FALSE
  ) {
    if (dry_run) {
      cat("git", c(...), "\n")
    } else {
      processx::run(
        "git",
        c(...),
        echo_cmd = echo_cmd,
        echo = echo,
        stderr_to_stdout = stderr_to_stdout
      )
    }
  }

  git_push <- function() {
    git("add", "-A", ".")
    git("commit", "-m", "Update repository")
    git("push", "--porcelain", "origin", stderr_to_stdout = TRUE)
  }

  repo_columns <- function() {
    dir.create(tmp <- tempfile())
    on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
    file.create(file.path(tmp, "PACKAGES"))
    paste0("* `", colnames(repo_list(path = tmp)), "`", collapse = "\n")
  }

  get_packages <- function(repo, subdir, target) {
    url <- paste0("https://github.com/", repo)
    files <- git_list_files(url)
    wh <- which(files$files$path == paste0(subdir, "/PACKAGES.gz"))[1]
    if (is.na(wh)) {
      wh <- which(files$files$path == paste0(subdir, "/PACKAGES"))[1]
    }
    if (is.na(wh)) {
      wh <- which(files$files$path == paste0(subdir, "/PACKAGES.rds"))[1]
    }
    if (is.na(wh)) {
      throw(pkg_error(
        "Cannot file `PACKAGES*` in {.val {subdir}} at repo {.val {repo}}."
      ))
    }
    hash <- files$files$hash[wh]
    git_download_file(url, hash, file.path(target, "PACKAGES"))
  }

  parse_slug <- function(slug) {
    parts <- strsplit(slug, "/", fixed = TRUE)[[1]]
    list(user = parts[1], owner = parts[1], repo = parts[2])
  }

  # -----------------------------------------------------------------------
  # Exported functions

  structure(
    list(
      .internal = environment(),
      add = repo_add,
      delete = repo_delete,
      list = repo_list,
      list_gh = repo_list_gh,
      update = repo_update,
      update_gh = repo_update_gh
    ),
    class = c("standalone_repo", "standalone")
  )
})

# -------------------------------------------------------------------------

#' Query and manipulate CRAN-like repositories
#'
#' These functions are currently experimental.
#'
#' @details
#'
#' ## List packages in a repository
#'
#' `repo$list()` lists packages in a repository. It reads the `PACKAGES`
#' file containing the repository metadata.
#'
#' ### Usage
#' ```
#' repo_list(..., path = ".")
#' ```
#'
#' ### Arguments
#'
#' * `...`: ignored currently.
#' * `path`: path to repository. Must contain a `PACKAGES` file.
#'
#' ### Value
#'
#' Data frame of package data, a data frame with at least the following
#' columns, possibly more if there are other entries in the metadata:
#' `r repo$.internal$repo_columns()`
#'
# -------------------------------------------------------------------------
#'
#' ## Delete packages from repository metadata
#'
#' `repo$delete()` deletes matching packages from the repository
#' metadata.
#'
#' ### Description
#'
#' All matching packages will be removed.
#' It does not delete the files themselves.
#'
#' ### Usage
#' ```
#' repo$delete(package, ..., path = ".")
#' ```
#'
#' ### Arguments
#'
#' * `package`: package name.
#' * `...`: other fields to match, they must be named. Matching is
#'   case insensitive.
#' * `path`: path to repository. Must contain a `PACKAGES` file.
#'
# -------------------------------------------------------------------------
#'
#' ## Add a package to a repository
#'
#' ### Description
#'
#' It does not check if any version of the package is already
#' in the repository. If you want to _update_ a package, use
#' `repo$update()`.
#'
#' ### Usage
#' ```
#' repo$add(file, ..., path = ".")
#' ```
#'
#' ### Arguments
#'
#' * `file Package file.
#' * `... Ignored currently.
#' * `path Path to repository. Must contain a `PACKAGES` file.
#'
# -------------------------------------------------------------------------
#'
#' ## Update a package in a repository
#'
#' ### Description
#'
#' Previous version of the same package are removed. In particular,
#' it removes packages with matching:
#' - package name (`Package` field),
#' - R version (`Rversion` field),
#' - same OS (`OS` field), or no `OS` field,
#' - same architecture (`Arch` field), or not `Arch` field.
#'
#' ### Usage
#' ```
#' repo$update(file, ..., path = ".")
#' ```
#'
#' ### Arguments
#'
#' * `file Package file.
#' * `... Ignored currently.
#' * `path Path to repository. Must contain a `PACKAGES` file.
#'
# -------------------------------------------------------------------------
#'
#' ## Update a file in a package metadata, stored on GitHub
#'
#' ### Description
#'
#' 1. Clones the GitHub repository.
#' 2. Calls `repo_update()` with `file`, in the `subdir` directory.
#' 3. Adds and commits changes.
#' 4. Pushes the git repository to GitHub.
#' If the push fails, then it resets the git repository, pulls it from
#' GitHub and tries the update process again, until the push succeeds.
#'
#' This function needs command line git installed.
#'
#' It sets up a `cache` credential helper, so the `git push` works
#' without interaction with the user.
#'
#' ### Usage
#' ```
#' repo$update_gh(repo, subdir, files)
#' ```
#'
#' ### Arguments
#'
#' * `repo`: GitHub slug, e.g. `r-hub/repos`.
#' * `subdir`: subdirectory in the GitHub repository, where the R package
#'   metadata should be updated. It must exist in the repository.
#'   If it does not have `PACKAGES*` files, then they will be created.
#' * `files`: package files to add. The files will _not_ be added to the
#'   repository, only to the metadata. They should be in the repository
#'   already.
#'
# -------------------------------------------------------------------------
#'
#' @name repo
#' @keywords internal
#' @export

repo

make_unzip_process <- function(
  zipfile,
  exdir = ".",
  post_process = NULL,
  stdout = "|",
  stderr = "2>&1",
  ...
) {
  up <- zip::unzip_process()
  up$new(
    zipfile,
    exdir = exdir,
    post_process = post_process,
    stdout = stdout,
    stderr = stderr,
    ...
  )
}

## ------------------------------------------------------------------------
## API

parse_remote_installed <- function(specs, config, ...) {
  parsed_specs <- re_match(specs, type_installed_rx())

  parsed_specs$ref <- parsed_specs$.text
  cn <- setdiff(colnames(parsed_specs), c(".match", ".text"))
  parsed_specs <- parsed_specs[, cn]
  parsed_specs$type <- "installed"
  lapply(
    seq_len(nrow(parsed_specs)),
    function(i) as.list(parsed_specs[i, ])
  )
}

resolve_remote_installed <- function(
  remote,
  direct,
  config,
  cache,
  dependencies,
  ...
) {
  deps <- setdiff(dependencies[[2 - direct]], c("LinkingTo", "linkingto"))
  resolve_installed(cache, remote, direct, deps)
}

download_remote_installed <- function(
  resolution,
  target,
  target_tree,
  config,
  cache,
  which,
  on_progress
) {
  "Had"
}

satisfy_remote_installed <- function(resolution, candidate, config, ...) {
  TRUE
}

installedok_remote_installed <- function(installed, solution, config, ...) {
  # Keep the plan, we were already using the installed package
  FALSE
}

## ----------------------------------------------------------------------
## Internal functions

type_installed_rx <- function() {
  paste0(
    "^",
    "(?:installed::)?",
    "(?<library>.*)/",
    "(?<package>",
    package_name_rx(),
    ")",
    "$"
  )
}

make_installed_cache <- function(library, packages = NULL, priority = NULL) {
  if (length(library) == 0) {
    return(NULL)
  }
  installed <- make_installed_cache1(library[1], packages, priority)
  for (lib in library[-1]) {
    installed <- merge_installed_caches(
      installed,
      make_installed_cache1(lib, packages, priority)
    )
  }
  installed
}

make_installed_cache1 <- function(library, packages = NULL, priority = NULL) {
  inst <- pkgcache::parse_installed(
    library = library,
    priority = priority,
    lowercase = TRUE,
    reencode = FALSE
  )

  # Need nrow(inst) > 0 because otherwise it drops to a vector, and
  # R ignores 'drop' in this case. (?!)
  if (!is.null(packages) && nrow(inst) > 0) {
    inst <- inst[inst$package %in% packages, ]
  }

  all_fields <- names(inst)
  fields <- tolower(unique(c(
    "Package",
    "Title",
    "Version",
    "Depends",
    "Suggests",
    "Imports",
    "LinkingTo",
    "Enhances",
    "Built",
    "MD5sum",
    "NeedsCompilation",
    "Platform",
    "License",
    "Priority",
    "Repository",
    "biocViews",
    grep("^remote", all_fields, value = TRUE),
    grep("^config/needs/", all_fields, value = TRUE)
  )))

  miss <- setdiff(fields, names(inst))
  for (nm in miss) {
    inst[[nm]] <- if (nrow(inst)) NA_character_ else character()
  }

  pkgs <- inst[, names(inst) %in% fields]

  # Called `sysreqs` in pkgcache, so we call it the same here
  pkgs$sysreqs <- inst[["systemrequirements"]] %||%
    rep(NA_character_, nrow(pkgs))

  if (nrow(pkgs) == 0) {
    pkgs$ref <- character()
  } else {
    pkgs$ref <- paste0("installed::", library, "/", pkgs$package)
  }
  pkgs$type <- rep("installed", nrow(pkgs))
  pkgs$status <- rep("OK", nrow(pkgs))
  built <- parse_built(inst$built)
  pkgs$rversion <- built$R
  pkgs$platform <- built$Platform

  # This is usually a broken installation, but 'Built' is also missing
  # for the core translations package
  pkgs$rversion[is.na(pkgs$rversion)] <- current_r_version()
  pkgs$platform[is.na(pkgs$platform)] <- current_r_platform()

  # On Windows, we need to check the Archs field
  winbin <- pkgs$platform != "" &
    !is.na(built$OStype) &
    built$OStype == "windows"
  if (any(winbin)) {
    # `archs` could be missing, because of a base R bug:
    # https://github.com/r-lib/pak/issues/448#issuecomment-1354807441
    if (is.null(inst[["archs"]])) {
      inst$archs <- rep(NA_character_, nrow(inst))
    }
    # we assume x64 on newer R, and i386 + x64 on older R, as this is
    # what typically happens on CRAN
    rver <- sub("R ", "", pkgs$rversion)
    inst[["archs"]] <- ifelse(
      winbin & is.na(inst[["archs"]]),
      ifelse(package_version(rver) < "4.2.0", "i386,x64", "x64"),
      inst[["archs"]]
    )
    archs <- gsub(" ", "", inst$archs[winbin])
    pkgs$platform[winbin] <- ifelse(
      is.na(archs) | archs %in% c("i386,x64", "x64,i386"),
      "i386+x86_64-w64-mingw32",
      pkgs$platform[winbin]
    )
  }
  pkgs$platform[pkgs$platform == ""] <- "*"

  pkgs$sources <- replicate(nrow(pkgs), character(), simplify = FALSE)
  pkgs$needscompilation <- ifelse(
    is.na(pkgs$needscompilation),
    NA,
    tolower(pkgs$needscompilation) %in% c("true", "yes")
  )

  cran <- !is.na(pkgs$repository) & pkgs$repository == "CRAN"
  bioc <- !is.na(pkgs$biocviews) & pkgs$biocviews != ""
  pkgs$repotype <- ifelse(cran, "cran", ifelse(bioc, "bioc", NA_character_))

  deps <- packages_parse_deps(pkgs)
  pkgs_deps <- split(
    deps[, -(1:2)],
    factor(deps$idx, levels = seq_len(nrow(pkgs)))
  )
  pkgs$deps <- unname(pkgs_deps)
  list(pkgs = pkgs, deps = deps)
}

merge_installed_caches <- function(c1, c2) {
  newdeps <- c2$deps
  newdeps$idx <- newdeps$idx + nrow(c1$pkgs)
  list(
    pkgs = rbind_expand(c1$pkgs, c2$pkgs),
    deps = rbind(c1$deps, newdeps)
  )
}

#' Status of packages in a library
#'
#' Query data of all packages in a package library.
#'
#' @param library Path to library.
#' @param packages If not `NULL`, then only these packages are shown.
#' @return Data frame that contains data about the packages
#'   installed in the library.
#'   ```{r child = "tools/doc/lib-status-return.Rmd"}
#'   ```
#'  `r doc_share_rmd("tools/doc/lib-status-return.Rmd", "inst/docs/lib-status-return.rds")`
#'
#' @export

lib_status <- function(library = .libPaths()[1], packages = NULL) {
  st <- make_installed_cache(library, packages)$pkgs
  st$library <- if (nrow(st) > 0) library else character()
  st <- st[, c("library", setdiff(colnames(st), "library")), drop = FALSE]
  rm <- extra_config_fields(colnames(st))
  st[, setdiff(colnames(st), rm), drop = FALSE]
}

# TODO: parse Remotes and Config/Needs/* fields

packages_parse_deps <- function(pkgs) {
  no_pkgs <- nrow(pkgs)
  cols <- intersect(colnames(pkgs), tolower(pkg_dep_types()))
  ## as.character is for empty data frames, e.g. from empty BioC repos
  deps <- as.character(unlist(pkgs[, cols], use.names = FALSE))
  nna <- which(!is.na(deps))
  if (length(nna)) {
    not_na_deps <- deps[nna]
    sp <- strsplit(not_na_deps, ",", fixed = TRUE)
    ll <- sapply(sp, length, USE.NAMES = FALSE)
    sp <- unlist(sp, use.names = FALSE)
    parsed <- re_match(
      sp,
      paste0(
        "^\\s*(?<package>[^(\\s]+)\\s*",
        "(?:\\((?<op>[^0-9\\s]+)\\s*(?<version>[^)\\s]+)\\))?\\s*$"
      )
    )
    parsed$idx <- rep(rep(seq_len(no_pkgs), length(cols))[nna], ll)
    parsed$type <- rep(rep(cols, each = no_pkgs)[nna], ll)
    parsed$ref <- parsed$package
    parsed$upstream <- pkgs$package[parsed$idx]
    parsed <- parsed[, c(
      "upstream",
      "idx",
      "ref",
      "type",
      "package",
      "op",
      "version"
    )]
    parsed <- parsed[order(parsed$idx), ]
  } else {
    parsed <- data_frame(
      upstream = character(),
      idx = integer(),
      ref = character(),
      type = character(),
      package = character(),
      version = character(),
      op = character()
    )
  }

  parsed
}

resolve_installed <- function(cache, remotes, direct, dependencies) {
  dependencies <- tolower(dependencies)

  ## Single remote, or a list of remotes
  if ("ref" %in% names(remotes)) {
    refs <- remotes$ref
    packages <- remotes$package
    params <- list(remotes$params %||% character())
  } else {
    refs <- vcapply(remotes, "[[", "ref")
    packages <- vcapply(remotes, "[[", "package")
    params <- lapply(remotes, "[[", "params")
  }

  pkgs <- cache$installed$pkgs
  cols <- c(
    "ref",
    "type",
    "status",
    "package",
    "version",
    "license",
    "needscompilation",
    "priority",
    "md5sum",
    "platform",
    "rversion",
    "sources",
    "built",
    "deps",
    "repotype",
    "sysreqs"
  )
  res <- pkgs[pkgs$package %in% packages, cols]
  repotype <- pkgs$repotype[pkgs$package %in% packages]

  res$direct <- direct
  res$metadata <- get_installed_metadata(res)
  res$deps <- lapply(res$deps, function(x) x[x$type %in% dependencies, ])
  # this might include extra rows from recommended packages
  idx <- match(res$ref, refs)
  res$params <- replicate(nrow(res), character())
  res$params[!is.na(idx)] <- params[na.omit(idx)]

  extracols <- c("repotype", grep("^remote", names(pkgs), value = TRUE))
  extra <- pkgs[pkgs$package %in% packages, extracols, drop = FALSE]
  res$extra <- lapply(seq_len(nrow(res)), function(i) extra[i, , drop = FALSE])

  attr(res, "unknown_deps") <-
    setdiff(unique(unlist(lapply(res$deps, "[[", "package"))), "R")

  res
}

get_installed_metadata <- function(tab) {
  meta <- replicate(nrow(tab), character(), simplify = FALSE)
  for (i in seq_len(nrow(tab))) {
    meta[[i]] <-
      c(
        RemoteType = tab$type[i],
        RemotePkgRef = tab$ref[i],
        RemotePkgPlatform = tab$platform[i],
        RemoteSha = tab$version[i]
      )
  }
  meta
}

parse_built <- function(x) {
  xp <- strsplit(x, ";", fixed = TRUE)
  data.frame(
    stringsAsFactors = FALSE,
    R = trimws(vapply(xp, "[", character(1), 1)),
    Platform = trimws(vapply(xp, "[", character(1), 2)),
    Date = trimws(vapply(xp, "[", character(1), 3)),
    OStype = trimws(vapply(xp, "[", character(1), 4))
  )
}

test_that("resolve", {
  setup_fake_apps()

  conf <- current_config()

  tt <- dirname(dirname(attr(packageDescription("testthat"), "file")))
  cache <- list(
    package = NULL,
    metadata = pkgcache::get_cranlike_metadata_cache(),
    installed = make_installed_cache(dirname(tt))
  )

  ref <- paste0("installed::", tt)
  res <- synchronise(
    resolve_remote_installed(
      parse_pkg_refs(ref)[[1]],
      TRUE,
      conf,
      cache,
      dependencies = "Imports"
    )
  )

  unun <- function(x) {
    attr(x, "unknown_deps") <- NULL
    x
  }

  expect_equal(
    unun(as.list(res[c(
      "ref",
      "type",
      "direct",
      "status",
      "package",
      "version"
    )])),
    list(
      ref = ref,
      type = "installed",
      direct = TRUE,
      status = "OK",
      package = "testthat",
      version = as.character(utils::packageVersion("testthat"))
    )
  )

  expect_true("cli" %in% attr(res, "unknown_deps"))

  expect_false(is.null(res$extra[[1]]$repotype))
})

test_that("download", {
  setup_fake_apps()

  dir.create(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  dir.create(tmp2 <- tempfile())
  on.exit(unlink(tmp2, recursive = TRUE), add = TRUE)

  tt <- dirname(dirname(attr(packageDescription("testthat"), "file")))
  ref <- paste0("installed::", tt)
  r <- pkg_plan$new(
    ref,
    library = dirname(tt),
    config = list(dependencies = FALSE, cache_dir = tmp)
  )
  expect_error(suppressMessages(r$resolve()), NA)
  expect_error(suppressMessages(r$download_resolution()), NA)
  dl <- r$get_resolution_download()
  expect_equal(dl$download_status, "Had")
})

test_that("satisfy", {
  ## Always TRUE, independently of arguments
  expect_true(satisfy_remote_installed())
})

test_that("installedok_remote_installed", {
  expect_false(installedok_remote_installed())
})

test_that("make_installed_cache", {
  # packages argument is used correctly
  cache <- make_installed_cache(.Library, "boot")
  expect_equal(cache$pkgs$package, "boot")
})

test_that("make_installed_cache edge case", {
  tmp <- withr::local_tempdir()
  cache <- make_installed_cache(tmp, "MASS")
  expect_equal(nrow(cache$pkgs), 0)
  expect_equal(nrow(cache$deps), 0)
  expect_true(ncol(cache$pkgs) >= 20) # approx
  expect_true(ncol(cache$deps) >= 5) # approx
})

#' A set of handy regular expressions related to R packages
#'
#' If you use these in R, make sure you specify `perl = TRUE`, see
#' [base::grep()].
#'
#' Currently included:
#' - `pkg_name`: A valid package name.
#' - `type_cran`: A `cran::` package reference.
#' - `type_bioc`: A `bioc::` package reference.
#' - `type_standard`: A `standard::` package reference.
#' - `type_github`: A `github::` package reference.
#' - `type_git`: A `git::` package reference.
#' - `type_local`: A `local::` package reference.
#' - `type_deps`: A `deps::` package reference.
#' - `type_installed`: An `installed::` package reference.
#' - `github_username`: A GitHub username.
#' - `github_repo`: A GitHub repository name.
#' - `github_url`: A GitHub URL.
#'
#' @return A named list of strings.
#'
#' @export
#' @examples
#' pkg_rx()

pkg_rx <- function() {
  list(
    pkg_name = package_name_rx(),
    type_cran = standard_rx("cran"),
    type_bioc = standard_rx("bioc"),
    type_standard = standard_rx("standard"),
    type_github = github_rx(),
    type_git = git_rx(),
    type_local = local_rx(),
    type_deps = type_deps_rx(),
    type_installed = type_installed_rx(),
    github_username = github_username_rx(),
    github_repo = github_repo_rx(),
    github_url = github_url_rx()
  )
}

# match any library(), require(), base::library(), base::require(), etc. calls
# we use these as fallbacks. If a call is not identified some other way
# we parse it with R and match the call.
q_library_0 <- function() {
  structure(
    c(
      '((call function: (identifier) @fn-name) @dep-code
      (#any-of? @fn-name
       "library" "require" "loadNamespace" "requireNamespace"
       "pkg_attach" "pkg_attach2"
       "p_load"
       "module"
       "tar_option_set"
       "glue"
       "ggsave"
       "set_engine"
       "R6Class" "test_package" "test_dir" "test_file"))',
      '((call function:
       (namespace_operator
        lhs: (identifier) @ns-name
        rhs: (identifier) @fn-name
       )
      ) @dep-code
      (#any-of? @ns-name
       "base" "xfun" "pacman" "modules" "import" "box" "targets" "glue"
       "ggplot2" "parsnip" "R6" "testthat")
      (#any-of? @fn-name
       "library" "require" "loadNamespace" "requireNamespace"
       "pkg_attach" "pkg_attach2"
       "p_load"
       "module" "import"
       "from" "here" "into"
       "use"
       "tar_option_set"
       "glue"
       "ggsave"
       "set_engine"
       "R6Class" "test_package" "test_dir" "test_file"))'
    ),
    names = rep("q_library_0", 2)
  )
}

q_module_import <- function() {
  c(
    '((call function: (identifier) @fn-name) @dep-code
      (#any-of? @fn-name "import"))',
    '((call function:
       (namespace_operator
        lhs: (identifier) @ns-name
        rhs: (identifier) @fn-name
       )
      ) @dep-code
      (#any-of? @ns-name "modules")
      (#any-of? @fn-name "import"))'
  )
}

# pkg::fun, pkg:::fun
q_colon <- function() {
  '((namespace_operator lhs: (identifier) @pkg-name) @dep-code
    (#not-eq? @pkg-name "base")
  )'
}

q_methods <- function() {
  structure(
    '((call function: (identifier) @fn-name) @dep-code
      (#any-of? @fn-name "setClass" "setGeneric"))',
    names = "methods"
  )
}

q_junit_reporter <- function() {
  structure(
    c(
      '((call function:
      (extract_operator
       lhs: (identifier) @class-name
       rhs: (identifier) @method-name
      )
     ) @dep-code
     (#eq? @class-name "JunitReporter")
     (#eq? @method-name "new"))',
      '((call function:
      (extract_operator
       lhs: (namespace_operator
             lhs: (identifier) @pkg-name
             rhs: (identifier) @class-name)
       rhs: (identifier) @method-name
      )
     ) @dep-code
     (#eq? @pkg-name "testthat")
     (#eq? @class-name "JunitReporter")
     (#eq? @method-name "new"))'
    ),
    names = rep("junit_reporter", 2)
  )
}

q_knitr_dev <- function() {
  structure(
    c(
      '((call function:
      (extract_operator
       lhs: (identifier) @object-name
       rhs: (identifier) @method-name
      )
     ) @dep-code
     (#eq? @object-name "opts_chunk")
     (#eq? @method-name "set"))',
      '((call function:
      (extract_operator
       lhs: (namespace_operator
             lhs: (identifier) @pkg-name
             rhs: (identifier) @object-name)
       rhs: (identifier) @method-name
      )
     ) @dep-code
     (#eq? @pkg-name "knitr")
     (#eq? @object-name "opts_chunk")
     (#eq? @method-name "set"))'
    ),
    names = rep("knitr_dev", 2)
  )
}

renv_dependencies_database <- function() {
  db <- getOption("renv.dependencies.database", default = list())
  db$ggplot2$geom_hex <- "hexbin"
  db$testthat$JunitReporter <- "xml2"
  db
}

q_database <- function() {
  db <- renv_dependencies_database()
  fns <- unlist(lapply(db, names))
  if (length(fns) == 0) {
    return(NULL)
  }
  structure(
    sprintf(
      '((identifier) @id
      (#any-of? @id %s))',
      paste0('"', fns, '"', collapse = " ")
    ),
    names = "database"
  )
}

q_deps <- function() {
  c(
    q_library_0(),
    q_colon(),
    q_methods(),
    q_junit_reporter(),
    q_knitr_dev(),
    q_database(),
    NULL
  )
}

q_deps_rmd <- function() {
  c(
    block = '(fenced_code_block
      (fenced_code_block_delimiter)
      (info_string (language) @language) @header
      (code_fence_content) @content
      (#any-of? @language "r" "R" "rscript" "Rscript")
      (#match? @header "^[{]")
    )',
    inline = '(inline) @inline',
    header = '(minus_metadata) @metadata'
  )
}

q_deps_rmd_inline <- function() {
  '(code_span
    (code_span_delimiter) @csd1
    (code_span_delimiter) @csd2
  ) @code'
}

q_deps_yaml_header <- function() {
  c(
    shiny = '(document (block_node (block_mapping (block_mapping_pair
         key: _ @key
         value: (_ [
           (plain_scalar)
           (block_scalar)
           (single_quote_scalar)
           (double_quote_scalar)
         ]) @value
         ) @code ))
       (#any-of? @key "server" "\'server\'" "\\"server\\""))',
    shiny = '(document (block_node (block_mapping (block_mapping_pair
         key: _ @key
         value: (block_node (block_mapping (block_mapping_pair
           key: _ @key2
           value: (_ [
             (plain_scalar)
             (block_scalar)
             (single_quote_scalar)
             (double_quote_scalar)
           ]) @value
         )))
         ) @code ))
         (#any-of? @key "server" "\'server\'" "\\"server\\"")
         (#any-of? @key2 "type" "\'type\'" "\\"type\\""))',
    pkgstring = '((string_scalar) @code
       (#match? @code "^[a-zA-Z][.a-zA-Z0-9]+[ ]*::"))',
    bslib = '(document (block_node (block_mapping (block_mapping_pair
        key: _ @key
        value: (block_node (block_mapping (block_mapping_pair
         key: _ @key2
         value: (block_node (block_mapping (block_mapping_pair
          key: _ @key3
          value: _ @value
         )))
        )))
        ) @code ))
        (#any-of? @key "output" "\'output\'" "\\"output\\"")
        (#any-of? @key2 "html_document" "\'html_document\'" "\\"html_document\\"")
        (#any-of? @key3 "theme" "\'theme\'" "\\"theme\\""))',
    tag = '((tag) @tag . _ @code
          (#any-of? @tag "!r" "\'!r\'" "\\"!r\\""))',
    NULL
  )
}

test_that("q_library_0", {
  local_reproducible_output(width = 500)
  do <- function(code) {
    code_query(code, q_library_0())[["matched_captures"]]
  }
  expect_snapshot({
    do("library(pkg, lib.loc = path)")
    do("library('pkg', lib.loc = path)")
    do("library(lib.loc = path, pkg)")
    do("require(lib.loc = path, character.only = TRUE, 'pkg')")
    do("library(foo, require(bar))")
  })
})

test_that("q_module_import", {
  local_reproducible_output(width = 500)
  do <- function(path) {
    apath <- test_path(path)
    code_query(readLines(apath), q_module_import())[["matched_captures"]]
  }
  expect_snapshot({
    do("fixtures/scan/modules.R")
    do("fixtures/scan/modules-empty.R")
  })
})

test_that("q_colon", {
  local_reproducible_output(width = 500)
  do <- function(code) {
    code_query(code, q_colon())[["matched_captures"]]
  }
  expect_snapshot({
    do("x <- foo::bar()")
    do("1 + 2 + foo:::bar")
  })
})

test_that("q_methods", {
  local_reproducible_output(width = 500)
  do <- function(code) {
    code_query(code, q_methods())[["matched_captures"]]
  }
  expect_snapshot({
    do("setClass('myclass')")
    do("setGeneric('props', function(object) attributes(object))")
  })
})

test_that("q_junit_reporter", {
  local_reproducible_output(width = 500)
  do <- function(code) {
    code_query(code, q_junit_reporter())[["matched_captures"]]
  }
  expect_snapshot({
    do("JunitReporter$new()")
    do("testthat::JunitReporter$new()")
  })
})

test_that("q_knitr_dev", {
  local_reproducible_output(width = 500)
  do <- function(code) {
    code_query(code, q_knitr_dev())[["matched_captures"]]
  }
  expect_snapshot({
    do("opts_chunk$set()")
    do("knitr::opts_chunk$set()")
  })
})

test_that("renv_dependencies_database", {
  expect_snapshot({
    renv_dependencies_database()
  })
})

test_that("q_database", {
  local_reproducible_output(width = 500)
  do <- function(code) {
    code_query(code, q_database())[["matched_captures"]]
  }
  expect_snapshot({
    do("geom_hex()")
    do("ggplot2::geom_hex()")
    do("JunitReporter")
    do("testthat::JunitReporter")
  })
})

test_that("q_database #2", {
  fake(q_database, "renv_dependencies_database", NULL)
  expect_null(q_database())
})

test_that("q_database #3", {
  local_reproducible_output(width = 500)
  withr::local_options(
    renv.dependencies.database = list(foopkg = list(foofun = "foodep"))
  )
  do <- function(code) {
    code_query(code, q_database())[["matched_captures"]]
  }
  expect_snapshot({
    do("geom_hex()")
    do("foopkg::foofun()")
    do("foofun")
  })
})

test_that("q_deps", {
  fake(q_deps, "q_library_0", 1)
  fake(q_deps, "q_colon", 2)
  fake(q_deps, "q_methods", 3)
  fake(q_deps, "q_junit_reporter", 4)
  fake(q_deps, "q_knitr_dev", 5)
  fake(q_deps, "q_database", 6)
  expect_snapshot(
    q_deps()
  )
})

test_that("q_deps_rmd", {
  local_reproducible_output(width = 500)

  expect_snapshot({
    code_query(
      readLines(test_path("fixtures/scan/chunk-errors.Rmd")),
      query = q_deps_rmd(),
      language = "markdown"
    )[["matched_captures"]]
  })

  expect_snapshot({
    code_query(
      readLines(test_path("fixtures/scan/inline-chunks.Rmd")),
      query = q_deps_rmd(),
      language = "markdown"
    )[["matched_captures"]]
  })
})

test_that("q_deps_rmd_inline", {
  local_reproducible_output(width = 500)

  code <- code_query(
    readLines(test_path("fixtures/scan/inline-chunks.Rmd")),
    query = q_deps_rmd(),
    language = "markdown"
  )[["matched_captures"]]

  expect_snapshot({
    code_query(
      readLines(test_path("fixtures/scan/inline-chunks.Rmd")),
      query = q_deps_rmd_inline(),
      language = "markdown-inline",
      ranges = code[, range_cols]
    )[["matched_captures"]]
  })
})

test_that("q_deps_yaml_header", {
  local_reproducible_output(width = 500)
  expect_snapshot({
    print(
      n = Inf,
      code_query(
        readLines(test_path("fixtures/scan/header-shiny.Rmd")),
        query = q_deps_yaml_header(),
        language = "yaml"
      )[["matched_captures"]]
    )
    print(
      n = Inf,
      code_query(
        readLines(test_path("fixtures/scan/header-shiny2.Rmd")),
        query = q_deps_yaml_header(),
        language = "yaml"
      )[["matched_captures"]]
    )
  })
})

#' Package configuration (internals)
#'
#' @description
#' Internal machinery for pkgdepends configuration. See [current_config()]
#' for how this machinery can be used to define the configuration of a
#' package.
#'
#' This is a standalone file that can be copied into another package without
#' changes.
#'
#' @format
#' `config` is a list of functions with a closure. You can use it two ways.
#'
#' For _one_ set of configuration values in a package, include this in
#' `.onLoad()`:
#'
#' ```r
#' conf <- config$new("prefix")
#' conf$add("myenrry", "string", "default-value")
#' ...
#' conf$lock()
#' ```
#'
#' For _multiple _ sets of configuration values (e.g. one per object),
#' include a function like this in the package:
#'
#' ```r
#' current_config <- function() {
#'   conf <- config$new("prefix")
#'   conf$add("myentry", "string", "default-value")
#'   ...
#'   conf$lock()
#'   conf
#' }
#' ```
#'
#' and then call `current_config()` every time you need to create a
#' new set of configuration values.
#'
#' @keywords internal
#'
#' @details
#' # Terminology
#'
#' ## Config entry (or entry)
#'
#' A single configurable entity. A config entry has a name. This name is
#' always standardized by replacing dashes with underscores.
#'
#' A config entry has a type: a string, a flag, or some
#' user defined type. It can also have a default value. Examples:
#'
#' * The URL of an API, which can be a string.
#' * Path to a directory to use for caching, either a string, or NULL
#'   (`string_or_null` type).
#' * Number of processor cores to use, a positive integer (`count` type).
#'
#' ## Configuration (or config)
#'
#' A set of configuration entries. `config$new()` returns a configuration.
#' A configuration can be extended by adding new entries to it, until it
#' is locked.
#'
#' ## R option (or option)
#'
#' An option set with [base::options()] and queried with [base::getOption()].
#'
#' ## Configuration prefix
#'
#' A prefix that is typically package specific, and it is used for all
#' config entries of a package, to avoid environment variable and option
#' name clashes between packages.
#'
#' ## Config entry type
#'
#' It defines the allowed values of configuration entries, and also how
#' a string from an environment variable is translated to the value of
#' the entry. See the list of builtin types below. You can also define new
#' types.
#'
#' # Built-in types
#'
#' ## `string`
#'
#' String scalar, `NA` is not allowed.
#'
#' ## `count`
#'
#' Non-negative integer, `NA` is not allowed.
#'
#' ## `flag`
#'
#' Logical scalar, `NA` is not allowed. For environment variables the
#' `TRUE` values are: `yes`, `true`, `1` and `on`, and the `FALSE` values
#' are: `no`, `false`, `0`, `off`. (All are case insensitive.)
#'
#' ## `string_or_null`
#'
#' String or `NULL`. `NA` is not allowed. In environment variables the
#' string `NULL` means an R `NULL` value.
#'
#' ## `character_or_null`
#'
#' Character vector or `NULL`. `NA` is not allowed. In environment variables
#' the entries are separated by a semicolon.
#'
#' ## `character`
#'
#' Character vector without `NA` values. In environment variables the
#' entries are separated by a semicolon.
#'
#' ## `custom`
#'
#' Custom type. An `env_decode` function must be specified for each config
#' entry of this type, otherwise an error is throw if the corresponding
#' environment variable is set.

config <- local({
  `%||%` <- function(l, r) if (is.null(l)) r else l

  # It is used as the check function if no check function is needed
  true <- function(...) TRUE

  is_config_name <- function(name) {
    if (is_string(name)) {
      return(TRUE)
    }
    structure(
      FALSE,
      msg = c(
        "A config entry name ({.arg name} argument) must be a string
         (character scalar).",
        i = "It is {.type {name}} instead."
      )
    )
  }

  is_config_check <- function(check) {
    if (is_string(check) || is.function(check) || is.null(check)) {
      return(TRUE)
    }
    structure(
      FALSE,
      msg = c(
        "A config check function ({.arg check} argument) must be of",
        "*" = "a predicate function that returns a logical scalar, or",
        "*" = "a config type name, a character scalar, or",
        "*" = "{.code NULL} for no value checks.",
        "i" = "{.arg check} is {.type {check}}."
      )
    )
  }

  is_config_env_decoder <- function(env_decode) {
    is_string(env_decode) || is.function(env_decode) || is.null(env_decode)
  }

  standard_name <- function(x) {
    gsub("-", "_", x)
  }

  # Built-in types
  builtin_types <- c(
    "character",
    "character_or_null",
    "count",
    "custom",
    "flag",
    "string",
    "string_or_null"
  )

  # Checks for builtin types
  builtin_checks <- list(
    character = function(x) is.character(x) && !anyNA(x),
    character_or_null = function(x) {
      (is.character(x) && !anyNA(x)) || is.null(x)
    },
    custom = function(x) true(x),
    flag = function(x) is_flag(x),
    string = function(x) is_string(x),
    string_or_null = function(x) is_optional_string(x),
    count = function(x) is_count(x)
  )

  # Builtin environment variable decoders
  builtin_env_decode <- list(
    character = function(x, ...) {
      strsplit(x, ";", fixed = TRUE)[[1]]
    },
    custom = function(x, name, env, ...) {
      throw(pkg_error(
        "Cannot decode config value from environment variable {.envvar {name}}.",
        i = "This is an probably an internal error in the {.pkg {env$package}} package."
      ))
    },
    flag = function(x, name, ...) {
      x <- tolower(x)
      if (tolower(x) %in% c("yes", "true", "1", "on")) {
        return(TRUE)
      }
      if (tolower(x) %in% c("no", "false", "0", "off")) {
        return(FALSE)
      }
      throw(pkg_error(
        "Invalid value for the {.envvar {name}} environment variable.",
        i = "It must be either {.code true} or {.code false}."
      ))
    },
    string = function(x, ...) x,
    string_or_null = function(x, ...) if (identical(x, "NULL")) NULL else x,
    character_or_null = function(x, ...) {
      if (x %in% c("", "NULL")) NULL else strsplit(x, ";", fixed = TRUE)[[1]]
    },
    count = function(x, name, ...) {
      num <- suppressWarnings(as.numeric(num))
      if (is.na(num) || !is_count(num)) {
        throw(pkg_error(
          "Cannot interpret environment variable {.envvar {name}} as a
           count: {.code {x}}."
        ))
      }
      as.integer(num)
    }
  )

  get_internal <- function(env, name) {
    assert_that(is_config_name(name))
    name <- standard_name(name)
    if (!name %in% names(env$data)) {
      throw(pkg_error(
        "Unknown configuration entry: {.code {name}}.",
        i = "This is an probably an internal error in the
             {.pkg {env$package}} package."
      ))
    }

    rec <- env$data[[name]]

    # was explicitly set?
    if (!is.null(rec$value)) {
      return(list("set", rec$value))
    }

    # set via options()
    optname <- paste0(env$prefix, name)
    opt <- getOption(optname)
    if (!is.null(opt)) {
      if (!is.null(chk <- env$data[[name]]$check)) {
        chk(opt)
      }
      return(list("option", opt))
    }

    # set via env var
    envvname <- toupper(chartr(".", "_", paste0(env$prefix, name)))
    envv <- Sys.getenv(envvname, NA_character_)
    if (!is.na(envv)) {
      return(list("envvar", rec$env_decode(envv, envvname, env)))
    }

    # otherwise the default, but if it is a function, then call it
    def <- rec$default
    if (is.function(def)) {
      if (length(formals(def)) > 0) {
        def <- def(env)
      } else {
        def <- def()
      }
      if (!is.null(chk <- env$data[[name]]$check)) chk(def)
    }
    list("default", def)
  }

  #' # `config$onload_hook()`: pretty-printing configuration
  #'
  #' Call `config$onload_hook()` from the package's `.onLoad()` function
  #' if you want to define S3 methods to print the package configuration
  #' nicer.
  #'
  #' ## Usage
  #' ```
  #' config$onload_hook()
  #' ```

  onload_hook <- function() {
    register_if_needed("print", "config_v1", print_config)
    register_if_needed("print", "config", print_config)
  }

  #' # `config$new()`: Create a new configuration
  #'
  #' Typically `new()` is called outside of the functions of the
  #' package, so the `config` object is created at install time.
  #'
  #' ## Usage
  #' ```
  #' conf <- config$new(prefix = utils::packageName(parent.frame()))
  #' ```
  #'
  #' ## Arguments
  #' * `prefix`: prefix of the config entry names. For environment
  #'   variables it is converted to uppercase, and dots are converted
  #'   to underscores. The default prefix is the name of the calling package.
  #'   An underscore separator is used between the prefix and the entry
  #'   name for environment variable names. A dot separator is used for
  #'   R option names.
  #'
  #' ## Value
  #' `new()` returns a `config` object, which is a list containing
  #' the configuration data and functions (methods) to query and change it.
  #'
  #' ```r
  #' conf <- config$new()
  #' ```
  #'
  #' # Configuration methods

  new <- function(
    prefix = utils::packageName(parent.frame()),
    package = utils::packageName(parent.frame())
  ) {
    env <- new.env(parent = emptyenv())
    env$prefix <- if (!is.null(prefix)) paste0(prefix, ".")
    env$data <- new.env(parent = emptyenv())
    env$package <- package

    # These can be modified by the user of the config class, as needed,
    # to add user-defined types
    env$types <- builtin_types
    env$checks <- builtin_checks
    env$env_decode = builtin_env_decode

    # --------------------------------------------------------------------
    #' ## `conf$add()`: add a new configuration entry
    #'
    #' ### Usage
    #'
    #' ```r
    #' conf$add(
    #'   name,
    #'   type = conf$types,
    #'   default = NULL,
    #'   check = type[1],
    #'   env_decode = type[1]
    #' )
    #' ```
    #'
    #' ### Arguments
    #'
    #' * `name`: name of the entry.
    #' * `type`: type of the entry, string.
    #' * `default`: default value of config entry. If this is a function,
    #'    then it is called to generate the default, at the time the
    #'    config entry's value is queried.
    #' * `check`: if it is a function, it is called to check the value of
    #'    the entry. The function should return `TRUE` for successful and
    #'    `FALSE` for unsuccessful checks. If `NULL`, then no check is
    #'    performed. It can also be the name of a type, then the default
    #'    check for that type is performed, if any.
    #' * `env_decode`: if it is a function, then it is used to decode the
    #'    value of the entry from an environment variable, i.e. a string.
    #'    It can also be the name of a type, then the `env_decode()`
    #'    function of that type is used for decoding.
    #'
    #' ### Value
    #'
    #' The configuration, invisibly.

    env$add <- function(
      name,
      type = env$types,
      default = NULL,
      check = type[1],
      env_decode = type[1]
    ) {
      # Need to explicitly add `env$types` on R 3.4.x
      type <- match.arg(type, env$types)
      assert_that(
        is_config_name(name),
        is_config_check(check),
        is_config_env_decoder(env_decode)
      )
      name <- standard_name(name)

      if (name %in% names(env$data)) {
        throw(pkg_error(
          "There is already a config entry called {.code {name}}."
        ))
      }

      if (is_string(check)) {
        check <- env$checks[[check]]
      }
      check <- check %||% true
      if (!is.null(default) && !is.function(default)) {
        check(default)
      }

      if (is_string(env_decode)) {
        env_decode <- env$env_decode[[env_decode]]
      }
      env_decode <- env_decode %||% identity

      env$data[[name]] <- list(
        type = type,
        check = check,
        default = default,
        env_decode = env_decode,
        value = NULL
      )
      invisible(env)
    }

    # --------------------------------------------------------------------
    #' ## `conf$get()`: query the value of a configuration entry
    #'
    #' ### Usage
    #'
    #' ```r
    #' conf$get(name)
    #' ```
    #'
    #' ### Arguments
    #'
    #' * `name`: name of the entry.
    #'
    #' ### Value
    #'
    #' Value of the entry.
    #'
    #' * If the entry was set via `conf$set()` or `conf$update()`, then
    #'   that value is returned.
    #' * Otherwise, if the entry is set via an R option (see
    #'   [base::getOption()]), then that value is returned. The config
    #'   prefix is used to get the option name, with a dot separator.
    #'   I.e. for an entry called 'foo', and prefix 'pkg', the `pkg.foo`
    #'   option is used. If the entry has a check function, that is called
    #'   before returning.
    #' * Otherwise, if the entry is set via an environment variable, then
    #'   that value is returned. The config prefix is used to get the name
    #'   of the environment variable. In addition, dots are replaced with
    #'   underscores and the name is converted to uppercase. I.e. for an
    #'   entry called 'foo.bar', and prefix 'pkg', the 'PKG_FOO_BAR`
    #'   environment variable is used.
    #' * Otherwise, if the entry has a default value, that is returned.
    #'   The default value might be a function, in which case that function
    #'   is called to produce a default value. If the entry has a check
    #'   function, that is called before returning.

    env$get <- function(name) {
      get_internal(env, name)[[2]]
    }

    # --------------------------------------------------------------------
    #' ## `conf$set()`: set the value of a configuration entry
    #'
    #' If you `$set()` a config entry, then the `value` used in `$set()`
    #' will be returned by `$get()`, without consulting R options or
    #' environment variables.
    #'
    #' ### Usage
    #'
    #' ```r
    #' conf$set(name, value)
    #' ```
    #'
    #' ### Arguments
    #'
    #' * `name`: name of the entry.
    #' * `value`: value of the entry. If the entry has a check function,
    #'   it is called here.
    #'
    #' ### Value
    #'
    #' The configuration, invisibly.

    env$set <- function(name, value) {
      assert_that(is_config_name(name))
      name <- standard_name(name)
      if (!name %in% names(env$data)) {
        throw(pkg_error(
          "Cannot set unknown config entry: {.code {name}}.",
          i = "See `$list()` for the list of all config entries."
        ))
      }
      if (!is.null(chk <- env$data[[name]]$check)) {
        chk(value)
      }
      env$data[[name]]$value <- value
      invisible(env)
    }

    # --------------------------------------------------------------------
    #' ## `$conf$unset()`: unset a configuration entry
    #'
    #' Note that this function does _not_ unset R options and environment
    #' variables. It merely removes a value that was assigned by `$set()`
    #' or `$update()`.
    #'
    #' ### Usage
    #'
    #' ```r
    #' conf$unset(name)
    #' ```
    #'
    #' ### Arguments
    #'
    #' * `name`: name of the entry.
    #'
    #' ### Value
    #'
    #' The configuration, invisibly.

    env$unset <- function(name) {
      assert_that(is_config_name(name))
      name <- standard_name(name)
      if (!name %in% names(env$data)) {
        throw(pkg_error(
          "Cannot unset unknown config entry: {.code {name}}.",
          i = "See `$list()` for the list of all config entries."
        ))
      }
      env$data[[name]]$value <- NULL
      invisible(env)
    }

    # --------------------------------------------------------------------
    #' ## `cond$update()`: update the values of configuration entries
    #'
    #' You can use this method to set()` and/or `$unset()` multiple config
    #' entries. `NULL` values in `new` will unset the corresponding entry.
    #'
    #' ### Usage
    #'
    #' ```r
    #' conf$update(new)
    #' ```
    #'
    #' ### Arguments
    #'
    #' * `new`: named list, where names are config entry names and values
    #'   are the corresponding config entry values. `NULL` values will
    #'   unset the config entry. The list is processed sequentially, so
    #'   for duplicates the latest values will be in effect. (But every
    #'   value is checked with its check function, if any.)
    #'
    #' ### Value
    #'
    #' The configuration, invisibly.

    env$update <- function(new) {
      for (i in seq_along(new)) {
        if (is.null(new[[i]])) {
          env$unset(names(new)[i])
        } else {
          env$set(names(new)[i], new[[i]])
        }
      }
      invisible(env)
    }

    # --------------------------------------------------------------------
    #' ## `conf$list()`: query the names of all config entries
    #'
    #' Note that their order is non-deterministic currently.
    #'
    #' ### Usage
    #'
    #' ```r
    #' conf$list()
    #' ```
    #'
    #' ### Value
    #'
    #' Character vector.

    env$list <- function() {
      names(env$data)
    }

    # --------------------------------------------------------------------
    #' ## `conf$exists()`: check if a config entry exists
    #'
    #' ### Usage
    #'
    #' ```r
    #' conf$exists(name)
    #' ```
    #'
    #' ### Arguments
    #'
    #' * `name`: name of the entry.
    #'
    #' ### Value
    #'
    #' Logical scalar, `TRUE` if the entry exists in the configuration,
    #' `FALSE` otherwise. Note that `TRUE` does not mean that the value
    #' of the entry was set.

    env$exists <- function(name) {
      assert_that(is_config_name(name))
      name <- standard_name(name)
      name %in% env$list()
    }

    # --------------------------------------------------------------------
    #' ## `conf$lock()`: lock the configuration
    #'
    #' If a configuration is locked, then no more entries can be added to
    #' it.
    #'
    #' ### Usage
    #'
    #' ```r
    #' conf$lock()
    #' ```
    #'
    #' ### Value
    #'
    #' The configuration, invisibly.

    env$lock <- function() {
      lockEnvironment(env$data)
      invisible(env)
    }

    # --------------------------------------------------------------------
    #' ## `conf$fix()`: fix the value of a single configuration entry
    #'
    #' The currently `$set()` value is fixed and cannot be changed any
    #' more. You can fix an entry without setting its value, in which case
    #' R options and environment variables can still be used to update it.
    #'
    #' ### Usage
    #'
    #' ```r
    #' conf$unset(name)
    #' ```
    #'
    #' ### Arguments
    #'
    #' * `name`: name of the entry.
    #'
    #' ### Value
    #'
    #' The configuration, invisibly.

    env$fix <- function(name) {
      assert_that(is_config_name(name))
      name <- standard_name(name)
      if (!name %in% names(env$data)) {
        throw(pkg_error(
          "Cannot fix unknown config entry: {.code {name}}.",
          i = "See `$list()` for the list of all config entries."
        ))
      }
      lockBinding(name, env$data)
      invisible(env)
    }

    # --------------------------------------------------------------------
    #' ## `conf$add_type()`: add a new config entry type to a configuration
    #'
    #' After this the new type can be used in `conf$add()` calls.
    #'
    #' ### Usage
    #'
    #' ```r
    #' conf$add_type(type_name, check, env_decode)
    #' ```
    #'
    #' ### Arguments
    #'
    #' * `type_name`: type name.
    #' * `check`: default check function of the type. Use `NULL` for no
    #'   checks.
    #' * `env_decode`: function to use to decode a value of the type from
    #'   an environment variable (i.e. string).
    #'
    #' ### Value
    #'
    #' The configuration, invisibly.

    env$add_type <- function(type_name, check, env_decode) {
      assert_that(
        is_string(type_name),
        is.function(check),
        is.function(env_decode)
      )
      if (type_name %in% env$types) {
        throw(pkg_error(
          "There is already a config entry type called {.code {type_name}}."
        ))
      }
      env$types <- c(env$types, type_name)
      env$checks[[type_name]] <- check
      env$env_decode[[type_name]] <- env_decode
      invisible(env)
    }

    structure(env, class = c("config_v1", "config"))
  }

  print_config <- function(x, ...) {
    nms <- names(x$data)
    all <- structure(
      lapply(nms, function(n) get_internal(x, n)),
      names = nms
    )
    cat0("# ", substr(x$prefix, 1, nchar(x$prefix) - 1), " config\n")
    for (i in seq_along(all)) {
      cat0("## ", names(all)[i], "\n")
      cat0("<", all[[i]][[1]], ">\n")
      print(all[[i]][[2]])
      cat0("\n")
    }
    invisible(x)
  }

  register_if_needed <- function(generic, class, fun) {
    if (is.null(utils::getS3method(generic, class, TRUE))) {
      registerS3method(generic, class, fun, baseenv())
    }
  }

  structure(
    list(
      .internal = environment(),
      new = new,
      onload_hook = onload_hook
    ),
    class = c("standalone_config", "standalone")
  )
})

test_that("uncompress error", {
  skip_on_cran()
  local_cli_config()

  tmp <- withr::local_tempdir()
  pkgzip <- file.path(tmp, "foo-t.zip")
  zip::zipr(pkgzip, test_path("fixtures", "foo"))
  bts <- readBin(pkgzip, "raw", file.size(pkgzip))
  bts <- head(bts, as.integer(length(bts) / 2))
  writeBin(bts, pkgzip)

  plan <- data_frame(
    type = "local",
    binary = FALSE,
    packaged = FALSE,
    dependencies = list(character()),
    needscompilation = TRUE,
    file = pkgzip,
    package = "foo"
  )

  expect_snapshot(
    error = TRUE,
    transform = function(x) sub("'.*/foo-t.zip", ".../foo-t.zip", x),
    install_package_plan(plan, lib = tempdir())
  )
})

test_that("build error", {
  withr::local_envvar(PKG_OMIT_TIMES = "true")
  pkg <- source_test_package("foo2")

  quo <- substitute(
    {
      plan <- data.frame(
        stringsAsFactors = FALSE,
        type = "local",
        binary = FALSE,
        dependencies = I(list(character())),
        file = pkg,
        needscompilation = TRUE,
        package = "foo"
      )

      lib <- withr::local_tempdir()
      pkgdepends:::install_package_plan(plan, lib = lib)
    },
    list(pkg = pkg)
  )

  tmp <- tempfile(fileext = ".R")
  on.exit(unlink(tmp), add = TRUE)
  writeLines(deparse(quo), tmp)

  out <- tempfile()
  on.exit(unlink(out), add = TRUE)
  res <- callr::rscript(
    tmp,
    stdout = out,
    stderr = out,
    fail_on_status = FALSE,
    show = FALSE
  )

  # We can't match the whole output or even the error message, because
  # different compilers have different output. Nevertheless this should
  # appear in the output.
  expect_match(res$stdout, "(return R_NilValue|error: expected)")
})

test_that("packaging error", {
  withr::local_envvar(PKG_OMIT_TIMES = "true")
  tmp <- withr::local_tempdir()
  pkgzip <- file.path(tmp, "foo3-t.zip")
  zip::zipr(pkgzip, test_path("fixtures", "foo3"))

  quo <- substitute(
    {
      plan <- data.frame(
        stringsAsFactors = FALSE,
        type = "local",
        binary = FALSE,
        dependencies = I(list(character())),
        file = pkgzip,
        needscompilation = TRUE,
        package = "foo3",
        packaged = FALSE
      )

      lib <- withr::local_tempdir()
      pkgdepends:::install_package_plan(plan, lib = lib)
    },
    list(pkgzip = pkgzip)
  )

  tmp <- tempfile(fileext = ".R")
  on.exit(unlink(tmp), add = TRUE)
  writeLines(deparse(quo), tmp)

  out <- tempfile()
  on.exit(unlink(out), add = TRUE)
  res <- callr::rscript(
    tmp,
    stdout = out,
    stderr = out,
    fail_on_status = FALSE,
    show = FALSE
  )

  expect_true(
    grepl("not in valid DCF format", res$stdout, fixed = TRUE) ||
      grepl("Error in read.dcf", res$stdout, fixed = TRUE)
  )
})

test_that("warn_for_long_paths", {
  # TODO: this is challenging to test, because it is hard to create these
  # long paths on Windows.
})

test_that("parse_deps", {
  expect_equal(
    parse_deps(character(), character()),
    list()
  )

  expect_equal(
    parse_deps("", "Imports"),
    list(data_frame(
      type = character(),
      package = character(),
      op = character(),
      version = character()
    ))
  )

  expect_equal(
    parse_deps("foobar", "Imports"),
    list(data_frame(
      type = "Imports",
      package = "foobar",
      op = "",
      version = ""
    ))
  )

  expect_equal(
    parse_deps("foobar (>= 1.0-5)", "Imports"),
    list(data_frame(
      type = "Imports",
      package = "foobar",
      op = ">=",
      version = "1.0-5"
    ))
  )

  expect_equal(
    parse_deps("foobar\n (>=\n 1.0-5), foobar2", "Imports"),
    list(data_frame(
      type = rep("Imports", 2),
      package = c("foobar", "foobar2"),
      op = c(">=", ""),
      version = c("1.0-5", "")
    ))
  )
})

zwnj <- function() {
  if (cli::is_utf8_output()) "\u200c" else ""
}

pkgd_data <- new.env(parent = emptyenv())

`%||%` <- function(l, r) if (is.null(l)) r else l

`%|0|%` <- function(l, r) if (length(l) == 0) r else l

`%|z|%` <- function(l, r) if (is.null(l) || identical(l, "")) r else l

`%&z&%` <- function(l, r) if (length(l) > 0 && l != "") r else ""

get_private <- function(x) x$.__enclos_env__$private

default_cran_mirror <- function() {
  mirror <- getOption("repos")["CRAN"]
  if (is.null(mirror) || is.na(mirror) || mirror == "@CRAN@") {
    "https://cran.rstudio.com"
  } else {
    mirror
  }
}

current_r_version <- function() {
  as.character(getRversion())
}

get_minor_r_version <- function(x) {
  x <- package_version(x)
  vapply(unclass(x), function(x) paste(x[1:2], collapse = "."), character(1))
}

str_trim <- function(x) {
  sub("\\s+$", "", sub("^\\s+", "", x))
}

## TODO: in theory the set of base packages can change over time,
## so we would need an R version specific vector here.
## Not an issue currently, might be in the future.

## CRAN does not want me to call installed.packages at all, so let's
## hardcode this for now.

base_packages <- function() {
  if (is.null(pkgd_data$base_packages)) {
    pkgd_data$base_packages <- c(
      "base",
      "compiler",
      "datasets",
      "graphics",
      "grDevices",
      "grid",
      "methods",
      "parallel",
      "splines",
      "stats",
      "stats4",
      "tcltk",
      "tools",
      "utils"
    )
  }
  pkgd_data$base_packages
}

recommended_packages <- function() {
  if (is.null(pkgd_data$recommended_packages)) {
    pkgd_data$recommended_packages <- c(
      "boot",
      "class",
      "cluster",
      "codetools",
      "foreign",
      "KernSmooth",
      "lattice",
      "MASS",
      "Matrix",
      "mgcv",
      "nlme",
      "nnet",
      "rpart",
      "spatial",
      "survival"
    )
  }
  pkgd_data$recommended_packages
}

lapply_with_names <- function(X, FUN, ...) {
  structure(
    lapply(X, FUN, ...),
    names = names(X) %||% (if (is.character(X)) X)
  )
}

vcapply <- function(X, FUN, ...) {
  vapply(X, FUN, FUN.VALUE = character(1), ...)
}

vlapply <- function(X, FUN, ...) {
  vapply(X, FUN, FUN.VALUE = logical(1), ...)
}

viapply <- function(X, FUN, ...) {
  vapply(X, FUN, FUN.VALUE = integer(1), ...)
}

vdapply <- function(X, FUN, ...) {
  vapply(X, FUN, FUN.VALUE = double(1), ...)
}

update_named_vector <- function(old, new) {
  assert_that(all_named(old), all_named(new))
  comm <- intersect(names(old), names(new))
  add <- setdiff(names(new), names(old))
  old[comm] <- new[comm]
  old <- c(old, new[add])
  old
}

make_dl_status <- function(status, url, target, bytes, error = NULL) {
  obj <- list(
    status = status,
    url = url,
    target = target,
    bytes = NA_real_,
    error = NULL
  )

  if (status == "Got") {
    obj$bytes <- as.double(bytes)
  } else if (status == "Failed") {
    obj$error <- error
  } else if (grepl("^Had", status)) {
    obj$bytes <- as.double(bytes)
  }

  obj
}

comma_wrap <- function(x, indent = 2, exdent = indent, sep = ", ") {
  w <- strwrap(paste(x, collapse = sep), indent = indent, exdent = exdent)
  paste(w, collapse = "\n")
}

add_class <- function(obj, classes, where = c("start", "end")) {
  where <- match.arg(where)
  nc <- c(
    if (where == "start") classes,
    class(obj),
    if (where == "end") classes
  )
  class(obj) <- unique(nc)
  obj
}

is_na_scalar <- function(x) {
  length(x) == 1 && is.na(x)
}

omit_cols <- function(df, omit) {
  if (!length(omit)) {
    df
  } else {
    df[, setdiff(names(df), omit), drop = FALSE]
  }
}

same_sha <- function(s1, s2) {
  assert_that(
    is.character(s1),
    length(s1) == 1,
    is.character(s2),
    length(s2) == 1
  )
  if (is.na(s1) || is.na(s2)) {
    return(FALSE)
  }
  assert_that(
    is_string(s1),
    is_string(s2)
  )

  len <- min(nchar(s1), nchar(s2))
  substr(s1, 1, len) == substr(s2, 1, len)
}

format_iso_8601 <- function(date) {
  format(as.POSIXlt(date, tz = "UTC"), "%Y-%m-%dT%H:%M:%S+00:00")
}

cat0 <- function(..., sep = "") {
  cat(..., sep = sep)
}

lapply_rows <- function(df, fun, ...) {
  lapply(seq_len(nrow(df)), function(i) fun(df[i, ], ...))
}

detect_download_cache_dir <- local({
  dir <- NULL
  function() {
    if (is.null(dir)) {
      dir <<- tempfile()
    }
    dir
  }
})

rbind_expand <- function(..., .list = list()) {
  data <- c(list(...), .list)
  cols <- unique(unlist(lapply(data, function(x) colnames(x))))
  for (i in seq_along(data)) {
    miss_cols <- setdiff(cols, colnames(data[[i]]))
    if (length(miss_cols)) {
      na_df <- as_data_frame(structure(
        replicate(
          length(miss_cols),
          if (nrow(data[[i]])) NA else logical(),
          simplify = FALSE
        ),
        names = miss_cols
      ))
      data[[i]] <- as_data_frame(cbind(data[[i]], na_df))
    }
  }

  do.call(rbind, data)
}

drop_nulls <- function(x) {
  x[!vlapply(x, is.null)]
}

## R CMD check fixes
self <- private <- "foobar"
dummy <- function() {
  callr::r
}

get_num_workers <- function() {
  n <- tryCatch(
    suppressWarnings(as.integer(getOption("Ncpus", NA_integer_))),
    error = function(e) NA_integer_
  )

  if (length(n) != 1 || is.na(n)) {
    n <- tryCatch(
      ps::ps_cpu_count(logical = TRUE),
      error = function(e) NA_integer_
    )
  }

  if (is.na(n)) {
    n <- 1L
  }

  n
}

is_rcmd_check <- function() {
  if (identical(Sys.getenv("NOT_CRAN"), "true")) {
    FALSE
  } else {
    Sys.getenv("_R_CHECK_PACKAGE_NAME_", "") != ""
  }
}

is_online <- local({
  online <- TRUE
  expires <- Sys.time()
  function() {
    if (is_rcmd_check()) {
      return(FALSE)
    }
    t <- Sys.time()
    if (t >= expires) {
      online <<- pingr::is_online()
      expires <<- t + as.difftime(10, units = "secs")
    }
    online
  }
})

# `reset` is useful for testing
once_per_session <- local({
  seen <- character()
  function(expr, reset = FALSE) {
    if (reset) {
      seen <<- character()
      return(invisible())
    } else {
      h <- cli::hash_obj_md5(substitute(expr))
      if (!h %in% seen) {
        seen <<- c(seen, h)
        expr
      }
    }
  }
})

# nocov start

synchronise <- synchronize <- sy <- function(...) {
  asNamespace("pkgcache")$synchronise(...)
}

async_constant <- function(...) {
  asNamespace("pkgcache")$async_constant(...)
}

http_get <- function(...) {
  asNamespace("pkgcache")$http_get(...)
}

http_post <- function(...) {
  asNamespace("pkgcache")$http_post(...)
}

http_delete <- function(...) {
  asNamespace("pkgcache")$http_delete(...)
}

when_all <- function(...) {
  asNamespace("pkgcache")$when_all(...)
}

async <- function(...) {
  asNamespace("pkgcache")$async(...)
}

async_map <- function(...) {
  asNamespace("pkgcache")$async_map(...)
}

async_timeout <- function(...) {
  asNamespace("pkgcache")$async_timeout(...)
}

download_file <- function(...) {
  asNamespace("pkgcache")$download_file(...)
}

download_one_of <- function(...) {
  asNamespace("pkgcache")$download_one_of(...)
}

http_stop_for_status <- function(...) {
  asNamespace("pkgcache")$http_stop_for_status(...)
}

new_async_timer <- function(...) {
  asNamespace("pkgcache")$async_timer$new(...)
}

async_delay <- function(...) {
  asNamespace("pkgcache")$delay(...)
}

external_process <- function(...) {
  asNamespace("pkgcache")$external_process(...)
}

get_user_cache_dir <- function(...) {
  asNamespace("pkgcache")$get_user_cache_dir(...)
}

# nocov end

format_error_with_stdout <- function(x, ...) {
  msg <- conditionMessage(x)
  if (is.null(x$stdout)) {
    paste0(msg, " (output not available)")
  } else {
    out <- last_stdout_lines(x$stdout, "stdout + stderr", "OE> ")
    c(paste0(msg, out[1]), out[-1])
  }
}

last_stdout_lines <- function(lines, std, prefix = "E> ") {
  if (err$is_interactive()) {
    pref <- paste0(
      ", ",
      std,
      if (length(lines) > 10) " (last 10 lines)",
      ":"
    )
    out <- paste0(prefix, utils::tail(lines, 10))
    c(pref, "", out)
  } else {
    out <- paste0(prefix, lines)
    c(paste0(", ", std, ":"), "", out)
  }
}

# I am a coward to test this :P
# nocov start

rimraf <- function(...) {
  x <- file.path(...)
  if ("~" %in% x) {
    throw(pkg_error(
      "Cowardly refusing to delete {.path ~}.",
      "i" = paste0(
        "You have a file or directory named {.path ~} and ",
        "because of an R bug deleting that could delete your ",
        "whole home directory."
      )
    ))
  }
  unlink(x, recursive = TRUE, force = TRUE)
}

# nocov end

is_windows <- function() {
  identical(tolower(Sys.info()[["sysname"]]), "windows")
}

# used in coverage condition

is_linux <- function() {
  identical(tolower(Sys.info()[["sysname"]]), "linux") # nocov
}

# This is a workaround for some RStudio bugs:
# https://github.com/r-lib/pkginstall/issues/42
# https://github.com/rstudio/rstudio/issues/2387
# https://github.com/rstudio/rstudio/issues/7278

is_older_rstudio <- function() {
  rs <- rstudio$detect()
  rs$type == "rstudio_console" &&
    !is.null(rs$version) &&
    rs$version <= "1.4.800"
}

ansi_align_width <- function(text) {
  if (length(text) == 0) {
    return(text)
  }
  width <- max(cli::ansi_nchar(text, type = "width"))
  cli::ansi_align(text, width = width)
}

get_id <- local({
  id <- 0L
  function() {
    id <<- id + 1L
    id
  }
})

# tools::md5sum has issues with UTF=8 file names on Windows, <= R 4.0
# TODO: switch to cli::hash_md5_file, once released. I tested that it works
# well on Windows non-ascii file names

safe_md5sum <- function(path) {
  assert_that(is_path(path))
  tryCatch(
    tools::md5sum(path),
    error = function(err) {
      tmp <- tempfile()
      on.exit(unlink(tmp, force = TRUE, recursive = TRUE), add = TRUE)
      file.copy(path, tmp)
      structure(tools::md5sum(tmp), names = path)
    }
  )
}

# TODO: rewrite this in C in ps

get_euid <- function() {
  euid <- tryCatch(
    as.integer(processx::run("id", "-u")$stdout),
    error = function(e) NA_integer_
  )
  if (length(euid) != 1 || is.na(euid)) {
    euid <- NA_integer_
  }
  euid
}

zip_list <- function(zipfile) {
  utils::unzip(zipfile, list = TRUE, unzip = "internal")[, 1]
}

os_type <- function() .Platform$OS.type # nocov

is.dir <- function(path) {
  assert_that(is_string(path), file.exists(path))
  file.info(path)$isdir
}

map_named <- function(x, fun) {
  mapply(names(x), x, SIMPLIFY = FALSE, FUN = fun)
}

sort_by_name <- function(x) {
  x[order(names(x))]
}

is_pak <- function() {
  Sys.getenv("THIS_IS_PAK") == "true" ||
    Sys.getenv("R_PKG_PKG_WORKER") == "true"
}

pak_or_pkgdepends <- function() {
  if (is_pak()) "pak" else "pkgdepends"
}

pakx_version <- function() {
  if (is_pak()) {
    utils::packageVersion("pak")
  } else {
    utils::packageVersion("pkgdepends")
  }
}

remove_entry <- function(l, n) {
  l[names(l) != n]
}

path_norm <- function(x) {
  normalizePath(x, mustWork = FALSE)
}

parse_platform <- function(x) {
  pcs <- strsplit(x, "-", fixed = TRUE)
  data_frame(
    cpu = vcapply(pcs, "[", 1),
    vendor = vcapply(pcs, "[", 2),
    os = vcapply(pcs, function(y) {
      if (length(y) < 3) NA_character_ else paste(y[-(1:2)], collapse = "-")
    })
  )
}

single_quote <- function(x) {
  encodeString(x, quote = "'", na.encode = FALSE)
}

backtick <- function(x) {
  encodeString(x, quote = "`", na.encode = FALSE)
}

collapse <- function(x, ...) {
  cli::ansi_collapse(x, ...)
}

na_omit <- function(x) {
  x[!is.na(x)]
}

file_ext <- function(x) {
  re_match(x, "[.]([[:alnum:]]+)$")[[".match"]]
}

# drop a prefix and a postfix, vectorized
omit_pre_post <- function(x, pre = 0, post = 0) {
  substr(x, 1L + pre, nchar(x) - post)
}

truthy_strings <- c("true", "t", "1", "on", "yes")

is_truthy <- function(x) {
  (is.logical(x) && length(x) >= 1 && !is.na(x[[1]]) && x[[1]]) ||
    (is.character(x) && length(x) >= 1 && tolower(x) %in% truthy_strings)
}

# check that all `path` are inside `root`
check_inside_dir <- function(root, path) {
  # we can assume that `root` and all `path` do exist
  # normalizePath removes trailing '/'
  nroot <- normalizePath(root, winslash = "/")
  npath <- normalizePath(path, winslash = "/")
  if (any(bad <- nroot != npath & !startsWith(paste0(npath, "/"), nroot))) {
    throw(pkg_error(
      "{.path {path[bad]}} {?is/are} outside of project
       root {.path {root}}."
    ))
  }
}

test_that("%|z|%", {
  expect_true("" %|z|% TRUE)

  bad <- list(
    character(),
    c("", ""),
    "foo",
    structure("", class = "foo")
  )
  for (b in bad) {
    expect_identical(b %|z|% FALSE, b)
  }
})

test_that("%&z&%", {
  expect_equal(
    NULL %&z&% "bar",
    ""
  )
  expect_equal(
    character() %&z&% "bar",
    ""
  )
  expect_equal(
    "" %&z&% "bar",
    ""
  )
  expect_equal(
    "foo" %&z&% "bar",
    "bar"
  )
})

test_that("get_private", {
  cls <- R6::R6Class(
    "foo",
    public = list(foo = function() "foo"),
    private = list(bar = function() "bar")
  )
  obj <- cls$new()
  expect_equal(get_private(obj)$bar(), "bar")
})

test_that("default_cran_mirror", {
  m1 <- withr::with_options(
    list(repos = c(CRAN = "@CRAN@")),
    default_cran_mirror()
  )
  m2 <- withr::with_options(
    list(repos = NULL),
    default_cran_mirror()
  )
  m3 <- withr::with_options(
    list(repos = c("foo" = "bar")),
    default_cran_mirror()
  )

  expect_true(is.character(m1) && length(m1) == 1 && !is.na(m1))
  expect_identical(m1, m2)
  expect_identical(m1, m3)

  m4 <- withr::with_options(
    list(repos = c(CRAN = "mymirror")),
    default_cran_mirror()
  )
  expect_identical(m4, c(CRAN = "mymirror"))
})

test_that("current_r_version", {
  ver <- current_r_version()
  expect_true(is.character(ver))
  expect_true(length(ver) == 1)
})

test_that("get_minor_r_version", {
  expect_equal(get_minor_r_version("4.2.2"), "4.2")
  expect_equal(get_minor_r_version("4.2.2.0"), "4.2")
})

test_that("recommended_packages", {
  expect_snapshot(
    recommended_packages()
  )
})

test_that("lapply_with_names", {
  expect_equal(
    lapply_with_names(list(a = 1, b = 2), function(x) x * 2),
    list(a = 2, b = 4)
  )
  expect_equal(
    lapply_with_names(c("a", "b"), function(x) paste0(x, x)),
    list(a = "aa", b = "bb")
  )
})

test_that("vlapply", {
  l <- list(NULL, "", character(), 1)
  expect_identical(
    vapply(l, is.character, logical(1)),
    vlapply(l, is.character)
  )
  expect_identical(
    vapply(list(), is.character, logical(1)),
    vlapply(list(), is.character)
  )
  expect_snapshot(error = TRUE, {
    vlapply(l, identity)
    vlapply(1:5, identity)
  })
})

test_that("viapply", {
  expect_equal(
    viapply(c(1L, 2L, 3L), function(x) x),
    c(1L, 2L, 3L)
  )
  expect_equal(
    viapply(c(a = 1L, b = 2L, c = 3L), function(x) x),
    c(a = 1L, b = 2L, c = 3L)
  )
  expect_snapshot(error = TRUE, viapply(c(a = 1L, b = 2L), function(x) 1))
})

test_that("vdapply", {
  l <- list(NULL, "", character(), 1)
  f <- function(x) as.double(length(x))
  expect_identical(
    vapply(l, f, double(1)),
    vdapply(l, f)
  )
  expect_identical(
    vapply(list(), f, double(1)),
    vdapply(list(), f)
  )
  expect_snapshot(error = TRUE, {
    vdapply(l, identity)
    vdapply(letters, identity)
  })
})

test_that("add_class", {
  expect_equal(
    add_class(structure(1, class = "foo"), "bar"),
    structure(1, class = c("bar", "foo"))
  )
})

test_that("cat0", {
  expect_snapshot({
    cat0("foo", "bar", "\n")
    cat0("foo", "bar", "\n", sep = " ")
  })
})

test_that("lapply_rows", {
  expect_equal(
    lapply_rows(data_frame(), function(...) 1),
    list()
  )
  expect_equal(
    lapply_rows(
      data_frame(foo = character(), bar = integer()),
      function(...) 1
    ),
    list()
  )
  expect_snapshot(
    lapply_rows(mtcars[1:3, ], function(row) row)
  )
})

test_that("detect_download_cache_dir", {
  expect_equal(
    detect_download_cache_dir(),
    detect_download_cache_dir()
  )
})

test_that("rbind_expand", {
  expect_snapshot({
    rbind_expand(data_frame(), data_frame())
    rbind_expand(data_frame(), data_frame(foo = 1:2))
    rbind_expand(data_frame(bar = c("a", "b")), data_frame())
    rbind_expand(data_frame(foo = 1:2), data_frame(foo = 3:4))
    rbind_expand(data_frame(foo = 1:2), data_frame(bar = 3:4))
    rbind_expand(
      data_frame(foo = list(1, 2), bar = letters[1:2]),
      data_frame(foo = list(3, 4), baz = list("x", "y"))
    )
  })
})

test_that("drop_nulls", {
  expect_equal(drop_nulls(list()), list())
  expect_equal(drop_nulls(list(1, 2, 3)), list(1, 2, 3))
  expect_equal(drop_nulls(list(NULL)), list())
  expect_equal(drop_nulls(list(NULL, NULL)), list())
  expect_equal(drop_nulls(list(NULL, 1, NULL)), list(1))
})

test_that("get_num_workers", {
  withr::local_options(Ncpus = NULL)

  # ps is not installed
  mockery::stub(get_num_workers, "ps::ps_cpu_count", function(...) stop("no"))
  expect_equal(get_num_workers(), 1L)

  # ps works
  mockery::stub(get_num_workers, "ps::ps_cpu_count", function(...) 13L)
  expect_equal(get_num_workers(), 13L)

  # option works
  withr::local_options(Ncpus = 11L)
  expect_equal(get_num_workers(), 11L)
})

test_that("is_rcmd_check", {
  withr::local_envvar(NOT_CRAN = "true")
  expect_false(is_rcmd_check())

  withr::local_envvar(
    NOT_CRAN = NA_character_,
    "_R_CHECK_PACKAGE_NAME_" = NA_character_
  )
  expect_false(is_rcmd_check())

  withr::local_envvar(
    NOT_CRAN = NA_character_,
    "_R_CHECK_PACKAGE_NAME_" = "foo"
  )
  expect_true(is_rcmd_check())
})

test_that("update_named_vector", {
  cases <- list(
    list(c(a = 1, b = 2), c(a = 2, c = 5), c(a = 2, b = 2, c = 5)),
    list(double(), c(a = 2), c(a = 2)),
    list(character(), character(), character()),
    list(c(a = 1), double(), c(a = 1))
  )

  for (c in cases) {
    expect_identical(update_named_vector(c[[1]], c[[2]]), c[[3]])
  }

  expect_snapshot(error = TRUE, {
    update_named_vector(1, c(a = 1))
    update_named_vector(c(a = 1), 1)
  })
})

test_that("make_dl_status", {
  obj <- list(
    status = "status",
    url = "url",
    target = "target",
    bytes = NA_real_,
    error = NULL
  )

  expect_identical(
    make_dl_status("Got", obj$url, obj$target, 100L),
    update_named_vector(obj, list(status = "Got", bytes = 100))
  )

  expect_identical(
    make_dl_status("Failed", obj$url, obj$target, error = "foobar"),
    update_named_vector(obj, list(status = "Failed", error = "foobar"))
  )

  expect_identical(
    make_dl_status("Had", obj$url, obj$target, 100),
    update_named_vector(obj, list(status = "Had", bytes = 100))
  )
})

test_that("comma_wrap", {
  expect_equal(
    withr::with_options(
      list(width = 15),
      comma_wrap(c("foo", "x", "foo2"))
    ),
    "  foo, x,\n  foo2"
  )

  expect_equal(
    withr::with_options(
      list(width = 10),
      comma_wrap(c("foo", "x", "foo2"), indent = 0)
    ),
    "foo, x,\nfoo2"
  )

  expect_equal(
    withr::with_options(
      list(width = 10),
      comma_wrap(c("foo", "x", "foo2"), indent = 0, exdent = 2)
    ),
    "foo, x,\n  foo2"
  )

  expect_equal(
    withr::with_options(
      list(width = 15),
      comma_wrap(c("foo", "x", "foo2"), sep = "xx ")
    ),
    "  fooxx xxx\n  foo2"
  )
})

test_that("is_na_scalar", {
  pos <- list(NA, NA_character_, NA_real_, NA_integer_, NA_complex_)
  neg <- list(logical(), integer(), 1, 1L, NULL, "foobar", c(NA, 1))

  for (p in pos) {
    expect_true(is_na_scalar(p))
  }
  for (n in neg) {
    expect_false(is_na_scalar(n))
  }
})

test_that("omit_cols", {
  df <- data.frame(a = 1:5, b = 5:1, c = letters[1:5])
  expect_identical(omit_cols(df, character()), df)
  expect_identical(omit_cols(df, "x"), df)
  expect_identical(omit_cols(df, "a"), df[, 2:3])
  expect_identical(omit_cols(df, c("a", "b")), df[, 3, drop = FALSE])
  expect_identical(omit_cols(df, c("a", "b", "c")), df[, c(), drop = FALSE])
})

test_that("same_sha", {
  expect_true(same_sha("badcafe", "b"))
  expect_true(same_sha("b", "badcafe"))
  expect_false(same_sha("badcafe1", "badcafebadcafe"))
  expect_false(same_sha("badcafe", NA_character_))
  expect_false(same_sha(NA_character_, "badcafe"))
})

test_that("format_iso_8601", {
  d <- structure(1266510204, class = c("POSIXct", "POSIXt"), tzone = "UTC")
  expect_equal(format_iso_8601(d), "2010-02-18T16:23:24+00:00")
})

test_that("is_online", {
  environment(is_online)$expires <- Sys.time() - 1
  on.exit(environment(is_online)$expires <- Sys.time() - 1, add = TRUE)
  mockery::stub(is_online, "is_rcmd_check", TRUE)
  expect_false(is_online())
  mockery::stub(is_online, "is_rcmd_check", FALSE)
  mockery::stub(is_online, "pingr::is_online", TRUE)
  expect_true(is_online())
  mockery::stub(is_online, "pingr::is_online", FALSE)
  # cached for 10 minutes
  expect_true(is_online())
  environment(is_online)$expires <- Sys.time() - 1
  expect_false(is_online())
})

test_that("once_per_session", {
  once_per_session(reset = TRUE)
  on.exit(once_per_session(reset = TRUE), add = TRUE)
  expect_snapshot({
    once_per_session(message("hello"))
    once_per_session(message("hello"))
    once_per_session(reset = TRUE)
    once_per_session(message("hello"))
    once_per_session(message("hello"))
  })
})

test_that("format_error_with_stdout", {
  err <- new_error("message")
  expect_snapshot(format_error_with_stdout(err))

  err$stdout <- c("this is", "the", "standard output")
  expect_snapshot(format_error_with_stdout(err))
})

test_that("last_stdout_lines", {
  expect_snapshot({
    last_stdout_lines(letters[1:3], "stdout + stderr")
  })

  # truncated in interactive sessions
  withr:::local_options(rlib_interactive = TRUE)
  expect_snapshot({
    last_stdout_lines(letters[1:11], "stdout + stderr")
  })

  # full in non-interactive sessions
  withr:::local_options(rlib_interactive = FALSE)
  expect_snapshot({
    last_stdout_lines(letters[1:11], "stdout + stderr")
  })
})

test_that("is_windows", {
  expect_true(is_flag(is_windows()))
})

test_that("is_older_rstudio", {
  expect_true(is_flag(is_older_rstudio()))
  mockery::stub(is_older_rstudio, "rstudio$detect", list(type = "foo"))
  expect_false(is_older_rstudio())

  mockery::stub(
    is_older_rstudio,
    "rstudio$detect",
    list(type = "rstudio_console", version = package_version("1.4.801"))
  )
  expect_false(is_older_rstudio())

  mockery::stub(
    is_older_rstudio,
    "rstudio$detect",
    list(type = "rstudio_console", version = package_version("1.4.700"))
  )
  expect_true(is_older_rstudio())
})

cli::test_that_cli(configs = c("plain", "ansi"), "ansi_align_width", {
  expect_equal(ansi_align_width(character()), character())
  expect_snapshot(
    ansi_align_width(c("foobar", cli::col_red("bar")))
  )
})

test_that("get_id", {
  expect_true(is_count(get_id()))
  expect_true(get_id() != get_id())
})

test_that("safe_md5sum", {
  mkdirp(tmp <- withr::local_tempdir())
  x <- file.path(tmp, "cs\u0151\u00fa\u0171")
  file.create(x)
  expect_equal(
    safe_md5sum(x)[[1]],
    "d41d8cd98f00b204e9800998ecf8427e"
  )

  mockery::stub(
    safe_md5sum,
    "tools::md5sum",
    function(files) {
      if (files == x) stop("no") else tools::md5sum(files)
    }
  )
  expect_equal(
    safe_md5sum(x)[[1]],
    "d41d8cd98f00b204e9800998ecf8427e"
  )
})

test_that("get_euid", {
  skip_on_os("windows")
  expect_true(is_count(get_euid()))

  mockery::stub(get_euid, "processx::run", function(...) stop("nope"))
  expect_equal(get_euid(), NA_integer_)
})

test_that("res_make_empty_df", {
  df <- res_make_empty_df()
  expect_true(inherits(df, "tbl"))
  expect_equal(nrow(df), 0L)
})

test_that("res_df_defaults", {
  empty <- res_make_empty_df()
  def <- res_df_defaults()
  expect_true(all_named(def))
  expect_true(all(names(def) %in% names(empty)))

  def_types <- vcapply(def, class)
  empty_types <- res_df_entry_types()[names(def)]
  expect_true(all(def_types == empty_types | def_types == "call"))

  lengths <- viapply(def, length)
  expect_true(all(lengths == 1L | vcapply(def, class) == "call"))
})

test_that("res_df_entry_types", {
  types <- res_df_entry_types()
  expect_true(all_named(types))
  expect_true(is.character(types))
  expect_false(any(types == ""))
})

test_that("res_df_must_have", {
  types <- res_df_must_have()
  def <- res_df_defaults()
  expect_true(length(intersect(types, def)) == 0)
})

test_that("res_add_df_entries", {
  good <- list(
    ref = "package",
    type = "standard",
    package = "package",
    version = "1.0.0",
    sources = c("url1", "url2")
  )

  empty <- res_make_empty_df()

  df <- res_add_df_entries(empty, good)
  expect_identical(names(df), names(empty))
  expect_equal(nrow(df), 1)

  good2 <- data_frame(
    ref = "package",
    type = "standard",
    package = "package",
    version = "1.0.0",
    sources = list(c("url1", "url2"))
  )

  empty <- res_make_empty_df()

  df2 <- res_add_df_entries(empty, good2)
  expect_identical(df2, df)

  good3 <- data_frame(
    ref = c("package1", "package2"),
    type = c("standard", "cran"),
    package = c("package1", "package2"),
    version = c("1.0.0", "2.0.0"),
    sources = list(c("url1", "url2"), "url3")
  )

  empty <- res_make_empty_df()

  df3 <- res_add_df_entries(empty, good3)
  expect_identical(names(df3), names(empty))
  expect_equal(nrow(df3), 2)
})

test_that("resolving with a list", {
  conf <- current_config()
  cache <- list(package = pkgcache::package_cache$new(), metadata = NULL)

  do <- function() {
    res <- new_resolution(config = conf, cache = cache)
    res$push(.list = parse_pkg_refs("foo::bar"))
    res$push(.list = parse_pkg_refs("foo::bar2"))
    res$when_complete()
  }

  foo_resolve <- function(remote, direct, config, cache, dependencies) {
    list(
      ref = remote$ref,
      type = remote$type,
      package = remote$rest,
      version = "1.0.0",
      sources = c("src1", "src2"),
      direct = direct
    )
  }

  res <- withr::with_options(
    list(pkg.remote_types = list(foo = list(resolve = foo_resolve))),
    synchronise(do())
  )

  expect_equal(nrow(res), 2)
  expect_identical(res$ref, c("foo::bar", "foo::bar2"))
  expect_identical(res$package, c("bar", "bar2"))
  expect_identical(res$status, c("OK", "OK"))
  expect_true(is.list(res$remote[[1]]))
  expect_true(is.list(res$remote[[2]]))
  expect_identical(res$sources, list(c("src1", "src2"), c("src1", "src2")))
})

test_that("resolving with a tbl", {
  conf <- current_config()
  cache <- list(package = pkgcache::package_cache$new(), metadata = NULL)
  do <- function() {
    res <- new_resolution(config = conf, cache = cache)
    res$push(.list = parse_pkg_refs("foo::bar"))
    res$push(.list = parse_pkg_refs("foo2::bar2"))
    res$when_complete()
  }

  foo_resolve <- function(remote, direct, config, cache, dependencies) {
    data_frame(
      ref = remote$ref,
      type = c("type1", "type2"),
      package = c("pkg1", "pkg2"),
      version = c("ver1", "ver2"),
      sources = list(c("s11", "s12"), c("s21", "s22")),
      direct = direct
    )
  }

  types <- list(
    foo = list(resolve = foo_resolve),
    foo2 = list(resolve = foo_resolve)
  )
  res <- withr::with_options(
    list(pkg.remote_types = types),
    synchronise(do())
  )

  expect_equal(nrow(res), 4)
  expect_identical(
    res$ref,
    c("foo::bar", "foo::bar", "foo2::bar2", "foo2::bar2")
  )
  expect_identical(res$package, c("pkg1", "pkg2", "pkg1", "pkg2"))
  expect_identical(res$status, c("OK", "OK", "OK", "OK"))
  expect_true(is.list(res$remote[[1]]))
  expect_true(is.list(res$remote[[2]]))
  expect_identical(
    res$sources,
    list(c("s11", "s12"), c("s21", "s22"), c("s11", "s12"), c("s21", "s22"))
  )
})

test_that("unknown deps are pushed in the queue", {
  conf <- current_config()
  cache <- list(package = pkgcache::package_cache$new(), metadata = NULL)
  do <- function() {
    res <- new_resolution(config = conf, cache = cache)
    res$push(.list = parse_pkg_refs("foo::bar"))
    res$when_complete()
  }

  foo_resolve <- function(remote, direct, config, cache, dependencies) {
    list(
      ref = remote$ref,
      type = remote$type,
      package = remote$rest,
      version = "1.0.0",
      sources = c("src1", "src2"),
      unknown_deps = "foo::bar2",
      direct = direct
    )
  }

  res <- withr::with_options(
    list(pkg.remote_types = list(foo = list(resolve = foo_resolve))),
    synchronise(do())
  )

  expect_equal(nrow(res), 2)
  expect_identical(res$ref, c("foo::bar", "foo::bar2"))
  expect_identical(res$package, c("bar", "bar2"))
  expect_identical(res$status, c("OK", "OK"))
  expect_true(is.list(res$remote[[1]]))
  expect_true(is.list(res$remote[[2]]))
  expect_identical(res$sources, list(c("src1", "src2"), c("src1", "src2")))
})

test_that("unknown deps, tble", {
  conf <- current_config()
  cache <- list(package = pkgcache::package_cache$new(), metadata = NULL)
  do <- function() {
    res <- new_resolution(config = conf, cache = cache)
    res$push(.list = parse_pkg_refs("foo::bar"))
    res$when_complete()
  }

  foo_resolve <- function(remote, direct, config, cache, dependencies) {
    data_frame(
      ref = remote$ref,
      type = c("type1", "type2"),
      package = c("pkg1", "pkg2"),
      version = c("ver1", "ver2"),
      sources = list(c("s11", "s12"), c("s21", "s22")),
      unknown_deps = "foo::bar2",
      direct = direct,
      params = list(character())
    )
  }

  types <- list(
    foo = list(resolve = foo_resolve),
    foo2 = list(resolve = foo_resolve)
  )
  res <- withr::with_options(
    list(pkg.remote_types = types),
    synchronise(do())
  )

  expect_equal(nrow(res), 4)
  expect_identical(res$ref, c("foo::bar", "foo::bar", "foo::bar2", "foo::bar2"))
  expect_identical(res$package, c("pkg1", "pkg2", "pkg1", "pkg2"))
  expect_identical(res$status, c("OK", "OK", "OK", "OK"))
  expect_true(is.list(res$remote[[1]]))
  expect_true(is.list(res$remote[[2]]))
  expect_identical(
    res$sources,
    list(c("s11", "s12"), c("s21", "s22"), c("s11", "s12"), c("s21", "s22"))
  )
})

test_that("error", {
  conf <- current_config()
  cache <- list(package = pkgcache::package_cache$new(), metadata = NULL)
  do <- function() {
    res <- new_resolution(config = conf, cache = cache)
    res$push(.list = parse_pkg_refs("foo::bar"))
    res$push(.list = parse_pkg_refs("foo::bar2"))
    res$when_complete()
  }

  foo_resolve <- function(remote, direct, config, cache, dependencies) {
    stop("foobar", call. = FALSE)
  }

  res <- withr::with_options(
    list(pkg.remote_types = list(foo = list(resolve = foo_resolve))),
    synchronise(do())
  )

  expect_equal(nrow(res), 2)
  expect_equal(res$ref, c("foo::bar", "foo::bar2"))
  expect_snapshot(res$error[[1]])
  expect_snapshot(res$error[[2]])
  expect_equal(res$status, c("FAILED", "FAILED"))
  expect_equal(res$type, c("foo", "foo"))
})

test_that("installed refs are also resolved", {
  conf <- current_config()
  mkdirp(lib <- tempfile())
  lib <- normalizePath(lib)
  on.exit(unlink(lib, recursive = TRUE), add = TRUE)
  mkdirp(file.path(lib, "bar"))
  mkdirp(file.path(lib, "bar2"))
  cache <- list(package = pkgcache::package_cache$new(), metadata = NULL)
  cache$installed <- list(
    pkgs = data_frame(
      ref = paste0("installed::", lib, c("/bar", "/bar2")),
      type = "installed",
      status = "OK",
      package = c("bar", "bar2"),
      version = "1.0.9",
      license = NA_character_,
      needscompilation = TRUE,
      priority = NA_character_,
      md5sum = NA_character_,
      platform = "source",
      rversion = "*",
      sources = list(NULL, NULL),
      built = NA_character_,
      deps = list(NULL, NULL),
      repotype = NA_character_,
      sysreqs = NA_character_
    )
  )

  do <- function() {
    res <- new_resolution(config = conf, cache = cache, library = lib)
    res$push(.list = parse_pkg_refs("foo::bar"))
    res$push(.list = parse_pkg_refs("foo::bar2"))
    res$when_complete()
  }

  foo_resolve <- function(remote, direct, config, cache, dependencies) {
    list(
      ref = remote$ref,
      type = remote$type,
      package = remote$rest,
      version = "1.0.0",
      sources = c("src1", "src2"),
      direct = direct
    )
  }

  types <- list(foo = list(resolve = foo_resolve))

  res <- withr::with_options(
    list(pkg.remote_types = types),
    synchronise(do())
  )

  expect_equal(nrow(res), 4)
  expect_equal(res$ref[1:2], c("foo::bar", "foo::bar2"))
  expect_equal(res$ref[3:4], paste0("installed::", lib, "/", c("bar", "bar2")))
  expect_equal(res$status, rep("OK", 4))
  expect_equal(res$package, c("bar", "bar2", "bar", "bar2"))
})

test_that("explicit cran", {
  setup_fake_apps()
  conf <- current_config()
  cache <- list(
    package = pkgcache::package_cache$new(),
    metadata = pkgcache::get_cranlike_metadata_cache()
  )
  do <- function(refs) {
    res <- new_resolution(config = conf, cache = cache)
    res$push(.list = parse_pkg_refs(refs), direct = TRUE)
    res$when_complete()
  }

  res <- suppressMessages(synchronise(do("cran::dplyr")))
  expect_true(inherits(res, "tbl"))
  expect_true("cran::dplyr" %in% res$ref)
  expect_true(all(grep("::", res$ref, value = TRUE) == "cran::dplyr"))
  expect_equal(res$type, ifelse(res$ref == "cran::dplyr", "cran", "standard"))
  expect_equal(res$direct, res$ref == "cran::dplyr")
  expect_equal(res$status, rep("OK", nrow(res)))
  expect_equal(res$package, sub("^.*::", "", res$ref))
  expect_true(all(grepl(".", res$version, fixed = TRUE)))
  expect_true(all(res$platform == "source" | !res$needscompilation))
  expect_true(all(is.na(res$built) | res$platform != "source"))
  good_platforms <- c(
    default_platforms(),
    if (is_windows()) "i386+x86_64-w64-mingw32"
  )
  expect_true(all(res$platform %in% good_platforms))
  expect_true(all(
    res$rversion %in%
      c(get_minor_r_version(current_r_version()), "*")
  ))
  expect_true(is_character(res$repodir))
  expect_true(is_character(res$target))
  expect_true(all(vlapply(res$deps, inherits, what = "tbl")))
  expect_true("imports" %in% unlist(lapply(res$deps, "[[", "type")))
  expect_true("suggests" %in% unlist(lapply(res$deps, "[[", "type")))
  expect_true(all(
    grepl("source|darwin", res$platform) |
      viapply(res$sources, length) == 1
  ))
  expect_true(all(vlapply(res$remote, inherits, what = "remote_ref")))
  expect_true(all(vlapply(res$error, identical, list())))
  ## TODO: metadata
})

test_that("standard", {
  setup_fake_apps()
  conf <- current_config()
  cache <- list(
    package = pkgcache::package_cache$new(),
    metadata = pkgcache::get_cranlike_metadata_cache()
  )
  do <- function(refs) {
    res <- new_resolution(config = conf, cache = cache)
    res$push(.list = parse_pkg_refs(refs), direct = TRUE)
    res$when_complete()
  }

  res <- synchronise(do("dplyr"))
  expect_true(inherits(res, "tbl"))
  expect_true("dplyr" %in% res$ref)
  expect_true(!any(grepl("::", res$ref)))
  expect_equal(res$type, ifelse(res$ref == "cran::dplyr", "cran", "standard"))
  expect_equal(res$direct, res$ref == "dplyr")
  expect_equal(res$status, rep("OK", nrow(res)))
  expect_equal(res$package, sub("^.*::", "", res$ref))
  expect_true(all(grepl(".", res$version, fixed = TRUE)))
  expect_true(all(res$platform == "source" | !res$needscompilation))
  expect_true(all(is.na(res$built) | res$platform != "source"))
  good_platforms <- c(
    default_platforms(),
    if (is_windows()) "i386+x86_64-w64-mingw32"
  )
  expect_true(all(res$platform %in% good_platforms))
  expect_true(all(
    res$rversion %in%
      c(get_minor_r_version(current_r_version()), "*")
  ))
  expect_true(is_character(res$repodir))
  expect_true(is_character(res$target))
  expect_true(all(vlapply(res$deps, inherits, what = "tbl")))
  expect_true("imports" %in% unlist(lapply(res$deps, "[[", "type")))
  expect_true("suggests" %in% unlist(lapply(res$deps, "[[", "type")))
  expect_true(all(
    grepl("source|darwin", res$platform) |
      viapply(res$sources, length) == 1
  ))
  expect_true(all(vlapply(res$remote, inherits, what = "remote_ref")))
  expect_true(all(vlapply(res$error, identical, list())))
  ## TODO: metadata
})

test_that("dependencies are honoured", {
  setup_fake_apps()
  conf <- current_config()
  cache <- list(
    package = pkgcache::package_cache$new(),
    metadata = pkgcache::get_cranlike_metadata_cache()
  )
  do <- function(refs, deps) {
    conf$set("dependencies", deps)
    res <- new_resolution(config = conf, cache = cache)
    res$push(.list = parse_pkg_refs(refs), direct = TRUE)
    res$when_complete()
  }

  ## FALSE means nothing
  res <- synchronise(do("cran::dplyr", FALSE))
  expect_true(all(res$direct))
  expect_true(all(res$ref == "cran::dplyr"))

  ## Exactly the specified ones
  res <- synchronise(do("cran::dplyr", "Imports"))
  imported <- unique(unlist(
    lapply(res$deps, function(x) x$package[x$type == "imports"])
  ))
  expect_true(all(res$package[!res$direct] %in% imported))

  ## NA means hard ones
  res <- synchronise(do("cran::dplyr", NA))
  hard <- c("imports", "depends", "linkingto")
  harddep <- unique(unlist(
    lapply(res$deps, function(x) x$package[x$type %in% hard])
  ))
  expect_true(all(res$package[!res$direct] %in% harddep))

  ## TRUE means hard + suggests on direct, hard only on rest
  res <- synchronise(do("cran::dplyr", TRUE))
  harddep <- unique(unlist(
    lapply(res$deps, function(x) x$package[x$type %in% hard])
  ))
  softdirectdep <- unique(unlist(
    lapply(res$deps[res$direct], function(x) x$package[x$type == "suggests"])
  ))
  indirect <- res$package[!res$direct]
  expect_true(all(indirect %in% harddep | indirect %in% softdirectdep))
})

test_that("error if cannot find package", {
  setup_fake_apps()
  conf <- current_config()
  cache <- list(
    package = pkgcache::package_cache$new(),
    metadata = pkgcache::get_cranlike_metadata_cache()
  )
  do <- function(refs) {
    res <- new_resolution(config = conf, cache = cache)
    res$push(.list = parse_pkg_refs(refs), direct = TRUE)
    res$when_complete()
  }

  bad <- c("cran::thiscannotexistxxx", "neitherthisonexxx")
  res <- suppressMessages(synchronise(do(bad)))
  expect_equal(res$ref, bad)
  expect_equal(res$type, c("cran", "standard"))
  expect_equal(res$direct, c(TRUE, TRUE))
  expect_equal(res$status, c("FAILED", "FAILED"))
})

test_that("dependency types", {
  setup_fake_apps()
  conf <- current_config()
  conf$set("dependencies", TRUE)
  cache <- list(
    package = pkgcache::package_cache$new(),
    metadata = pkgcache::get_cranlike_metadata_cache()
  )
  do <- function(refs) {
    res <- new_resolution(config = conf, cache = cache)
    res$push(.list = parse_pkg_refs(refs), direct = TRUE)
    res$when_complete()
  }

  res <- suppressMessages(synchronise(do("processx")))
  dt <- res$dep_types
  expect_equal(
    dt[res$package == "processx"],
    replicate(
      sum(res$package == "processx"),
      c("Depends", "Imports", "LinkingTo", "Suggests"),
      simplify = FALSE
    )
  )
  expect_equal(
    dt[res$package != "processx"],
    replicate(
      sum(res$package != "processx"),
      c("Depends", "Imports", "LinkingTo"),
      simplify = FALSE
    )
  )
})

windows_archs <- function() c("prefer-x64", "both")

default_windows_archs <- function() {
  if (getRversion() < "4.2.0") "both" else "prefer-x64"
}

default_update_after <- function() as.difftime(24, units = "hours")

env_decode_dependencies <- function(x, name, ...) {
  if (tolower(x) %in% c("yes", "true", "1", "on")) {
    return(TRUE)
  }
  if (tolower(x) %in% c("no", "false", "0", "off")) {
    return(FALSE)
  }
  if (tolower(x) == "na") {
    return(NA)
  }
  strsplit(x, ";", fixed = TRUE)[[1]]
}

env_decode_difftime <- function(x, name, ...) {
  if (nchar(x) >= 2) {
    unit <- substr(x, nchar(x), nchar(x))
    unit <- c(s = "secs", m = "mins", h = "hours", d = "days")[unit]
    qty <- suppressWarnings(as.numeric(substr(x, 1, nchar(x) - 1)))
    if (!is.na(unit) && !is.na(qty)) {
      return(as.difftime(qty, units = unit))
    }
  }
  throw(pkg_error(
    "Invalid time interval specification in {.envvar {name}} environment
     variable: {.str {x}}",
    i = "It must have the form {.code <number><unit>}.",
    i = "The unit must be a single letter: {.code s} (seconds),
     {.code m} (minutes), {.code h} (hours) or {.code d} (days).",
    i = "Examples: {.emph 60s}, {.emph 2h}, {.emph 1d}."
  ))
}

default_sysreqs_platform <- function() {
  pkgcache::current_r_platform()
}

default_sysreqs <- function(config) {
  plt <- config$get("sysreqs_platform")
  sysreqs_is_supported(plt) &&
    (is_root() || can_sudo_without_pw())
}

default_sysreqs_sudo <- function() {
  if (os_type() != "unix") {
    FALSE
  } else {
    euid <- get_euid()
    is.na(euid) || euid != 0
  }
}

default_sysreqs_update <- function() {
  Sys.getenv("CI") == "true"
}

default_sysreqs_verbose <- function() {
  Sys.getenv("CI") != ""
}

default_sysreqs_rspm_url <- function() {
  Sys.getenv("RSPM_ROOT", "https://packagemanager.posit.co")
}

default_sysreqs_rspm_repo_id <- function() {
  Sys.getenv("RSPM_REPO_ID", "1")
}

pkgdepends_config <- sort_by_name(list(
  # -----------------------------------------------------------------------
  build_vignettes = list(
    type = "flag",
    default = FALSE,
    docs = "Whether to build vignettes for package trees.
       This is only used if the package is obtained from a package tree,
       and not from a source (or binary) package archive. By default
       vignettes are not built in this case. If you set this to `TRUE`,
       then you need to make sure that the vignette builder packages are
       available, as these are not installed by default currently."
  ),

  # -----------------------------------------------------------------------
  cache_dir = list(
    type = "string",
    default = function() detect_download_cache_dir(),
    docs = "Directory to download the packages to. Defaults to a temporary
       directory within the R session temporary directory, see
       [base::tempdir()]."
  ),

  # -----------------------------------------------------------------------
  cran_mirror = list(
    type = "string",
    default = function() default_cran_mirror(),
    docs = "CRAN mirror to use. Defaults to the `repos` option
       (see [base::options()]), if that's not set then
       `https://cran.rstudio.com`.",
    docs_pak = "CRAN mirror to use. Defaults to the `repos` option
       (see [base::options()]), if that's not set then
       `https://cran.rstudio.com`. See also [pak::repo_add()] and
       [pak::repo_get()]"
  ),

  # -----------------------------------------------------------------------
  dependencies = list(
    type = "dependencies",
    default = function() pkg_dep_types_hard(),
    # pak functions take this as an argument, so we do not need it, for now
    pak = FALSE,
    docs = "Dependencies to consider or download or install.
       Defaults to the hard dependencies, see
       [pkgdepends::pkg_dep_types_hard()]. The following values are
       supported in the `PKG_DEPENDENCIES` environment variable:
       `\"TRUE\"`, `\"FALSE\"`, `\"NA\"`, or a semicolon separated list of
       dependency types. See [pkgdepends::as_pkg_dependencies()] for
       details."
  ),

  # -----------------------------------------------------------------------
  git_submodules = list(
    type = "flag",
    default = FALSE,
    docs = "Whether or not to update submodules in git repositories. This
       affects `git::` and `gitlab::` package sources only.
       If the R package is in a subdirectory then only the submodules
       within that directory are updated. If a submodule appears in
       `.Rbuildignore`, then it is skipped."
  ),

  # -----------------------------------------------------------------------
  include_linkingto = list(
    type = "flag",
    default = FALSE,
    docs = "Whether to always include `LinkingTo` dependencies in the solution
       of and installation, even if they are needed because the packages
       are installed from binaries. This is sometimes useful, see e.g.
       <https://github.com/r-lib/pak/issues/485> for an example use case."
  ),

  # -----------------------------------------------------------------------
  library = list(
    type = "character_or_null",
    docs = "Package library to install packages to. It is also used for
       already installed packages when considering dependencies in
       [dependency lookup][pkg_deps] or
       [package installation][pkg_installation_proposal]. Defaults to the
       first path in [.libPaths()].",
    docs_pak = "Package library to install packages to. It is also used for
       already installed packages when considering dependencies."
  ),

  # -----------------------------------------------------------------------
  metadata_cache_dir = list(
    type = "string",
    default = function() tempfile(),
    # TODO: do not link to pkgcache docs from pak
    docs = "Location of metadata replica of
       [`pkgcache::cranlike_metadata_cache`]. Defaults to a temporary
       directory within the R session temporary directory, see
       [base::tempdir()]."
  ),

  # -----------------------------------------------------------------------
  metadata_update_after = list(
    type = "difftime",
    default = function() default_update_after(),
    docs = "A time interval as a [difftime] object. pkgdepends will update the
       metadata cache if it is older than this. The default is one day.
       The `PKG_METADATA_UPDATE_AFTER` environment variable may be set
       in seconds (`s` suffix), minutes (`m` suffix), hours (`h` suffix),
       or days (`d` suffix). E.g: `1d` means one day.",
    docs_pak = "A time interval as a [difftime] object. pak will update the
       metadata cache if it is older than this. The default is one day.
       The `PKG_METADATA_UPDATE_AFTER` environment variable may be set
       in seconds (`s` suffix), minutes (`m` suffix), hours (`h` suffix),
       or days (`d` suffix). E.g: `1d` means one day."
  ),

  # -----------------------------------------------------------------------
  package_cache_dir = list(
    type = "string",
    docs = "Package cache location of [`pkgcache::package_cache`]. The default
       is the pkgcache default.",
    docs_pak = "Location of the package cache on the disk. See
       [pak::cache_summary()]. Default is selected by pkgcache."
  ),

  # -----------------------------------------------------------------------
  platforms = list(
    type = "character",
    default = function() default_platforms(),
    # TODO: do not link to pkgdepends docs from pak
    docs = "Character vector of platforms to _download_ or _install_ packages
       for. See [pkgdepends::default_platforms()] for possible platform
       names. Defaults to the platform of the current R session, plus
       `\"source\"`."
  ),

  # -----------------------------------------------------------------------
  r_versions = list(
    type = "character",
    default = function() current_r_version(),
    check = is_r_version_list,
    docs = "Character vector, R versions to download or install
       packages for. It defaults to the current R version."
  ),

  # -----------------------------------------------------------------------
  sysreqs = list(
    type = "flag",
    default = function(config) default_sysreqs(config),
    docs = "Whether to automatically look up and install system requirements.
       If `TRUE`, then `r pak_or_pkgdepends()` will try to install required
       system packages. If `FALSE`, then system requirements are still
       printed (including OS packages on supported platforms), but they
       are not installed.
       By default it is `TRUE` on supported platforms,
       if the current user is the root user or password-less `sudo` is
       configured for the current user."
  ),

  # -----------------------------------------------------------------------
  sysreqs_db_update = list(
    type = "flag",
    default = TRUE,
    docs = "Whether to try to update the system requirements database from
       GitHub. If the update fails, then the cached or the build-in
       database if used. Defaults to TRUE."
  ),

  # -----------------------------------------------------------------------
  sysreqs_db_update_timeout = list(
    type = "difftime",
    default = function() {
      as.difftime(if (Sys.getenv("CI") == "") 5 else 60, units = "secs")
    },
    docs = "Timeout for the system requirements database update.
       Defaults to five seconds, except if the `CI` environment variable
       is set, then it is one minute."
  ),

  # -----------------------------------------------------------------------
  sysreqs_dry_run = list(
    type = "flag",
    default = FALSE,
    docs = "If `TRUE`, then pkgdepends only prints the system commands to
       install system requirements, but does not execute them.",
    docs_pak = "If `TRUE`, then pak only prints the system commands to
       install system requirements, but does not execute them."
  ),

  # -----------------------------------------------------------------------
  sysreqs_platform = list(
    type = "string",
    default = function() default_sysreqs_platform(),
    docs = "The platform to use for system requirements lookup. On Linux, where
       system requirements are currently supported, it must be a string
       containing the distribution name and release, separated by a dash.
       E.g.: `\"ubuntu-22.04\"`, or `\"rhel-9\"`."
  ),

  # -----------------------------------------------------------------------
  sysreqs_rspm_repo_id = list(
    type = "string",
    default = function() default_sysreqs_rspm_repo_id(),
    docs = "Posit Package Manager (formerly RStudio Package Manager) repository
       id to use for CRAN system requirements lookup. Defaults to the
       `RSPM_REPO_ID` environment variable, if set. If not set, then it
        defaults to `1`."
  ),

  # -----------------------------------------------------------------------
  sysreqs_rspm_url = list(
    type = "string",
    default = function() default_sysreqs_rspm_url(),
    docs = "Root URL of Posit Package Manager (formerly RStudio Package
       Manager) for system requirements lookup. By default the `RSPM_ROOT`
       environment variable is used, if set. If not set,
       it defaults to `https://packagemanager.posit.co`."
  ),

  # -----------------------------------------------------------------------
  sysreqs_sudo = list(
    type = "flag",
    default = function() default_sysreqs_sudo(),
    docs = "Whether to use `sudo` to install system requirements,
       on Unix. By default it is `TRUE` on Linux if the effective user id
       of the current process is not the `root` user."
  ),

  # -----------------------------------------------------------------------
  sysreqs_update = list(
    type = "flag",
    default = function() default_sysreqs_update(),
    docs = "Whether to try to update system packages that are already installed.
       It defaults to `TRUE` on CI systems: if the `CI` environment
       variable is set to `true`."
  ),

  # -----------------------------------------------------------------------
  sysreqs_verbose = list(
    type = "flag",
    default = function() default_sysreqs_verbose(),
    docs = "Whether to echo the output of system requirements installation.
       Defaults to `TRUE` if the `CI` environment variable is set."
  ),

  # -----------------------------------------------------------------------
  use_bioconductor = list(
    type = "flag",
    default = TRUE,
    docs = "Whether to automatically use the Bioconductor repositories.
       Defaults to `TRUE`."
  ),

  # -----------------------------------------------------------------------
  windows_archs = list(
    type = "string",
    default = function() default_windows_archs(),
    # we can't indent this "correctly" because markdown will take it as code
    docs = "Character scalar specifying which architectures
   to download/install for on Windows. Its possible values are:

   - `\"prefer-x64\"`: Generally prefer x64 binaries. If the current R
     session is `x64`, then we download/install x64 packages.
     (These packages might still be multi-architecture binaries!)
     If the current R session is `i386`, then we download/install
     packages for both architectures. This might mean compiling
     packages from source if the binary packages are for `x64` only,
     like the CRAN Windows binaries for R 4.2.x currently.
     `\"prefer-x64\"` is the default for R 4.2.0 and later.
   - `\"both\"`: Always download/install packages for both `i386` and
     `x64` architectures. This might need compilation from source
     if the available binaries are for `x64` only, like the CRAN
     Windows binaries for R 4.2.x currently. `\"both\"` is the default
     for R 4.2.0 and earlier."
  ),

  # -----------------------------------------------------------------------
  # Internal
  goal = list(
    type = "string",
    default = "unknown"
  ),
  sysreqs_lookup_system = list(
    type = "flag",
    default = TRUE
  )
))

#' pkgdepends configuration
#' @name pkg_config
#' @aliases pkgdepends-config pkgdepends_config
#'
#' @description
#' Configuration entries for several pkgdepends classes.
#'
#' @export
#' @details
#' pkgdepends configuration is set from several source. They are, in the
#' order of preference:
#' * Function arguments, e.g. the `config` argument of
#'   [new_pkg_installation_proposal()].
#' * Global options, set via [options()]. The name of the global option
#'   is the `pkg.` prefix plus the name of the pkgdepends configuration
#'   entry. E.g. `pkg.platforms`.
#' * Environment variables. The name of the environment variable is the
#'   `PKG_` prefix, plus the name of the pkgdepends configuration entry, in
#'   uppercase. E.g. `PKG_PLATFORMS`.
#' * Default values.
#'
#' Not all classes use all entries. E.g. a [`pkg_download_proposal`] is not
#' concerned about package libraries, so it'll ignore the `library`
#' configuration entry.
#'
#' Call `current_config()` to print the current configuration.
#'
#' # Configuration entries
#'
#' `r generate_config_docs()`

current_config <- function() {
  conf <- config$new("pkg")
  conf$add_type("dependencies", is_dependencies, env_decode_dependencies)
  conf$add_type("difftime", is_difftime, env_decode_difftime)

  map_named(pkgdepends_config, function(name, entry) {
    type <- entry$type %||% "character"
    conf$add(
      name,
      type,
      default = entry$default,
      check = entry$check %||% type,
      env_decode = entry$env_decode %||% type
    )
  })

  conf$lock()
  conf
}

test_that("windows_archs", {
  expect_snapshot(windows_archs())
})

test_that("default_windows_archs", {
  mockery::stub(default_windows_archs, "getRversion", package_version("4.1.0"))
  expect_equal(default_windows_archs(), "both")

  mockery::stub(default_windows_archs, "getRversion", package_version("4.2.0"))
  expect_equal(default_windows_archs(), "prefer-x64")
})

test_that("default_update_after", {
  expect_true(length(default_update_after()) == 1)
  expect_true(inherits(default_update_after(), "difftime"))
})

test_that("env_decode_dependencies", {
  expect_true(env_decode_dependencies("yes"))
  expect_true(env_decode_dependencies("true"))
  expect_true(env_decode_dependencies("1"))
  expect_true(env_decode_dependencies("on"))

  expect_false(env_decode_dependencies("no"))
  expect_false(env_decode_dependencies("false"))
  expect_false(env_decode_dependencies("0"))
  expect_false(env_decode_dependencies("off"))

  expect_equal(env_decode_dependencies("na"), NA)

  expect_equal(env_decode_dependencies("foo"), "foo")
  expect_equal(env_decode_dependencies("foo;bar"), c("foo", "bar"))
})

test_that("env_decode_difftime", {
  expect_snapshot({
    env_decode_difftime("0s")
    env_decode_difftime("120s")
    env_decode_difftime("0m")
    env_decode_difftime("120m")
    env_decode_difftime("0h")
    env_decode_difftime("24h")
    env_decode_difftime("0d")
    env_decode_difftime("1d")
  })

  expect_snapshot(error = TRUE, env_decode_difftime("", "PKG_UPDATE"))
  expect_snapshot(error = TRUE, env_decode_difftime("123", "PKG_UPDATE"))
  expect_snapshot(error = TRUE, env_decode_difftime("1k", "PKG_UPDATE"))
  expect_snapshot(error = TRUE, env_decode_difftime("k1k", "PKG_UPDATE"))
})

test_that("default_sysreqs_sudo", {
  mockery::stub(default_sysreqs_sudo, "os_type", "windows")
  expect_false(default_sysreqs_sudo())

  mockery::stub(default_sysreqs_sudo, "os_type", "unix")
  mockery::stub(default_sysreqs_sudo, "get_euid", 0L)
  expect_false(default_sysreqs_sudo())

  mockery::stub(default_sysreqs_sudo, "get_euid", 100L)
  expect_true(default_sysreqs_sudo())

  mockery::stub(default_sysreqs_sudo, "get_euid", NA_integer_)
  expect_true(default_sysreqs_sudo())
})

test_that("default_sysreqs_verbose", {
  withr::local_envvar(CI = "true")
  expect_true(default_sysreqs_verbose())

  withr::local_envvar(CI = NA_character_)
  expect_false(default_sysreqs_verbose())
})

test_that("default_sysreqs_rspm_url", {
  withr::local_envvar(RSPM_ROOT = "https://custom")
  expect_equal(default_sysreqs_rspm_url(), "https://custom")

  withr::local_envvar(RSPM_ROOT = NA_character_)
  expect_equal(
    default_sysreqs_rspm_url(),
    "https://packagemanager.posit.co"
  )
})

test_that("default_sysreqs_rspm_repo_id", {
  withr::local_envvar(RSPM_REPO_ID = "2")
  expect_equal(default_sysreqs_rspm_repo_id(), "2")

  withr::local_envvar(RSPM_REPO_ID = NA_character_)
  expect_equal(default_sysreqs_rspm_repo_id(), "1")
})

test_that("current_config", {
  expect_snapshot(
    sort(current_config()$list())
  )
})

#' Dependency resolution
#'
#' Collect information about dependencies of R packages, recursively.
#'
#' [`pkg_deps`], [`pkg_download_proposal`] and [`pkg_installation_proposal`]
#' all resolve their dependencies recursively, to obtain information about
#' all packages needed for the specified [package references][pkg_refs].
#'
#' ## CRAN and Bioconductor packages
#'
#' Resolution currently start by downloading the CRAN and Bioconductor
#' metadata, if it is out of date. For CRAN, we also download additional
#' metadata, that includes file sizes, SHA hashes, system requirements,
#' and "built" (for binary packages) and "packaged" time stamps. The extra
#' meta information is updated daily currently, so for some packages it
#' might be incorrect or missing.
#'
#' ## GitHub packages
#'
#' For GitHub packages, we query their download URL to be able to
#' download the package later, and also download their `DESCRIPTION`
#' file, to learn about their dependencies.
#'
#' ## Local packages
#'
#' From local package files we extract the `DESCRIPTION` file, to learn
#' about their dependencies.
#'
#' ## The `remotes` field in `DESCRIPTION`
#'
#' We support the non-standard `Remotes` field in the package `DESCRIPTION`
#' file. This field may contain a list of package references for any of the
#' dependencies that are specified in one of the `Depends`, `Includes`,
#' `Suggests` or `Enhances` fields. The syntax is a comma separated list of
#' [package references][pkg_refs].
#'
#' ## The result
#'
#' The result of the resolution is a data frame with information about the
#' packages and their dependencies.
#'
#' ```{r child = {options(rx_downloads=NULL); "tools/doc/resolution-result.Rmd"}}
#' ```
#' `r { options(rx_downloads = NULL); doc_share_rmd("tools/doc/resolution-result.Rmd", "inst/docs/resolution-result.rds")}`
#'
#' ## Resolution failures
#'
#' The resolution process does not stop on error. Instead, failed
#' resolutions return and error object in the `error` column of the result
#' data frame.
#'
#' @name pkg_resolution
#' @aliases pkg_resolution_result
NULL

pkgplan_resolve <- function(self, private) {
  "!DEBUG pkgplan_resolve (sync)"
  synchronise(self$async_resolve())
}

pkgplan_async_resolve <- function(self, private) {
  "!DEBUG pkgplan_resolve (async)"
  ## We remove this, to avoid a discrepancy between them
  private$downloads <- NULL
  private$solution <- NULL
  private$system_packages <- NULL
  private$sysreqs <- NULL

  private$dirty <- TRUE
  private$resolution <- new_resolution(
    config = private$config,
    cache = private$cache,
    library = private$config$get("library"),
    remote_types = private$remote_types
  )

  private$resolution$push(direct = TRUE, .list = private$remotes)

  private$resolution$when_complete()$then(function(x) {
    private$dirty <- FALSE
    x
  })
}

pkgplan_get_resolution <- function(self, private) {
  if (is.null(private$resolution$result)) {
    throw(pkg_error(
      "No resolution yet.",
      i = "You need to call {.code $resolve()} first."
    ))
  }
  private$resolution$result
}

pkgplan__subset_resolution <- function(self, private, which) {
  if (is.null(private$resolution$result)) {
    throw(pkg_error(
      "No resolution yet.",
      i = "You need to call {.code $resolve()} first."
    ))
  }
  res <- private$resolution$result[which, ]
  attr(res, "metadata") <- attr(private$resolution$result, "metadata")
  res
}

new_resolution <- function(config, cache, library = NULL, remote_types = NULL) {
  resolution$new(config, cache, library, remote_types)
}

resolution <- R6::R6Class(
  "resolution",
  public = list(
    result = NULL,
    initialize = function(config, cache, library = NULL, remote_types = NULL) {
      res_init(self, private, config, cache, library, remote_types)
    },
    push = function(..., direct = FALSE, .list = list()) {
      res_push(self, private, ..., direct = direct, .list = .list)
    },
    when_complete = function() private$deferred
  ),

  private = list(
    remote_types = NULL,
    config = NULL,
    cache = NULL,
    library = NULL,
    deferred = NULL,
    state = NULL,
    params = NULL,
    dependencies = NULL,
    metadata = NULL,
    system_packages = NULL,
    sysreqs = NULL,
    bar = NULL,

    delayed = list(),
    delayed_refs = character(),
    resolve_delayed = function(resolve) {
      res__resolve_delayed(self, private, resolve)
    },

    create_progress_bar = function() res__create_progress_bar(self, private),
    done_progress_bar = function() res__done_progress_bar(self, private),

    set_result = function(row_idx, value) {
      res__set_result(self, private, row_idx, value)
    },
    sysreqs_match = function() res__sysreqs_match(self, private),
    try_finish = function(resolve) res__try_finish(self, private, resolve)
  )
)

res_init <- function(self, private, config, cache, library, remote_types) {
  "!DEBUG resolution init"
  private$config <- config
  private$cache <- cache
  private$library <- library
  private$remote_types <- remote_types %||% default_remote_types()
  private$metadata <- list(resolution_start = Sys.time())
  private$dependencies <- as_pkg_dependencies(config$get("dependencies"))
  private$bar <- private$create_progress_bar()

  self$result <- res_make_empty_df()

  private$state <- data_frame(
    ref = character(),
    remote = list(),
    status = character(),
    direct = logical(),
    async_id = integer(),
    started_at = Sys.time()[FALSE]
  )

  private$deferred <- asNamespace("pkgcache")$deferred$new(
    type = "resolution_queue",
    parent_resolve = function(value, resolve) {
      "!DEBUG resolution done"
      # maybe a non-resolution task ended, e.g. sysreqs
      if (!"id" %in% names(value)) {
        return(private$try_finish(resolve))
      }
      id <- value$id
      value <- value$value
      wh <- which(id == private$state$async_id)
      private$state$status[wh] <- "OK"

      ## Rule out installed:: refs and packages with ?source param
      not_inst <- value$type != "installed"
      prms <- value[["params"]]
      if (!is.data.frame(value)) {
        prms <- list(prms)
      }
      want_source <- vlapply(prms, is_true_param, "source")
      want_reinst <- vlapply(prms, is_true_param, "reinstall")

      package <- value$package
      for (par in private$params) {
        if (is_true_param(par$params, "source")) {
          want_source[package %in% par$package | par$package == ""] <- TRUE
        }
        if (is_true_param(par$params, "reinstall")) {
          want_reinst[package %in% par$package | par$package == ""] <- TRUE
        }
      }

      npkgs <- value$package[not_inst & !want_source & !want_reinst]

      ## Installed already? Resolve that as well
      if (length(npkgs)) {
        for (lib in private$library) {
          ex <- file.exists(file.path(lib, npkgs))
          if (any(ex)) {
            lib <- normalizePath(lib, winslash = "/", mustWork = FALSE)
            refs <- paste0("installed::", lib, "/", npkgs[ex])
            refs <- setdiff(refs, private$state$ref)
            self$push(.list = parse_pkg_refs(refs))
            npkgs <- npkgs[!ex]
          }
        }
        lib <- .Library
        ex <- file.exists(file.path(lib, npkgs)) &
          npkgs %in% recommended_packages()
        if (any(ex)) {
          lib <- normalizePath(lib, winslash = "/", mustWork = FALSE)
          refs <- paste0("installed::", lib, "/", npkgs[ex])
          refs <- setdiff(refs, private$state$ref)
          self$push(.list = parse_pkg_refs(refs))
        }
      }

      private$set_result(wh, value)
      private$try_finish(resolve)
    },

    parent_reject = function(value, resolve) {
      "!DEBUG resolution failed"
      id <- value$id
      wh <- which(id == private$state$async_id)
      private$state$status[wh] <- "FAILED"
      rec <- private$state[wh, ]
      fail_val <- list(
        ref = rec$ref,
        type = rec$remote[[1]]$type,
        package = rec$remote[[1]]$package %|z|% NA_character_,
        version = NA_character_,
        sources = NA_character_,
        direct = rec$direct,
        status = "FAILED",
        remote = rec$remote,
        error = list(value)
      )
      private$set_result(wh, fail_val)
      private$try_finish(resolve)
    }
  )

  # If sysreqs is supported on this platform, then
  # 1. look up system packages, and
  # 2. (try to) update sysreqs mapping
  sys_sup <- sysreqs_is_supported(private$config$get("sysreqs_platform"))
  sys_lookup <- private$config$get("sysreqs_lookup_system")
  if (sys_sup && sys_lookup) {
    private$system_packages <- NA # nocovif !is_linux()
    async_system_list_packages(private$config)$then(function(x) {
      # nocovif !is_linux()
      private$system_packages <- x
      NULL
    })$then(private$deferred) # nocovif !is_linux() # nocovif !is_linux()
  }
  if (sys_sup) {
    private$sysreqs <- NA
    sysreqs2_async_update_metadata(config = private$config)$then(function() {
      private$sysreqs <- TRUE
      NULL
    })$then(private$deferred)
  }
}

res_push <- function(self, private, ..., direct, .list = .list) {
  self
  private
  new <- c(list(...), .list)

  ## Set (new) parameters
  new_types <- vcapply(new, "[[", "type")
  params <- new[new_types == "param"]
  new <- new[new_types != "param"]

  if (length(params) + length(private$params)) {
    private$params <- c(private$params, params)
    update_params(self, private, private$params)
  }

  ## Drop the ones already resolving up front
  new_refs <- vcapply(new, "[[", "ref")
  keep <- !new_refs %in% c(private$state$ref, private$delayed_refs)
  new <- new[keep]
  new_refs <- new_refs[keep]

  ## Drop duplicates as well
  uni_refs <- !duplicated(new_refs)
  if (!all(uni_refs)) {
    new <- new[uni_refs]
    new_refs <- new_refs[uni_refs]
  }

  ## We do CRAN/BioC/standard in batches
  ## TODO: direct ones in one go as well
  batch_types <- setdiff(
    c("cran", "standard", "bioc", "installed"),
    private$remote_types
  )
  delay <- vcapply(new, "[[", "type") %in% batch_types
  if (!direct && any(delay)) {
    refs <- vcapply(new[delay], "[[", "ref")
    private$delayed <- c(private$delayed, new[delay])
    private$delayed_refs <- c(private$delayed_refs, refs)
    new <- new[!delay]
    "!DEBUG pushing `sum(delay)` batch resolutions"
  }

  for (n in new) {
    "!DEBUG resolution push `n$ref`"
    dx <- resolve_remote(
      n,
      direct,
      private$config,
      private$cache,
      private$dependencies,
      remote_types = private$remote_types
    )

    private$state <- rbind(
      private$state,
      data_frame(
        ref = n$ref,
        remote = list(n),
        status = NA_character_,
        direct = direct,
        async_id = dx$id,
        started_at = Sys.time()
      )
    )

    dx$dx$then(private$deferred)
  }
}

res__resolve_delayed <- function(self, private, resolve) {
  n <- private$delayed
  private$delayed <- list()
  private$delayed_refs <- character()

  refs <- vcapply(n, "[[", "ref")
  done <- refs %in% private$state$ref
  n <- n[!done]
  refs <- vcapply(n, "[[", "ref")
  "!DEBUG resolving `length(private$delayed)` delayed remotes"

  if (length(n)) {
    types <- vcapply(n, "[[", "type")
    utypes <- unique(types)
    for (t in utypes) {
      n2 <- n[types == t]

      dx <- resolve_remote(
        n2,
        direct = FALSE,
        private$config,
        private$cache,
        private$dependencies,
        remote_types = private$remote_types
      )

      private$state <- rbind(
        private$state,
        data_frame(
          ref = vcapply(n2, "[[", "ref"),
          remote = n2,
          status = NA_character_,
          direct = FALSE,
          async_id = dx$id,
          started_at = Sys.time()
        )
      )
      dx$dx$then(private$deferred)
    }
  }

  private$try_finish(resolve)
}

res__set_result <- function(self, private, row_idx, value) {
  unknown <- if ("unknown_deps" %in% names(value)) value$unknown_deps
  if (is.null(unknown) && !is.null(attr(value, "unknown_deps"))) {
    unknown <- attr(value, "unknown_deps")
    attr(value, "unknown_deps") <- NULL
  }
  value <- value[setdiff(names(value), "unknown_deps")]

  if (is.data.frame(value)) {
    res__set_result_df(self, private, row_idx, value)
  } else {
    res__set_result_list(self, private, row_idx, value)
  }

  "!DEBUG resolution setting result, total: `nrow(self$result)`"
  if (length(unknown)) self$push(.list = parse_pkg_refs(unknown))
}

res__set_result_df <- function(self, private, row_idx, value) {
  # remove the ones that are already done
  done <- value$ref %in% self$result$ref
  value <- value[!done, ]

  # avoid removing the ones that are direct and already on the TODO list
  running <- intersect(value$ref, private$state$ref)
  avoid <-
    !value$direct[match(running, value$ref)] &
    private$state$direct[match(running, private$state$ref)]
  value <- value[!value$ref %in% running[avoid], ]

  if (nrow(value)) self$result <- res_add_df_entries(self$result, value)
}

res__set_result_list <- function(self, private, row_idx, value) {
  # already done?
  if (all(value$ref %in% self$result$ref)) {
    return()
  }
  if (is.data.frame(value)) {
    # if a data frame, then some might be done
    done <- value$ref %in% private$state$ref
    value <- value[!done, drop = FALSE]
  }
  self$result <- res_add_df_entries(self$result, value)
}

res__sysreqs_match <- function(self, private) {
  if ("sysreqs" %in% names(self$result)) {
    sys <- sysreqs2_match(self$result$sysreqs, config = private$config)
    if (!is.null(spkgs <- private$system_packages)) {
      spkgs <- spkgs[grepl("^.i$", spkgs$status), ] # nocovif !is_linux()
      allspkgs <- unique(unlist(c(spkgs$package, spkgs$provides))) # nocovif !is_linux()
      for (i in seq_along(sys)) {
        # nocovif !is_linux()
        elt <- sys[[i]] # nocovif !is_linux()
        for (j in seq_along(elt)) {
          # nocovif !is_linux()
          elt[[j]]$packages_missing <- setdiff(elt[[j]]$packages, allspkgs) # nocovif !is_linux()
        } # nocovif !is_linux()
        if (!is.null(elt)) sys[[i]] <- elt # nocovif !is_linux()
      }
    }
    self$result$sysreqs_packages <- sys
    platform <- private$config$get("sysreqs_platform")
    cmd_upd <- sysreqs2_command(platform, "update_command")
    cmd_inst <- sysreqs2_command(platform, "install_command")
    pre <- lapply(sys, function(x) unlist(lapply(x, "[[", "pre_install")))
    post <- lapply(sys, function(x) unlist(lapply(x, "[[", "post_install")))
    spkg <- lapply(sys, function(x) {
      unlist(lapply(x, function(xx) xx$packages_missing %||% xx$packages))
    })

    pre <- vcapply(pre, paste, collapse = ";")
    post <- vcapply(post, paste, collapse = ";")
    spkg <- vcapply(spkg, paste, collapse = " ")
    if (!is.na(cmd_upd)) {
      pre <- ifelse(spkg == "", pre, paste0(cmd_upd, ";", pre))
    }
    spkg <- ifelse(spkg == "", "", paste(cmd_inst, spkg))

    self$result$sysreqs_pre_install <- pre
    self$result$sysreqs_post_install <- post
    self$result$sysreqs_install <- spkg
  } else {
    self$result$sysreqs_packages <- list(NULL) # nocovif !is_linux()
  }
}

res__try_finish <- function(self, private, resolve) {
  "!DEBUG resolution trying to finish with `nrow(self$result)` results"
  if (length(private$delayed)) {
    return(private$resolve_delayed(resolve))
  }
  if (
    all(!is.na(private$state$status)) &&
      !identical(private$system_packages, NA) &&
      !identical(private$sysreqs, NA)
  ) {
    "!DEBUG resolution finished"
    update_params(self, private, private$params)
    update_dep_types(self, private)
    private$metadata$resolution_end <- Sys.time()
    self$result$cache_status <-
      calculate_cache_status(self$result, private$cache)
    attr(self$result, "metadata") <- private$metadata
    class(self$result) <- c("pkg_resolution_result", class(self$result))
    private$done_progress_bar()
    if (sysreqs_is_supported(private$config$get("sysreqs_platform"))) {
      private$sysreqs_match()
    }
    resolve(self$result)
  }
}

resolve_remote <- function(
  remote,
  direct,
  config,
  cache,
  dependencies,
  remote_types = NULL
) {
  remote
  direct
  config
  cache
  dependencies
  remote_types

  remote_types <- c(default_remote_types(), remote_types)

  type <- remote$type %||% unique(vcapply(remote, "[[", "type"))
  if (length(type) != 1) {
    throw(pkg_error(
      "Invalid remote or remote list, multiple types: {.val {type}}.",
      i = msg_internal_error()
    ))
  }

  resolve <- remote_types[[type]]$resolve
  if (is.null(resolve)) {
    throw(pkg_error(
      "Don't know how to resolve remote type {.val {type}}.",
      i = msg_internal_error()
    ))
  }

  id <- get_id()
  dx <- async(resolve)(
    remote,
    direct = direct,
    config = config,
    cache = cache,
    dependencies = dependencies
  )$then(function(value) list(value = value, id = id))$catch(error = function(
    err
  ) {
    err$id <- id
    err$call <- NULL
    throw(
      pkg_error(
        "{pak_or_pkgdepends()} resolution error for {.pkg {remote$ref}}.",
        .data = list(id = id)
      ),
      parent = err
    )
  })

  list(dx = dx, id = id)
}

update_params <- function(self, private, params) {
  for (par in params) {
    if (par$package %in% c("", "*") && length(self$result$params)) {
      self$result$params <- lapply(self$result$params, function(p) {
        p[names(par$params)] <- par$params
        p
      })
      self$result$remote <- lapply(self$result$remote, function(rem) {
        if (is.list(rem)) {
          rem$params[names(par$params)] <- par$params
        }
        rem
      })
    } else {
      sel <- self$result$package == par$package
      if (any(sel)) {
        self$result$params[sel] <- lapply(self$result$params[sel], function(p) {
          p[names(par$params)] <- par$params
          p
        })
        self$result$remote[sel] <- lapply(
          self$result$remote[sel],
          function(rem) {
            if (is.list(rem)) {
              rem$params[names(par$params)] <- par$params
            }
            rem
          }
        )
      }
    }
  }
}

update_dep_types <- function(self, private) {
  # these are the default
  def <- vlapply(
    self$result$dep_types,
    function(x) length(x) == 1 && x == "default"
  )
  dep_types <- self$result$dep_types

  directdef <- self$result$direct[def]
  directpkgdef <- self$result$directpkg[def]

  # dependencies get the default
  dep_types[def][!directpkgdef] <- list(private$dependencies[["indirect"]])

  # direct refs that use the default
  dep_types[def][directdef] <- list(private$dependencies[["direct"]])

  # direct packages that are not direct refs will get their default from
  # their direct ref(s)
  dpkgs <- unique(self$result$package[def][directpkgdef & !directdef])
  for (pkg in dpkgs) {
    myrefs <- self$result$package == pkg & self$result$direct
    myrefdeps <- unique(unlist(dep_types[myrefs]))
    mypkgs <- self$result$package[def] == pkg
    dep_types[def][mypkgs & directpkgdef & !directdef] <- list(myrefdeps)
  }

  self$result$dep_types <- dep_types
}

resolve_from_description <- function(
  path,
  sources,
  remote,
  direct,
  config,
  cache,
  dependencies
) {
  dsc <- desc::desc(file = path)
  deps <- resolve_ref_deps(
    dsc$get_deps(),
    dsc$get("Remotes")[[1]],
    dsc$get(extra_config_fields(dsc$fields()))
  )

  rversion <- tryCatch(
    get_minor_r_version(dsc$get_built()$R),
    error = function(e) "*"
  )

  platform <- if (dsc$has_fields("Built")) {
    built <- dsc$get_built()
    archs <- gsub(" ", "", dsc$get("Archs"))
    if (built$OStype == "windows") {
      if (is.na(archs) || archs == "i386,x64" || archs == "x64,i386") {
        "i386+x86_64-w64-mingw32"
      } else {
        built$Platform
      }
    } else {
      built$Platform
    }
  } else {
    "source"
  }

  nc <- dsc$get_field("NeedsCompilation", NA)
  if (!is.na(nc)) {
    nc <- tolower(nc) %in% c("true", "yes")
  }

  unknown <- deps$ref[deps$type %in% dependencies]

  meta <- c(
    RemotePkgRef = remote$ref,
    RemoteType = remote$type,
    RemoteSha = NULL, # TODO
    RemoteUrl = NULL, # TODO
    RemoteUsername = NULL, # TODO
    RemoteRepo = NULL, # TODO
    RemoteBranch = NULL # TODO
  )

  list(
    ref = remote$ref,
    type = remote$type,
    direct = direct,
    status = "OK",
    package = dsc$get_field("Package"),
    version = dsc$get_field("Version"),
    license = dsc$get_field("License", NA_character_),
    needscompilation = nc,
    md5sum = dsc$get_field("MD5sum", NA_character_),
    built = dsc$get_field("Built", NA_character_),
    platform = platform,
    rversion = rversion,
    deps = list(deps),
    sources = sources,
    remote = list(remote),
    unknown_deps = setdiff(unknown, "R"),
    extra = list(list(description = dsc)),
    metadata = meta,
    params = list(remote$params),
    sysreqs = dsc$get_field("SystemRequirements", "")
  )
}

drop_config_needs <- function(x) {
  x[!grepl("^config/needs", x, ignore.case = TRUE)]
}

# TODO: Parse remotes and Config/Needs/* fields

resolve_from_metadata <- function(
  remotes,
  direct,
  config,
  cache,
  dependencies
) {
  remotes
  direct
  config
  cache
  dependencies
  if (is.list(dependencies)) {
    dependencies$direct <- drop_config_needs(dependencies$direct)
    dependencies$indirect <- drop_config_needs(dependencies$indirect)
  }

  ## Single remote, or a list of remotes
  if ("ref" %in% names(remotes)) {
    packages <- remotes$package
    refs <- remotes$ref
    types <- remotes$type
    params <- list(remotes$params)
  } else {
    packages <- vcapply(remotes, "[[", "package")
    refs <- vcapply(remotes, "[[", "ref")
    types <- vcapply(remotes, "[[", "type")
    params <- lapply(remotes, "[[", "params")
  }

  if (!direct) {
    dependencies <- dependencies$indirect
  }
  "!DEBUG resolving `length(refs)` batch resolution"
  cache$metadata$async_deps(packages, dependencies = dependencies)$then(
    function(data) {
      cols <- c(
        "ref",
        "type",
        "status",
        "package",
        "version",
        "license",
        "needscompilation",
        "priority",
        "md5sum",
        "platform",
        "rversion",
        "repodir",
        "target",
        "deps",
        "sources",
        "mirror",
        "filesize",
        "sha256",
        "sysreqs",
        "os_type"
      )

      cols <- intersect(names(data), cols)

      res <- data[cols]
      res$built <- data[["built"]] %||% rep(NA_character_, nrow(res))
      idx <- match(res$package, packages)
      res$ref[!is.na(idx)] <- na.omit(refs[idx])
      res$repotype <- res$type
      res$type[] <- "standard"
      res$type[!is.na(idx)] <- na.omit(types[idx])
      res$needscompilation <-
        tolower(res$needscompilation) %in% c("yes", "true")
      res$direct <- direct & res$ref %in% refs
      res$params <- replicate(nrow(res), character())
      res$params[!is.na(idx)] <- params[na.omit(idx)]

      res$metadata <- get_standard_metadata(res)

      # Drop binaries if source package was requested
      want_source <- vlapply(res$params, is_true_param, "source")
      todrop <- res$platform != "source" & want_source
      if (any(todrop)) {
        res <- res[!todrop, ]
      }

      if (length(bad <- attr(data, "unknown"))) {
        idx <- match(bad, packages)
        bad[!is.na(idx)] <- na.omit(refs[idx])
        failed <- make_failed_resolution(
          bad,
          ifelse(!is.na(idx), types[idx], "standard"),
          direct & bad %in% refs
        )
        failed <- res_add_defaults(failed)
        ## This is added later, we don't want to add it here, because
        ## it messes up the dep_types of the successful resolutions
        failed <- failed[names(failed) != "dep_types"]
        res <- rbind_expand(res, failed)
      }

      res
    }
  )
}

get_standard_metadata <- function(tab) {
  meta <- replicate(nrow(tab), character(), simplify = FALSE)
  for (i in seq_len(nrow(tab))) {
    meta[[i]] <-
      c(
        RemoteType = tab$type[i],
        RemotePkgRef = tab$ref[i],
        RemoteRef = tab$ref[i],
        RemoteRepos = tab$mirror[i],
        RemotePkgPlatform = tab$platform[i],
        RemoteSha = tab$version[i]
      )
  }
  meta
}

make_failed_resolution <- function(refs, type, direct) {
  rstr <- paste(refs, collapse = ", ")
  err <- structure(
    list(message = paste0("Can't find package called ", rstr, ".")),
    class = c("error", "condition")
  )
  data_frame(
    ref = refs,
    type = type,
    package = sub("^[a-z]+::", "", refs),
    version = NA_character_,
    sources = replicate(length(refs), NA_character_, simplify = FALSE),
    direct = direct,
    status = "FAILED",
    remote = parse_pkg_refs(refs),
    error = replicate(length(refs), err, simplify = FALSE)
  )
}

#' @export

`[.pkg_resolution_result` <- function(x, i, j, drop = FALSE) {
  class(x) <- setdiff(class(x), "pkg_resolution_result")
  NextMethod("[")
}

install_binary <- function(
  filename,
  lib = .libPaths()[[1L]],
  metadata = NULL,
  quiet = FALSE
) {
  assert_that(
    is_existing_file(filename),
    is_string(lib),
    all_named(metadata),
    is.null(quiet) || is_flag(quiet)
  )

  stdout <- ""

  px <- make_install_process(filename, lib = lib, metadata = metadata)

  repeat {
    px$poll_io(100)
    stdout <- paste0(stdout, px$read_output())
    if (!px$is_alive() && !px$is_incomplete_output()) {
      break
    }
  }

  if (px$get_exit_status() != 0) {
    stop("Package installation failed\n", stdout)
  }

  cli::cli_alert_success(paste0("Installed ", filename))

  invisible(px$get_result())
}


local_binary_package <- function(pkgname, ..., envir = parent.frame()) {
  # All arguments must be named
  args <- list(...)
  stopifnot(length(args) == 0 || all_named(args))

  d <- create_temp_dir()
  pkgdir <- file.path(d, pkgname)
  dir.create(pkgdir)
  nms <- names(args)
  for (i in seq_along(args)) {
    dir.create(
      file.path(pkgdir, dirname(nms[[i]])),
      showWarnings = FALSE,
      recursive = TRUE
    )
    withr::with_connection(
      list(con = file(file.path(pkgdir, nms[[i]]), open = "wb")),
      {
        writeLines(args[[i]], con, sep = "\n")
      }
    )
  }

  filename <- file.path(d, paste0(pkgname, ".tgz"))
  withr::with_dir(
    dirname(filename),
    utils::tar(basename(filename), pkgname, compression = "gzip")
  )

  # We do not want to unlink files if we are calling this from the R console,
  # useful when debugging.
  is_globalenv <- identical(envir, globalenv())
  if (!is_globalenv) {
    withr::defer(unlink(d, recursive = TRUE), envir = envir)
  }
  filename
}

binary_test_package <- function(name) {
  mkdirp(tmp <- tempfile())
  file.copy(test_path("fixtures", name), tmp, recursive = TRUE)
  zip_path <- system.file(package = "zip", "bin", .Platform$r_arch)
  withr::with_path(
    zip_path,
    pkgbuild::build(file.path(tmp, name), binary = TRUE, quiet = TRUE)
  )
}

source_test_package <- function(name) {
  mkdirp(tmp <- tempfile())
  file.copy(test_path("fixtures", name), tmp, recursive = TRUE)
  pkgbuild::build(file.path(tmp, name), binary = FALSE, quiet = TRUE)
}

dummy_worker_process <- R6::R6Class(
  "dummy_worker_process",
  inherit = callr::r_process,
  public = list(
    initialize = function(...) {
      super$initialize(...)
    },
    get_built_file = function() NA_character_
  )
)

make_dummy_worker_process <- function(n_iter = 10, sleep = 1, status = 0) {
  n_iter
  sleep
  status
  function(...) {
    dummy_worker_process$new(callr::r_process_options(
      func = function(n_iter, sleep, status) {
        # nocov start
        for (i in seq_len(n_iter)) {
          cat("out ", i, "\n", sep = "")
          message("err ", i)
          Sys.sleep(sleep)
        }
        status
        .GlobalEnv$.Last <- function() {
          rm(list = ".Last", envir = .GlobalEnv)
          quit(save = "no", status = status)
        }
        # nocov end
      },
      args = list(n_iter = n_iter, sleep = sleep, status = status),
      stderr = "2>&1"
    ))
  }
}

make_install_plan <- function(ref, lib = .libPaths()[1]) {
  r <- pkg_plan$new(ref, library = lib)
  r$resolve()
  r$solve()
  r$download_solution()
  r$get_install_plan()
}

describe("install_packages", {
  setup_fake_apps()
  local_cli_config()

  it("works with source packages", {
    pkg <- source_test_package("foo")

    libpath <- test_temp_dir()

    expect_snapshot({
      plan <- make_install_plan(
        paste0("local::", pkg, "?nocache"),
        lib = libpath
      )
      install_package_plan(plan, lib = libpath)
    })

    callr::r(function(l) library("foo", lib.loc = l), list(libpath))
  })
})

# Test the fix for https://github.com/r-lib/pkgdepends/issues/160
test_that("install_plan dependency order with binary and source pkgs", {
  plan <- structure(
    list(
      package = c(
        "pkgdepends",
        "assertthat",
        "backports",
        "callr",
        "cli",
        "crayon",
        "curl",
        "desc",
        "digest",
        "ellipsis",
        "fansi",
        "filelock",
        "glue",
        "jsonlite",
        "lpSolve",
        "pillar",
        "pkgbuild",
        "pkgcache",
        "pkgconfig",
        "prettyunits",
        "ps",
        "R6",
        "rappdirs",
        "rematch2",
        "rlang",
        "rprojroot",
        "tibble",
        "utf8",
        "uuid",
        "vctrs",
        "withr",
        "zip",
        "processx"
      ),
      binary = c(
        FALSE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        TRUE,
        FALSE
      ),
      dependencies = structure(
        list(
          c(
            "assertthat",
            "backports",
            "callr",
            "cli",
            "crayon",
            "curl",
            "desc",
            "filelock",
            "glue",
            "jsonlite",
            "lpSolve",
            "pkgbuild",
            "pkgcache",
            "prettyunits",
            "ps",
            "rematch2",
            "rprojroot",
            "R6",
            "tibble",
            "withr",
            "zip"
          ),
          character(0),
          character(0),
          c("processx", "R6"),
          c("assertthat", "crayon", "glue", "fansi"),
          character(0),
          character(0),
          c("assertthat", "R6", "crayon", "rprojroot"),
          character(0),
          "rlang",
          character(0),
          character(0),
          character(0),
          character(0),
          character(0),
          c("cli", "crayon", "fansi", "rlang", "utf8", "vctrs"),
          c(
            "callr",
            "cli",
            "crayon",
            "desc",
            "prettyunits",
            "R6",
            "rprojroot",
            "withr"
          ),
          c(
            "assertthat",
            "callr",
            "cli",
            "curl",
            "digest",
            "filelock",
            "glue",
            "prettyunits",
            "R6",
            "processx",
            "rappdirs",
            "rlang",
            "tibble",
            "uuid"
          ),
          character(0),
          character(0),
          character(0),
          character(0),
          character(0),
          "tibble",
          character(0),
          "backports",
          c("cli", "crayon", "fansi", "pillar", "pkgconfig", "rlang"),
          character(0),
          character(0),
          c("ellipsis", "digest", "glue", "rlang"),
          character(0),
          character(0),
          c("ps", "R6")
        ),
        class = "AsIs"
      )
    ),
    metadata = list(
      download_start = structure(
        1583239144.75085,
        class = c("POSIXct", "POSIXt")
      ),
      download_end = structure(
        1583239146.91793,
        class = c("POSIXct", "POSIXt")
      )
    ),
    row.names = c(NA, -33L),
    class = c("tbl_df", "tbl", "data.frame")
  )

  plan2 <- add_recursive_dependencies(plan)
  expect_equal(plan$binary, plan2$binary)
  expect_equal(
    plan$dependencies[plan$binary],
    plan2$dependencies[plan2$binary]
  )
  expect_equal(
    sort(plan2$dependencies[[1]]),
    sort(c(
      plan$dependencies[[1]],
      c(
        "processx",
        "fansi",
        "digest",
        "rappdirs",
        "rlang",
        "uuid",
        "pillar",
        "pkgconfig",
        "utf8",
        "vctrs",
        "ellipsis"
      )
    ))
  )
  expect_equal(
    sort(plan$dependencies[[33]]),
    sort(plan2$dependencies[[33]])
  )
})

test_that("install_binary metadata", {
  pkg <- binary_test_package("foo")

  libpath <- test_temp_dir()

  metadata <- c("Foo" = "Bar", "Foobar" = "baz")
  suppressMessages(
    install_binary(pkg, lib = libpath, metadata = metadata, quiet = TRUE)
  )

  dsc <- desc::desc(file.path(libpath, "foo"))
  expect_equal(dsc$get("Foo")[[1]], "Bar")
  expect_equal(dsc$get("Foobar")[[1]], "baz")

  rds <- readRDS(file.path(libpath, "foo", "Meta", "package.rds"))
  dsc2 <- rds$DESCRIPTION
  expect_equal(dsc2[["Foo"]], "Bar")
  expect_equal(dsc2[["Foobar"]], "baz")
})

test_that("install_package_plan metadata", {
  setup_fake_apps()
  local_cli_config()

  pkg <- source_test_package("foo")
  libpath <- test_temp_dir()

  expect_snapshot({
    plan <- make_install_plan(
      paste0("local::", pkg, "?nocache"),
      lib = libpath
    )
    plan$metadata[[1]] <- c("Foo" = "Bar", "Foobar" = "baz")
    plan$vignettes <- FALSE
    install_package_plan(plan, lib = libpath, num_workers = 1)
  })

  dsc <- desc::desc(file.path(libpath, "foo"))
  expect_equal(dsc$get("Foo")[[1]], "Bar")
  expect_equal(dsc$get("Foobar")[[1]], "baz")

  rds <- readRDS(file.path(libpath, "foo", "Meta", "package.rds"))
  dsc2 <- rds$DESCRIPTION
  expect_equal(dsc2[["Foo"]], "Bar")
  expect_equal(dsc2[["Foobar"]], "baz")
})

test_that("folders with potentially problematic characters", {
  skip_on_cran()
  local_cli_config()

  tmp <- tempfile()
  on.exit(tryCatch(unloadNamespace("foo"), error = identity), add = TRUE)
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  on.exit(environment(need_internal_tar)$internal <- NULL, add = TRUE)

  pkg <- binary_test_package("foo")

  folders <- c(
    "s p a c e s",
    "\u00fa\u00e1\u00f6\u0151\u00e9",
    "s' p' a' c' e' s'"
  )

  skipped <- 0

  for (f in folders) {
    error <- FALSE
    tryCatch(
      {
        if ("foo" %in% loadedNamespaces()) {
          unloadNamespace("foo")
        }
        unlink(tmp, recursive = TRUE)
        dir.create(tmp)
        dir.create(file.path(tmp, f))
        libdir <- dir(tmp)
        libpath <- file.path(tmp, libdir)
      },
      warning = function(e) error <<- TRUE,
      error = function(e) error <<- TRUE
    )
    if (error) {
      skipped <- skipped + 1
      next
    }

    ## Reset this
    environment(need_internal_tar)$internal <- NULL

    suppressMessages(
      install_binary(pkg, lib = libpath, quiet = TRUE)
    )
    library("foo", lib.loc = libpath)
    expect_equal(foo::foo(), NULL)
    unloadNamespace("foo")

    ## Make sure tar is internal
    unlink(tmp, recursive = TRUE)
    dir.create(tmp)
    dir.create(libpath)
    environment(need_internal_tar)$internal <- NULL
    withr::with_envvar(
      c(TAR = NA),
      withr::with_path("foobar", action = "replace", {
        suppressMessages(install_binary(pkg, lib = libpath, quiet = TRUE))
      })
    )

    library("foo", lib.loc = libpath)
    expect_equal(foo::foo(), NULL)
    unloadNamespace("foo")
  }

  if (skipped) skip(paste(skipped, " path tests were skipped"))
})

project_root_anchors <-
  c("DESCRIPTION", ".git", ".Rproj.user", "renv.lock", "renv")

find_package_root <- function(path = ".") {
  find_project_root(path = path, anchors = "DESCRIPTION")
}

is_package_root <- function(path = ".") {
  file.exists(file.path(path, "DESCRIPTION"))
}

find_project_root <- function(path = ".", anchors = project_root_anchors) {
  is_root <- function(path) {
    identical(
      normalizePath(path, winslash = "/"),
      normalizePath(dirname(path), winslash = "/")
    )
  }

  if (!file.exists(path)) {
    stop("Path does not exist: ", path)
  }
  cur_path <- normalizePath(path, winslash = "/")
  errmsg <- paste0(
    "Could not find R package in `",
    path,
    "` or its parent directories."
  )
  max_depth <- 100
  for (i in 1:max_depth) {
    if (any(file.exists(file.path(cur_path, anchors)))) {
      return(cur_path)
    } else if (is_root(cur_path)) {
      stop(errmsg)
    } else {
      cur_path <- dirname(cur_path)
    }
  }
  stop(errmsg, " Checked ", max_depth, " parent directories.") # nocov
}

expect_equal_named_lists <- function(object, expected, ...) {
  expect_true(!is.null(names(object)) && !is.null(names(expected)))
  expect_true(is.list(object) && is.list(expected))
  object <- object[order(names(object))]
  expected <- expected[order(names(expected))]
  expect_equal(object, expected)
}

test_temp_file <- function(
  fileext = "",
  pattern = "test-file-",
  envir = parent.frame(),
  create = TRUE
) {
  tmp <- tempfile(pattern = pattern, fileext = fileext)
  if (identical(envir, .GlobalEnv)) {
    message("Temporary files will _not_ be cleaned up")
  } else {
    withr::defer(
      try(unlink(tmp, recursive = TRUE, force = TRUE), silent = TRUE),
      envir = envir
    )
  }
  if (create) {
    cat("", file = tmp)
    normalizePath(tmp)
  } else {
    tmp
  }
}

test_temp_dir <- function(pattern = "test-dir-", envir = parent.frame()) {
  tmp <- test_temp_file(pattern = pattern, envir = envir, create = FALSE)
  dir.create(tmp, recursive = TRUE, showWarnings = FALSE)
  normalizePath(tmp)
}

# Create a temporary directory and switch the working directory to it
local_temp_dir <- function(..., .local_envir = parent.frame()) {
  tmp <- withr::local_tempdir(..., .local_envir = .local_envir)
  withr::local_dir(tmp, .local_envir = .local_envir)
  invisible(tmp)
}

test_package_root <- function() {
  x <- tryCatch(
    find_package_root(),
    error = function(e) NULL
  )

  if (!is.null(x)) {
    return(x)
  }

  pkg <- testthat::testing_package()
  x <- tryCatch(
    find_package_root(
      path = file.path("..", "..", "00_pkg_src", pkg)
    ),
    error = function(e) NULL
  )

  if (!is.null(x)) {
    return(x)
  }

  stop("Cannot find package root")
}

skip_in_covr <- function() {
  if (Sys.getenv("R_COVR") == "true") skip("In covr")
}

# TODO: update this to cli.num_colors

local_cli_config <- function(
  unicode = FALSE,
  dynamic = FALSE,
  ansi = FALSE,
  num_colors = 1,
  .local_envir = parent.frame()
) {
  withr::local_options(
    cli.dynamic = dynamic,
    cli.ansi = ansi,
    cli.unicode = unicode,
    crayon.enabled = num_colors > 1,
    crayon.colors = num_colors,
    .local_envir = .local_envir
  )
  withr::local_envvar(
    PKG_OMIT_TIMES = "true",
    PKG_OMIT_SIZES = "true",
    .local_envir = .local_envir
  )
}

pst <- function(...) suppressMessages(...)

long_basename <- function(x) {
  # remove potential trailing slash
  l <- nchar(x)
  x <- ifelse(substr(x, l, l) %in% c("/", "\\"), substr(x, 1, l - 1), x)
  sub("^.*[/\\]", "", x)
}

read_all <- function(path) {
  bytes <- readBin(path, "raw", file.size(path))
  chr <- rawToChar(bytes)
  Encoding(chr) <- "UTF-8"
  chr
}

#' Create a binary package from an installed package
#'
#' The built package will be in the current working directory.
#'
#' This function is currently experimental.
#'
#' @param pkg Package name.
#' @param library Library path.
#' @param flavor Platform flavor. Defaults to the `PKG_BUILD_FLAVOR`
#'   environment variable. If not `NULL` or an empty string, then it is
#'   appended to the platform string with a dash.
#' @param build_number An integer number that is added to the file name,
#'   after the version number, to be able to have multiple builds for the
#'   same package version.
#' @return Path to the built package.
#'
#' @export
#' @keywords internal

pkg_build <- function(
  pkg,
  library = .libPaths()[1],
  flavor = Sys.getenv("PKG_BUILD_FLAVOR"),
  build_number = 1L
) {
  pkgdir <- file.path(library, pkg)
  if (!dir.exists(pkgdir)) {
    throw(pkg_error(
      "Cannot find package {.pkg {pkg}} in library at {.path {library}}."
    ))
  }
  platform <- pkgcache::current_r_platform()
  if (nzchar(flavor %||% "")) {
    platform <- paste0(platform, "-", flavor)
  }
  meta <- c(
    RemoteBuildPlatform = platform,
    GraphicsAPIVersion = pkgcache::get_graphics_api_version(),
    InternalsId = pkgcache::get_internals_id()
  )
  add_metadata(pkgdir, meta)
  dsc <- desc::desc(file = pkgdir)
  version <- dsc$get_field("Version")
  rversion <- get_minor_r_version(getRversion())

  sys <- sysname()
  if (sys == "windows") {
    install_md5_sums(pkg)
    fn <- paste0(
      pkg,
      "_",
      version,
      "_",
      "b",
      build_number,
      "_",
      "R",
      rversion,
      if (nzchar(flavor %||% "")) paste0("_", flavor),
      ".zip"
    )
    zip::zip(fn, pkgdir, mode = "cherry-pick")
  } else {
    ext <- if (sys == "mac") ".tgz" else ".tar.gz"
    fn <- paste0(
      pkg,
      "_",
      version,
      "_",
      "b",
      build_number,
      "_",
      "R",
      rversion,
      "_",
      platform,
      ext
    )
    ffn <- file.path(normalizePath("."), fn)
    old <- getwd()
    on.exit(setwd(old), add = TRUE)
    setwd(dirname(pkgdir))
    utils::tar(ffn, pkg, compression = "gzip", compression_level = 9)
  }

  fn
}

install_md5_sums <- function(pkgdir) {
  old <- getwd()
  on.exit(setwd(old), add = TRUE)

  setwd(pkgdir)
  fns <- setdiff(dir(".", recursive = TRUE), "MD5")
  md5 <- cli::hash_file_md5(fns)
  writeLines(
    paste0(md5, " *", fns),
    "MD5"
  )
}

# -----------------------------------------------------------------------
# API

parse_remote_url <- function(specs, config, ...) {
  parsed_specs <- re_match(specs, type_url_rx())
  parsed_specs$ref <- parsed_specs$.text
  cn <- setdiff(colnames(parsed_specs), c(".match", ".text"))
  parsed_specs <- parsed_specs[, cn]
  parsed_specs$type <- "url"
  parsed_specs$hash <- vcapply(specs, url_hash)
  parsed_specs$package[parsed_specs$package == ""] <- NA_character_

  # Special case downloads from GH
  parsed_gh_specs <- re_match(specs, type_github_download_url_rx())
  parsed_specs$package <- ifelse(
    is.na(parsed_specs$package) & !is.na(parsed_gh_specs$repo),
    parsed_gh_specs$repo,
    parsed_specs$package
  )

  lapply(
    seq_len(nrow(parsed_specs)),
    function(i) as.list(parsed_specs[i, ])
  )
}

resolve_remote_url <- function(
  remote,
  direct,
  config,
  cache,
  dependencies,
  ...
) {
  remote
  direct
  config
  cache
  dependencies
  list(...)
  nocache <- is_true_param(remote$params, "nocache")
  type_url_resolve(
    remote,
    cache,
    config,
    direct = direct,
    dependencies = dependencies[[2 - direct]],
    nocache = nocache
  )$then(function(x) x$resolution)
}

download_remote_url <- function(
  resolution,
  target,
  target_tree,
  config,
  cache,
  which,
  on_progress
) {
  resolution
  target
  target_tree
  config
  cache
  which
  on_progress

  remote <- resolution$remote[[1]]
  packaged <- resolution$metadata[[1]][["RemotePackaged"]]
  tmpd <- type_url_tempdir(remote, config)

  rimraf(c(target, target_tree))

  status <- NULL

  # If we are installing an install plan, then it might not be there
  nocache <- is_true_param(resolution$params[[1]], "nocache")
  if (!file.exists(tmpd$ok)) {
    dx <- type_url_resolve(remote, cache, config, nocache)$then(function(x) {
      newres <- x$resolution
      status <<- x$data$status
      res_etag <- resolution$metadata[[1]][["RemoteEtag"]]
      new_etag <- newres$metadata[["RemoteEtag"]]
      if (res_etag != new_etag) {
        warning("Package file at `", remote$url, "` has changed") # nocov
      }
    })
  } else {
    status <- "Had"
    dx <- async_constant(0)
  }

  dx$then(function() {
    if (packaged == "TRUE") {
      mkdirp(dirname(target))
      if (!file.copy(tmpd$archive, target, overwrite = TRUE)) {
        throw(pkg_error(
          "Failed to copy package from {.path {tmpd$archive}} to
             {.path{ target}}.",
          i = "It was downloaded from {.url {remote$url}}."
        ))
      }
    } else {
      mkdirp(target_tree)
      if (!file.copy(tmpd$extract, target_tree, recursive = TRUE)) {
        throw(pkg_error(
          # nocov start
          "Failed to copy package from {.path {tmpd$archive}} to
             {.path{ target}}.",
          i = "It was downloaded from {.url {remote$url}}."
        )) # nocov end
      }
    }
  })$then(function() status)
}

satisfy_remote_url <- function(resolution, candidate, config, ...) {
  ## 1. package name must match
  if (resolution$package != candidate$package) {
    return(structure(FALSE, reason = "Package names differ"))
  }

  ## 2. installed ref is good, if it has the same etag
  if (candidate$type == "installed") {
    want_reinst <- is_true_param(resolution$params[[1]], "reinstall")
    if (want_reinst) {
      return(structure(FALSE, reason = "Re-install requested"))
    }
    t1 <- tryCatch(candidate$extra[[1]]$remoteetag, error = function(e) "")
    t2 <- resolution$metadata[[1]][["RemoteEtag"]]
    ok <- is_string(t1) && is_string(t2) && t1 == t2
    if (!ok) {
      return(structure(FALSE, reason = "Installed URL etag mismatch"))
    } else {
      return(TRUE)
    }
  }

  ## 3. same url is good
  if (candidate$type == "url") {
    if (resolution$ref == candidate$ref) {
      return(TRUE)
    }
    return(structure(FALSE, reason = "URL mismatch"))
  }

  structure(FALSE, reason = "Repo type mismatch")
}

installedok_remote_url <- function(installed, solution, config, ...) {
  identical(installed$package, solution$package) &&
    identical(installed$version, solution$version) &&
    identical(installed[["remotetype"]], "url") &&
    identical(installed[["remoteetag"]], solution$metadata[[1]][["RemoteEtag"]])
}

# -----------------------------------------------------------------------
# Internal functions

type_url_rx <- function() {
  paste0(
    "^",
    ## Optional package name
    "(?:(?<package>",
    package_name_rx(),
    ")=)?",
    "(?:url::)",
    "(?<url>.*)",
    "$"
  )
}

url_hash <- function(x) {
  cli::hash_obj_md5(x)
}

# E.g. url::https://github.com/r-lib/tidyselect/archive/main.tar.gz

type_github_download_url_rx <- function() {
  paste0(
    "^",
    "(?:url::)",
    "https://github.com/",
    github_username_rx(),
    "/",
    github_repo_rx(),
    "/archive/",
    "(?<branch>[^/.]+)",
    "[.]tar[.]gz$"
  )
}

type_url_tempdir <- function(remote, config) {
  base <- basename(remote$url)
  filename <- paste0(substr(remote$hash, 1, 7), "-", basename(remote$url))
  archive <- file.path(config$get("cache_dir"), filename)
  extract <- file.path(config$get("cache_dir"), paste0(filename, "-t"))
  ok <- file.path(config$get("cache_dir"), paste0(filename, "-ok"))
  list(
    archive = archive,
    extract = extract,
    cachepath = file.path("archives", filename),
    ok = ok
  )
}

type_url_download_and_extract <- function(
  remote,
  cache,
  config,
  tmpd,
  nocache
) {
  id <- NULL
  tmpd <- tmpd
  async_constant(
    if (nocache) {
      mkdirp(dirname(tmpd$archive))
      download_one_of(remote$url, tmpd$archive)$then(function(dl) {
        attr(dl, "action") <- "Got"
        dl
      })
    } else {
      cache$package$async_update_or_add(
        tmpd$archive,
        remote$url,
        path = tmpd$cachepath,
        http_headers = default_download_headers(remote$url)
      )
    }
  )$then(function(dl) {
    tmpd$status <<- attr(dl, "action")
    tmpd$etag <<- if (is.na(dl$etag)) substr(dl$sha256, 1, 16) else dl$etag
    tmpd$id <<- cli::hash_obj_md5(tmpd$etag)
    rimraf(c(tmpd$extract, tmpd$ok))
    mkdirp(tmpd$extract)
    run_uncompress_process(tmpd$archive, tmpd$extract)
  })$then(function(status) {
    tmpd$pkgdir <<- get_pkg_dir_from_archive_dir(tmpd$extract)
    cat("ok\n", file = tmpd$ok)
    tmpd
  })
}

type_url_resolve <- function(
  remote,
  cache,
  config,
  direct = FALSE,
  dependencies = character(),
  nocache = FALSE
) {
  tmpd <- type_url_tempdir(remote, config)
  xdirs <- NULL
  type_url_download_and_extract(remote, cache, config, tmpd, nocache)$then(
    function(dirs) {
      xdirs <<- dirs
      resolve_from_description(
        path = dirs$pkgdir,
        sources = remote$url,
        remote = remote,
        direct = direct,
        config = config,
        cache = cache,
        dependencies = dependencies
      )
    }
  )$then(function(x) {
    x$target <- file.path(
      "src/contrib",
      paste0(x$package, "_", x$version, "-", substr(xdirs$id, 1, 10), ".tar.gz")
    )
    x$metadata[["RemoteEtag"]] <- xdirs$etag
    x$extra[[1]][["resolve_download_status"]] <- tmpd$status
    x$metadata[["RemotePackaged"]] <-
      x$extra[[1]]$description$has_fields("Packaged")
    x$params[[1]] <- remote$params
    list(resolution = x, data = xdirs)
  })
}


get_pkg_dir_from_archive_dir <- function(x) {
  top <- dir(x)
  if (length(top) != 1) {
    throw(pkg_error(
      "Package archive at {.path {x}} should contain exactly one directory.",
      i = "It has {cli::no(length(top))} file{?s}/director{?y/ies}:
       {.path {top}}"
    ))
  }
  file.path(x, top)
}

ghr <- local({
  # -------------------------------------------------------------------------

  ghr_list <- function(repo) {
    synchronize(async_ghr_list(repo))
  }

  async_ghr_list <- function(repo) {
    repo <- parse_slug(repo)
    query <- glue::glue(
      "{
    rateLimit {
      cost
      remaining
    }
    repository(owner: \"<repo$user>\", name: \"<repo$repo>\") {
      releases(last: 100) {
        nodes {
          id
          name
          createdAt
          tagName
        }
      }
    }
  }",
      .open = "<",
      .close = ">"
    )

    github_query(query)$then(function(resp) {
      rls <- resp$obj$data$repository$releases$nodes
      data_frame(
        id = vcapply(rls, "[[", "id"),
        name = vcapply(rls, "[[", "name"),
        tag_name = vcapply(rls, "[[", "tagName"),
        created_at = parse_iso_8601(vcapply(rls, "[[", "createdAt"))
      )
    })
  }

  # -------------------------------------------------------------------------

  ghr_get <- function(repo, tag) {
    synchronize(async_ghr_get(repo, tag))
  }

  async_ghr_get <- function(repo, tag) {
    prepo <- parse_slug(repo)
    ep <- glue::glue("/repos/{prepo$owner}/{prepo$repo}/releases/tags/{tag}")
    async_github_v3_query(ep)$then(function(resp) {
      jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = FALSE)
    })
  }

  # -------------------------------------------------------------------------

  ghr_list_assets <- function(repo, tag) {
    synchronize(async_ghr_list_assets(repo, tag))
  }

  async_ghr_list_assets <- function(repo, tag) {
    repo <- parse_slug(repo)
    query <- glue::glue(
      "{
    rateLimit {
      cost
      remaining
    }
    repository(owner: \"<repo$user>\", name: \"<repo$repo>\") {
      release(tagName: \"<tag>\") {
        releaseAssets(last: 100) {
          nodes {
            id
            name
            downloadUrl
            size
            createdAt
            updatedAt
            contentType
          }
        }
      }
    }
  }",
      .open = "<",
      .close = ">"
    )

    github_query(query)$then(function(resp) {
      asts <- resp$obj$data$repository$release$releaseAssets$nodes
      data_frame(
        id = vcapply(asts, "[[", "id"),
        name = vcapply(asts, "[[", "name"),
        download_url = vcapply(asts, "[[", "downloadUrl"),
        size = viapply(asts, "[[", "size"),
        created_at = parse_iso_8601(vcapply(asts, "[[", "createdAt")),
        updated_at = parse_iso_8601(vcapply(asts, "[[", "updatedAt")),
        content_type = vcapply(asts, "[[", "contentType")
      )
    })
  }

  # -------------------------------------------------------------------------

  ghr_add_asset <- function(repo, file, tag, name = basename(file)) {
    invisible(synchronize(async_ghr_add_asset(repo, file, tag, name)))
  }

  async_ghr_add_asset <- function(repo, file, tag, name = basename(file)) {
    repo
    file
    tag
    name

    async_ghr_delete_asset(repo, tag, name)$then(
      function(res) res$release$upload_url
    )$catch(async_http_404 = function(err) {
      async_ghr_create(repo, tag)$then(function(rel) rel$upload_url)
    })$then(function(upload_url) {
      upload_url <- sub("[{].*[}]", "", upload_url)
      prepo <- parse_slug(repo)
      async_github_v3_query(
        url = upload_url,
        endpoint = "",
        query = c(name = name),
        method = "POST",
        headers = c("Content-Type" = "application/octet-stream"),
        file = file
      )
    })$then(function(resp) {
      jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = FALSE)
    })
  }

  # -------------------------------------------------------------------------

  ghr_delete_asset <- function(repo, tag, name) {
    synchronize(async_ghr_delete_asset(repo, tag, name))
  }

  async_ghr_delete_asset <- function(repo, tag, name) {
    prepo <- parse_slug(repo)
    async_ghr_get(repo, tag)$then(function(res) {
      release_id <- res$id
      asset_names <- vcapply(res$assets, "[[", "name")
      if (name %in% asset_names) {
        asset_id <- res$assets[[match(name, asset_names)]]$id
        ep <- glue::glue(
          "/repos/{prepo$user}/{prepo$repo}/releases/assets/{asset_id}"
        )
        async_github_v3_query(ep, method = "DELETE")$then(function(resp) {
          list(
            release = res,
            deleted = TRUE
          )
        })
      } else {
        list(release = res, deleted = FALSE)
      }
    })
  }

  # -------------------------------------------------------------------------

  ghr_create <- function(repo, tag) {
    invisible(synchronize(async_ghr_create(repo, tag)))
  }

  async_ghr_create <- function(
    repo,
    tag,
    description = "",
    draft = FALSE,
    prerelease = FALSE,
    generate_release_notes = FALSE
  ) {
    prepo <- parse_slug(repo)
    ep <- glue::glue("/repos/{prepo$owner}/{prepo$repo}/releases")
    data <- tojson$write_str(
      list(
        tag_name = tag,
        name = paste0(prepo$repo, " ", tag),
        body = description,
        draft = draft,
        prerelease = prerelease,
        generate_release_notes = generate_release_notes
      ),
      list(auto_unbox = TRUE)
    )

    async_github_v3_query(
      ep,
      data = data,
      method = "POST"
    )$then(function(resp) {
      jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = FALSE)
    })
  }

  # -------------------------------------------------------------------------

  parse_slug <- function(slug) {
    parts <- strsplit(slug, "/", fixed = TRUE)[[1]]
    list(user = parts[1], owner = parts[1], repo = parts[2])
  }

  async_github_v3_query <- function(
    endpoint,
    query = NULL,
    method = c("GET", "POST", "DELETE"),
    headers = NULL,
    data = NULL,
    file = NULL,
    url = NULL
  ) {
    method <- match.arg(method)

    headers <- update_named_vector(type_github_get_headers(), headers)

    base <- url %||%
      Sys.getenv("R_PKG_GITHUB_API_URL", "https://api.github.com")
    query_str <- paste(
      glue::glue(
        "{curl::curl_escape(names(query))}={curl::curl_escape(query)}"
      ),
      collapse = "&"
    )
    url <- paste0(base, endpoint, "?", query_str)

    px <- if (method == "GET") {
      http_get(url, headers = headers)
    } else if (method == "POST") {
      if (is.null(data) + is.null(file) != 1) {
        throw(pkg_error(
          "Must specify exactly of {.arg data} and {.arg file} for POST requests."
        ))
      }
      if (is.null(data)) {
        data <- readBin(file, "raw", file.size(file))
      }
      http_post(url, data, headers = headers)
    } else if (method == "DELETE") {
      http_delete(url, headers = headers)
    }

    px$then(http_stop_for_status)
  }

  # github_query() is in type-github.R

  # -----------------------------------------------------------------------
  # Exported functions

  structure(
    list(
      .internal = environment(),

      async_add_asset = async_ghr_add_asset,
      async_delete_asset = async_ghr_delete_asset,
      async_create = async_ghr_create,
      async_get = async_ghr_get,
      async_list = async_ghr_list,
      async_list_assets = async_ghr_list_assets,

      add_asset = ghr_add_asset,
      delete_asset = ghr_delete_asset,
      create = ghr_create,
      get = ghr_get,
      list = ghr_list,
      list_assets = ghr_list_assets
    )
  )
})

# -------------------------------------------------------------------------

#' GitHub Releases
#'
#' Functions to query and manipulate GitHub releases. These functions are
#' currently experimental.
#'
#' @details
#'
#' ## List releases
#'
#' ### Description
#'
#' `ghr$list()` lists the last 100 releases for a GitHub repository.
#' `ghr$async_list()` is the async version of `ghr$list()`.
#'
#' ### Usage
#' ```
#' ghr$list(repo)
#' ghr$async_list(repo)
#' ```
#'
#' ### Arguments
#'
#' * `repo`: repository slug, e.g. `"cran/cli"`.
#'
#' ### Value
#'
#' Data frame with columns:
#'   * `id`: release id,
#'   * `name`: release name, usually the version number, possibly with
#'     a `v` prefix: `3.6.1` or `v3.6.1`, but can be different.
#'   * `tag_name`: usually the same as `name`.
#'   * `created_at`: `POSIXct` vector.
#'
# -------------------------------------------------------------------------
#'
#' ## Get information about a release
#'
#' ### Description
#'
#' `ghr$get()` downloads information about a release, including
#' release assets.
#'
#' `ghr$async_get` is the async version of `ghr$get`.
#'
#' ### Usage
#' ```
#' ghr$get(repo, tag)
#' ```
#'
#' ### Arguments
#'
#' * `repo`: repository slug, e.g. `"cran/cli"`.
#' * `tag`: tag to get.
#'
#' ### Value
#'
#' Named list, see
#' <https://docs.github.com/en/rest/releases/releases#get-a-release>
#' for the entries.
#'
# -------------------------------------------------------------------------
#'
#' ## List assets of a release
#'
#' ### Description
#'
#' `ghr$list_assets()` lists the last 100 assets of a release.
#'
#' `ghr$async_list_assets()` is the async version of `ghr$list_assets()`
#'
#' ### Usage
#' ```
#' ghr$list_assets(repo, tag)
#' ```
#'
#' ### Arguments
#'
#' * `repo`: repository slug, e.g. `"cran/cli"`.
#' * `tag`: tag to query.
#'
#' ### Value
#'
#' Data frame with columns:
#' * `id`: asset id,
#' * `name`: file name of the asset,
#' * `download_url`: download URL,
#' * `size`: size in bytes,
#' * `created_at`: `POSIXct` vector,
#' * `updated_at`: `POSXct` vector,
#' * `content_type`: content type of asset.
#'
# -------------------------------------------------------------------------
#'
#' ## Add a release asset
#'
#' ### Description
#'
#' `ghr$add_asset()` adds an asset to a GitHub release.
#'
#' `ghr$async_add_asset()` is the async version of `ghr$add_asset()`.
#'
#' ### Usage
#' ```
#' ghr%add_asset(repo, file, tag, name = basename(file))
#' ```
#'
#' ### Arguments
#'
#' * `repo`: repository slug, e.g. `cran/cli`.
#' * `file`: path to file to upload as an asset.
#' * `tag`: tag name to add the asset to. It must exist on GitHub.
#' * `name`: file name of the asset in the release.
#'
#' ### Details
#'
#' If an asset with the same name already exists, then that will be deleted
#' first.
#'
#' ### Value
#'
#' Response from GitHub as a named list. See
#' <https://docs.github.com/en/rest/releases/assets#upload-a-release-asset>
#' for the structure.
#'
# -------------------------------------------------------------------------
#' ## Delete a release asset
#'
#' ### Description
#'
#' `ghr$delete_asset()` deleted a release asset.
#'
#' `ghr$async_delete_asset()` is an async version of `ghr$delete_asset()`.
#'
#' ### Usage
#' ```
#' ghr$delete_asset(repo, tag, name)
#' ghr$async_delete_asset(repo, tag, name)
#' ```
#'
#' ### Arguments
#'
#' * `repo`: repository slug, e.g. `cran/cli`.
#' * `tag`: tag name to create a release for. It must exist on GitHub.
#' * `name`: name of the asset.
#'
#' ### Value
#'
#' A list with entries:
#' * `release`: a list with the data about the release, before the
#'   deletion. It has the same format as the return value of `ghr$get()`.
#' * `deleted`: `TRUE` if the asset was deleted. `FALSE` if the asset
#'   did not exist.
#'
# -------------------------------------------------------------------------
#'
#' ## Create a GitHub release
#'
#' ### Description
#'
#' `ghr$create()` creates a GitHub release from a tag.
#'
#' `ghr$async_create()` is an async version of `ghr$create()`.
#'
#' ### Usage
#' ```
#' ghr$create(
#'   repo,
#'   tag,
#'   description = "",
#'   draft = FALSE,
#'   prerelease = FALSE,
#'   generage_release_notes = FALSE
#' )
#' ```
#'
#' ### Arguments
#'
#' * `repo`: repository slug, e.g. `cran/cli`.
#' * `tag`: tag name to create a release for. It must exist on GitHub.
#' * `description`: release description.
#' * `draft`: whether to create a draft release.
#' * `prerelease`: whether to create a prerelease.
#' * `generate_release_notes`: whether to auto-generate release notes.
#'
#' ### Value
#'
#' Response from GitHub as a named list. See
#' <https://docs.github.com/en/rest/releases/releases#create-a-release>
#' for the structure.
#'
# -------------------------------------------------------------------------
#'
#' @name ghr
#' @keywords internal
#' @export

ghr

install_extracted_binary <- function(
  filename,
  lib_cache,
  pkg_cache,
  lib,
  metadata,
  now
) {
  pkg <- verify_extracted_package(filename, pkg_cache)
  add_metadata(pkg$path, metadata)
  pkg_name <- pkg$name

  lockfile <- lock_cache(lib_cache, pkg_name, getOption("install.lock"))
  if (!is.null(lockfile)) {
    on.exit(filelock::unlock(lockfile), add = TRUE)
  }

  installed_path <- file.path(lib, pkg_name)
  if (file.exists(installed_path)) {
    if (is_windows()) {
      # First move the existing library (which still works even if a process has
      # the DLL open), then try to delete it, which may fail if another process
      # has the file open. Some points:
      # - the <lib_cache> / <pkg_name> directory might exist with the leftovers
      #   of a previous installation, typically because the DLL file was/is
      #   locked, so we could not delete it after the move.
      # - so we create a random path component to avoid interference
      # - we also unlink() the whole package-specific cache directory,
      #   to avoid accumulating junk there. This is safe, well, if we are
      #   locking, which is strongly suggested.
      move_to <- file.path(lib_cache, pkg_name, basename(tempfile()))
      unlink(dirname(move_to), recursive = TRUE, force = TRUE)
      dir.create(dirname(move_to), showWarnings = FALSE, recursive = TRUE)
      ret <- file.rename(installed_path, move_to)
      if (!ret) {
        throw(pkg_error(
          "Failed to move installed package {.pkg {pkg_name}} at
          {.path {installed_path}}.",
          .data = list(package = pkg_name),
          .class = "install_filesystem_error"
        ))
      }
      ret <- unlink(move_to, recursive = TRUE, force = TRUE)
      if (ret != 0) {
        throw(pkg_warning(
          "Failed to remove installed package at {.path {move_to}}.",
          .data = list(package = pkg_name),
          .class = "install_filesystem_warning"
        ))
      }
    } else {
      # On Unix we are fine with just deleting the old package
      ret <- unlink(installed_path, recursive = TRUE, force = TRUE)
      if (ret != 0) {
        throw(pkg_warning(
          "Failed to remove installed package at {.path {installed_path}}.",
          .data = list(package = pkg_name),
          .class = "install_filesystem_warning"
        ))
      }
    }
  }
  ret <- file.rename(pkg$path, installed_path)
  if (!ret) {
    throw(pkg_error(
      "Unable to move package from {.path {pkg$path}} to
      {.path {installed_path}}",
      .data = list(package = pkg_name),
      .class = "install_filesystem_error"
    ))
  }

  installed_path
}

#' @importFrom utils modifyList
add_metadata <- function(pkg_path, metadata) {
  if (!length(metadata)) {
    return()
  }

  ## During installation, the DESCRIPTION file is read and an package.rds
  ## file created with most of the information from the DESCRIPTION file.
  ## Functions that read package metadata may use either the DESCRIPTION
  ## file or the package.rds file, therefore we attempt to modify both of
  ## them, and return an error if neither one exists.

  source_desc <- file.path(pkg_path, "DESCRIPTION")
  binary_desc <- file.path(pkg_path, "Meta", "package.rds")
  if (file.exists(source_desc)) {
    do.call(
      desc::desc_set,
      c(as.list(metadata), list(file = source_desc, check = FALSE))
    )
  }

  if (file.exists(binary_desc)) {
    pkg_desc <- base::readRDS(binary_desc)
    desc <- as.list(pkg_desc$DESCRIPTION)
    desc <- modifyList(desc, as.list(metadata))
    pkg_desc$DESCRIPTION <- stats::setNames(as.character(desc), names(desc))
    base::saveRDS(pkg_desc, binary_desc)
  }

  if (!file.exists(source_desc) && !file.exists(binary_desc)) {
    throw(pkg_error(
      "Could not find {.file DESCRIPTION} file when installing package
       into {.path {pkg_path}}.",
      i = msg_internal_error()
    ))
  }

  md5 <- file.path(pkg_path, "MD5")
  if (file.exists(md5)) {
    lines <- readLines(md5)
    f1 <- grep("^[a-fA-F0-9]*[ |*][ |*]*DESCRIPTION", lines)[1]
    f2 <- grep("^[a-fA-F0-9]*[ |*][ |*]*Meta[/\\\\]package.rds", lines)[1]
    if (!is.na(f1)) {
      h1 <- safe_md5sum(source_desc)
      lines[f1] <- sub("^[a-fA-F0-9]*", h1, lines[f1])
    }
    if (!is.na(f2)) {
      h2 <- safe_md5sum(binary_desc)
      lines[f2] <- sub("[a-fA-F0-9]*", h2, lines[f2])
    }
    if (!is.na(f1) || !is.na(f2)) cat(lines, file = md5, sep = "\n")
  }
}

make_install_process <- function(
  filename,
  lib = .libPaths()[[1L]],
  metadata = NULL
) {
  filename
  lib
  metadata

  now <- Sys.time()

  type <- detect_package_archive_type(filename)
  if (type == "unknown") {
    throw(pkg_error(
      "Cannot extract {.path {filename}}, unknown archive type.",
      .class = "install_input_error"
    ))
  }

  lib_cache <- library_cache(lib)
  mkdirp(pkg_cache <- tempfile(tmpdir = lib_cache))

  ppfun <- function() {
    install_extracted_binary(filename, lib_cache, pkg_cache, lib, metadata, now)
  }

  p <- if (type == "zip") {
    make_unzip_process(filename, exdir = pkg_cache, post_process = ppfun)
  } else {
    ## TODO: we already know the package type, no need to detect again
    make_untar_process(filename, exdir = pkg_cache, post_process = ppfun)
  }

  reg.finalizer(
    p,
    function(...) unlink(pkg_cache, recursive = TRUE),
    onexit = TRUE
  )

  p
}

test_that("make_install_process error", {
  tmp <- tempfile()
  on.exit(unlink(tmp), add = TRUE)
  cat("foobar\n", file = tmp)

  expect_error(
    make_install_process(tmp, lib = tempdir()),
    "Cannot extract",
    class = "install_input_error"
  )
})

test_that("install_extracted_binary, add_metadata", {
  pkg <- binary_test_package("foo")
  type <- detect_package_archive_type(pkg)
  pkg_cache <- withr::local_tempdir()

  p <- if (type == "zip") {
    make_unzip_process(pkg, exdir = pkg_cache)
  } else {
    make_untar_process(pkg, exdir = pkg_cache)
  }

  p$wait(50000)
  p$kill()
  expect_snapshot(dir(pkg_cache))

  lib <- withr::local_tempdir()
  mkdirp(lib_cache <- file.path(lib, "_cache"))
  install_extracted_binary(
    basename(pkg),
    lib_cache,
    pkg_cache,
    lib,
    metadata = c(Foo = "bar", Foobar = "baz")
  )

  x <- callr::r(
    function(l) {
      library("foo", lib.loc = l)
      foo::foo()
    },
    list(lib)
  )
  expect_null(x)

  expect_snapshot({
    d <- desc::desc(file.path(lib, "foo"))
    d$get_field("Package")
    d$get_field("Foo")
    d$get_field("Foobar")
  })

  rds <- readRDS(file.path(lib, "foo", "Meta", "package.rds"))
  expect_snapshot(rds$DESCRIPTION[c("Package", "Foo", "Foobar")])

  # add more metadata
  add_metadata(
    file.path(lib, "foo"),
    c(Another = "field")
  )

  expect_snapshot({
    d <- desc::desc(file.path(lib, "foo"))
    d$get_field("Package")
    d$get_field("Another")
  })

  rds <- readRDS(file.path(lib, "foo", "Meta", "package.rds"))
  expect_snapshot(rds$DESCRIPTION[c("Package", "Another")])

  # install again, to test overwriting. we need to extract again,
  # because it was probably moved
  p <- if (type == "zip") {
    make_unzip_process(pkg, exdir = pkg_cache)
  } else {
    make_untar_process(pkg, exdir = pkg_cache)
  }

  p$wait(50000)
  p$kill()
  expect_snapshot(dir(pkg_cache))

  install_extracted_binary(
    basename(pkg),
    lib_cache,
    pkg_cache,
    lib,
    metadata = c(Foo = "bar2", Foobar = "baz2")
  )

  expect_snapshot({
    d <- desc::desc(file.path(lib, "foo"))
    d$get_field("Package")
    d$get_field("Foo")
    d$get_field("Foobar")
  })

  rds <- readRDS(file.path(lib, "foo", "Meta", "package.rds"))
  expect_snapshot(rds$DESCRIPTION[c("Package", "Foo", "Foobar")])

  # check that MD5 sums were updated on Windows
  if (is_windows()) {
    tools::checkMD5sums(dir = file.path(lib, "foo"))
  }
})

test_that("add_metadata error", {
  tmp <- withr::local_tempdir()
  expect_snapshot(
    error = TRUE,
    transform = transform_tempdir,
    add_metadata(tmp, c(foo = "bar"))
  )
})

test_that("failing to move and/or remove existing installation", {
  # this is challenging to test, but we should do it, nevertheless
  skip("TODO")
})

## ------------------------------------------------------------------------
## API

parse_remote_standard <- function(specs, config, ...) {
  ## This is the same as CRAN, but possibly with standard::
  parsed_specs <- re_match(specs, standard_rx())
  parsed_specs$ref <- parsed_specs$.text
  cn <- setdiff(colnames(parsed_specs), c(".match", ".text"))
  parsed_specs <- parsed_specs[, cn]
  eq <- parsed_specs$atleast == "" & parsed_specs$version != ""
  parsed_specs$atleast[eq] <- "=="
  parsed_specs$type <- "standard"
  lapply(
    seq_len(nrow(parsed_specs)),
    function(i) as.list(parsed_specs[i, ])
  )
}

resolve_remote_standard <- function(
  remote,
  direct,
  config,
  cache,
  dependencies,
  ...
) {
  force(remote)
  force(direct)
  force(dependencies)
  versions <- if ("type" %in% names(remote)) {
    remote$version
  } else {
    vcapply(remote, "[[", "version") # nocov
  }

  if (all(versions %in% c("", "current"))) {
    resolve_from_metadata(remote, direct, config, cache, dependencies)
  } else {
    type_cran_resolve_version(remote, direct, config, cache, dependencies) # nocov
  }
}

download_remote_standard <- function(
  resolution,
  target,
  target_tree,
  config,
  cache,
  which,
  on_progress
) {
  # This is slightly different for cran and bioc packages.
  rptp <- resolution$repotype
  if (identical(rptp, "cran")) {
    download_remote_cran(
      resolution,
      target,
      target_tree,
      config,
      cache,
      which,
      on_progress
    )
  } else if (identical(rptp, "bioc")) {
    download_remote_bioc(
      resolution,
      target,
      target_tree,
      config,
      cache,
      which,
      on_progress
    )
  } else {
    # this will always ping in practice, because we only have a sha in
    # the metadata for CRAN currently
    download_ping_if_no_sha(resolution, target, config, cache, on_progress)
  }
}

satisfy_remote_standard <- function(resolution, candidate, config, ...) {
  ## A standard ref is special, in that any ref source can satisfy it,
  ## as long as the package name is the same, and the version
  ## requirements are satisfied.

  ## 1. package name must be the same
  if (resolution$package != candidate$package) {
    return(structure(FALSE, reason = "Package names differ"))
  }

  ## 2. if this is a direct ref, then it has to be a CRAN or
  ## bioc package. If the candidate is an installed package, we
  ## need to check where it was installed from.
  ## Also, an installed package is only accepted if it is not older,
  ## for direct refs.
  if (resolution$direct) {
    if (candidate$type == "installed") {
      type <- candidate$extra[[1]][["repotype"]] %||% "unknown"
      if (is.na(type)) {
        type <- "unknown"
      }
      remotetype <- candidate$extra[[1]][["remotetype"]] %||% "unknown"
      if (is.na(remotetype)) remotetype <- "unknown"
    } else {
      type <- candidate$type
      remotetype <- "unknown"
    }
    if (!type %in% c("cran", "bioc", "standard") && remotetype != "standard") {
      return(structure(FALSE, reason = "User requested CRAN package"))
    }
    if (
      candidate$type == "installed" &&
        package_version(resolution$version) > candidate$version
    ) {
      return(structure(FALSE, reason = "Direct ref needs update"))
    }
  }

  ## 3. version requirements must be satisfied
  version <- tryCatch(resolution$remote[[1]]$version, error = function(e) "")
  if (version == "") {
    return(TRUE)
  }

  if (
    !version_satisfies(
      candidate$version,
      resolution$remote[[1]]$atleast,
      version
    )
  ) {
    return(structure(FALSE, reason = "Insufficient version"))
  }

  TRUE
}

installedok_remote_standard <- function(installed, solution, config, ...) {
  if (identical(solution$repotype, "cran")) {
    installedok_remote_cran(installed, solution, config, ...)
  } else if (identical(solution$repotype, "bioc")) {
    installedok_remote_bioc(installed, solution, config, ...)
  } else if (solution$platform != "source") {
    # the repo must match
    identical(installed$package, solution$package) &&
      identical(installed$version, solution$version) &&
      installedok_remote_standard_platform(installed, solution, config, ...) &&
      identical(installed$remoterepos, solution$metadata[[1]][["RemoteRepos"]])
  } else {
    # source package from unknown repo
    identical(installed$package, solution$package) &&
      identical(installed$version, solution$version) &&
      identical(installed$remoterepos, solution$metadata[[1]][["RemoteRepos"]])
  }
}

installedok_remote_standard_platform <- function(
  installed,
  solution,
  config,
  ...
) {
  if (identical(installed[["platform"]], solution[["platform"]])) {
    return(TRUE)
  }
  if (identical(installed[["platform"]], "*")) {
    return(TRUE)
  }
  if (identical(installed$remotepkgplatform, solution$platform)) {
    return(TRUE)
  }
  if (
    is_string(solution$platform) &&
      is_string(installed$platform) &&
      startsWith(solution$platform, installed$platform)
  ) {
    return(TRUE)
  }
  FALSE
}

options(asciicast_timeout = 600)
if (exists(".knitr_asciicast_process", envir = .GlobalEnv)) {
  rm(list = ".knitr_asciicast_process", envir = .GlobalEnv)
}

asciicast::init_knitr_engine(
  echo = TRUE,
  echo_input = FALSE,
  interactive = FALSE,
  timeout = as.integer(Sys.getenv("ASCIICAST_TIMEOUT", 300)),
  startup = quote({
    options(width = 72)
    options(cli.width = 72)
    options(cli.progress_show_after = 0)
    options(cli.progress_clear = FALSE)
    library(cli)
    pkgload::load_all()
    set.seed(1)
  })
)

knitr::opts_chunk$set(
  asciicast_knitr_output = "html",
  asciicast_include_style = FALSE,
  asciicast_theme = "pkgdown",
  asciicast_width = 72,
  asciicast_cols = 72
)

list(
  markdown = TRUE,
  knitr_chunk_options = list(
    cache = TRUE,
    cache_lazy = FALSE,
    cache.path = file.path(getwd(), "man/_cache/"),
    fig.path = file.path(getwd(), "man/figures"),
    error = TRUE
  ),
  restrict_image_formats = TRUE
)

progress_chars <- function() {
  if (cli::is_utf8_output()) {
    list(
      build = "\U0001f4e6",
      inst = "\u2705",
      lpar = "\u2e28",
      rpar = "\u2e29",
      fill = "\u2588",
      half = "\u2592"
    )
  } else {
    list(
      build = "[BLD]",
      inst = "[INS]",
      lpar = "(",
      rpar = ")",
      fill = "#",
      half = "-"
    )
  }
}

res__create_progress_bar <- function(self, private) {
  self
  private

  if (!should_show_progress_bar()) {
    return()
  }

  bar <- new.env(parent = emptyenv())
  bar$spinner <- cli::get_spinner()
  bar$spinner_state <- 1L
  bar$chars <- progress_chars()

  bar$status <- cli::cli_status("", .auto_close = FALSE)

  bar$timer <- new_async_timer(
    1 / 10,
    function() res__show_progress_bar(self, private)
  )
  bar$timer$listen_on("error", function(e) {
    stop(e)
  })

  bar
}

res__show_progress_bar <- function(self, private) {
  # Maybe progress is off
  if (is.null(private$bar)) {
    return()
  }

  # This can be called _after_ the resolution is over
  if (is.null(private$bar$status)) {
    return()
  }

  deps <- nrow(private$state)
  direct <- sum(private$state$direct)
  direct_done <- sum(!is.na(private$state$status) & private$state$direct)

  bar <- if (direct >= 5) {
    make_bar(private$bar$chars, direct_done / direct, width = 15)
  } else {
    ""
  }

  state <- make_progress_main(deps, direct_done, direct)
  spinner <- make_progress_spinner(self, private)
  msg <- make_trailing_progress_msg(self, private)

  # TODO: check width
  str <- "{bar} {state} {spinner} {msg}"
  cli::cli_status_update(private$bar$status, str)
}

make_bar <- function(chars, p, width = 15) {
  width <- width - 2L

  w <- if (isTRUE(all.equal(p, 1))) width else trunc(width * p)

  pchars <- rep(chars$fill, w)
  xchars <- rep(" ", max(width - w, 0))
  bar <- paste(
    c(chars$lpar, pchars, xchars, chars$rpar),
    collapse = ""
  )

  if (is_older_rstudio()) bar else cli::col_green(bar)
}

make_progress_main <- function(deps, done, total) {
  if (is_older_rstudio()) {
    bggrey <- fgdark <- function(x) x
  } else {
    bggrey <- cli::make_ansi_style("grey", bg = TRUE)
    fgdark <- cli::col_black
  }
  paste0(
    "Found ",
    bggrey(fgdark(paste0(" ", deps, " "))),
    " deps for ",
    bggrey(fgdark(paste0(" ", done, "/", total, " "))),
    " pkgs"
  )
}

make_progress_spinner <- function(self, private) {
  bar <- private$bar
  spin <- bar$spinner$frames[[bar$spinner_state]]
  bar$spinner_state <-
    bar$spinner_state %% length(bar$spinner$frames) + 1L
  private$bar <- bar
  paste0("[", spin, "]")
}

make_trailing_progress_msg <- function(self, private) {
  ongoing <- private$state[is.na(private$state$status), ]
  if (nrow(ongoing) == 0) {
    return("Done")
  }

  types <- vcapply(ongoing$remote, "[[", "type")
  remote <- if (all(types %in% c("cran", "bioc", "standard"))) {
    ongoing$remote[[order(ongoing$started_at)[1]]]
  } else {
    nonstd <- ongoing[!types %in% c("cran", "bioc", "special"), ]
    nonstd$remote[[order(nonstd$started_at)[1]]]
  }

  if (remote$type %in% c("cran", "bioc", "standard")) {
    "Resolving standard (CRAN/BioC) packages"
  } else if (remote$type %in% "installed") {
    "Checking installed packages"
  } else {
    paste0("Resolving ", remote$ref)
  }
}

res__done_progress_bar <- function(self, private) {
  if (is.null(private$bar)) {
    return()
  }
  cli::cli_status_clear(private$bar$status, result = "clear")
  private$bar <- NULL
}

#' @param refs Package names or references. See
#'   ['Package references'][pkg_refs] for the syntax.
#' @param ... Additional arguments, passed to
#'   [`pkg_deps$new()`](#method-new).
#' @return `new_pkg_deps()` returns a new `pkg_deps` object.
#'
#' @details
#' `new_pkg_deps()` creates a new object from the `pkg_deps` class.
#' The advantage of `new_pkg_deps()` compared to using the [pkg_deps]
#' constructor directly is that it avoids making pkgdepends a build time
#' dependency.
#'
#' @export
#' @rdname pkg_deps

new_pkg_deps <- function(refs, ...) {
  pkg_deps$new(refs, ...)
}

#' R6 class for package dependency lookup
#'
#' @description
#' Look up dependencies of R packages from various sources.
#'
#' @details
#' The usual steps to query package dependencies are:
#'
#' 1. Create a `pkg_deps` object with `new_pkg_deps()`.
#' 1. Resolve all possible dependencies with
#'    [`pkg_deps$resolve()`](#method-resolve).
#' 1. Solve the dependencies, to obtain a subset of all possible
#'    dependencies that can be installed together, with
#'    [`pkg_deps$solve()`](#method-solve).
#' 1. Call [`pkg_deps$get_solution()`](#method-get-solution) to list the
#'    result of the dependency solver.
#'
#' @export

pkg_deps <- R6::R6Class(
  "pkg_deps",
  public = list(
    #' @description
    #' Create a new `pkg_deps` object. Consider using `new_pkg_deps()`
    #' instead of calling the constructor directly.
    #'
    #' The returned object can be used to look up (recursive) dependencies
    #' of R packages from various sources. To perform the actual lookup,
    #' you'll need to call the [`resolve()`](#method-resolve) method.
    #' @param refs Package names or references. See
    #'   ['Package references'][pkg_refs] for the syntax.
    #' @param config Configuration options, a named list. See
    #'   ['Configuration'][pkgdepends-config].
    #' @param policy Solution policy. See ['The dependency
    #'   solver'][pkg_solution].
    #' @param remote_types Custom remote ref types, this is for advanced
    #'   use, and experimental currently.
    #'
    #' @return
    #'  A new `pkg_deps` object.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method initialize()
    #' pd <- pkg_deps$new("r-lib/pkgdepends")
    #' pd

    initialize = function(
      refs,
      config = list(),
      policy = c("lazy", "upgrade"),
      remote_types = NULL
    ) {
      private$library <- tempfile()
      config$sysreqs_lookup_system <- FALSE
      policy <- match.arg(policy)
      private$policy <- policy
      dir.create(private$library)
      private$plan <- pkg_plan$new(
        refs,
        config,
        library = private$library,
        remote_types
      )
    },

    #' @description
    #' The package refs that were used to create the `pkg_deps` object.
    #'
    #' @return
    #' A character vector of package refs that were used to create the
    #' `pkg_deps` object.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method get_refs()
    #' pd <- new_pkg_deps(c("pak", "jsonlite"))
    #' pd$get_refs()

    get_refs = function() private$plan$get_refs(),

    #' @description
    #' Configuration options for the `pkg_deps` object. See
    #' ['Configuration'][pkgdepends-config] for details.
    #'
    #' @return
    #' See ['Configuration'][pkgdepends-config] for the configuration entries.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method get_config()
    #' pd <- new_pkg_deps("pak")
    #' pd$get_config()

    get_config = function() private$plan$get_config(),

    #' @description
    #' Resolve the dependencies of the specified package references. This
    #' usually means downloading metadata from CRAN and Bioconductor, unless
    #' already cached, and also from GitHub if GitHub refs were included,
    #' either directly or indirectly. See
    #' ['Dependency resolution'][pkg_resolution] for details.
    #'
    #'@return
    #' The `pkg_deps` object itself, invisibly.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method resolve()
    #' pd <- new_pkg_deps("pak")
    #' pd$resolve()
    #' pd$get_resolution()

    resolve = function() {
      private$plan$resolve()
      invisible(self)
    },

    #' @description
    #' The same as [`resolve()`](#method-resolve), but asynchronous.
    #' This method is for advanced use.
    #'
    #' @return
    #' A deferred value.

    async_resolve = function() private$plan$async_resolve(),

    #' @description
    #' Query the result of the dependency resolution. This method can be
    #' called after [`resolve()`](#method-resolve) has completed.
    #'
    #' @return
    #' A [pkg_resolution_result] object, which is also a data frame. See
    #' ['Dependency resolution'][pkg_resolution] for its columns.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method get_resolution()
    #' pd <- new_pkg_deps("r-lib/pkgdepends")
    #' pd$resolve()
    #' pd$get_resolution()

    get_resolution = function() private$plan$get_resolution(),

    #' @description
    #' Returns the current policy of the dependency solver.
    #' See ['The dependency solver'][pkg_solution] for details.
    #'
    #' @return
    #' A character vector of length one.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method get_solve_policy()
    #' pdi <- new_pkg_deps("r-lib/pkgdepends")
    #' pdi$get_solve_policy()
    #' pdi$set_solve_policy("upgrade")
    #' pdi$get_solve_policy()

    get_solve_policy = function() private$policy,

    #' @description
    #' Set the current policy of the dependency solver.
    #' If the object already contains a solution and the new policy is
    #' different than the old policy, then the solution is deleted.
    #' See ['The dependency solver'][pkg_solution] for details.
    #'
    #' @param policy Policy to set.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method set_solve_policy()
    #' pdi <- new_pkg_deps("r-lib/pkgdepends")
    #' pdi$get_solve_policy()
    #' pdi$set_solve_policy("upgrade")
    #' pdi$get_solve_policy()

    set_solve_policy = function(policy = c("lazy", "upgrade")) {
      policy <- match.arg(policy)
      if (private$policy != policy) {
        private$plan$delete_solution()
        private$policy <- policy
      }
    },
    #' @description
    #' Solve the package dependencies. Out of the resolved dependencies, it
    #' works out a set of packages, that can be installed together to
    #' create a functional installation. The set includes all directly
    #' specified packages, and all required (or suggested, depending on
    #' the configuration) packages as well. It includes every package at
    #' most once. See ['The dependency solver'][pkg_solution] for details.
    #'
    #' `solve()` calls [`resolve()`](#method-resolve) automatically, if it
    #' hasn't been called yet.
    #'
    #' @return
    #' The `pkg_deps` object itself, invisibly.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method solve()
    #' pd <- new_pkg_deps("r-lib/pkgdepends")
    #' pd$resolve()
    #' pd$solve()
    #' pd$get_solution()

    solve = function() {
      private$plan$solve(policy = private$policy)
      invisible(self)
    },

    #' @description
    #' Returns the solution of the package dependencies.
    #' @return
    #' A [pkg_solution_result] object, which is a list. See
    #' [pkg_solution_result] for details.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method get_solution()
    #' pd <- new_pkg_deps("pkgload")
    #' pd$resolve()
    #' pd$solve()
    #' pd$get_solution()

    get_solution = function() private$plan$get_solution(),

    #' @description
    #' Error if the dependency solver failed to find a consistent set of
    #' packages that can be installed together.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method stop_for_solution_error()
    #' # This is an error, because the packages conflict:
    #' pd <- new_pkg_deps(
    #'   c("r-lib/pak", "cran::pak"),
    #'   config = list(library = tempfile())
    #' )
    #' pd$resolve()
    #' pd$solve()
    #' pd
    #' # This fails:
    #' # pd$stop_for_solution_error()

    stop_for_solution_error = function() {
      private$plan$stop_for_solve_error()
    },

    #' @description
    #' Draw a tree of package dependencies. It returns a `tree` object, see
    #' [cli::tree()]. Printing this object prints the dependency tree to the
    #' screen.
    #'
    #' @return
    #' A `tree` object from the cli package, see [cli::tree()].
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method draw()
    #' pd <- new_pkg_deps("pkgload")
    #' pd$solve()
    #' pd$draw()

    draw = function() private$plan$draw_solution_tree(),

    #' @description
    #' Format a `pkg_deps` object, typically for printing.
    #'
    #' @param ... Not used currently.
    #'
    #' @return
    #' A character vector, each element should be a line in the printout.

    format = function(...) {
      refs <- private$plan$get_refs()

      has_res <- private$plan$has_resolution()
      res <- if (has_res) private$plan$get_resolution()
      res_err <- has_res && any(res$status != "OK")

      has_sol <- private$plan$has_solution()
      sol <- if (has_sol) private$plan$get_solution()
      sol_err <- has_sol && sol$status != "OK"

      deps <- if (has_res) length(unique(res$package[!res$direct]))

      c(
        "<pkg_dependencies>",
        "+ refs:",
        paste0("  - ", refs),
        if (has_res) paste0("+ has resolution (+", deps, " dependencies)"),
        if (res_err) "x has resolution errors",
        if (has_sol) "+ has solution",
        if (sol_err) "x has solution errors",
        if (!has_res) "(use `$resolve()` to resolve dependencies)",
        if (has_res) "(use `$get_resolution()` to see resolution results)",
        if (!has_sol) "(use `$solve()` to solve dependencies)",
        if (has_sol) "(use `$show_solution()` to see the dependencies",
        if (has_sol) "(use `$get_solution()` to see the full solution results)",
        if (has_sol && !sol_err) "(use `$draw()` to draw the dependency tree)"
      )
    },

    #' @description
    #' Prints a `pkg_deps` object to the screen. The printout includes:
    #'
    #' * The package refs.
    #' * Whether the object has the resolved dependencies.
    #' * Whether the resolution had errors.
    #' * Whether the object has the solved dependencies.
    #' * Whether the solution had errors.
    #' * Advice on which methods to call next.
    #'
    #' See the example below.
    #'
    #' @param ... not used currently.
    #' @return
    #' The `pkg_deps` object itself, invisibly.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method print()
    #' pd <- new_pkg_deps("r-lib/pkgdepends")
    #' pd
    #'
    #' pd$resolve()
    #' pd
    #'
    #' pd$solve()
    #' pd

    print = function(...) {
      cat(self$format(...), sep = "\n")
      invisible(self)
    }
  ),

  private = list(
    plan = NULL,
    policy = NULL,
    library = NULL
  )
)

test_that("new_pkg_deps", {
  pkgcache::pkg_cache_delete_files()
  setup_fake_apps()

  deps <- new_pkg_deps("pkg3", config = list(library = tempfile()))
  expect_snapshot(deps)
  expect_snapshot(deps$get_refs())
  expect_snapshot(sort(deps$get_config()$list()))

  suppressMessages(deps$resolve())
  expect_snapshot(deps)

  suppressMessages(deps$solve())
  expect_snapshot(deps)

  expect_snapshot(
    deps$draw(),
    transform = transform_bytes
  )

  expect_snapshot(
    deps$get_solution()
  )
})

test_that("async", {
  setup_fake_apps()

  deps <- new_pkg_deps("pkg3", config = list(library = tempfile()))
  expect_snapshot(
    synchronize(
      deps$async_resolve()$then(function() "done")
    )
  )
  expect_snapshot(
    deps$get_resolution()[, c("ref", "type", "directpkg", "package", "error")]
  )
})

test_that("solve policy", {
  setup_fake_apps()

  deps <- new_pkg_deps("pkg3", config = list(library = tempfile()))
  deps$get_solve_policy()
  deps$set_solve_policy("upgrade")
  suppressMessages(deps$solve())
  expect_snapshot(deps)
})

test_that("errors", {
  setup_fake_apps()

  deps <- new_pkg_deps("needsfuturama", config = list(library = tempfile()))
  expect_silent(deps$solve())
  expect_snapshot(
    error = TRUE,
    deps$stop_for_solution_error()
  )
})

# There is no easy way to update the database in yum/dnf/zypper.
# But FIXME. We'll need to solve this issue when (and if) we'll check the
# installed system packages, and only install/update the ones that are
# missing or out of date. For now this is OK, the first `install` command
# will update the DB (cache it is called I believe), and the rest will
# use it.

# TODO: get this data from the DB

sysreqs2_cmds <- utils::read.table(
  stringsAsFactors = FALSE,
  header = TRUE,
  textConnection(
    "
   name                       os      id                  distribution version  version_match update_command      install_command                     query_command
   'Ubuntu Linux'             linux   ubuntu              ubuntu       *        NA            'apt-get -y update' 'apt-get -y install'                dpkg-query
   'Debian Linux'             linux   debian              debian       *        NA            'apt-get -y update' 'apt-get -y install'                dpkg-query
   'CentOS Linux'             linux   centos              centos       *        major          NA                 'yum install -y'                    rpm
   'Rocky Linux'              linux   rocky               rockylinux   *        major          NA                 'dnf install -y'                    rpm
   'Rocky Linux'              linux   rockylinux          rockylinux   *        major          NA                 'dnf install -y'                    rpm
   'AlmaLinux'                linux   almalinux           rockylinux   *        major          NA                 'dnf install -y'                    rpm
   'Red Hat Enterprise Linux' linux   rhel                redhat       6        major          NA                 'yum install -y'                    rpm
   'Red Hat Enterprise Linux' linux   rhel                redhat       7        major          NA                 'yum install -y'                    rpm
   'Red Hat Enterprise Linux' linux   rhel                redhat       *        major          NA                 'dnf install -y'                    rpm
   'Red Hat Enterprise Linux' linux   redhat              redhat       6        major          NA                 'yum install -y'                    rpm
   'Red Hat Enterprise Linux' linux   redhat              redhat       7        major          NA                 'yum install -y'                    rpm
   'Red Hat Enterprise Linux' linux   redhat              redhat       *        major          NA                 'dnf install -y'                    rpm
   'Fedora Linux'             linux   fedora              fedora       *        NA             NA                 'dnf install -y'                    rpm
   'openSUSE Linux'           linux   opensuse            opensuse     *        NA             NA                 'zypper --non-interactive install'  rpm
   'openSUSE Linux'           linux   opensuse-leap       opensuse     *        NA             NA                 'zypper --non-interactive install'  rpm
   'openSUSE Linux'           linux   opensuse-tumbleweed opensuse     *        NA             NA                 'zypper --non-interactive install'  rpm
   'SUSE Linux Enterprise'    linux   sles                sle          *        NA             NA                 'zypper --non-interactive install'  rpm
   'SUSE Linux Enterprise'    linux   sle                 sle          *        NA             NA                 'zypper --non-interactive install'  rpm
   'Alpine Linux'             linux   alpine              alpine       *        minor          NA                 'apk add --no-cache'                apk
"
  )
)

# do not use package_version, in case the distro is nor semver
get_major_version <- function(x) {
  sub("[.].*$", "", x)
}
get_minor_version <- function(x) {
  sub("([.][^.]+)[.].*$", "\\1", x)
}

find_sysreqs_platform <- function(sysreqs_platform = NULL, parsed = NULL) {
  plt <- parsed %||%
    parse_sysreqs_platform(
      sysreqs_platform %||% current_config()$get("sysreqs_platform")
    )
  plt$version_major <- get_major_version(plt$version)
  plt$version_minor <- get_minor_version(plt$version)
  which(
    sysreqs2_cmds$os == plt$os &
      sysreqs2_cmds$id == plt$distribution &
      (sysreqs2_cmds$version %in%
        c("*", plt$version) |
        (sysreqs2_cmds$version_match == "major" &
          sysreqs2_cmds$version == plt$version_major) |
        (sysreqs2_cmds$version_match == "minor" &
          sysreqs2_cmds$version == plt$version_minor))
  )[1]
}

canonize_sysreqs_platform <- function(sysreqs_platform) {
  parsed <- parse_sysreqs_platform(sysreqs_platform)
  known <- find_sysreqs_platform(parsed = parsed)
  if (is.na(known)) {
    sysreqs_platform
  } else {
    plt <- sysreqs_platforms()[known, ]
    paste0(
      plt$distribution,
      "-",
      if (plt$version != "*") {
        plt$version
      } else if (identical(plt$version_match, "major")) {
        get_major_version(parsed$version)
      } else if (identical(plt$version_match, "minor")) {
        get_minor_version(parsed$version)
      } else {
        parsed$version
      }
    )
  }
}

sysreqs2_command <- function(
  sysreqs_platform = NULL,
  cmd = c("install_command", "update_command", "query_command")
) {
  cmd <- match.arg(cmd)
  sel <- find_sysreqs_platform(sysreqs_platform)
  if (is.na(sel)) {
    throw(pkg_error(paste0(
      "Unknown OS. Don't know how to query or install system packages for ",
      sysreqs_platform,
      "."
    )))
  }

  sysreqs2_cmds[[cmd]][sel]
}

sysreqs2_resolve <- function(
  sysreqs,
  sysreqs_platform = NULL,
  config = NULL,
  ...
) {
  synchronize(sysreqs2_async_resolve(sysreqs, sysreqs_platform, config, ...))
}

sysreqs2_async_resolve <- function(sysreqs, sysreqs_platform, config, ...) {
  sysreqs
  sysreqs_platform
  config
  list(...)

  config <- config %||% current_config()
  sysreqs_platform <- sysreqs_platform %||% config$get("sysreqs_platform")

  sysreqs2_async_update_metadata(config = config)$then(function() {
    sysreqs2_match(
      sysreqs,
      sysreqs_platform = sysreqs_platform,
      config = config,
      ...
    )
  })$then(function(recs) {
    sysreqs2_scripts(recs, sysreqs_platform)
  })
}

sysreqs2_scripts <- function(recs, sysreqs_platform, missing = FALSE) {
  sysreqs_platform <- canonize_sysreqs_platform(sysreqs_platform)
  plt <- parse_sysreqs_platform(sysreqs_platform)
  flatrecs <- unlist(recs, recursive = FALSE)
  upd <- sysreqs2_command(sysreqs_platform, "update")
  pre <- unlist(lapply(flatrecs, "[[", "pre_install"))
  post <- unlist(lapply(flatrecs, "[[", "post_install"))
  if (is.na(upd)) {
    upd <- character()
  }
  cmd <- sysreqs2_command(sysreqs_platform, "install")
  allpkgs <- unique(unlist(lapply(flatrecs, "[[", "packages")))
  misspkgs <- unique(unlist(lapply(flatrecs, function(x) {
    if ("packages_missing" %in% names(x)) {
      x$packages_missing
    } else {
      x$packages
    }
  })))
  pkgs <- if (missing) misspkgs else allpkgs
  ipkgs <- if (length(pkgs)) paste(pkgs, collapse = " ") else character()
  ipkgs <- if (length(ipkgs)) paste(cmd, ipkgs)
  # no need to update if nothing to do
  if (length(pre) + length(ipkgs) + length(post) == 0) {
    upd <- character()
  }
  res <- list(
    os = plt$os,
    distribution = plt$distribution,
    version = plt$version,
    url = NA_character_,
    pre_install = c(upd, pre),
    install_scripts = ipkgs,
    post_install = post,
    packages = allpkgs
  )
  if (missing) {
    res$misspkgs <- misspkgs
  }
  res
}

sysreqs2_git_repo <- function() {
  list(
    repo = Sys.getenv(
      "R_PKG_SYSREQS_GIT_REPO_URL",
      "https://github.com/r-hub/r-system-requirements.git"
    ),
    ref = Sys.getenv(
      "R_PKG_SYSREQS_GIT_REPO_REF",
      "HEAD"
    )
  )
}

sysreqs2_update_metadata <- function(path = NULL, config = NULL) {
  synchronize(sysreqs2_async_update_metadata(path = path, config = config))
}

# Increase this if the syntax breaks or if we need more / different files
# from the r-system-requirements repo.
sysreqs_db_version <- "1"

sysreqs2_async_update_metadata <- function(path = NULL, config = NULL) {
  config <- config %||% current_config()
  if (!config$get("sysreqs_db_update")) {
    return(async_constant())
  }

  path <- path %||% file.path(get_user_cache_dir()$root, "sysreqs")
  head_file <- file.path(path, "HEAD")
  ver_file <- file.path(path, "VERSION")

  if (file.exists(head_file) && file.exists(ver_file)) {
    mt <- file.mtime(head_file)
    ver <- readLines(ver_file)
    upd <- config$get("metadata_update_after")
    if (ver == sysreqs_db_version && !is.na(mt) && Sys.time() - mt < upd) {
      return(async_constant())
    }
  }

  upd <- function() {
    head <- if (file.exists(head_file)) readLines(head_file)[1] else ""
    repo <- sysreqs2_git_repo()
    async_git_list_refs_v1(repo$repo)$then(function(refs) {
      rem_head <- refs$refs$hash[refs$refs$ref == "HEAD"]
      if (rem_head == head) {
        # still update the time stamps
        Sys.setFileTime(head_file, Sys.time())
        return()
      }
      tmp <- paste0(path, "-new")
      async_git_download_repo(repo$repo, rem_head, tmp)$then(function() {
        unlink(path, recursive = TRUE, force = TRUE)
        file.rename(tmp, path)
        writeLines(rem_head, head_file)
        writeLines(sysreqs_db_version, ver_file)
      })
    })
  }

  timeout <- as.double(
    config$get("sysreqs_db_update_timeout"),
    units = "secs"
  )
  async_timeout(upd, timeout)$catch(error = function(e) {
    cli::cli_alert_warning(
      # nocov start
      "Failed to update system requirement mappings,
         will use cached mappings.",
      wrap = TRUE
    ) # nocov end
    invisible()
  })
}

sysreqs2_list_rules <- function(path = NULL) {
  if (is.null(path)) {
    cached_path <- file.path(get_user_cache_dir()$root, "sysreqs")
    head_file <- file.path(cached_path, "HEAD")
    ver_file <- file.path(cached_path, "VERSION")
    if (file.exists(ver_file) && readLines(ver_file) == sysreqs_db_version) {
      path <- cached_path
    } else {
      path <- system.file("sysreqs", package = "pkgdepends")
    }
  }
  rules <- dir(
    file.path(path, "rules"),
    pattern = "[.]json$",
    full.names = TRUE
  )
}

sysreqs2_match <- function(
  sysreqs,
  path = NULL,
  sysreqs_platform = NULL,
  config = NULL
) {
  rules <- sysreqs2_list_rules(path)

  result <- structure(
    vector(mode = "list", length(sysreqs)),
    names = names(sysreqs)
  )
  todo <- !is.na(sysreqs) & sysreqs != ""

  config <- config %||% current_config()
  plt <- parse_sysreqs_platform(canonize_sysreqs_platform(
    sysreqs_platform %||% config$get("sysreqs_platform")
  ))

  rsysreqs <- sysreqs[todo]
  for (r in rules) {
    rule <- jsonlite::fromJSON(r, simplifyVector = FALSE)
    pats <- unlist(rule$patterns)
    mch <- vapply(
      pats,
      grepl,
      logical(length(rsysreqs)),
      x = rsysreqs,
      ignore.case = TRUE
    )
    mch <- apply(rbind(mch), 1, any)
    if (!any(mch)) {
      next
    }
    for (dep in rule$dependencies) {
      appl <- FALSE
      for (const in dep$constraints) {
        if (
          identical(const$os, plt$os) &&
            identical(const$distribution, plt$distribution) &&
            (is.null(const$versions) || plt$version %in% const$versions)
        ) {
          appl <- TRUE
          break
        }
      }
      if (!appl) {
        next
      }
      sysreq_name <- tools::file_path_sans_ext(basename(r))
      rec <- list(
        sysreq = sysreq_name,
        packages = unname(unlist(dep$packages)),
        pre_install = unname(unlist(dep$pre_install)),
        post_install = unname(unlist(dep$post_install))
      )
      for (idx in which(mch)) {
        result[todo][[idx]] <- c(result[todo][[idx]], list(rec))
      }
      break
    }
  }

  result
}

sysreqs_update_state <- function(sys, spkgs = NULL) {
  spkgs <- spkgs %||% sysreqs_list_system_packages()
  spkgs <- spkgs[grepl("^.i$", spkgs$status), ]
  allspkgs <- unique(unlist(c(spkgs$package, spkgs$provides)))
  for (i in seq_along(sys)) {
    elt <- sys[[i]]
    for (j in seq_along(elt)) {
      elt[[j]]$packages_missing <- setdiff(elt[[j]]$packages, allspkgs)
    }
    if (!is.null(elt)) sys[[i]] <- elt
  }
  sys
}

test_that("sysreqs_is_supported", {
  expect_true(sysreqs_is_supported("ubuntu-20.04"))
  expect_true(sysreqs_is_supported("ubuntu-22.04"))
  expect_true(sysreqs_is_supported("ubuntu-24.04"))

  expect_true(sysreqs_is_supported("debian-11"))
  expect_true(sysreqs_is_supported("debian-12"))
  expect_true(sysreqs_is_supported("debian-unstable"))

  expect_true(sysreqs_is_supported("opensuse-15.4"))
  expect_true(sysreqs_is_supported("opensuse-15.5"))
  expect_true(sysreqs_is_supported("opensuse-15.6"))
  expect_true(sysreqs_is_supported("opensuse-leap-15.4"))
  expect_true(sysreqs_is_supported("opensuse-leap-15.5"))
  expect_true(sysreqs_is_supported("opensuse-leap-15.6"))
  expect_true(sysreqs_is_supported("opensuse-tumbleweed-20250509"))

  expect_true(sysreqs_is_supported("centos-6"))
  expect_true(sysreqs_is_supported("centos-7"))
  expect_true(sysreqs_is_supported("centos-8"))

  expect_true(sysreqs_is_supported("rhel-7"))
  expect_true(sysreqs_is_supported("rhel-7.9"))
  expect_true(sysreqs_is_supported("rhel-8"))
  expect_true(sysreqs_is_supported("rhel-8.10"))
  expect_true(sysreqs_is_supported("rhel-9"))
  expect_true(sysreqs_is_supported("rhel-9.6"))

  expect_true(sysreqs_is_supported("fedora-39"))
  expect_true(sysreqs_is_supported("fedora-40"))
  expect_true(sysreqs_is_supported("fedora-41"))
  expect_true(sysreqs_is_supported("fedora-42"))
  expect_true(sysreqs_is_supported("fedora-43"))

  expect_true(sysreqs_is_supported("sles-15.4"))
  expect_true(sysreqs_is_supported("sles-15.5"))
  expect_true(sysreqs_is_supported("sles-15.6"))

  expect_true(sysreqs_is_supported("almalinux-8.10"))
  expect_true(sysreqs_is_supported("almalinux-9.6"))

  expect_true(sysreqs_is_supported("rocky-8.9"))
  expect_true(sysreqs_is_supported("rocky-9.3"))

  expect_true(sysreqs_is_supported("alpine-3.19"))
  expect_true(sysreqs_is_supported("alpine-3.20"))
  expect_true(sysreqs_is_supported("alpine-3.21"))
  expect_true(sysreqs_is_supported("alpine-edge"))

  # these should not happen, these are the PPM names
  expect_true(sysreqs_is_supported("redhat-8"))
  expect_true(sysreqs_is_supported("redhat-8.10"))
  expect_true(sysreqs_is_supported("rockylinux-8"))
  expect_true(sysreqs_is_supported("rockylinux-8.9"))

  expect_false(sysreqs_is_supported("foo-bar"))
})

test_that("update", {
  # TODO: needs internet
  skip_on_cran()
  tmp <- withr::local_tempdir()
  synchronize(sysreqs2_async_update_metadata(tmp))
  expect_true(file.exists(tmp))
  expect_true("rules" %in% dir(tmp))

  # no update if already there
  synchronize(sysreqs2_async_update_metadata(tmp))
  expect_true(file.exists(tmp))
  expect_true("rules" %in% dir(tmp))
})

test_that("match", {
  skip_on_cran()
  sr1 <- sysreqs2_resolve("libcurl and Java", "ubuntu-22.04")
  expect_snapshot(sr1)

  sr2 <- sysreqs2_resolve("libcurl and Java", "debian-unstable")
  expect_snapshot(sr2)
})

test_that("sysreqs2_command error", {
  expect_snapshot(
    error = TRUE,
    sysreqs2_command("foobar-2023")
  )
})

test_that("do not run update if nothing to do", {
  skip_on_cran()
  sr1 <- sysreqs2_resolve("nothing needed", "ubuntu-22.04")
  expect_snapshot(sr1)
})

test_that("sysreqs_platforms", {
  skip_on_cran()
  sysreqs <- function(sysreqs_platform) {
    sysreqs_resolve("libxml2", sysreqs_platform)[["install_scripts"]]
  }
  expect_snapshot({
    sysreqs("ubuntu-20.04")
    sysreqs("ubuntu-22.04")
    sysreqs("ubuntu-24.04")

    sysreqs("debian-11")
    sysreqs("debian-12")
    sysreqs("debian-unstable")

    sysreqs("opensuse-15.4")
    sysreqs("opensuse-15.5")
    sysreqs("opensuse-15.6")
    sysreqs("opensuse-leap-15.4")
    sysreqs("opensuse-leap-15.5")
    sysreqs("opensuse-leap-15.6")
    sysreqs("opensuse-tumbleweed-20250509")

    sysreqs("centos-6")
    sysreqs("centos-7")
    sysreqs("centos-8")

    sysreqs("rhel-7")
    sysreqs("rhel-7.9")
    sysreqs("rhel-8")
    sysreqs("rhel-8.10")
    sysreqs("rhel-9")
    sysreqs("rhel-9.6")

    sysreqs("fedora-39")
    sysreqs("fedora-40")
    sysreqs("fedora-42")
    sysreqs("fedora-42")
    sysreqs("fedora-43")

    sysreqs("sles-15.4")
    sysreqs("sles-15.5")
    sysreqs("sles-15.6")

    sysreqs("almalinux-8.10")
    sysreqs("almalinux-9.6")

    sysreqs("rocky-8.9")
    sysreqs("rocky-9.3")

    sysreqs("alpine-3.19")
    sysreqs("alpine-3.20")
    sysreqs("alpine-3.21")
    sysreqs("alpine-3.edge")

    sysreqs("redhat-8")
    sysreqs("redhat-8.10")
    sysreqs("rockylinux-8")
    sysreqs("rockylinux-8.9")
  })
})

test_that("canonize_sysreqs_platform", {
  expect_snapshot({
    canonize_sysreqs_platform("ubuntu-20.04")
    canonize_sysreqs_platform("ubuntu-22.04")
    canonize_sysreqs_platform("ubuntu-24.04")

    canonize_sysreqs_platform("debian-11")
    canonize_sysreqs_platform("debian-12")
    canonize_sysreqs_platform("debian-unstable")

    canonize_sysreqs_platform("opensuse-15.4")
    canonize_sysreqs_platform("opensuse-15.5")
    canonize_sysreqs_platform("opensuse-15.6")
    canonize_sysreqs_platform("opensuse-leap-15.4")
    canonize_sysreqs_platform("opensuse-leap-15.5")
    canonize_sysreqs_platform("opensuse-leap-15.6")
    canonize_sysreqs_platform("opensuse-tumbleweed-20250509")

    canonize_sysreqs_platform("centos-6")
    canonize_sysreqs_platform("centos-7")
    canonize_sysreqs_platform("centos-8")

    canonize_sysreqs_platform("rhel-7")
    canonize_sysreqs_platform("rhel-7.9")
    canonize_sysreqs_platform("rhel-8")
    canonize_sysreqs_platform("rhel-8.10")
    canonize_sysreqs_platform("rhel-9")
    canonize_sysreqs_platform("rhel-9.6")

    canonize_sysreqs_platform("fedora-39")
    canonize_sysreqs_platform("fedora-40")
    canonize_sysreqs_platform("fedora-42")
    canonize_sysreqs_platform("fedora-42")
    canonize_sysreqs_platform("fedora-43")

    canonize_sysreqs_platform("sles-15.4")
    canonize_sysreqs_platform("sles-15.5")
    canonize_sysreqs_platform("sles-15.6")

    canonize_sysreqs_platform("almalinux-8.10")
    canonize_sysreqs_platform("almalinux-9.6")

    canonize_sysreqs_platform("rocky-8.9")
    canonize_sysreqs_platform("rocky-9.3")

    canonize_sysreqs_platform("alpine-3.19")
    canonize_sysreqs_platform("alpine-3.20")
    canonize_sysreqs_platform("alpine-3.21")
    canonize_sysreqs_platform("alpine-3.19.1")
    canonize_sysreqs_platform("alpine-3.20.2")
    canonize_sysreqs_platform("alpine-3.21.3")
    canonize_sysreqs_platform("alpine-3.edge")

    canonize_sysreqs_platform("redhat-8")
    canonize_sysreqs_platform("redhat-8.10")
    canonize_sysreqs_platform("rockylinux-8")
    canonize_sysreqs_platform("rockylinux-8.9")
  })
})

snapshot <- function(x, width = Inf, ...) {
  UseMethod("snapshot")
}

snapshot.default <- function(x, width = Inf, ...) {
  if (width == Inf) {
    width <- 10000
  }
  print(x, width = width, ...)
}

#' @export

print.snapshot <- function(x, ...) {
  writeLines(x)
}

#' @export

snapshot.pkg_resolution_result <- function(x, width = Inf, extra = NULL, ...) {
  if (!"pillar" %in% loadedNamespaces()) {
    throw(pkg_error("Needs pillar loaded!"))
  }
  if ("md5sum" %in% colnames(x)) {
    x$md5sum <- sub("^[0-9a-fA-F]{32}$", "<md5>", x$md5sum)
  }
  if ("sha256" %in% colnames(x)) {
    x$sha256 <- sub("^[0-9a-fA-F]{64}$", "<sha256>", x$sha256)
  }
  if ("filesize" %in% colnames(x)) {
    x$filesize[!is.na(x$filesize)] <- 42L
  }

  chr <- format(x, width = width, ...)

  if ("sources" %in% colnames(x) && any(c("all", "sources") %in% extra)) {
    chr <- c(chr, "+ sources:", vcapply(x$sources, paste, collapse = ", "))
  }
  if ("remote" %in% colnames(x) && any(c("all", "remote") %in% extra)) {
    chr <- c(chr, "+ remote:", vcapply(x$remote, format))
  }
  if ("error" %in% colnames(x) && any(c("all", "error") %in% extra)) {
    chr <- c(chr, "+ error:", vcapply(x$error, format_error))
  }
  ## TODO: error
  if ("metadata" %in% colnames(x) && any(c("all", "metadata") %in% extra)) {
    chr <- c(chr, "+ metadata:", vcapply(x$metadata, format.pkg_metadata))
  }
  ## TODO: extra
  if ("dep_types" %in% colnames(x) && any(c("all", "dep_types") %in% extra)) {
    chr <- c(chr, "+ dep_types:", vcapply(x$dep_types, paste, collapse = ", "))
  }
  ## TODO: params

  structure(chr, class = "snapshot")
}

#' @export

format.remote_ref <- function(x, ...) {
  x$params <- paste(vcapply(x$params, format), collapse = ", ")
  paste0(
    "<",
    paste(class(x), collapse = "/"),
    "> ",
    paste0(names(x), ": ", vcapply(x, format), collapse = "; ")
  )
}

format.pkg_metadata <- function(x, ...) {
  if (length(x) == 0) {
    return("-")
  }
  paste0(names(x), ": ", vcapply(x, format), collapse = "; ")
}

format_error <- function(x, ...) {
  if (length(x) == 0) {
    return("-")
  }
  paste0(
    "<",
    paste0(class(x), collapse = "/"),
    "> ",
    conditionMessage(x)
  )
}

#' @importFrom utils modifyList

build_package <- function(path, build_args = list()) {
  default_args <- list(
    path = path,
    dest_path = NULL,
    binary = FALSE,
    vignettes = TRUE,
    manual = TRUE,
    args = NULL,
    quiet = TRUE
  )
  args <- modifyList(default_args, build_args)
  zip_path <- system.file(package = "zip", "bin", .Platform$r_arch)
  withr_with_path(zip_path, do.call(pkgbuild::build, args))
}

#' Installation plans
#'
#' An installation plan contains all data that is needed to install a
#' set of package files. It is usually created from an
#' [installation proposal][pkg_installation_proposal] with
#' [solving][pkg_solution] the dependencies and [downloading][pkg_downloads]
#' the package files.
#'
#' It is also possible to create an installation plan a different way. An
#' installation plan object must be a data frame, with at least the
#' following columns:
#'
#' * `package`: The name of the package.
#' * `type`: The type of the [package reference][pkg_refs].
#' * `binary`: Whether the package is a binary package.
#' * `file`: Full path to the package file or directory.
#' * `dependencies`: A list column that lists the names of the dependent
#'   packages for each package.
#' * `needscompilation`: Whether the package needs compilation. This should
#'   be `FALSE` for binary packages.
#'
#' For installation plans created via [pkg_installation_proposal], the plan
#' contains all columns from [`pkg_download_result`][pkg_download_result]
#' objects, and some additional ones:
#'
#' * `library`: the library the package is supposed to be installed to.
#' * `direct`: whether the package was directly requested or it is
#'   installed as a dependency.
#' * vignettes: whether the vignettes need to be (re)built.
#' * `packaged`: whether `R CMD build` was already called for the package.
#'
#' @seealso [pkg_installation_proposal] to create install plans,
#' [install_package_plan()] to install plans from any source.
#'
#' @examples
#' \dontrun{
#' pdi <- new_pkg_installation_proposal(
#'   "pak",
#'   config = list(library = tempfile())
#' )
#' pdi$resolve()
#' pdi$solve()
#' pdi$download()
#' pdi$get_install_plan()
#' }
#'
#' @name install_plans
NULL

#' Perform a package installation plan
#'
#' See ['Installation plans'][install_plans] for the details and the format.
#'
#' @param plan Package plan object, a data frame, see
#' ['Installation plans'][install_plans] for the format.
#' @param lib Library directory to install to.
#' @param num_workers Number of worker processes to use.
#' @param cache Package cache to use, or `NULL`.
#' @return Information about the installation process.
#'
#' @export

install_package_plan <- function(
  plan,
  lib = .libPaths()[[1]],
  num_workers = 1,
  cache = NULL
) {
  start <- Sys.time()
  cli::ansi_hide_cursor()
  on.exit(cli::ansi_show_cursor())

  lib <- lib[1]

  cli::cli_div(
    theme = list(
      ".timestamp" = list(
        color = "darkgrey",
        before = "(",
        after = ")"
      )
    )
  )

  required_columns <- c(
    "type",
    "binary",
    "dependencies",
    "file",
    "needscompilation",
    "package"
  )
  assert_that(
    inherits(plan, "data.frame"),
    all(required_columns %in% colnames(plan)),
    is_string(lib),
    is_count(num_workers, min = 1L)
  )

  if (!"vignettes" %in% colnames(plan)) {
    plan$vignettes <- FALSE
  }
  if (!"metadata" %in% colnames(plan)) {
    plan$metadata <- replicate(nrow(plan), character(), simplify = FALSE)
  }
  if (!"packaged" %in% colnames(plan)) {
    plan$packaged <- TRUE
  }
  if (!"used_cached_binary" %in% colnames(plan)) {
    plan$used_cached_binary <- FALSE
  }

  plan <- add_recursive_dependencies(plan)

  config <- list(
    lib = lib,
    num_workers = num_workers,
    show_time = tolower(Sys.getenv("PKG_OMIT_TIMES")) != "true"
  )
  state <- make_start_state(plan, config)
  state$cache <- cache
  state$progress <- create_progress_bar(state)
  on.exit(done_progress_bar(state), add = TRUE)

  withCallingHandlers(
    {
      ## Initialise one task for each worker
      for (i in seq_len(state$config$num_workers)) {
        task <- select_next_task(state)
        state <- start_task(state, task)
      }

      repeat {
        if (are_we_done(state)) {
          break
        }
        update_progress_bar(state)

        events <- poll_workers(state)
        state <- handle_events(state, events)
        task <- select_next_task(state)
        state <- start_task(state, task)
      }
    },
    error = function(e) kill_all_processes(state)
  )

  create_install_result(state)
}

# See https://github.com/r-lib/pkgdepends/issues/160
# for why this is needed. We only need to do it for the packages that
# are built from source, binaries can always be extracted.

add_recursive_dependencies <- function(plan) {
  # If all packages are source packages then there is nothing to do,
  # we cannot get into a bad situation. Similarly if we only have
  # binary packages.
  if (all(plan$binary) || all(!plan$binary)) {
    return(plan)
  }

  # Otherwise for every source package, we need to add its recursive
  # dependencies.
  idx <- structure(seq_len(nrow(plan)), names = plan$package)
  xdps <- structure(plan$dependencies, names = plan$package)
  done <- structure(logical(nrow(plan)), names = plan$package)
  srcidx <- which(!plan$binary)

  do <- function(i) {
    if (done[[i]]) {
      return()
    }
    miss <- idx[names(!done[xdps[[i]]])]
    done[[i]] <<- TRUE
    for (m in miss) {
      do(m)
    }
    xdps[[i]] <<- unique(c(xdps[[i]], unlist(xdps[xdps[[i]]])))
  }

  for (i in srcidx) {
    do(i)
  }

  plan$dependencies[srcidx] <- unname(xdps[srcidx])
  plan
}

make_start_state <- function(plan, config) {
  ## We store the data about build and installation here
  install_cols <- data.frame(
    stringsAsFactors = FALSE,
    row.names = NULL,
    package_done = plan$packaged,
    package_time = I(rep_list(nrow(plan), as.POSIXct(NA))),
    package_error = I(rep_list(nrow(plan), list())),
    package_stdout = I(rep_list(nrow(plan), character())),
    build_done = (plan$type %in% c("deps", "installed")) |
      plan$binary |
      plan$used_cached_binary,
    build_time = I(rep_list(nrow(plan), as.POSIXct(NA))),
    build_error = I(rep_list(nrow(plan), list())),
    build_stdout = I(rep_list(nrow(plan), character())),
    install_done = plan$type %in% c("deps", "installed"),
    install_time = I(rep_list(nrow(plan), as.POSIXct(NA))),
    install_error = I(rep_list(nrow(plan), list())),
    install_stdout = I(rep_list(nrow(plan), character())),
    worker_id = NA_character_
  )
  plan <- cbind(plan, install_cols)

  installed <- plan$package[plan$install_done]
  plan$deps_left <- lapply(plan$dependencies, setdiff, installed)

  list(
    plan = plan,
    workers = list(),
    config = config
  )
}

are_we_done <- function(state) {
  all(state$plan$install_done)
}

poll_workers <- function(state) {
  if (length(state$workers)) {
    timeout <- get_timeout(state)
    procs <- lapply(state$workers, "[[", "process")
    res <- processx::poll(procs, ms = timeout)
    vlapply(res, function(x) "ready" %in% x)
  } else {
    logical()
  }
}

get_timeout <- function(state) 100

handle_events <- function(state, events) {
  for (i in which(events)) {
    state <- handle_event(state, i)
  }
  state$workers <- drop_nulls(state$workers)
  state
}

handle_event <- function(state, evidx) {
  proc <- state$workers[[evidx]]$process

  ## Read out stdout. If process is done, then read out all
  if (proc$is_alive()) {
    deadline <- Sys.time() + as.difftime(1, units = "secs")
    while (Sys.time() < deadline && proc$poll_io(0)[["output"]] == "ready") {
      state$workers[[evidx]]$stdout <-
        c(state$workers[[evidx]]$stdout, out <- proc$read_output(n = 10000))
    }
  } else {
    state$workers[[evidx]]$stdout <-
      c(state$workers[[evidx]]$stdout, out <- proc$read_all_output())
  }

  ## If there is still output, then wait a bit more
  if (proc$is_alive() || proc$is_incomplete_output()) {
    return(state)
  }

  ## Otherwise we are done. Remove worker
  worker <- state$workers[[evidx]]
  state$workers[evidx] <- list(NULL)

  if (is.function(proc$get_result)) {
    proc$get_result()
  }

  ## Cut stdout to lines
  worker$stdout <- cut_into_lines(worker$stdout)

  ## Record what was done
  stop_task(state, worker)
}

select_next_task <- function(state) {
  ## Cannot run more workers?
  if (length(state$workers) >= state$config$num_workers) {
    return(task("idle"))
  }

  ## Can we select a package tree to build into a source package? Do that.
  can_package <- which(
    !state$plan$package_done &
      viapply(state$plan$deps_left, length) == 0 &
      is.na(state$plan$worker_id)
  )

  if (any(can_package)) {
    pkgidx <- can_package[1]
    return(task("package", pkgidx = pkgidx, phase = "uncompress"))
  }

  ## Can we select a source package build? Do that.
  can_build <- which(
    !state$plan$build_done &
      viapply(state$plan$deps_left, length) == 0 &
      is.na(state$plan$worker_id)
  )

  if (any(can_build)) {
    pkgidx <- can_build[1]
    return(task("build", pkgidx = pkgidx))
  }

  ## TODO: can we select a binary that is depended on by a source package?

  ## Otherwise select a binary if there is one
  can_install <- which(
    state$plan$build_done &
      !state$plan$install_done &
      is.na(state$plan$worker_id)
  )

  if (any(can_install)) {
    pkgidx <- can_install[1]
    return(task("install", pkgidx = pkgidx))
  }

  ## Detect internal error
  if (!all(state$plan$install_done) && all(is.na(state$plan$worker_id))) {
    todo <- state$plan$package[!state$plan$install_done]
    throw(pkg_error(
      "Cannot select new package installation task.",
      i = "{length(todo)} package{?s} still waiting to install: {.pkg {todo}}.",
      i = msg_internal_error()
    ))
  }

  ## Looks like nothing else to do
  task("idle")
}

task <- function(name, ...) {
  list(name = name, args = list(...))
}

start_task <- function(state, task) {
  if (task$name == "idle") {
    state
  } else if (task$name == "package") {
    start_task_package(state, task)
  } else if (task$name == "build") {
    start_task_build(state, task)
  } else if (task$name == "install") {
    start_task_install(state, task)
  } else {
    throw(pkg_error(
      "Unknown task: {.val {task$name}}.",
      i = msg_internal_error()
    ))
  }
}

get_worker_id <- (function() {
  id <- 0
  function() {
    id <<- id + 1
    as.character(id)
  }
})()

# TODO: test this on Windows
# nocov start

get_rtools_path <- function() {
  if (!is.null(pkgd_data$rtools_path)) {
    return(pkgd_data$rtools_path)
  }
  # no need to mess with Rtools on R >= 4.3.0, R puts Rtools on the PATH
  # the error message for missing Rtools will be worse, though :(
  if (getRversion() >= "4.3.0") {
    pkgd_data$rtools_path <- character()
  } else {
    pkgd_data$rtools_path <- pkgbuild::without_cache({
      if (suppressMessages(pkgbuild::has_rtools())) {
        gsub("/", "\\", pkgbuild::rtools_path(), fixed = TRUE)
      }
    })
  }
  pkgd_data$rtools_path
}

# nocov end

make_build_process <- function(
  path,
  pkg,
  tmp_dir,
  lib,
  vignettes,
  needscompilation,
  binary,
  cmd_args,
  metadata = NULL
) {
  # For windows, we need ensure the zip.exe bundled with the zip package is on the PATH
  # TODO: test this on Windows
  # nocov start
  if (is_windows()) {
    zip_tool_path <- asNamespace("zip")$get_tool("zip")
    rtools <- get_rtools_path()
    withr_local_path(c(dirname(zip_tool_path), rtools))
  }
  # nocov end

  # We also allow an extra subdirectory, e.g. in .tar.gz files downloaded
  # from GHA
  if (
    !file.exists(file.path(path, "DESCRIPTION")) &&
      length(subdir <- dir(path)) == 1
  ) {
    path <- file.path(path, subdir)
  }
  if (
    !file.exists(file.path(path, "DESCRIPTION")) &&
      length(subdir <- dir(path)) == 1
  ) {
    path <- file.path(path, subdir)
  }

  warn_for_long_paths(path, pkg)

  ## with_libpath() is needed for newer callr, which forces the current
  ## lib path in the child process.
  mkdirp(tmplib <- tempfile("pkg-lib"))
  # Otherwise it is loaded with a modified library path, so potentially from the
  # wrong place, especially in pak
  loadNamespace("pkgbuild")
  loadNamespace("desc")
  loadNamespace("callr")
  loadNamespace("processx")
  loadNamespace("cli")
  loadNamespace("ps")
  # loadNamespace("R6") # not needed, built time dependency

  add_metadata(path, metadata)

  withr_with_libpaths(
    c(tmplib, lib),
    action = "prefix",
    pkgbuild::pkgbuild_process$new(
      path,
      tmp_dir,
      binary = binary,
      vignettes = vignettes,
      needs_compilation = needscompilation,
      compile_attributes = FALSE,
      args = c(
        "--no-lock",
        cmd_args,
        if (binary) sprintf("--library=%s", tmplib)
      )
    )
  )
}

# TODO: test this, if possible
# nocov start

warn_for_long_paths <- function(path, pkg) {
  if (!is_windows()) {
    return()
  }
  pkg_paths <- dir(path, recursive = TRUE, full.names = TRUE)
  # No files here for R CMD INSTALL --build, path is a file there
  max_len <- max(c(0, nchar(pkg_paths)))
  if (max_len < 255) {
    return(path)
  }
  alert(
    "warning",
    paste0(
      "The {.pkg {pkg}} package has very long paths. The installation ",
      "will probably fail because most Windows versions do not support ",
      "long paths. Please tell the package authors about this."
    ),
    wrap = TRUE
  )
}

# nocov end

start_task_package <- function(state, task) {
  pkgidx <- task$args$pkgidx
  path <- state$plan$file[pkgidx]
  pkg <- state$plan$package[pkgidx]
  version <- state$plan$version[pkgidx]

  state$plan$package_time[[pkgidx]] <- Sys.time()
  alert("info", "Packaging {.pkg {pkg}} {.version {version}}")

  if (file.info(path)$isdir) {
    ## Just build tree_dir
    task$args$tree_dir <- path
    start_task_package_build(state, task)
  } else {
    ## Uncompress to tree_dir, then build it
    task$args$tree_dir <- file.path(tempdir(), "X", pkg)
    unlink(task$args$tree_dir, recursive = TRUE)
    start_task_package_uncompress(state, task)
  }
}

start_task_package_uncompress <- function(state, task) {
  pkgidx <- task$args$pkgidx
  path <- state$plan$file[pkgidx]

  needscompilation <- !identical(state$plan$needscompilation[pkgidx], "no")
  lib <- state$config$lib

  task$args$phase <- "uncompress"
  px <- make_uncompress_process(path, task$args$tree_dir)
  worker <- list(
    id = get_worker_id(),
    task = task,
    process = px,
    stdout = character()
  )
  state$workers <- c(
    state$workers,
    structure(list(worker), names = worker$id)
  )
  state$plan$worker_id[pkgidx] <- worker$id
  state
}

start_task_package_build <- function(state, task) {
  pkgidx <- task$args$pkgidx
  pkg <- state$plan$package[pkgidx]

  # GH (and similar) packages might be in a subdirectory
  subdir <- c(state$plan$metadata[[pkgidx]]["RemoteSubdir"])[1]
  if (is.null(subdir) || is.na(subdir)) {
    subdir <- "."
  }

  # The actual package might be in a subdirectory of the tarball,
  # e.g. when the tree was downloaded from GitHub. So the package is
  # either in tree_dir/subdir or tree_dir/something/subdir
  root_dir <- task$args$tree_dir
  pkg_dir <- file.path(root_dir, subdir)
  dir_root_dir <- dir(root_dir)

  if (!"DESCRIPTION" %in% dir_root_dir && length(dir_root_dir) == 1) {
    pkg_dir <- file.path(root_dir, dir_root_dir, subdir)
  }

  vignettes <- state$plan$vignettes[pkgidx]
  needscompilation <- !identical(state$plan$needscompilation[pkgidx], "no")
  lib <- state$config$lib
  metadata <- state$plan$metadata[[pkgidx]]

  task$args$phase <- "build"
  px <- make_build_process(
    pkg_dir,
    pkg,
    create_temp_dir(),
    lib,
    vignettes,
    needscompilation,
    binary = FALSE,
    cmd_args = NULL,
    metadata = metadata
  )
  worker <- list(
    id = get_worker_id(),
    task = task,
    process = px,
    stdout = character()
  )
  state$workers <- c(
    state$workers,
    structure(list(worker), names = worker$id)
  )
  state$plan$worker_id[pkgidx] <- worker$id
  state$plan$build_time[[pkgidx]] <- Sys.time()
  state
}

start_task_build <- function(state, task) {
  pkgidx <- task$args$pkgidx
  path <- state$plan$file[pkgidx]
  vignettes <- state$plan$vignettes[pkgidx]
  needscompilation <- !identical(state$plan$needscompilation[pkgidx], "no")
  tmp_dir <- create_temp_dir()
  lib <- state$config$lib

  pkg <- state$plan$package[pkgidx]
  version <- state$plan$version[pkgidx]
  alert("info", "Building {.pkg {pkg}} {.version {version}}")

  if ("install_args" %in% names(state$plan)) {
    cmd_args <- state$plan$install_args[pkgidx]
    if (identical(cmd_args, "")) cmd_args <- NULL
  } else {
    cmd_args <- NULL
  }
  px <- make_build_process(
    path,
    pkg,
    tmp_dir,
    lib,
    vignettes,
    needscompilation,
    binary = TRUE,
    cmd_args = cmd_args
  )
  worker <- list(
    id = get_worker_id(),
    task = task,
    process = px,
    stdout = character()
  )
  state$workers <- c(
    state$workers,
    structure(list(worker), names = worker$id)
  )
  state$plan$worker_id[pkgidx] <- worker$id
  state$plan$build_time[[pkgidx]] <- Sys.time()
  state
}

start_task_install <- function(state, task) {
  pkgidx <- task$args$pkgidx
  filename <- state$plan$file[pkgidx]
  lib <- state$config$lib
  metadata <- state$plan$metadata[[pkgidx]]

  pkg <- state$plan$package[pkgidx]
  version <- state$plan$version[pkgidx]
  update_progress_bar(state)

  px <- make_install_process(filename, lib = lib, metadata = metadata)
  worker <- list(
    id = get_worker_id(),
    task = task,
    process = px,
    stdout = character()
  )

  state$workers <- c(
    state$workers,
    structure(list(worker), names = worker$id)
  )
  state$plan$worker_id[pkgidx] <- worker$id
  state$plan$install_time[[pkgidx]] <- Sys.time()
  state
}

stop_task <- function(state, worker) {
  if (worker$task$name == "package") {
    stop_task_package(state, worker)
  } else if (worker$task$name == "build") {
    stop_task_build(state, worker)
  } else if (worker$task$name == "install") {
    stop_task_install(state, worker)
  } else {
    throw(pkg_error(
      "Unknown task: {.val {worker$task$name}}.",
      i = msg_internal_error()
    ))
  }
}

stop_task_package <- function(state, worker) {
  if (worker$task$args$phase == "uncompress") {
    stop_task_package_uncompress(state, worker)
  } else {
    stop_task_package_build(state, worker)
  }
}

stop_task_package_uncompress <- function(state, worker) {
  pkgidx <- worker$task$args$pkgidx
  success <- worker$process$get_exit_status() == 0

  if (!success) {
    pkg <- state$plan$package[pkgidx]
    version <- state$plan$version[pkgidx]
    time <- Sys.time() - state$plan$package_time[[pkgidx]]
    ptime <- format_time$pretty_sec(as.numeric(time, units = "secs"))
    alert("danger", "Failed to uncompress {.pkg {pkg}} {.version {version}}")
    update_progress_bar(state, 1L)

    state$plan$package_done[[pkgidx]] <- TRUE
    state$plan$package_time[[pkgidx]] <- time
    state$plan$package_error[[pkgidx]] <- !success
    state$plan$package_stdout[[pkgidx]] <- worker$stdout
    state$plan$worker_id[[pkgidx]] <- NA_character_

    throw(pkg_error(
      "Failed to uncompress {.pkg {pkg}} from
      {.path {state$plan$file[[pkgidx]]}}.",
      .data = list(
        package = pkg,
        version = version,
        time = time,
        stdout = worker$stdout
      ),
      .class = "package_uncompress_error"
    ))
  }

  start_task_package_build(state, worker$task)
}

stop_task_package_build <- function(state, worker) {
  pkgidx <- worker$task$args$pkgidx
  success <- worker$process$get_exit_status() == 0

  pkg <- state$plan$package[pkgidx]
  version <- state$plan$version[pkgidx]
  time <- Sys.time() - state$plan$package_time[[pkgidx]]
  ptime <- format_time$pretty_sec(as.numeric(time, units = "secs"))

  if (success) {
    alert(
      "success",
      paste0(
        "Packaged {.pkg {pkg}} {.version {version}}",
        if (isTRUE(state$config$show_time)) " {.timestamp {ptime}}"
      )
    )
    ## Need to save the name of the built package
    state$plan$file[pkgidx] <- worker$process$get_built_file()
  } else {
    alert(
      "danger",
      "Failed to create source package {.pkg {pkg}} \\
           {.version {version}}"
    )
    if (!identical(worker$stdout, "")) {
      cli::cli_h1("Standard output")
      cli::cli_verbatim(worker$stdout)
    } else {
      alert("info", "Standard output is empty") # nocov
    }
  }
  update_progress_bar(state, 1L)

  state$plan$package_done[[pkgidx]] <- TRUE
  state$plan$package_time[[pkgidx]] <- time
  state$plan$package_error[[pkgidx]] <- !success
  state$plan$package_stdout[[pkgidx]] <- worker$stdout
  state$plan$worker_id[[pkgidx]] <- NA_character_

  if (!success) {
    throw(pkg_error(
      "Failed to create source package {.pkg {pkg}} from source tree at
      {.path {state$plan$file[[pkgidx]]}}",
      .data = list(
        package = pkg,
        version = version,
        stdout = worker$stdout,
        time = time
      ),
      .class = "package_packaging_error"
    ))
  }

  prms <- state$plan$params[[pkgidx]]
  if (!is.null(state$cache) && !is_true_param(prms, "nocache")) {
    tryCatch(
      state$cache$add(
        state$plan$file[pkgidx],
        state$plan$target[pkgidx],
        package = pkg,
        version = version,
        built = TRUE,
        sha256 = state$plan$extra[[pkgidx]]$remotesha,
        vignettes = state$plan$vignettes[pkgidx],
        platform = "source"
      ),
      error = function(err) {
        alert(
          "warning",
          "Failed to add {.pkg {pkg}} \\
               {.version {version}} to the cache"
        )
      }
    )
  }

  state
}

stop_task_build <- function(state, worker) {
  ## TODO: make sure exit status is non-zero on build error!
  success <- worker$process$get_exit_status() == 0

  pkgidx <- worker$task$args$pkgidx
  pkg <- state$plan$package[pkgidx]
  version <- state$plan$version[pkgidx]
  time <- Sys.time() - state$plan$build_time[[pkgidx]]
  ptime <- format_time$pretty_sec(as.numeric(time, units = "secs"))
  prms <- state$plan$params[[pkgidx]]

  if (success) {
    alert(
      "success",
      paste0(
        "Built {.pkg {pkg}} {.version {version}}",
        if (isTRUE(state$config$show_time)) " {.timestamp {ptime}}"
      )
    )
    ## Need to save the name of the built package
    state$plan$file[pkgidx] <- worker$process$get_built_file()
  } else {
    ignore_error <- is_true_param(prms, "ignore-build-errors")
    alert(
      if (ignore_error) "warning" else "danger",
      paste0(
        "Failed to build {.pkg {pkg}} {.version {version}}",
        if (isTRUE(state$config$show_time)) " {.timestamp {ptime}}"
      )
    )
  }
  update_progress_bar(state, 1L)

  state$plan$build_done[[pkgidx]] <- TRUE
  state$plan$build_time[[pkgidx]] <- time
  state$plan$build_stdout[[pkgidx]] <- worker$stdout
  state$plan$worker_id[[pkgidx]] <- NA_character_

  if (success) {
    # a bit silly, but for compatibility...
    state$plan$build_error[[pkgidx]] <- FALSE
  } else {
    build_error <- list(
      package = pkg,
      version = version,
      stdout = worker$stdout,
      time = time
    )
    state$plan$build_error[[pkgidx]] <- build_error
    if (ignore_error) {
      # upstream will probably fail as well, but march on, neverthelesss
      state$plan$install_done[[pkgidx]] <- TRUE
      ## Need to remove from the dependency list
      state$plan$deps_left <- lapply(state$plan$deps_left, setdiff, pkg)
    } else {
      throw(pkg_error(
        "Failed to build source package {.pkg {pkg}}.",
        .data = build_error,
        .class = "package_build_error"
      ))
    }
  }

  if (success && !is.null(state$cache) && !is_true_param(prms, "nocache")) {
    ptfm <- current_r_platform()
    rv <- current_r_version()
    target <- paste0(state$plan$target[pkgidx], "-", ptfm, "-", rv)
    tryCatch(
      state$cache$add(
        state$plan$file[pkgidx],
        target,
        package = pkg,
        version = version,
        built = TRUE,
        sha256 = state$plan$extra[[pkgidx]]$remotesha,
        vignettes = state$plan$vignettes[pkgidx],
        platform = ptfm,
        rversion = rv
      ),
      error = function(err) {
        alert(
          "warning",
          "Failed to add {.pkg {pkg}} \\
               {.version {version}} ({ptfm}) to the cache"
        )
      }
    )
  }

  state
}

installed_note <- function(pkg) {
  github_note <- function() {
    meta <- pkg$metadata[[1]]
    paste0(
      "(github::",
      meta[["RemoteUsername"]],
      "/",
      meta[["RemoteRepo"]],
      "@",
      substr(meta[["RemoteSha"]], 1, 7),
      ")"
    )
  }

  gitlab_note <- function() {
    meta <- pkg$metadata[[1]]
    paste0(
      "(gitlab::",
      meta[["RemoteUsername"]],
      "/",
      meta[["RemoteRepo"]],
      "@",
      substr(meta[["RemoteSha"]], 1, 7),
      ")"
    )
  }

  git_note <- function() {
    meta <- pkg$metadata[[1]]
    paste0(
      "(git::",
      meta[["RemoteUrl"]],
      "@",
      substr(meta[["RemoteSha"]], 1, 7),
      ")"
    )
  }

  switch(
    pkg$type,
    cran = "",
    bioc = "(Bioconductor)",
    standard = "",
    local = "(local)",
    github = github_note(),
    gitlab = gitlab_note(),
    git = git_note()
  )
}

stop_task_install <- function(state, worker) {
  ## TODO: make sure the install status is non-zero on exit
  success <- worker$process$get_exit_status() == 0

  pkgidx <- worker$task$args$pkgidx
  pkg <- state$plan$package[pkgidx]
  version <- state$plan$version[pkgidx]
  time <- Sys.time() - state$plan$install_time[[pkgidx]]
  ptime <- format_time$pretty_sec(as.numeric(time, units = "secs"))
  note <- installed_note(state$plan[pkgidx, ])

  if (success) {
    alert(
      "success",
      paste0(
        "Installed {.pkg {pkg}} {.version {version}} {note}",
        if (isTRUE(state$config$show_time)) " {.timestamp {ptime}}"
      )
    )
  } else {
    alert("danger", "Failed to install {.pkg {pkg}} {.version {version}}")
  }
  update_progress_bar(state, 1L)

  state$plan$install_done[[pkgidx]] <- TRUE
  state$plan$install_time[[pkgidx]] <- time
  state$plan$install_error[[pkgidx]] <- !success
  state$plan$install_stdout[[pkgidx]] <- worker$stdout
  state$plan$worker_id[[pkgidx]] <- NA_character_

  if (!success) {
    throw(pkg_error(
      "Failed to install binary package {.pkg {pkg}}.",
      .class = "package_install_error"
    ))
  }

  ## Need to remove from the dependency list
  state$plan$deps_left <- lapply(state$plan$deps_left, setdiff, pkg)

  state
}

create_install_result <- function(state) {
  result <- state$plan
  class(result) <- c("pkginstall_result", class(result))
  result
}

#' @export

print.pkginstall_result <- function(x, ...) {
  newly <- sum(x$lib_status == "new" & x$type != "deps")
  upd <- sum(x$lib_status == "update")
  noupd <- sum(x$lib_status == "no-update")
  curr <- sum(x$lib_status == "current")

  build_time <- sum(unlist(x$build_time), na.rm = TRUE)
  inst_time <- sum(unlist(x$install_time), na.rm = TRUE)

  res <- c(
    "Summary:",
    if (newly) paste0(emoji("sparkles", ""), " ", newly, " new"),
    if (upd) paste0(emoji("rocket", ""), " ", upd, " updated"),
    if (noupd + curr) paste0(emoji("hand", ""), " ", noupd + curr, " kept"),
    if (!tolower(Sys.getenv("PKG_OMIT_TIMES")) == "true") {
      paste0("in ", format_time$pretty_sec(build_time + inst_time))
    }
  )

  cli::cli_alert_success(paste0(res, collapse = "  "))

  invisible(x)
}


kill_all_processes <- function(state) {
  alive <- FALSE
  for (i in seq_along(state$workers)) {
    proc <- state$workers[[i]]$process
    proc$signal(tools::SIGINT)
    alive <- alive || proc$is_alive()
  }

  if (alive) {
    for (i in seq_along(state$workers)) {
      proc <- state$workers[[i]]$process
      proc$wait(200)
      if (ps::ps_is_supported()) proc$kill_tree() else proc$kill()
    }
  }
}

test_that("make_start_state", {
  plan <- readRDS(test_path("fixtures/sample_plan.rds"))
  state <- make_start_state(plan, list(foo = "bar"))

  expect_equal(names(state), c("plan", "workers", "config"))
  xcols <- c(
    "build_done",
    "build_time",
    "build_error",
    "build_stdout",
    "install_done",
    "install_time",
    "install_error",
    "install_stdout"
  )
  expect_true(all(xcols %in% colnames(state$plan)))
  eq_cols <- setdiff(colnames(plan), "deps_left")
  expect_identical(
    as.data.frame(plan[, eq_cols]),
    as.data.frame(state$plan[, eq_cols])
  )
})

test_that("are_we_done", {
  plan <- readRDS(test_path("fixtures/sample_plan.rds"))
  state <- make_start_state(plan, list(foo = "bar"))
  expect_false(are_we_done(state))

  state$plan$install_done <- TRUE
  state$plan$install_done[1] <- FALSE
  expect_false(are_we_done(state))

  state$plan$install_done[1] <- TRUE
  expect_true(are_we_done(state))
})

test_that("poll_workers", {
  state <- list(workers = list())
  expect_equal(poll_workers(state), logical())

  skip_on_os("windows")

  ## These might fail, but that does not matter much here
  p1 <- processx::process$new("true", stdout = "|", stderr = "2>&1")
  p2 <- processx::process$new("true", stdout = "|", stderr = "2>&1")

  state <- list(workers = list(list(process = p1)))
  expect_equal(poll_workers(state), TRUE)

  state <- list(workers = c(state$workers, list(list(process = p2))))
  expect_true(any(poll_workers(state)))

  opts <- callr::r_process_options(func = function() Sys.sleep(5))
  p3 <- callr::r_process$new(opts)
  on.exit(p3$kill(), add = TRUE)
  state <- list(workers = c(state$workers, list(list(process = p3))))
  p <- poll_workers(state)
  expect_true(any(p))
  expect_false(p[3])
  p3$kill()
})

test_that("handle_event, process still running", {
  ## If just output, but the process is still running, then collect
  ## stdout and stderr
  plan <- readRDS(test_path("fixtures/sample_plan.rds"))
  state <- make_start_state(plan, list(num_workers = 2))

  mockery::stub(
    start_task_build,
    "make_build_process",
    make_dummy_worker_process()
  )

  ## Run a dummy worker that runs for 10s, writes to stdout & stderr
  expect_snapshot(
    state <- start_task_build(state, task("build", pkgidx = 1))
  )
  proc <- state$workers[[1]]$process
  on.exit(proc$kill(), add = TRUE)

  for (i in 1:2) {
    proc$poll_io(-1)
    state <- handle_event(state, 1)
    expect_false(is.null(state$workers[[1]]))
    ## We cannot be sure that both stdout and stderr are already there,
    ## but one of them must be
    out <- paste(state$workers[[1]]$stdout, collapse = "")
    err <- paste(state$workers[[1]]$stderr, collapse = "")
    expect_true(grepl("^out ", out) || grepl("^err ", err))
    expect_true(proc$is_alive())
    expect_false(is.na(state$plan$worker_id[1]))
  }

  proc$kill()
})

test_that("handle_event, build process finished", {
  local_cli_config()
  plan <- readRDS(test_path("fixtures/sample_plan.rds"))
  state <- make_start_state(plan, list(foo = "bar"))
  state$plan$build_done[1] <- FALSE

  mockery::stub(
    start_task_build,
    "make_build_process",
    make_dummy_worker_process(n_iter = 2, sleep = 0)
  )

  expect_snapshot(
    state <- start_task_build(state, task("build", pkgidx = 1))
  )

  proc <- state$workers[[1]]$process
  on.exit(proc$kill(), add = TRUE)

  expect_snapshot(
    repeat {
      events <- poll_workers(state)
      state <- handle_events(state, events)
      if (all(state$plan$build_done)) break
    }
  )

  expect_false(proc$is_alive())
  expect_false(state$plan$build_error[[1]])
  expect_equal(
    state$plan$build_stdout[[1]],
    c("out 1", "err 1", "out 2", "err 2")
  )
  expect_identical(state$plan$worker_id[[1]], NA_character_)
  expect_equal(length(state$workers), 0)
})

test_that("handle event, build process finished, but failed", {
  local_cli_config()
  plan <- readRDS("fixtures/sample_plan.rds")
  expect_snapshot(state <- make_start_state(plan, list(foo = "bar")))
  state$plan$build_done[1] <- FALSE

  mockery::stub(
    start_task_install,
    "make_install_process",
    make_dummy_worker_process(n_iter = 2, sleep = 0, status = 1)
  )

  expect_snapshot(
    state <- start_task_install(state, task("install", pkgidx = 1))
  )
  proc <- state$workers[[1]]$process
  on.exit(proc$kill(), add = TRUE)

  expect_snapshot(
    error = TRUE,
    repeat {
      events <- poll_workers(state)
      state <- handle_events(state, events)
      if (all(state$plan$build_done)) break
    }
  )
})

test_that("handle_event, install process finished", {
  plan <- readRDS("fixtures/sample_plan.rds")
  state <- make_start_state(plan, list(foo = "bar"))

  mockery::stub(
    start_task_install,
    "make_install_process",
    make_dummy_worker_process(n_iter = 2, sleep = 0)
  )

  state <- start_task_install(state, task("install", pkgidx = 1))
  proc <- state$workers[[1]]$process
  on.exit(proc$kill(), add = TRUE)

  done <- FALSE
  expect_snapshot(
    repeat {
      events <- poll_workers(state)
      state <- handle_events(state, events)
      if (done) {
        break
      }
      if (!proc$is_alive()) done <- TRUE
    }
  )

  expect_false(proc$is_alive())
  expect_false(state$plan$install_error[[1]])
  expect_equal(
    state$plan$install_stdout[[1]],
    c("out 1", "err 1", "out 2", "err 2")
  )
  expect_identical(state$plan$worker_id[[1]], NA_character_)
  expect_equal(length(state$workers), 0)
})

test_that("handle event, install process finished, but failed", {
  plan <- readRDS("fixtures/sample_plan.rds")
  state <- make_start_state(plan, list(foo = "bar"))

  mockery::stub(
    start_task_install,
    "make_install_process",
    make_dummy_worker_process(n_iter = 2, sleep = 0, status = 1)
  )

  state <- start_task_install(state, task("install", pkgidx = 1))
  proc <- state$workers[[1]]$process
  on.exit(proc$kill(), add = TRUE)

  expect_snapshot(error = TRUE, {
    done <- FALSE
    repeat {
      events <- poll_workers(state)
      state <- handle_events(state, events)
      if (done) {
        break
      }
      if (!proc$is_alive()) done <- TRUE
    }
  })
})

test_that("select_next_task", {
  plan <- readRDS("fixtures/sample_plan.rds")
  state <- make_start_state(plan, list(num_workers = 2))

  ## If no more workers are available
  state$workers <- list(list("dummy1"), list("dummy2"))
  expect_equal(select_next_task(state), task("idle"))

  ## An ongoing install task is not selected again
  state <- make_start_state(plan, list(num_workers = 2))
  state$plan$worker_id[-nrow(state$plan)] <- 42
  expect_equal(
    select_next_task(state),
    task("install", pkgidx = nrow(state$plan))
  )

  ## An ongoing build task is not selected again
  state <- make_start_state(plan, list(num_workers = 2))
  state$plan$build_done <- FALSE
  state$plan$deps_left[] <- rep_list(nrow(state$plan), character())
  state$plan$worker_id[-nrow(state$plan)] <- 42
  expect_equal(
    select_next_task(state),
    task("build", pkgidx = nrow(state$plan))
  )

  ## Source is preferred over binary
  state <- make_start_state(plan, list(num_workers = 2))
  state$plan$build_done[nrow(state$plan)] <- FALSE
  state$plan$deps_left[] <- rep_list(nrow(state$plan), character())
  expect_equal(
    select_next_task(state),
    task("build", pkgidx = nrow(state$plan))
  )

  ## Source is selected only if dependencies are done
  state <- make_start_state(plan, list(num_workers = 2))
  state$plan$build_done <- FALSE
  state$plan$deps_left[] <- rep_list(nrow(state$plan), "foobar")
  state$plan$deps_left[[nrow(state$plan)]] <- character()
  expect_equal(
    select_next_task(state),
    task("build", pkgidx = nrow(state$plan))
  )

  ## Binary is selected irrespectively of dependencies
  state <- make_start_state(plan, list(num_workers = 2))
  state$plan$deps_left[] <- rep_list(nrow(state$plan), "foobar")
  expect_equal(
    select_next_task(state),
    task("install", pkgidx = 1L)
  )

  ## We cannot select anything, because of the dependencies
  state <- make_start_state(plan, list(num_workers = 2))
  state$plan$build_done <- FALSE
  state$plan$worker_id[1] <- 1
  state$plan$deps_left[] <- rep_list(nrow(state$plan), "foobar")
  expect_equal(
    select_next_task(state),
    task("idle")
  )
})

test_that("start_task", {
  expect_snapshot(error = TRUE, start_task(list(), task("foobar")))
})

test_that("stop_task", {
  expect_snapshot(
    error = TRUE,
    stop_task(list(), list(task = task("foobar")))
  )
})

test_that("get_worker_id", {
  expect_true(get_worker_id() != get_worker_id())
})

test_that("kill_all_processes", {
  skip_on_os("windows")

  p1 <- processx::process$new("true", stdout = "|")
  on.exit(p1$kill(), add = TRUE)
  p2 <- processx::process$new("true", stdout = "|")
  on.exit(p2$kill(), add = TRUE)
  opts <- callr::r_process_options(func = function() Sys.sleep(5))
  p3 <- callr::r_process$new(opts)
  on.exit(p3$kill(), add = TRUE)

  state <- list(
    workers = list(
      list(process = p1),
      list(process = p2),
      list(process = p3)
    )
  )

  kill_all_processes(state)

  expect_false(p1$is_alive())
  expect_false(p2$is_alive())
  expect_false(p3$is_alive())

  p1$kill()
  p2$kill()
  p3$kill()
})

test_that("kill_all_processes that catch/ignore SIGINT", {
  skip_on_cran()
  skip_on_os("windows")
  if (Sys.which("bash") == "") {
    skip("Needs 'bash'")
  }

  sh <- "trap '>&2 echo \"Hold on\"' INT
    for ((n=5; n; n--))
    do
      echo going
      sleep 1
    done"

  px <- processx::process$new("bash", c("-c", sh), stdout = "|", stderr = "|")
  expect_true(px$is_alive())

  state <- list(workers = list(list(process = px)))

  ## Need to wait until the shell starts and traps SIGINT
  px$poll_io(2000)

  tic <- Sys.time()
  kill_all_processes(state)
  limit <- Sys.time() + as.difftime(3, units = "secs")
  while (px$is_alive() && Sys.time() < limit) {
    Sys.sleep(0.05)
  }
  expect_true(Sys.time() < limit)
  expect_true(Sys.time() - tic > as.difftime(0.2, units = "secs"))
  expect_false(px$is_alive())

  ## We can't get the output of the signal handler, because SIGKILL
  ## does not ensure emptying the buffers....

  px$kill()
})

test_that("deadlock detection", {
  plan <- data_frame(
    package = c("p1", "p2", "p3"),
    type = "cran",
    binary = FALSE,
    dependencies = list("p2", "p3", "p1"),
    file = NA_character_,
    needscompilation = FALSE
  )

  expect_snapshot(
    error = TRUE,
    install_package_plan(plan, lib = tempfile())
  )
})

test_that("make_build_process", {
  tmp <- withr::local_tempdir()
  mkdirp(file.path(tmp, "subdir"))
  file.copy(
    test_path("fixtures", "foo"),
    file.path(tmp, "subdir"),
    recursive = TRUE
  )

  p <- make_build_process(
    tmp,
    "foo",
    tempdir(),
    .libPaths(),
    vignettes = FALSE,
    needscompilation = TRUE,
    binary = FALSE,
    cmd_args = character()
  )

  p$wait(5000)
  p$kill()
  blt <- p$get_built_file()
  expect_equal(basename(blt), "foo_0.0.0.9000.tar.gz")
  expect_true(file.exists(blt))

  p <- make_build_process(
    file.path(tmp, "subdir"),
    "foo",
    tempdir(),
    .libPaths(),
    vignettes = FALSE,
    needscompilation = TRUE,
    binary = FALSE,
    cmd_args = character()
  )

  p$wait(5000)
  p$kill()
  blt <- p$get_built_file()
  expect_equal(basename(blt), "foo_0.0.0.9000.tar.gz")
  expect_true(file.exists(blt))
})

test_that("install_args are passed", {
  withr::local_envvar(PKG_OMIT_TIMES = "true")

  pkg <- source_test_package("foo")
  plan <- data_frame(
    type = "local",
    binary = FALSE,
    dependencies = list(character()),
    file = pkg,
    needscompilation = TRUE,
    package = "foo",
    install_args = "--no-inst"
  )

  lib <- withr::local_tempdir()
  expect_snapshot(install_package_plan(plan, lib = lib))

  expect_false(file.exists(file.path(lib, "foo", "installed-file")))
})

test_that("built package is added to the cache", {})

test_that("installed_note", {
  expect_snapshot({
    installed_note(list(type = "cran"))
    installed_note(list(type = "bioc"))
    installed_note(list(type = "standard"))
    installed_note(list(type = "local"))
    installed_note(list(
      type = "github",
      metadata = list(list(
        RemoteUsername = "r-lib",
        RemoteRepo = "pak",
        RemoteSha = "5a4da54df42528545af8a64e83112be21273907c6dfa0f31a0982ca88db6527d"
      ))
    ))
  })
})

# nocov start

gitcreds_get <- NULL
gitcreds_set <- NULL
gitcreds_delete <- NULL
gitcreds_list_helpers <- NULL
gitcreds_cache_envvar <- NULL
gitcreds_fill <- NULL
gitcreds_approve <- NULL
gitcreds_reject <- NULL
gitcreds_parse_output <- NULL

gitcreds <- local({
  # ------------------------------------------------------------------------
  # Public API
  # ------------------------------------------------------------------------

  gitcreds_get <<- function(
    url = "https://github.com",
    use_cache = TRUE,
    set_cache = TRUE
  ) {
    stopifnot(
      is_string(url),
      has_no_newline(url),
      is_flag(use_cache),
      is_flag(set_cache)
    )

    cache_ev <- gitcreds_cache_envvar(url)
    if (use_cache && !is.null(ans <- gitcreds_get_cache(cache_ev))) {
      return(ans)
    }

    check_for_git()

    out <- gitcreds_fill(list(url = url), dummy = TRUE)
    creds <- gitcreds_parse_output(out, url)

    if (set_cache) {
      gitcreds_set_cache(cache_ev, creds)
    }

    creds
  }

  gitcreds_set <<- function(url = "https://github.com") {
    if (!is_interactive()) {
      throw(new_error(
        "gitcreds_not_interactive_error",
        message = "`gitcreds_set()` only works in interactive sessions"
      ))
    }
    stopifnot(is_string(url), has_no_newline(url))
    check_for_git()

    current <- tryCatch(
      gitcreds_get(url, use_cache = FALSE, set_cache = FALSE),
      gitcreds_no_credentials = function(e) NULL
    )

    if (!is.null(current)) {
      gitcreds_set_replace(url, current)
    } else {
      gitcreds_set_new(url)
    }

    msg("-> Removing credentials from cache...")
    gitcreds_delete_cache(gitcreds_cache_envvar(url))

    msg("-> Done.")
    invisible()
  }

  #' Replace credentials with new ones
  #'
  #' It only works interactively, because of `menu()` in `ack()` and
  #' `readline()`.
  #'
  #' We need to set a username, it is compulsory for git credential.
  #' 1. If there was one in the url, then we use that.
  #' 2. Otherwise if git has a username configured for this URL, we use that.
  #' 3. Otherwise we use the username in the credentials we are replacing.
  #'
  #' @param url URL.
  #' @param current Must not be `NULL`, and it must contain a
  #' `gitcreds` object. (Well, a named list, really.)
  #' @noRd
  #' @return Nothing.

  gitcreds_set_replace <- function(url, current) {
    # Potentially take username from the credential we are replacing
    current_username <- current$username

    # Keep warning until there is a credential to replace.
    # In case there are multiple credentials for the same URL.
    while (!is.null(current)) {
      if (!ack(url, current, "Replace")) {
        throw(new_error("gitcreds_abort_replace_error"))
      }

      msg("\n-> Removing current credentials...")
      gitcreds_reject(current)

      current <- tryCatch(
        gitcreds_get(url, use_cache = FALSE, set_cache = FALSE),
        gitcreds_no_credentials = function(e) NULL
      )
      if (!is.null(current)) msg("\n!! Found more matching credentials!")
    }

    msg("")
    pat <- readline("? Enter new password or token: ")

    username <- get_url_username(url) %||%
      gitcreds_username(url) %||%
      current_username

    msg("-> Adding new credentials...")
    gitcreds_approve(list(url = url, username = username, password = pat))

    invisible()
  }

  #' Set new credentials
  #'
  #' This should not replace or remove any old credentials, but of course
  #' we cannot be sure, because credential helpers pretty much do what they
  #' want.
  #'
  #' We need to set a username, it is compulsory for git credential.
  #' 1. If there was one in the url, then we use that.
  #' 2. Otherwise if git has a username configured for this URL, we use that.
  #' 3. Otherwise we use a default username.
  #'
  #' @param url URL.
  #' @noRd
  #' @return Nothing.

  gitcreds_set_new <- function(url) {
    msg("\n")
    pat <- readline("? Enter password or token: ")

    username <- get_url_username(url) %||%
      gitcreds_username(url) %||%
      default_username()

    msg("-> Adding new credentials...")
    gitcreds_approve(list(url = url, username = username, password = pat))

    invisible()
  }

  gitcreds_delete <<- function(url = "https://github.com") {
    if (!is_interactive()) {
      throw(new_error(
        "gitcreds_not_interactive_error",
        message = "`gitcreds_delete()` only works in interactive sessions"
      ))
    }
    stopifnot(is_string(url))
    check_for_git()

    current <- tryCatch(
      gitcreds_get(url, use_cache = FALSE, set_cache = FALSE),
      gitcreds_no_credentials = function(e) NULL
    )

    if (is.null(current)) {
      return(invisible(FALSE))
    }

    if (!ack(url, current, "Delete")) {
      throw(new_error("gitcreds_abort_delete_error"))
    }

    msg("-> Removing current credentials...")
    gitcreds_reject(current)

    msg("-> Removing credentials from cache...")
    gitcreds_delete_cache(gitcreds_cache_envvar(url))

    msg("-> Done.")

    invisible(TRUE)
  }

  gitcreds_list_helpers <<- function() {
    check_for_git()
    out <- git_run(c("config", "--get-all", "credential.helper"))
    clear <- rev(which(out == ""))
    if (length(clear)) {
      out <- out[-(1:clear[1])]
    }
    out
  }

  gitcreds_cache_envvar <<- function(url) {
    pcs <- parse_url(url)
    bad <- is.na(pcs$protocol) | is.na(pcs$host)
    if (any(bad)) {
      stop("Invalid URL(s): ", paste(url[bad], collapse = ", "))
    }

    proto <- sub("^https?_$", "", paste0(pcs$protocol, "_"))
    user <- ifelse(pcs$username != "", paste0(pcs$username, "_AT_"), "")
    host0 <- sub("^api[.]github[.]com$", "github.com", pcs$host)
    host1 <- gsub("[.:]+", "_", host0)
    host <- gsub("[^a-zA-Z0-9_-]", "x", host1)

    slug1 <- paste0(proto, user, host)

    # fix the user name ambiguity, not that it happens often...
    slug2 <- ifelse(grepl("^AT_", slug1), paste0("AT_", slug1), slug1)

    # env vars cannot start with a number
    slug3 <- ifelse(grepl("^[0-9]", slug2), paste0("AT_", slug2), slug2)

    paste0("GITHUB_PAT_", toupper(slug3))
  }

  gitcreds_get_cache <- function(ev) {
    val <- Sys.getenv(ev, NA_character_)
    if (is.na(val) && ev == "GITHUB_PAT_GITHUB_COM") {
      val <- Sys.getenv("GITHUB_PAT", NA_character_)
    }
    if (is.na(val) && ev == "GITHUB_PAT_GITHUB_COM") {
      val <- Sys.getenv("GITHUB_TOKEN", NA_character_)
    }
    if (is.na(val) || val == "") {
      return(NULL)
    }
    if (val == "FAIL" || grepl("^FAIL:", val)) {
      class <- strsplit(val, ":", fixed = TRUE)[[1]][2]
      if (is.na(class)) {
        class <- "gitcreds_no_credentials"
      }
      throw(new_error(class))
    }

    unesc <- function(x) {
      gsub("\\\\(.)", "\\1", x)
    }

    # split on `:` that is not preceded by a `\`
    spval <- strsplit(val, "(?<!\\\\):", perl = TRUE)[[1]]
    spval0 <- unesc(spval)

    # Single field, then the token
    if (length(spval) == 1) {
      return(new_gitcreds(
        protocol = NA_character_,
        host = NA_character_,
        username = NA_character_,
        password = unesc(val)
      ))
    }

    # Two fields? Then it is username:password
    if (length(spval) == 2) {
      return(new_gitcreds(
        protocol = NA_character_,
        host = NA_character_,
        username = spval0[1],
        password = spval0[2]
      ))
    }

    # Otherwise a full record
    if (length(spval) %% 2 == 1) {
      warning(
        "Invalid gitcreds credentials in env var `",
        ev,
        "`. ",
        "Maybe an unescaped ':' character?"
      )
      return(NULL)
    }

    creds <- structure(
      spval0[seq(2, length(spval0), by = 2)],
      names = spval[seq(1, length(spval0), by = 2)]
    )
    do.call("new_gitcreds", as.list(creds))
  }

  gitcreds_set_cache <- function(ev, creds) {
    esc <- function(x) gsub(":", "\\:", x, fixed = TRUE)
    keys <- esc(names(creds))
    vals <- esc(unlist(creds, use.names = FALSE))
    value <- paste0(keys, ":", vals, collapse = ":")
    do.call("set_env", list(structure(value, names = ev)))
    invisible(NULL)
  }

  gitcreds_delete_cache <- function(ev) {
    Sys.unsetenv(ev)
  }

  # ------------------------------------------------------------------------
  # Raw git credential API
  # ------------------------------------------------------------------------

  gitcreds_fill <<- function(input, args = character(), dummy = TRUE) {
    if (dummy) {
      helper <- paste0(
        "credential.helper=\"! echo protocol=dummy;",
        "echo host=dummy;",
        "echo username=dummy;",
        "echo password=dummy\""
      )
      args <- c(args, "-c", helper)
    }

    gitcreds_run("fill", input, args)
  }

  gitcreds_approve <<- function(creds, args = character()) {
    gitcreds_run("approve", creds, args)
  }

  gitcreds_reject <<- function(creds, args = character()) {
    gitcreds_run("reject", creds, args)
  }

  gitcreds_parse_output <<- function(txt, url) {
    if (is.null(txt) || txt[1] == "protocol=dummy") {
      throw(new_error("gitcreds_no_credentials", url = url))
    }
    nms <- sub("=.*$", "", txt)
    vls <- sub("^[^=]+=", "", txt)
    structure(as.list(vls), names = nms, class = "gitcreds")
  }

  #' Run a `git credential` command
  #'
  #' @details
  #' We set the [gitcreds_env()] environment variables, to avoid dialog boxes
  #' from some credential helpers and also validation that potentiall needs
  #' an internet connection.
  #'
  #' @param command Command name, e.g. `"fill"`.
  #' @param input Named list of input, see
  #' https://git-scm.com/docs/git-credential#IOFMT
  #' @param args Extra command line arguments, added after `git` and
  #' _before_ `command`, to allow `git -c ... fill`.
  #' @return Standard output, line by line.
  #'
  #' @noRd
  #' @seealso [git_run()].

  gitcreds_run <- function(command, input, args = character()) {
    env <- gitcreds_env()
    oenv <- set_env(env)
    on.exit(set_env(oenv), add = TRUE)

    stdin <- create_gitcreds_input(input)

    git_run(c(args, "credential", command), input = stdin)
  }

  # ------------------------------------------------------------------------
  # Helpers specific to git
  # ------------------------------------------------------------------------

  #' Run a git command
  #'
  #' @details
  #' Currently we don't set the credential specific environment variables
  #' here, and credential helpers invoked by `git` behave the same way as
  #' they would from the command line.
  #'
  #' ## Errors
  #'
  #' On error `git_run()` returns an error with class `git_error` and
  #' also `gitcreds_error`. The error object includes
  #' * `args` the command line arguments,
  #' * `status`: the exit status of the command,
  #' * `stdout`: the standard output of the command, line by line.
  #' * `stderr`: the standard error of the command, line by line.
  #'
  #' @param args Command line arguments.
  #' @param input The standard input (the `input` argument of [system2()].
  #' @noRd
  #' @return Standard output, line by line.

  git_run <- function(args, input = NULL) {
    stderr_file <- tempfile("gitcreds-stderr-")
    on.exit(unlink(stderr_file, recursive = TRUE), add = TRUE)
    if (!is.null(input)) {
      stdin_file <- tempfile("gitcreds-stdin-")
      on.exit(unlink(stdin_file, recursive = TRUE), add = TRUE)
      writeBin(charToRaw(input), stdin_file)
      stdin <- stdin_file
    } else {
      stdin <- ""
    }
    out <- tryCatch(
      suppressWarnings(system2(
        "git",
        args,
        stdin = stdin,
        stdout = TRUE,
        stderr = stderr_file
      )),
      error = function(e) NULL
    )

    if (!is.null(attr(out, "status")) && attr(out, "status") != 0) {
      throw(new_git_error(
        "git_error",
        args = args,
        stdout = out,
        status = attr(out, "status"),
        stderr = read_file(stderr_file)
      ))
    }

    out
  }

  #' Request confirmation from the user, to replace or delete credentials
  #'
  #' This function only works in interactive sessions.
  #'
  #' @param url URL to delete or set new credentials for.
  #' @param current The current credentials.
  #' @return `FALSE` is the user changed their mind, to keep the current
  #' credentials. `TRUE` for replacing/deleting them.
  #'
  #' @noRd
  #' @seealso [gitcreds_set()].

  ack <- function(url, current, what = "Replace") {
    msg("\n-> Your current credentials for ", squote(url), ":\n")
    msg(paste0(format(current, header = FALSE), collapse = "\n"), "\n")

    choices <- c(
      "Abort update with error, and keep the existing credentials",
      paste(what, "these credentials"),
      if (has_password(current)) "See the password / token"
    )

    repeat {
      ch <- utils::menu(title = "-> What would you like to do?", choices)

      if (ch == 1) {
        return(FALSE)
      }
      if (ch == 2) {
        return(TRUE)
      }

      msg("\nCurrent password: ", current$password, "\n\n")
    }
  }

  #' Whether a `gitcreds` credential has a non-empty `password`
  #'
  #' This is usually `TRUE`.
  #'
  #' @param creds `gitcreds`
  #' @noRd
  #' @return `TRUE` is there is a `password`

  has_password <- function(creds) {
    is_string(creds$password) && creds$password != ""
  }

  #' Create a string that can be passed as standard input to `git credential`
  #' commands
  #'
  #' @param args Usually a `gitcreds` object, but can be a named list in
  #' general. This is a format: https://git-scm.com/docs/git-credential#IOFMT
  #' @noRd
  #' @return String.

  create_gitcreds_input <- function(args) {
    paste0(
      paste0(names(args), "=", args, collapse = "\n"),
      "\n\n"
    )
  }

  #' Environment to set for all `git credential` commands.
  #' @noRd
  #' @return Named character vector.

  gitcreds_env <- function() {
    # Avoid interactivity and validation with some common credential helpers
    c(
      GCM_INTERACTIVE = "Never",
      GCM_MODAL_PROMPT = "false",
      GCM_VALIDATE = "false",
      GCM_GUI_PROMPT = "false"
    )
  }

  #' Check if `git` is installed and can run
  #'
  #' If not installed, a `gitcreds_nogit_error` is thrown.
  #'
  #' @noRd
  #' @return Nothing

  check_for_git <- function() {
    # This is simpler than Sys.which(), and also less fragile
    has_git <- tryCatch(
      {
        suppressWarnings(system2(
          "git",
          "--version",
          stdout = TRUE,
          stderr = null_file()
        ))
        TRUE
      },
      error = function(e) FALSE
    )

    if (!has_git) throw(new_error("gitcreds_nogit_error"))
  }

  #' Query the `username` to use for `git config credential`
  #'
  #' @details
  #' The current working directory matters for this command, as you can
  #' configure `username` in a local `.git/config` file (via
  #' `git config --local`).
  #'
  #' @param url URL to query the username for, or `NULL`. If not `NULL`,
  #' then we first try to query an URL-specific username. See
  #' https://git-scm.com/docs/gitcredentials for more about URL-specific
  #' credential config
  #' @noRd
  #' @return A string with the username, or `NULL` if no default was found.

  gitcreds_username <- function(url = NULL) {
    gitcreds_username_for_url(url) %||% gitcreds_username_generic()
  }

  gitcreds_username_for_url <- function(url) {
    if (is.null(url)) {
      return(NULL)
    }
    tryCatch(
      git_run(c(
        "config",
        "--get-urlmatch",
        "credential.username",
        shQuote(url)
      )),
      git_error = function(err) {
        if (err$status == 1) NULL else throw(err)
      }
    )
  }

  gitcreds_username_generic <- function() {
    tryCatch(
      git_run(c("config", "credential.username")),
      git_error = function(err) {
        if (err$status == 1) NULL else throw(err)
      }
    )
  }

  #' User name to use when creating a credential, if there is nothing better
  #'
  #' These user names are typical for some git tools, e.g.
  #' [Git Credential Manager for Windows](http://microsoft.github.io/Git-Credential-Manager-for-Windows/)
  #' (`manager`) and
  #' [Git Credential Manager Core](https://github.com/Microsoft/Git-Credential-Manager-Core)
  #' (`manager-core`).
  #'
  #' @noRd
  #' @return Character string

  default_username <- function() {
    "PersonalAccessToken"
  }

  new_gitcreds <- function(...) {
    structure(list(...), class = "gitcreds")
  }

  # ------------------------------------------------------------------------
  # Errors
  # ------------------------------------------------------------------------

  gitcred_errors <- function() {
    c(
      git_error = "System git failed",
      gitcreds_nogit_error = "Could not find system git",
      gitcreds_not_interactive_error = "gitcreds needs an interactive session",
      gitcreds_abort_replace_error = "User aborted updating credentials",
      gitcreds_abort_delete_error = "User aborted deleting credentials",
      gitcreds_no_credentials = "Could not find any credentials",
      gitcreds_no_helper = "No credential helper is set",
      gitcreds_multiple_helpers = "Multiple credential helpers, only using the first",
      gitcreds_unknown_helper = "Unknown credential helper, cannot list credentials"
    )
  }

  new_error <- function(class, ..., message = "", call. = TRUE, domain = NULL) {
    if (message == "") {
      message <- gitcred_errors()[[class]]
    }
    message <- .makeMessage(message, domain = domain)
    cond <- list(message = message, ...)
    if (call.) {
      cond$call <- sys.call(-1)
    }
    class(cond) <- c(class, "gitcreds_error", "error", "condition")
    cond
  }

  new_git_error <- function(class, ..., stderr) {
    cond <- new_error(class, ..., stderr = stderr)
    cond$message <- paste0(cond$message, ": ", stderr)
    cond
  }

  new_warning <- function(
    class,
    ...,
    message = "",
    call. = TRUE,
    domain = NULL
  ) {
    if (message == "") {
      message <- gitcred_errors()[[class]]
    }
    message <- .makeMessage(message, domain = domain)
    cond <- list(message = message, ...)
    if (call.) {
      cond$call <- sys.call(-1)
    }
    class(cond) <- c(class, "gitcreds_warning", "warning", "condition")
    cond
  }

  throw <- function(cond) {
    cond
    if ("error" %in% class(cond)) {
      stop(cond)
    } else if ("warning" %in% class(cond)) {
      warning(cond)
    } else if ("message" %in% class(cond)) {
      message(cond)
    } else {
      signalCondition(cond)
    }
  }

  # ------------------------------------------------------------------------
  # Genetic helpers
  # ------------------------------------------------------------------------

  #' Set/remove env var and return the old values
  #'
  #' @param envs Named character vector or list of env vars to set. `NA`
  #' values will un-set an env var.
  #' @noRd
  #' @return Character vector, the old values of the supplied environment
  #' variables, `NA` for the ones that were not set.

  set_env <- function(envs) {
    current <- Sys.getenv(names(envs), NA_character_, names = TRUE)
    na <- is.na(envs)
    if (any(na)) {
      Sys.unsetenv(names(envs)[na])
    }
    if (any(!na)) {
      do.call("Sys.setenv", as.list(envs[!na]))
    }
    invisible(current)
  }

  #' Get the user name from a `protocol://username@host/path` URL.
  #'
  #' @param url URL
  #' @noRd
  #' @return String or `NULL` if `url` does not have a username.

  get_url_username <- function(url) {
    nm <- parse_url(url)$username
    if (nm == "") NULL else nm
  }

  #' Parse URL
  #'
  #' It does not parse query parameters, as we don't deal with them here.
  #' The port number is included in the host name, if present.
  #'
  #' @param url Character vector of one or more URLs.
  #' @noRd
  #' @return Data frame with string columns: `protocol`, `username`,
  #' `password`, `host`, `path`.

  parse_url <- function(url) {
    re_url <- paste0(
      "^(?<protocol>[a-zA-Z0-9]+)://",
      "(?:(?<username>[^@/:]+)(?::(?<password>[^@/]+))?@)?",
      "(?<host>[^/]+)",
      "(?<path>.*)$" # don't worry about query params here...
    )

    mch <- re_match(url, re_url)
    mch[, setdiff(colnames(mch), c(".match", ".text")), drop = FALSE]
  }

  is_string <- function(x) {
    is.character(x) && length(x) == 1 && !is.na(x)
  }

  is_flag <- function(x) {
    is.logical(x) && length(x) == 1 && !is.na(x)
  }

  has_no_newline <- function(url) {
    !grepl("\n", url, fixed = TRUE)
  }

  # From the rematch2 package

  re_match <- function(text, pattern, perl = TRUE, ...) {
    stopifnot(is.character(pattern), length(pattern) == 1, !is.na(pattern))
    text <- as.character(text)

    match <- regexpr(pattern, text, perl = perl, ...)

    start <- as.vector(match)
    length <- attr(match, "match.length")
    end <- start + length - 1L

    matchstr <- substring(text, start, end)
    matchstr[start == -1] <- NA_character_

    res <- data.frame(
      stringsAsFactors = FALSE,
      .text = text,
      .match = matchstr
    )

    if (!is.null(attr(match, "capture.start"))) {
      gstart <- attr(match, "capture.start")
      glength <- attr(match, "capture.length")
      gend <- gstart + glength - 1L

      groupstr <- substring(text, gstart, gend)
      groupstr[gstart == -1] <- NA_character_
      dim(groupstr) <- dim(gstart)

      res <- cbind(groupstr, res, stringsAsFactors = FALSE)
    }

    names(res) <- c(attr(match, "capture.names"), ".text", ".match")
    res
  }

  null_file <- function() {
    if (get_os() == "windows") "nul:" else "/dev/null"
  }

  get_os <- function() {
    if (.Platform$OS.type == "windows") {
      "windows"
    } else if (Sys.info()[["sysname"]] == "Darwin") {
      "macos"
    } else if (Sys.info()[["sysname"]] == "Linux") {
      "linux"
    } else {
      "unknown"
    }
  }

  `%||%` <- function(l, r) if (is.null(l)) r else l

  #' Like [message()], but print to standard output in interactive
  #' sessions
  #'
  #' To avoid red output in RStudio, RGui, and R.app.
  #'
  #' @inheritParams message
  #' @noRd
  #' @return Nothing

  msg <- function(..., domain = NULL, appendLF = TRUE) {
    cnd <- .makeMessage(..., domain = domain, appendLF = appendLF)
    withRestarts(muffleMessage = function() NULL, {
      signalCondition(simpleMessage(cnd))
      output <- default_output()
      cat(cnd, file = output, sep = "")
    })
    invisible()
  }

  #' Where to print messages to
  #'
  #' If the session is not interactive, then it potentially matters
  #' whether we print to stdout or stderr, so we print to stderr.
  #'
  #' The same applies when there is a sink for stdout or stderr.
  #'
  #' @noRd
  #' @return The connection to print to.

  default_output <- function() {
    if (is_interactive() && no_active_sink()) stdout() else stderr()
  }

  no_active_sink <- function() {
    # See ?sink.number for the explanation
    sink.number("output") == 0 && sink.number("message") == 2
  }

  #' Smarter `interactive()`
  #'
  #' @noRd
  #' @return Logical scalar.

  is_interactive <- function() {
    opt <- getOption("rlib_interactive")
    opt2 <- getOption("rlang_interactive")
    if (isTRUE(opt)) {
      TRUE
    } else if (identical(opt, FALSE)) {
      FALSE
    } else if (isTRUE(opt2)) {
      TRUE
    } else if (identical(opt2, FALSE)) {
      FALSE
    } else if (tolower(getOption("knitr.in.progress", "false")) == "true") {
      FALSE
    } else if (identical(Sys.getenv("TESTTHAT"), "true")) {
      FALSE
    } else {
      base::interactive()
    }
  }

  #' Squote wrapper to avoid smart quotes
  #'
  #' @inheritParams sQuote
  #' @inherit sQuote return
  #' @noRd
  #' @return Character vector.

  squote <- function(x) {
    old <- options(useFancyQuotes = FALSE)
    on.exit(options(old), add = TRUE)
    sQuote(x)
  }

  #' Read all of a file
  #'
  #' @param path File to read.
  #' @param ... Passed to [readChar()].
  #' @noRd
  #' @return String.

  read_file <- function(path, ...) {
    readChar(path, nchars = file.info(path)$size, ...)
  }

  environment()
})

# nocov end

res_make_empty_df <- local({
  data <- NULL
  function() {
    if (is.null(data)) {
      data <<- data_frame(
        ref = character(),
        type = character(),
        direct = logical(),
        directpkg = logical(),
        status = character(), # "OK" or "FAILED"
        package = character(),
        version = character(),
        license = character(),
        needscompilation = logical(),
        priority = character(),
        md5sum = character(),
        sha256 = character(),
        filesize = integer(),
        built = character(),
        platform = character(), # "source" or platform string
        rversion = character(), # * or version number (prefix)
        repotype = character(),
        repodir = character(),
        target = character(),
        deps = list(),
        mirror = character(), # for CRAN/BioC
        sources = list(), # list of URLs
        remote = list(), # parsed remote ref
        error = list(), # list of errors
        metadata = list(), # named character of entries
        extra = list(), # any extra data (e.g. GitHub sha)
        dep_types = list(),
        params = list(),
        sysreqs = character(),
        os_type = character()
      )
    }
    data
  }
})

res_df_defaults <- local({
  data <- NULL
  function() {
    if (is.null(data)) {
      data <<- list(
        direct = FALSE,
        directpkg = FALSE,
        status = "OK",
        license = NA_character_,
        needscompilation = TRUE,
        priority = NA_character_,
        md5sum = NA_character_,
        sha256 = NA_character_,
        filesize = NA_integer_,
        built = NA_character_,
        platform = "source",
        rversion = "*",
        repotype = NA_character_,
        repodir = "src/contrib",
        target = quote(file.path(
          "src/contrib",
          paste0(package, "_", version, ".tar.gz")
        )),
        deps = list(make_null_deps()),
        remote = quote(parse_pkg_refs(ref)),
        error = list(list()),
        metadata = list(list()),
        mirror = NA_character_,
        extra = list(list()),
        dep_types = list("default"),
        params = list(character()),
        sysreqs = NA_character_,
        os_type = NA_character_
      )
    }
    data
  }
})

res_df_entry_types <- local({
  data <- NULL
  function() {
    if (is.null(data)) {
      data <<- vcapply(res_make_empty_df(), class)
    }
    data
  }
})

res_df_must_have <- local({
  data <- NULL
  function() {
    if (is.null(data)) {
      data <<- setdiff(names(res_make_empty_df()), names(res_df_defaults()))
    }
    data
  }
})

#' @noRd
#' @param df Resolution data frame.
#' @param entries List of entries to add.

res_add_df_entries <- function(df, entries) {
  if (!is.data.frame(entries)) {
    entries <- res_one_row_df(entries)
  }
  entries <- res_add_defaults(entries)
  ret <- as_data_frame(rbind(df, entries))[names(df)]
  direct <- ret$package[ret$direct]
  ret$directpkg <- ret$package %in% direct
  ret
}

res_one_row_df <- function(l) {
  assert_that(is.list(l) && all_named(l))
  samp <- res_make_empty_df()[names(l)]
  bad <- vlapply(samp, is.list) & !vlapply(l, is.list)
  # If col is isn't a list col, but is supposed to be.
  bad_len <- vlapply(samp, is.list) & !vlapply(l, function(x) length(x) == 1)
  # If col is supposed to be a list, but isn't length one.
  bad <- bad | bad_len
  l[bad] <- lapply(l[bad], list)
  out <- as_data_frame(l)
  assert_that(nrow(out) == 1)
  out
}

res_list_to_df <- function(reslist) {
  unk <- unlist(lapply(reslist, "[[", "unknown_deps"))
  df <- res_make_empty_df()
  for (res in reslist) {
    df <- res_add_df_entries(df, remove_entry(res, "unknown_deps"))
  }
  attr(df, "unknown_deps") <- unk
  df
}

res_add_defaults <- function(df) {
  if (length(bad <- setdiff(res_df_must_have(), names(df)))) {
    throw(pkg_error(
      "Entr{?y/ies} missing from package remote: {.val {bad}}",
      i = msg_internal_error()
    ))
  }

  all_types <- res_df_entry_types()
  miss <- setdiff(names(all_types), names(df))
  def <- lapply(
    res_df_defaults()[miss],
    eval,
    envir = df,
    enclos = environment()
  )
  df[names(def)] <- def
  df <- df[, names(all_types)]

  if ("filesize" %in% names(df)) {
    df$filesize <- as.integer(df$filesize)
  }

  ent_types <- vcapply(df, typeof)
  exp_types <- all_types[names(df)]
  if (any(bad <- ent_types != exp_types)) {
    items <- structure(
      paste0(
        names(df)[bad],
        " (",
        ent_types[bad],
        ", expected ",
        exp_types[bad],
        ")"
      ),
      names = rep("*", sum(bad))
    )
    throw(pkg_error(
      "{sum(bad)} wrong entry type{?s} in package remotes:",
      items
    ))
  }

  df
}

ts_languages <- c(r = 0L, markdown = 1L, "markdown-inline" = 2L, yaml = 3L)

s_expr <- function(
  code,
  language = c("r", "markdown", "markdown-inline", "yaml"),
  ranges = NULL
) {
  language <- tolower(language)
  language <- ts_languages[match.arg(language)]
  if (is.character(code)) {
    code <- charToRaw(paste(code, collapse = "\n"))
  }
  call_with_cleanup(c_s_expr, code, language, ranges)
}

code_query <- function(
  code = NULL,
  query,
  file = NULL,
  language = c("r", "markdown", "markdown-inline", "yaml"),
  ranges = NULL
) {
  language <- tolower(language)
  language <- ts_languages[match.arg(language)]
  qlen <- nchar(query, type = "bytes") + 1L # + \n
  qbeg <- c(1L, cumsum(qlen))
  qnms <- names(query) %||% rep(NA_character_, length(query))
  query1 <- paste0(query, "\n", collapse = "")

  if (!is.null(code)) {
    if (is.character(code)) {
      code <- charToRaw(paste(code, collapse = "\n"))
    }
    res <- call_with_cleanup(c_code_query, code, query1, language, ranges)
  } else {
    res <- call_with_cleanup(c_code_query_path, file, query1, language, ranges)
  }

  qorig <- as.integer(cut(res[[1]][[3]], breaks = qbeg, include.lowest = TRUE))

  list(
    patterns = data_frame(
      id = seq_along(res[[1]][[1]]),
      name = qnms[qorig],
      pattern = res[[1]][[1]],
      match_count = res[[1]][[2]]
    ),
    matched_captures = data_frame(
      id = viapply(res[[2]], "[[", 3L),
      pattern = viapply(res[[2]], "[[", 1L),
      match = viapply(res[[2]], "[[", 2L),
      start_byte = viapply(res[[2]], "[[", 6L),
      end_byte = viapply(res[[2]], "[[", 7L),
      start_row = viapply(res[[2]], "[[", 8L),
      start_column = viapply(res[[2]], "[[", 9L),
      end_row = viapply(res[[2]], "[[", 10L),
      end_column = viapply(res[[2]], "[[", 11L),
      name = vcapply(res[[2]], "[[", 4L),
      code = vcapply(res[[2]], "[[", 5L)
    )
  )
}

test_that("s_expr", {
  local_reproducible_output(width = 500)
  expect_snapshot(
    s_expr("f(arg, arg2 = 2); 1:100")
  )
})

test_that("code_query", {
  local_reproducible_output(width = 500)
  expect_snapshot({
    code_query("f(arg, arg2)", "(call(arguments))")
    code_query("f(arg, arg2)", "(call(arguments)) @call")
  })
})

test_that("code_query, multiple patterns", {
  local_reproducible_output(width = 500)
  expect_snapshot({
    code_query(
      "f(arg, arg2)",
      "(call(arguments)) @call (arguments) @args"
    )
    code_query(
      "f(arg, arg2)",
      c("(call(arguments)) @call", "(arguments) @args")
    )
  })
})

test_that("pattern names", {
  local_reproducible_output(width = 500)
  expect_snapshot({
    code_query("f('x')", c("(call) (call)", "(call)"))[["patterns"]]
    code_query("f('x')", c(a = "(call) (call)", "(call)"))[["patterns"]]
    code_query("f('x')", c("(call) (call)", b = "(call)"))[["patterns"]]
    code_query("f('x')", c(a = "(call) (call)", b = "(call)"))[["patterns"]]
  })
})

test_that("syntax error is handled", {
  local_reproducible_output(width = 500)
  expect_snapshot({
    code_query("f(1); g(1,2); 1+; h(3)", "(call) @call-code")
  })
})
test_that("code_query, field names", {
  local_reproducible_output(width = 500)
  expect_snapshot({
    code_query(
      "f(arg, x = arg2)",
      "(argument name: (identifier) @name value: (identifier) @value)"
    )
  })
})

test_that("code_query, negated fields", {
  local_reproducible_output(width = 500)
  expect_snapshot({
    # function calls without arguments
    code_query(
      "f(arg1, x = arg2); g(); a",
      "((call (arguments !argument)) @call-name)"
    )
  })
})

test_that("code_query, anonymous nodes", {
  local_reproducible_output(width = 500)
  expect_snapshot({
    code_query("1 + 2", '((float) @lhs "+" @op (float)) @rhs')
  })
})

test_that("code_query, capturing nodes", {
  local_reproducible_output(width = 500)
  expect_snapshot({
    # function calls
    code_query(
      "library(testthat); foo(bar); library(pak); bar(baz)",
      "(call function: (identifier) @fun-name)"
    )
  })
})

test_that("code_query, quantification operators", {
  local_reproducible_output(width = 500)
  expect_snapshot({
    # blocks of comments
    code_query(
      "# comment\n#comment 2\n1 + 5\n#comment 3\n",
      "(comment)+ @comments"
    )
  })
  expect_snapshot({
    # comments before function calls
    code_query(
      "f()\n# comment\ng()# comment1\n# comment 2\nh()\n",
      "((comment)* @comments (call function: (identifier)) @call)"
    )
  })
  expect_snapshot({
    # single optional comment before function call
    code_query(
      "f()\n# comment\ng()# comment1\n# comment 2\nh()\n",
      "((comment)? @comment (call function: (identifier)) @call)"
    )
  })
})

test_that("code_query, grouping sibling nodes", {
  local_reproducible_output(width = 500)
  expect_snapshot({
    code_query(
      "f(); g(); h()",
      "((call) @call-1 (call) @call-2 (call) @call-3)"
    )
  })
})

test_that("code_query, alternations", {
  local_reproducible_output(width = 500)
  # named or unnamed argument, capture the arg name if any
  expect_snapshot({
    code_query(
      "f(x=1); f(1)",
      paste0(
        "(call function: (identifier) arguments: (arguments argument: [",
        "  (argument value: (_)) ",
        "  (argument name: (identifier) @arg-name value: (_))",
        "]))"
      )
    )
  })
})

test_that("code_query, wildcard node", {
  local_reproducible_output(width = 500)
  expect_snapshot({
    # anything in a loop sequence
    code_query(
      "for (i in 1:10) foo",
      "(for_statement sequence: (_) @seq)"
    )
  })
})

test_that("code_query, anchors", {
  local_reproducible_output(width = 500)
  expect_snapshot({
    # single optional comment _right_ before function call
    code_query(
      "f()\n# comment\ng()# comment1\n# comment 2\nh()\n",
      "((comment)? @comment . (call function: (identifier)) @call)"
    )
  })
  expect_snapshot({
    # first argument
    code_query(
      "f(); f(x); f(x,y,z); f(x = 1, 2)",
      "(call arguments: (arguments . (argument) @arg))"
    )
  })
})

test_that("code_query, predicates, #eq?", {
  local_reproducible_output(width = 500)
  # multiple calls to the same function, from before and after comment
  q <- "
    (program
      (call function: (identifier) @fn-1) %s
      (comment)
      (call function: (identifier) @fn-2) %s
      (#eq? @fn-1 @fn-2)
    )"
  do <- function(q1, q2, code) {
    q_ <- sprintf(q, q1, q2)
    code_query(code, q_)
  }
  expect_snapshot({
    # no auantification
    do("", "", c("f()", "# c", "f()"))
    do("", "", c("f()", "# c", "g()"))
    # *
    do("*", "*", c("f()", "# c", "f()"))
    do("*", "*", c("f()", "# c", "g()"))
    do("*", "*", c("f()", "f(1)", "f(2)", "# c", "f()", "f(5)"))
    do("*", "*", c("f()", "f(1)", "f(2)", "# c", "f()", "g(5)"))
    do("*", "*", c("f()", "# c"))
    do("*", "*", c("# c", "f()"))
    # +
    do("+", "+", c("f()", "# c", "f()"))
    do("+", "+", c("f()", "# c", "g()"))
    do("+", "+", c("f()", "f(1)", "f(2)", "# c", "f()", "f(5)"))
    do("+", "+", c("f()", "f(1)", "f(2)", "# c", "f()", "g(5)"))
    do("+", "+", c("f()", "# c"))
    do("+", "+", c("# c", "f()"))
    # ?
    do("?", "?", c("f()", "# c", "f()"))
    do("?", "?", c("f()", "# c", "g()"))
    do("?", "?", c("f()", "f(1)", "f(2)", "# c", "f(5)"))
    do("?", "?", c("f()", "# c", "f(5)", "f()"))
    do("?", "?", c("f()", "# c"))
    do("?", "?", c("# c", "f()"))
  })
})

test_that("code_query, predicates, #not-eq?", {
  local_reproducible_output(width = 500)
  # multiple calls to the same function, from before and after comment
  q <- "
    (program
      (call function: (identifier) @fn-1) %s
      (comment)
      (call function: (identifier) @fn-2) %s
      (#not-eq? @fn-1 @fn-2)
    )"
  do <- function(q1, q2, code) {
    q_ <- sprintf(q, q1, q2)
    code_query(code, q_)
  }
  expect_snapshot({
    # no auantification
    do("", "", c("f()", "# c", "f()"))
    do("", "", c("f()", "# c", "g()"))
    # *
    do("*", "*", c("f()", "# c", "f()"))
    do("*", "*", c("f()", "# c", "g()"))
    do("*", "*", c("f()", "f(1)", "f(2)", "# c", "f()", "f(5)"))
    do("*", "*", c("f()", "f(1)", "f(2)", "# c", "h()", "g(5)"))
    do("*", "*", c("f()", "# c"))
    do("*", "*", c("# c", "f()"))
    # +
    do("+", "+", c("f()", "# c", "f()"))
    do("+", "+", c("f()", "# c", "g()"))
    do("+", "+", c("f()", "f(1)", "f(2)", "# c", "f()", "f(5)"))
    do("+", "+", c("f()", "f(1)", "f(2)", "# c", "h()", "g(5)"))
    do("+", "+", c("f()", "# c"))
    do("+", "+", c("# c", "f()"))
    # ?
    do("?", "?", c("f()", "# c", "f()"))
    do("?", "?", c("f()", "# c", "g()"))
    do("?", "?", c("f()", "f(1)", "f(2)", "# c", "g(5)"))
    do("?", "?", c("f()", "# c", "g(5)", "g()"))
    do("?", "?", c("f()", "# c"))
    do("?", "?", c("# c", "f()"))
  })
})

test_that("code_query, predicates, #any-eq?", {
  local_reproducible_output(width = 500)
  # multiple calls to the same function, from before and after comment
  q <- "
    (program
      (call function: (identifier) @fn-1) %s
      (comment)
      (call function: (identifier) @fn-2) %s
      (#any-eq? @fn-1 @fn-2)
    )"
  do <- function(q1, q2, code) {
    q_ <- sprintf(q, q1, q2)
    code_query(code, q_)
  }
  expect_snapshot({
    # no auantification
    do("", "", c("f()", "# c", "f()"))
    do("", "", c("f()", "# c", "g()"))
    # *
    do("*", "*", c("f()", "# c", "f()"))
    do("*", "*", c("f()", "# c", "g()"))
    do("*", "*", c("f()", "f(1)", "f(2)", "# c", "f()", "f(5)"))
    do("*", "*", c("f()", "f(1)", "f(2)", "# c", "g()", "g(5)"))
    do("*", "*", c("f()", "# c"))
    do("*", "*", c("# c", "f()"))
    # +
    do("+", "+", c("f()", "# c", "f()"))
    do("+", "+", c("f()", "# c", "g()"))
    do("+", "+", c("f()", "f(1)", "g(2)", "# c", "f()", "f(5)"))
    do("+", "+", c("f()", "f(1)", "f(2)", "# c", "g()", "g(5)"))
    do("+", "+", c("f()", "# c"))
    do("+", "+", c("# c", "f()"))
    # ?
    do("?", "?", c("f()", "# c", "f()"))
    do("?", "?", c("f()", "# c", "g()"))
    do("?", "?", c("f()", "f(1)", "g(2)", "# c", "f(5)"))
    do("?", "?", c("g()", "# c", "f(5)", "f()"))
    do("?", "?", c("f()", "# c"))
    do("?", "?", c("# c", "f()"))
  })
})

test_that("code_query, predicates, #any-not-eq?", {
  local_reproducible_output(width = 500)
  # multiple calls to the same function, from before and after comment
  q <- "
    (program
      (call function: (identifier) @fn-1) %s
      (comment)
      (call function: (identifier) @fn-2) %s
      (#any-not-eq? @fn-1 @fn-2)
    )"
  do <- function(q1, q2, code) {
    q_ <- sprintf(q, q1, q2)
    code_query(code, q_)
  }
  expect_snapshot({
    # no auantification
    do("", "", c("f()", "# c", "f()"))
    do("", "", c("f()", "# c", "g()"))
    # *
    do("*", "*", c("f()", "# c", "f()"))
    do("*", "*", c("f()", "# c", "g()"))
    do("*", "*", c("f()", "f(1)", "f(2)", "# c", "f()", "f(5)"))
    do("*", "*", c("f()", "f(1)", "f(2)", "# c", "g()", "g(5)"))
    do("*", "*", c("f()", "# c"))
    do("*", "*", c("# c", "f()"))
    # +
    do("+", "+", c("f()", "# c", "f()"))
    do("+", "+", c("f()", "# c", "g()"))
    do("+", "+", c("f()", "f(1)", "g(2)", "# c", "f()", "f(5)"))
    do("+", "+", c("f()", "f(1)", "f(2)", "# c", "g()", "g(5)"))
    do("+", "+", c("f()", "# c"))
    do("+", "+", c("# c", "f()"))
    # ?
    do("?", "?", c("f()", "# c", "f()"))
    do("?", "?", c("f()", "# c", "g()"))
    do("?", "?", c("f()", "f(1)", "g(2)", "# c", "f(5)"))
    do("?", "?", c("g()", "# c", "f(5)", "f()"))
    do("?", "?", c("f()", "# c"))
    do("?", "?", c("# c", "f()"))
  })
})

test_that("code_query, predicates, #eq? vs string", {
  local_reproducible_output(width = 500)
  # multiple calls to the same function, from before and after comment
  q <- "
    (program .
      (call function: (identifier) @fn) %s
      (#eq? @fn \"f\")
    . )"
  do <- function(q1, code) {
    q_ <- sprintf(q, q1)
    code_query(code, q_)
  }
  expect_snapshot({
    # no auantification
    do("", "f()")
    do("", "g()")
    # *
    do("*", "")
    do("*", c("f()", "f()"))
    do("*", c("f()", "g()", "f()"))
    # +
    do("+", "")
    do("+", c("f()", "f()"))
    do("+", c("f()", "g(1)"))
    # ?
    do("?", "")
    do("?", c("f()", "f()"))
    do("?", c("f()", "g()"))
  })
})

test_that("code_query, predicates, #not-eq? vs string", {
  local_reproducible_output(width = 500)
  # multiple calls to the same function, from before and after comment
  q <- "
    (program .
      (call function: (identifier) @fn) %s
      (#not-eq? @fn \"f\")
    . )"
  do <- function(q1, code) {
    q_ <- sprintf(q, q1)
    code_query(code, q_)
  }
  expect_snapshot({
    # no auantification
    do("", "f()")
    do("", "g()")
    # *
    do("*", "")
    do("*", c("f()", "f()"))
    do("*", c("g()", "h()"))
    do("*", c("f()", "g()", "f()"))
    # +
    do("+", "")
    do("+", c("f()", "f()"))
    do("+", c("g()", "h()"))
    do("+", c("f()", "g(1)"))
    # ?
    do("?", "")
    do("?", c("f()", "f()"))
    do("?", c("g()", "h()"))
    do("?", c("f()", "g()"))
  })
})

test_that("code_query, predicates, #any-eq? vs string", {
  local_reproducible_output(width = 500)
  # multiple calls to the same function, from before and after comment
  q <- "
    (program .
      (call function: (identifier) @fn) %s
      (#any-eq? @fn \"f\")
    . )"
  do <- function(q1, code) {
    q_ <- sprintf(q, q1)
    code_query(code, q_)
  }
  expect_snapshot({
    # no auantification
    do("", "f()")
    do("", "g()")
    # *
    do("*", "")
    do("*", c("f()", "f()"))
    do("*", c("f()", "g()", "f()"))
    # +
    do("+", "")
    do("+", c("f()", "f()"))
    do("+", c("f()", "g(1)"))
    # ?
    do("?", "")
    do("?", c("f()", "f()"))
    do("?", c("f()", "g()"))
  })
})

test_that("code_query, predicates, #any-not-eq? vs string", {
  local_reproducible_output(width = 500)
  # multiple calls to the same function, from before and after comment
  q <- "
    (program .
      (call function: (identifier) @fn) %s
      (#any-not-eq? @fn \"f\")
    . )"
  do <- function(q1, code) {
    q_ <- sprintf(q, q1)
    code_query(code, q_)
  }
  expect_snapshot({
    # no auantification
    do("", "f()")
    do("", "g()")
    # *
    do("*", "")
    do("*", c("f()", "f()"))
    do("*", c("g()", "h()"))
    do("*", c("f()", "g()", "f()"))
    # +
    do("+", "")
    do("+", c("f()", "f()"))
    do("+", c("g()", "h()"))
    do("+", c("f()", "g(1)"))
    # ?
    do("?", "")
    do("?", c("f()", "f()"))
    do("?", c("g()", "h()"))
    do("?", c("f()", "g()"))
  })
})

test_that("code_query, predicates, #match?", {
  local_reproducible_output(width = 500)
  # multiple calls to the same function, from before and after comment
  q <- "
    (program .
      (call function: (identifier) @fn) %s
      (#match? @fn \"^f\")
    . )"
  do <- function(q1, code) {
    q_ <- sprintf(q, q1)
    code_query(code, q_)
  }
  expect_snapshot({
    # no auantification
    do("", "f()")
    do("", "g()")
    # *
    do("*", "")
    do("*", c("f()", "f()"))
    do("*", c("f()", "g()", "f()"))
    # +
    do("+", "")
    do("+", c("f()", "f()"))
    do("+", c("f()", "g(1)"))
    # ?
    do("?", "")
    do("?", c("f()", "f()"))
    do("?", c("f()", "g()"))
  })
})

test_that("code_query, predicates, #not-match?", {
  local_reproducible_output(width = 500)
  # multiple calls to the same function, from before and after comment
  q <- "
    (program .
      (call function: (identifier) @fn) %s
      (#not-match? @fn \"^f\")
    . )"
  do <- function(q1, code) {
    q_ <- sprintf(q, q1)
    code_query(code, q_)
  }
  expect_snapshot({
    # no auantification
    do("", "f()")
    do("", "g()")
    # *
    do("*", "")
    do("*", c("f()", "f()"))
    do("*", c("g()", "h()"))
    do("*", c("f()", "g()", "f()"))
    # +
    do("+", "")
    do("+", c("f()", "f()"))
    do("+", c("g()", "h()"))
    do("+", c("f()", "g(1)"))
    # ?
    do("?", "")
    do("?", c("f()", "f()"))
    do("?", c("g()", "h()"))
    do("?", c("f()", "g()"))
  })
})

test_that("code_query, predicates, #any-match?", {
  local_reproducible_output(width = 500)
  # multiple calls to the same function, from before and after comment
  q <- "
    (program .
      (call function: (identifier) @fn) %s
      (#any-match? @fn \"^f\")
    . )"
  do <- function(q1, code) {
    q_ <- sprintf(q, q1)
    code_query(code, q_)
  }
  expect_snapshot({
    # no auantification
    do("", "f()")
    do("", "g()")
    # *
    do("*", "")
    do("*", c("f()", "f()"))
    do("*", c("f()", "g()", "f()"))
    # +
    do("+", "")
    do("+", c("f()", "f()"))
    do("+", c("f()", "g(1)"))
    # ?
    do("?", "")
    do("?", c("f()", "f()"))
    do("?", c("f()", "g()"))
  })
})

test_that("code_query, predicates, #any-not-match?", {
  local_reproducible_output(width = 500)
  # multiple calls to the same function, from before and after comment
  q <- "
    (program .
      (call function: (identifier) @fn) %s
      (#any-not-match? @fn \"^f\")
    . )"
  do <- function(q1, code) {
    q_ <- sprintf(q, q1)
    code_query(code, q_)
  }
  expect_snapshot({
    # no auantification
    do("", "f()")
    do("", "g()")
    # *
    do("*", "")
    do("*", c("f()", "f()"))
    do("*", c("g()", "h()"))
    do("*", c("f()", "g()", "f()"))
    # +
    do("+", "")
    do("+", c("f()", "f()"))
    do("+", c("g()", "h()"))
    do("+", c("f()", "g(1)"))
    # ?
    do("?", "")
    do("?", c("f()", "f()"))
    do("?", c("g()", "h()"))
    do("?", c("f()", "g()"))
  })
})

test_that("code_query, predicates, #any-of?", {
  local_reproducible_output(width = 500)
  # multiple calls to the same function, from before and after comment
  q <- "
    (program .
      (call function: (identifier) @fn) %s
      (#any-of? @fn \"f\" \"f2\")
    . )"
  do <- function(q1, code) {
    q_ <- sprintf(q, q1)
    code_query(code, q_)
  }
  expect_snapshot({
    # no auantification
    do("", "f()")
    do("", "g()")
    # *
    do("*", "")
    do("*", c("f()", "f2()"))
    do("*", c("f()", "g()", "f2()"))
    # +
    do("+", "")
    do("+", c("f()", "f2()"))
    do("+", c("f()", "g(1)"))
    # ?
    do("?", "")
    do("?", c("f()", "f2()"))
    do("?", c("f()", "g()"))
  })
})

test_that("code_query, predicates, #not-any-of?", {
  local_reproducible_output(width = 500)
  # multiple calls to the same function, from before and after comment
  q <- "
    (program .
      (call function: (identifier) @fn) %s
      (#not-any-of? @fn \"f\" \"f2\")
    . )"
  do <- function(q1, code) {
    q_ <- sprintf(q, q1)
    code_query(code, q_)
  }
  expect_snapshot({
    # no auantification
    do("", "f()")
    do("", "g()")
    # *
    do("*", "")
    do("*", c("f()", "f2()"))
    do("*", c("g()", "h()"))
    do("*", c("f()", "g()", "f2()"))
    # +
    do("+", "")
    do("+", c("f()", "f2()"))
    do("+", c("g()", "h()"))
    do("+", c("f()", "g(1)"))
    # ?
    do("?", "")
    do("?", c("f()", "f2()"))
    do("?", c("g()", "h()"))
    do("?", c("f()", "g()"))
  })
})

is_valid_package <- function(file) {
  if (!file.exists(file)) {
    FALSE
  } else if (grepl("\\.zip$", file)) {
    is_valid_package_zip(file)
  } else if (grepl("\\.tgz$|\\.tar\\.gz$", file)) {
    is_valid_package_targz(file)
  } else {
    ## Just ignore other files
    FALSE
  }
}

is_valid_package_zip <- function(file) {
  if (file.info(file)$size == 0) {
    return(FALSE)
  }
  tryCatch(
    is_package_file_list(file, suppressWarnings(zip_list(file))),
    error = function(e) FALSE
  )
}

#' @importFrom utils untar

is_valid_package_targz <- function(file) {
  if (file.info(file)$size == 0) {
    return(FALSE)
  }
  con <- gzfile(file, open = "rb")
  on.exit(close(con), add = TRUE)
  tryCatch(
    is_package_file_list(file, untar(con, list = TRUE)),
    error = function(e) FALSE
  )
}

is_package_file_list <- function(file, list) {
  pkgname <- pkg_name_from_file(file)

  ## A single directory, named after the package
  if (any(!grepl(paste0("^", pkgname, "\\b"), list))) {
    return(FALSE)
  }

  ## DESCRIPTION file
  if (!paste0(pkgname, "/DESCRIPTION") %in% list) {
    return(FALSE)
  }

  return(TRUE)
}

pkg_name_from_file <- function(x) {
  sub("^([a-zA-Z0-9\\.]+)_.*$", "\\1", basename(x))
}

#' The download progress bar
#'
#' This is a short summary of the design of the download progress bar.
#'
#' ## Package sizes
#'
#' One difficulty here is that we don't always know the sizes of the files
#' we are downloading. We basically have three cases for downloads:
#' * For some, we know the sizes from the extra CRAN metadata. This is
#'   usually correct, unless CRAN has changed the file on the web server.
#'   If we get a size from the web server, then we'll use that.
#' * For some, we do not know the size beforehand, but the web server
#'   sends us the correct size, together with the first chunk of data.
#'   This is typically for CRAN packages that we don't have metadata
#'   about, Bioconductor packages, or GitHub packages for which GitHub
#'   already knows the size.
#' * For some, we do not know the size, and the web server does not tell
#'   us the size, either. This is typically for GH packages that GH is
#'   creating on the fly.
#'
#' So ideally we would have a different progress bar when the total
#' size if known.
#'
#' ## Possible information to show:
#'
#' * Visual progress bar
#' * Number of packages: done / in-progress / total
#' * Number of bytes: done / total (if known)
#' * ETA
#' * Download rate, i.e. x kB/s
#' * Elapsed time
#' * Percent of packages done
#' * Percent of bytes done
#' * Spinner to show that we are still working, although updating ETA is
#'   good as well...
#' * Events, e.g. got a package, downloading, or starting...
#'
#' ## All package sizes are known
#'
#' ```
#' 230 kB/s (#####      ) 45% | 14/56 pkgs | ETA ~34s | Getting dplyr
#' ```
#'
#' * Rate is informative, serves as an "alive" indicator as well.
#' * Bar is visual.
#' * Percent is the best quantitative measure.
#' * Done/total packages is informative, not so much for the download
#'   but the whole installation.
#' * ETA is informative.
#' * Event is nice, and it fills the space...
#' * We don't show current/total bytes, shoulwd we? Quite wide...
#'
#' ## Some package sizes are unknown
#'
#' ```
#' 230 kB/s (#####      ) 45% | 14/56 pkgs | Getting dplyr
#' ```
#'
#' * Percent is for the packages, not bytes.
#' * We can't really show ETA.
#'
#' ## Events
#'
#' This is quite simple currently, we just print the packages we are
#' "getting" currently, i.e. the packages we received data for in the
#' last time slot.
#'
#' ## Rate (download "speed")
#'
#' To calculate the rate we need to record when the various data chunks
#' have arrived. We only do this with a resolution of 1s. We create an
#' environment called chunks, and for every second since the downloads have
#' started we record the number of bytes that arrived in that second.
#'
#' When calculating the rate, we take the data for the last 4 seconds,
#' and normalize it to per second, while taking into account that the
#' data for the last (current) second is incomplete.
#'
#' In the first 3 seconds we use all the available data.
#'
#' @name pkgdepends-download-progress-bar
#' @noRd
NULL

pkgplan__create_progress_bar <- function(what) {
  bar <- new.env(parent = emptyenv())

  bar$what <- what[, c("type", "filesize", "package", "cache_status")]
  bar$what$idx <- seq_len(nrow(what))
  bar$what$current <- 0L # We got this many bytes
  bar$what$need <- bar$what$filesize
  bar$what$status <- "todo" # "todo", "data", "got", "had", "error"
  bar$what$skip <-
    what$type %in% c("installed", "deps") | what$cache_status != "miss"
  bar$what$status[bar$what$skip] <- "skip"
  bar$what$cache_status[what$type %in% c("installed", "deps")] <- NA_character_

  pkgplan__initial_pb_message(bar)

  bar$chars <- progress_chars()
  bar$chunks <- new.env(parent = emptyenv())
  bar$start_at <- Sys.time()
  bar$events <- list()
  bar$lastmsg <- "Connecting..."
  bar$show_time <- tolower(Sys.getenv("PKG_OMIT_TIMES")) != "true"
  bar$show_size <- tolower(Sys.getenv("PKG_OMIT_SIZES")) != "true"

  bar
}

pkgplan__init_progress_bar <- function(bar) {
  bar$timer <- new_async_timer(
    1 / 10,
    function() pkgplan__show_progress_bar(bar)
  )
  bar$timer$listen_on("error", function(e) {
    stop(e)
  })

  bar
}

pkgplan__initial_pb_message <- function(bar) {
  # number of packages to _download_
  num <- sum(bar$what$status == "todo")
  # number of packages to download, with unknown size
  unk <- sum(is.na(bar$what$filesize[bar$what$status == "todo"]))
  # number of bytes to download
  bts <- sum(bar$what$filesize[bar$what$status == "todo"], na.rm = TRUE)
  # number of packages that are (surely) cached
  # for installed and deps refs this should be NA
  nch <- sum(bar$what$cache_status %in% "hit")
  # the number of bytes cached (for known file sizes)
  cbt <- sum(bar$what$filesize[bar$what$cache_status %in% "hit"], na.rm = TRUE)

  if (num == 0) {
    cli::cli_alert_info(c(
      "No downloads are needed",
      if (nch > 0) ", {nch} pkg{?s} ",
      if (cbt > 0) "{.size ({format_bytes$pretty_bytes(cbt)})} ",
      if (nch > 0) "{cli::qty(nch)}{?is/are} cached"
    ))
  } else {
    cli::cli_alert_info(c(
      "Getting",
      if (bts > 0) {
        " {num-unk} pkg{?s} {.size ({format_bytes$pretty_bytes(bts)})}"
      },
      if (bts > 0 && unk > 0) " and",
      if (unk > 0) " {unk} pkg{?s} with unknown size{?s}",
      if (nch > 0) ", {nch} ",
      if (cbt > 0) "{.size ({format_bytes$pretty_bytes(cbt)})} ",
      if (nch > 0) "cached"
    ))
  }
  if (should_show_progress_bar()) {
    bar$status <- cli::cli_status("", .auto_close = FALSE)
  }
}

#' Update the progress bar data
#'
#' This is triggered by libcurl.
#'
#' @param bar The progress bar.
#' @param idx The row index in the download data table.
#' @param data Data about this row, a list with entries `current` and
#'   `total`, measured in the number of bytes. If `total` is unknown,
#'   then it is set to zero. At the end of the download we also get
#'   `"done"`, and on error `"error"`.
#'
#' @noRd

pkgplan__update_progress_bar <- function(bar, idx, event, data) {
  # Record the time here, and use it in this function, so that this
  # function runs in a single point of time
  time <- Sys.time()

  # Work out which second we are in. We record the received data per second
  sec <- as.character(floor(as.double(time - bar$start_at, units = "secs")))

  # If it is "done", then we are done. We don't make assumptions about
  # libcurl/async signalling the end of every file via file sizes in the
  # progress bar callback, so we assume that when we get "done", then we
  # are indeed "done" and set the status and sizes accordingly.
  if (event == "done") {
    if (data$download_status == "Got") {
      bar$what$status[idx] <- "got"
      sz <- na.omit(file.size(c(data$fulltarget, data$fulltarget_tree)))[1]
      if (!is.na(sz)) {
        bar$what$filesize[idx] <- sz
      }
      cli::cli_alert_success(c(
        "Got {.pkg {data$package}} ",
        "{.version {data$version}} ({data$platform})",
        if (!is.na(sz) && bar$show_size) {
          " {.size ({format_bytes$pretty_bytes(sz)})}"
        }
      ))
      if (!is.na(bar$what$filesize[idx])) {
        bar$chunks[[sec]] <- (bar$chunks[[sec]] %||% 0) -
          bar$what$current[idx] +
          bar$what$filesize[idx]
        bar$what$current[idx] <- bar$what$filesize[idx]
      }
    } else if (grepl("^(Had|Current)", data$download_status)) {
      bar$what$status[idx] <- "had"
      bar$what$current[idx] <- 0L
      bar$what$need[idx] <- 0L
      if (identical(data$cache_status, "miss") && data$type != "deps") {
        cli::cli_alert_success(c(
          "Cached copy of {.pkg {data$package}} ",
          "{.version {data$version}} ({data$platform}) is the latest build"
        ))
      }
    } else if (data$download_status == "Failed") {
      cli::cli_alert_danger(c(
        "Failed to download {.pkg {data$package}} ",
        "{.version {data$version}} ({data$platform})"
      ))
      bar$what$status[idx] <- "error"
      bar$what$need[idx] <- bar$what$current[idx]
    } else {
      stop("Unknown download status, internal pkgdepends error :(")
    }

    return(TRUE)
  }

  if (event == "error") {
    cli::cli_alert_danger(c(
      "Failed to download {.pkg {data$package}} ",
      "{.version {data$version}} ({data$platform})"
    ))
    bar$what$status[idx] <- "error"
    bar$what$need[idx] <- bar$what$current[idx]

    return(TRUE)
  }

  # Otherwise we got a chunk of data
  bar$what$status[idx] <- "data"
  bar$events$data <- unique(c(bar$events$data, idx))

  # Update data chunks
  bar$chunks[[sec]] <- (bar$chunks[[sec]] %||% 0) -
    bar$what$current[idx] +
    data$current

  # Update current and total
  bar$what$current[idx] <- data$current
  if (data$total > 0) {
    bar$what$filesize[idx] <- bar$what$need[idx] <- data$total
  }

  TRUE
}

#' Show / update the progress bar
#'
#' @param bar The progress bar object.
#'
#' @noRd

pkgplan__show_progress_bar <- function(bar) {
  if (is.null(bar$status)) {
    return()
  }

  # Don't show if there is nothing to download
  if (sum(!bar$what$skip) == 0) {
    return()
  }
  parts <- calculate_progress_parts(bar)

  # Ready to update. We can't use the package emoji because its
  # width is not calculated properly
  str <- paste0(
    " {parts$rate} {parts$line} {parts$percent} ",
    "| {parts$pkg_done}/{parts$pkg_total} pkg{?s} ",
    if (!is.na(parts$bytes_total)) "| ETA {parts$eta} ",
    "| {parts$msg}"
  )

  bar$events <- list()
  cli::cli_status_update(bar$status, str)
}

calculate_rate <- function(start, now, chunks) {
  # Rate, see above how this works
  time_at <- as.double(now - start, units = "secs")
  time_at_s <- as.integer(floor(time_at))
  labels <- as.character(seq(time_at_s, time_at_s - 3L, by = -1L))
  data <- unlist(mget(labels, envir = chunks, ifnotfound = 0L))
  fact <- time_at - max(time_at_s - 3, 0)
  rate <- sum(data) / fact
  if (is.nan(rate)) {
    rate <- 0
  }
  if (rate == 0 && time_at < 4) {
    rstr <- strrep(" ", 8)
  } else {
    rstr <- paste0(format_bytes$pretty_bytes(rate, style = "6"), "/s")
  }
  list(rate = rate, rstr = rstr)
}

calculate_eta <- function(total, current, rate) {
  if (rate == 0) {
    etas <- NA
    estr <- "??s "
  } else {
    todo <- total - current
    etas <- as.difftime(todo / rate, units = "secs")
    if (etas < 1) {
      estr <- "<1s   "
    } else {
      estr <- format(format_time$pretty_dt(etas, compact = TRUE), width = 6)
    }
  }
  list(etas = etas, estr = estr)
}

calculate_progress_parts <- function(bar) {
  parts <- list()

  # We filter these here, instead at the beginning, because otherwise
  # the indices would not work in the update function
  whatx <- bar$what[!bar$what$skip, ]

  now <- Sys.time()

  # Simple numbers
  pkg_done <- sum(whatx$status %in% c("got", "had", "error"))
  pkg_total <- nrow(whatx)
  parts$pkg_done <- format(c(pkg_done, pkg_total), justify = "right")[1]
  parts$pkg_total <- as.character(pkg_total)
  pkg_percent <- pkg_done / pkg_total
  bytes_done <- sum(whatx$current, na.rm = TRUE)
  bytes_total <- sum(whatx$need) # could be NA
  parts$bytes_total <- bytes_total
  bytes_percent <- bytes_done / bytes_total # could be NA
  percent <- if (!is.na(bytes_percent)) bytes_percent else pkg_percent
  if (round(percent * 100) == 100 && percent < 1) {
    percent <- 0.99
  }
  parts$percent <- format(
    paste0(round(100 * percent), "%"),
    width = 4,
    justify = "right"
  )

  rate <- calculate_rate(bar$start_at, now, bar$chunks)
  parts$rate <- rate$rstr

  # Message
  parts$msg <- bar$lastmsg
  if (length(bar$events$data) > 0) {
    pkgs <- bar$what$package[bar$what$idx %in% bar$events$data]
    parts$msg <- paste0(
      "Getting ",
      cli::ansi_collapse(pkgs, sep = ", ", last = " and ")
    )
    bar$lastmsg <- parts$msg
  }

  # Line
  parts$line <- make_bar(bar$chars, percent, width = 15)

  # ETA
  if (!is.na(bytes_total)) {
    parts$eta <- calculate_eta(bytes_total, bytes_done, rate$rate)$estr
  }

  parts
}

pkgplan__done_progress_bar <- function(bar) {
  if (is.null(bar$status)) {
    return()
  }

  end_at <- Sys.time()
  dt <- format_time$pretty_dt(Sys.time() - bar$start_at)

  cli::cli_status_clear(bar$status)
  bar$status <- NULL

  bts <- format_bytes$pretty_bytes(sum(bar$what$current))
  dld <- sum(bar$what$status == "got")
  cch <- sum(bar$what$status == "had")
  err <- sum(bar$what$status == "error")

  if (sum(!bar$what$skip) == 0) {
    # Print nothing, we already printed that no downloads are needed
  } else if (err == 0 && dld == 0) {
    cli::cli_alert_success("No downloads needed, all packages are cached")
  } else if (err == 0) {
    cli::cli_alert_success(
      paste0(
        "Downloaded {dld} package{?s} {.size ({bts})}",
        if (bar$show_time) " in {.time {dt}}"
      )
    )
  } else {
    cli::cli_alert_danger(
      "Failed to download {err} package{?s}. "
    )
  }
}

test_that("make_bar", {
  chars <- withr::with_options(list(cli.unicode = FALSE), progress_chars())
  lapply(
    seq(0, 1, by = 0.2),
    function(p) {
      expect_equal(nchar(cli::ansi_strip(make_bar(chars, p, 10))), 10)
    }
  )
})

test_that("initial state", {
  local_cli_config()

  # num = 0, nch = 0, cbt = 0
  # no downloads are needed, all installed
  what <- tibble::tribble(
    ~type         ,
    ~filesize     ,
    ~package      ,
    ~cache_status ,
    "installed"   ,
            10000 ,
    "foo"         ,
    NA_character_ ,
    "installed"   ,
            10000 ,
    "foo2"        ,
    "miss"        ,
  )
  expect_snapshot(invisible(
    pkgplan__create_progress_bar(what)
  ))

  # num = 0, nch = 1, cbt > 0
  # no downloads needed, all cached or installed
  what <- tibble::tribble(
    ~type         ,
    ~filesize     ,
    ~package      ,
    ~cache_status ,
    "cran"        ,
            10000 ,
    "foo"         ,
    "hit"         ,
    "installed"   ,
            10000 ,
    "foo2"        ,
    "hit"
  )
  expect_snapshot(invisible(
    pkgplan__create_progress_bar(what)
  ))

  # num = 0, nch = 1, cbt > 0
  # no downloads needed, but cached size known
  what <- tibble::tribble(
    ~type         ,
    ~filesize     ,
    ~package      ,
    ~cache_status ,
    "cran"        ,
            10000 ,
    "foo"         ,
    "hit"         ,
    "cran"        ,
            10000 ,
    "foo2"        ,
    "hit"         ,
    "installed"   ,
            10000 ,
    "foo3"        ,
    "hit"
  )
  expect_snapshot(invisible(
    pkgplan__create_progress_bar(what)
  ))

  # num = 0, nch = 1, cbt = 0
  # no downloads needed, cached size unknown
  what <- tibble::tribble(
    ~type         ,
    ~filesize     ,
    ~package      ,
    ~cache_status ,
    "cran"        ,
    NA_integer_   ,
    "foo"         ,
    "hit"         ,
    "installed"   ,
            10000 ,
    "foo2"        ,
    "hit"
  )
  expect_snapshot(invisible(
    pkgplan__create_progress_bar(what)
  ))

  # num > 0, bts > 0, unk = 0, nch = 0, cbt = 0
  # nothing is cached, sizes known
  what <- tibble::tribble(
    ~type         ,
    ~filesize     ,
    ~package      ,
    ~cache_status ,
    "cran"        ,
            10000 ,
    "foo"         ,
    "miss"        ,
    "cran"        ,
            20000 ,
    "bar"         ,
    "miss"
  )
  expect_snapshot(invisible(
    pkgplan__create_progress_bar(what)
  ))

  # num > 0, bts > 0, unk = 0, nch = 1, cbt = 0
  # 1 package cached, 2 downloaded, cached size unknown
  what <- tibble::tribble(
    ~type         ,
    ~filesize     ,
    ~package      ,
    ~cache_status ,
    "cran"        ,
            10000 ,
    "foo"         ,
    "miss"        ,
    "cran"        ,
            20000 ,
    "bar"         ,
    "miss"        ,
    "installed"   ,
            10000 ,
    "foo2"        ,
    NA_character_ ,
    "cran"        ,
    NA_integer_   ,
    "bar2"        ,
    "hit"
  )
  expect_snapshot(invisible(
    pkgplan__create_progress_bar(what)
  ))

  # num > 0, bts > 0, unk = 0, nch = 1, cbt > 0
  # 1 package cached, 2 downloaded, cached size known
  what <- tibble::tribble(
    ~type         ,
    ~filesize     ,
    ~package      ,
    ~cache_status ,
    "cran"        ,
            10000 ,
    "foo"         ,
    "miss"        ,
    "cran"        ,
            20000 ,
    "bar"         ,
    "miss"        ,
    "installed"   ,
            10000 ,
    "foo2"        ,
    NA_character_ ,
    "cran"        ,
            20000 ,
    "bar2"        ,
    "hit"
  )
  expect_snapshot(invisible(
    pkgplan__create_progress_bar(what)
  ))

  # num > 0, bts > 0, unk > 0, nch = 0, cbt = 0
  # downloads with unknown sizes
  what <- tibble::tribble(
    ~type         ,
    ~filesize     ,
    ~package      ,
    ~cache_status ,
    "cran"        ,
            20000 ,
    "foo"         ,
    "miss"        ,
    "cran"        ,
    NA_integer_   ,
    "bar"         ,
    "miss"        ,
    "installed"   ,
            10000 ,
    "foo2"        ,
    NA_character_
  )
  expect_snapshot(invisible(
    pkgplan__create_progress_bar(what)
  ))

  # num > 0, bts > 0, unk > 0, nch = 1, cbt > 0
  # downloads with known & unknown sizes, 1 cached with known size
  what <- tibble::tribble(
    ~type         ,
    ~filesize     ,
    ~package      ,
    ~cache_status ,
    "cran"        ,
            10000 ,
    "foo"         ,
    "miss"        ,
    "cran"        ,
    NA_integer_   ,
    "bar"         ,
    "miss"        ,
    "installed"   ,
            10000 ,
    "foo2"        ,
    NA_character_ ,
    "cran"        ,
            30000 ,
    "bar2"        ,
    "hit"
  )
  expect_snapshot(invisible(
    pkgplan__create_progress_bar(what)
  ))

  # num > 0, bts > 0, unk > 0, nch = 1, cbt = 0
  # downloads with known sizes, 1 cached with unknown size
  what <- tibble::tribble(
    ~type         ,
    ~filesize     ,
    ~package      ,
    ~cache_status ,
    "cran"        ,
            10000 ,
    "foo"         ,
    "miss"        ,
    "cran"        ,
    NA_integer_   ,
    "bar"         ,
    "miss"        ,
    "installed"   ,
            10000 ,
    "foo2"        ,
    NA_character_ ,
    "cran"        ,
    NA_integer_   ,
    "bar2"        ,
    "hit"
  )
  expect_snapshot(invisible(
    pkgplan__create_progress_bar(what)
  ))

  # num > 0, bts = 0, unk > 0, nch = 0, cbt = 0
  # downloads with only unknown sizes, nothing cached
  what <- tibble::tribble(
    ~type         ,
    ~filesize     ,
    ~package      ,
    ~cache_status ,
    "cran"        ,
    NA_integer_   ,
    "foo"         ,
    "miss"        ,
    "cran"        ,
    NA_integer_   ,
    "bar"         ,
    "miss"        ,
    "installed"   ,
            10000 ,
    "foo2"        ,
    NA_character_
  )
  expect_snapshot(invisible(
    pkgplan__create_progress_bar(what)
  ))

  # num > 0, bts = 0, unk > 0, nch = 1, cbt = 0
  # downloads with unknown sizes, 1 cached with unknown size
  what <- tibble::tribble(
    ~type         ,
    ~filesize     ,
    ~package      ,
    ~cache_status ,
    "cran"        ,
    NA_integer_   ,
    "foo"         ,
    "miss"        ,
    "cran"        ,
    NA_integer_   ,
    "bar"         ,
    "miss"        ,
    "installed"   ,
            10000 ,
    "foo2"        ,
    NA_character_ ,
    "cran"        ,
    NA_integer_   ,
    "bar2"        ,
    "hit"
  )
  expect_snapshot(invisible(
    pkgplan__create_progress_bar(what)
  ))

  # num > 0, bts = 0, unk > 0, nch = 1, cbt > 0
  # downloads with unknown sizes, 1 cached
  what <- tibble::tribble(
    ~type         ,
    ~filesize     ,
    ~package      ,
    ~cache_status ,
    "cran"        ,
    NA_integer_   ,
    "foo"         ,
    "miss"        ,
    "cran"        ,
    NA_integer_   ,
    "bar"         ,
    "miss"        ,
    "installed"   ,
            10000 ,
    "foo2"        ,
    NA_character_ ,
    "cran"        ,
            30000 ,
    "bar2"        ,
    "hit"
  )
  expect_snapshot(invisible(
    pkgplan__create_progress_bar(what)
  ))
})

test_that("data updates", {
  what <- tibble::tribble(
    ~type         ,
    ~filesize     ,
    ~package      ,
    ~cache_status ,
    "cran"        ,
            10000 ,
    "foo"         ,
    "miss"        ,
    "installed"   ,
            10000 ,
    "foo2"        ,
    "miss"        , # no download, installed
    "cran"        ,
            20000 ,
    "bar"         ,
    "hit"         , # no download, cached
    "cran"        ,
            20000 ,
    "bar2"        ,
    "miss"
  )

  expect_snapshot({
    bar <- pkgplan__create_progress_bar(what = what)
    pkgplan__update_progress_bar(
      bar,
      1L,
      "data",
      list(current = 5000, total = 10000)
    )
    pkgplan__update_progress_bar(
      bar,
      4L,
      "data",
      list(current = 20000, total = 20000)
    )
    pkgplan__done_progress_bar(bar)
    bar$what
  })
})

test_that("all finish messages for updates", {
  do <- function(
    idx,
    event = "done",
    download_status = "Got",
    make_tempfile = TRUE
  ) {
    bar <- pkgplan__create_progress_bar(what = what)
    tmp <- tempfile()
    if (make_tempfile) {
      writeBin(raw(7), tmp)
      on.exit(unlink(tmp), add = TRUE)
    }
    pkgplan__update_progress_bar(
      bar,
      idx,
      event,
      list(
        download_status = download_status,
        package = "foo",
        platform = "source",
        version = "1.0.0",
        fulltarget = tmp,
        fulltarget_tree = tempfile()
      )
    )
    pkgplan__done_progress_bar(bar)
    bar$what
  }

  what <- tibble::tribble(
    ~type         ,
    ~filesize     ,
    ~package      ,
    ~cache_status ,
    "cran"        ,
            10000 ,
    "foo"         ,
    "miss"        ,
    "installed"   ,
            10000 ,
    "foo2"        ,
    "miss"        , # no download, installed
    "cran"        ,
            20000 ,
    "bar"         ,
    "hit"         , # no download, cached
    "cran"        ,
            20000 ,
    "bar2"        ,
    "miss"
  )

  expect_snapshot(do(1L))
  expect_snapshot(do(1L, make_tempfile = FALSE))
  expect_snapshot(do(1L, event = "error"))
})

test_that("rate", {
  now <- Sys.time()
  chunks <- as.environment(list(
    "0" = 1000,
    "1" = 2000,
    "2" = 0,
    "3" = 400,
    "4" = 100,
    "5" = 200
  ))

  r <- calculate_rate(now - 5.5, now, chunks)
  expect_equal(r, list(rate = 200, rstr = "0.2 kB/s"))
})

test_that("eta", {
  expect_equal(
    calculate_eta(100, 50, 0),
    list(etas = NA, estr = "??s ")
  )
  expect_equal(
    calculate_eta(100, 50, 10),
    list(etas = as.difftime(5, units = "secs"), estr = "~5s   ")
  )
})

test_that("parts are calculated properly", {
  local_cli_config()
  what <- tibble::tribble(
    ~type         ,
    ~filesize     ,
    ~package      ,
    ~cache_status ,
    "cran"        ,
            10000 ,
    "foo"         ,
    "miss"        ,
    "cran"        ,
            20000 ,
    "bar"         ,
    "miss"        ,
    "installed"   ,
            10000 ,
    "foo2"        ,
    NA_character_ ,
    "cran"        ,
            20000 ,
    "bar2"        ,
    "hit"
  )

  expect_snapshot(
    bar <- pkgplan__create_progress_bar(what)
  )
  parts <- calculate_progress_parts(bar)
  expect_equal(parts$pkg_done, "0")
  expect_equal(parts$pkg_total, "2")
  expect_equal(parts$percent, "  0%")
  expect_match(parts$rate, "^\\s+$")
  expect_equal(parts$msg, "Connecting...")
  expect_match(cli::ansi_strip(parts$line), "^[(][\\s\u00a0]+[)]$", perl = TRUE)
  expect_equal(parts$eta, "??s ")
})

test_that("parts if file size is unknown", {
  # TODO
  expect_true(TRUE)
})

test_that("status bar is cleaned up on interrupt / error", {
  # TODO
  expect_true(TRUE)
})

#' List platforms with system requirements support
#'
#' @usage
#' sysreqs_platforms()
#'
#' @return
#' Data frame with columns:
#' * `name`: human readable OS name.
#' * `os`: OS name, e.g. `linux`.
#' * `distribution`: OS id, e.g. `ubuntu` or `redhat`.
#' * `version`: distribution version. A star means that all versions are
#'    supported, that are also supported by the vendor.
#' * `update_command`: command to run to update the system package metadata.
#' * `install_command`: command to run to install packages.
#' * `query_command`: name of the tool to use to query system package
#'    information.
#'
#' @name sysreqs_platforms
#' @export
#' @family system requirements functions
#' @examplesIf Sys.getenv("IN_PKGDOWN") == "true"
#' sysreqs_platforms()

sysreqs_platforms

#' Check if a platform has system requirements support
#'
#' @usage
#' sysreqs_is_supported(sysreqs_platform = NULL)
#'
#' @param sysreqs_platform System requirements platform. If `NULL`, then the
#'   `sysreqs_platform` \eval{man_config_link("configuration option")}
#'   is used, which defaults to the current platform. Set this option if
#'   \eval{.packageName} does not detect your platform correctly.
#' @return Logical scalar.
#'
#' @name sysreqs_is_supported
#' @export
#' @family system requirements functions
#' @seealso The `sysreqs_platform`
#'   \eval{man_config_link("configuration option")}.
#' @examplesIf Sys.getenv("IN_PKGDOWN") == "true"
#' sysreqs_is_supported()

sysreqs_is_supported

#' List contents of the system requirements DB, for a platform
#'
#' It also tries to update the system dependency database, if it is
#' outdated. (I.e. older than allowed in the `metadata_update_after`
#' \eval{man_config_link("configuration option")}.
#'
#' @usage
#' sysreqs_db_list(sysreqs_platform = NULL)
#'
#' @param sysreqs_platform System requirements platform. If `NULL`, then the
#'   `sysreqs_platform` \eval{man_config_link("configuration option")}
#'   is used, which defaults to the current platform. Set this option if
#'   \eval{.packageName} does not detect your platform correctly.
#' @return Data frame with columns:
#' * `name`: cross platform system dependency name in the database.
#' * `patterns`: one or more regular expressions to match to
#'   `SystemRequirements` fields.
#' * `packages`: one or more system package names to install.
#' * `pre_install`: command(s) to run before installing the packages.
#' * `post_install`:: command(s) to run after installing the packages.
#'
#' @name sysreqs_db_list
#' @export
#' @family system requirements functions
#' @examplesIf Sys.getenv("IN_PKGDOWN") == "true"
#' sysreqs_db_list(sysreqs_platform = "ubuntu-22.04")

sysreqs_db_list

#' Match system requirement descriptions to the database
#'
#' In the usual workflow `r pak_or_pkgdepends()` matches the
#' `SystemRequirements` fields of the `DESCRIPTION` files to the database.
#'
#' The `sysreqs_db_match()` function lets you match any string, and it is
#' mainly useful for debugging.
#'
#' @usage
#' sysreqs_db_match(specs, sysreqs_platform = NULL)
#'
#' @param specs Character vector of system requirements descriptions.
#' @param sysreqs_platform System requirements platform. If `NULL`, then the
#'   `sysreqs_platform` \eval{man_config_link("configuration option")}
#'   is used, which defaults to the current platform. Set this option if
#'   \eval{.packageName} does not detect your platform correctly.
#' @return Data frame with columns:
#' * `spec`: the input `specs`.
#' * `sysreq`:  name of the system library or tool.
#' * `packages`: system packages, list column of character vectors.
#'    Rarely it can be an empty string, e.g. if a `pre_install` script
#'    performs the installation.
#' * `pre_install`: list column of character vectors. Shell script(s) to
#'    run before the installation.
#' * `post_install`: list column of character vectors. Shell script(s) to
#'    run after the installation.
#'
#' @name sysreqs_db_match
#' @export
#' @family system requirements functions
#' @examplesIf Sys.getenv("IN_PKGDOWN") == "true"
#' sysreqs_db_match(
#'   c("Needs libcurl", "Java, libssl"),
#'   sysreqs_platform = "ubuntu-22.04"
#' )

sysreqs_db_match

#' Update the cached copy of the system requirements database
#'
#' @usage
#' sysreqs_db_update()
#'
#' @details
#' If the the cached copy is recent, then no update is attempted. See the
#' `metadata_update_after` \eval{man_config_link("configuration option")}.
#'
#' @name sysreqs_db_update
#' @export
#' @family system requirements functions

sysreqs_db_update

#' Check if installed packages have all their system requirements
#'
#' @usage
#' sysreqs_check_installed(packages = NULL, library = .libPaths()[1])
#' sysreqs_fix_installed(packages = NULL, library = .libPaths()[1])
#'
#' @description
#' `sysreqs_check_installed()` checks if the system requirements of all
#' packages (or a subset of packages) are installed.
#'
#' `sysreqs_fix_installed()` installs the missing system packages.
#'
#' @details
#' These functions use the `sysreqs_platform` configuration option,
#' see \eval{man_config_link("Configuration")}. Set this if
#' `r pak_or_pkgdepends()` does not detect your platform correctly.
#'
#' @param packages If not `NULL`, then only these packages are checked.
#'   If a package in `packages` is not installed, then
#'   `r pak_or_pkgdepends()` throws a warning.
#' @param library Library or libraries to check.
#' @return Data frame with a custom print and format method, and a
#'   `pkg_sysreqs_check_result` class. Its columns are:
#'    * `system_package`: string, name of the required system package.
#'    * `installed`: logical, whether the system package is correctly
#'      installed.
#'    * `packages`: list column of character vectors. The names of the
#'      installed R packages that need this system package.
#'    * `pre_install`: list column of character vectors. Commands to run
#'      before the installation of the the system package.
#'    * `post_install`: list column of character vectors. Commands to run
#'      after the installation of the system package.
#'
#' The data frame also have two attributes with additional data:
#' * `sysreqs_records`: the raw system requirements records, and
#' * `system_packages`: the list of the installed system packages.
#'
#' `sysreqs_fix_packages()` returns the same value, but invisibly.
#'
#' @name sysreqs_check_installed
#' @aliases sysreqs_fix_installed
#' @export
#' @family system requirements functions
#' @examplesIf Sys.getenv("IN_PKGDOWN") == "true" && Sys.info()[["sysname"]] == "Linux"
#' # This only works on supported platforms
#' sysreqs_check_installed()

sysreqs_check_installed

#' @export

sysreqs_fix_installed <- sysreqs_fix_installed

#' List installed system packages
#'
#' @usage
#' sysreqs_list_system_packages()
#'
#' @details
#' This function uses the `sysreqs_platform` configuration option,
#' see \eval{man_config_link("Configuration")}. Set this if
#' `r pak_or_pkgdepends()` does not detect your platform correctly.
#'
#' @return Data frame with columns:
#'   * `status`. two or three characters, the notation of `dpkg` on Debian
#'     based systems. `"ii"` means the package is correctly installed.
#'     On `RPM` based systems it is always `"ii"` currently.
#'   * `package`: name of the system package.
#'   * `version`: installed version of the system package.
#'   * `capabilities`: list column of character vectors, the capabilities
#'     provided by the package.
#'
#' @name sysreqs_list_system_packages
#' @export
#' @family system requirements functions
#' @examplesIf Sys.getenv("IN_PKGDOWN") == "true" && Sys.info()[["sysname"]] == "Linux"
#' sysreqs_list_system_packages()[1:10,]

sysreqs_list_system_packages

# -------------------------------------------------------------------------
# Public API
# -------------------------------------------------------------------------

sysreqs_platforms <- function() {
  as_data_frame(sysreqs2_cmds)
}

sysreqs_is_supported <- function(sysreqs_platform = NULL) {
  sysreqs_platform <- sysreqs_platform %||%
    current_config()$get("sysreqs_platform")
  !is.na(find_sysreqs_platform(sysreqs_platform))
}

sysreqs_db_list <- function(sysreqs_platform = NULL) {
  sysreqs_platform <- sysreqs_platform %||%
    current_config()$get("sysreqs_platform")
  plt <- parse_sysreqs_platform(sysreqs_platform)

  sysreqs_db_update()
  rule_files <- sysreqs2_list_rules()
  rules <- lapply(rule_files, jsonlite::fromJSON, simplifyVector = FALSE)

  matching_deps <- function(rule) {
    for (dep in rule$dependencies) {
      appl <- FALSE
      for (const in dep$constraints) {
        if (
          identical(const$os, plt$os) &&
            identical(const$distribution, plt$distribution) &&
            (is.null(const$versions) || plt$version %in% const$versions)
        ) {
          appl <- TRUE
          break
        }
      }
      if (appl) {
        return(list(
          packages = unname(unlist(dep$packages)),
          pre_install = unname(unlist(dep$pre_install)),
          post_install = unname(unlist(dep$post_install))
        ))
      }
    }
    list(
      packages = NULL,
      pre_install = NULL,
      post_install = NULL
    )
  }

  dependencies <- lapply(rules, matching_deps)
  data_frame(
    name = tools::file_path_sans_ext(basename(rule_files)),
    patterns = lapply(rules, function(r) unlist(r$patterns)),
    packages = lapply(dependencies, "[[", "packages"),
    pre_install = lapply(dependencies, "[[", "pre_install"),
    post_install = lapply(dependencies, "[[", "post_install")
  )
}

sysreqs_db_match <- function(specs, sysreqs_platform = NULL) {
  sysreqs_db_update()
  recs <- sysreqs2_match(specs, sysreqs_platform = sysreqs_platform)
  mapply(
    specs,
    recs,
    SIMPLIFY = FALSE,
    USE.NAMES = FALSE,
    FUN = function(spec, rec) {
      data_frame(
        spec = spec,
        sysreq = vcapply(rec, "[[", "sysreq"),
        packages = lapply(rec, "[[", "packages"),
        pre_install = lapply(rec, "[[", "pre_install"),
        post_install = lapply(rec, "[[", "post_install")
      )
    }
  )
}

sysreqs_db_update <- function() {
  invisible(sysreqs2_update_metadata())
}

#' Create an installation plan for system requirements
#'
#' This function uses [new_pkg_installation_proposal()] and its methods
#' to create an installation plan for one or more packages, and then print
#' their system requirements.
#'
#' @param refs Packages to install.
#' @param upgrade If `TRUE`, pkgdepends will choose the latest available
#'   versions of packages, instead of preferring binary packages over
#'   source packages.
#' @param config Configuration options. See
#'   ['Configuration'][pkgdepends-config]. If it does not include
#'   `library`, then a temporary library is used, which is equivalent to
#'   not assuming any preinstalled packages. Pass `sysreqs_platform` here
#'   if you want a different platform than the one R is running on.
#' @return List with entries:
#'   * `os`: character string. Operating system.
#'   * `distribution`: character string. Linux distribution, `NA` if the
#'     OS is not Linux.
#'   * `version`: character string. Distribution version, `NA` is the OS
#'     is not Linux.
#'   * `pre_install`: character vector. Commands to run before the
#'     installation of system packages.
#'   * `install_scripts`: character vector. Commands to run to install the
#'     system packages.
#'   * `post_install`: character vector. Commands to run after the
#'     installation of system packages.
#'   * `packages`: data frame. Information about the system packages that
#'     are needed. It has columns:
#'     * `sysreq`: string, cross-platform name of the system requirement.
#'     * `packages`: list column of character vectors. The names of the R
#'       packages that have this system requirement.
#'     * `pre_install`: list column of character vectors. Commands run
#'       before the package installation for this system requirement.
#'     * `system_packages`: list column of character vectors. Names of
#'       system packages to install.
#'     * `post_install`: list column of character vectors. Commands run
#'       after the package installation for this system requirement.
#'
#' @export
#' @family system requirements functions
#' @seealso [new_pkg_installation_proposal()] to actually install
#' packages, and potentially system requirements.
#' @examplesIf !pkgdepends:::is_rcmd_check()
#' sysreqs_install_plan(
#'   "tidyverse",
#'   config = list(sysreqs_platform = "ubuntu-22.04")
#' )

sysreqs_install_plan <- function(refs, upgrade = TRUE, config = list()) {
  if (!"library" %in% names(config)) {
    dir.create(lib <- tempfile())
    on.exit(unlink(lib, recursive = TRUE), add = TRUE)
    config$library <- lib
  }
  config$sysreqs <- TRUE
  config$sysreqs_lookup_system <- FALSE

  prop <- new_pkg_installation_proposal(refs, config = config)
  if (upgrade) {
    prop$set_solve_policy("upgrade")
  } else {
    prop$set_solve_policy("lazy")
  }
  prop$solve()
  prop$stop_for_solution_error()
  sol <- prop$get_solution()

  sysreqs_platform <- prop$get_config()$get("sysreqs_platform")
  scripts <- sysreqs2_scripts(sol$data$sysreqs_packages, sysreqs_platform)
  res <- scripts[c(
    "os",
    "distribution",
    "version",
    "pre_install",
    "install_scripts",
    "post_install"
  )]

  res$pre_install <- as.character(res$pre_install)
  res$install_scripts <- as.character(res$install_scripts)
  res$post_install <- as.character(res$post_install)

  sysreqs <- lapply(sol$data$sysreqs_packages, sapply, "[[", "sysreq")
  names <- as.character(sort(unique(unlist(sysreqs))))
  if (length(names)) {
    records <- unlist(sol$data$sysreqs_packages, recursive = FALSE)
    names(records) <- vcapply(records, "[[", "sysreq")
    records <- records[!duplicated(names(records))]
  }

  spkgs <- data_frame(
    sysreq = names,
    packages = lapply(names, function(n) {
      sol$data$package[vlapply(sysreqs, function(s) n %in% s)]
    }),
    pre_install = lapply(
      names,
      function(n) as.character(records[[n]]$pre_install)
    ),
    system_packages = lapply(
      names,
      function(n) as.character(records[[n]]$packages)
    ),
    post_install = lapply(
      names,
      function(n) as.character(records[[n]]$post_install)
    )
  )

  res$packages <- spkgs
  res
}

sysreqs_check_installed <- function(packages = NULL, library = .libPaths()[1]) {
  data <- synchronize(when_all(
    sysreqs2_async_update_metadata(),
    async_system_list_packages(),
    async_parse_installed(library = library, packages = packages)
  ))

  spkgs <- data[[2]]
  spkgs <- spkgs[grepl("^.i$", spkgs$status), ]
  rpkgs <- data[[3]]
  if (is.null(rpkgs$SystemRequirements)) {
    rpkgs$SystemRequirements <- rep(NA, nrow(rpkgs))
  }
  sys <- sysreqs2_match(rpkgs$SystemRequirements)
  sys <- sysreqs_update_state(sys, spkgs)
  rpkgs$sysreqs_packages <- sys
  rpkgs <- rpkgs[!vlapply(rpkgs$sysreqs_packages, is.null), ]
  rpkgs$sys_package_name <- lapply(
    rpkgs$sysreqs_packages,
    function(s) as.character(unlist(lapply(s, "[[", "packages")))
  )
  rpkgs <- rpkgs[lengths(rpkgs$sys_package_name) > 0, ]

  upkgs <- unique(sort(unlist(rpkgs$sys_package_name)))
  provided <- c(spkgs$package, unlist(spkgs$provides))

  pre <- vector("list", length(upkgs))
  post <- vector("list", length(upkgs))
  for (pi in seq_along(upkgs)) {
    p <- upkgs[pi]
    for (s in sys) {
      sp <- unlist(lapply(s, "[[", "packages"))
      if (p %in% sp) {
        newpre <- unlist(lapply(s, "[[", "pre_install"))
        if (length(newpre) > 0) {
          pre[[pi]] <- c(pre[[pi]], newpre)
        }
        newpost <- unlist(lapply(s, "[[", "post_install"))
        if (length(newpost) > 0) post[[pi]] <- c(post[[pi]], newpost)
      }
    }
  }

  res <- data_frame(
    system_package = upkgs,
    installed = upkgs %in% provided,
    packages = lapply(upkgs, function(p) {
      rpkgs$Package[vlapply(rpkgs$sys_package_name, function(s) p %in% s)]
    }),
    pre_install = pre,
    post_install = post
  )

  attr(res, "sysreqs_records") <- sys
  attr(res, "system_packages") <- spkgs

  class(res) <- c("pkg_sysreqs_check_result", class(res))
  res
}

#' @export

format.pkg_sysreqs_check_result <- function(x, ...) {
  ok <- cli::col_green(cli::symbol$tick)
  notok <- cli::col_red(cli::symbol$cross)
  req <- vcapply(x$packages, paste, collapse = ", ")
  paste(
    ansi_align_width(c("system package", "--------------", x$system_package)),
    ansi_align_width(c("installed", "--", ifelse(x$installed, ok, notok))),
    ansi_align_width(c("required by", "-----------", req))
  )
}

#' @export

print.pkg_sysreqs_check_result <- function(x, ...) {
  writeLines(format(x, ...))
}

#' @export

`[.pkg_sysreqs_check_result` <- function(x, i, j, drop = FALSE) {
  class(x) <- setdiff(class(x), "pkg_sysreqs_check_result")
  NextMethod("[")
}

sysreqs_fix_installed <- function(packages = NULL, library = .libPaths()[1]) {
  chk <- sysreqs_check_installed(packages = packages, library = library)
  if (nrow(chk) == 0) {
    cli::cli_alert("No system requirements.")
  } else if (all(chk$installed)) {
    cli::cli_alert_success("All system requirements are installed.")
  } else {
    cli::cli_alert_info(
      "Need to install {sum(!chk$installed)} system package{?s}."
    )
    config <- current_config()
    cmds <- sysreqs2_scripts(
      attr(chk, "sysreqs_records"),
      sysreqs_platform = config$get("sysreqs_platform"),
      missing = TRUE
    )
    sysreqs_install(cmds, config)
  }
  invisible(chk)
}

async_parse_installed <- function(library, packages) {
  pkgs <- pkgcache::parse_installed(library = library, packages = packages)
  miss <- unique(setdiff(packages, pkgs$Package))
  if (length(miss)) {
    warning(cli::format_warning(
      "Ignored {length(miss)} package{?s} that {?is/are} not
       installed: {.pkg {miss}}."
    ))
  }

  async_constant(pkgs)
}

sysreqs_list_system_packages <- function() {
  synchronize(async_system_list_packages()) # nocov
}

# -------------------------------------------------------------------------
# Implementation
# -------------------------------------------------------------------------

parse_sysreqs_platform <- function(x) {
  stopifnot(length(x) == 1)

  # full form or only distro [+ version]
  n_hyphens <- sum(strsplit(x, "")[[1]] == "-")
  if (
    n_hyphens >= 2 &&
      !grepl("^opensuse-leap", x) &&
      !grepl("^opensuse-tumbleweed", x) &&
      !(n_hyphens == 2 && endsWith(x, "libc++"))
  ) {
    osplt <- parse_platform(x)
    if (startsWith(osplt$os, "linux-")) {
      rest <- sub(
        "^linux[-]((dietlibc|gnu|musl|uclibc|unknown)[-])?",
        "",
        osplt$os
      )
      osplt$os <- "linux"
    } else {
      rest <- ""
    }
  } else {
    osplt <- data_frame(
      cpu = NA_character_,
      vendor = NA_character_,
      os = "linux"
    )
    rest <- x
  }
  osplt$distribution <- NA_character_
  osplt$version <- NA_character_

  if (nchar(rest) == 0) {
    return(osplt)
  }

  if (grepl("^opensuse-leap-", rest)) {
    osplt$distribution <- "opensuse-leap"
    osplt$version <- sub("^opensuse-leap-", "", rest)
  } else if (grepl("^opensuse-tumbleweed-", rest)) {
    osplt$distribution <- "opensuse-tumbleweed"
    osplt$version <- sub("^opensuse-tumbleweed-", "", rest)
  } else {
    restpcs <- strsplit(rest, "-", fixed = TRUE)[[1]]
    if (length(restpcs) == 1) {
      osplt$distribution <- restpcs
    } else if (length(restpcs) == 2) {
      osplt$distribution <- restpcs[1]
      osplt$version <- restpcs[2]
    } else {
      osplt$distribution <- restpcs[1]
      osplt$version <- paste0(restpcs[-1], collapse = "-")
    }
  }

  osplt
}

sysreqs_resolve <- function(
  sysreqs,
  sysreqs_platform = NULL,
  config = NULL,
  ...
) {
  if (tolower(Sys.getenv("R_PKG_SYSREQS2")) != "false") {
    synchronize(sysreqs2_async_resolve(sysreqs, sysreqs_platform, config, ...))
  } else {
    synchronise(sysreqs_async_resolve(sysreqs, sysreqs_platform, config, ...))
  }
}

sysreqs_async_resolve <- function(sysreqs, sysreqs_platform, config) {
  sysreqs
  sysreqs_platform
  config
  sysreqs_async_resolve_query(sysreqs, sysreqs_platform, config)$then(function(
    resp
  ) {
    if (resp$status_code < 400) {
      return(resp)
    }
    throw(pkg_error(
      call. = FALSE,
      "Failed to look up system requirements for OS {sysreqs_platform}.",
      i = "HTTP error {resp$status_code} for {.url {resp$url}}.",
      i = "Response: {.val {rawToChar(resp$content)}}."
    ))
  })$then(
    function(resp) sysreqs_resolve_process(sysreqs, sysreqs_platform, resp)
  )$then(function(res) add_class(res, "pkg_sysreqs_result"))
}

sysreqs_async_resolve_query <- function(sysreqs, sysreqs_platform, config) {
  config <- config %||% current_config()
  sysreqs_platform <- sysreqs_platform %||% config$get("sysreqs_platform")
  rspm <- config$get("sysreqs_rspm_url")
  rspm_repo_id <- config$get("sysreqs_rspm_repo_id")
  rspm_repo_url <- sprintf("%s/__api__/repos/%s", rspm, rspm_repo_id)

  plt <- parse_sysreqs_platform(sysreqs_platform)
  req_url <- sprintf(
    "%s/sysreqs?distribution=%s&release=%s",
    rspm_repo_url,
    plt$distribution,
    plt$version
  )

  headers <- c("Content-Type" = "text/plain")

  data <- sysreqs_resolve_make_data(sysreqs)

  http_post(req_url, data = data, headers = headers)
}

sysreqs_resolve_process <- function(sysreqs, sysreqs_platform, resp) {
  hdr <- curl::parse_headers_list(resp$headers)
  cnt <- rawToChar(resp$content)
  Encoding(cnt) <- "UTF-8"

  data <- jsonlite::fromJSON(cnt, simplifyVector = FALSE)

  pre_install <- unique(as.character(unlist(c(
    data[["pre_install"]],
    lapply(data[["dependencies"]], `[[`, "pre_install")
  ))))
  install_scripts <- unique(as.character(unlist(c(
    data[["install_scripts"]],
    lapply(data[["dependencies"]], `[[`, "install_scripts")
  ))))
  post_install <- unique(as.character(unlist(c(
    data[["post_install"]],
    lapply(data[["dependencies"]], `[[`, "post_install")
  ))))

  plt <- parse_sysreqs_platform(sysreqs_platform)
  list(
    os = plt$os,
    distribution = plt$distribution,
    version = plt$version,
    url = resp$url,
    total = resp$times["total"],
    pre_install = pre_install,
    install_scripts = install_scripts,
    post_install = post_install
  )
}

sysreqs_canonise_query <- function(sysreqs) {
  sysreqs <- str_trim(sysreqs)
  sysreqs <- sort(unique(sysreqs[!is.na(sysreqs) & sysreqs != ""]))
  sysreqs <- gsub("\n", "\n ", sysreqs)
  sysreqs
}

sysreqs_resolve_make_data <- function(sysreqs) {
  sysreqs <- sysreqs_canonise_query(sysreqs)
  paste(
    collapse = "\n",
    c(
      "Package: pkgdependssysreqs",
      "Version: 1.0.0",
      "SystemRequirements: ",
      paste0("    ", sysreqs),
      "Note: and thank you!",
      ""
    )
  )
}

sysreqs_install <- function(sysreqs_cmds, config = NULL) {
  config <- config %||% current_config()
  sudo <- config$get("sysreqs_sudo")
  verbose <- config$get("sysreqs_verbose")
  dry_run <- config$get("sysreqs_dry_run")

  cmds <- unlist(sysreqs_cmds[c(
    "pre_install",
    "install_scripts",
    "post_install"
  )])
  if (length(cmds) == 0) {
    return()
  }

  cli::cli_alert_info("Installing system requirements")

  # TODO: fix 'R' commands (e.g. `R CMD javareconf`) to call the current
  # version of R and not the one on the PATH

  cmds <- compact_cmds(cmds)

  if (dry_run) {
    cmds <- paste("echo", cmds)
  }

  if (verbose) {
    callback <- function(x, ...) {
      x <- str_trim(x)
      if (nchar(x)) cli::cli_verbatim(x)
    }
  } else {
    callback <- function(x, ...) invisible()
  }

  output <- lapply(cmds, function(cmd) {
    if (sudo) {
      sh <- "sudo" # nocov
      cmdline <- c("sh", "-c", cmd) # nocov
    } else {
      sh <- "sh"
      cmdline <- c("-c", cmd)
    }
    fullcmd <- paste(c(sh, cmdline), collapse = " ")
    cli::cli_alert_info("Executing {.code {fullcmd}}")
    processx::run(
      sh,
      cmdline,
      stdout_callback = callback,
      stderr_to_stdout = TRUE
    )
  })

  invisible(output)
}

compact_cmds <- function(x) {
  rx <- "^apt-get install -y ([a-z0-9-]+)$"
  if (length(x) == 0 || !all(grepl(rx, x))) {
    return(x)
  }

  paste0(
    "apt-get install -y ",
    paste(gsub(rx, "\\1", x), collapse = " ")
  )
}

is_root <- function() {
  if (os_type() != "unix") {
    return(FALSE)
  }
  ps::ps_uids()[["effective"]] == 0
}

can_sudo_without_pw <- function() {
  if (os_type() != "unix") {
    return(FALSE)
  }
  tryCatch(
    {
      processx::run("sudo", c("-s", "id"))
      TRUE
    },
    error = function(err) FALSE
  )
}

test_that("sysreqs_platforms", {
  withr::local_options(width = 200)
  expect_snapshot(print(sysreqs_platforms(), n = Inf))
})

test_that("sysreqs_db_list", {
  skip_on_cran()
  withr::local_options(pkg.sysreqs_db_update = FALSE)

  lst <- sysreqs_db_list("ubuntu-22.04")
  lst <- lst[lst$name %in% c("cairo", "chrome", "libcurl"), ]
  expect_snapshot({
    lst$packages
    lst$pre_install
    lst$post_install
  })
})

test_that("sysreqs_db_match", {
  skip_on_cran()
  withr::local_options(pkg.sysreqs_db_update = FALSE)

  res <- sysreqs_db_match(
    c("java and libcurl", "openssl would be good"),
    sysreqs_platform = "ubuntu-22.04"
  )
  expect_snapshot({
    res
    lapply(res, "[[", "packages")
    lapply(res, "[[", "post_install")
  })
})

test_that("sysreqs_install_plan", {
  skip_on_cran()
  setup_fake_apps()
  setup_fake_gh_app()
  withr::local_options(pkg.sysreqs_db_update = FALSE)

  res <- suppressMessages(sysreqs_install_plan(
    "curl",
    config = list(sysreqs_platform = "ubuntu-22.04")
  ))
  expect_snapshot({
    res
    res$packages$packages
    res$packages$system_packages
  })

  res2 <- suppressMessages(sysreqs_install_plan(
    "cran/rJava",
    config = list(sysreqs_platform = "centos-7")
  ))
  expect_snapshot({
    res2
    res2$packages$packages
    res2$packages$system_packages
  })
})

test_that("sysreqs_check_installed", {
  skip_on_cran()
  withr::local_options(pkg.sysreqs_db_update = FALSE)

  fakesys <- data_frame(
    status = "ii",
    package = c("gsfonts", "libcurl4-openssl-dev"),
    provides = list(
      character(),
      c(
        "libcurl-dev",
        "libcurl-ssl-dev",
        "libcurl3-dev",
        "libcurl3-openssl-dev",
        "libcurl4-dev"
      )
    )
  )

  faker <- data_frame(
    Package = c("magick", "curl"),
    SystemRequirements = c(
      "ImageMagick++: ImageMagick-c++-devel (rpm) or\nlibmagick++-dev (deb)",
      "libcurl: libcurl-devel (rpm) or\nlibcurl4-openssl-dev (deb)."
    )
  )

  mockery::stub(
    sysreqs_check_installed,
    "async_system_list_packages",
    function(...) async_constant(fakesys)
  )
  mockery::stub(
    sysreqs_check_installed,
    "async_parse_installed",
    function(...) async_constant(faker)
  )

  withr::local_options(pkg.sysreqs_platform = "ubuntu-22.04")
  res <- sysreqs_check_installed()
  expect_snapshot(res)

  # can drop class
  expect_snapshot(res[])

  # edge case of no sysreqs at all
  faker <- data_frame(
    Package = c("filelock")
  )
  expect_snapshot(sysreqs_check_installed())
})

test_that("async_parse_installed", {
  mockery::stub(
    async_parse_installed,
    "pkgcache::parse_installed",
    data_frame(Package = c("foo", "bar"))
  )

  expect_snapshot(synchronize(async_parse_installed(
    library = .libPaths()[1],
    packages = c("foo", "bar", "baz")
  )))
})

test_that("parse_sysreqs_platform", {
  expect_snapshot({
    parse_sysreqs_platform("x86_64-pc-linux-gnu-ubuntu-22.04")
    parse_sysreqs_platform("x86_64-pc-linux-musl-alpine-3.14.1")
    parse_sysreqs_platform("x86_64-pc-linux-ubuntu-22.04")
    parse_sysreqs_platform("ubuntu-22.04")
    parse_sysreqs_platform("aarch64-apple-darwin20")
    parse_sysreqs_platform("i386+x86_64-w64-mingw32")
    parse_sysreqs_platform("x86_64-w64-mingw32")
    parse_sysreqs_platform("ubuntu")
    parse_sysreqs_platform("x86_64-px-linux-gnu-ubuntu-bar-baz")
    parse_sysreqs_platform("ubuntu-22.04-libc++")
  })
})

test_that("is_root", {
  mockery::stub(is_root, "os_type", "windows")
  expect_false(is_root())

  mockery::stub(is_root, "os_type", "unix")
  mockery::stub(is_root, "ps::ps_uids", c(effective = 0L))
  expect_true(is_root())

  mockery::stub(is_root, "ps::ps_uids", c(effective = 500L))
  expect_false(is_root())
})

test_that("can_sudo_without_pw", {
  mockery::stub(can_sudo_without_pw, "os_type", "windows")
  expect_false(can_sudo_without_pw())

  mockery::stub(can_sudo_without_pw, "os_type", "unix")
  mockery::stub(can_sudo_without_pw, "processx::run", list(status = 0))
  expect_true(can_sudo_without_pw())

  mockery::stub(can_sudo_without_pw, "processx::run", function(...) stop("no"))
  expect_false(can_sudo_without_pw())
})

# -------------------------------------------------------------------------
# Tests for the old implementation

test_that("query, post_install", {
  setup_fake_sysreqs_app()
  withr::local_envvar(R_PKG_SYSREQS2 = "false")
  expect_snapshot(
    {
      srq <- sysreqs_resolve("java", "ubuntu-22.04")
      srq$total <- 1 / 3
      srq
    },
    transform = transform_sysreqs_server
  )
})

test_that("pre_install", {
  setup_fake_sysreqs_app()
  withr::local_envvar(R_PKG_SYSREQS2 = "false")
  expect_snapshot(
    {
      srq <- sysreqs_resolve("this needs geos please", "ubuntu-16.04")
      srq$total <- 1 / 3
      srq
    },
    transform = transform_sysreqs_server
  )
})

test_that("multiple sysreqs", {
  setup_fake_sysreqs_app()
  withr::local_envvar(R_PKG_SYSREQS2 = "false")
  expect_snapshot(
    {
      srq <- sysreqs_resolve("java and also libcurl", "ubuntu-22.04")
      srq$total <- 1 / 3
      srq
    },
    transform = transform_sysreqs_server
  )
})

test_that("error, unknown os", {
  setup_fake_sysreqs_app()
  withr::local_envvar(R_PKG_SYSREQS2 = "false")
  expect_snapshot(
    error = TRUE,
    transform = transform_sysreqs_server,
    sysreqs_resolve("java", "foobar-11")
  )
})

test_that("sysreqs_install", {
  skip_on_os("windows")
  setup_fake_sysreqs_app()
  withr::local_envvar(
    PKG_SYSREQS_DRY_RUN = "true",
    PKG_SYSREQS_SUDO = "false",
    R_PKG_SYSREQS2 = "false"
  )

  # not verbose
  withr::local_envvar(PKG_SYSREQS_VERBOSE = "false")
  srq <- sysreqs_resolve("libcurl and openssl", "ubuntu-22.04")
  expect_snapshot(sysreqs_install(srq))

  srq <- sysreqs_resolve("java and also libcurl", "ubuntu-22.04")
  expect_snapshot(sysreqs_install(srq))

  # verbose
  withr::local_envvar(PKG_SYSREQS_VERBOSE = "true")
  expect_snapshot(sysreqs_install(srq))

  # nothing to do
  expect_silent(expect_null(sysreqs_install(list())))
})

test_that("compact_cmds", {
  expect_snapshot({
    compact_cmds(character())
    compact_cmds(c(
      "apt-get install -y libssl-dev"
    ))
    compact_cmds(c(
      "apt-get install -y libssl-dev",
      "apt-get install -y libcurl4-openssl-dev"
    ))
  })
})

test_that("highlight_sysreqs", {
  # edge case
  expect_equal(highlight_sysreqs(NULL), "")

  sq <- list(
    NULL,
    list(
      list(
        sysreq = "fontconfig",
        packages = "libfontconfig1-dev",
        pre_install = NULL,
        post_install = NULL,
        packages_missing = character(0)
      ),
      list(
        sysreq = "freetype",
        packages = "libfreetype6-dev",
        pre_install = NULL,
        post_install = NULL,
        packages_missing = character(0)
      )
    ),
    list(
      list(
        sysreq = "freetype",
        packages = "libfreetype6-dev",
        pre_install = NULL,
        post_install = NULL,
        packages_missing = character(0)
      ),
      list(
        sysreq = "fribidi",
        packages = "libfribidi-dev",
        pre_install = NULL,
        post_install = NULL,
        packages_missing = "libfribidi-dev"
      ),
      list(
        sysreq = "harfbuzz",
        packages = "libharfbuzz-dev",
        pre_install = NULL,
        post_install = NULL,
        packages_missing = "libharfbuzz-dev"
      )
    ),
    NULL
  )
  sq2 <- lapply(sq, function(x) {
    lapply(x, function(xx) xx[setdiff(names(xx), "packages_missing")])
  })

  expect_snapshot({
    highlight_sysreqs(sq)
    highlight_sysreqs(sq2)
  })

  if (!l10n_info()[["UTF-8"]]) {
    skip("No UTF-8 support")
  }
  withr::local_options(cli.unicode = TRUE, cli.num_colors = 256)
  expect_snapshot({
    highlight_sysreqs(sq)
    highlight_sysreqs(sq2)
  })

  # installer
  sq[[2]] <- list(list(
    sysreq = "chrome",
    packages = NULL,
    pre_install = "download and install chrome",
    post_install = NULL,
    packages_missing = character()
  ))
  expect_snapshot({
    highlight_sysreqs(sq)
  })
})

test_that("default_sysreqs", {
  config <- current_config()
  config$set("sysreqs_platform", "aarch64-apple-darwin20")
  expect_false(default_sysreqs(config))

  config$set("sysreqs_platform", "x86_64-pc-linux-gnu-ubuntu-22.04")
  mockery::stub(default_sysreqs, "is_root", FALSE)
  mockery::stub(default_sysreqs, "can_sudo_without_pw", FALSE)
  expect_false(default_sysreqs(config))

  mockery::stub(default_sysreqs, "is_root", TRUE)
  mockery::stub(default_sysreqs, "can_sudo_without_pw", FALSE)
  expect_true(default_sysreqs(config))

  mockery::stub(default_sysreqs, "is_root", FALSE)
  mockery::stub(default_sysreqs, "can_sudo_without_pw", TRUE)
  expect_true(default_sysreqs(config))
})

pkg_error <- function(
  ...,
  .data = NULL,
  .class = NULL,
  .envir = parent.frame(),
  call. = TRUE
) {
  .hide_from_trace <- TRUE
  cnd <- new_error(
    call. = call.,
    cli::format_error(
      .envir = .envir,
      c(
        ...
      )
    )
  )

  if (length(.data)) {
    cnd[names(.data)] <- .data
  }
  if (length(.class)) {
    class(cnd) <- c(.class, class(cnd))
  }

  cnd
}

pkg_warning <- function(
  ...,
  .data = NULL,
  .class = NULL,
  .envir = parent.frame(),
  call. = TRUE
) {
  .hide_from_trace <- TRUE
  cnd <- new_cond(
    call. = call.,
    cli::format_error(
      .envir = .envir,
      c(
        ...
      )
    )
  )
  class(cnd) <- c(.class, "warning", "condition")

  if (length(.data)) {
    cnd[names(.data)] <- .data
  }
  if (length(.class)) {
    class(cnd) <- c(.class, class(cnd))
  }

  cnd
}

stop <- function(..., call. = TRUE, domain = NA) {
  .hide_from_trace <- TRUE
  args <- list(...)
  if (length(args) == 1L && inherits(args[[1L]], "condition")) {
    throw(
      add_class(args[[1]], c("rlib_error_3_0", "rlib_error"), "end"),
      frame = parent.frame()
    )
  } else {
    throw(new_error(..., call. = call., domain = domain))
  }
}

stopifnot <- function(...) {
  assert_that(..., env = parent.frame())
}

msg_internal_error <- function() {
  "This is an internal error in {pak_or_pkgdepends()}, please report
   an issue at {.url https://github.com/r-lib/{pak_or_pkgdepends()}/issues}."
}

msg_package_sources <- function() {
  if (is_pak()) {
    "See {.help pak::pak_package_sources} for the package sources
     pak supports."
  } else {
    "See {.help pkgdepends::pkg_refs} for supported package sources."
  }
}

test_that("spell check", {
  skip_on_cran()
  skip_in_covr()
  if (utils::packageVersion("spelling") <= "2.1") {
    skip("Needs newer spelling package")
  }
  pkg_dir <- test_package_root()
  suppressMessages(results <- spelling::spell_check_package(pkg_dir))

  if (nrow(results)) {
    output <- sprintf(
      "Potential spelling errors: %s\n",
      paste(results$word, collapse = ", ")
    )
    stop(output, call. = FALSE)
  } else {
    expect_true(TRUE)
  }
})

async_system_list_packages <- function(config = NULL) {
  config <- config %||% current_config()
  sysreqs_platform <- config$get("sysreqs_platform")
  pkgtool <- sysreqs2_command(sysreqs_platform, "query")
  if (pkgtool == "dpkg-query") {
    async_system_list_packages_dpkg_query(config)
  } else if (pkgtool == "apk") {
    async_system_list_packages_apk(config)
  } else {
    async_system_list_packages_rpm(config)
  }
}

# For DEB systems we need
# dpkg-query -W -f '${db:Status-Abbrev} ${Package} ${Version} ${Provides}\n' '*'
# This also lists the packages that are not installed currently.
# First field is the status code, second is package name, then version
# number (if installed), then capabilities in a comma separated list.
# Capabilities may include version requirements.

async_system_list_packages_dpkg_query <- function(config) {
  args <- c(
    "-W",
    "-f",
    "${db:Status-Abbrev} ${Package} ${Version} ${Provides}\\n",
    "*"
  )
  stdout <- tempfile()
  external_process(function(...) {
    processx::process$new(
      "dpkg-query",
      stdout = stdout,
      stderr = stdout,
      args = args,
      ...
    )
  })$then(function(ret) {
    parse_dpkg_query_output(strsplit(ret$stdout, "\n")[[1]])
  })$finally(function() unlink(stdout))
}

parse_dpkg_query_output <- function(lines) {
  # If not installed, then not interesting
  lines <- lines[grepl("^.[^n]", lines)]
  status <- trimws(sub(" .*$", "", lines))
  rest <- sub("^[^ ]+[ ]+", "", lines)
  package <- trimws(sub(" .*$", "", rest))
  rest <- sub("^[^ ]+[ ]+", "", rest)
  version <- trimws(sub(" .*$", "", rest))
  rest <- sub("^[^ ]+[ ]+", "", rest)
  provides <- lapply(strsplit(rest, ",[ ]?"), trimws)
  # just drop version requirements, we probably don't need them
  provides <- lapply(provides, sub, pattern = "[ ].*$", replacement = "")
  provides <- lapply(provides, function(x) x[x != ""])
  # sorted by default
  data_frame(
    status = status,
    package = package,
    version = version,
    provides = provides
  )
}

# For RPM, we need this query:
# rpm -qa --provides --qf '---%{NAME}\n'
# The output is a bit weird, for each package first the capabilities are
# listed, then the package name, after `---`.
# We can use %{NAME} %{VERSION} if we want the version numbers as well.

async_system_list_packages_rpm <- function(config) {
  args <- c(
    "-qa",
    "--provides",
    "--qf",
    "---%{NAME} %{VERSION}\\n"
  )
  stdout <- tempfile()
  external_process(function(...) {
    processx::process$new(
      "rpm",
      stdout = stdout,
      stderr = stdout,
      args = args,
      ...
    )
  })$then(function(ret) {
    parse_rpm_output(strsplit(ret$stdout, "\n")[[1]])
  })$finally(function() unlink(stdout))
}

parse_rpm_output <- function(lines) {
  last <- which(grepl("^---", lines))
  from <- c(1L, last[-length(last)] + 1L)
  to <- c(last)
  blocks <- mapply(from, to, FUN = function(f, t) lines[f:t])
  pkglines <- sub("^---", "", vcapply(blocks, utils::tail, 1))
  package <- trimws(sub(" .*$", "", pkglines))
  version <- trimws(sub("^[^ ]+[ ]+", "", pkglines))
  provides <- mapply(
    package,
    blocks,
    USE.NAMES = FALSE,
    FUN = function(pkg, blk) {
      blk <- utils::head(blk, -1)
      # the package itself is a capability, drop that
      # TODO: drop the pkg(arch) = version capabilties as well?
      blk <- blk[!startsWith(blk, paste0(pkg, " = "))]
      # drop version numbers
      blk <- sub(" .*$", "", blk)
      trimws(blk)
    }
  )

  pkgs <- data_frame(
    status = "ii",
    package = package,
    version = version,
    provides = provides
  )

  pkgs <- pkgs[order(tolower(pkgs$package)), ]
  pkgs
}

# For APK, we need this query:

async_system_list_packages_apk <- function(config) {
  args <- c(
    "list",
    "-I"
  )
  stdout <- tempfile()
  external_process(function(...) {
    processx::process$new(
      "apk",
      stdout = stdout,
      stderr = stdout,
      args = args,
      ...
    )
  })$then(function(ret) {
    parse_apk_output(strsplit(ret$stdout, "\n")[[1]])
  })$finally(function() unlink(stdout))
}

parse_apk_output <- function(lines) {
  package = sub("-[0-9].*", "", lines)
  version = sub(".*?-([0-9][^ ]*).*", "\\1", lines)
  provides = ""

  pkgs <- data_frame(
    status = "ii",
    package = package,
    version = version,
    provides = provides
  )

  pkgs <- pkgs[order(tolower(pkgs$package)), ]
  pkgs
}

test_that("async_system_list_packages, deb", {
  config <- current_config()
  config$set("sysreqs_platform", "x86_64-pc-linux-gnu-ubuntu-22.04")
  mockery::stub(
    async_system_list_packages,
    "async_system_list_packages_dpkg_query",
    function(...) async_constant("OK")
  )
  expect_equal(
    synchronise(async_system_list_packages(config)),
    "OK"
  )
})

test_that("async_system_list_packages, rpm", {
  config <- current_config()
  config$set("sysreqs_platform", "x86_64-pc-linux-gnu-redhat-9")
  mockery::stub(
    async_system_list_packages,
    "async_system_list_packages_rpm",
    function(...) async_constant("OK")
  )
  expect_equal(
    synchronise(async_system_list_packages(config)),
    "OK"
  )
})

test_that("async_system_list_packages_dpkg_query", {
  fake <- read_all(test_path("fixtures", "dpkg-query.txt"))
  mockery::stub(
    async_system_list_packages_dpkg_query,
    "external_process",
    function(...) async_constant(list(stdout = fake))
  )
  ans <- synchronise(async_system_list_packages_dpkg_query())
  expect_snapshot(ans)
})

test_that("async_system_list_packages_rpm", {
  fake <- read_all(test_path("fixtures", "rpm.txt"))
  mockery::stub(
    async_system_list_packages_rpm,
    "external_process",
    function(...) async_constant(list(stdout = fake))
  )
  ans <- synchronise(async_system_list_packages_rpm())
  expect_snapshot(ans)
})

ghmirror <- local({
  ghmirror_update <- function(pkg, source_repo = "https://cran.r-project.org") {
    cran_versions <- get_package_versions(pkg, source_repo)
    atgh_versions <- get_github_versions(pkg)

    missing <- setdiff(cran_versions$version, atgh_versions$version)

    add_missing_versions(
      pkg,
      missing,
      new_pkg = length(atgh_versions) == 0,
      repo = source_repo
    )

    invisible(missing)
  }

  # -----------------------------------------------------------------------
  # Internal functions

  cran_versions <- NULL

  get_all_cran_versions <- function(
    source_repo = "https://cran.r-project.org",
    forget = FALSE
  ) {
    if (!is.null(cran_versions) && !forget) {
      return(cran_versions)
    }

    r_archive <- pkgcache::cran_archive_cache$new(
      cran_mirror = source_repo,
      update_after = as.difftime(30, units = "mins")
    )
    r_archive$check_update()
    r_archive_pkgs <- r_archive$list()

    r_source <- pkgcache::cranlike_metadata_cache$new(
      platforms = "source",
      bioc = FALSE,
      cran_mirror = source_repo,
      repos = NULL,
      update_after = as.difftime(30, units = "mins")
    )
    r_source$check_update()
    r_source_pkgs <- r_source$list()

    # we'll just drop versions with invalid version numbers, these are
    # old packages only
    badver <- is.na(package_version(r_archive_pkgs$version, strict = FALSE))
    r_archive_pkgs <- r_archive_pkgs[!badver, , drop = FALSE]

    res <- data_frame(
      package = c(r_archive_pkgs$package, r_source_pkgs$package),
      version = c(r_archive_pkgs$version, r_source_pkgs$version)
    )

    res <- res[order(tolower(res$package), package_version(res$version)), ]
    cran_versions <<- res

    res
  }

  get_package_versions <- function(
    pkg,
    source_repo = "https://cran.r-project.org",
    forget = FALSE
  ) {
    all <- get_all_cran_versions(source_repo, forget = forget)
    res <- all[all$package == pkg, , drop = FALSE]
    res
  }

  get_github_versions <- function(pkg) {
    github_versions <- tryCatch(
      gh::gh(
        "/repos/:owner/:repo/tags",
        owner = "cran",
        repo = pkg,
        .limit = Inf,
        .token = get_gh_token()
      ),
      error = function(e) list(),
      warning = function(e) list()
    )
    github_versions <- vapply(github_versions, "[[", FUN.VALUE = "", "name")
    github_versions <- grep("R-", github_versions, value = TRUE, invert = TRUE)

    data_frame(
      version = rev(github_versions)
    )
  }

  get_gh_token <- function() {
    token <- Sys.getenv("CRANATGH_GITHUB_TOKEN", NA_character_)
    if (is.na(token)) {
      token <- Sys.getenv("GITHUB_PAT", NA_character_)
    }
    if (is.na(token)) {
      token <- Sys.getenv("GITHUB_TOKEN", NA_character_)
    }
    token
  }

  add_missing_versions <- function(pkg, versions, new_pkg, repo) {
    if (length(versions) == 0) {
      return()
    }

    oldwd <- getwd()
    on.exit(setwd(oldwd), add = TRUE)
    change_to_cranatgh_home()

    if (new_pkg) {
      create_git_repo(pkg)
    }

    if (!file.exists(pkg)) {
      clone_git_repo(pkg)
    }

    set_git_user(pkg)

    for (ver in versions) {
      metadata <- add_missing_version(pkg, ver, repo)
    }

    desc <- make_description(metadata)

    if (new_pkg) {
      create_gh_repo(pkg, desc)
    }

    push_to_github(pkg)

    if (!new_pkg) {
      update_description(pkg, desc)
    }

    invisible()
  }

  make_description <- function(pkg) {
    if (is.character(pkg)) {
      pkg <- tryCatch(
        pkgsearch::cran_package(pkg),
        error = function(e) NULL
      )
    } else if (inherits(pkg, "description")) {
      pkg <- list(
        Package = pkg$get("Package"),
        Title = pkg$get("Title"),
        URL = pkg$get("URL"),
        BugReports = pkg$get("BugReports")
      )
    } else {
      stop("'pkg' must be a character scalar or a 'description' object")
    }

    dsc <- paste(
      sep = "  ",
      dont_break(
        ":exclamation: This is a read-only mirror of the CRAN R package repository."
      ),
      dont_break(
        pkg$Package,
        " \u2014 ",
        pkg$Title,
        nullna_or(pkg$URL, paste0(". Homepage: ", pkg$URL))
      ),
      nullna_or(
        pkg$BugReports,
        dont_break("Report bugs for this package: ", pkg$BugReports)
      )
    )

    # Limit is 350 characters, but be conservative
    if (nchar(dsc) > 320) {
      dsc <- paste0(substr(dsc, 1, 320), " ...")
    }

    dsc
  }

  add_missing_version <- function(package, version, repo) {
    proc <- cli::cli_process_start("Adding {.pkg {package}} {version}")

    ## Rename the .git directory. We'll need it later
    file.rename(file.path(package, ".git"), "dot-git")

    ## Remove everything from the old version
    unlink(package, recursive = TRUE, force = TRUE)
    dir.create(package)

    ## Put the new version in place
    tar_file <- get_package_tarball(package, version, repo)
    untar(tar_file)
    unlink(tar_file)

    ## Put back the .git directory
    ## The unlink is for the occasional case when there is already
    ## a .git directory in the package. This is junk anyway, and it
    ## should not be there
    unlink(file.path(package, ".git"))
    file.rename("dot-git", file.path(package, ".git"))

    setwd(package)
    on.exit(setwd(".."), add = TRUE)

    ## To prevent an error like "detected dubious ownership in repository"
    if (.Platform$OS.type == "unix") {
      system("chown -R `id -un`:`id -gn` .")
    }

    ## Add all the new files
    git("status")
    git("add", "-A", ".")
    git("status")

    ## Package information from DESCRIPTION
    metadata <- desc::description$new()
    maint <- metadata$get_maintainer()
    auth <- metadata$get("Author")

    ## Commit the new version
    date <- format_iso_8601(metadata$get("Date/Publication"))
    git(
      env = c("GIT_COMMITTER_DATE" = date),
      "commit",
      "--allow-empty",
      "-m",
      paste0("version ", version),
      "--date",
      date,
      "--author",
      fix_maintainer(maint, auth)
    )

    git("tag", version)

    cli::cli_process_done(proc)

    metadata
  }

  change_to_cranatgh_home <- function() {
    home <- default_tree_location()
    if (is.na(home)) {
      dir.create(home <- tempfile())
    }
    setwd(home)
  }

  default_tree_location <- function() {
    Sys.getenv("CRANATGH_TREES", NA_character_)
  }

  create_git_repo <- function(path) {
    dir.create(path)
    wd <- getwd()
    on.exit(setwd(wd), add = TRUE)
    setwd(path)
    cli::cli_alert_info("Creating git repo in {.path {path}}")
    git("init", ".")
  }

  set_git_user <- function(path, user = NULL, email = NULL) {
    user <- user %||% default_cranatgh_user()
    email <- email %||% default_cranatgh_email()

    withr::with_dir(path, {
      git("config", "--local", "user.name", user)
      git("config", "--local", "user.email", email)
    })
  }

  default_cranatgh_user <- function() {
    Sys.getenv("CRANATGH_USER", "cran-robot")
  }

  default_cranatgh_email <- function() {
    Sys.getenv("CRANATGH_EMAIL", "csardi.gabor+cran@gmail.com")
  }

  default_cranatgh_org <- function() {
    Sys.getenv("CRANATGH_ORG", "cran")
  }

  clone_git_repo <- function(pkg) {
    url <- get_clone_url(pkg)
    proc <- cli::cli_process_start(
      "Cloning GitHub repo from {.url {safe_url(url)}}"
    )
    git("clone", url)
    cli::cli_process_done(proc)
  }

  safe_url <- function(url) {
    sub("://[-:a-z0-9]+@", "://<token>@", url)
  }

  get_clone_url <- function(package) {
    owner <- default_cranatgh_org()
    token <- get_gh_token()
    token <- if (is.na(token)) "" else paste0(token, "@")

    sprintf("https://%sgithub.com/%s/%s.git", token, owner, package)
  }

  git <- function(..., env = character(), timeout = 60 * 60) {
    processx::run(
      "git",
      args = unlist(list(...)),
      env = c(Sys.getenv(), env)
    )
  }

  get_package_tarball <- function(package, version, repo) {
    urls <- package_urls(package, version, repo)
    ok <- FALSE
    for (url in urls) {
      dest_file <- basename(url)
      tryCatch(
        {
          curl::curl_download(url, dest_file)
          ok <- TRUE
          break
        },
        error = function(err) err
      )
    }

    if (!ok) {
      stop("Cannot download package ", package)
    }
    dest_file
  }

  package_urls <- function(package, version, repo) {
    # work around some mistakes, file names do not match the version number
    if (package == "HTML" && version == "0.4") {
      "https://cran.rstudio.com/src/contrib/Archive/HTML/HTML_0.4-1.tar.gz"
    } else if (package == "timeslab" && version == "1.0") {
      "https://cran.r-project.org/src/contrib/Archive/timeslab/timeslab_1.0-1.tar.gz"
    } else {
      c(
        sprintf("%s/src/contrib/%s_%s.tar.gz", repo, package, version),
        sprintf(
          "%s/src/contrib/Archive/%s/%s_%s.tar.gz",
          repo,
          package,
          package,
          version
        ),
        sprintf(
          "%s/src/contrib/%s/Other/%s_%s.tar.gz",
          repo,
          as.character(getRversion()),
          package,
          version
        )
      )
    }
  }

  fix_maintainer <- function(maint, auth) {
    if (is.na(maint)) {
      maint <- auth
    }
    maint <- sub("\\s*<", " <", maint)

    ## ': end of single quote
    ## ": start of double quote
    ## ': single quote (within double quotes)
    ## ": end of double quote
    ## ': start of single quote for the rest of the string
    maint <- gsub("'", paste0("'", '"', "'", '"', "'"), maint)

    if (is.na(maint)) {
      maint <- "??? <???@???>"
    }
    if (!grepl("<.*>", maint)) {
      maint <- paste0(maint, " <???@???>")
    }
    if (toupper(maint) == "ORPHANED") {
      maint <- "ORPHANED <cran@R-project.org>"
    }
    maint
  }

  dont_break <- function(...) {
    gsub("\\s+", "\u00a0", paste0(...))
  }

  nullna_or <- function(x, expr) {
    if (is.null(x) || (length(x) == 1 && is.na(x))) "" else expr
  }

  push_to_github <- function(package, forced_push = FALSE) {
    wd <- getwd()
    on.exit(setwd(wd), add = TRUE)
    setwd(package)

    add_gh_remote(package)

    proc <- cli::cli_process_start("Pushing {.pkg {package}} to GitHub")
    current <- sub("* ", "", fixed = TRUE, system("git branch", intern = TRUE))
    git("push", "--tags", if (forced_push) "-f", "-u", "origin", current)
    cli::cli_process_done(proc)
  }

  add_gh_remote <- function(package) {
    current <- git("remote")$stdout
    if (!grepl("\\borigin\\b", current)) {
      git(
        "remote",
        "add",
        "origin",
        get_clone_url(package)
      )
    }
  }

  update_description <- function(
    package,
    description = make_description(package)
  ) {
    description <- clean_description(description)

    proc <- cli::cli_process_start(
      "Updating repo description for {.pkg {package}}"
    )
    gh::gh(
      "PATCH /repos/:owner/:repo",
      owner = default_cranatgh_org(),
      repo = package,
      name = package,
      description = description,
      homepage = "",
      .token = get_gh_token()
    )
    cli::cli_process_done(proc)
  }

  clean_description <- function(description) {
    description <- unname(description)
    description <- gsub("\n", " ", description)
    description
  }

  structure(
    list(
      .internal = environment(),

      update = ghmirror_update
    )
  )
})

## ------------------------------------------------------------------------
## API

#' @importFrom stats na.omit

parse_remote_cran <- function(specs, ...) {
  parsed_specs <- re_match(specs, standard_rx("cran"))

  parsed_specs$ref <- parsed_specs$.text
  cn <- setdiff(colnames(parsed_specs), c(".match", ".text"))
  parsed_specs <- parsed_specs[, cn]
  eq <- parsed_specs$atleast == "" & parsed_specs$version != ""
  parsed_specs$atleast[eq] <- "=="
  parsed_specs$type <- "cran"
  lapply(
    seq_len(nrow(parsed_specs)),
    function(i) as.list(parsed_specs[i, ])
  )
}

resolve_remote_cran <- function(
  remote,
  direct,
  config,
  cache,
  dependencies,
  ...
) {
  force(remote)
  force(direct)
  force(dependencies)
  versions <- if ("type" %in% names(remote)) {
    remote$version
  } else {
    vcapply(remote, "[[", "version") # nocov
  }

  if (all(versions %in% c("", "current"))) {
    type_cran_resolve_current(remote, direct, config, cache, dependencies)
  } else {
    type_cran_resolve_version(remote, direct, config, cache, dependencies) # nocov
  }
}

download_remote_cran <- function(
  resolution,
  target,
  target_tree,
  config,
  cache,
  which,
  on_progress
) {
  # if we don't have a shain the metadata, then we need to ping.
  # otherwise we can use the cache.
  download_ping_if_no_sha(resolution, target, config, cache, on_progress)
}

satisfy_remote_cran <- function(resolution, candidate, config, ...) {
  ## 1. candidate must be a cran, standard or installed ref
  if (!candidate$type %in% c("cran", "standard", "installed")) {
    return(structure(
      FALSE,
      reason = "Type must be 'cran', 'standard' or 'installed'"
    ))
  }

  ## 2. installed refs must be from CRAN
  if (candidate$type == "installed") {
    repotype <- candidate$extra[[1]]$repotype
    if (!identical(repotype, "cran")) {
      return(structure(FALSE, reason = "Installed package not from CRAN"))
    }
  }

  ## 3. package names must match
  if (resolution$package != candidate$package) {
    return(structure(FALSE, reason = "Package names differ"))
  }

  ## 4. installed package must not be older for direct refs
  if (resolution$direct) {
    if (
      candidate$type == "installed" &&
        package_version(resolution$version) > candidate$version
    ) {
      return(structure(FALSE, reason = "Direct ref needs update"))
    }
  }

  ## 5. version requirements must be satisfied. Otherwise good.
  if (resolution$remote[[1]]$version == "") {
    return(TRUE)
  }

  if (
    !version_satisfies(
      candidate$version,
      resolution$remote[[1]]$atleast,
      resolution$remote[[1]]$version
    )
  ) {
    return(structure(FALSE, reason = "Insufficient version"))
  }

  TRUE
}

installedok_remote_cran <- function(installed, solution, config, ...) {
  if (solution$platform != "source") {
    # Binary packages are simple. We need to match `$built` to make sure
    # that we are installing the same build.
    identical(installed$package, solution$package) &&
      identical(installed$version, solution$version) &&
      installedok_remote_standard_platform(installed, solution, config, ...) &&
      identical(installed[["built"]], solution[["built"]])
  } else {
    # Source packages are different, because only the installed one has a
    # 'Built' field. So we require `Repository == "CRAN"` for these.
    identical(installed$package, solution$package) &&
      identical(installed$version, solution$version) &&
      identical(installed[["repository"]], "CRAN")
  }
}

## ----------------------------------------------------------------------
## Internal functions

type_cran_resolve_current <- function(
  remote,
  direct,
  config,
  cache,
  dependencies
) {
  resolve_from_metadata(remote, direct, config, cache, dependencies)
}

type_cran_resolve_version <- function(
  remote,
  direct,
  config,
  cache,
  dependencies
) {
  remote
  direct
  config
  cache
  dependencies

  # remote can be a list
  if (!inherits(remote, "remote_ref")) {
    return(async_map(
      remote,
      type_cran_resolve_version,
      direct = direct,
      config = config,
      cache = cache,
      dependencies = dependencies
    )$then(function(reslist) res_list_to_df(reslist)))
  }

  if (remote$atleast == ">=") {
    throw(pkg_error(
      "Version ranges are not implemented yet."
    ))
  }

  url_remote <- remote
  url_remote$url <- paste0(
    config$get("cran-mirror")[[1]],
    "/src/contrib/Archive/",
    remote$package,
    "/",
    remote$package,
    "_",
    remote$version,
    ".tar.gz"
  )
  url_remote$hash <- url_hash(paste0(remote$package, "=url::", remote$url))

  resolve_remote_url(url_remote, direct, config, cache, dependencies)$then(
    function(res) {
      # Also resolve as the main package. This should not be needed,
      # but we do it in case the package file is both in the main repo
      # and Archive, or in case the binary version matches the requested
      # one, which is typical shortly after a new release.
      res$unknown_deps <- c(res$unknown_deps, remote$package)

      # Metadata for standard::, not url::
      res$metadata[["RemoteRef"]] <- remote$package
      res$metadata[["RemoteRepos"]] <- config$get("cran-mirror")[[1]]
      res$metadata[["RemotePkgPlatform"]] <- "source"
      res$metadata[["RemoteSha"]] <- remote$version
      res
    }
  )$catch(error = function(err) {
    # if it is not in Archive, try the main package.
    # This will fail if not the right vesion
    resolve_from_metadata(remote, direct, config, cache, dependencies)
  })
}

test_that("satisfies_remote", {
  res <- make_fake_resolution(`cran::crayon@>=1.0.0` = list())

  ## GitHub type is never good
  bad1 <- make_fake_resolution(`github::r-lib/crayon` = list())
  expect_false(ans <- satisfy_remote_cran(res, bad1))
  expect_match(attr(ans, "reason"), "Type must be")

  ## Missing DESCRIPTION for installed type
  bad2 <- make_fake_resolution(`installed::foobar` = list())
  expect_false(ans <- satisfy_remote_cran(res, bad2))
  expect_match(attr(ans, "reason"), "not from CRAN")

  ## installed, but not from CRAN
  fake_desc <- desc::desc("!new")
  fake_desc$set(Repository = "Not CRAN")
  bad3 <- make_fake_resolution(
    `installed::foobar` = list(
      extra = list(list(description = fake_desc))
    )
  )
  expect_false(ans <- satisfy_remote_cran(res, bad3))
  expect_match(attr(ans, "reason"), "not from CRAN")

  ## CRAN type, but package name does not match
  bad4 <- make_fake_resolution(`cran::crayon2` = list())
  expect_false(ans <- satisfy_remote_cran(res, bad4))
  expect_match(attr(ans, "reason"), "names differ")

  ## installed type, but package name does not match
  bad5 <- make_fake_resolution(
    `installed::foobar` = list(
      package = "crayon2",
      extra = list(list(repotype = "cran"))
    )
  )
  expect_false(ans <- satisfy_remote_cran(res, bad5))
  expect_match(attr(ans, "reason"), "names differ")

  ## CRAN type, but version is not good enough
  bad6 <- make_fake_resolution(`cran::crayon` = list(version = "0.0.1"))
  expect_false(ans <- satisfy_remote_cran(res, bad6))
  expect_match(attr(ans, "reason"), "Insufficient version")

  ## Same version, CRAN
  ok1 <- make_fake_resolution(`cran::crayon` = list())
  expect_true(satisfy_remote_cran(res, ok1))

  ## Newer version, CRAN
  ok2 <- make_fake_resolution(`cran::crayon` = list(version = "2.0.0"))
  expect_true(satisfy_remote_cran(res, ok2))

  ## Same version, installed
  ok3 <- make_fake_resolution(
    `installed::foobar` = list(
      package = "crayon",
      extra = list(list(repotype = "cran"))
    )
  )
  expect_true(satisfy_remote_cran(res, ok3))

  ## Newer version, installed
  ok4 <- make_fake_resolution(
    `installed::foobar` = list(
      package = "crayon",
      version = "2.0.0",
      extra = list(list(repotype = "cran"))
    )
  )
  expect_true(satisfy_remote_cran(res, ok4))

  # direct ref is slightly different
  res <- make_fake_resolution(
    `cran::crayon@>=1.0.0` = list(
      direct = TRUE
    )
  )

  ## Installed, but old
  bad6 <- make_fake_resolution(
    `installed::foobar` = list(
      package = "crayon",
      version = "0.0.1",
      extra = list(list(repotype = "cran"))
    )
  )
  expect_false(ans <- satisfy_remote_cran(res, bad6))
  expect_match(attr(ans, "reason"), "Direct ref needs update")

  # without version req
  res <- make_fake_resolution(`cran::crayon` = list())

  ok5 <- make_fake_resolution(`cran::crayon` = list())
  expect_true(satisfy_remote_cran(res, ok5))
})

test_that("installedok", {
  sol <- list(
    package = "pkg1",
    version = "1.0.0",
    platform = "aarch64-apple-darwin20",
    built = "R 4.2.0; aarch64-apple-darwin20; 2022-09-13 21:32:41 UTC; unix"
  )

  inst <- list(
    package = "pkg1",
    version = "1.0.0",
    platform = "aarch64-apple-darwin20",
    built = "R 4.2.0; aarch64-apple-darwin20; 2022-09-13 21:32:41 UTC; unix",
    repository = "CRAN"
  )

  expect_true(installedok_remote_cran(inst, sol))
  expect_true(installedok_remote_cran(
    utils::modifyList(inst, list(platform = "*")),
    sol
  ))

  sol <- list(
    package = "pkg1",
    version = "1.0.0",
    platform = "source"
  )

  expect_true(installedok_remote_cran(inst, sol))
})

## A simple class with two slots
track <- setClass("track", slots = c(x = "numeric", y = "numeric"))
## an object from the class
t1 <- track(x = 1:10, y = 1:10 + rnorm(10))

setGeneric("plot")

setMethod(
  "plot",
  signature(x = "track", y = "missing"),
  function(x, y, ...) plot(x@x, x@y, ...)
)

ghcr_const_annotations <- list(
  com.github.package.type = "r_package"
  #  org.opencontainers.image.authors = "Gabor Csardi",
  #  org.opencontainers.image.url = "https://github.com/r-lib/pak",
  #  org.opencontainers.image.documentation = "https://pak.r-lib.org/",
  #  org.opencontainers.image.source = "https://github.com/r-lib/pak",
  #  org.opencontainers.image.title = "pak R package",
  #  org.opencontainers.image.description = "Package manager for R",
  #  org.opencontainers.image.licenses = "GPL-3"
)

ghcr_const_annotations_js <- function() {
  paste0(
    glue::glue('"{names(ghcr_const_annotations)}": "{ghcr_const_annotations}"'),
    collapse = ",\n      "
  )
}

ghcr_get_package_data <- function(path) {
  if (!file.exists(path)) {
    throw(pkg_error("{.path {path}} does not exist."))
  }

  plt <- pkgcache::current_r_platform_data()
  if (plt$os != "linux" && !grepl("^linux-", plt$os)) {
    throw(pkg_error(
      "{.fn ghcr_push_package} only works on Linux.",
      "i" = "It does not work on this platform ({.var {plt$platform}}) yet."
    ))
  }

  chain_error(
    dsc <- desc::desc(path),
    pkg_error(
      "Cannot read {.code DESCRIPTION} from {.file {path}}.",
      "i" = "A valid R package must have a {.code DESCRIPTION} file."
    )
  )

  chain_error(
    built <- dsc$get_built(),
    pkg_error(
      "Cannot find {.code Built} field in package at {.file {path}}",
      "i" = "A binary R package must have a {.code Built} field in
             {.code DESCRIPTION}."
    )
  )

  sha256 <- cli::hash_file_sha256(path)
  data_frame(
    path = path,
    md5 = cli::hash_file_md5(path),
    sha256 = sha256,
    digest = paste0("sha256:", sha256),
    image_title = paste("R package", dsc$get_field("Package")),
    size = file.size(path),
    buildtime = format_iso_8601(built$Date),
    r_version = as.character(built$R),
    r_platform = built$Platform,
    package_name = dsc$get_field("Package"),
    package_version = dsc$get_field("Version"),
    license = dsc$get_field("License"),
    arch = ghcr_canonize_arch(built$Platform),
    os = ghcr_canonize_os(built$Platform)
  )
}

ghcr_create_oci_repo <- function(
  pkgs,
  repo = NULL,
  oci_path = NULL,
  cleanup = TRUE
) {
  if (nrow(pkgs) > 1) {
    throw(pkg_error("Multiple package versions are not supported yet."))
  }

  repo <- repo %||% Sys.getenv("GITHUB_REPOSITORY", NA_character_)
  if (is.na(repo)) {
    throw(pkg_error("Cannot detect GitHub repository"))
  }

  oci_path <- oci_path %||% tempfile()
  mkdirp(oci_path)
  blob_path <- file.path(oci_path, "blobs", "sha256")
  mkdirp(blob_path)

  # Create oci-layout file
  cat(
    '{"imageLayoutVersion": "1.0.0"}\n',
    file = file.path(oci_path, "oci-layout")
  )

  # Copy package files to blobs
  file.copy(pkgs$path, file.path(blob_path, pkgs$sha256))

  # Config is dummy now
  pkgs$image_config <- "{}"
  pkgs$image_config_hash <- cli::hash_sha256(pkgs$image_config)
  write_files(pkgs$image_config, file.path(blob_path, pkgs$image_config_hash))

  # Create manifests
  pkgs$manifest <- ghcr_image_manifest(pkgs)
  pkgs$manifest_hash <- cli::hash_sha256(pkgs$manifest)
  write_files(pkgs$manifest, file.path(blob_path, pkgs$manifest_hash))

  # Create index
  imidx <- ghcr_image_index(pkgs)
  imidx_hash <- cli::hash_sha256(imidx)
  write_files(imidx, file.path(blob_path, imidx_hash))

  # index.json
  idxjs <- jsonlite::prettify(glue::glue(
    '{
       "schemaVersion": 2,
       "manifests": [
          {
            "mediaType": "application/vnd.oci.image.index.v1+json",
            "digest": "sha256:<<imidx_hash>>",
            "size": <<nchar(imidx, "bytes")>>
          }
       ]
     }',
    .open = "<<",
    .close = ">>"
  ))
  write_files(idxjs, file.path(oci_path, "index.json"))

  # policy.json
  policy_file <- file.path(oci_path, "policy.json")
  policy <- jsonlite::prettify(
    '{
       "default": [
         {
           "type": "insecureAcceptAnything"
         }
       ],
       "transports":
         {
           "docker-daemon":
             {
               "": [{"type":"insecureAcceptAnything"}]
             }
         }
     }'
  )
  write_files(policy, policy_file)

  oci_path
}

ghcr_uri <- function() {
  Sys.getenv(
    "PAK_GHCR_URI",
    "docker://ghcr.io/r-lib/pak"
  )
}

ghcr_user <- function() {
  Sys.getenv("PAK_GHCR_USER", "gaborcsardi")
}

ghcr_token <- function() {
  Sys.getenv("PAK_GHCR_TOKEN", gitcreds::gitcreds_get()$password)
}

ghcr_push_oci_repo <- function(oci_path, ghcr_tag = NULL) {
  if (!file.exists(oci_path)) {
    throw(pkg_error("Could not find OCI repository at {.path {oci_path}}."))
  }
  ghcr_tag <- ghcr_tag %||% "TODO"

  skopeo <- check_skopeo()
  oci_path <- normalizePath(oci_path, winslash = "/")
  policy_file <- file.path(oci_path, "policy.json")
  args <- c(
    "copy",
    "--all",
    "--retry-times",
    20,
    "--preserve-digests",
    "--policy",
    policy_file,
    paste0("--dest-creds=", ghcr_user(), ":", ghcr_token()),
    paste0("oci:", oci_path),
    paste0(ghcr_uri(), ":", ghcr_tag)
  )
  "TODO"
}

ghcr_canonize_arch <- function(platform) {
  arch <- strsplit(platform, "-", fixed = TRUE)[[1]][1]
  c("aarch64" = "arm64", "x86_64" = "amd64", "s390x" = "s390s")[[arch]]
}

ghcr_canonize_os <- function(platform) {
  os <- strsplit(platform, "-", fixed = TRUE)[[1]][3]
  if (substr(os, 1, 6) == "darwin") {
    os <- "darwin"
  }
  if (substr(os, 1, 5) == "mingw") {
    os <- "windows"
  }
  if (substr(os, 1, 7) == "solaris") {
    os <- "solaris"
  }
  os
}

ghcr_image_index <- function(pkgs) {
  tmpl <- '
    {
      "mediaType": "application/vnd.oci.image.manifest.v1+json",
      "digest": "sha256:<<manifest_hash>>",
      "size": <<nchar(manifest, "bytes")>>,
      "platform": {
        "architecture": "<<arch>>",
        "os": "<<os>>",
        "os.version": "R <<r_version>>",
        "r_version": "<<r_version>>",
        "r_platform": "<<r_platform>>"
      },
      "annotations": {
        "org.opencontainers.image.ref.name": "<<package_version>>--<<r_version>>--<<r_platform>>",
        "io.r-hub.package.digest": "<<digest>>"
      }
    }'
  mnfts <- glue::glue_data(pkgs, tmpl, .open = "<<", .close = ">>")

  buildtime <- max(pkgs$buildtime)
  package_name <- unique(package_name)
  if (length(package_name) != 1) {
    throw(pkg_error(
      "Cannot push different packages at the same time.",
      "i" = "Found packages {.pkg {package_name}}."
    ))
  }

  package_version <- unique(pkgs$package_version)
  if (length(package_version) != 1) {
    throw(pkg_error(
      "Non-matching package versions when pushing package {.pkg {package_name}}.",
      "i" = "All package versions must be the same.",
      "i" = "Found package versions: {.var {package_version}}."
    ))
  }

  tmpl2 <- '
    {
      "schemaVersion": 2,
      "manifests": [
         <<paste0(mnfts, collapse = ",\n")>>
      ],
      "annotations": {
         <<ghcr_const_annotations_js()>>,
         "org.opencontainers.image.created": "<<buildtime>>",
         "org.opencontainers.image.version": "<<package_version>>"
      }
    }'

  jsonlite::prettify(glue::glue(tmpl2, .open = "<<", .close = ">>"))
}

ghcr_image_manifest <- function(pkgs) {
  tmpl <- '
    {
      "schemaVersion": 2,
      "config": {
        "mediaType": "application/vnd.oci.image.config.v1+json",
        "digest": "sha256:<<image_config_hash>>",
        "size": <<nchar(image_config, "bytes")>>
      },
      "layers": [
        {
          "mediaType": "application/vnd.oci.image.layer.v1.tar+gzip",
          "digest": "<<digest>>",
          "size": <<size>>,
          "annotations": {
            "org.opencontainers.image.title": "<<image_title>>"
          }
        }
      ],
      "annotations": {
         <<ghcr_const_annotations_js()>>,
         "org.opencontainers.image.created": "<<buildtime>>",
         "org.opencontainers.image.version": "<<package_version>>"
      }
   }'

  glue::glue_data(pkgs, tmpl, .open = "<<", .close = ">>")
}


write_file <- function(txt, path) {
  out <- file(path, open = "wb")
  on.exit(close(out), add = TRUE)
  cat(txt, file = out, sep = "")
}

write_files <- function(txts, paths) {
  invisible(mapply(write_file, txts, paths))
}

find_skopeo <- function() {
  path <- Sys.which("skopeo")
  if (path != "") {
    return(path)
  }
  if (file.exists(cand <- "/usr/local/bin/skopeo")) {
    return(cand)
  }
  if (file.exists(cand <- "/opt/homebrew/bin/skopeo")) {
    return(cand)
  }
  throw(pkg_error("Need skopeo to push packages."))
}

skopeo_version <- function() {
  skopeo <- find_skopeo()
  out <- processx::run(skopeo, "--version")
  re_ver <- "[ ]([0-9]+[.][0-9]+[.][0-9]+)"
  if (!grepl(re_ver, out$stdout)) {
    stop("Cannot determine skopeo version")
  }
  mch <- regexpr(re_ver, out$stdout, perl = TRUE)
  beg <- attr(mch, "capture.start")[1]
  end <- beg + attr(mch, "capture.length")[1] - 1L
  package_version(substr(out$stdout, beg, end))
}

check_skopeo <- function() {
  skopeo <- find_skopeo()
  skopeo_ver <- skopeo_version()
  if (skopeo_ver < "1.6.0") {
    throw(pkg_error(
      "Need at least skopeo 1.6.0 to push packages",
      "i" = "Found skopeo {.emph {skopoe_ver}} at {.path {skopeo}}."
    ))
  }
  skopeo[[1]]
}

#' @export

format.pkg_scan_deps <- function(x, ...) {
  labels <- c(
    prod = "Dependencies",
    test = "Test dependencies",
    dev = "Development dependencies",
    # TODO: generic label for others
    NULL
  )
  lns <- lapply(seq_along(labels), function(i) {
    deps <- x[x$type == names(labels)[i], , drop = FALSE]
    if (nrow(deps) == 0) {
      return(NULL)
    }
    fls <- tapply(deps$path, deps$package, "c", simplify = FALSE)
    fls[] <- lapply(fls, unique)
    fls <- vcapply(fls, paste, collapse = ", ")
    pkg <- format(names(fls))
    flsw <- cli::console_width() - nchar(pkg[1]) - 5
    c(
      "",
      cli::col_yellow(paste0(labels[i], ":")),
      paste0(
        cli::col_grey("+ "),
        cli::col_blue(pkg),
        cli::col_grey(" @ "),
        cli::col_silver(cli::ansi_strtrim(fls, flsw))
      )
    )
  })

  unlist(lns)
}

#' @export

print.pkg_scan_deps <- function(x, ...) {
  writeLines(format(x, ...))
  invisible(x)
}

#' @export

`[.pkg_scan_deps` <- function(x, i, j, drop = FALSE) {
  class(x) <- setdiff(class(x), "pkg_scan_deps")
  NextMethod("[")
}

#' @export
#' @rdname pkg_dep_types

pkg_dep_types_hard <- function() c("Depends", "Imports", "LinkingTo")

#' @export
#' @rdname pkg_dep_types

pkg_dep_types_soft <- function() c("Suggests", "Enhances")

#' Possible package dependency types
#'
#' Hard dependencies are needed for a package to load, soft dependencies
#' are optional.
#'
#' @return A string vector of dependency types, capitalized.
#'
#' @family package dependency utilities
#' @export

pkg_dep_types <- function() c(pkg_dep_types_hard(), pkg_dep_types_soft())

make_null_deps <- function() {
  data_frame(
    ref = character(),
    type = character(),
    package = character(),
    op = character(),
    version = character()
  )
}

parse_deps <- function(deps, type) {
  assert_that(length(deps) == length(type))
  deps <- lapply(strsplit(deps, ","), str_trim)
  rx <- paste0(
    "(?<type>)",
    "^\\s*",
    "(?<package>[^\\s]+)",
    "\\s*",
    "(?:[(](?<op>>|>=|==|<|<=)\\s*(?<version>[-0-9\\.]+)[)])?\\s*$"
  )
  base <- base_packages()
  lapply(seq_along(deps), function(i) {
    x <- omit_cols(re_match(deps[[i]], pattern = rx), c(".text", ".match"))
    x$type <- if (length(x$type) > 0) type[[i]] else character()
    x[!x$package %in% base, ]
  })
}

deps_from_desc <- function(deps, last) {
  op_ver <- strsplit(deps$version, "\\s+")
  deps$op <- vcapply(op_ver, "[", 1)
  deps$op[deps$op == "*"] <- ""
  deps$version <- vcapply(op_ver, "[", 2)
  deps$version[is.na(deps$version)] <- ""
  deps$ref <- paste0(deps$package, if (last) "@last")
  base <- base_packages()
  res <- as_data_frame(deps[
    !deps$package %in% base,
    c("ref", "type", "package", "op", "version")
  ])
  rownames(res) <- NULL
  res
}

parse_all_deps <- function(deps) {
  deps <- na.omit(deps)
  res <- do.call(rbind, parse_deps(deps, names(deps)))
  if (is.null(res)) {
    res <- parse_deps("", "")[[1]]
  }
  res$ref <- res$package
  res[, c("ref", setdiff(names(res), "ref"))]
}

resolve_ref_deps <- function(deps, remotes, extra) {
  deps <- deps_from_desc(deps, last = FALSE)

  parse <- function(x) {
    str_trim(strsplit(x, "\\s*,\\s*", perl = TRUE)[[1]])
  }

  check_names <- function(x) {
    nms <- vcapply(x, function(e) e$package %||% NA_character_)
    bad <- is.na(nms)
    if (any(bad)) {
      badpkgs <- vcapply(x, "[[", "ref")[bad]
      badtypes <- vcapply(x, "[[", "type")[bad]
      throw(pkg_error(
        "Cannot determine package {cli::qty(sum(bad))}name{?s} for
         {sum(bad)} package{?s}: {.val {badpkgs}}.",
        i = "Maybe you need to add a {.code <packagename>{zwnj()}=} prefix?"
      ))
    }
    nms
  }

  if (!is.na(remotes)) {
    remotes <- str_trim(na.omit(remotes))
    remotes <- parse(remotes)
    remotes_packages <- check_names(parse_pkg_refs(remotes))
    keep <- which(remotes_packages %in% deps$package)
    deps$ref[match(remotes_packages[keep], deps$package)] <- remotes[keep]
  }

  if (length(extra) > 0) {
    xdeps <- parse_all_deps(extra)
    xdeps$package <- check_names(parse_pkg_refs(xdeps$package))
    deps <- rbind(deps, xdeps)
  }

  deps
}

#' Shorthands for dependency specifications
#'
#' @details
#' ```{r child = "tools/doc/deps.Rmd"}
#' ```
#' `r doc_share_rmd("tools/doc/deps.Rmd", "inst/docs/deps.rds")`
#'
#' @param deps See below.
#' @return A named list with two character vectors: `direct`, `indirect`,
#' the dependency types to use for direct installations and dependent
#' packages.
#'
#' @family package dependency utilities
#' @export

as_pkg_dependencies <- function(deps) {
  assert_that(is_dependencies(deps))

  hard <- pkg_dep_types_hard()
  soft <- pkg_dep_types()

  res <- if (isTRUE(deps)) {
    list(c(hard, "Suggests"), hard)
  } else if (identical(deps, FALSE)) {
    list(character(), character())
  } else if (is_na_scalar(deps)) {
    list(hard, hard)
  } else if (is.list(deps) && all(names(deps) == c("direct", "indirect"))) {
    deps
  } else {
    list(deps, hard)
  }

  res <- lapply(res, function(x) {
    if ("hard" %in% x) {
      x <- unique(c(hard, setdiff(x, "hard")))
    }
    if (any(c("soft", "all") %in% x)) {
      x <- unique(c(soft, setdiff(x, c("all", "soft"))))
    }
    x
  })

  names(res) <- c("direct", "indirect")
  res
}

extra_config_fields <- function(x) {
  grep("^config/needs/", x, value = TRUE, ignore.case = TRUE)
}

test_that("pkg_dep_types_*", {
  expect_snapshot(pkg_dep_types_hard())
  expect_snapshot(pkg_dep_types_soft())
  expect_snapshot(pkg_dep_types())
})

test_that("make_null_deps", {
  expect_snapshot(
    make_null_deps(),
    variant = paste0("pillar-", packageVersion("pillar"))
  )
})

test_that("parse_deps", {
  # edge cases
  expect_snapshot(parse_deps(character(), character()))
  expect_snapshot(parse_deps("foo", "Imports"))

  # vectorized
  expect_snapshot(parse_deps(c("foo", "foo"), c("Imports", "LinkingTo")))

  # requirement
  expect_snapshot(parse_deps("foo (>= 1.0.0)", "Imports"))

  # multiple packages
  expect_snapshot(parse_deps("foo (>= 1.0.0), bar", "Imports"))

  # newlines are fine
  expect_snapshot(parse_deps("foo (>= 1.0.0)\n bar", "Imports"))
  expect_snapshot(parse_deps("foo (>= \n 1.0.0)\n bar", "Imports"))

  # various relations
  expect_snapshot(parse_deps("R (== 4.2.1)", "Depends"))
  expect_snapshot(parse_deps("R (< 4.2.1)", "Depends"))
  expect_snapshot(parse_deps("R (> 4.2.1)", "Depends"))
  expect_snapshot(parse_deps("R (<= 4.2.1)", "Depends"))

  # base packages are not included
  expect_snapshot(parse_deps("foo, stats", "Imports"))
  expect_snapshot(
    parse_deps("grid, stats", "Imports"),
    variant = paste0("pillar-", packageVersion("pillar"))
  )
})

test_that("parse_all_deps", {
  # edge cases
  expect_snapshot(
    parse_all_deps(c(Imports = NA_character_)),
    variant = paste0("pillar-", packageVersion("pillar"))
  )
  expect_snapshot(
    parse_all_deps(c(Imports = "grid")),
    variant = paste0("pillar-", packageVersion("pillar"))
  )

  expect_snapshot(
    parse_all_deps(c(
      Imports = "foo (>= 1.0.0), bar",
      Suggests = "baz, foobaz (>= 2.0.0)"
    )),
    variant = paste0("pillar-", packageVersion("pillar"))
  )
})

test_that("resolve_ref_deps", {
  dsc <- desc::desc(
    text = c(
      "Suggests: covr, jsonlite, testthat (>= 3.1.0)",
      "Imports: assertthat, curl, R6, rlang (>= 1.0.0)",
      "Remotes: r-lib/covr",
      "Config/Needs/website: pkgdown, r-lib/downlit"
    )
  )
  expect_snapshot(resolve_ref_deps(dsc$get_deps(), NA_character_, NULL))
  expect_snapshot(resolve_ref_deps(
    dsc$get_deps(),
    dsc$get("Remotes"),
    dsc$get(extra_config_fields(dsc$fields()))
  ))
})

test_that("resolve_ref_deps, package name from remote", {
  dsc <- desc::desc(
    text = c(
      "Imports: foo",
      "Remotes: foo=r-lib/bar",
      "Config/Needs/website: foobar=r-lib/baz"
    )
  )
  expect_snapshot(resolve_ref_deps(
    dsc$get_deps(),
    dsc$get("Remotes"),
    dsc$get(extra_config_fields(dsc$fields()))
  ))

  dsc <- desc::desc(
    text = c(
      "Imports: foo",
      "Remotes: url::http://example.com"
    )
  )
  expect_snapshot(
    error = TRUE,
    resolve_ref_deps(
      dsc$get_deps(),
      dsc$get("Remotes"),
      NULL
    )
  )
})

test_that("as_pkg_dependencies", {
  expect_snapshot(as_pkg_dependencies(TRUE))
  expect_snapshot(as_pkg_dependencies("all"))
  expect_snapshot(as_pkg_dependencies("hard"))
  expect_snapshot(as_pkg_dependencies(FALSE))
  expect_snapshot(as_pkg_dependencies(NA))
  expect_snapshot(as_pkg_dependencies(
    list(direct = "Imports", indirect = c("Imports", "Suggests"))
  ))
  expect_snapshot(as_pkg_dependencies(c("Depends", "Imports", "LinkingTo")))
  expect_snapshot(as_pkg_dependencies(
    list(direct = "all", indirect = "hard")
  ))
  expect_snapshot(as_pkg_dependencies(
    list(direct = "all", indirect = "all")
  ))
  expect_snapshot(as_pkg_dependencies(
    list(direct = "all", indirect = c("hard", "soft"))
  ))
})

# -------------------------------------------------------------------------
# Dummy CRAN app

cran_app <- function(...) {
  asNamespace("pkgcache")$cran_app(...)
}

bioc_app <- function(...) {
  asNamespace("pkgcache")$bioc_app(...)
}

dcf <- function(...) {
  asNamespace("pkgcache")$dcf(...)
}

fix_port <- function(...) {
  asNamespace("pkgcache")$fix_port(...)
}

cran_app_pkgs <- dcf(
  "
  Package: pkg1
  Version: 1.0.0

  Package: pkg1
  Version: 0.9.0

  Package: pkg1
  Version: 0.8.0

  Package: pkg2
  Version: 1.0.0
  Depends: pkg1

  Package: pkg3
  Version: 1.0.0
  Depends: pkg2

  Package: pkg3
  Version: 0.9.9

  Package: crayon
  Version: 1.0.0

  Package: needspak
  Imports: pak

  Package: pak

  Package: futurama
  Depends: R (>= 3000.0)

  Package: needsfuturama
  Imports: futurama

  Package: dplyr
  Imports: tibble
  Suggests: testthat

  Package: tibble

  Package: testthat

  Package: curl
  SystemRequirements: libcurl: libcurl-devel (rpm) or libcurl4-openssl-dev (deb).
"
)

fake_cran <- webfakes::local_app_process(
  cran_app(cran_app_pkgs),
  opts = webfakes::server_opts(num_threads = 3)
)

bioc_app_pkgs <- dcf(
  "
  Package: Biobase
  Version: 1.2.3
  Depends: R (>= 2.10), BiocGenerics(>= 0.27.1), utils
  Imports: methods
  Suggests: tools, tkWidgets, ALL, RUnit, golubEsets
"
)

fake_bioc <- webfakes::local_app_process(
  bioc_app(bioc_app_pkgs),
  opts = webfakes::server_opts(num_threads = 3)
)

setup_fake_apps <- function(
  cran_app = NULL,
  bioc_app = NULL,
  cran_repo = NULL,
  bioc_repo = NULL,
  cran_options = NULL,
  bioc_options = NULL,
  .local_envir = parent.frame()
) {
  cran_app <- if (!is.null(cran_app)) {
    cran_app
  } else if (!is.null(cran_repo)) {
    app <- webfakes::local_app_process(
      cran_app(cran_repo, options = as.list(cran_options)),
      opts = webfakes::server_opts(num_threads = 3),
      .local_envir = .local_envir
    )
    assign(".cran_app", app, envir = .local_envir)
    app
  } else {
    fake_cran
  }

  bioc_app <- if (!is.null(bioc_app)) {
    bioc_app
  } else if (!is.null(bioc_repo)) {
    app <- webfakes::local_app_process(
      bioc_app(bioc_repo, options = as.list(bioc_options)),
      opts = webfakes::server_opts(num_threads = 3),
      .local_envir = .local_envir
    )
    assign(".bioc_app", app, envir = .local_envir)
    app
  } else {
    fake_bioc
  }

  withr::local_options(
    repos = c(CRAN = cran_app$url()),
    pkg.cran_metadata_url = cran_app$url(),
    .local_envir = .local_envir
  )
  withr::local_envvar(
    R_PKG_CRAN_METADATA_URL = cran_app$url(),
    R_BIOC_CONFIG_URL = paste0(bioc_app$url(), "/config.yaml"),
    R_BIOC_VERSION = NA_character_,
    R_BIOC_MIRROR = bioc_app$url(),
    .local_envir = .local_envir
  )
}

# -------------------------------------------------------------------------
# GH app

gh_app_desc <- function(pkg) {
  sprintf("Package: %s\nVersion: 1.0.0\n", pkg)
}

random_sha <- function() {
  paste(
    sample(c(0:9, letters[1:6]), 64, replace = TRUE),
    collapse = ""
  )
}

gh_app_repos <- list(
  users = list(
    "r-lib" = list(
      repos = list(
        pak = list(
          commits = list(
            list(
              sha = "111ef906acb58fe406370f7bc0a72cac55dbbb231ea687494c25742ca521255a",
              branch = "main",
              tag = "HEAD",
              files = list("DESCRIPTION" = gh_app_desc("pak"), NAMESPACE = "")
            ),
            list(
              sha = "a503fe843f11c279864f29d58137f8de319d115b239ce48ccc15406306019480",
              branch = "main",
              tag = "v0.1.2",
              files = list("DESCRIPTION" = gh_app_desc("pak"), NAMESPACE = "")
            ),
            list(
              sha = "e65de1e9630dbfcaf1044718b742bf806486b107239ce48ccc15406306019480",
              branch = "main",
              files = list("DESCRIPTION" = gh_app_desc("pak"), NAMESPACE = "")
            ),
            list(
              sha = "b001d6ddeab1589ad367b62baabbeeb2af3b0ebac2e61d239df660c1d63e3232",
              branch = "somebranch",
              pull = 90,
              files = list("DESCRIPTION" = gh_app_desc("pak"), NAMESPACE = "")
            ),
            list(
              sha = "b001d6ddeab1589ad367b62baabbeeb2af3b0ebac2e61d239df660c1d63e3232",
              latestRelease = TRUE,
              tagName = "v1.2.3",
              files = list("DESCRIPTION" = gh_app_desc("pak"), NAMESPACE = "")
            )
          )
        ),
        bad = list(
          commits = list(
            list(
              sha = "546d9eab84b002c35302dda3822560950c7528cfc9ef1b916cecd9dbef3cf6b6",
              tag = "HEAD",
              branch = "main",
              files = list(
                DESCRIPTION = "this is not\na good file\n",
                "bin/DESCRIPTION" = charToRaw("\xf0\xb0\xa0")
              )
            ),
            list(
              sha = "546d9eab84b002c35302dda3822560950c7528cfc9ef1b916cecd9dbef3cf6b6",
              pull = 100,
              branch = "other",
              files = list(DESCRIPTION = "this is not\na good file\n")
            )
          )
        ),
        crayon = list(
          commits = list(
            list(
              sha = "bdd9a1bcf062396790c341cf1dba563eb0277f2ca0a6d524bc3da98a9a6f2975",
              tag = "HEAD",
              branch = "main",
              files = list(DESCRIPTION = gh_app_desc("crayon"), NAMESPACE = "")
            ),
            list(
              sha = "b5221ab024605019800ddea474f7a0981a4d53f719f5af2b1af627b34e0760b2",
              branch = "b5221ab024605019800ddea474f7a0981a4d53f719f5af2b1af627b34e0760b2",
              files = list(DESCRIPTION = gh_app_desc("crayon"), NAMESPACE = "")
            ),
            list(
              sha = "9d93692f8f7c1d6b2308d0c4aa83cdc2d99ec1fd0097cede1d9aa1301247cb01",
              branch = "pr61",
              pull = 79,
              files = list(DESCRIPTION = gh_app_desc("crayon"), NAMESPACE = "")
            )
          )
        ),
        pkgconfig = list(
          commits = list(
            list(
              sha = "c9be9cde5e91ad771d1b5150781e6e8d32a7be0e9ab227bdf45cb41ad513004c",
              branch = "pr7",
              pull = 7,
              files = list(
                DESCRIPTION = gh_app_desc("pkgconfig"),
                NAMESPACE = ""
              )
            )
          )
        )
      )
    ),

    "wesm" = list(
      repos = list(
        "feather" = list(
          commits = list(
            list(
              sha = "ec40c1eae1ac83b86fc41bb2f5cd916152d19015649c3d209f2c08115dd993b1",
              tag = "HEAD",
              branch = "main",
              files = list(
                "R/DESCRIPTION" = gh_app_desc("feather"),
                NAMESPACE = ""
              )
            )
          )
        )
      )
    ),

    "gaborcsardi" = list(
      repos = list(
        "secret-test" = list(
          commits = list(
            list(
              sha = "599cc5d745d2079eddf1ff582b83d381e885cd30f33bafebbe83e73d010cfa93",
              tag = "HEAD",
              branch = "main",
              token = "b9984750bea6a170081ca98255c3b43fe5fb0978",
              files = list(
                "DESCRIPTION" = gh_app_desc("secret"),
                NAMESPACE = ""
              )
            )
          )
        ),
        "secret" = list(
          commits = list(
            list(
              sha = "7f9fb08e26015e05529cd4d7fc2a7edbd88c783d456ff83a96dcc58ace1d3ea5",
              tag = "HEAD",
              branch = "x",
              files = list(
                "DESCRIPTION" = gh_app_desc("secret"),
                NAMESPACE = ""
              )
            )
          )
        )
      )
    ),

    "tidyverse" = list(
      repos = list(
        "tidyverse.org" = list(
          commits = list(
            list(
              sha = "d998eab68c66d862c31a6091f9e71200b13bb44ea754e3371d098dcaa20e51a4",
              tag = "HEAD",
              branch = "main",
              files = list("foo" = "bar")
            )
          )
        )
      )
    ),

    "cran" = list(
      repos = list(
        "rJava" = list(
          commits = list(
            list(
              sha = "dfb3b64b13343e07b2db038777d9dc2aba5d824c5eca8c891c87bd4fd38d7256",
              tag = "HEAD",
              branch = "master",
              files = list(
                DESCRIPTION = "Package: rJava\nVersion: 1.0-6\nSystemRequirements: Java JDK 1.2 or higher (for JRI/REngine JDK 1.4 or higher), GNU make\n",
                NAMESPACE = ""
              )
            )
          )
        )
      )
    )
  )
)

fake_gh <- webfakes::local_app_process(
  asNamespace("pkgdepends")$gh_app(gh_app_repos),
  opts = webfakes::server_opts(num_threads = 3)
)

setup_fake_gh_app <- function(.local_envir = parent.frame()) {
  withr::local_envvar(
    .local_envir = .local_envir,
    R_PKG_GITHUB_API_URL = fake_gh$url()
  )
}


# -------------------------------------------------------------------------
# Name check app

new_check_app <- function() {
  `%||%` <- function(l, r) if (is.null(l)) r else l

  app <- webfakes::new_app()

  app$use(webfakes::mw_json())
  app$use(webfakes::mw_urlencoded())

  app$get("/crandb", function(req, res) {
    pkg <- sub("\"$", "", sub("^\"", "", req$query$key))
    if (pkg == "dbi") {
      res$send_json(
        list(
          total_rows = 20000,
          offset = 14000,
          rows = list(list(id = "DBI", key = "dbi", value = "DBI"))
        ),
        auto_unbox = TRUE
      )
    } else {
      res$send_json(list(
        total_rows = 20000,
        offset = 14000,
        rows = list()
      ))
    }
  })

  app$post("/wikipedia", function(req, res) {
    titles <- strsplit(req$form$titles, "|", fixed = TRUE)[[1]]
    Titles <- tools::toTitleCase(titles)
    ret <- list(
      query = list(
        normalized = list(list(from = titles, to = Titles)),
        pages = list(
          `11178` = list(
            pageid = 11178,
            title = Titles,
            extract = "The terms foobar (), foo, bar, and others are used ..."
          )
        )
      )
    )
    res$send_json(ret, auto_unbox = TRUE)
  })

  app$all(c("/echo", "/echo/define"), function(req, res) {
    out <- list(
      method = req$method,
      query = req$query_string,
      type = req$get_header("Content-Type") %||% NA_character_,
      body = rawToChar(req$.body %||% raw())
    )
    res$send_json(out, auto_unbox = TRUE)
  })

  app$get("/sentiment", function(req, res) {
    txt <- "abuse\t-3\nirony\t-1\nxo\t3\nxoxoxo\t4\n"
    res$send(txt)
  })

  app$get("/bioc/a", function(req, res) {
    res$send(paste0(
      collapse = "",
      c(
        "hello nobody, this is httpd@ip-172-30-0-33 running gitolite3 v3.6.6-6-g7c8f0ab on git 2.28.0",
        "",
        " R  \tpackages/a4",
        " R  \tpackages/a4Base",
        " R  \tpackages/a4Classif",
        " R  \tpackages/a4Core",
        " R  \tpackages/a4Preproc",
        " R  \tpackages/a4Reporting",
        " R  \tpackages/aCGH",
        " R  \tpackages/abseqR",
        " R  \tpackages/ag.db"
      ),
      "\n"
    ))
  })

  app$get("/bioc/A", function(req, res) {
    res$send(paste0(
      collapse = "",
      c(
        "hello nobody, this is httpd@ip-172-30-0-33 running gitolite3 v3.6.6-6-g7c8f0ab on git 2.28.0",
        "",
        " R  \tpackages/ABAData",
        " R  \tpackages/ABAEnrichment",
        " R  \tpackages/ABSSeq",
        " R  \tpackages/AGDEX",
        " R  \tpackages/AHPathbankDbs",
        " R  \tpackages/AIMS",
        " R  \tpackages/ALDEx2",
        " R  \tpackages/ALL",
        " R  \tpackages/ALLMLL",
        " R  \tpackages/ALPS",
        " R  \tpackages/AMARETTO"
      ),
      "\n"
    ))
  })

  app$get("/biocann/src/contrib/PACKAGES.gz", function(req, res) {
    tmp <- tempfile(fileext = ".gz")
    on.exit(unlink(tmp), add = TRUE)
    l <- c(
      "Package: adme16cod.db",
      "Version: 3.4.0",
      "Depends: R (>= 2.7.0), methods, AnnotationDbi (>= 1.31.18),",
      "        org.Rn.eg.db (>= 3.2.1)",
      "Imports: methods, AnnotationDbi",
      "Suggests: annotate, RUnit",
      "License: Artistic-2.0",
      "MD5sum: 3902516a40a503302ef732143b2394b9",
      "NeedsCompilation: no",
      "",
      "Package: ag.db",
      "Version: 3.2.3",
      "Depends: R (>= 2.7.0), methods, AnnotationDbi (>= 1.34.3),",
      "        org.At.tair.db (>= 3.3.0)",
      "Imports: methods, AnnotationDbi",
      "Suggests: DBI, annotate, RUnit",
      "License: Artistic-2.0",
      "MD5sum: e5913da38fe4487202306cacd885840d",
      "NeedsCompilation: no",
      "",
      "Package: agcdf",
      "Version: 2.18.0",
      "Depends: utils",
      "Imports: AnnotationDbi",
      "License: LGPL",
      "MD5sum: 5dd14bc6a6d2729f5e7b170105c78e48",
      "NeedsCompilation: no"
    )
    writeLines(l, con <- gzfile(tmp, open = "wb"))
    close(con)

    # We don't use send_file, because of a webfakes bug on Windows
    # with absolute paths. Webfakes prepends '/' to 'c:/...'.
    blob <- readBin(tmp, what = "raw", n = 10000)
    res$set_type("application/gzip")$send(blob)
  })

  app
}


new_sysreqs_app <- function() {
  app <- webfakes::new_app()
  app$use(webfakes::mw_json())
  app$use(webfakes::mw_urlencoded())

  db <- list(
    ubuntu = list(
      "22.04" = list(
        "java" = list(
          install_scripts = list("apt-get install -y default-jdk"),
          post_install = list(list(command = "R CMD javareconf"))
        ),
        "openssl" = list(
          install_scripts = list("apt-get install -y libssl-dev")
        ),
        "libcurl" = list(
          install_scripts = list("apt-get install -y libcurl4-openssl-dev")
        )
      ),
      "16.04" = list(
        "\\bgeos\\b" = list(
          pre_install = list(
            list(command = "apt-get install -y software-properties-common"),
            list(command = "add-apt-repository -y ppa:ubuntugis/ppa"),
            list(command = "apt-get update")
          ),
          install_scripts = list("apt-get install -y libgeos-dev")
        )
      )
    )
  )

  app$post("/__api__/repos/:id/sysreqs", function(req, res) {
    dist <- req$query$distribution
    rele <- req$query$release

    dsc <- desc::desc(text = rawToChar(req$.body))
    pkgsrq <- trimws(dsc$get("SystemRequirements"))
    if (is.na(pkgsrq)) {
      pkgsrq <- ""
    }

    if (dist == "ubuntu" && rele %in% c("16.04", "18.04", "20.04", "22.04")) {
      mydb <- db[[dist]][[rele]]
      srq <- lapply(names(mydb), function(nm) {
        if (grepl(nm, pkgsrq)) mydb[[nm]] else NULL
      })

      bf <- unlist(lapply(srq, "[[", "pre_install"), recursive = FALSE)
      is <- unlist(lapply(srq, "[[", "install_scripts"))
      af <- unlist(lapply(srq, "[[", "post_install"), recursive = FALSE)

      res$send_json(list(
        name = jsonlite::unbox("pkgdependssysreqs"),
        pre_install = bf,
        install_scripts = is,
        post_install = af
      ))
    } else {
      res$set_status(400)
      res$send_json(list(
        code = jsonlite::unbox(14),
        error = jsonlite::unbox("Unsupported system"),
        payload = jsonlite::unbox(NA)
      ))
    }
  })

  app
}

fake_sysreqs <- webfakes::local_app_process(new_sysreqs_app())

setup_fake_sysreqs_app <- function(.local_envir = parent.frame()) {
  withr::local_envvar(
    .local_envir = .local_envir,
    RSPM_ROOT = sub("/$", "", fake_sysreqs$url())
  )
}

transform_sysreqs_server <- function(x) {
  x <- gsub("https://packagemanager.posit.co", "<server>", x, fixed = TRUE)
  x <- gsub("http://127.0.0.1:[0-9]+", "<server>", x)
  x <- gsub("http://localhost:[0-9]+", "<server>", x)
  x
}

show_request <- function(req) {
  x <- jsonlite::fromJSON(rawToChar(req$content))
  cat(toupper(x$method), " ", x$type, sep = "", "\n")
  cat("Query string: ", x$query, sep = "", "\n")
  cat("Body: ", x$body, sep = "", "\n")
}

check_app <- webfakes::new_app_process(
  new_check_app(),
  opts = webfakes::server_opts(num_threads = 4)
)

transform_no_srcref <- function(x) {
  x <- sub("[ ]*at [-a-zA-Z0-9]+[.]R:[0-9]+:[0-9]+", "", x)
  x <- sub("[ ]*at line [0-9]+", "", x)
  x <- sub("\033[90m\033[39m", "", x, fixed = TRUE)
  x <- sub("Caused by error: ", "Caused by error:", x, fixed = TRUE)
  if (x[length(x)] == "") {
    x <- x[-length(x)]
  }
  x
}

transform_local_port <- function(x) {
  gsub("127\\.0\\.0\\.1:[0-9]+", "127.0.0.1:<port>", x)
}

transform_bioc_version <- function(x) {
  sub("3[.][0-9]+/bioc", "<bioc-version>/bioc", x)
}

transform_bytes <- function(x) {
  sub("[(][0-9]+ B[)]", "(<size>)", x)
}

transform_ext <- function(x) {
  x <- sub("[.](zip|tgz)", ".zip/.tgz/.tar.gz", x)
  x <- sub("_R_[-_a-z0-9A-Z]+[.]tar[.]gz", ".zip/.tgz/.tar.gz", x)
  x
}

transform_sha <- function(x) {
  gsub("[a-fA-F0-9]{64}", "<sha>", x)
}

transform_hash <- function(x) {
  x <- gsub("[a-f0-9]{32}", "<hash>", x)
  x <- gsub("[a-f0-9]{10}", "<hash>", x)
  x
}

transform_etag <- function(x) {
  sub("RemoteEtag: \"[a-z0-9]+\"", "RemoteEtag: \"<etag>\"", x)
}

transform_tempdir <- function(x) {
  x <- sub(tempdir(), "<tempdir>", x, fixed = TRUE)
  x <- sub(normalizePath(tempdir()), "<tempdir>", x, fixed = TRUE)
  x <- sub(
    normalizePath(tempdir(), winslash = "/"),
    "<tempdir>",
    x,
    fixed = TRUE
  )
  x <- sub("\\R\\", "/R/", x, fixed = TRUE)
  x <- sub("[\\\\/]file[a-zA-Z0-9]+", "/<tempfile>", x)
  x <- sub("[A-Z]:.*Rtmp[a-zA-Z0-9]+[\\\\/]", "<tempdir>/", x)
  x
}

transform_show_cursor <- function(x) {
  gsub("\033[?25h", "", x, fixed = TRUE)
}

transform_no_links <- function(x) {
  cli::ansi_strip(x, sgr = FALSE, csi = FALSE, link = TRUE)
}

transform_installed_in_temp <- function(x) {
  m <- regexpr("installed::.*$", x)
  regmatches(x, m) <- paste0("installed::.../", long_basename(regmatches(x, m)))
  x
}

fake_git <- local({
  dir.create(tmp <- tempfile())
  untar(testthat::test_path("fixtures/git-repo.tar.gz"), exdir = tmp)
  app <- asNamespace("pkgdepends")$git_app(file.path(tmp, "repo"))
  webfakes::local_app_process(app)
})

test_that("resolve_remote", {
  setup_fake_apps()

  cols <- c(
    "package",
    "version",
    "needscompilation",
    "repodir",
    "rversion",
    "platform",
    "priority",
    "ref",
    "type",
    "status",
    "target",
    "mirror",
    "sources",
    "sysreqs",
    "deps",
    "built",
    "repotype",
    "direct",
    "params",
    "metadata"
  )

  conf <- current_config()
  cache <- list(
    package = NULL,
    metadata = pkgcache::get_cranlike_metadata_cache()
  )

  # CRAN package as standard package --------------------------------------

  res <- suppressMessages(synchronise(resolve_remote_standard(
    parse_pkg_ref("crayon"),
    TRUE,
    conf,
    cache,
    dependencies = FALSE
  )))

  expect_snapshot(
    as.list(res[, intersect(names(res), cols)]),
    transform = transform_local_port,
    variant = paste0("pillar-", packageVersion("pillar"))
  )

  # Bioc package as standard package --------------------------------------

  res <- suppressMessages(synchronise(resolve_remote_standard(
    parse_pkg_ref("Biobase"),
    TRUE,
    conf,
    cache,
    dependencies = FALSE
  )))

  expect_snapshot(
    as.list(res[, intersect(names(res), cols)]),
    transform = function(x) transform_local_port(transform_bioc_version(x)),
    variant = if (getRversion() < "3.6.0") "old-r" else "new-r"
  )

  # Proper error for non-existing package ---------------------------------

  nonpkg <- "thispackagecannotexistforsure"
  res <- suppressMessages(synchronise(resolve_remote_standard(
    parse_pkg_ref(nonpkg),
    TRUE,
    conf,
    cache,
    dependencies = FALSE
  )))

  expect_snapshot(
    as.list(res[, intersect(names(res), cols)]),
    transform = transform_local_port,
    variant = paste0("pillar-", packageVersion("pillar"))
  )
})

test_that("download_remote", {
  setup_fake_apps()

  dir.create(tmp <- tempfile())
  dir.create(tmp2 <- tempfile())
  on.exit(unlink(c(tmp, tmp2), recursive = TRUE), add = TRUE)

  conf <- current_config()
  conf$platforms <- "source"
  conf$cache_dir <- tmp
  conf$package_cache_dir <- tmp2
  cache <- list(
    package = pkgcache::package_cache$new(conf$package_cache_dir),
    metadata = pkgcache::get_cranlike_metadata_cache()
  )

  # CRAN package ----------------------------------------------------------

  res <- suppressMessages(synchronise(resolve_remote_standard(
    parse_pkg_ref("crayon"),
    TRUE,
    conf,
    cache,
    dependencies = FALSE
  )))

  target <- file.path(conf$cache_dir, res$target[1])
  tree <- paste0(target, "-t")
  dl <- synchronise(download_remote_standard(
    res[1, ],
    target,
    tree,
    conf,
    cache,
    on_progress = NULL
  ))

  expect_equal(dl, "Got")
  expect_true(file.exists(target))
  unlink(target)

  # bioc package ----------------------------------------------------------

  res <- suppressMessages(synchronise(resolve_remote_standard(
    parse_pkg_ref("Biobase"),
    TRUE,
    conf,
    cache,
    dependencies = FALSE
  )))

  target <- file.path(conf$cache_dir, res$target[1])
  tree <- paste0(target, "-t")
  dl <- synchronise(download_remote_standard(
    res[1, ],
    target,
    tree,
    conf,
    cache,
    on_progress = NULL
  ))

  expect_equal(dl, "Got")
  expect_true(file.exists(target))
  unlink(target)

  # caching ---------------------------------------------------------------

  dl2 <- synchronise(download_remote_standard(
    res[1, ],
    target,
    tree,
    conf,
    cache,
    on_progress = NULL
  ))
  expect_true(dl2 %in% c("Had", "Current"))
  expect_true(file.exists(target))
})

test_that("download cran-like repos", {
  setup_fake_apps()

  x_app_pkgs <- dcf(
    "
    Package: foobar
    Version: 1.0.0
  "
  )

  fake_cranlike <- webfakes::local_app_process(cran_app(x_app_pkgs))
  withr::local_options(
    repos = c(X = fake_cranlike$url(), getOption("repos"))
  )

  dir.create(tmp <- tempfile())
  dir.create(tmp2 <- tempfile())
  on.exit(unlink(c(tmp, tmp2), recursive = TRUE), add = TRUE)

  conf <- current_config()
  conf$platforms <- "source"
  conf$cache_dir <- tmp
  conf$package_cache_dir <- tmp2
  cache <- list(
    package = pkgcache::package_cache$new(conf$package_cache_dir),
    metadata = pkgcache::get_cranlike_metadata_cache()
  )

  res <- suppressMessages(synchronise(resolve_remote_standard(
    parse_pkg_ref("foobar"),
    TRUE,
    conf,
    cache,
    dependencies = FALSE
  )))

  target <- file.path(conf$cache_dir, res$target[1])
  tree <- paste0(target, "-t")
  dl <- synchronise(download_remote_standard(
    res[1, ],
    target,
    tree,
    conf,
    cache,
    on_progress = NULL
  ))

  expect_equal(dl, "Got")
  expect_true(file.exists(target))
  unlink(target)

  dl2 <- synchronise(download_remote_standard(
    res[1, ],
    target,
    tree,
    conf,
    cache,
    on_progress = NULL
  ))
  expect_true(dl2 %in% c("Had", "Current"))
  expect_true(file.exists(target))
})

test_that("satisfy_remote", {
  res <- make_fake_resolution(`crayon@>=1.0.0` = list(type = "standard"))

  ## Package names differ
  bad1 <- make_fake_resolution(`crayon2` = list())
  expect_false(ans <- satisfy_remote_standard(res, bad1))
  expect_match(attr(ans, "reason"), "Package names differ")

  ## Insufficient version
  bad2 <- make_fake_resolution(`crayon` = list(version = "0.0.1"))
  expect_false(ans <- satisfy_remote_standard(res, bad2))
  expect_match(attr(ans, "reason"), "Insufficient version")

  ## Version is OK
  ok1 <- make_fake_resolution(`local::foobar` = list(package = "crayon"))
  expect_true(satisfy_remote_standard(res, ok1))

  ## No version req
  res <- make_fake_resolution(`crayon` = list(type = "standard"))
  ok2 <- make_fake_resolution(
    `local::foobar` = list(
      package = "crayon",
      version = "0.0.1"
    )
  )
  expect_true(satisfy_remote_standard(res, ok2))

  ## Direct package must be from CRAN(like), must update
  res <- make_fake_resolution(
    `crayon` = list(
      type = "standard",
      package = "crayon",
      direct = TRUE,
      version = "1.1.1"
    )
  )

  bad3 <- make_fake_resolution(
    `installed::crayon` = list(
      type = "installed",
      package = "crayon",
      extra = list(list(repotype = NA_character_, remotetype = NA_character_))
    )
  )
  expect_false(satisfy_remote_standard(res, bad3))

  bad4 <- make_fake_resolution(
    `installed::crayon` = list(
      type = "installed",
      package = "crayon",
      extra = list(list())
    )
  )
  expect_false(satisfy_remote_standard(res, bad4))

  bad5 <- make_fake_resolution(
    `installed::crayon` = list(
      type = "installed",
      package = "crayon",
      extra = list(list(repotype = "cran", remotetype = "standard")),
      version = "1.0.0"
    )
  )
  expect_false(satisfy_remote_standard(res, bad5))

  bad6 <- make_fake_resolution(
    `installed::crayon` = list(
      type = "local",
      package = "crayon",
      extra = list(list(repotype = "cran", remotetype = "standard")),
      version = "1.0.0"
    )
  )
  expect_false(satisfy_remote_standard(res, bad6))

  ok3 <- make_fake_resolution(
    `installed::crayon` = list(
      type = "installed",
      package = "crayon",
      extra = list(list(repotype = "cran", remotetype = "standard")),
      version = "1.2.0"
    )
  )
  expect_true(satisfy_remote_standard(res, ok3))
})

test_that("installedok", {
  # CRAN binary
  sol <- list(
    repotype = "cran",
    package = "package",
    version = "1.0.0",
    platform = "aarch64-apple-darwin20",
    built = "R 4.2.0; aarch64-apple-darwin20; 2022-10-08 08:37:45 UTC; unix"
  )

  expect_true(installedok_remote_standard(
    list(
      repotype = "cran",
      package = "package",
      version = "1.0.0",
      platform = "aarch64-apple-darwin20",
      built = "R 4.2.0; aarch64-apple-darwin20; 2022-10-08 08:37:45 UTC; unix"
    ),
    sol
  ))
  expect_false(installedok_remote_standard(
    list(
      package = "package",
      version = "1.0.0",
      platform = "aarch64-apple-darwin20",
      built = "R 4.2.0; aarch64-apple-darwin20; 2022-11-08 08:37:45 UTC; unix"
    ),
    sol
  ))

  # CRAN source
  sol <- list(
    repotype = "cran",
    package = "package",
    version = "1.0.0",
    platform = "source"
  )
  expect_true(installedok_remote_standard(
    list(
      package = "package",
      version = "1.0.0",
      repository = "CRAN"
    ),
    sol
  ))
  expect_false(installedok_remote_standard(
    list(
      package = "package",
      version = "1.0.0"
    ),
    sol
  ))

  # Bioc binary
  sol <- list(
    repotype = "bioc",
    package = "package",
    version = "1.0.0",
    platform = "aarch64-apple-darwin20",
    built = "R 4.2.0; aarch64-apple-darwin20; 2022-10-08 08:37:45 UTC; unix"
  )
  expect_true(installedok_remote_standard(
    list(
      repotype = "cran",
      package = "package",
      version = "1.0.0",
      platform = "aarch64-apple-darwin20",
      built = "R 4.2.0; aarch64-apple-darwin20; 2022-10-08 08:37:45 UTC; unix"
    ),
    sol
  ))
  expect_false(installedok_remote_standard(
    list(
      package = "package",
      version = "1.0.0",
      platform = "aarch64-apple-darwin20",
      built = "R 4.2.0; aarch64-apple-darwin20; 2022-11-08 08:37:45 UTC; unix"
    ),
    sol
  ))

  # Bioc source
  sol <- list(
    repotype = "bioc",
    package = "package",
    version = "1.0.0",
    platform = "source"
  )
  expect_true(installedok_remote_standard(
    list(
      package = "package",
      version = "1.0.0",
      repotype = "bioc"
    ),
    sol
  ))
  expect_false(installedok_remote_standard(
    list(
      package = "package",
      version = "1.0.0"
    ),
    sol
  ))

  # Other binary
  sol <- list(
    repotype = NA_character_,
    package = "package",
    version = "1.0.0",
    platform = "*",
    built = "R 4.2.0; aarch64-apple-darwin20; 2022-10-08 08:37:45 UTC; unix",
    metadata = list(list(RemoteRepos = "foobar"))
  )
  expect_true(installedok_remote_standard(
    list(
      package = "package",
      version = "1.0.0",
      repotype = "bioc",
      remoterepos = "foobar",
      platform = "*",
      built = "R 4.2.0; aarch64-apple-darwin20; 2022-10-08 08:37:45 UTC; unix"
    ),
    sol
  ))
  expect_false(installedok_remote_standard(
    list(
      package = "package",
      version = "1.0.0"
    ),
    sol
  ))

  # Other source
  sol <- list(
    repotype = NA_character_,
    package = "package",
    version = "1.0.0",
    platform = "source",
    metadata = list(list(RemoteRepos = "foobar"))
  )
  expect_true(installedok_remote_standard(
    list(
      package = "package",
      version = "1.0.0",
      repotype = "bioc",
      remoterepos = "foobar"
    ),
    sol
  ))
  expect_false(installedok_remote_standard(
    list(
      package = "package",
      version = "1.0.0"
    ),
    sol
  ))
})

test_that("installedok Linux binaries", {
  solution <- list(
    package = "package",
    version = "1.0.0",
    platform = "x86_64-pc-linux-gnu-ubuntu-22.04"
  )
  installed <- list(
    package = "package",
    version = "1.0.0",
    platform = "x86_64-pc-linux-gnu",
    built = "R 4.3.1; x86_64-pc-linux-gnu; 2023-09-21 14:53:54 UTC; unix"
  )

  expect_true(installedok_remote_standard(installed, solution))

  solution2 <- list(
    package = "package",
    version = "1.0.0",
    platform = "x86_64-pc-linux-gnu-ubuntu-22.04"
  )
  installed2 <- list(
    package = "package",
    version = "1.0.0",
    remotepkgplatform = "x86_64-pc-linux-gnu-ubuntu-22.04",
    platform = "x86_64-pc-linux-gnu"
  )

  expect_true(installedok_remote_standard(installed2, solution2))
})

test_that("different versions required", {
  setup_fake_gh_app()
  setup_fake_apps()

  lib <- withr::local_tempdir()
  p <- new_pkg_installation_proposal(
    c("r-lib/crayon", "crayon"),
    config = list(library = lib)
  )

  suppressMessages(p$solve())
  expect_snapshot(
    error = TRUE,
    p$stop_for_solution_error(),
    # need to sort this, otherwise it is not deterministic
    transform = function(x) {
      x[-1] <- sort(x[-1], decreasing = TRUE)
      x
    }
  )
})

test_that("direct CRAN conflicts with downstream GH dep", {
  gh <- list(
    users = list(
      "r-lib" = list(
        repos = list(
          foo = list(
            commits = list(
              list(
                sha = "b69f0b34da3b538a666c02944d67ebae9535488fa9d29a010a79980822b56780",
                branch = "main",
                tag = "HEAD",
                files = list(
                  DESCRIPTION = "Package: foo\nVersion: 1.0.0\nImports: crayon\nRemotes: r-lib/crayon\n",
                  NAMESPACE = ""
                )
              )
            )
          ),
          crayon = list(
            commits = list(
              list(
                sha = "09ec8a1bbab45ba6a54f896f721fdf00798559ef71852063ea4934d9656101d0",
                branch = "main",
                tag = "HEAD",
                files = list(
                  DESCRIPTION = gh_app_desc("crayon"),
                  NAMESPACE = ""
                )
              )
            )
          )
        )
      )
    )
  )

  fake_gh <- webfakes::local_app_process(
    gh_app(gh)
  )
  withr::local_envvar(R_PKG_GITHUB_API_URL = fake_gh$url())
  setup_fake_apps()

  lib <- withr::local_tempdir()
  p <- new_pkg_installation_proposal(
    c("r-lib/foo", "crayon"),
    config = list(library = lib)
  )

  suppressMessages(p$solve())
  expect_snapshot(
    error = TRUE,
    p$stop_for_solution_error()
  )
})

test_that("no required version", {
  repo <- dcf(
    "
    Package: pkg1
    Depends: pkg2 (>= 2.0.0)

    Package: pkg2
    Version: 1.0.0
  "
  )
  setup_fake_apps(cran_repo = repo)

  lib <- withr::local_tempdir()
  p <- new_pkg_installation_proposal("pkg1", config = list(library = lib))

  suppressMessages(p$solve())
  expect_snapshot(error = TRUE, p$stop_for_solution_error())
})

test_that("failed resolution", {
  setup_fake_apps()
  setup_fake_gh_app()

  lib <- withr::local_tempdir()
  p1 <- new_pkg_installation_proposal("SDF", config = list(library = lib))
  suppressMessages(p1$solve())
  expect_snapshot(error = TRUE, p1$stop_for_solution_error())

  p2 <- new_pkg_installation_proposal("SDF/SDF", config = list(library = lib))
  suppressMessages(p2$solve())
  expect_snapshot(error = TRUE, p2$stop_for_solution_error())
})

test_that("ruled out direct dep", {
  # I.e. there are multiple versions, the newer version is ruled out
  # for the current R version, but the older vresion is fine.

  # pkgcache::cran_app cannot seem to do this with a single repo
  repo1 <- dcf(
    "
    Package: pkg1
    Version: 1.0.0
  "
  )

  repo2 <- dcf(
    "
    Package: pkg1
    Version: 1.0.1
    Depends: R (>= 20000.0.0)
  "
  )

  fake1 <- webfakes::new_app_process(cran_app(repo1))
  fake2 <- webfakes::new_app_process(cran_app(repo2))
  withr::local_options(repos = c(CRAN = fake1$url(), X = fake2$url()))

  lib <- tempfile()
  lock <- tempfile()
  on.exit(unlink(c(lib, lock), recursive = TRUE), add = TRUE)

  p <- suppressMessages(new_pkg_installation_proposal(
    c("pkg1"),
    config = list(
      dependencies = TRUE,
      library = lib
    )
  ))
  suppressMessages(p$solve())
  expect_equal(p$get_solution()$data$version, "1.0.0")
})

#' @param lockfile Path to the lock file to use.
#' @param config Configuration options, a named list. See
#'  ['Configuration'][pkgdepends-config]. If it does not include `library`, then
#'  `.libPaths()[1]` is added as `library`.
#' @param ... Additional arguments, passed to
#'   [`pkg_installation_plan$new()`](#method-new).
#' @return `new_pkg_installation_plan()` returns a `pkg_installation_plan`
#'   object.
#'
#' @export
#' @rdname pkg_installation_plan

new_pkg_installation_plan <- function(
  lockfile = "pkg.lock",
  config = list(),
  ...
) {
  config$library <- config$library %||% .libPaths()[[1]]
  pkg_installation_plan$new(lockfile, config = config, ...)
}

#' R6 class for installation from a lock file
#'
#' @description
#' An installation plan is similar to an installation proposal
#' (i.e. [pkg_installation_proposal]), but it already contains the solved
#' dependencies, complete with download URLs.
#'
#' @details
#' Typically you create a `pkg_installation_plan` object with
#' `new_pkg_installation_plan()` and then call its `$download()` method
#' to download the packages and then its `$install()` method to install
#' them.
#'
#' @export

pkg_installation_plan <- R6::R6Class(
  "pkg_installation_plan",
  inherit = pkg_installation_proposal,
  public = list(
    #' @description
    #' Create a new `pkg_installation_plan` object. Consider using
    #' `new_pkg_installation_plan()` instead of calling the constructor
    #' directly.
    #'
    #' The returned object can be used to download and install
    #' packages, according to the plan.
    #'
    #' @param lockfile Path to the lock file to use.
    #' @param config Configuration options. See
    #'   ['Configuration'][pkgdepends-config]. It needs to include the package
    #'   library to install to, in `library`.
    #' @param remote_types Custom remote ref types, this is for advanced
    #'   use, and experimental currently.

    initialize = function(
      lockfile = "pkg.lock",
      config = list(),
      remote_types = NULL
    ) {
      assert_that(is_character(config$library))
      private$library <- config$library
      config$goal <- "install"
      private$plan <- pkg_plan$new(
        lockfile = lockfile,
        config = config,
        library = config$library,
        remote_types = remote_types
      )
    },

    #' @description
    #' This function is implemented for installation plans, and will error.

    resolve = function() {
      throw(pkg_error(
        "Cannot resolve an installation plan, it is already resolved."
      ))
    },

    #' @description
    #' This function is implemented for installation plans, and will error.

    async_resolve = function() {
      throw(pkg_error(
        "Cannot resolve an installation plan, it is already resolved."
      ))
    },

    #' @description
    #' Installation plans are already solved, and this method will return
    #' `NA_character_`, always.

    get_solve_policy = function() NA_character_,

    #' @description
    #' This function is implemented for installation plans, and will error.

    set_solve_policy = function() {
      throw(pkg_error(
        "Cannot solve an installation plan, it is already solved."
      ))
    },

    #' @description
    #' This function is implemented for installation plans, and will error.

    solve = function() {
      throw(pkg_error(
        "Cannot solve an installation plan, it is already solved."
      ))
    },

    #' @description
    #' Update the plan to the current state of the library. If the library
    #' has not changed since the plan was created, then it does nothing.
    #' If new packages have been installed, then it might not be necessary
    #' to download and install all packages in the plan.
    #'
    #' @details
    #' This operation is different than creating a new proposal with the
    #' updated library, because it uses the the packages and package
    #' versions of the original plan. E.g. if the library has a newer
    #' version of a package, then `$update()` will downgrade it to the
    #' version in the plan.

    update = function() pkg_lockfile_update(self, private),

    #' @description
    #' Update information about installed and missing system requirements.

    update_sysreqs = function() pkg_lockfile_update_sysreqs(self, private),

    #' @description
    #' Format a `pkg_installation_plan` object, typically for printing.
    #'
    #' @param ... not used currently.
    #'
    #' @return
    #' A character vector, each element should be a line in the printout.

    format = function(...) {
      refs <- private$plan$get_refs()

      has_dls <- private$plan$has_solution_downloads()
      dls <- if (has_dls) private$plan$get_solution_download()
      dls_err <- has_dls && any(dls$status == "Failed")

      has_sys <- !is.null(private$plan$get_solution()$sysreqs)

      c(
        "<pkg_installation_plan>",
        "+ refs:",
        paste0("  - ", refs),
        "+ has solution",
        if (has_dls) "+ has downloads",
        if (dls_err) "x has download errors",
        "(use `$update()` to update the plan for an updated library)",
        "(use `$show_solution()` to see the packages to install",
        "(use `$get_solution()` to see the full solution results)",
        "(use `$draw()` to draw the dependency tree)",
        if (!has_dls) "(use `$download()` to download packages)",
        if (has_dls) "(use `$get_downloads()` to get download data)",
        if (has_dls) "(use `$get_install_plan()` to get the installation plan)",
        if (has_sys) {
          "(use `$install_sysreqs()` to install system requirements)"
        },
        if (has_dls) "(use `$install()` to install the packages)"
      )
    }
  ),

  private = list(
    plan = NULL,
    library = NULL
  )
)

pkg_lockfile_update <- function(self, private) {
  private$plan$update()
  invisible(self)
}

pkg_lockfile_update_sysreqs <- function(self, private) {
  private$plan$update_sysreqs()
  invisible(self)
}

parse_remote_param <- function(specs, config, ...) {
  parsed_specs <- re_match(specs, param_rx())
  parsed_specs$ref <- "*"
  cn <- setdiff(colnames(parsed_specs), c(".match", ".text"))
  parsed_specs <- parsed_specs[, cn]
  parsed_specs$type <- "param"
  lapply(
    seq_len(nrow(parsed_specs)),
    function(i) as.list(parsed_specs[i, ])
  )
}

resolve_remote_param <- function(
  remote,
  direct,
  config,
  cache,
  dependencies,
  ...
) {
  ## Handled specially by the resolver
  NULL
}

download_remote_param <- function(
  resolution,
  target,
  target_tree,
  config,
  cache,
  which,
  on_progress
) {
  ## this is never actually called
  stop("Internal error")
}

satisfy_remote_param <- function(resolution, candidate, config, ...) {
  ## this is never actually called
  stop("Internal error")
}

installedok_remote_param <- function(installed, solution, config, ...) {
  ## this is never actually called
  stop("Internal error")
}

#' The dependency solver
#'
#' The dependency solver takes the resolution information, and works out
#' the exact versions of each package that must be installed, such that
#' version and other requirements are satisfied.
#'
#' ## Solution policies
#'
#' The dependency solver currently supports two policies: `lazy` and
#' `upgrade`. The `lazy` policy prefers to minimize installation time,
#' and it does not perform package upgrades, unless version requirements
#' require them. The `upgrade` policy prefers to update all package to
#' their latest possible versions, but it still considers that version
#' requirements.
#'
#' ## The integer problem
#'
#' Solving the package dependencies requires solving an integer linear
#' problem (ILP). This subsection briefly describes how the problem is
#' represented as an integer problem, and what the solution policies
#' exactly mean.
#'
#' Every row of the package resolution is a candidate for the dependency
#' solver. In the integer problem, every candidate corresponds to a binary
#' variable. This is 1 if that candidate is selected as part of the
#' solution, and 0 otherwise.
#'
#' The objective of the ILP minimization is defined differently for
#' different solution policies. The ILP conditions are the same.
#'
#' 1. For the `lazy` policy, `installed::` packaged get 0 points, binary
#'    packages 1 point, sources packages 5 points.
#' 2. For the 'upgrade' policy, we rank all candidates for a given package
#'    according to their version numbers, and assign more points to older
#'    versions. Points are assigned by 100 and candidates with equal
#'    versions get equal points. We still prefer installed packages to
#'    binaries to source packages, so also add 0 point for already
#'    installed candidates, 1 extra points for binaries and 5 points for
#'    source packages.
#' 3. For directly specified refs, we aim to install each package exactly
#'    once. So for these we require that the variables corresponding to
#'    the same package sum up to 1.
#' 4. For non-direct refs (i.e. dependencies), we require that the
#'    variables corresponding to the same package sum up to at most one.
#'    Since every candidate has at least 1 point in the objective function
#'    of the minimization problem, non-needed dependencies will be
#'    omitted.
#' 5. For direct refs, we require that their candidates satisfy their
#'    references. What this means exactly depends on the ref types. E.g.
#'    for CRAN packages, it means that a CRAN candidate must be selected.
#'    For a standard ref, a GitHub candidate is OK as well.
#' 6. We rule out candidates for which the dependency resolution failed.
#' 7. We go over all the dependency requirements and rule out packages
#'    that do not meet them. For every package `A`, that requires
#'    package `B`, we select the `B(i, i=1..k)` candidates of `B` that
#'    satisfy `A`'s requirements and add a `A - B(1) - ... - B(k) <= 0`
#'    rule. To satisfy this rule, either we cannot install `A`, or if `A`
#'    is installed, then one of the good `B` candidates must be installed
#'    as well.
#' 8. We rule out non-installed CRAN and Bioconductor candidates for
#'    packages that have an already installed candidate with the same exact
#'    version.
#' 9. We also rule out source CRAN and Bioconductor candidates for
#'    packages that have a binary candidate with the same exact version.
#'
#' ## Explaining why the solver failed
#'
#' To be able to explain why a solution attempt failed, we also add a dummy
#' variable for each directly required package. This dummy variable has a
#' very large objective value, and it is only selected if there is no
#' way to install the directly required package.
#'
#' After a failed solution, we look the dummy variables that were selected,
#' to see which directly required package failed to solve. Then we check
#' which rule(s) ruled out the installation of these packages, and their
#' dependencies, recursively.
#'
#' ## The result
#'
#' The result of the solution is a `pkg_solution_result` object. It is a
#' named list with entries:
#'
#' * `status`: Status of the solution attempt, `"OK"` or `"FAILED"`.
#' * `data`: The selected candidates. This is very similar to a
#'   [pkg_resolution_result] object, but it has two extra columns:
#'     * `lib_status`: status of the package in the library, after the
#'        installation. Possible values: `new` (will be newly installed),
#'        `current` (up to date, not installed), `update` (will be updated),
#'        `no-update` (could update, but will not).
#'     * `old_version`: The old (current) version of the package in the
#'        library, or `NA` if the package is currently not installed.
#' * `problem`: The ILP problem. The exact representation is an
#'   implementation detail, but it does have an informative print method.
#' * `solution`: The return value of the internal solver.
#'
#' @name pkg_solution
#' @aliases pkg_solution_result
NULL

solve_dummy_obj <- 1000000000

pkgplan_solve <- function(self, private, policy) {
  "!DEBUG starting to solve `length(private$resolution$packages)` packages"
  if (is.null(private$config$get("library"))) {
    throw(pkg_error(
      # nocov start
      "No package library specified for installation plan.",
      i = "Maybe you need to specify {.code config = list(library = ...)}
       in {.code pkg_installation_plan$new()} or another initializer?"
    )) # nocov end
  }
  if (is.null(private$resolution)) {
    self$resolve()
  }
  if (private$dirty) {
    throw(pkg_error(
      # nocov start
      "Package list has changed, you need to call the {.code $resolve()}
       method again?"
    )) # nocov end
  }

  metadata <- list(solution_start = Sys.time())
  pkgs <- self$get_resolution()
  rversion <- private$config$get("r_versions")

  prb <- private$create_lp_problem(pkgs, policy)
  sol <- private$solve_lp_problem(prb)

  if (sol$status != 0) {
    throw(pkg_error(
      # nocov start
      "Error in dependency solver, cannot solve installation.",
      i = "Solver status: {sol$status}.",
      i = msg_internal_error()
    )) # nocov end
  }

  selected <- as.logical(sol$solution[seq_len(nrow(pkgs))])
  res <- list(
    status = if (sol$objval < solve_dummy_obj - 1) "OK" else "FAILED",
    data = private$subset_resolution(selected),
    problem = prb,
    solution = sol
  )

  lib_status <- calculate_lib_status(res$data, pkgs)
  res$data <- as_data_frame(cbind(res$data, lib_status))
  res$data$cache_status <-
    calculate_cache_status(res$data, private$cache)

  metadata$solution_end <- Sys.time()
  attr(res, "metadata") <- modifyList(attr(pkgs, "metadata"), metadata)
  class(res) <- unique(c("pkg_solution_result", class(res)))

  if (res$status == "FAILED") {
    res$failures <- describe_solution_error(pkgs, res)
  }

  private$solution$result <- res
  self$get_solution()
}

pkgplan_stop_for_solve_error <- function(self, private) {
  if (is.null(private$solution)) {
    throw(pkg_error(
      # nocov start
      "You need to call the {.code $solve()} method first."
    )) # nocov end
  }

  sol <- self$get_solution()

  if (sol$status != "OK") {
    msg <- paste(format(sol$failures), collapse = "\n")
    throw(new_error(
      "Could not solve package dependencies:\n",
      msg,
      call. = FALSE
    ))
  }
}

pkgplan__create_lp_problem <- function(self, private, pkgs, policy) {
  pkgplan_i_create_lp_problem(pkgs, private$config, policy)
}

## Add a condition, for a subset of variables, with op and rhs
pkgplan_i_lp_add_cond <- function(
  lp,
  vars,
  op = "<=",
  rhs = 1,
  coef = rep(1, length(vars)),
  type = NA_character_,
  note = NULL
) {
  lp$conds[[length(lp$conds) + 1]] <-
    list(vars = vars, coef = coef, op = op, rhs = rhs, type = type, note = note)
  lp
}

## This is a separate function to make it testable without a `remotes`
## object.
##
## Variables:
## * 1:num are candidates
## * (num+1):(num+num_direct_pkgs) are the relax variables for direct refs

pkgplan_i_create_lp_problem <- function(pkgs, config, policy) {
  "!DEBUG creating LP problem"

  ## TODO: we could already rule out (standard) source packages if binary
  ## with the same version is present

  ## TODO: we could already rule out (standard) source and binary packages
  ## if an installed ref with the same version is present

  rversion <- config$get("r_versions")

  lp <- pkgplan_i_lp_init(pkgs, config, policy)
  lp <- pkgplan_i_lp_objectives(lp)
  lp <- pkgplan_i_lp_os_type(config, lp)
  lp <- pkgplan_i_lp_force_source(lp)
  lp <- pkgplan_i_lp_failures(lp)
  lp <- pkgplan_i_lp_ignore(lp)
  lp <- pkgplan_i_lp_platforms(lp)
  lp <- pkgplan_i_lp_no_multiples(lp)
  lp <- pkgplan_i_lp_rversion(lp, rversion)
  lp <- pkgplan_i_lp_satisfy_direct(lp)
  lp <- pkgplan_i_lp_latest_direct(lp)
  lp <- pkgplan_i_lp_latest_within_repo(lp)
  lp <- pkgplan_i_lp_prefer_installed(lp)
  lp <- pkgplan_i_lp_deduplicate(lp, config)
  lp <- pkgplan_i_lp_prefer_binaries(lp)
  lp <- pkgplan_i_lp_prefer_new_binaries(lp)
  lp <- pkgplan_i_lp_dependencies(lp, config)

  lp
}

pkgplan_i_lp_init <- function(pkgs, config, policy) {
  num_candidates <- nrow(pkgs)
  packages <- unique(pkgs$package)
  direct_packages <- unique(pkgs$package[pkgs$direct])
  indirect_packages <- setdiff(packages, direct_packages)
  num_direct <- length(direct_packages)

  structure(
    list(
      ## Configuration
      config = config,
      ## Number of package candidates
      num_candidates = num_candidates,
      ## Number of directly specified ones
      num_direct = num_direct,
      ## Total number of variables. For direct ones, we have an extra variable
      total = num_candidates + num_direct,
      ## Constraints to fill in
      conds = list(),
      pkgs = pkgs,
      policy = policy,
      ## All package names
      packages = packages,
      ## The names of the direct packages
      direct_packages = direct_packages,
      ## The names of the indirect packages
      indirect_packages = indirect_packages,
      ## Candidates (indices) that have been ruled out. E.g. resolution failed
      ruled_out = integer()
    ),
    class = "pkgplan_lp_problem"
  )
}

## Coefficients of the objective function, this is very easy
## TODO: use rversion as well, for installed and binary packages

pkgplan_i_lp_objectives <- function(lp) {
  pkgs <- lp$pkgs
  policy <- lp$policy
  num_candidates <- lp$num_candidates

  if (policy == "lazy") {
    ## Simple: installed < binary < source
    lp$obj <- ifelse(
      pkgs$type == "installed",
      0,
      ifelse(pkgs$platform == "source", 5, 1)
    )
  } else if (policy == "upgrade") {
    ## Sort the candidates of a package according to version number
    lp$obj <- rep((num_candidates + 1) * 100, num_candidates)
    whpp <- pkgs$status == "OK" & !is.na(pkgs$version)
    pn <- unique(pkgs$package[whpp])
    for (p in pn) {
      whp <- whpp & pkgs$package == p
      v <- pkgs$version[whp]
      r <- rank(package_version(v), ties.method = "min")
      lp$obj[whp] <- (max(r) - r + 1) * 100
      lp$obj[whp] <- lp$obj[whp] - min(lp$obj[whp])
    }
    lp$obj <- lp$obj +
      ifelse(
        pkgs$type == "installed",
        1,
        ifelse(pkgs$platform == "source", 3, 2)
      )
    lp$obj <- lp$obj - min(lp$obj)
  } else {
    throw(pkg_error(
      # nocov start
      "Unknown version selection policy: {.val {policy}}.",
      i = "It has to be one of {.val lazy} or {.val upgrade}."
    )) # nocov end
  }

  lp$obj <- c(lp$obj, rep(solve_dummy_obj, lp$num_direct))

  lp
}

pkgplan_i_lp_os_type <- function(config, lp) {
  if (config$get("goal") != "install") {
    return(lp)
  }
  if (!"os_type" %in% names(lp$pkgs)) {
    return(lp)
  }
  os <- os_type()
  bad <- which(!is.na(lp$pkgs$os_type) & lp$pkgs$os_type != os)
  for (wh in bad) {
    lp <- pkgplan_i_lp_add_cond(
      lp,
      wh,
      op = "==",
      rhs = 0,
      type = "matching-platform"
    )
  }
  lp$ruled_out <- c(lp$ruled_out, bad)

  lp
}

pkgplan_i_lp_force_source <- function(lp) {
  # if source package is forced, then rule out binaries
  src_req <- vlapply(lp$pkgs$params, is_true_param, "source")
  not_src <- lp$pkgs$platform != "source"
  bad <- which(src_req & not_src)
  for (wh in bad) {
    lp <- pkgplan_i_lp_add_cond(
      lp,
      wh,
      op = "==",
      rhs = 0,
      type = "source-required"
    )
  }
  lp$ruled_out <- c(lp$ruled_out, bad)

  lp
}

pkgplan_i_lp_failures <- function(lp) {
  ## 5. Can't install failed resolutions
  failedconds <- function(wh) {
    if (lp$pkgs$status[wh] != "FAILED") {
      return()
    }
    lp <<- pkgplan_i_lp_add_cond(
      lp,
      wh,
      op = "==",
      rhs = 0,
      type = "ok-resolution"
    )
    lp$ruled_out <<- c(lp$ruled_out, wh)
  }
  lapply(seq_len(lp$num_candidates), failedconds)

  lp
}

pkgplan_i_lp_ignore <- function(lp) {
  ignored <- which(vlapply(lp$pkgs$params, is_true_param, "ignore"))
  for (wh in ignored) {
    lp <- pkgplan_i_lp_add_cond(
      lp,
      wh,
      op = "==",
      rhs = 0,
      type = "ignored-by-user"
    )
  }
  lp$ruled_out <- c(lp$ruled_out, ignored)

  lp
}

pkgplan_i_lp_platforms <- function(lp) {
  ## check if platform is good
  badplatform <- function(wh) {
    if (lp$pkgs$type[wh] %in% c("deps", "param")) {
      return()
    }
    ok <- platform_is_ok(
      lp$pkgs$platform[wh],
      lp$config$get("platforms"),
      lp$config$get("windows_archs")
    )
    if (!ok) {
      lp <<- pkgplan_i_lp_add_cond(
        lp,
        wh,
        op = "==",
        rhs = 0,
        type = "matching-platform"
      )
      lp$ruled_out <<- c(lp$ruled_out, wh)
    }
  }
  lapply(seq_len(lp$num_candidates), badplatform)

  lp
}

pkgplan_i_lp_no_multiples <- function(lp) {
  ## 1. Each directly specified package exactly once.
  ##    (We also add a dummy variable to catch errors.)
  for (p in seq_along(lp$direct_packages)) {
    pkg <- lp$direct_packages[p]
    wh <- which(lp$pkgs$package == pkg)
    lp <- pkgplan_i_lp_add_cond(
      lp,
      c(wh, lp$num_candidates + p),
      op = "==",
      type = "exactly-once"
    )
  }

  ## 2. Each non-direct package must be installed at most once
  for (p in seq_along(lp$indirect_packages)) {
    pkg <- lp$indirect_packages[p]
    wh <- which(lp$pkgs$package == pkg)
    lp <- pkgplan_i_lp_add_cond(lp, wh, op = "<=", type = "at-most-once")
  }

  lp
}

pkgplan_i_lp_rversion <- function(lp, rversion) {
  rversion <- package_version(rversion)
  pkgs <- lp$pkgs
  num_candidates <- lp$num_candidates
  ruled_out <- lp$ruled_out
  base <- base_packages()

  depconds <- function(wh) {
    if (pkgs$status[wh] != "OK") {
      return()
    }
    deps <- pkgs$deps[[wh]]
    deps <- deps[deps$ref == "R", , drop = FALSE]
    if (nrow(deps) == 0) {
      return()
    }
    type <- NA
    for (idx in seq_len(nrow(deps))) {
      need <- deps$version[idx]
      needrver <- paste0(deps$op[[idx]], " ", need)
      switch(
        deps$op[[idx]],
        "<" = if (!rversion < need) type <- "new-rversion",
        "<=" = if (!rversion <= need) type <- "new-rversion",
        "==" = if (!rversion == need) type <- "different-rversion",
        ">=" = if (!rversion >= need) type <- "old-rversion",
        ">" = if (!rversion > need) type <- "old-rversion",
        warning(paste0("Ignoring R version requirement: ", needrver))
      )
      # Enough to have one to rule out
      if (!is.na(type)) break
    }
    if (!is.na(type)) {
      lp <<- pkgplan_i_lp_add_cond(
        lp,
        wh,
        op = "==",
        rhs = 0,
        type = type,
        note = needrver
      )
      lp$ruled_out <<- c(lp$ruled_out, wh)
    }
  }

  lapply(setdiff(seq_len(num_candidates), ruled_out), depconds)

  lp
}

pkgplan_i_lp_satisfy_direct <- function(lp) {
  ## 3. Direct refs must be satisfied
  satisfy <- function(wh) {
    pkgname <- lp$pkgs$package[[wh]]
    res <- lp$pkgs[wh, ]
    others <- setdiff(which(lp$pkgs$package == pkgname), wh)
    for (o in others) {
      res2 <- lp$pkgs[o, ]
      if (!isTRUE(satisfies_remote(res, res2))) {
        lp <<- pkgplan_i_lp_add_cond(
          lp,
          o,
          op = "==",
          rhs = 0,
          type = "satisfy-refs",
          note = wh
        )
      }
    }
  }
  direct <- setdiff(which(lp$pkgs$direct), lp$ruled_out)
  lapply(direct, satisfy)

  lp
}

## Order matters. By the time this is called, the failed resolutions
## are ruled out, and R version requirements are also checked.
## So we only work with the packages that are not ruled out, and select
## the latest version. This cannot be ruled out later, only because of
## version requirements, but then it is up to the user to solve this.

pkgplan_i_lp_latest_direct <- function(lp) {
  pkgs <- lp$pkgs
  # these have version requirements
  vreq <- vlapply(
    lp$pkgs$remote,
    function(r) is.list(r) && !is.null(r$version) && r$version != ""
  )
  dirpkgs <- unique(lp$pkgs$package[lp$pkgs$direct & !vreq])
  for (pkg in dirpkgs) {
    cand <- which(
      pkgs$package == pkg &
        pkgs$type %in% c("cran", "bioc", "standard")
    )
    cand <- setdiff(cand, lp$ruled_out)
    if (length(cand) == 0) {
      next
    }
    vers <- package_version(pkgs$version[cand])
    bad <- vers < max(vers)
    for (wh in cand[bad]) {
      lp <- pkgplan_i_lp_add_cond(
        lp,
        wh,
        op = "==",
        rhs = 0,
        type = "direct-update"
      )
    }
    lp$ruled_out <- c(lp$ruled_out, cand[bad])
  }

  lp
}

# CRAN's repo sometimes relies on selecting the latest version of
# a package, if multiple versions are available. (This is after considering
# R version requirements.) So we need to do the same, within repo.
# Otherwise pak/pkgdepends would select the first candidate, and while that
# always (?) OK for CRAN, the order is not the same in RSPM, apparently.

pkgplan_i_lp_latest_within_repo <- function(lp) {
  nbr <- seq_len(nrow(lp$pkgs))
  oid <- ifelse(nbr %in% lp$ruled_out, nbr, 0)
  key <- paste0(
    oid,
    "/",
    lp$pkgs$mirror,
    "/",
    lp$pkgs$repodir,
    "/",
    lp$pkgs$platform,
    "/",
    lp$pkgs$ref
  )
  dups <- unique(key[duplicated(key)])
  for (dupkey in dups) {
    cand <- which(key == dupkey)
    if (length(cand) == 0) {
      next
    }
    vers <- package_version(lp$pkgs$version[cand])
    bad <- vers < max(vers)
    for (wh in cand[bad]) {
      lp <- pkgplan_i_lp_add_cond(
        lp,
        wh,
        op = "==",
        rhs = 0,
        type = "choose-latest"
      )
    }
    lp$ruled_out <- c(lp$ruled_out, cand[bad])
  }

  lp
}

pkgplan_i_lp_prefer_installed <- function(lp) {
  pkgs <- lp$pkgs
  inst <- which(
    pkgs$type == "installed" & !seq_along(pkgs$type) %in% lp$ruled_out
  )
  for (i in inst) {
    ## If not a CRAN or BioC package, skip it
    repotype <- pkgs$extra[[i]]$repotype
    if (is.null(repotype) || !repotype %in% c("cran", "bioc")) {
      next
    }

    ## Look for others with cran/bioc/standard type and same name & ver
    package <- pkgs$package[i]
    version <- pkgs$version[i]

    ruledout <- which(
      pkgs$type %in%
        c("cran", "bioc", "standard") &
        pkgs$package == package &
        pkgs$version == version
    )

    lp$ruled_out <- c(lp$ruled_out, ruledout)
    for (r in ruledout) {
      lp <- pkgplan_i_lp_add_cond(
        lp,
        r,
        op = "==",
        rhs = 0,
        type = "prefer-installed"
      )
    }
  }

  lp
}

# We only do this for source packages, because we already prefer new
# binaries, via `pkgplan_i_lp_prefer_new_binaries()`.
#
# We do this before we prefer binaries, because that rule will rule out
# any source package version that has a binary, and we want to compare all
# source versions here.

pkgplan_i_lp_deduplicate <- function(lp, config) {
  pkgs <- lp$pkgs
  whpp <- pkgs$status == "OK" & !is.na(pkgs$version)
  pn <- unique(pkgs$package[whpp])
  ruled_out <- integer()
  for (p in pn) {
    whp <- which(
      whpp &
        pkgs$package == p &
        pkgs$platform == "source" &
        pkgs$type %in% c("cran", "bioc", "standard")
    )
    whp <- setdiff(whp, lp$ruled_out)
    if (length(whp) <= 1) {
      next
    }
    v <- package_version(pkgs$version[whp])
    mv <- max(v)
    best <- which(v == mv)[1]
    for (i in whp[-best]) {
      if (same_deps(pkgs$deps[[i]], pkgs$deps[[whp[best]]])) {
        ruled_out <- c(ruled_out, i)
      }
    }
  }

  for (r in ruled_out) {
    lp <- pkgplan_i_lp_add_cond(
      lp,
      r,
      op = "==",
      rhs = 0,
      type = "choose-latest"
    )
  }

  lp$ruled_out <- unique(c(lp$ruled_out, ruled_out))
  lp
}

same_deps <- function(d1, d2) {
  d1 <- d1[order(d1$package, d1$type), ]
  rownames(d1) <- NULL
  d2 <- d2[order(d2$package, d2$type), ]
  rownames(d2) <- NULL
  identical(d1, d2)
}

pkgplan_i_lp_prefer_binaries <- function(lp) {
  pkgs <- lp$pkgs
  str <- paste0(pkgs$type, "::", pkgs$package, "@", pkgs$version)
  for (ustr in unique(str)) {
    same <- which(ustr == str)
    ## We can't do this for other packages, because version is not
    ## exclusive for those
    if (!pkgs$type[same[1]] %in% c("cran", "bioc", "standard")) {
      next
    }

    ## TODO: choose the right one for the current R version
    selected <- setdiff(same[pkgs$platform[same] != "source"], lp$ruled_out)[1]
    ## No binary package, maybe there is RSPM. This is temporary,
    ## until we get proper RSPM support.
    ## It would be better to merge the download URLs in this case.
    if (is.na(selected)) {
      selected <- setdiff(
        same[grepl("__linux__", pkgs$mirror[same])],
        lp$ruled_out
      )[1]
    }
    ## Same on Windows, to work around
    ## https://github.com/r-lib/pkgdepends/issues/276
    ## It would be better to merge the download URLs in this case.
    if (is.na(selected)) {
      selected <- setdiff(
        same[grepl(
          "^https://packagemanager[.]rstudio[.]com",
          pkgs$mirror[same]
        )],
        lp$ruled_out
      )[1]
    }
    if (is.na(selected)) {
      next
    }
    ruledout <- setdiff(same, selected)
    lp$ruled_out <- c(lp$ruled_out, ruledout)
    for (r in ruledout) {
      lp <- pkgplan_i_lp_add_cond(
        lp,
        r,
        op = "==",
        rhs = 0,
        type = "prefer-binary"
      )
    }
  }

  lp
}

pkgplan_i_lp_prefer_new_binaries <- function(lp) {
  # We rule out older binaries if there is a new one available
  # This is not always correct, but otherwise the solver will be slow.
  # https://github.com/r-lib/pkgdepends/issues/276
  # I tried adding a penalty to older versions, but that did not work.
  pkgs <- lp$pkgs
  whpp <- pkgs$status == "OK" & !is.na(pkgs$version)
  # Skip packages directly referenced with a version pin. Those are
  # handled by pkgplan_i_lp_latest_direct, and ruling out older binaries
  # here would conflict with the satisfy-direct constraint.
  vreq <- vlapply(
    lp$pkgs$remote,
    function(r) is.list(r) && !is.null(r$version) && r$version != ""
  )
  pinned <- unique(lp$pkgs$package[lp$pkgs$direct & vreq])
  pn <- setdiff(unique(pkgs$package[whpp]), pinned)
  ruled_out <- integer()
  for (p in pn) {
    whp <- which(
      whpp &
        pkgs$package == p &
        pkgs$platform != "source" &
        pkgs$type %in% c("cran", "bioc", "standard")
    )
    whp <- setdiff(whp, lp$ruled_out)
    if (length(whp) == 0) {
      next
    }
    v <- package_version(pkgs$version[whp])
    ruled_out <- c(ruled_out, whp[v != max(v)])
  }

  for (r in ruled_out) {
    lp <- pkgplan_i_lp_add_cond(
      lp,
      r,
      op = "==",
      rhs = 0,
      type = "prefer-new-binary"
    )
  }

  lp$ruled_out <- unique(c(lp$ruled_out, ruled_out))
  lp
}

pkgplan_i_lp_dependencies <- function(lp, config) {
  pkgs <- lp$pkgs
  linkingto <- config$get("include_linkingto")
  num_candidates <- lp$num_candidates
  ruled_out <- lp$ruled_out
  base <- base_packages()

  ignored <- vlapply(pkgs$params, is_true_param, "ignore")
  ignore_rver <- vcapply(pkgs$params, get_param_value, "ignore-before-r")
  if (any(!is.na(ignore_rver))) {
    ignore_rver[is.na(ignore_rver)] <- "0.0.0"
    current <- min(lp$config$get("r_versions"))
    ignored2 <- package_version(ignore_rver) > current
    ignored <- ignored | ignored2
  }
  ignore_unavail <- vlapply(
    pkgs$params,
    is_true_param,
    "ignore-unavailable"
  )
  failed <- pkgs$status == "FAILED"
  ignored <- ignored | (ignore_unavail & failed)

  # ignore packages with the wrong OS type
  if (config$get("goal") == "install") {
    os <- os_type()
    bad <- which(!is.na(pkgs$os_type) & pkgs$os_type != os)
    if (length(bad) > 0) ignored[bad] <- TRUE
  }

  soft_deps <- tolower(pkg_dep_types_soft())

  ## 4. Package dependencies must be satisfied
  depconds <- function(wh) {
    if (pkgs$status[wh] != "OK") {
      return()
    }
    deps <- pkgs$deps[[wh]]
    deptypes <- pkgs$dep_types[[wh]]
    deps <- deps[deps$ref != "R", ]
    deps <- deps[!deps$ref %in% base, ]
    deps <- deps[tolower(deps$type) %in% tolower(deptypes), ]
    if (!linkingto && pkgs$platform[wh] != "source") {
      deps <- deps[tolower(deps$type) != "linkingto", ]
    }
    for (i in seq_len(nrow(deps))) {
      depref <- deps$ref[i]
      depver <- deps$version[i]
      depop <- deps$op[i]
      deppkg <- deps$package[i]
      deptyp <- tolower(deps$type[i])

      # candidates
      res <- pkgs[match(depref, pkgs$ref), ]
      cand <- which(pkgs$package == deppkg)

      # if all candidates are ignored and the package is a soft
      # dependency, then nothing to do
      if (all(ignored[cand]) && deptyp %in% soft_deps) {
        next
      }

      # good candidates
      good_cand <- Filter(
        x = cand,
        function(c) {
          candver <- pkgs$version[c]
          pkgs$status[[c]] != "FAILED" &&
            isTRUE(satisfies_remote(res, pkgs[c, ])) &&
            (depver == "" || version_satisfies(candver, depop, depver))
        }
      )
      bad_cand <- setdiff(cand, good_cand)

      report <- c(
        if (length(good_cand)) {
          gc <- paste(pkgs$ref[good_cand], pkgs$version[good_cand])
          paste0("version ", paste(gc, collapse = ", "))
        },
        if (length(bad_cand)) {
          bc <- paste(pkgs$ref[bad_cand], pkgs$version[bad_cand])
          paste0("but not ", paste(bc, collapse = ", "))
        },
        if (!length(cand)) "but no candidates"
      )
      txt <- sprintf(
        "%s depends on %s: %s",
        pkgs$ref[wh],
        depref,
        collapse(report, sep = ", ")
      )
      note <- list(
        wh = wh,
        ref = depref,
        cand = cand,
        good_cand = good_cand,
        txt = txt,
        depop = depop,
        depver = depver
      )

      lp <<- pkgplan_i_lp_add_cond(
        lp,
        c(wh, good_cand),
        "<=",
        rhs = 0,
        coef = c(1, rep(-1, length(good_cand))),
        type = "dependency",
        note = note
      )
    }
  }
  lapply(setdiff(seq_len(num_candidates), ruled_out), depconds)

  lp
}

#' @export

print.pkgplan_lp_problem <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
}

format_cond <- function(x, cond) {
  if (cond$type == "dependency") {
    paste0(cond$note$txt)
  } else if (cond$type == "satisfy-refs") {
    ref <- x$pkgs$ref[cond$note]
    cand <- x$pkgs$ref[cond$vars]
    sprintf("`%s` is not satisfied by `%s`", ref, cand)
  } else if (cond$type == "ok-resolution") {
    ref <- x$pkgs$ref[cond$vars]
    sprintf("`%s` resolution failed", ref)
  } else if (cond$type == "source-required") {
    ref <- x$pkgs$ref[cond$vars]
    sprintf("a source package was required for `%s` by the user", ref)
  } else if (cond$type == "ignored-by-user") {
    ref <- x$pkgs$ref[cond$vars]
    sprintf("`%s` explicitly ignored by user", ref)
  } else if (cond$type == "matching-platform") {
    ref <- x$pkgs$ref[cond$vars]
    plat <- x$pkgs$platform[cond$vars]
    sprintf("Platform `%s` does not match for `%s`", plat, ref)
  } else if (cond$type == "old-rversion") {
    ref <- x$pkgs$ref[cond$vars]
    sprintf("`%s` needs a newer R version: %s", ref, cond$note)
  } else if (cond$type == "new-rversion") {
    ref <- x$pkgs$ref[cond$vars]
    sprintf("`%s` needs an older R version: %s", ref, cond$node)
  } else if (cond$type == "different-rversion") {
    ref <- x$pkgs$ref[cond$vars]
    sprintf("`%s` needs a different R version: %s", ref, cond$note)
  } else if (cond$type == "direct-update") {
    package <- x$pkgs$package[cond$vars]
    sprintf("`%s` is direct, needs latest version", package)
  } else if (cond$type == "choose-latest") {
    ref <- x$pkgs$ref[cond$vars]
    sprintf("`%s` has a newer version of the same platform", ref)
  } else if (cond$type == "prefer-installed") {
    ref <- x$pkgs$ref[cond$vars]
    sprintf("installed is preferred for `%s`", ref)
  } else if (cond$type == "prefer-binary") {
    ref <- x$pkgs$ref[cond$vars]
    sprintf("binary is preferred for `%s`", ref)
  } else if (cond$type == "prefer-new-binary") {
    ref <- x$pkgs$ref[cond$vars]
    sprintf("newer binary is preferred for `%s`", ref)
  } else if (cond$type == "source-requested") {
    ref <- x$pkgs$ref[cond$vars]
    sprintf("source package is requested for `%s`", ref)
  } else if (cond$type == "exactly-once") {
    ref <- na.omit(x$pkgs$package[cond$vars])[1]
    sprintf("select %s exactly once", ref)
  } else if (cond$type == "at-most-once") {
    ref <- na.omit(x$pkgs$package[cond$vars])[1]
    sprintf("select %s at most once", ref)
  } else {
    "Unknown constraint"
  }
}

#' @export

format.pkgplan_lp_problem <- function(x, ...) {
  result <- character()
  push <- function(...) result <<- c(result, ...)

  push("<pkgplan_lp_problem>")
  push(sprintf("+ refs (%s):", x$num_candidates))
  pn <- sort(x$pkgs$ref)
  push(paste0("  - ", x$pkgs$ref))

  if (length(x$conds)) {
    push(sprintf("+ constraints (%s):", length(x$conds)))
    conds <- drop_nulls(lapply(x$conds, format_cond, x = x))
    push(paste0("  - ", conds))
  } else {
    push("+ no constraints")
  }

  result
}

pkgplan__solve_lp_problem <- function(self, private, problem) {
  res <- pkgplan_i_solve_lp_problem(problem)
  res
}

pkgplan_i_solve_lp_problem <- function(problem) {
  "!DEBUG solving LP problem"
  condmat <- matrix(0, nrow = length(problem$conds), ncol = problem$total)
  for (i in seq_along(problem$conds)) {
    cond <- problem$conds[[i]]
    condmat[i, cond$vars] <- cond$coef
  }

  dir <- vcapply(problem$conds, "[[", "op")
  rhs <- vapply(problem$conds, "[[", "rhs", FUN.VALUE = double(1))
  lpSolve::lp(
    "min",
    problem$obj,
    condmat,
    dir,
    rhs,
    int.vec = seq_len(problem$total)
  )
}

pkgplan_get_solution <- function(self, private) {
  if (is.null(private$solution)) {
    throw(pkg_error(
      "You need to call the {.code $solve()} method first."
    ))
  }
  private$solution$result
}

#' Highlight version number changes
#'
#' @param old Character vector, old versions. `NA` for new installs.
#' @param new Character vector, the new versions to highlight.
#' @return Character vector, like `new`, but the change highlighted
#'
#' @noRd

highlight_version <- function(old, new) {
  if (length(old) != length(new)) {
    throw(pkg_error(
      "Lengtgs of `old` and `new` must match",
      i = msg_internal_error()
    ))
  }
  if (length(new) == 0) {
    return(new)
  }

  wch <- !is.na(old) & old != new

  oldv <- strsplit(old[wch], "(?=[.-])", perl = TRUE)
  newv <- strsplit(new[wch], "(?=[.-])", perl = TRUE)
  new[wch] <- as.character(mapply(oldv, newv, FUN = function(o, n) {
    length(o) <- length(n) <- max(length(o), length(n))
    idx <- which(is.na(o) | is.na(n) | (o != n & o != "." & o != "-"))[1]
    n <- na.omit(n)
    paste0(
      if (idx > 1) paste(n[1:(idx - 1)], collapse = ""),
      if (idx <= length(n)) {
        cli::style_bold(paste(n[idx:length(n)]), collapse = "")
      }
    )
  }))

  new
}

#' Highlight package list
#'
#' @param sol Solution data, data frame, with at least these columns:
#' `type`, `package`, `old_version`, `version`, `lib_status`,
#' `cache_status`, `platform`, `needscompilation`. Just what
#' `$get_solution()$data` returns, basically.
#' @return Character vector of highlighted list. All strings will have the
#' same (printed) length. Packages that do not involve installation will
#' have `NA` in the result.
#'
#' @noRd

highlight_package_list <- function(sol) {
  arrow <- cli::symbol$arrow_right

  ins <- sol$type != "installed" & sol$type != "deps"
  sol <- sol[ins, ]

  pkg <- ansi_align_width(cli::col_blue(sol$package))
  old <- ansi_align_width(ifelse(is.na(sol$old_version), "", sol$old_version))
  arr <- ansi_align_width(ifelse(is.na(sol$old_version), "", arrow))
  new <- ansi_align_width(highlight_version(sol$old_version, sol$version))

  bld <- sol$lib_status %in% c("new", "update") & sol$platform == "source"
  cmp <- sol$lib_status %in%
    c("new", "update") &
    !is.na(sol$needscompilation) &
    sol$needscompilation
  dnl <- !is.na(sol$cache_status) & sol$cache_status == "miss"

  gh <- sol$type == "github"
  hash <- character(nrow(sol))
  hash[gh] <- vcapply(sol$metadata[gh], function(x) x["RemoteSha"])

  sysreqs <- highlight_sysreqs(sol$sysreqs_packages)

  ann <- paste0(
    ifelse(
      bld,
      if (has_emoji()) emo_builder(sum(ins)) else emoji("builder"),
      ""
    ),
    ifelse(cmp, emoji("wrench"), ""),
    ifelse(dnl, emoji("dl"), ""),
    ifelse(
      dnl & !is.na(sol$filesize),
      paste0(" ", format_file_size(sol$filesize)),
      ""
    ),
    ifelse(gh, paste0(" (GitHub: ", substr(hash, 1, 7), ")"), ""),
    sysreqs
  )

  lns <- paste0(pkg, " ", old, " ", arr, " ", new, " ", ann)

  ret <- rep(NA_character_, length(ins))
  ret[ins] <- lns

  key <- paste0(
    c(
      if (any(bld)) paste(emoji("builder"), "build"),
      if (any(cmp)) paste(emoji("wrench"), "compile"),
      if (any(dnl)) paste(emoji("dl"), "download")
    ),
    collapse = " | "
  )

  attr(ret, "key") <- if (key == "") "" else paste("[", key, "]")
  ret
}

highlight_sysreqs <- function(sysreqs) {
  if (is.null(sysreqs)) {
    return("")
  }
  vcapply(sysreqs, function(p) {
    if (length(p) == 0) {
      return("")
    }
    tick <- paste0(cli::col_green(cli::symbol$tick), " ")
    cross <- paste0(cli::col_red(cli::symbol$cross), " ")
    pkgs <- unlist(lapply(p, function(x) {
      if (length(x$packages) != 0) {
        if ("packages_missing" %in% names(x)) {
          paste0(
            ifelse(x$packages %in% x$packages_missing, cross, tick),
            x$packages
          )
        } else {
          x$packages
        }
      } else {
        paste0(x$sysreq, " (installer)")
      }
    }))
    if (length(pkgs) == 0) {
      return("")
    } # nocov
    paste0(
      cli::col_silver(" + "),
      paste(cli::col_cyan(pkgs), collapse = ", ")
    )
  })
}

pkgplan_show_solution <- function(self, private, key = FALSE) {
  self$stop_for_solve_error()
  sol <- self$get_solution()$data
  sol <- sol[order(sol$package), ]

  hl <- highlight_package_list(sol)
  hl2 <- na.omit(hl)

  if (length(hl2)) {
    hl2 <- paste0(cli::col_silver("+ "), hl2)
    if (key && attr(hl, "key") != "") {
      hl2 <- c(hl2, " ", attr(hl, "key"))
    }
    out <- paste(hl2, collapse = "\n")
    cli::cli_verbatim(hl2)
  }

  invisible(self$get_solution())
}

categorize_sysreqs <- function(rpkgs) {
  rpkgs <- rpkgs[vlapply(rpkgs$sysreqs_packages, function(x) length(x) > 0), ]

  inst <- structure(list(), names = character())
  miss <- structure(list(), names = character())
  upd <- structure(list(), names = character())

  for (i in seq_along(rpkgs$sysreqs_packages)) {
    elt <- rpkgs$sysreqs_packages[[i]]
    pkg <- rpkgs$package[i]
    for (j in seq_along(elt)) {
      if ("packages_missing" %in% names(elt[[j]])) {
        # we know exactly what is missing
        miss1 <- elt[[j]][["packages_missing"]]
        inst1 <- setdiff(elt[[j]][["packages"]], miss)
        upd1 <- character()
      } else {
        # we don't know what is missing
        upd1 <- elt[[j]][["packages"]]
        inst1 <- miss1 <- character()
      }

      for (p in miss1) {
        miss[[p]] <- c(miss[[p]], pkg)
      }
      for (p in inst1) {
        inst[[p]] <- c(inst[[p]], pkg)
      }
      for (p in upd1) {
        upd[[p]] <- c(upd[[p]], pkg)
      }
    }
  }

  list(inst = inst, miss = miss, upd = upd)
}

pkgplan_get_sysreqs <- function(self, private) {
  categorize_sysreqs(self$get_solution()[["data"]])
}

pkgplan_show_sysreqs <- function(self, private) {
  rpkgs <- self$get_solution()[["data"]]
  if (is.null(rpkgs[["sysreqs_packages"]])) {
    return(invisible())
  }
  cats <- categorize_sysreqs(rpkgs)
  inst <- cats$inst
  miss <- cats$miss
  upd <- cats$upd

  col1 <- col2 <- character()
  if (length(miss)) {
    miss <- miss[order(tolower(names(miss)))]
    col1 <- paste0(cli::col_silver("+ "), cli::col_cyan(names(miss)))
    col2 <- paste0(
      cli::col_silver("- "),
      vcapply(miss, function(x) paste(cli::col_blue(x), collapse = ", "))
    )
  }

  if (length(upd)) {
    upd <- upd[order(tolower(names(upd)))]
    col1 <- c(col1, paste0(cli::col_silver("* "), cli::col_cyan(names(upd))))
    col2 <- c(
      col2,
      paste0(
        cli::col_silver("- "),
        vcapply(upd, function(x) paste(cli::col_blue(x), collapse = ", "))
      )
    )
  }

  col1 <- ansi_align_width(col1)
  col2 <- ansi_align_width(col2)
  out <- paste0(col1, "  ", col2)
  if (length(out)) {
    cli::cli_verbatim(paste(out, collapse = "\n"))
  }

  invisible(cats)
}

pkgplan_install_plan <- function(self, private, downloads) {
  "!DEBUG creating install plan"
  sol <- if (downloads) {
    self$get_solution_download()
  } else {
    self$stop_for_solve_error()
    self$get_solution()$data
  }
  if (inherits(sol, "pkgplan_solve_error")) {
    return(sol)
  }

  # If it is coming from an install plan, then there is no 'deps' column,
  # but the dependencies column is already set.
  has_deps <- "deps" %in% names(sol)
  if (has_deps) {
    linkingto <- private$config$get("include_linkingto")
    if ("dep_types" %in% names(sol)) {
      selected_deps <- lapply(sol$dep_types, intersect, pkg_dep_types_hard())
    } else {
      selected_deps <- list(pkg_dep_types_hard())
    }
    deps <- lapply(
      seq_len(nrow(sol)),
      function(i) {
        x <- sol$deps[[i]]
        if (!linkingto && sol$platform[[i]] != "source") {
          x <- x[tolower(x$type) != "linkingto", ]
        }

        mydeps <- if (length(selected_deps) > 1) {
          selected_deps[[i]]
        } else {
          selected_deps[[1]]
        }
        pkgs <- x$package[tolower(x$type) %in% tolower(mydeps)]
        pkgs <- intersect(pkgs, sol$package)
        # Yes, some packages Suggest themselves, apparently
        setdiff(pkgs, sol$package[[i]])
      }
    )
    deps <- lapply(deps, setdiff, y = c("R", base_packages()))
  }

  installed <- rep(NA_character_, nrow(sol))
  if (any(sol$type == "installed")) {
    for (lib in c(private$config$get("library"), .Library)) {
      todo <- is.na(installed) & sol$type == "installed"
      ex <- file.exists(file.path(lib, sol$package[todo]))
      installed[todo][ex] <- file.path(lib, sol$package[todo][ex])
    }
  }

  res <- self$get_resolution()
  direct_packages <- res$package[res$direct]
  direct <- sol$direct |
    (sol$type == "installed" & sol$package %in% direct_packages)

  binary <- sol$platform != "source"
  if (downloads) {
    had <- paste("Had", current_r_platform())
    binary <- binary | sol$download_status == had
  }
  vignettes <- !binary &
    !sol$type %in% c("cran", "bioc", "standard") &
    private$config$get("build-vignettes")

  sol$library <- private$config$get("library")
  sol$binary <- binary
  sol$direct <- direct
  if (has_deps) {
    sol$dependencies <- I(deps)
  }
  sol$installed <- installed
  sol$vignettes <- vignettes

  # If we are on X64 Windows and we prefer x64, then only compile
  # for x64
  nomulti <- sol$platform %in%
    c("*", "source") &
    sol$type != "installed" &
    "x86_64-w64-mingw32" %in% private$config$get("platforms") &
    private$config$get("windows-archs") == "prefer-x64"
  sol$install_args <- ifelse(nomulti, "--no-multiarch", "")

  if (downloads) {
    tree <- file.exists(sol$fulltarget_tree)
    sol$packaged <- !tree
    sol$file <- ifelse(tree, sol$fulltarget_tree, sol$fulltarget)
  }

  sol
}

pkgplan_export_install_plan <- function(self, private, plan_file, version) {
  pkgs <- pkgplan_install_plan(self, private, downloads = FALSE)
  cols <- unique(c(
    "ref",
    "package",
    "version",
    "type",
    "direct",
    "binary",
    "dependencies",
    "vignettes",
    "needscompilation",
    "metadata",
    "sources",
    "target",
    "platform",
    "rversion",
    "built",
    "directpkg",
    "license",
    "sha256",
    "filesize",
    "dep_types",
    "params",
    "install_args",
    "repotype"
  ))

  packages <- pkgs[, cols]
  if ("sysreqs" %in% names(pkgs)) {
    packages[["sysreqs"]] <- pkgs[["sysreqs"]]
  }

  # drop missing system packages from sysreqs, we don't want these in
  # the lock file
  if ("sysreqs_packages" %in% names(pkgs)) {
    spkgs <- pkgs[["sysreqs_packages"]]
    for (i in seq_along(spkgs)) {
      elt <- spkgs[[i]]
      for (j in seq_along(elt)) {
        elt[[j]]$sysreq <- tojson$unbox(elt[[j]]$sysreq)
        elt[[j]]$packages_missing <- NULL
      }
      if (!is.null(elt)) spkgs[[i]] <- elt
    }
    packages[["sysreqs_packages"]] <- spkgs
  }

  packages$params <- lapply(
    packages$params,
    function(x) lapply(as.list(x), tojson$unbox)
  )

  plan <- list(
    lockfile_version = tojson$unbox(version),
    os = tojson$unbox(utils::sessionInfo()$running),
    r_version = tojson$unbox(R.Version()$version.string),
    platform = tojson$unbox(R.Version()$platform),
    packages = packages
  )

  if (private$config$get("sysreqs")) {
    sysreqs <- sysreqs2_scripts(
      self$get_solution()$data$sysreqs_packages,
      private$config$get("sysreqs_platform")
    )
    sysreqs$os <- tojson$unbox(sysreqs$os)
    sysreqs$distribution <- tojson$unbox(sysreqs$distribution)
    sysreqs$version <- tojson$unbox(sysreqs$version)
    sysreqs$url <- tojson$unbox(sysreqs$url)
    sysreqs$total <- NULL
    plan$sysreqs <- sysreqs
  }

  txt <- as_json_lite_plan(plan)
  writeLines(txt, plan_file)
}

as_json_lite_plan <- function(liteplan) {
  tolist1 <- function(x) lapply(x, function(v) lapply(as.list(v), tojson$unbox))
  liteplan$packages$metadata <- tolist1(liteplan$packages$metadata)
  json <- tojson$write_str(liteplan, opts = list(pretty = TRUE))
  json
}

calculate_lib_status <- function(sol, res) {
  ## Possible values at the moment:
  ## - virtual: not really a package
  ## - new: newly installed
  ## - current: up to date, not installed
  ## - update: will be updated
  ## - no-update: could update, but won't

  sres <- res[res$package %in% sol$package, c("package", "version", "type")]

  ## Check if it is not new
  lib_ver <- vcapply(sol$package, function(p) {
    c(
      sres$version[sres$package == p & sres$type == "installed"],
      NA_character_
    )[1]
  })

  ## If not new, and not "installed" type, that means update
  status <- ifelse(
    sol$type == "deps",
    "virtual",
    ifelse(
      is.na(lib_ver),
      "new",
      ifelse(sol$type == "installed", "current", "update")
    )
  )

  # Return NA for empty sets
  mmax <- function(...) c(suppressWarnings(max(...)), NA)[1]

  ## Check for no-update
  new_version <- vcapply(seq_along(sol$package), function(i) {
    p <- sol$package[i]
    v <- if (is.na(sol$version[i])) {
      NA_character_
    } else {
      package_version(sol$version[i])
    }
    g <- sres$package == p & !is.na(sres$version)
    mmax(sres$version[g][v < sres$version[g]])
  })
  status[status == "current" & !is.na(new_version)] <- "no-update"

  data_frame(
    lib_status = status,
    old_version = lib_ver,
    new_version = new_version
  )
}

## TODO: non-CRAN packages? E.g. GH based on sha.

calculate_cache_status <- function(soldata, cache) {
  toinst <- soldata$sha256[soldata$type != "installed"]
  nocache <- vlapply(soldata$params, is_true_param, "nocache")
  cached <- cache$package$find(sha256 = toinst)
  ifelse(
    soldata$type == "installed",
    NA_character_,
    ifelse(soldata$sha256 %in% cached$sha256 & !nocache, "hit", "miss")
  )
}

describe_solution_error <- function(pkgs, solution) {
  assert_that(
    !is.null(pkgs),
    !is.null(solution),
    solution$solution$objval >= solve_dummy_obj - 1L
  )

  num <- nrow(pkgs)
  if (!num) {
    throw(pkg_error(
      "No solution errors to describe",
      i = msg_internal_error()
    ))
  }
  sol <- solution$solution$solution
  sol_pkg <- sol[1:num]
  sol_dum <- sol[(num + 1):solution$problem$total]

  ## For each candidate, we work out if it _could_ be installed, and if
  ## not, why not. Possible cases:
  ## 1. it is in the install plan, so it can be installed, YES
  ## 2. it failed resolution, so NO
  ## 3. it needs a newer R version, so NO
  ## 4. it does not satisfy a direct ref for the same package, so NO
  ## 5. it conflicts with another to-be-installed candidate of the
  ##    same package, so NO
  ## 6. one of its (downstream) dependencies cannot be installed, so NO
  ## 7. otherwise YES

  FAILS <- c(
    "failed-res",
    "satisfy-direct",
    "conflict",
    "dep-failed",
    "old-rversion",
    "new-rvresion",
    "different-rversion",
    "matching-platform",
    "ignored-by-user",
    "binary-preferred",
    "source-required",
    "installed-preferred"
  )

  state <- rep("maybe-good", num)
  note <- replicate(num, NULL)
  downstream <- replicate(num, character(), simplify = FALSE)

  ## If a candidate of a package is OK, then we should not report errors
  ## about other candidates.
  ok_pkgs <- unique(pkgs$package[sol_pkg == 1])
  state[pkgs$package %in% ok_pkgs] <- "installed"

  cnd <- solution$problem$conds
  typ <- vcapply(cnd, "[[", "type")
  var <- lapply(cnd, "[[", "vars")

  ## Ignored by user
  ign_vars <- unlist(var[typ == "ignored-by-user"])
  ign_vars <- intersect(ign_vars, which(state == "maybe-good"))
  state[ign_vars] <- "ignored-by-user"

  ## Source required
  ign_vars <- unlist(var[typ == "source-required"])
  ign_vars <- intersect(ign_vars, which(state == "maybe-good"))
  state[ign_vars] <- "source-required"

  ## Ruled out in favor of an installed package
  ins_vars <- unlist(var[typ == "prefer-installed"])
  ins_vars <- intersect(ins_vars, which(state == "maybe-good"))
  state[ins_vars] <- "installed-preferred"

  ## Ruled out in favor of a binary package
  bin_vars <- unlist(var[typ %in% c("prefer-binary", "prefer-new-binary")])
  bin_vars <- intersect(bin_vars, which(state == "maybe-good"))
  state[bin_vars] <- "binary-preferred"

  ## Candidates that failed resolution
  fres_vars <- unlist(var[typ == "ok-resolution"])
  fres_vars <- intersect(fres_vars, which(state == "maybe-good"))
  state[fres_vars] <- "failed-res"
  for (fv in fres_vars) {
    if (length(e <- pkgs$error[[fv]])) {
      note[[fv]] <- c(note[[fv]], conditionMessage(e))
    }
  }

  ## Candidates that need a newer/older/different R version
  for (w in which(
    typ %in% c("old-rversion", "new-rversion", "different-rversion")
  )) {
    sv <- var[[w]]
    if (state[sv] != "maybe-good") {
      next
    }
    needs <- cnd[[w]]$note
    state[sv] <- typ[[w]]
    note[[sv]] <- c(note[[sv]], sprintf("Needs R %s", needs))
  }

  ## Candidates with platform mismatch
  for (w in which(typ == "matching-platform")) {
    sv <- var[[w]]
    if (state[sv] != "maybe-good") {
      next
    }
    state[sv] <- "matching-platform"
    note[[sv]] <- c(note[[sv]], "Platform mismatch")
  }

  ## Candidates that conflict with a direct package
  for (w in which(typ == "satisfy-refs")) {
    sv <- var[[w]]
    down <- pkgs$ref[sv]
    up <- pkgs$ref[cnd[[w]]$note]
    state[sv] <- "satisfy-direct"
    note[[sv]] <- c(note[[sv]], sprintf("Conflicts with %s", up))
  }

  ## Find "conflict". These are candidates that are not installed,
  ## and have an "at-most-once" constraint with another package that will
  ## be installed. So we just go over these constraints.
  for (c in cnd[typ == "at-most-once"]) {
    is_in <- sol_pkg[c$vars] != 0
    if (any(is_in)) {
      state[c$vars[!is_in]] <- "conflict"
      package <- pkgs$package[c$vars[1]]
      inst <- pkgs$ref[c$vars[is_in]]
      vv <- c$vars[!is_in]
      for (v in vv) {
        note[[v]] <- c(
          note[[v]],
          sprintf("%s conflicts with %s, to be installed", pkgs$ref[v], inst)
        )
      }
    }
  }

  ## Find "dep-failed". This is the trickiest. First, if there are no
  ## condidates at all
  type_dep <- typ == "dependency"
  dep_up <- viapply(cnd[type_dep], function(x) x$vars[1])
  dep_cands <- lapply(cnd[type_dep], function(x) x$vars[-1])
  no_cands <- which(
    !viapply(dep_cands, length) &
      state[dep_up] == "maybe-good"
  )
  for (x in no_cands) {
    pkg <- cnd[type_dep][[x]]$note$ref
    state[dep_up[x]] <- "dep-failed"
    note[[dep_up[x]]] <-
      c(note[[dep_up[x]]], sprintf("Can't install dependency %s", pkg))
    depop <- cnd[type_dep][[x]]$note$depop %||% ""
    depver <- cnd[type_dep][[x]]$note$depver %||% ""
    if (nzchar(depver)) {
      note[[dep_up[x]]] <- paste0(
        note[[dep_up[x]]],
        " (",
        depop,
        " ",
        depver,
        ")"
      )
    }
    downstream[[dep_up[x]]] <- c(downstream[[dep_up[x]]], pkg)
  }

  ## Then we start with the already known
  ## NO answers, and see if they rule out upstream packages
  new <- which(state %in% FAILS)
  while (length(new)) {
    dep_cands <- lapply(dep_cands, setdiff, new)
    which_new <- which(
      !viapply(dep_cands, length) & state[dep_up] == "maybe-good"
    )
    for (x in which_new) {
      pkg <- cnd[type_dep][[x]]$note$ref
      state[dep_up[x]] <- "dep-failed"
      note[[dep_up[x]]] <- c(
        note[[dep_up[x]]],
        sprintf("Can't install dependency %s", pkg)
      )
      depop <- cnd[type_dep][[x]]$note$depop %||% ""
      depver <- cnd[type_dep][[x]]$note$depver %||% ""
      if (nzchar(depver)) {
        note[[dep_up[x]]] <- paste0(
          note[[dep_up[x]]],
          " (",
          depop,
          " ",
          depver,
          ")"
        )
      }
      downstream[[dep_up[x]]] <- c(downstream[[dep_up[x]]], pkg)
    }
    new <- dep_up[which_new]
  }

  ## The rest is good
  state[state == "maybe-good"] <- "could-be"

  wh <- state %in% FAILS | (pkgs$direct & sol_pkg == 0)
  fails <- pkgs[wh, ]
  fails$failure_type <- state[wh]
  fails$failure_message <- note[wh]
  fails$failure_down <- downstream[wh]
  class(fails) <- unique(c("pkg_solution_failures", class(fails)))

  fails
}

#' @export

print.pkg_solution_failures <- function(x, ...) {
  writeLines(format(x, ...))
  invisible(x)
}

#' @export

format.pkg_solution_failures <- function(x, ...) {
  fails <- x
  if (!nrow(fails)) {
    return()
  }

  done <- rep(FALSE, nrow(x))
  res <- character()

  do <- function(i) {
    if (done[i]) {
      return()
    }
    if (
      fails$failure_type[i] %in% c("installed-preferred", "binary-preferred")
    ) {
      return()
    }
    done[i] <<- TRUE
    msgs <- unique(fails$failure_message[[i]])

    fail <- paste0("* ", cli::style_bold(fails$ref[i]), ":")
    if (length(msgs) == 0) {
      fail <- paste0(fail, " ", "dependency conflict")
    } else if (length(msgs) == 1) {
      fail <- paste0(fail, " ", msgs)
    } else if (length(msgs) > 1) {
      fail <- paste0(fail, "\n", paste0("  * ", msgs, collapse = "\n"))
    }

    res <<- c(res, fail)
    down <- which(
      fails$package %in%
        fails$failure_down[[i]] |
        fails$ref %in% fails$failure_down[[i]]
    )
    lapply(down, do)
  }

  direct_refs <- which(fails$direct)
  lapply(direct_refs, do)

  unique(res)
}

#' @export

`[.pkg_solution_result` <- function(x, i, j, drop = FALSE) {
  class(x) <- setdiff(class(x), "pkg_solution_result")
  NextMethod("[")
}

#' @export

format.pkg_solution_result <- function(x, ...) {
  ok <- is.null(x$failures) || nrow(x$failures) == 0
  refs <- sort(unique(x$problem$pkgs$ref[x$problem$pkgs$direct]))
  nc <- length(x$problem$conds)
  cnst <- unlist(lapply(x$problem$conds, format_cond, x = x$problem))
  solrefs <- sort(x$data$ref)
  c(
    "<pkg_solution>",
    paste0("+ result: ", if (ok) "OK" else "FAILED"),
    "+ refs:",
    paste0("  - ", refs),
    if (nc == 0) "+ no constraints",
    if (nc > 0) paste0("+ constraints (", length(cnst), "):"),
    if (nc > 0) paste0("  - ", utils::head(cnst, 10)),
    if (nc > 10) "  ...",
    if (ok) c("+ solution:", paste0("  - ", solrefs)),
    if (!ok) c("x failures:", format(x$failures))
  )
}

#' @export

print.pkg_solution_result <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
}

test_that("newer version preferred", {
  pkgs <- make_fake_resolution(
    `aa` = list(version = "1-0-1", platform = "source", direct = TRUE),
    `installed::/tmp/aa` = list(version = "1.2.0", type = "cran")
  )
  lp <- pkgplan_i_create_lp_problem(
    pkgs,
    config = current_config(),
    policy = "upgrade"
  )
  sol <- pkgplan_i_solve_lp_problem(lp)
  expect_true(sol$objval < 1000)
  expect_equal(as.logical(sol$solution), c(FALSE, TRUE, FALSE))

  pkgs <- make_fake_resolution(
    `aa` = list(version = "1-0-1", platform = "source", direct = TRUE),
    `installed::/tmp/aa` = list(version = "1.0.0", type = "bioc")
  )
  lp <- pkgplan_i_create_lp_problem(
    pkgs,
    config = current_config(),
    policy = "upgrade"
  )
  sol <- pkgplan_i_solve_lp_problem(lp)
  expect_true(sol$objval < 1000)
  expect_equal(as.logical(sol$solution), c(TRUE, FALSE, FALSE))
})

test_that("if newer fails, older is used", {
  pkgs <- make_fake_resolution(
    `aa` = list(version = "0.9.9", direct = TRUE),
    `cran::aa` = list(version = "1-0-1", status = "FAILED"),
    `installed::/tmp/aa` = list(version = "1.0.0", type = "cran")
  )
  lp <- pkgplan_i_create_lp_problem(
    pkgs,
    config = current_config(),
    policy = "upgrade"
  )
  sol <- pkgplan_i_solve_lp_problem(lp)
  expect_true(sol$objval < 1000)
  expect_equal(as.logical(sol$solution), c(FALSE, FALSE, TRUE, FALSE))
})

test_that("newer version if preferred in the dependencies", {
  pkgs <- make_fake_resolution(
    `a0` = list(direct = TRUE, deps = list(make_fake_deps(Imports = "aa"))),
    `aa` = list(version = "1-0-1", platform = "source"),
    `installed::/tmp/aa` = list(version = "1.0.0")
  )
  lp <- pkgplan_i_create_lp_problem(
    pkgs,
    config = current_config(),
    policy = "upgrade"
  )
  sol <- pkgplan_i_solve_lp_problem(lp)
  expect_true(sol$objval < 1000)
  expect_equal(as.logical(sol$solution), c(TRUE, TRUE, FALSE, FALSE))
})

test_that("binaries are still preferred over source", {
  pkgs <- make_fake_resolution(
    `aa` = list(version = "1.0.0", direct = TRUE, platform = "macos"),
    `cran::aa` = list(version = "1.0.0", platform = "source"),
    `installed::/tmp/aa` = list(version = "0.9.9")
  )
  config <- current_config()$set("platforms", c("macos", "source"))
  lp <- pkgplan_i_create_lp_problem(pkgs, config = config, policy = "upgrade")
  sol <- pkgplan_i_solve_lp_problem(lp)
  expect_true(sol$objval < 1000)
  expect_equal(as.logical(sol$solution), c(TRUE, FALSE, FALSE, FALSE))

  pkgs <- make_fake_resolution(
    `cran::aa` = list(version = "1.0.0", platform = "source"),
    `aa` = list(version = "1.0.0", direct = TRUE, platform = "macos"),
    `installed::/tmp/aa` = list(version = "0.9.9")
  )
  lp <- pkgplan_i_create_lp_problem(pkgs, config = config, policy = "upgrade")
  sol <- pkgplan_i_solve_lp_problem(lp)
  expect_true(sol$objval < 1000)
  expect_equal(as.logical(sol$solution), c(FALSE, TRUE, FALSE, FALSE))
})

test_that("installed is still preferred over binaries", {
  pkgs <- make_fake_resolution(
    `aa` = list(version = "1.0.0", direct = TRUE, platform = "macos"),
    `cran::aa` = list(version = "1.0.0", platform = "source"),
    `installed::/tmp/aa` = list(
      version = "1.0.0",
      extra = list(list(repotype = "cran"))
    )
  )
  config <- current_config()$set("platforms", c("macos", "source"))
  lp <- pkgplan_i_create_lp_problem(pkgs, config = config, policy = "upgrade")
  sol <- pkgplan_i_solve_lp_problem(lp)
  expect_true(sol$objval < 1000)
  expect_equal(as.logical(sol$solution), c(FALSE, FALSE, TRUE, FALSE))

  pkgs <- make_fake_resolution(
    `cran::aa` = list(version = "1.0.0", platform = "source"),
    `aa` = list(version = "1.0.0", direct = TRUE, platform = "macos"),
    `installed::/tmp/aa` = list(
      version = "1.0.0",
      extra = list(list(repotype = "cran"))
    )
  )
  lp <- pkgplan_i_create_lp_problem(pkgs, config = config, policy = "upgrade")
  sol <- pkgplan_i_solve_lp_problem(lp)
  expect_true(sol$objval < 1000)
  expect_equal(as.logical(sol$solution), c(FALSE, FALSE, TRUE, FALSE))

  pkgs <- make_fake_resolution(
    `installed::/tmp/aa` = list(
      version = "1.0.0",
      extra = list(list(repotype = "cran"))
    ),
    `cran::aa` = list(version = "1.0.0", platform = "source"),
    `aa` = list(version = "1.0.0", direct = TRUE, platform = "macos")
  )
  lp <- pkgplan_i_create_lp_problem(pkgs, config = config, policy = "upgrade")
  sol <- pkgplan_i_solve_lp_problem(lp)
  expect_true(sol$objval < 1000)
  expect_equal(as.logical(sol$solution), c(TRUE, FALSE, FALSE, FALSE))

  pkgs <- make_fake_resolution(
    `cran::aa` = list(version = "1.0.0", platform = "source"),
    `installed::/tmp/aa` = list(
      version = "1.0.0",
      extra = list(list(repotype = "cran"))
    ),
    `aa` = list(version = "1.0.0", direct = TRUE, platform = "macos")
  )
  lp <- pkgplan_i_create_lp_problem(pkgs, config = config, policy = "upgrade")
  sol <- pkgplan_i_solve_lp_problem(lp)
  expect_true(sol$objval < 1000)
  expect_equal(as.logical(sol$solution), c(FALSE, TRUE, FALSE, FALSE))
})

test_that("unqualified is cran/bioc", {
  ## If installed is GH, that's not good
  pkgs <- make_fake_resolution(
    `aa` = list(version = "1.0.0", direct = TRUE, platform = "macos"),
    `cran::aa` = list(version = "1.0.0", platform = "source"),
    `installed::/tmp/aa` = list(
      version = "1.0.0",
      extra = list(list(repotype = "github"))
    )
  )
  config <- current_config()$set("platforms", c("macos", "source"))
  lp <- pkgplan_i_create_lp_problem(pkgs, config = config, policy = "upgrade")
  sol <- pkgplan_i_solve_lp_problem(lp)
  expect_true(sol$objval < 1000)
  expect_equal(as.logical(sol$solution), c(TRUE, FALSE, FALSE, FALSE))
})

make_fake_deps <- function(...) {
  assert_that(all_named(list(...)))

  d <- desc::desc("!new")
  if (length(list(...))) {
    d$set(...)
  }
  resolve_ref_deps(
    d$get_deps(),
    d$get("Remotes")[[1]],
    NULL
  )
}

make_fake_resolution1 <- function(ref, args = list()) {
  pref <- parse_pkg_refs(ref)[[1]]
  if (!is.null(args$extra)) {
    pref[names(args$extra)] <- args$extra
  }

  mirror <- args$mirror %||% current_config()$get("cran-mirror")
  repodir <- args$repodir %||% "src/contrib"
  version <- args$version %||% "1.0.0"
  filename <- paste0(pref$package, "_", version, ".tar.gz")

  def <- list(
    ref = ref,
    type = pref$type,
    package = pref$package,
    version = version,
    sources = c(
      sprintf("%s/%s/%s", mirror, repodir, filename),
      sprintf("%s/%s/Archive/%s/%s", mirror, repodir, pref$package, filename)
    ),
    dep_types = tolower(pkg_dep_types_hard())
  )

  modifyList(def, args)
}

make_fake_metadata <- function() {
  list(
    resolution_start = Sys.time(),
    resolution_end = Sys.time()
  )
}

make_fake_resolution <- function(...) {
  pkgs <- list(...)
  assert_that(all_named(pkgs))
  ress <- lapply_with_names(
    names(pkgs),
    function(n) make_fake_resolution1(n, pkgs[[n]])
  )

  res <- res_make_empty_df()
  for (r in ress) {
    res <- res_add_df_entries(res, r)
  }
  res
}

describe_fake_error <- function(
  pkgs,
  policy = "lazy",
  config = current_config()
) {
  lp <- pkgplan_i_create_lp_problem(pkgs, config, policy)
  sol <- pkgplan_i_solve_lp_problem(lp)

  expect_true(sol$objval >= solve_dummy_obj - 1)
  solution <- list(status = "FAILED", data = pkgs, problem = lp, solution = sol)
  describe_solution_error(pkgs, solution)
}

test_that("pkgplan_i_create_lp_init", {
  pkgs <- read_fixture("resolution-simple.rds")
  config <- current_config()
  lp <- pkgplan_i_lp_init(pkgs, config, policy = "lazy")
  expect_equal(lp$num_candidates, 2)
  expect_equal(lp$num_direct, 1)
  expect_equal(lp$total, 3)
  expect_equal(lp$conds, list())
  expect_equal(lp$pkgs, pkgs)
  expect_equal(lp$policy, "lazy")
  expect_equal(lp$packages, "pkgconfig")
  expect_equal(lp$direct_packages, "pkgconfig")
  expect_equal(lp$indirect_packages, character())
  expect_equal(lp$ruled_out, integer())
})

test_that("pkgplan_i_lp_objectives lazy policy", {
  pkgs <- read_fixture("resolution-simple.rds")
  config <- current_config()
  lp0 <- pkgplan_i_lp_init(pkgs, config, policy = "lazy")
  lp <- pkgplan_i_lp_objectives(lp0)
  expect_equal(lp0[setdiff(names(lp0), "obj")], lp[setdiff(names(lp), "obj")])
  expect_true(
    lp$obj[which(pkgs$platform != "source")] <
      lp$obj[which(pkgs$platform == "source")]
  )
  expect_equal(lp$obj[3], solve_dummy_obj)
})

test_that("pkgplan_i_lp_objectives upgrade policy", {
  ## TODO
})

test_that("pkgplan_i_lp_no_multiples", {
  pkgs <- read_fixture("resolution-progress.rds")
  config <- current_config()
  lp <- pkgplan_i_lp_init(pkgs, config, "lazy")
  lp <- pkgplan_i_lp_no_multiples(lp)
  expect_equal(
    vcapply(lp$conds, "[[", "type"),
    c(
      rep("exactly-once", length(lp$direct_packages)),
      rep("at-most-once", length(lp$indirect_packages))
    )
  )
  prvar <- which(pkgs$package == "progress")
  expect_equal(
    lp$conds[[1]],
    structure(
      list(
        vars = c(prvar, nrow(pkgs) + 1L),
        coef = c(1, 1, 1),
        op = "==",
        rhs = 1,
        type = "exactly-once",
        note = NULL
      )
    )
  )
  prvar2 <- which(pkgs$package == "assertthat")
  expect_equal(
    lp$conds[[2]],
    structure(list(
      vars = prvar2,
      coef = c(1, 1),
      op = "<=",
      rhs = 1,
      type = "at-most-once",
      note = NULL
    ))
  )
})

test_that("pkgplan_i_lp_satisfy_direct", {
  pkgs <- read_fixture("resolution-gh-vs-cran.rds")
  config <- current_config()
  lp <- pkgplan_i_lp_init(pkgs, config, "lazy")
  lp <- pkgplan_i_lp_satisfy_direct(lp)
  expect_equal(
    vcapply(lp$conds, "[[", "type"),
    rep("satisfy-refs", 4)
  )
  expect_equal(
    sort(viapply(lp$conds, "[[", "vars")),
    c(1, 2, 3, 3)
  )
  expect_equal(vcapply(lp$conds, "[[", "op"), rep("==", 4))
  expect_equal(vdapply(lp$conds, "[[", "rhs"), rep(0, 4))
})

test_that("pkgplan_i_lp_failures", {
  ## TODO
})

test_that("pkgplan_i_lp_prefer_installed", {
  ## TODO
})

test_that("pkgplan_i_lp_prefer_binaries", {
  ## TODO
})

test_that("pkgplan_i_lp_prefer_new_binaries skips version-pinned direct refs", {
  # A newer binary (pkg@4.0.2) is available, but the user directly requested
  # pkg@3.5.2 (as a PPM binary with empty platform string). The 3.5.2 candidate
  # must NOT be ruled out by prefer_new_binaries;
  pkgs <- res_make_empty_df()
  for (entry in list(
    make_fake_resolution1(
      "pkg@3.5.2",
      list(
        direct = TRUE,
        version = "3.5.2",
        platform = "" # PPM binary: empty platform for pure-R packages
      )
    ),
    make_fake_resolution1(
      "pkg",
      list(
        direct = FALSE,
        version = "4.0.2",
        platform = "x86_64-pc-linux-gnu"
      )
    ),
    make_fake_resolution1(
      "pkg1",
      list(
        direct = FALSE,
        version = "1.0.0",
        platform = "x86_64-pc-linux-gnu"
      )
    ),
    make_fake_resolution1(
      "pkg1",
      list(
        direct = FALSE,
        version = "2.0.0",
        platform = "x86_64-pc-linux-gnu"
      )
    )
  )) {
    pkgs <- res_add_df_entries(pkgs, entry)
  }

  config <- current_config()
  lp <- pkgplan_i_lp_init(pkgs, config, "lazy")
  lp <- pkgplan_i_lp_prefer_new_binaries(lp)

  types <- vcapply(lp$conds, "[[", "type")
  pnb_vars <- unlist(lapply(
    lp$conds[types == "prefer-new-binary"],
    "[[",
    "vars"
  ))

  pkg352_idx <- which(pkgs$package == "pkg" & pkgs$version == "3.5.2")
  pkg402_idx <- which(pkgs$package == "pkg" & pkgs$version == "4.0.2")
  pkg1_old_idx <- which(pkgs$package == "pkg1" & pkgs$version == "1.0.0")
  pkg1_new_idx <- which(pkgs$package == "pkg1" & pkgs$version == "2.0.0")

  # The pinned pkg@3.5.2 must not be ruled out
  expect_false(pkg352_idx %in% pnb_vars)
  expect_false(pkg352_idx %in% lp$ruled_out)

  # pkg@4.0.2 is also unaffected (pkg as a whole is skipped when pinned)
  expect_false(pkg402_idx %in% pnb_vars)

  # pkg1@1.0.0 (no pin, indirect) is correctly ruled out in favour of 2.0.0
  expect_true(pkg1_old_idx %in% pnb_vars)
  expect_true(pkg1_old_idx %in% lp$ruled_out)
  expect_false(pkg1_new_idx %in% pnb_vars)
  expect_false(pkg1_new_idx %in% lp$ruled_out)
})

test_that("pkgplan_i_lp_dependencies", {
  pkgs <- read_fixture("resolution-progress.rds")
  config <- current_config()
  lp <- pkgplan_i_lp_init(pkgs, config, "lazy")
  lp <- pkgplan_i_lp_dependencies(lp, config)
  expect_equal(length(lp$conds), 32)
  expect_snapshot(lp$conds[[3]])
})

test_that("pkgplan_i_lp_rversion", {
  pkgs <- read_fixture("resolution-progress.rds")
  config <- current_config()
  lp <- pkgplan_i_lp_init(pkgs, config, "lazy")
  lp <- pkgplan_i_lp_rversion(lp, "3.1.3")
  fmt <- format(lp)
  expect_true(any(grepl("`rlang` needs a newer R version", fmt, fixed = TRUE)))
  expect_true(any(grepl("`vctrs` needs a newer R version", fmt, fixed = TRUE)))
})

test_that("highlight_version", {
  b <- cli::style_bold
  g <- function(x, .envir = parent.frame()) {
    unname(vcapply(x, glue::glue, .envir = .envir))
  }

  cases <- list(
    list(
      c("2.3.4", "2.3.4", "2.3.4", "2.3.4"),
      c("2.3.5", "2.3.2", "2.4.0", "3.0.0"),
      c("2.3.{b(5)}", "2.3.{b(2)}", "2.{b('4.0')}", "{b('3.0.0')}")
    ),
    list(character(), character(), character()),
    list("1.0.0", "1.0.0", "1.0.0"),
    list(
      c("1.0.0", "1.0.0"),
      c("2.0.0", "1.0.0"),
      c("{b('2.0.0')}", "1.0.0")
    ),
    list("1.0.0.9000", "1.0.0", "1.0.0")
  )

  for (case in cases) {
    expect_equal(highlight_version(case[[1]], case[[2]]), g(case[[3]]))
  }
})

# TODO: allow restriting to dumb, v1, v2 protocol

git_app <- function(
  git_root,
  git_timeout = as.difftime(1, units = "mins"),
  filter = TRUE,
  cleanup = TRUE
) {
  app <- webfakes::new_app()
  app$locals$git_root <- git_root
  app$locals$git_timeout <- as.double(git_timeout, units = "secs") * 1000
  app$locals$git_config <- tempfile()

  reg.finalizer(app, function(app0) unlink(app$locals$git_config), TRUE)
  writeLines(
    c(
      "[uploadpack]",
      paste0("\tallowFilter = ", if (isTRUE(filter)) "true" else "false")
    ),
    app$locals$git_config
  )

  if (cleanup) {
    reg.finalizer(
      app,
      function(app) unlink(app$locals$git_root, recursive = TRUE),
      TRUE
    )
  }

  app$get(webfakes::new_regexp("^(?<path>.*)$"), function(req, res) {
    out <- tempfile()
    err <- tempfile()
    on.exit(unlink(c(out, err)), add = TRUE)
    px <- processx::process$new(
      "git",
      "http-backend",
      env = git_env_vars(req),
      stdout = out,
      stderr = err
    )
    px$wait(req$app$locals$git_timeout)
    parse_cgi_output(px, out, err, res)
  })

  app$post(webfakes::new_regexp("^(?<path>.*)$"), function(req, res) {
    tmp <- tempfile()
    out <- tempfile()
    err <- tempfile()
    on.exit(unlink(c(out, err, tmp)), add = TRUE)
    writeBin(req$.body, tmp)
    px <- processx::process$new(
      "git",
      "http-backend",
      env = git_env_vars(req),
      stdin = tmp,
      stdout = out,
      stderr = err
    )
    px$wait(req$app$locals$git_timeout)
    parse_cgi_output(px, out, err, res)
  })

  app
}

git_env_vars <- function(req) {
  url <- parse_url(req$url)
  c(
    "current",

    # For git
    GIT_CONFIG_GLOBAL = req$app$locals$git_config,
    GIT_HTTP_EXPORT_ALL = "true",
    GIT_PROJECT_ROOT = req$app$locals$git_root,
    GIT_PROTOCOL = req$get_header("Git-Protocol") %||% "",
    HTTP_GIT_PROTOCOL = req$get_header("Git-Protocol") %||% "",

    # general CGI
    CONTENT_LENGTH = if (length(req$.body)) length(req$.body),
    CONTENT_TYPE = req$get_header("content-type") %||% "",
    GATEWAY_INTERFACE = "CGI/1.1",
    PATH_INFO = req$path,
    QUERY_STRING = req$query_string,
    REMOTE_ADDR = req$remote_addr,
    REMOTE_HOST = req$remote_addr,
    REMOTE_USER = "anonymous",
    REQUEST_METHOD = toupper(req$method),
    SERVER_NAME = url$host,
    SERVER_PORT = url$port,
    SERVER_PROTOCOL = paste0("http/", req$http_version),
    SERVER_SOFTWARE = "https://github.com/r-lib/webfakes"
  )
}

parse_cgi_output <- function(px, out, err, res) {
  if (px$is_alive() || px$get_exit_status() != 0) {
    px$kill()
    res$set_status(500)$send(paste0("Internal git error: ", err))
  }

  out <- read_bin(out)
  err <- read_char(err)

  cgi_res <- split_cgi_output(out)
  headers <- cgi_res$headers

  for (idx in seq_along(headers)) {
    if (tolower(names(headers)[idx]) == "status") {
      res$set_status(parse_status(headers[[idx]]))
    } else {
      res$set_header(names(headers)[idx], headers[[idx]])
    }
  }

  if (!"status" %in% names(headers)) {
    res$set_status(200L)
  }

  res$send(cgi_res$body)
}

split_cgi_output <- function(x) {
  nlnl <- grepRaw("\r?\n\r?\n", x)[1]
  if (is.na(nlnl)) {
    stop("Invalid response from git cgi, no headers?")
  }

  headers <- parse_headers(rawToChar(x[1:(nlnl - 1L)]))

  body <- x[nlnl:length(x)]
  ndrop <- 1L
  while (body[ndrop] != 0x0a) {
    ndrop <- ndrop + 1L
  }
  ndrop <- ndrop + 1L
  while (body[ndrop] != 0x0a) {
    ndrop <- ndrop + 1L
  }
  body <- utils::tail(body, -ndrop)

  list(headers = headers, body = body)
}

parse_status <- function(x) {
  status <- as.integer(strsplit(x, " ", fixed = TRUE)[[1]][1])
  if (is.na(status)) {
    stop("Invalid status from git cgi: ", x)
  }
}

read_bin <- function(path) {
  readBin(path, "raw", file.info(path)$size)
}

parse_headers <- function(txt) {
  headers <- grep(":", parse_headers0(txt), fixed = TRUE, value = TRUE)
  out <- lapply(headers, split_header)
  names <- tolower(vapply(out, `[[`, character(1), 1))
  values <- lapply(lapply(out, `[[`, 2), trimws)
  names(values) <- names
  values
}

parse_headers0 <- function(txt, multiple = FALSE) {
  if (!length(txt)) {
    return(NULL)
  }
  if (is.raw(txt)) {
    txt <- rawToChar(txt)
  }
  stopifnot(is.character(txt))
  if (length(txt) > 1) {
    txt <- paste(txt, collapse = "\n")
  }
  sets <- strsplit(txt, "\\r\\n\\r\\n|\\n\\n|\\r\\r")[[1]]
  headers <- strsplit(sets, "\\r\\n|\\n|\\r")
  if (multiple) {
    headers
  } else {
    headers[[length(headers)]]
  }
}

split_header <- function(x) {
  pos <- grepRaw(":", x, fixed = TRUE)[1]
  if (is.na(pos)) {
    stop("Invalid response header from git cgi: ", x)
  }
  c(substr(x, 1, pos - 1L), substr(x, pos + 1L, nchar(x)))
}

parse_url <- function(url) {
  re_url <- paste0(
    "^(?<protocol>[a-zA-Z0-9]+)://",
    "(?:(?<username>[^@/:]+)(?::(?<password>[^@/]+))?@)?",
    "(?<url>(?<host>[^:/]+)",
    "(?::(?<port>[0-9]+))?",
    "(?<path>/.*))$" # don't worry about query params here...
  )
  re_match(url, re_url)
}

version_satisfies <- function(ver, op, cond) {
  ver <- package_version(ver)
  switch(
    op,
    "<" = ver < cond,
    "<=" = ver <= cond,
    "==" = ver == cond,
    ">=" = ver >= cond,
    ">" = ver > cond,
    "!=" = ver != cond
  )
}

satisfies_remote <- function(
  resolution,
  candidate,
  config,
  remote_types = NULL,
  ...
) {
  remote_types <- c(default_remote_types(), remote_types)
  sat <- remote_types[[resolution$type]]$satisfy
  if (is.null(sat)) {
    return(resolution$ref == candidate$ref)
  }

  sat(resolution, candidate, config, ...)
}

installedok_remote <- function(
  installed,
  solution,
  config,
  remote_types = NULL,
  ...
) {
  remote_types <- c(default_remote_types(), remote_types)
  ok <- remote_types[[solution$type]]$installedok
  if (is.null(ok)) {
    return(FALSE)
  }
  ok(installed, solution, config, ...)
}

test_that("CRAN packages", {
  ## Satisfies itself
  pkgs <- make_fake_resolution(
    `cran::pkgA` = list()
  )
  expect_true(satisfies_remote(pkgs[1, ], pkgs[1, ]))

  ## A certain version satisfies a ref without version requirements
  pkgs <- make_fake_resolution(
    `cran::pkgA` = list(),
    `cran::pkgA@2.0.0` = list()
  )
  expect_true(satisfies_remote(pkgs[1, ], pkgs[2, ]))
})

verify_extracted_package <- function(filename, parent_path) {
  pkg_name <- dir(parent_path)
  pkg_path <- file.path(parent_path, pkg_name)

  if (length(pkg_name) == 0) {
    throw(pkg_error(
      "{.path {filename}} is not a valid R package, it is an empty archive.",
      .class = "install_input_error"
    ))
  } else if (length(pkg_name) > 1) {
    throw(pkg_error(
      "{.path {filename}} is not a valid R package, it should contain a
      single directory",
      .class = "install_input_error"
    ))
  }

  rel_package_files <- c(
    file.path(pkg_name, "Meta", "package.rds"),
    file.path(pkg_name, "DESCRIPTION")
  )
  package_files <- file.path(parent_path, rel_package_files)

  has_files <- file.exists(package_files)
  if (!all(has_files)) {
    miss <- rel_package_files[!has_files]
    throw(pkg_error(
      "{.path {filename}} is not a valid binary, it is missing {miss}.",
      .data = list(package = pkg_name),
      .class = "install_input_error"
    ))
  }

  rel_dsc_file <- file.path(pkg_name, "DESCRIPTION")
  dsc_file <- file.path(pkg_path, "DESCRIPTION")
  dsc <- tryCatch(
    desc::desc(dsc_file),
    error = function(e) {
      throw(pkg_error(
        "{.path {filename}} is not a valid binary, invalid DESCRIPTION
        file at {.path {rel_dsc_file}}.",
        .data = list(package = pkg_name),
        .class = "install_input_error"
      ))
    }
  )

  if (!length(dsc$fields())) {
    throw(pkg_error(
      "{.path {filename}} is not a valid binary, empty DESCRIPTION file
      at {.path {rel_dsc_file}}.",
      .data = list(package = pkg_name),
      .class = "install_input_error"
    ))
  }

  dsc_pkg <- dsc$get("Package")
  if (is.na(dsc_pkg)) {
    throw(pkg_error(
      "{.path {filename}} has no `Package` entry in DESCRIPTION at
      {.path {rel_dsc_file}}.",
      .data = list(package = pkg_name),
      .class = "install_input_error"
    ))
  }

  if (pkg_name != str_trim(dsc_pkg[[1]])) {
    throw(pkg_error(
      "{.path {filename}} is not a valid binary, package name mismatch in
      archive and in {.path {rel_dsc_file}}.",
      .data = list(package = pkg_name),
      .class = "install_input_error"
    ))
  }

  if (is.na(dsc$get("Built"))) {
    throw(pkg_error(
      "{.path {filename}} is not a valid binary, no 'Built' entry in
      {.path {rel_dsc_file}}.",
      .data = list(package = pkg_name),
      .class = "install_input_error"
    ))
  }

  list(name = pkg_name, path = pkg_path, desc = dsc)
}

#' Package references
#'
#' @description
#' A package reference (ref) specifies a location from which an R package
#' can be obtained from. The full syntax of a reference is `type::ref`, but
#' `type` can be often omitted, the common ref types have shortcuts.
#'
#' @details
#' ```{r child = "tools/doc/pkg-refs.Rmd"}
#' ```
#' `r doc_share_rmd("tools/doc/pkg-refs.Rmd", "inst/docs/pkg-refs.rds")`
#'
#' @name pkg_refs
NULL

package_name_rx <- function() "[a-zA-Z][a-zA-Z0-9.]*[a-zA-Z0-9]"

## CRAN and GitHub are special, because they have shorthands,
## so we need to know their regexes to find the type of the remotes

standard_rx <- function(remote_name = "standard") {
  paste0(
    "^",
    ## Optional remote type
    "(?:",
    remote_name,
    "::)?",
    ## Package name, only valid names
    "(?<package>",
    package_name_rx(),
    ")",
    ## Package version, only valid version numbers
    "(?:@(?:(?:(?<atleast>>=)?",
    "(?<version>[0-9]+[-\\.][0-9]+(?:[-\\.][0-9]+)*|current|last))))?",
    "$"
  )
}

#' Match a GitHub username
#'
#' * may only contain alphanumeric characters or hyphens
#' * cannot have multiple consecutive hyphens
#' * cannot begin or end with a hyphen
#' * maximum 39 characters
#'
#' Based on <https://github.com/shinnn/github-username-regex>
#'
#' @noRd

github_username_rx <- function() {
  "(?<username>(?:[a-zA-Z\\d](?:[a-zA-Z\\d-]){0,38}))"
}

github_repo_rx <- function() "(?<repo>[^/@#]+)"
github_subdir_rx <- function() "(?:/(?<subdir>(?:[^@#]*[^@#/])/?))"
github_commitish_rx <- function() "(?:@(?<commitish>[^*].*))"
github_pull_rx <- function() "(?:#(?<pull>[0-9]+))"
github_release_rx <- function() "(?:@(?<release>[*]release))"
github_detail_rx <- function() {
  sprintf(
    "(?:(?:%s)|(?:%s)|(?:%s))?",
    github_commitish_rx(),
    github_pull_rx(),
    github_release_rx()
  )
}

github_rx <- function() {
  paste0(
    "^",
    ## Optional package name
    "(?:(?<package>",
    package_name_rx(),
    ")=)?",
    ## Optional remote type
    "(?:github::)?",
    github_username_rx(),
    "/",
    github_repo_rx(),
    github_subdir_rx(),
    "?",
    ## Commit / PR / Release
    github_detail_rx(),
    "$"
  )
}

github_url_commitish_rx <- function() {
  "(?:(?:tree|commit|releases/tag)/(?<commitish>.+$))"
}

github_url_pull_rx <- function() "(?:pull/(?<pull>.+$))"

github_url_release_rx <- function() "(?:releases/)(?<release>.+$)"

github_url_detail_rx <- function() {
  paste0(
    "(?:/(?:",
    github_url_commitish_rx(),
    "|",
    github_url_pull_rx(),
    "|",
    github_url_release_rx(),
    "))?"
  )
}

## We need to select the shortest match here, to avoid matching a
## a .git suffix

github_url_repo_rx <- function() "(?<repo>[^/@#]+?)"

github_url_rx <- function() {
  paste0(
    "^",
    ## Optional package name
    "(?:(?<package>",
    package_name_rx(),
    ")=)?",
    ## Optional remote type
    "(?:github::)?",
    ## Protocol
    "(?:(?:https?://)|(?:(?:ssh://|[^@]+@)))",
    ## Servername
    ## TODO: should probably restrict this is (configurable) GH servers
    "(?:[^/:]+)[/:]",
    ## Username
    github_username_rx(),
    "/",
    ## Repo
    github_url_repo_rx(),
    ## subdir, always empty
    "(?<subdir>)",
    ## Optional Extension
    "(?:[.]git)?",
    ## Commit / PR / Release
    github_url_detail_rx(),
    "$"
  )
}

remote_type_rx <- function() {
  paste0(
    "^",
    ## Optional package name
    "(?:(?<package>",
    package_name_rx(),
    ")=)?",
    ## Remote type
    "(?:(?<type>[-_[:alnum:]]+)::)?",
    ## Rest of ref
    "(?<rest>.*)$"
  )
}

local_rx <- function() {
  typed <- "local::(?<path>.*)"
  sugar <- "(?<path>(?:/|\\\\|~|[.]/|[.]\\\\|[.]$).*)"
  paste0(
    "^",
    ## Optional package name
    "(?:(?<package>",
    package_name_rx(),
    ")=)?",
    "(?|",
    typed,
    "|",
    sugar,
    ")",
    "$"
  )
}

type_default_parse <- function(refs, ...) {
  m <- re_match(refs, remote_type_rx())
  lapply_rows(
    m,
    function(x) {
      list(package = x$package, type = x$type, rest = x$rest, ref = x$.text)
    }
  )
}

get_remote_types <- function(refs) {
  m <- re_match(refs, remote_type_rx())
  types <- m$type

  types[types == "" & grepl(standard_rx(), refs, perl = TRUE)] <- "standard"
  types[types == "" & grepl(github_rx(), refs, perl = TRUE)] <- "github"
  types[types == "" & grepl(github_url_rx(), refs, perl = TRUE)] <- "github"
  types[types == "" & grepl(local_rx(), refs, perl = TRUE)] <- "local"
  types[types == "" & grepl(param_rx(), refs, perl = TRUE)] <- "param"
  types[is.na(types)] <- ""

  if (any(bad <- types == "")) {
    throw(pkg_error(
      "Cannot parse package{?s}: {.pkg {refs[bad]}}.",
      i = msg_package_sources()
    ))
  }

  types
}

#' Parse package location references
#'
#' See [pkg_refs] for more about supported package references.
#'
#' @param refs Character vector of references.
#' @param remote_types Custom remote types can be added here, this is
#'   for advanced use, and experimental currently.
#' @param ... Additional arguments are passed to the individual parser
#'   functions.
#' @return `parse_pkg_refs()` returns a list of parsed references.
#' `parse_pkg_ref()` returns one parsed reference. A parsed reference is
#' a list, with at least elements:
#' - `ref`: The original reference string.
#' - `type`: The reference type.
#' - `package`: The package name.
#' It typically contains additional data, specific to the various
#' reference types. See [pkg_refs] for details.
#' The parsed reference always has class `remote_ref_<type>` and
#' `remote_ref`.
#'
#' @export

parse_pkg_refs <- function(refs, remote_types = NULL, ...) {
  remote_types <- c(default_remote_types(), remote_types)
  params <- parse_ref_params(refs)
  refs <- params$refs
  types <- get_remote_types(refs)
  unique_types <- unique(types)
  res <- vector("list", length(refs))

  if (length(bad <- setdiff(unique_types, names(remote_types)))) {
    throw(pkg_error(
      "Unknown package source{?s}: {.val {bad}}.",
      i = msg_package_sources()
    ))
  }

  for (this in unique_types) {
    parser <- remote_types[[this]]$parse %||% type_default_parse
    this_refs <- refs[types == this]
    new_remotes <- parser(this_refs, ...)
    new_remotes <- lapply(new_remotes, function(x) {
      x$type <- this
      x
    })
    new_remotes <- lapply(
      new_remotes,
      add_class,
      c(paste0("remote_ref_", this), "remote_ref")
    )
    res[types == this] <- new_remotes
  }

  add_ref_params(res, params$params)
}

#' @param ref A package reference, like `refs`, but a length one vector,
#' for convenience.
#' @export
#' @rdname parse_pkg_refs

parse_pkg_ref <- function(ref, remote_types = NULL, ...) {
  assert_that(is_string(ref))
  parse_pkg_refs(ref, remote_types = remote_types, ...)[[1]]
}

# TODO: allow omitting the package name: `?nocache` and then it applies
# to every package

param_rx <- function() {
  paste0(
    "(?:(?<package>",
    package_name_rx(),
    "|[*])=)",
    "$"
  )
}

parse_ref_params <- function(refs) {
  list(
    refs = sub("[?].*$", "", refs),
    params = lapply(refs, parse_query)
  )
}

add_ref_params <- function(res, params) {
  if (length(res) != length(params)) {
    throw(pkg_error(
      "Internal error in the parameter parser, parameter length mismath.",
      i = msg_internal_error()
    ))
  }
  for (i in seq_along(res)) {
    res[[i]]$params <- if (length(params[[i]])) params[[i]] else character()
  }

  res
}

known_query_params <- c(
  "ignore",
  "ignore-before-r",
  "ignore-build-errors",
  "ignore-unavailable",
  "nocache",
  "reinstall",
  "source"
)

parse_query <- function(ref) {
  query <- sub("^[^?]*(\\?|$)", "", ref)
  query <- sub("^[?]", "", query)
  query <- chartr("+", " ", query)
  argstr <- strsplit(query, "&", fixed = TRUE)[[1]]
  argstr <- strsplit(argstr, "=", fixed = TRUE)
  keys <- vcapply(argstr, function(x) utils::URLdecode(x[[1]]))
  vals <- vcapply(argstr, function(x) {
    if (length(x) == 2) utils::URLdecode(x[[2]]) else ""
  })

  if (length(bad <- unique(setdiff(keys, known_query_params)))) {
    cli::cli_alert_warning(c(
      "Unknown package{cli::qty(bad)} parameter{?s}: ",
      "{.val {bad}} in {.val {ref}}."
    ))
  }

  structure(vals, names = keys)
}

is_true_param <- function(params, which) {
  which %in%
    names(params) &&
    tolower(params[[which]]) %in% c("", "true", "yes", "y", "on", "1")
}

get_param_value <- function(params, which) {
  if (which %in% names(params)) params[[which]] else NA_character_
}

test_that("package_name_rx", {
  good <- c("A1", "a1", "Z1", "z1", "foo.bar", "foo.bar.baz", "a1.b2")

  for (t in good) {
    expect_true(is_valid_package_name(t), info = t)
  }

  bad <- list(
    c("pkg", "forbidden"),
    c("pak\u00e1ge", "ASCII"),
    c("good-package", "letters, numbers and dot"),
    c("x", "two characters"),
    c("1stpackage", "must start with a letter"),
    c("dots.", "must not end with a dot")
  )

  for (t in bad) {
    ans <- is_valid_package_name(t[1])
    expect_false(ans, info = t[1])
    expect_match(attr(ans, "reason"), t[2], info = t[1])
  }
})

test_that("parse_pkg_refs, standard", {
  cases <- list(
    list("pkg", list(package = "pkg", atleast = "", version = "")),
    list("pkg@0.1-2", list(package = "pkg", atleast = "==", version = "0.1-2")),
    list("pkg@>=2.9", list(package = "pkg", atleast = ">=", version = "2.9")),
    list("pkg@last", list(package = "pkg", atleast = "==", version = "last")),
    list("standard::pkg", list(package = "pkg", atleast = "", version = "")),
    list(
      "standard::pkg@0.1-2",
      list(package = "pkg", atleast = "==", version = "0.1-2")
    )
  )

  expect_equal(
    get_remote_types(vcapply(cases, "[[", 1)),
    rep("standard", length(cases))
  )

  for (case in cases) {
    expect_equal_named_lists(
      p <- parse_pkg_refs(case[[1]])[[1]],
      c(
        case[[2]],
        list(ref = case[[1]], type = "standard", params = character())
      )
    )
    expect_s3_class(p, c("remote_ref_cran", "remote_ref"))
  }
})

test_that("parse_pkg_refs, cran", {
  cases <- list(
    list("cran::pkg", list(package = "pkg", atleast = "", version = "")),
    list(
      "cran::pkg@0.1-2",
      list(package = "pkg", atleast = "==", version = "0.1-2")
    )
  )

  expect_equal(
    get_remote_types(vcapply(cases, "[[", 1)),
    rep("cran", length(cases))
  )

  for (case in cases) {
    expect_equal_named_lists(
      p <- parse_pkg_refs(case[[1]])[[1]],
      c(case[[2]], list(ref = case[[1]], type = "cran", params = character()))
    )
    expect_s3_class(p, c("remote_ref_cran", "remote_ref"))
  }
})

test_that("github regexes", {
  username <- list(
    list("foobar", "foobar"),
    list("", NA_character_),
    list("-bad", NA_character_),
    list(
      "123456789012345678901234567890123456789",
      "123456789012345678901234567890123456789"
    ),
    list("1234567890123456789012345678901234567890", NA_character_)
  )
  for (c in username) {
    rx <- paste0("^", github_username_rx(), "$")
    expect_equal(
      re_match(c[[1]], rx)$username,
      c[[2]]
    )
  }

  commitish <- list(
    list("", NA_character_),
    list("x", NA_character_),
    list("foobar", NA_character_),
    list("@foobar", "foobar"),
    list("@*foobar", NA_character_)
  )
  for (c in commitish) {
    expect_equal(
      re_match(c[[1]], github_commitish_rx())$commitish,
      c[[2]]
    )
  }

  pull <- list(
    list("", NA_character_),
    list("x", NA_character_),
    list("@foobar", NA_character_),
    list("#", NA_character_),
    list("#123", "123"),
    list("#1", "1"),
    list("#foobar", NA_character_),
    list("#1foo", "1")
  )
  for (c in pull) {
    expect_equal(
      re_match(c[[1]], github_pull_rx())$pull,
      c[[2]]
    )
  }

  release <- list(
    list("", NA_character_),
    list("x", NA_character_),
    list("@foobar", NA_character_),
    list("@*foobar", NA_character_),
    list("@*release", "*release")
  )
  for (c in release) {
    expect_equal(
      re_match(c[[1]], github_release_rx())$release,
      c[[2]]
    )
  }

  detail <- list(
    list("@foobar", c("foobar", "", "")),
    list("#123", c("", "123", "")),
    list("@*release", c("", "", "*release")),
    list("foobar", c("", "", ""))
  )
  for (c in detail) {
    expect_equal(
      unlist(re_match(
        c[[1]],
        github_detail_rx()
      )[, c("commitish", "pull", "release")]),
      structure(c[[2]], names = c("commitish", "pull", "release"))
    )
  }
})

test_that("parse_pkg_refs error on unknown type", {
  expect_snapshot(
    error = TRUE,
    parse_pkg_refs(c("notgood::pkg", "good", "my_package"))
  )
})

test_that("custom remote types", {
  xspecs <- NULL
  xargs <- NULL
  parse_remote_foo <- function(specs, ...) {
    xspecs <<- specs
    xargs <<- list(...)
    list(list())
  }
  res <- withr::with_options(
    list(pkg.remote_types = list(foo = list(parse = parse_remote_foo))),
    parse_pkg_refs("foo::arbitrary_string/xxx", ex1 = "1", ex2 = "2")
  )
  expect_identical(
    res,
    list(structure(
      list(type = "foo", params = character()),
      class = c("remote_ref_foo", "remote_ref", "list")
    ))
  )
  expect_identical(xspecs, "foo::arbitrary_string/xxx")
  expect_identical(xargs, list(ex1 = "1", ex2 = "2"))

  res2 <- parse_pkg_refs(
    "foo::arbitrary_string/xxx",
    ex1 = "1",
    ex2 = "2",
    remote_types = list(foo = list(parse = parse_remote_foo))
  )
  expect_identical(res, res2)
})

test_that("type_default_parse", {
  res <- type_default_parse(c("foo::bar", "package=foo2::bar2"))
  expect_identical(
    res,
    list(
      list(package = "", type = "foo", rest = "bar", ref = "foo::bar"),
      list(
        package = "package",
        type = "foo2",
        rest = "bar2",
        ref = "package=foo2::bar2"
      )
    )
  )
})

test_that("default parse function", {
  res <- withr::with_options(
    list(pkg.remote_types = list(foo = list(), foo2 = list())),
    parse_pkg_refs(c("foo::bar", "package=foo2::bar2"))
  )
  expect_identical(
    res,
    list(
      structure(
        list(
          package = "",
          type = "foo",
          rest = "bar",
          ref = "foo::bar",
          params = character()
        ),
        class = c("remote_ref_foo", "remote_ref", "list")
      ),
      structure(
        list(
          package = "package",
          type = "foo2",
          rest = "bar2",
          ref = "package=foo2::bar2",
          params = character()
        ),
        class = c("remote_ref_foo2", "remote_ref", "list")
      )
    )
  )

  res2 <- parse_pkg_refs(
    c("foo::bar", "package=foo2::bar2"),
    remote_types = list(foo = list(), foo2 = list())
  )
  expect_identical(res, res2)
})

test_that("parse_pkg_refs, local", {
  cases <- list(
    list("local::path", "path"),
    list("local::/path", "/path"),
    list("local::~/path", "~/path"),
    list("local::./path", "./path"),
    list("local::\\path", "\\path"),
    list("/path", "/path"),
    list("~/path", "~/path"),
    list("./path", "./path"),
    list(".\\path", ".\\path"),
    list("\\path", "\\path"),
    list(".", ".")
  )

  for (c in cases) {
    expect_equal(
      parse_pkg_ref(c[[1]]),
      structure(
        list(
          package = NA_character_,
          path = c[[2]],
          ref = paste0("local::", c[[2]]),
          type = "local",
          params = character()
        ),
        class = c("remote_ref_local", "remote_ref", "list")
      )
    )
  }
})

test_that("parse_query", {
  empty <- c(a = "1")[-1]
  cases <- list(
    list("", empty),
    list("?", empty),
    list("?foo", c(foo = "")),
    list("?foo=1", c(foo = "1")),
    list("?foo&bar", c(foo = "", bar = "")),
    list("?foo=1&bar=2&foo", c(foo = "1", bar = "2", foo = "")),
    list("?foo=1%202&bar=x", c(foo = "1 2", bar = "x"))
  )

  for (c in cases) {
    suppressMessages(expect_equal(parse_query(c[[1]]), c[[2]]))
  }
})

test_that("parameters", {
  cases <- list(
    list("foo", "?bar", c(bar = "")),
    list("foo", "?bar=1&foo&bar=11", c(bar = "1", foo = "", bar = "11")),
    list("user/repo", "?source", c(source = ""))
  )

  for (c in cases) {
    wo <- parse_pkg_ref(c[[1]])
    suppressMessages(wi <- parse_pkg_ref(paste0(c[[1]], c[[2]])))
    wo$params <- c[[3]]
    expect_equal(wi, wo)
  }
})

test_that("explicit package names", {
  expect_snapshot({
    parse_pkg_ref("package=user/notpackage")
    parse_pkg_ref("package=user/notpackage@tag")
    parse_pkg_ref("package=github::user/notpackage@tag")

    parse_pkg_ref("package=local::/abs/path")
    parse_pkg_ref("package=local::rel/path")
    parse_pkg_ref("package=local::~/home/path")
    parse_pkg_ref("package=/abs/path")
    parse_pkg_ref("package=./rel/path")
    parse_pkg_ref("package=~/home/path")

    parse_pkg_ref("package=url::https://example.com/p1.tar.gz")
  })
})

test_that("gitlab", {
  expect_snapshot({
    parse_pkg_ref("gitlab::user/project")
    parse_pkg_ref("gitlab::user/project@ref")
    parse_pkg_ref("gitlab::user/project/-/sub/dir")
    parse_pkg_ref("gitlab::user/project/-/sub/dir@ref")
    parse_pkg_ref("gitlab::group/subgroup/project")
    parse_pkg_ref("gitlab::group/subgroup/project@ref")
    parse_pkg_ref("gitlab::group/subgroup/project/-/sub/dir")
    parse_pkg_ref("gitlab::group/subgroup/project/-/sub/dir@ref")
    parse_pkg_ref(
      "gitlab::https://acme.co/group/subgroup/project/-/sub/dir@ref"
    )

    parse_pkg_ref("pkg=gitlab::user/project")
    parse_pkg_ref("pkg=gitlab::user/project@ref")
    parse_pkg_ref("pkg=gitlab::user/project/-/sub/dir")
    parse_pkg_ref("pkg=gitlab::user/project/-/sub/dir@ref")
    parse_pkg_ref("pkg=gitlab::group/subgroup/project")
    parse_pkg_ref("pkg=gitlab::group/subgroup/project@ref")
    parse_pkg_ref("pkg=gitlab::group/subgroup/project/-/sub/dir")
    parse_pkg_ref("pkg=gitlab::group/subgroup/project/-/sub/dir@ref")
    parse_pkg_ref(
      "pkg=gitlab::https://acme.co/group/subgroup/project/-/sub/dir@ref"
    )
  })
})

#' @param refs Package names or references. See
#'   ['Package references'][pkg_refs] for the syntax.
#' @param ... Additional arguments, passed to
#'   [`pkg_download_proposal$new()`](#method-new).
#' @return `new_pkg_download_proposal()` returns a new
#'   `pkg_download_proposal` object.
#'
#' @details
#' `new_pkg_download_proposal()` creates a new object from the
#' `pkg_download_proposal` class, that can be used to look up and download
#' R packages and their dependencies. The advantage of
#' `new_pkg_download_proposal()` compared to using the
#' [pkg_download_proposal] constructor directly is that it avoids making
#' pkgdepends a build time dependency.
#'
#' @export
#' @rdname pkg_download_proposal

new_pkg_download_proposal <- function(refs, ...) {
  pkg_download_proposal$new(refs, ...)
}

#' R6 class for package downloads
#'
#' @description
#' Download packages with their dependencies, from various sources.
#'
#' @details
#' Typical workflow to download a set of packages:
#'
#' 1. Create a `pkg_download_proposal` object with
#'    `new_pkg_download_proposal()`.
#' 1. Resolve all possible dependencies with
#'    [`pkg_download_proposal$resolve()`](#method-resolve).
#' 1. Download all files with
#'    [`pkg_download_proposal$download()`](#method-download).
#' 1. Get the data about the packages and downloads with
#'    [`pkg_download_proposal$get_downloads()`](#method-get-downloads).
#'
#' @export

pkg_download_proposal <- R6::R6Class(
  "pkg_download_proposal",
  public = list(
    #' @description
    #' Create a new `pkg_download_proposal` object. Consider using
    #' `new_pkg_download_proposal()` instead of calling the constructor
    #' directly.
    #'
    #' The returned object can be used to look up (recursive) dependencies
    #' of R packages from various sources, and then to download the
    #' package files.
    #'
    #' @param refs Package names or references. See
    #'   ['Package references'][pkg_refs] for the syntax.
    #' @param config Configuration options, a named list. See
    #'   ['Configuration'][pkgdepends-config].
    #' @param remote_types Custom remote ref types, this is for advanced
    #'   use, and experimental currently.
    #'
    #' ## Examples
    #' ```r
    #' pdl <- pkg_download_proposal$new("r-lib/pkgdepends")
    #' pdl
    #' ```

    initialize = function(refs, config = list(), remote_types = NULL) {
      private$plan <- pkg_plan$new(refs, config, library = NULL, remote_types)
      invisible(self)
    },

    #' @description
    #' The package refs that were used to create the
    #' `pkg_download_proposal` object.
    #'
    #' @return
    #' A character vector of package refs that were used to create the
    #' `pkg_download_proposal` object.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method get_refs()
    #' pdl <- new_pkg_download_proposal(c("pak", "jsonlite"))
    #' pdl$get_refs()

    get_refs = function() private$plan$get_refs(),

    #' @description
    #' Configuration options for the `pkg_download_proposal` object. See
    #' ['Configuration'][pkgdepends-config] for details.
    #'
    #' @return
    #' Named list. See ['Configuration'][pkgdepends-config] for the configuration
    #' options.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method get_config()
    #' pdl <- new_pkg_download_proposal("pak")
    #' pdl$get_config()

    get_config = function() private$plan$get_config(),

    #' @description
    #' Resolve the dependencies of the specified package references. This
    #' usually means downloading metadata from CRAN and Bioconductor,
    #' unless already cached, and also from GitHub if GitHub refs were
    #' included, either directly or indirectly. See
    #' ['Dependency resolution'][pkg_resolution] for details.
    #'
    #' @return
    #' The `pkg_download_proposal` object itself, invisibly.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method resolve()
    #' pdl <- new_pkg_download_proposal("pak")
    #' pdl$resolve()
    #' pdl$get_resolution()

    resolve = function() {
      private$plan$resolve()
      invisible(self)
    },

    #' @description
    #' The same as [`resolve()`](#method-resolve), but asynchronous.
    #' This method is for advanced use.
    #'
    #' @return
    #' A deferred value.

    async_resolve = function() private$plan$async_resolve(),

    #' @description
    #' Query the result of the dependency resolution. This method can be
    #' called after [`resolve()`](#method-resolve) has completed.
    #'
    #' @return
    #' A [pkg_resolution_result] object, which is also a data frame. See
    #' ['Dependency resolution'][pkg_resolution] for its columns.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method get_resolution()
    #' pdl <- new_pkg_download_proposal("r-lib/pkgdepends")
    #' pdl$resolve()
    #' pdl$get_resolution()

    get_resolution = function() private$plan$get_resolution(),

    #' @description
    #' Download all resolved packages. It uses the package cache in the
    #' pkgcache package by default, to avoid downloads if possible.
    #'
    #' @return
    #' The `pkg_download_proposal` object, invisibly.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method download()
    #' pdl <- new_pkg_download_proposal("r-lib/pkgdepends")
    #' pdl$resolve()
    #' pdl$download()
    #' pdl$get_downloads()

    download = function() {
      private$plan$download_resolution()
      invisible(self)
    },

    #' @description
    #' The same as [`download()`](#method-download), but asynchronous.
    #' This method is for advanced use.
    #'
    #' @return
    #' A deferred value.

    async_download = function() private$plan$async_download_resolution(),

    #' @description
    #' Returns the summary of the package downloads.
    #'
    #' @return
    #' A [pkg_download_result] object, which is a list. See
    #' [pkg_download_result] for details.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method get_downloads()
    #' pdl <- new_pkg_download_proposal("pkgload")
    #' pdl$resolve()
    #' pdl$download()
    #' pdl$get_downloads()

    get_downloads = function() private$plan$get_resolution_download(),

    #' @description
    #' Throw and error if the some of the downloads have failed for the
    #' most recent
    #' [`pkg_download_proposal$download()`](#method-download) call.

    stop_for_download_error = function() {
      private$plan$stop_for_resolution_download_error()
    },

    #' @description
    #' Format a `pkg_download_proposal` object, typically for printing.
    #'
    #' @param ... not used currently.
    #' @return Nothing.
    #' A character vector, each element should be a line in the printout.

    format = function(...) {
      refs <- private$plan$get_refs()

      has_res <- private$plan$has_resolution()
      res <- if (has_res) private$plan$get_resolution()
      res_err <- has_res && any(res$status != "OK")

      has_dls <- private$plan$has_resolution_downloads()
      dls <- if (has_dls) private$plan$get_resolution_download()
      dls_err <- has_dls && any(dls$status == "Failed")

      deps <- if (has_res) length(unique(res$package[!res$direct]))

      c(
        "<pkg_download_proposal>",
        "+ refs:",
        paste0("  - ", refs),
        if (has_res) paste0("+ has resolution (+", deps, " dependencies)"),
        if (res_err) "x has resolution errors",
        if (has_dls) "+ has downloads",
        if (dls_err) "x has download errors",
        if (!has_res) "(use `$resolve()` to resolve dependencies)",
        if (has_res && !res_err && !has_dls) {
          "(use `$download()` to download packages)"
        },
        if (has_res) "(use `$get_resolution()` to see resolution results)",
        if (has_dls) "(use `$get_downloads()` to get download data)"
      )
    },

    #' @description
    #' Prints a `pkg_download_proposal` object to the screen. The printout
    #' includes:
    #'
    #' * The package refs.
    #' * Whether the object has the resolved dependencies.
    #' * Whether the resolution had errors.
    #' * Whether the downloads were completed.
    #' * Whether the downloads had errors.
    #' * Advice on which methods to call next.
    #'
    #' See the example below.
    #'
    #' @param ... not used currently.
    #' @return
    #' The `pkg_download_proposal` object itself, invisibly.
    #'
    #' @examplesIf pkgdepends:::is_online()
    #' # Method print()
    #' pdl <- new_pkg_download_proposal("r-lib/pkgdepends")
    #' pdl
    #'
    #' pdl$resolve()
    #' pdl
    #'
    #' pdl$download()
    #' pdl

    print = function(...) cat(self$format(...), sep = "\n")
  ),

  private = list(
    plan = NULL
  )
)

test_that("new_pkg_download_proposal", {
  setup_fake_apps()

  dl <- new_pkg_download_proposal(
    "pkg3",
    config = list(library = tempfile())
  )
  expect_snapshot(dl)
  expect_snapshot(dl$get_refs())
  expect_snapshot(sort(dl$get_config()$list()))

  suppressMessages(dl$resolve())
  expect_snapshot(dl)
  expect_snapshot(
    dl$get_resolution()[, c("ref", "type", "directpkg", "package", "error")]
  )
})

test_that("async_resolve", {
  setup_fake_apps()

  dl <- new_pkg_download_proposal(
    "pkg3",
    config = list(library = tempfile())
  )
  expect_snapshot(
    synchronize(
      dl$async_resolve()$then(function() "done")
    )
  )
  expect_snapshot(
    dl$get_resolution()[, c("ref", "type", "directpkg", "package", "error")]
  )
})

test_that("download", {
  setup_fake_apps()

  tmp <- withr::local_tempdir()
  dl <- new_pkg_download_proposal(
    "pkg3",
    config = list(library = tempfile(), cache_dir = tmp)
  )

  suppressMessages(dl$resolve())
  suppressMessages(dl$download())

  expect_snapshot(dir(tmp, recursive = TRUE, pattern = "[.]tar[.]gz$"))
})

test_that("async_download", {
  setup_fake_apps()

  tmp <- withr::local_tempdir()
  dl <- new_pkg_download_proposal(
    "pkg3",
    config = list(library = tempfile(), cache_dir = tmp)
  )

  suppressMessages(dl$resolve())
  suppressMessages(synchronize(dl$async_download()))

  expect_snapshot(dir(tmp, recursive = TRUE, pattern = "[.]tar[.]gz$"))

  expect_snapshot(dl$get_downloads()$target)

  expect_silent(dl$stop_for_download_error())
})

#' git protocol notes, for developers
#'
#' Assumptions, they might be relaxed or checked for later:
#' - The server must speak the smart protocol, version 1 or 2.
#'   (Although I added some functions specifically for the dumb protocol,
#'   but they are pretty limited.)
#' - We use HTTP transport, not SSH.
#' - The server should have the `shallow` capability.
#' - The server should have the `filter` capability if protocol version 2.
#' - Only SHA-1 hashing is supported.
#'
#' Improvements needed:
#' - DONE Tests. (Can always have more.)
#' - DONE Use async HTTP.
#' - DONE Support packfiles with deltas. (`ofs-delta` objects are still
#'   not supported.)
#' - DONE Optionally send authorization. Already possibly in the URL.
#' - DONE Better error messages.
#' - Better errors for non-existing user, repository, ref, PR, etc.
#'
#' Optional improvements:
#' - Support `ofs-delta` objects in packfiles. Not necessarily, unless we
#'   send this capability, the server is not sending `ofs-delta` objects.
#' - Make unpacking faster. It is not fast currently, with all the bit
#'   arithmetic in R. But it is already faster than a `tar.gz` + uncompress
#'   download from GitHub, so not really needed.
#'
#' ## Docs and other helpful links:
#' - <https://github.com/git/git/blob/master/Documentation/gitprotocol-common.adoc>
#' - <https://github.com/git/git/blob/master/Documentation/gitprotocol-pack.adoc>
#' - <https://github.com/git/git/blob/master/Documentation/gitprotocol-v2.adoc>
#' - <https://github.com/calebsander/git-internals/blob/part2/src/main.rs>
#' - <https://dev.to/calebsander/git-internals-part-1-the-git-object-model-474m>
#'
#' @keywords internal
#' @name git-protocol
NULL

# -------------------------------------------------------------------------

git_creds_for_url <- function(url) {
  creds <- tryCatch(
    gitcreds_get(url)[c("username", "password")],
    error = function(e) NULL
  )
  if (is.null(creds)) {
    do.call(
      Sys.setenv,
      structure(list("FAIL"), names = gitcreds_cache_envvar(url))
    )
  }
  creds
}

git_http_get <- function(url, options = list(), ...) {
  options <- c(options, git_creds_for_url(url))
  http_get(url, options = options, ...)
}

git_http_post <- function(url, options = list(), ...) {
  options <- c(options, git_creds_for_url(url))
  http_post(url, options = options, ...)
}

#' List references in a remote git repository
#'
#' @details
#' * Branches have references named `refs/heads/<branch>`, e.g.
#' `refs/heads/main`.
#' * Tags have references named `refs/tags/<tag>`, e.g. `refs/tags/v1.0.2`.
#' * On GitHub pull requests have references named
#'   `refs/pull/<pr-number>/head`, e.g. `refs/pull/37/head`.
#'   For open pull requests there should be a `refs/pull/<pr-number>/merge`
#'   reference as well, which is the branch after the pull request has been
#'   merged.
#' * There is a special reference called `HEAD`. This points to the default
#'   branch on GitHub.
#'
#' @param url Repository URL, e.g. `https://github.com/r-lib/pak.git`.
#'   It might include authentication information, e.g. a GitHub token.
#' @param prefixes If not `NULL`, then only references of which one of
#'   `prefixes` is a prefix, are listed.
#' @return A list with entries: `refs` and `caps`.
#'   `caps` is a character vector of capabilities advertised by the server.
#'   `refs` is a data frame of git refs, it has columns `ref` and `hash`.
#'
#' @keywords internal
#' @section Examples:
#'
#' ```{r, git-list-refs, cache = TRUE}
#' git_list_refs("https://github.com/r-lib/filelock.git")
#' ```
#'
#' List only references having a certain prefix:
#'
#' ```{r git-list-refs-2, cache = TRUE}
#' git_list_refs("https://github.com/r-lib/filelock.git", "refs/heads/main")
#' ```
#'
#' Various services:
#' ```{r git-list-refs-2, cache = TRUE}
#' git_list_refs("https://gitlab.com/Linaro/tuxmake.git", "HEAD")
#' git_list_refs("https://bitbucket.org/gaborcsardi/cli.git")
#' git_list_refs("https://git.savannah.nongnu.org/git/administration/savane.git")
#' ```

git_list_refs <- function(url, prefixes = NULL) {
  synchronize(async_git_list_refs(url, prefixes))
}

async_git_list_refs <- function(url, prefixes = NULL) {
  async_git_list_refs_v2(url, prefixes)
}

# -------------------------------------------------------------------------

#' List files in a remote git repository
#'
#' @inheritParams git_list_refs
#' @param ref Either a SHA or a ref name. It may also be a branch name
#'   without the `refs/heads` prefix, or a partial (but unique) SHA of
#'   at least seven hexadecimal digits.
#'   See [git_list_refs()] for how branches, tags and GitHub pull
#'   requests are named.
#' @return A list with entries:
#'   * `ref`: The `ref` the function was called with.
#'   * `sha`: SHA of `ref`.
#'   * `commit`: named character vector of the data of the commit object
#'      belonging to `ref`. It has fields named: `tree`, `parent`,
#'      `author`, `committer`, `encoding`, `message`.
#'   * `tree`: SHA of the `tree` object belonging to `commit`.
#'   * `files`: a data frame of files and directories. It has columns:
#'        - `hash`: SHA hash of the tree or blob.
#'        - `type`: Either `"tree"` or `"blob"`.
#'        - `mode`: Unix mode.
#'        - `path`: Relative path from the repository's root directory.
#'
#' @keywords internal
#' @section Examples:
#' ```{r git-list-files, cache = TRUE}
#' fls <- git_list_files("https://github.com/r-lib/pak.git", "refs/heads/main")
#' fls$files
#' ```

git_list_files <- function(url, ref = "HEAD") {
  synchronize(async_git_list_files(url, ref))
}

async_git_resolve_ref <- function(url, ref) {
  url
  ref
  sha <- ref

  if (!grepl("^[0-9a-f]{40}$", ref)) {
    # Only use 'ref' as a filter if it is not a sha prefix of at least 7 chars
    filt <- if (nchar(ref) < 7 || !grepl("^[0-9a-f]+$", ref)) {
      paste0(c("", "refs/heads/", "refs/tags/"), ref)
    }
    async_git_list_refs(url, filt)$catch(
      error = function(e) async_git_list_refs_v1(url)
    )$then(function(refs) {
      result <- if (ref %in% refs$refs$ref) {
        refs$refs$hash[refs$refs$ref == ref]
      } else if (paste0("refs/tags/", ref) %in% refs$refs$ref) {
        refs$refs$hash[refs$refs$ref == paste0("refs/tags/", ref)]
      } else if (paste0("refs/heads/", ref) %in% refs$refs$ref) {
        refs$refs$hash[refs$refs$ref == paste0("refs/heads/", ref)]
      } else if (any(startsWith(refs$refs$hash, ref))) {
        sha <- unique(refs$refs$hash[startsWith(refs$refs$hash, ref)])
        if (length(sha) > 1) {
          throw(pkg_error(
            "Found multiple git refs with prefix {.val {ref}}, it is ambiguous.",
            "i" = "Matching git refs: {.val {sha}}.",
            "i" = "Specify a longer prefix to choose a single git ref."
          ))
        }
        sha
      } else {
        throw(pkg_error(
          "Unknown git ref: {.val {ref}}.",
          .class = "git_proto_error_unknown_ref",
          .data = list(ref = ref, url = redact_url(url))
        ))
      }

      attr(result, "protocol") <- if ("version 2" %in% refs$caps) 2 else 1
      attr(result, "filter") <- any(grepl("\\bfilter\\b", refs$caps))
      result
    })
  } else {
    async_constant(ref)
  }
}

async_git_list_files <- function(url, ref = "HEAD") {
  url
  ref
  sha2 <- ref
  async_git_resolve_ref(url, ref)$then(function(sha) {
    sha2 <<- sha
    async_git_fetch(url, sha)
  })$then(function(pf) async_git_list_files_process(pf, ref, sha2, url))
}

async_git_list_files_process <- function(packfile, ref, sha, url) {
  names(packfile) <- vcapply(packfile, "[[", "hash")
  types <- unname(vcapply(packfile, "[[", "type"))
  tree_sizes <- viapply(packfile, function(x) nrow(x$object) %||% NA_integer_)
  num_files <- sum(tree_sizes, na.rm = TRUE)

  files = data_frame(
    hash = character(num_files),
    type = character(num_files),
    mode = character(num_files),
    path = character(num_files)
  )

  trees <- packfile[types == "tree"]
  done <- logical(length(trees))
  idx <- 1L
  wd <- character()

  commit <- parse_commit(packfile[[which(types == "commit")]]$object)
  tree <- commit[["tree"]]

  process_tree <- function(i) {
    if (done[i]) {
      return()
    }
    done[i] <<- TRUE
    tr <- trees[[i]]$object
    for (l in seq_len(nrow(tr))) {
      files$hash[idx] <<- tr$hash[l]
      files$type[idx] <<- tr$type[l]
      files$mode[idx] <<- tr$mode[l]
      files$path[idx] <<- paste(c(wd, tr$path[l]), collapse = "/")
      idx <<- idx + 1L
      if (tr$type[l] == "tree") {
        tidx <- which(tr$hash[l] == names(trees))[1]
        if (is.na(tidx)) {
          next
        } # nocov
        wd <<- c(wd, tr$path[l])
        process_tree(tidx)
        wd <<- utils::head(wd, -1)
      }
    }
  }

  # start with the root tree
  root <- match(tree, names(trees))
  if (is.na(root)) {
    throw(pkg_error(
      "Invalid git response from {.url {url}}, cannot find commit tree"
    ))
  }
  process_tree(root)
  if (any(!done)) {
    warning(
      "Some trees are unreachable when listing files from git repo from ",
      url
    )
  }

  list(
    ref = ref,
    sha = sha,
    commit = commit,
    tree = tree,
    files = files
  )
}

# -------------------------------------------------------------------------

#' Download a blob for a remote git repository
#'
#' @inheritParams git_list_refs
#' @param sha SHA hash of the blob.
#' @param output Path where the blob will be written. It's directory is
#'   created if it does not exist.
#' @return A list that corresponds to a git packfile entry, invisibly.
#'   It has entries: `type` (always `"blob"`), `object` (raw object, the
#'   blob itself), `size`, `packed_size`, `hash`.
#'
#' @keywords internal
#' @section Examples:
#' Download a `DESCRIPTION` file from GitHub:
#' ```{r git-download-file, cache = TRUE}
#' pak_repo <- "https://github.com/r-lib/pak.git"
#' fls <- git_list_files(pak_repo, "HEAD")
#' git_download_file(
#'   pak_repo,
#'   fls$files$hash[fls$files$path == "DESCRIPTION"],
#'   output = tmp <- tempfile()
#' )
#' readLines(tmp)[1:5]
#' ```

git_download_file <- function(url, sha, output = sha) {
  synchronize(async_git_download_file(url, sha, output))
}

async_git_download_file <- function(url, sha, output = sha) {
  url
  sha
  output
  async_git_fetch(url, sha)$then(function(packfile) {
    if (length(packfile) != 1 || packfile[[1]]$type != "blob") {
      # nocov start
      throw(pkg_error(
        "Invalid response from git server, packfile should have a single blob.",
        .class = "git_proto_error_unexpected_response",
        .data = list(url = redact_url(url), sha = sha)
      ))
      # nocov end
    }
    if (!is.null(output)) {
      mkdirp(dirname(output))
      writeBin(packfile[[1]]$raw, output)
    }
    invisible(packfile[[1]])
  })
}

# -------------------------------------------------------------------------

#' Get a packfile for an object from a remote git repository
#'
#' @details
#' It uses `filter blob:none`, so extra blobs are not downloaded, only
#' trees.
#'
#' @inheritParams git_list_refs
#' @param sha SHA of the object to get.
#' @param blobs Whether we want to download all blobs as well or not.
#'   We are still always requesting a shallow (depth = 1) clone.
#' @return A list of git objects. Each element has entries:
#'   * `type`: `commit`, `tree`, `blob` or `tag`.
#'   * `object`: the object itself. It is a
#'       - character scalar for commits,
#'       - a data frame for trees, with columns: `type`, `mode`, `path`, `hash`,
#'       - a raw vector for blobs,
#'       - a raw vector for tags. (We could probably do better here.)
#'
#' @keywords internal
#' @section Examples:
#'
#' ```{r, git-fetch, cache = TRUE}
#' ft <- git_fetch(
#'   "https://github.com/r-lib/filelock.git",
#'   "9fdba75a62facaa3e818902f58891166e45eabe9"
#' )
#' ft
#' ```

git_fetch <- function(url, sha, blobs = FALSE) {
  synchronize(async_git_fetch(url, sha, blobs))
}

async_git_fetch <- function(url, sha, blobs = FALSE) {
  url
  sha

  if (!is.null(attr(sha, "protocol"))) {
    if (attr(sha, "protocol") == 1) {
      async_git_fetch_v1(url, sha, blobs)
    } else {
      async_git_fetch_v2(url, sha, blobs)
    }
  } else {
    async_git_fetch_v2(url, sha, blobs)$catch(
      error = function(err) async_git_fetch_v1(url, sha, blobs)
    )
  }
}

async_git_fetch_v1 <- function(url, sha, blobs) {
  async_git_send_message_v1(
    url,
    args = c(
      paste0("want ", sha),
      "deepen 1",
      "",
      "done",
      ""
    ),
    caps = c("multi_ack", "no-done", "no-progress", paste0("agent=", git_ua()))
  )$then(function(reply) git_fetch_process_v1(reply, url, sha))
}

async_git_fetch_v2 <- function(url, sha, blobs) {
  # If 'filter' is not supported, then we need to get the blobs
  if (!is.null(attr(sha, "filter")) && !attr(sha, "filter")) {
    blobs <- TRUE
  }
  async_git_send_message_v2(
    url,
    "fetch",
    caps = c(
      paste0("agent=", git_ua(), "\n"),
      "object-format=sha1\n"
    ),
    args = c(
      "deepen 1\n",
      "no-progress\n",
      if (!blobs) "filter blob:none\n",
      paste0("want ", sha, "\n"),
      paste0("want ", sha, "\n"),
      "done\n"
    )
  )$then(function(reply) git_fetch_process(reply, url, sha))
}

git_fetch_process_v1 <- function(reply, url, sha) {
  if (length(reply) == 0) {
    throw(pkg_error(
      "Empty reply from git server (protocol v1) at {.url {redact_url(url)}}."
    ))
  }
  if (reply[[length(reply)]]$type != "pack") {
    throw(pkg_error(
      "No PACK in git server response (protocol v1) from {.url {redact_url(url)}}."
    ))
  }

  git_unpack(reply[[length(reply)]]$data)
}

git_fetch_process <- function(reply, url, sha) {
  # https://github.com/git/git/blob/7c2ef319c52c4997256f5807564523dfd4acdfc7/Documentation/gitprotocol-v2.txt#L240
  #
  #     output = acknowledgements flush-pkt |
  # 	     [acknowledgments delim-pkt] [shallow-info delim-pkt]
  # 	     [wanted-refs delim-pkt] [packfile-uris delim-pkt]
  # 	     packfile flush-pkt
  #
  #     acknowledgments = PKT-LINE("acknowledgments" LF)
  # 		      (nak | *ack)
  # 		      (ready)
  #     ready = PKT-LINE("ready" LF)
  #     nak = PKT-LINE("NAK" LF)
  #     ack = PKT-LINE("ACK" SP obj-id LF)
  #
  #     shallow-info = PKT-LINE("shallow-info" LF)
  # 		   *PKT-LINE((shallow | unshallow) LF)
  #     shallow = "shallow" SP obj-id
  #     unshallow = "unshallow" SP obj-id
  #
  #     wanted-refs = PKT-LINE("wanted-refs" LF)
  # 		  *PKT-LINE(wanted-ref LF)
  #     wanted-ref = obj-id SP refname
  #
  #     packfile-uris = PKT-LINE("packfile-uris" LF) *packfile-uri
  #     packfile-uri = PKT-LINE(40*(HEXDIGIT) SP *%x20-ff LF)
  #
  #     packfile = PKT-LINE("packfile" LF)
  # 	       *PKT-LINE(%x01-03 *%x00-ff)
  #

  # Since we sent a 'done', there must be no acknowledgements section

  if (is.null(reply[[1]]$text) || reply[[1]]$text != "shallow-info") {
    # nocov start
    throw(pkg_error(
      "Expected {.code shallow-info} section from git server.",
      .class = "git_proto_error_unexpected_response",
      .data = list(url = redact_url(url), sha = sha)
    ))
    # nocov end
  }

  # There should be only one 'shallow' line I think
  idx <- 2L
  while (idx <= length(reply) && !identical(reply[[idx]]$text, "packfile")) {
    idx <- idx + 1L
  }
  idx <- idx + 1L

  if (idx > length(reply)) {
    # nocov start
    throw(pkg_error(
      "Response from git server does not have a {.code packgile} section.",
      .class = "git_proto_error_unexpected_response",
      .data = list(url = redact_url(url), sha = sha)
    ))
    # nocov end
  }

  # check closing flush-pkt
  if (reply[[length(reply)]]$type != "flush-pkt") {
    # nocov start
    throw(pkg_error(
      "Response from git server does not have a closing {.code flush-pkt}.",
      .class = "git_proto_error_unexpected_response",
      .data = list(url = redact_url(url), sha = sha)
    ))
    # nocov end
  }

  data <- reply[idx:(length(reply) - 1)]

  if (any(vcapply(data, "[[", "type") != "data-pkt")) {
    # nocov start
    throw(pkg_error(
      "Response from git server must contain {.code data-pkt} sections.",
      .class = "git_proto_error_unexpected_response",
      .data = list(url = redact_url(url), sha = sha)
    ))
    # nocov end
  }

  # drop progress messages, although we did not requrest them...
  stream <- viapply(data, function(x) as.integer(x$data[1]))
  if (any(stream == 3)) {
    # nocov start
    msg <- tryCatch(
      rawToChar(reply[[which(stream == 3)[1]]]$data[-1]),
      error = function(err) "error message not available"
    )
    throw(pkg_error(
      "Got fatal error from git server: {msg}.",
      .class = "git_proto_error_unexpected_response",
      .data = list(url = redact_url(url), sha = sha)
    ))
    # nocov end
  }
  data <- data[stream == 1L]

  packfile <- do.call("c", lapply(data, function(x) x$data[-1]))

  git_unpack(packfile)
}

# -------------------------------------------------------------------------

git_download_repo <- function(
  url,
  ref = "HEAD",
  output = ref,
  submodules = FALSE
) {
  synchronize(async_git_download_repo(url, ref, output, submodules))
}

async_git_download_repo <- function(
  url,
  ref = "HEAD",
  output = ref,
  submodules = FALSE
) {
  url
  ref
  async_git_resolve_ref(url, ref)$then(function(sha) {
    async_git_download_repo_sha(url, sha, output, submodules)
  })
}

async_git_download_repo_sha <- function(url, sha, output, submodules = FALSE) {
  url
  sha
  output
  p <- async_git_fetch(url, sha, blobs = TRUE)$then(
    function(packfile) unpack_packfile_repo(packfile, output, url)
  )
  if (!submodules) {
    p
  } else {
    p$then(function() async_update_git_submodules(output))
  }
}

unpack_packfile_repo <- function(parsed, output, url) {
  types <- unname(vcapply(parsed, "[[", "type"))
  trees <- parsed[types == "tree"]
  done <- logical(length(trees))
  idx <- 1L
  wd <- character()

  mkdirp(output)

  process_tree <- function(i) {
    if (done[i]) {
      return()
    }
    done[i] <<- TRUE
    tr <- trees[[i]]$object
    for (l in seq_len(nrow(tr))) {
      idx <<- idx + 1L
      opath <- file.path(output, paste(c(wd, tr$path[l]), collapse = "/"))
      if (tr$type[l] == "tree") {
        tidx <- which(tr$hash[l] == names(trees))[1]
        if (is.na(tidx)) {
          throw(new_error(
            # nocov
            "git tree missing from packfile: {.val {tr$hash[i]}}.", # nocov
            .class = "git_proto_error_unexpected_response" # nocov
          )) # nocov
        }
        wd <<- c(wd, tr$path[l])
        mkdirp(opath)
        process_tree(tidx)
        wd <<- utils::head(wd, -1)
      } else if (tr$type[l] == "blob") {
        # for submodules this is NULL
        if (!is.null(parsed[[tr$hash[l]]])) {
          writeBin(parsed[[tr$hash[l]]]$raw, opath)
        }
      }
    }
  }

  commit <- parse_commit(parsed[[which(types == "commit")]]$object)
  tree <- commit[["tree"]]
  root <- match(tree, names(trees))
  if (is.na(root)) {
    throw(pkg_error(
      "Invalid git response from {.url {url}}, cannot find commit tree"
    ))
  }
  process_tree(root)
  if (any(!done)) {
    warning(
      "Some trees are unreachable when listing files from git repo from ",
      url
    )
  }

  for (i in seq_along(trees)) {
    process_tree(i)
  }

  invisible()
}

# -------------------------------------------------------------------------
# Utility functions
# -------------------------------------------------------------------------

git_ua <- function() {
  "git/2.38.1"
}

#' Comvert raw message from a pkt_line to text, if possible
#'
#' If the input is binary it returns `NA_character_`.
#'
#' @param x Raw vector.
#' @return Character string, if `x` can be an UTF-8 string. Otherwise
#'   `NA_character_`.
#'
#' @noRd

raw_as_utf8 <- function(x) {
  if (is.raw(x)) {
    if (any(x == 0x0)) {
      return(NA_character_)
    }
    if (length(x) > 0 && x[[length(x)]] == 0x0a) {
      x <- x[1:(length(x) - 1)]
    }
    x <- rawToChar(x)
  }

  iconv(x, "UTF-8", "UTF-8")
}

#' Parse a message from git
#'
#' @param msg A raw vector, the full message.
#' @return A list of objects that correspond to _pkt_lines_ of the git
#' message. Each such object is a named list with potential entries:
#' * `type`: this entry is always present, and it is one of
#'   `flush-pkt`, `delim-pkt`, `response-end-pkt` or `data-pkt`. See the
#'   git protocol docs for what these are.
#' * `data`: for `data-pkt` lines this is a raw vector of the data.
#' * `text`: for `data-pkt` lines that are text, this is the text of the
#'   data. We use [raw_as_utf8()] to convert raw data to text, and sometimes
#'   it might interpret binary data as text, especially if the data is
#'   short. So this field is for convenience only.
#'
#' @noRd

git_parse_message <- function(msg) {
  assert_that(is.raw(msg))
  lines <- list()
  pos <- 1L

  while (pos <= length(msg)) {
    if (pos + 3L > length(msg)) {
      throw(pkg_error(
        "Invalid pkt-line at the enf of message from git.",
        .class = "git_proto_error_invalid_data"
      ))
    }
    pkg_len <- msg[pos:(pos + 3)]
    if (!all(pkg_len == charToRaw("PACK"))) {
      if (any(pkg_len < charToRaw('0') | pkg_len > charToRaw('f'))) {
        throw(pkg_error(
          "Invalid pkt-len field in message from git, must be four hexa digits.",
          .class = "git_proto_error_invalid_data"
        ))
      }
      len <- as.integer(as.hexmode(rawToChar(pkg_len)))
    }

    if (all(pkg_len == charToRaw("PACK"))) {
      lines[[length(lines) + 1L]] <- list(
        type = "pack",
        data = msg[pos:length(msg)]
      )
      pos <- length(msg) + 1L
    } else if (len == 0) {
      lines[[length(lines) + 1L]] <- list(type = "flush-pkt")
      pos <- pos + 4L
    } else if (len == 1) {
      lines[[length(lines) + 1L]] <- list(type = "delim-pkt")
      pos <- pos + 4L
    } else if (len == 2) {
      lines[[length(lines) + 1L]] <- list(type = "response-end-pkt") # nocov
      pos <- pos + 4L # nocov
    } else {
      if (pos + len - 1L > length(msg)) {
        throw(pkg_error(
          "Invalid pkt-payload in message from git.",
          i = "Need {len} byte{?s}, found {length(msg) - pos + 1L}.",
          .class = "git_proto_error_invalid_data"
        ))
      }

      data <- msg[(pos + 4):(pos + len - 1L)]
      lines[[length(lines) + 1L]] <- list(
        type = "data-pkt",
        data = data,
        text = raw_as_utf8(data)
      )

      pos <- pos + len
    }
  }

  lines
}

# After receiving the capability advertisement, a client can then issue a
# request to select the command it wants with any particular capabilities
# or arguments.  There is then an optional section where the client can
# provide any command specific parameters or queries.  Only a single
# command can be requested at a time.
#
#     request = empty-request | command-request
#     empty-request = flush-pkt
#     command-request = command
#                       capability-list
#                       delim-pkt
#                       command-args
#                       flush-pkt
#     command = PKT-LINE("command=" key LF)
#     command-args = *command-specific-arg
#
#     command-specific-args are packet line framed arguments defined by
#     each individual command.
#
# https://github.com/git/git/blob/master/Documentation/gitprotocol-v2.txt

#' Create a git protocol version 2 message, to be sent to git
#'
#' @param cmd Command to send. E.g. `ls-refs` or `fetch`.
#' @param caps Capabilities to advertise, a character vector.
#'   If they don't include the trailing `\n` character, it will be added.
#' @param args Arguments to send to git, a character vector. If they
#'   don't include the trailing `\n` character, it will be added.
#'
#' @return The message as a raw vector.
#'
#' @noRd

git_create_message_v2 <- function(
  cmd,
  caps = character(),
  args = character()
) {
  caps <- ifelse(last_char(caps) == "\n", caps, paste0(caps, "\n"))
  args <- ifelse(last_char(args) == "\n", args, paste0(args, "\n"))

  c(
    pkt_line("command=", cmd, "\n"),
    unlist(lapply(caps, pkt_line)),
    delim_pkt(),
    unlist(lapply(args, pkt_line)),
    flush_pkt()
  )
}

git_create_message_v1 <- function(args = character(), caps = character()) {
  if (length(args) == 0) {
    throw(pkg_error(
      "Invalid git protocol (v1) message, must have at least one argument"
    ))
  }

  args[1] <- paste0(args[1], " ", paste(caps, collapse = " "))
  args <- ifelse(
    last_char(args) == "\n" | args == "",
    args,
    paste0(args, "\n")
  )

  res <- lapply(args, pkt_line)
  res[args == ""] <- list(charToRaw("0000"))
  unlist(res)
}

#' Send a protocol version 2 message to a git server
#'
#' @inheritParams git_list_refs
#' @inheritParams git_create_message_v2
#' @return Response from git, already parsed with [git_parse_message()].
#'
#' @noRd

git_send_message_v2 <- function(
  url,
  cmd,
  caps = character(),
  args = character()
) {
  synchronize(async_git_send_message_v2(url, cmd, caps = caps, args = args)) # nocov
}

#' `async_git_send_message_v2()` is the asynchronous variant of
#' `git_send_message_v2()`.
#' @rdname git_send_message_v2
#' @noRd

async_git_send_message_v2 <- function(
  url,
  cmd,
  caps = character(),
  args = character()
) {
  msg <- git_create_message_v2(cmd, caps = caps, args = args)

  url2 <- paste0(url, "/git-upload-pack")
  headers <- c(
    "Content-Type" = "application/x-git-upload-pack-request",
    "User-Agent" = git_ua(),
    "accept-encoding" = "deflate, gzip",
    "accept" = "application/x-git-upload-pack-result",
    "git-protocol" = "version=2",
    "content-length" = as.character(length(msg))
  )

  git_http_post(
    url2,
    data = msg,
    headers = headers
  )$then(http_stop_for_status)$then(
    function(res) git_parse_message(res$content)
  )
}

async_git_send_message_v1 <- function(url, args, caps) {
  msg <- git_create_message_v1(caps = caps, args = args)

  url2 <- paste0(url, "/git-upload-pack")
  headers <- c(
    "Content-Type" = "application/x-git-upload-pack-request",
    "User-Agent" = git_ua(),
    "accept" = "application/x-git-upload-pack-result",
    "content-length" = as.character(length(msg))
  )
  git_http_post(
    url2,
    data = msg,
    headers = headers
  )$then(http_stop_for_status)$then(
    function(res) git_parse_message(res$content)
  )
}

delim_pkt <- function() {
  charToRaw("0001")
}

flush_pkt <- function() {
  charToRaw("0000")
}

#' Create a `pkt_line`, to be sent as part of a message to git
#'
#' Currently it errors if the complete payload is longer than 65516 bytes.
#'
#' @param payload Raw data, or a string to send.
#' @param ... More strings to send. `payload` and `...` are all
#'   concatenated to form the line.
#' @return Raw vector containing the `pkt_line`.
#' @noRd

pkt_line <- function(payload, ...) {
  if (!is.raw(payload)) {
    payload <- charToRaw(payload)
  }
  line <- c(
    payload,
    charToRaw(paste0("", ...))
  )
  len <- length(line) + 4L
  if (len > 65516) {
    throw(pkg_error(
      "packet line longer than 65516 bytes is not implemented yet."
    ))
  }

  line <- c(
    charToRaw(format(as.hexmode(len), width = 4)),
    line
  )

  line
}

#' List references in a remote git repositoty, protocol version 1
#'
#' We use this if we don't want to filter the references, because it
#' needs a single HTTP query, whereas with version 2, we would need two
#' queries.
#'
#' @inheritParams git_list_refs
#' @return Same as [git_list_refs()].
#' @noRd

git_list_refs_v1 <- function(url) {
  synchronize(async_git_list_refs_v1(url))
}

#' `async_git_list_refs_v1()` is the asynchronous variant of
#' `git_list_refs_v1()`.
#' @rdname git_list_refs_v1
#' @noRd

# https://github.com/git/git/blob/master/Documentation/gitprotocol-http.txt

async_git_list_refs_v1 <- function(url) {
  url
  url1 <- paste0(url, "/info/refs?service=git-upload-pack")
  git_http_get(url1, headers = c("User-Agent" = git_ua()))$then(
    http_stop_for_status
  )$then(function(response) git_list_refs_v1_process(response, url))
}

#' Process the response to a version 1 reference list query
#'
#' @param response Response from git, as returned by `async::http_get()`, which
#'   is the same as the object from `curl::curl_fetch_memory()`.
#' @return Same as [git_list_refs()].
#' @noRd

git_list_refs_v1_process <- function(response, url) {
  psd <- git_parse_message(response$content)
  psd <- drop_service_line(psd)
  check_initial_response(psd, url)

  # The stream MUST include capability declarations behind a NUL on the
  # first ref.
  nul <- which(psd[[1]]$data == 0)
  if (length(nul) != 1) {
    # nocov start
    throw(pkg_error(
      "Unexpected response from git server, third pkt-line should contain
       exactly one {.code NUL} byte.",
      .class = "git_proto_error_unexpected_response",
      .data = list(url = redact_url(url))
    ))
    # nocov end
  }

  caps <- if (nul < length(psd[[1]]$data)) {
    rawToChar(psd[[1]]$data[(nul + 1):length(psd[[1]]$data)])
  } else {
    "" # nocov
  }
  caps <- strsplit(trimws(caps), " ", fixed = TRUE)[[1]]
  psd[[1]]$data <- if (nul == 1) raw() else psd[[1]]$data[1:(nul - 1)]
  psd[[1]]$text <- rawToChar(psd[[1]]$data)

  if (psd[[length(psd)]]$type != "flush-pkt") {
    # nocov start
    throw(pkg_error(
      "Response from git server does not have a closing {.code flush-pkt}.",
      .class = "git_proto_error_unexpected_response",
      .data = list(url = redact_url(url))
    ))
    # nocov end
  }

  refs <- git_parse_pkt_line_refs(psd[1:(length(psd) - 1)], url)
  list(refs = refs, caps = caps)
}

git_list_refs_v1_process_1 <- function(response, url, prefixes) {
  # We wanted v2, but the server replied in v1, so we'll use v1 now
  refs <- git_list_refs_v1_process(response, url)
  if (length(prefixes) > 0) {
    keep <- logical(length(refs$refs$ref))
    for (pfx in prefixes) {
      keep <- keep | startsWith(refs$refs$ref, pfx)
    }
    refs$refs <- refs$refs[keep, ]
  }
  refs
}

#' Helper function to parse a `pkt_line` containing a git ref
#'
#' E.g. an `ls-refs` commans responds with such lines.
#'
#' @param lines List of `pkt_line` objects, they must be all data
#'   packets (`data-pkt`), in the right format.
#' @return Data frame with columns `ref` (name of the reference), and
#'   `hash` (SHA).
#' @noRd

git_parse_pkt_line_refs <- function(lines, url) {
  res <- data_frame(
    ref = character(length(lines)),
    hash = character(length(lines))
  )

  for (idx in seq_along(lines)) {
    line <- lines[[idx]]
    if (
      line$type != "data-pkt" ||
        is.null(line$text) ||
        !grepl("^[0-9a-f]{40} .+$", line$text)
    ) {
      # nocov start
      throw(pkg_error(
        "Expected response from git server.",
        i = "Line {idx} must be a text {.code data-pkt} with a sha and a ref name.",
        .class = "git_proto_error_unexpected_response",
        .data = list(url = redact_url(url))
      ))
      # nocov end
    }

    res$ref[idx] <- substr(line$text, 42, nchar(line$text))
    res$hash[idx] <- substr(line$text, 1, 40)
  }

  res
}

#' List references in a remote git repositoty, protocol version 2
#'
#' We use this to filter all refs using prefixes, if we are only
#' interested in a subset of refs.
#'
#' @inheritParams git_list_refs
#' @return Same as [git_list_refs()].
#' @noRd

git_list_refs_v2 <- function(url, prefixes = character()) {
  synchronize(async_git_list_refs_v2(url, prefixes))
}

#' `async_git_list_refs_v2()` is the asynchronous version of
#' `git_list_refs_v2()`.
#' @rdname git_list_refs_v2
#' @noRd

async_git_list_refs_v2 <- function(url, prefixes = character()) {
  url
  prefixes

  url1 <- paste0(url, "/info/refs?service=git-upload-pack")

  headers <- c(
    "User-Agent" = git_ua(),
    "git-protocol" = "version=2"
  )

  git_http_get(url1, headers = headers)$then(http_stop_for_status)$then(
    function(res) async_git_list_refs_v2_process_1(res, url, prefixes)
  )
}

#' Helper function to post-process the response to the initial client
#' request, with protocol version 2
#'
#' @param response The HTTP response from git, from `async::http_get()`,
#'   i.e. from `curl::curl_fetch_memory()`.
#' @param url The original URL is passed in here.
#' @param prefixes The original prefixes are passed in here.
#' @return `async_git_list_refs_v2_process_1()` will send out the
#'   second query, an `ls-refs` command, filtered for `prefixes`, and
#'   return with (the deferred value of) the HTTP response.
#'
#' @noRd

async_git_list_refs_v2_process_1 <- function(response, url, prefixes) {
  psd <- git_parse_message(response$content)
  psd <- drop_service_line(psd)
  check_initial_response(psd, url)
  async_git_list_refs_v2_process_2(response, psd, url, prefixes)
}


async_git_list_refs_v2_process_2 <- function(response, psd, url, prefixes) {
  # capability-advertisement = protocol-version
  #                capability-list
  #                flush-pkt
  #
  # protocol-version = PKT-LINE("version 2" LF)
  # capability-list = *capability
  # capability = PKT-LINE(key[=value] LF)
  #
  # key = 1*(ALPHA | DIGIT | "-_")
  # value = 1*(ALPHA | DIGIT | " -_.,?\/{}[]()<>!@#$%^&*+=:;")

  if (is.null(psd[[1]]$text)) {
    throw(pkg_error(
      "Invalid git protocol message from {.url {redact_url(url)}}."
    ))
  }

  if (is.na(psd[[1]]$text)) {
    return(git_list_refs_v1_process_1(response, url, prefixes))
  }

  if (psd[[1]]$text != "version 2") {
    throw(pkg_error(
      "Only git protocol version 2 is supported, not {psd[[1]]$text}.",
      .class = "git_proto_error_not_implemented"
    ))
  }

  if (psd[[length(psd)]]$type != "flush-pkt") {
    throw(pkg_error(
      "Response from git server does not have a closing {.code flush-pkt}.",
      .class = "git_proto_error_unexpected_response",
      .data = list(url = redact_url(url))
    ))
  }

  caps <- unlist(lapply(psd[1:(length(psd) - 1)], "[[", "text"))

  args <- if (length(prefixes)) paste0("ref-prefix ", prefixes, "\n")

  async_git_send_message_v2(url, "ls-refs", args = as.character(args))$then(
    function(res) async_git_list_refs_v2_process_3(res, caps, url)
  )
}

drop_service_line <- function(psd) {
  # Actually, this might not be present in V2 servers...
  if (
    length(psd) > 0 &&
      !is.null(psd[[1]]$text) &&
      grepl("^# service=.+", psd[[1]]$text)
  ) {
    psd <- psd[-1]
  }

  if (length(psd) > 0 && psd[[1]]$type == "flush-pkt") {
    psd <- psd[-1]
  }

  psd
}

check_initial_response <- function(psd, url) {
  if (length(psd) < 1 || psd[[1]]$type != "data-pkt") {
    throw(pkg_error(
      "Unexpected response from git server, no {.code data-pkt} line.",
      .class = "git_proto_error_unexpected_response",
      .data = list(url = redact_url(url))
    ))
  }
}

#' Helper function to process the response to an `ls-refs` command
#'
#' @param reply The parsed message from the server.
#' @param caps Capabilities that are passed in from the response to
#'   the initial client request.
#' @return Same as [git_list_refs()].
#' @noRd

async_git_list_refs_v2_process_3 <- function(reply, caps, url) {
  if (reply[[length(reply)]]$type != "flush-pkt") {
    throw(pkg_error(
      "Response from git server does not have a closing {.code flush-pkt}.",
      .class = "git_proto_error_unexpected_response",
      .data = list(url = redact_url(url))
    ))
  }

  # Any refs at all?
  refs <- if (length(reply) == 1) {
    git_parse_pkt_line_refs(list(), url)
  } else {
    git_parse_pkt_line_refs(reply[1:(length(reply) - 1)], url)
  }
  list(refs = refs, caps = caps)
}

#' Unpack a git packfile in memory
#'
#' @details
#' It cannot currently unpack packfiles with deltas.
#'
#' It checks the packfile checksum.
#'
#' @param pack The git packfile as a raw vector, or the name of the file
#'   containing the packfile.
#' @return List of git objects. Each object is a named list with entries:
#'   * `type`: `commit`, `tree`, `blob` or `tag`.
#'   * `object`: the contents of the object. For `commit` the commit
#'     data and message in a string. For `tree` a data frame, as returned
#'     from [parse_tree()]. For `blob` and `tag` a raw vector.
#'   * `size`: unpacked size.
#'   * `packed_size`: packed size (size header not included).
#'
#' @noRd

git_unpack <- function(pack) {
  # allow file names as well
  if (is.character(pack)) {
    pack <- readBin(pack, "raw", file.size(pack))
  }

  if (length(pack) < 32) {
    throw(pkg_error(
      "Invalid packfile from git, too short.",
      .class = "git_proto_error_invalid_data"
    ))
  }
  if (!all(pack[1:4] == charToRaw("PACK"))) {
    throw(pkg_error(
      "Not a git packfile, it does not have a {.code PACK} header.",
      .class = "git_proto_error_invalid_data"
    ))
  }
  if (!all(pack[5:8] == as.raw(c(0x00, 0x00, 0x00, 0x02)))) {
    throw(pkg_error(
      "Unexpected packfile version, must be version 2.",
      .class = "git_proto_error_unexpected_response"
    ))
  }

  chksum <- cli::hash_raw_sha1(utils::head(pack, -20))
  chksum_exp <- paste(format(utils::tail(pack, 20)), collapse = "")
  if (chksum != chksum_exp) {
    throw(pkg_error(
      "Checksum mismatch in git packfile.",
      .class = "git_proto_error_invalid_data"
    ))
  }

  n_obj <- parse_int32_nwb(pack[9:12])

  idx <- 13L
  objects <- vector("list", n_obj)
  object_starts <- integer()

  types <- c(
    "commit",
    "tree",
    "blob",
    "tag",
    "reserved",
    "ofs_delta",
    "ref_delta"
  )

  unpack_object <- function() {
    start <- idx
    type <- bitwShiftR(bitwAnd(as.integer(pack[idx]), 0x7f), 4L)
    size <- parse_size(pack, idx)
    idx <<- size$idx + 1
    if (type == 6L) {
      offset <- parse_ofs_delta_offset(pack, idx)
      idx <<- offset$idx + 1
    } else if (type == 7L) {
      if (idx + 19 > length(pack)) {
        # nocov start
        throw(pkg_error(
          "Invalid packfile, unexpected end of file"
        ))
        # nocov end
      }
      base <- bin_to_sha(pack[idx:(idx + 19)])
      idx <<- idx + 20
    }
    obj <- zip::inflate(pack, idx, size$size)
    idx <<- idx + obj$bytes_read
    if (type == 6L) {
      baseidx <- object_starts[[as.character(start - offset$size)]]
      deltified_object(obj$output, baseidx = baseidx)
    } else if (type == 7L) {
      deltified_object(obj$output, base = base)
    } else {
      list(
        type = types[type],
        raw = obj$output,
        size = size$size,
        packed_size = obj$bytes_read
      )
    }
  }

  deltified_object <- function(delta, base = NULL, baseidx = NULL) {
    baseidx <- baseidx %||% match(base, names(objects))
    if (is.na(baseidx) || objects[[baseidx]]$type == "delta") {
      return(list(
        type = "delta",
        data = delta,
        base = base,
        baseidx = if (is.na(baseidx)) NULL else baseidx
      ))
    }
    baseobj <- objects[[baseidx]]$raw
    didx <- 1L
    basesize <- parse_delta_size(delta, didx)
    didx <- basesize$idx + 1L
    size <- parse_delta_size(delta, didx)
    didx <- size$idx + 1L
    newobj <- raw(size$size)
    nidx <- 1L
    while (didx <= length(delta)) {
      c <- as.integer(delta[didx])
      if (c < 128) {
        datasize <- bitwAnd(c, 0x7f)
        newobj[nidx:(nidx + datasize - 1L)] <-
          delta[(didx + 1L):(didx + datasize)]
        nidx <- nidx + datasize
        didx <- didx + datasize + 1L
      } else {
        ofs <- parse_delta_offset(delta, didx)
        newobj[nidx:(nidx + ofs$size - 1L)] <-
          baseobj[(ofs$offset + 1L):(ofs$offset + ofs$size)]
        nidx <- nidx + ofs$size
        didx <- ofs$idx + 1L
      }
    }
    list(
      type = objects[[baseidx]]$type,
      raw = newobj,
      size = size,
      packed_size = length(delta)
    )
  }

  finalize_object <- function(x) {
    if (x$type == "commit") {
      x$object <- rawToChar(x$raw)
    } else if (x$type == "tree") {
      x$object <- parse_tree(x$raw)
    }

    if (x$type %in% c("commit", "tree", "blob", "tag")) {
      raw2 <- c(
        charToRaw(paste0(x$type, " ", length(x$raw))),
        as.raw(0L),
        x$raw
      )
      x$hash <- cli::hash_raw_sha1(raw2)
    } else if (x$type == "delta") {
      # do nothing
    } else {
      # nocov start
      throw(pkg_error(
        "git packfile object type {.cls {x$type}} is not
         implemented yet.",
        .class = "git_proto_error_not_implemented"
      ))
      # nocov end
    }
    x
  }

  for (i in seq_len(n_obj)) {
    object_starts[[as.character(idx)]] <- i
    objects[[i]] <- unpack_object()
    objects[[i]] <- finalize_object(objects[[i]])
    if (!is.null(objects[[i]]$hash)) names(objects)[i] <- objects[[i]]$hash
  }

  # now need to resolve the deltas
  n_delta <- sum(sapply(objects, "[[", "type") == "delta")
  while (n_delta > 0) {
    for (i in seq_len(n_obj)) {
      if (objects[[i]]$type == "delta") {
        objects[[i]] <- deltified_object(
          objects[[i]]$data,
          objects[[i]]$base,
          objects[[i]]$baseidx
        )
        objects[[i]] <- finalize_object(objects[[i]])
        if (!is.null(objects[[i]]$hash)) names(objects)[i] <- objects[[i]]$hash
      }
    }
    n_delta2 <- sum(sapply(objects, "[[", "type") == "delta")
    if (n_delta2 == n_delta) {
      throw(pkg_error(
        "Found circular references while resolving deltas in git pack file."
      ))
    }
    n_delta <- n_delta2
  }

  objects
}

git_list_pack_index <- function(idx) {
  if (is.character(idx)) {
    idx <- readBin(idx, "raw", file.size(idx))
  }

  if (length(idx) < 4 + 4 + 256) {
    throw(pkg_error(
      "Invalid pack index file, too short: {length(idx)} byte{?s}."
    ))
  }

  if (any(idx[1:4] != c(0xff, 0x74, 0x4f, 0x63))) {
    throw(pkg_error(
      "Invalid pack index file, no {.code \\377tOc} header."
    ))
  }
  if (any(idx[5:8] != c(0, 0, 0, 2))) {
    throw(pkg_error(
      "Only version 2 pack index files are supported"
    ))
  }

  tab <- matrix(as.integer(idx[9L:(9L + (256L * 4L) - 1L)]), nrow = 4)
  tab <- tab[1, ] * 256**3 + tab[2, ] * 256**2 + tab[3, ] * 256 + tab[4, ]
  n_obj <- tab[256]

  hash_off <- 256L * 4L + 9L
  hash_len <- n_obj * 20L
  hash <- matrix(
    as.character(idx[hash_off:(hash_off + hash_len - 1L)]),
    nrow = 20
  )
  hash <- apply(hash, 2, paste, collapse = "")

  crc_off <- hash_off + hash_len
  crc_len <- n_obj * 4
  crc <- matrix(as.integer(idx[crc_off:(crc_off + crc_len - 1L)]), nrow = 4)
  crc <- crc[1, ] * 256**3 + crc[2, ] * 256**2 + crc[3, ] * 256 + crc[4, ]

  off_off <- crc_off + crc_len
  off_len <- n_obj * 4
  off <- matrix(as.integer(idx[off_off:(off_off + off_len - 1L)]), nrow = 4)
  off <- off[1, ] * 256**3 + off[2, ] * 256**2 + off[3, ] * 256 + off[4, ]

  data_chksum <- idx[(off_off + off_len):(off_off + off_len + 20L - 1L)]
  data_chksum <- paste(as.character(data_chksum), collapse = "")

  idx_chksum <- idx[(off_off + off_len):(off_off + off_len + 20L - 1L) + 20L]
  idx_chksum <- paste(as.character(idx_chksum), collapse = "")

  if (length(idx) > off_off + off_len + 40L) {
    warning("Ignored 8 byte offsets in git pack file")
  }

  objects <- data.frame(
    stringsAsFactors = FALSE,
    hash = hash,
    crc = crc,
    offset = off
  )
  objects <- objects[order(objects$offset), ]
  rownames(objects) <- NULL

  list(
    objects = objects,
    data_chksum = data_chksum,
    idx_chksum = idx_chksum
  )
}

#' Parse a four byte integer in network byte order
#'
#' @param x Raw vector or four bytes.
#' @return Integer scalar.
#' @noRd

parse_int32_nwb <- function(x) {
  if (length(x) != 4L || !is.raw(x)) {
    throw(pkg_error(
      "Cannot parse integer, not raw or number of bytes is wrong."
    ))
  }
  sum(as.integer(x) * (256L**(3:0)))
}

#' Parse a variable length size field
#'
#' @details
#' As in https://git-scm.com/docs/pack-format#_size_encoding.
#' Some notes, because that docs is pretty dense.
#'
#' * First bit is zero for the last byte.
#' * Next three bits of the first byte are the object type.
#' * So we throw away the first four bits of the first byte and use the
#'   last four bits only.
#' * Then we keep processing the last seven bits of bytes, until the
#'   first bit is zero. (We still process that byte.)
#'
#' @param x Raw vector, typically the whole packfile.
#' @param idx Integer scalar, where to start reading the size field in the
#'   `x` raw vector.
#' @return Named list with entries:
#'   * `size`: parsed size,
#'   * `idx`: Integer, point to the last byte of the parsed size field.
#'
#' @noRd

parse_size <- function(x, idx) {
  c <- as.integer(x[idx])
  size <- bitwAnd(c, 0x0f)
  shft <- 4L
  while (c >= 128) {
    idx <- idx + 1L
    if (idx > length(x)) {
      # nocov start
      throw(pkg_error(
        "Invalid git packfile, invalid size field.",
        .class = "git_proto_error_invalid_data"
      ))
      # nocov end
    }
    c <- as.integer(x[idx])
    size <- size + bitwShiftL(bitwAnd(c, 0x7f), shft)
    shft <- shft + 7L
  }

  list(size = size, idx = idx)
}

parse_ofs_size <- function(x, idx) {}

parse_delta_size <- function(x, idx) {
  c <- as.integer(x[idx])
  size <- bitwAnd(c, 0x7f)
  shft <- 7L
  while (c >= 128) {
    idx <- idx + 1L
    if (idx > length(x)) {
      # nocov start
      throw(pkg_error(
        "Invalid git packfile, invalid size field.",
        .class = "git_proto_error_invalid_data"
      ))
      # nocov end
    }
    c <- as.integer(x[idx])
    size <- size + bitwShiftL(bitwAnd(c, 0x7f), shft)
    shft <- shft + 7L
  }

  list(size = size, idx = idx)
}

parse_ofs_delta_offset <- function(x, idx) {
  c <- as.integer(x[idx])
  size <- bitwAnd(c, 0x7f)
  while (c >= 128) {
    idx <- idx + 1L
    c <- as.integer(x[idx])
    size <- size + 1L
    size <- bitwShiftL(size, 7) + (bitwAnd(c, 0x7f))
  }

  list(size = size, idx = idx)
}

parse_delta_offset <- function(x, idx) {
  c <- as.integer(x[idx])
  offset <- 0L
  size <- 0L
  if (bitwAnd(c, 0x01)) {
    idx <- idx + 1L
    offset <- offset + as.integer(x[idx])
  }
  if (bitwAnd(c, 0x02)) {
    idx <- idx + 1L
    offset <- offset + as.integer(x[idx]) * 256
  }
  if (bitwAnd(c, 0x04)) {
    idx <- idx + 1L
    offset <- offset + as.integer(x[idx]) * 256 * 256
  }
  if (bitwAnd(c, 0x08)) {
    idx <- idx + 1L
    offset <- offset + as.integer(x[idx]) * 256 * 256 * 256
  }
  if (bitwAnd(c, 0x10)) {
    idx <- idx + 1L
    size <- size + as.integer(x[idx])
  }
  if (bitwAnd(c, 0x20)) {
    idx <- idx + 1L
    size <- size + as.integer(x[idx]) * 256
  }
  if (bitwAnd(c, 0x40)) {
    idx <- idx + 1L
    size <- size + as.integer(x[idx]) * 256 * 256
  }

  # exception for easily including a block
  if (size == 0L) {
    size <- 0x10000
  }

  list(offset = offset, size = size, idx = idx)
}

#' Parse a git tree object
#'
#' @param tree The tree object in a raw vector, _without_ the 'tree' +
#'   content size + `\0` header. Git packfiles do not have these headers,
#'   so we don't require them here, either.
#' @return Data frame with character columns:
#'   * `type`: Either `"blob"` or `"tree"`.
#'   * `mode`: Unix permissions.
#'   * `path`: File or directory name.
#'   * `hash`: Hash of `blob` or tree`.
#'
#' @noRd

parse_tree <- function(tree) {
  nul <- which(tree == 0)
  last <- nul[1]
  for (i in seq_along(nul)[-1]) {
    nl <- nul[i]
    if (nl > last + 20) {
      last <- nl
    } else {
      nul[i] <- NA_integer_
    }
  }

  nul <- nul[!is.na(nul)]
  num <- length(nul)
  res <- data_frame(
    type = character(num),
    mode = character(num),
    path = character(num),
    hash = character(num)
  )

  beg <- c(1L, nul + 21)
  for (i in seq_along(nul)) {
    txt <- rawToChar(tree[beg[i]:nul[i]])
    res$mode[i] <- sub("[ ].*$", "", txt)
    res$path[i] <- sub("[^ ]*[ ]", "", txt)
    res$hash[i] <- bin_to_sha(tree[(nul[i] + 1):(nul[i] + 20)])
  }
  res$type <- ifelse(substr(res$mode, 1, 1) == "1", "blob", "tree")

  res
}

#' Format a raw vector as a hexa string
#'
#' @param x Raw vector.
#' @return String.
#'
#' @noRd

bin_to_sha <- function(x) {
  paste(format(x), collapse = "")
}

#' Parse a commit object
#'
#' @param commit Commit object as an UTF-8 string.
#' @return Named character vector. Names are taken from the commit object,
#'   plus `message` is the name of the commit message. The commit object
#'   should have names `tree`, `parent` (one for each parent commit),
#'   `author`, `committer`, `encoding`.
#'
#' @noRd

parse_commit <- function(commit) {
  lines <- strsplit(commit, "\n", fixed = TRUE)[[1]]
  delim <- which(lines == "")[1]
  if (delim == 1) {
    # nocov start
    throw(pkg_error(
      "Invalid git commit object, no {.code tree} field.",
      .class = "git_proto_error_invalid_data"
    ))
    # nocov end
  }

  message <- if (is.na(delim)) {
    NA_character_
  } else if (delim == length(lines)) {
    "" # nocov
  } else {
    paste(lines[(delim + 1):length(lines)], collapse = "\n")
  }

  fields <- lines[1:(delim - 1)]
  nms <- sub("[ ].*$", "", fields)
  vls <- sub("^[^ ]+ ", "", fields)

  structure(c(vls, message), names = c(nms, "message"))
}

# Returns `""` for empty strings!

last_char <- function(x) {
  nc <- nchar(x)
  substr(x, nc, nc)
}

redact_url <- function(x) {
  sub("://[^/]+@", "://<auth>@", x)
}

git_dumb_list_refs <- function(url) {
  synchronize(async_git_dumb_list_refs(url))
}

async_git_dumb_list_refs <- function(url) {
  url

  url1 <- paste0(url, "/info/refs")
  url2 <- paste0(url, "/HEAD")
  headers <- c(
    "User-Agent" = git_ua()
  )
  when_all(
    git_http_get(url1, headers = headers)$then(http_stop_for_status),
    git_http_get(url2, headers = headers)$then(http_stop_for_status)
  )$then(function(res) async_git_dumb_list_refs_process(res, url))
}

async_git_dumb_list_refs_process <- function(res, url) {
  res_refs <- res[[1]]
  res_head <- res[[2]]
  lines <- strsplit(rawToChar(res_refs$content), "\n", fixed = TRUE)[[1]]
  lines2 <- strsplit(lines, "\t", fixed = TRUE)
  ref <- vcapply(lines2, "[[", 2)
  hash <- vcapply(lines2, "[[", 1)

  head <- trimws(rawToChar(res_head$content))
  head <- sub("^[^ ]* ", "", head)
  has_head <- head != ""
  if (has_head && !head %in% ref) {
    throw(pkg_error(
      "HEAD does not refer to a ref in git repo at {.url {url}}.",
      "i" = "HEAD is {.val {head}}."
    ))
  }

  list(
    refs = data_frame(
      ref = c(if (has_head) "HEAD", ref),
      hash = c(if (has_head) hash[match(head, ref)], hash)
    ),
    caps = character()
  )
}

git_dumb_download_file <- function(url, sha, path, output = basename(path)) {
  invisible(synchronize(async_git_dumb_download_file(url, sha, path, output)))
}

# This only works for blobs at the root of the tree currently!
# Also, it only works if the object with sha is not in a pack file!
# So it is pretty limited, and can only be used with a fallback.

async_git_dumb_download_file <- function(
  url,
  sha,
  path,
  output = basename(path)
) {
  async_git_dumb_get_commit(url, sha)$then(function(cmt) {
    async_git_dumb_get_tree(url, cmt[["tree"]])
  })$then(function(tree) {
    wh <- match(path, tree$path)
    if (is.na(wh)) {
      throw(pkg_error(
        "Could not find path {.val {path}} in git repo at {.url {url}}."
      ))
    }
    if (tree$type[wh] != "blob") {
      throw(pkg_error(
        "Path {.val {path}} is not a blob in git repo at {.url {url}}."
      ))
    }
    tree$hash[wh]
  })$then(function(blob) {
    async_git_dumb_get_blob(url, blob)
  })$then(function(bytes) {
    if (!is.null(output)) {
      mkdirp(dirname(output))
      writeBin(bytes, output)
    }
    invisible(bytes)
  })
}

async_git_dumb_get_commit <- function(url, sha) {
  url1 <- paste0(
    url,
    "/objects/",
    substr(sha, 1, 2),
    "/",
    substr(sha, 3, nchar(sha))
  )
  headers <- c(
    "User-Agent" = git_ua(),
    "accept-encoding" = "deflate, gzip"
  )
  git_http_get(url = url1, headers = headers)$then(http_stop_for_status)$then(
    function(res) {
      cmt <- zip::inflate(res$content)$output
      if (any(utils::head(cmt, 6) != charToRaw("commit"))) {
        throw(pkg_error(
          "Git object {.val {substr(sha, 1, 7)}} is not a commit object
           in git repo at {.url {url}}."
        ))
      }
      nul <- which(cmt == 0)
      if (length(nul) != 1) {
        throw(pkg_error(
          "Git commit object {.val {substr(sha, 1, 7)}} is invalid, does not
           contain a single zero byte, in git repo at {.url {url}}."
        ))
      }
      parse_commit(rawToChar(cmt[(nul + 1):length(cmt)]))
    }
  )
}

async_git_dumb_get_tree <- function(url, sha) {
  url1 <- paste0(
    url,
    "/objects/",
    substr(sha, 1, 2),
    "/",
    substr(sha, 3, nchar(sha))
  )
  headers <- c(
    "User-Agent" = git_ua(),
    "accept-encoding" = "deflate, gzip"
  )
  git_http_get(url = url1, headers = headers)$then(http_stop_for_status)$then(
    function(res) {
      cmt <- zip::inflate(res$content)$output
      if (any(utils::head(cmt, 4) != charToRaw("tree"))) {
        throw(pkg_error(
          "Git object {.val {substr(sha, 1, 7)}} is not a tree object
           in git repo at {.url {url}}."
        ))
      }
      nul <- which(cmt == 0)[1]
      if (is.na(nul)) {
        throw(pkg_error(
          "Git tree object {.val {substr(sha, 1, 7)}} is invalid, does not
           contain any zero bytes, in git repo at {.url {url}}."
        ))
      }
      parse_tree(cmt[(nul + 1):length(cmt)])
    }
  )
}

async_git_dumb_get_blob <- function(url, sha) {
  url1 <- paste0(
    url,
    "/objects/",
    substr(sha, 1, 2),
    "/",
    substr(sha, 3, nchar(sha))
  )
  headers <- c(
    "User-Agent" = git_ua(),
    "accept-encoding" = "deflate, gzip"
  )
  git_http_get(url = url1, headers = headers)$then(http_stop_for_status)$then(
    function(res) {
      cmt <- zip::inflate(res$content)$output
      if (any(utils::head(cmt, 4) != charToRaw("blob"))) {
        throw(pkg_error(
          "Git object {.val {substr(sha, 1, 7)}} is not a blob object
           in git repo at {.url {url}}."
        ))
      }
      nul <- which(cmt == 0)[1]
      if (is.na(nul)) {
        throw(pkg_error(
          "Git blob object {.val {substr(sha, 1, 7)}} is invalid, does not
           contain any zero bytes, in git repo at {.url {url}}."
        ))
      }
      cmt[(nul + 1):length(cmt)]
    }
  )
}

withr::local_envvar(GITHUB_PAT = "FAIL")

test_that("git_list_refs", {
  skip_on_cran()
  expect_snapshot(
    git_list_refs(fake_git$url("/pak-test.git"))$refs
  )

  # filter
  expect_snapshot(
    git_list_refs_v2(
      fake_git$url("/pak-test.git"),
      "refs/heads/"
    )$refs
  )
})

test_that("git_list_files", {
  skip_on_cran()
  expect_snapshot(error = TRUE, {
    git_list_files(
      fake_git$url("/pak-test.git"),
      "foobar"
    )
  })
  if (!l10n_info()[["UTF-8"]]) {
    skip("UTF-8 snapshot")
  }
  expect_snapshot({
    git_list_files(
      fake_git$url("/pak-test.git"),
      "cefdc0eebcd7f757efb9a80652fd8aaf1a87508e"
    )
    git_list_files(
      fake_git$url("/pak-test.git"),
      "refs/tags/v1"
    )
  })
})

test_that("async_git_list_files_process", {
  # reordering objects in a PACK file does not matter
  ref <- "cefdc0eebcd7f757efb9a80652fd8aaf1a87508e"
  pack <- git_fetch(fake_git$url("/pak-test.git"), ref, blobs = FALSE)

  withr::local_seed(13L)
  pack <- sample(pack)

  expect_snapshot(
    sort(
      async_git_list_files_process(
        pack,
        ref = ref,
        sha = ref,
        url = "url"
      )$files$path
    )
  )
})

test_that("git_download_file", {
  skip_on_cran()
  tmp <- tempfile()
  on.exit(unlink(tmp), add = TRUE)
  expect_snapshot({
    out <- git_download_file(
      fake_git$url("/pak-test.git"),
      "a1e2d6741374d1f32ec138ee2020eae36b859e99",
      tmp
    )
    out
    readLines(tmp)
  })
})

test_that("git_fetch", {
  skip_on_cran()

  # force v1
  pack <- git_fetch(
    fake_git$url("/pak-test.git"),
    structure("cefdc0eebcd7f757efb9a80652fd8aaf1a87508e", protocol = "1")
  )

  if (!l10n_info()[["UTF-8"]]) {
    skip("UTF-8 snapshot")
  }
  expect_snapshot(cat(pack[[1]]$object))
})

test_that("raw_as_utf8", {
  # TODO
  expect_true(TRUE)
})

test_that("git_parse_message errors", {
  expect_snapshot(
    error = TRUE,
    git_parse_message(raw(3))
  )
  expect_snapshot(
    error = TRUE,
    git_parse_message(charToRaw("foobvar"))
  )
  expect_snapshot(
    error = TRUE,
    git_parse_message(charToRaw("00bbnoteonugh"))
  )
})

test_that("git_create_message_v1", {
  expect_snapshot(
    error = TRUE,
    git_create_message_v1(character())
  )
})

test_that("git_create_message_v2", {
  # TODO
  expect_true(TRUE)
})

test_that("git_send_message", {
  # TODO
  expect_true(TRUE)
})

test_that("pkt_line", {
  expect_snapshot(
    error = TRUE,
    pkt_line(raw(70000))
  )
})

test_that("git_list_refs_v1", {
  skip_on_cran()
  expect_snapshot(
    git_list_refs_v1(fake_git$url("/pak-test.git"))$refs
  )
})

test_that("git_list_refs_v1_process_1", {
  skip_on_cran()
  resp <- readRDS(test_path("fixtures", "git-response-v1.rds"))
  expect_snapshot(
    git_list_refs_v1_process_1(
      resp,
      fake_git$url("/pak-test.git"),
      "refs/tags/v1"
    )$refs
  )
})

test_that("async_git_list_refs_v2_process_1", {
  skip_on_cran()
  # works with a v1 response as well
  resp <- readRDS(test_path("fixtures", "git-response-v1.rds"))
  expect_snapshot(
    sy(async_git_list_refs_v2_process_1(
      resp,
      fake_git$url("/pak-test.git"),
      "refs/tags/v1"
    ))$refs
  )
})

test_that("async_git_list_refs_v2_process_2", {
  psd <- readRDS(test_path("fixtures", "git-response-parsed-v2.rds"))
  url <- "http://localhost:3000/git/cli"
  psd2 <- psd
  psd2[[1]]$text <- NULL
  expect_snapshot(
    error = TRUE,
    async_git_list_refs_v2_process_2(NULL, psd2, url, NULL)
  )

  psd2 <- psd
  psd2[[1]]$text <- "version 10"
  expect_snapshot(
    error = TRUE,
    async_git_list_refs_v2_process_2(NULL, psd2, url, NULL)
  )

  psd2 <- psd
  psd2[[length(psd2)]]$type <- "data-pkt"
  expect_snapshot(
    error = TRUE,
    async_git_list_refs_v2_process_2(NULL, psd2, url, NULL)
  )
})

test_that("check_initial_response", {
  expect_snapshot(
    error = TRUE,
    check_initial_response(list(), "http://localhost:3000/git/cli")
  )
})

test_that("async_git_list_refs_v2_process_2", {
  url <- "http://localhost:3000/git/cli"
  expect_snapshot(
    error = TRUE,
    async_git_list_refs_v2_process_3(
      list(list(type = "data-pkt")),
      character(),
      url
    )
  )
})

test_that("git_parse_okt_line_refs", {
  # TODO
  expect_true(TRUE)
})

test_that("git_list_refs_v2", {
  # TODO
  expect_true(TRUE)
})

test_that("git_unpack", {
  path <- test_path("fixtures/git-test-1.pack")
  up1 <- git_unpack(path)
  up2 <- git_unpack(readBin(path, "raw", file.size(path)))
  expect_equal(up1, up2)
  if (!l10n_info()[["UTF-8"]]) {
    skip("UTF-8 snapshot")
  }
  expect_snapshot(git_unpack(path))
})

test_that("git_list_pack_index", {
  path <- test_path("fixtures/git-test-1.idx")
  expect_snapshot(git_list_pack_index(path))
})

test_that("git_unpack errors", {
  path <- test_path("fixtures/git-test-1.pack")
  pack <- readBin(path, "raw", file.size(path))
  expect_snapshot(
    error = TRUE,
    git_unpack(pack[1:30])
  )

  expect_snapshot(
    error = TRUE,
    git_unpack(charToRaw("nope and some more so we have enough bytes"))
  )

  pack2 <- pack
  pack2[8] <- as.raw(1)
  expect_snapshot(
    error = TRUE,
    git_unpack(pack2)
  )

  pack2 <- pack
  pack2[length(pack2)] <- as.raw(11)
  expect_snapshot(
    error = TRUE,
    git_unpack(pack2)
  )
})

test_that("parse_int32_nwb", {
  expect_snapshot(
    error = TRUE,
    parse_int32_nwb(raw(3))
  )
})

test_that("parse_size", {
  # TODO
  expect_true(TRUE)
})

test_that("parse_tree", {
  # TODO
  expect_true(TRUE)
})

test_that("bin_to_sha", {
  # TODO
  expect_true(TRUE)
})

test_that("parse_commit", {
  # TODO
  expect_true(TRUE)
})

test_that("async_git_resolve_ref", {
  skip_on_cran()

  # branches w/o refs/heads/ prefix
  expect_snapshot(
    sy(async_git_resolve_ref(
      fake_git$url("/pak-test.git"),
      "main"
    ))
  )

  # tags w/o refs/rags/ prefix
  expect_snapshot(
    sy(async_git_resolve_ref(
      fake_git$url("/pak-test.git"),
      "v1"
    ))
  )

  # sha prefix
  expect_snapshot(
    sy(async_git_resolve_ref(
      fake_git$url("/pak-test.git"),
      "3f3b0b4ee8a0ff"
    ))
  )

  # unknown sha prefix -> error
  expect_snapshot(
    error = TRUE,
    sy(async_git_resolve_ref(
      fake_git$url("/pak-test.git"),
      "badcafe"
    ))
  )

  mockery::stub(
    async_git_resolve_ref,
    "async_git_list_refs",
    function(...) {
      async_constant(list(
        refs = data_frame(
          ref = c("one", "two"),
          hash = c("badcafe1", "badcafe2")
        ),
        caps = character()
      ))
    }
  )

  expect_snapshot(
    error = TRUE,
    sy(async_git_resolve_ref(
      fake_git$url("/pak-test.git"),
      "badcafe"
    ))
  )
})

test_that("git_fetch_process_v1", {
  url <- "https://user:secret@example.com"
  expect_snapshot(
    error = TRUE,
    git_fetch_process_v1(list(), url, "badcafe")
  )

  expect_snapshot(
    error = TRUE,
    git_fetch_process_v1(list(list(type = "boo")), url, "badcafe")
  )
})

test_that("git_download_repo", {
  skip_on_cran()
  dir.create(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  git_download_repo(
    fake_git$url("/pak-test.git"),
    ref = "v1",
    output = file.path(tmp, "v1")
  )
  expect_snapshot(dir(tmp, recursive = TRUE))
})

test_that("unpack_packfile_repo", {
  # reordering objects in a PACK file does not matter
  ref <- "cefdc0eebcd7f757efb9a80652fd8aaf1a87508e"
  pack <- git_fetch(fake_git$url("/pak-test.git"), ref, blobs = TRUE)

  withr::local_seed(13L)
  pack <- sample(pack)

  output <- tempfile()
  on.exit(unlink(output, recursive = TRUE), add = TRUE)
  unpack_packfile_repo(pack, output, url = "url")
  expect_snapshot(
    sort(dir(output, recursive = TRUE))
  )
})

roxy_to_rd <- function(text) {
  roxy_text <- paste0(
    "#' @name foo\n",
    "#' @title title\n",
    "#' @details\n",
    "#' ",
    gsub("\n", "\n#' ", text),
    "\n",
    "#' @md\n",
    "NULL\n"
  )
  # We change the current package's name to make sure that we create
  # cross-references that are fully qualified.
  pkg <- asNamespace("roxygen2")$roxy_meta_get("current_package")
  asNamespace("roxygen2")$roxy_meta_set("current_package", "no-package")
  on.exit(
    asNamespace("roxygen2")$roxy_meta_set("current_package", pkg),
    add = TRUE
  )
  # Suppress the warning caused by the changed package name
  out <- suppressWarnings(asNamespace("roxygen2")$roc_proc_text(
    asNamespace("roxygen2")$rd_roclet(),
    roxy_text
  ))

  det <- out$foo$get_rd("details")
  sec <- out$foo$get_rd("section")
  if (det == "NULL" && sec == "NULL") {
    cli::cli_alert_warning("No content in {.fn roxy_to_rd}.")
    ""
  } else if (det == "NULL") {
    sec
  } else if (sec == "NULL") {
    if (startsWith(det, "\\details")) {
      det <- sub("^[\\]details[{]\n?", "", det)
      det <- sub("\n?[}]$", "", det)
    }
    det
  } else {
    cli::cli_alert_warning("\\details{{}} content ignored in {.fn roxy_to_rd}.")
    sec
  }
}

generate_config_docs <- function() {
  # for the dynamic help in pak
  config <- Filter(function(x) !is.null(x$docs), pkgdepends_config)
  nms <- names(config)
  dcs <- lapply(config, function(x) x[["docs_pak"]] %||% x[["docs"]])
  inc <- vlapply(config, function(x) x$pak %||% TRUE)
  nms <- nms[inc]
  dcs <- dcs[inc]

  rd <- lapply(dcs, roxy_to_rd)
  outp <- paste(utils::capture.output(print(rd)), collapse = "\n")
  mdfile <- "tools/doc/pak-config-docs.md"
  oldp <- read_char(mdfile)
  if (outp != oldp) {
    outfile <- file.path("inst/docs/pak-config-docs.rds")
    cli::cli_alert_info("Writing {.path {outfile}}")
    saveRDS(rd, outfile, version = 2)
    write_char(outp, mdfile)
  }

  # for roxygen2 in pkgdepends
  items <- map_named(config, function(name, entry) {
    paste0("* `", name, "`: ", entry$docs)
  })

  alldocs <- paste(items, collapse = "\n")
  alldocs
}

doc_share_rmd <- function(rmd, rds) {
  md <- sub("[.]Rmd$", ".md", rmd)
  if (md == rmd) {
    stop("Docs Rmd file must have extension `.Rmd`")
  }
  withr_with_envvar(c(THIS_IS_PAK = "true"), {
    cmd <- paste0("```{r child=\"", rmd, "\"}\n```\n")
    rd <- roxy_to_rd(cmd)
  })
  oldp <- read_char(md)
  if (oldp != rd) {
    cli::cli_alert_info("Writing {.path {rds}.}")
    saveRDS(rd, rds, version = 2)
    write_char(rd, md)
  }

  return("")
}

read_char <- function(path) {
  bin <- readBin(path, "raw", file.info(path)$size)
  txt <- rawToChar(bin)
  Encoding(txt) <- "UTF-8"
  txt
}

write_char <- function(x, path) {
  writeBin(charToRaw(x), path)
}

man_config_link <- function(txt) {
  paste0("\\link[=pkgdepends-config]{", txt, "}")
}

# From remotes
parse_submodules <- function(file) {
  if (grepl("\n", file)) {
    # fix windows line endings
    file <- gsub("\r\n", "\n", file, fixed = TRUE)
    x <- strsplit(file, "\n")[[1]]
  } else {
    x <- readLines(file)
  }

  # https://git-scm.com/docs/git-config#_syntax
  # Subsection names are case sensitive and can contain any characters except
  # newline and the null byte. Doublequote " and backslash can be included by
  # escaping them as \" and \\
  double_quoted_string_with_escapes <- '(?:\\\\.|[^"])*'

  # Otherwise extract section names
  section_names <- re_match(
    x,
    sprintf(
      '^[[:space:]]*\\[submodule "(?<submodule>%s)"\\][[:space:]]*$',
      double_quoted_string_with_escapes
    )
  )$submodule

  # If no sections found return the empty list
  if (all(is.na(section_names))) {
    return(list())
  }

  # Extract name = value
  # The variable names are case-insensitive, allow only alphanumeric characters
  # and -, and must start with an alphabetic character.
  variable_name <- "[[:alpha:]][[:alnum:]\\-]*"
  mapping_values <- re_match(
    x,
    sprintf(
      '^[[:space:]]*(?<name>%s)[[:space:]]*=[[:space:]]*(?<value>.*)[[:space:]]*$',
      variable_name
    )
  )

  values <- cbind(
    submodule = fill(section_names),
    mapping_values[c("name", "value")],
    stringsAsFactors = FALSE
  )
  values <- values[!is.na(mapping_values$.match), ]

  # path and valid url are required
  if (!all(c("path", "url") %in% values$name)) {
    warning(
      "Invalid submodule definition, skipping submodule installation",
      immediate. = TRUE,
      call. = FALSE
    )
    return(list())
  }

  # Roughly equivalent to tidyr::spread(values, name, value)
  res <- stats::reshape(
    values,
    idvar = "submodule",
    timevar = "name",
    v.name = "value",
    direction = "wide"
  )

  # Set the column names, reshape prepends `value.` to path, url and branch
  colnames(res) <- gsub("value[.]", "", colnames(res))

  # path and valid url are required
  if (any(is.na(res$url), is.na(res$path))) {
    warning(
      "Invalid submodule definition, skipping submodule installation",
      immediate. = TRUE,
      call. = FALSE
    )
    return(list())
  }

  # branch is optional
  if (!exists("branch", res)) {
    res$branch <- NA_character_
  }

  # Remove unneeded attribute
  attr(res, "reshapeWide") <- NULL

  # Remove rownames
  rownames(res) <- NULL

  res
}

# Adapted from https://stackoverflow.com/a/9517731/2055486
fill <- function(x) {
  not_missing <- !is.na(x)

  res <- x[not_missing]
  res[cumsum(not_missing)]
}

update_submodule <- function(url, path, branch) {
  synchronize(async_update_submodule(url, path, branch)) # nocov
}

async_update_submodule <- function(url, path, branch) {
  url
  path
  branch
  # if the directory already exists and not empty, we assume that
  # it was already downloaded. We still to update the submodules
  # recursively. This is problematic if a git download is interrupted
  # and then stated again with the same output, but that does not happen
  # during normal operation of pkgdepends, I think. A better solution
  # would be to download the submodule to a temporary directory, and if
  # successful, then move the temporary directory to the correct place.
  if (
    file.exists(path) &&
      length(dir(path, all.files = TRUE, no.. = TRUE)) > 0
  ) {
    # message(path, " exists")
    async_update_git_submodules(path)
  } else {
    if (is.null(branch) || is.na(branch)) {
      branch <- "HEAD"
    }
    # message("getting ", path)
    async_git_download_repo(
      url,
      ref = branch,
      output = path,
      submodules = TRUE
    )
  }
}

update_git_submodules_r <- function(path, subdir) {
  synchronize(async_update_git_submodules_r(path, subdir)) # nocov
}

async_update_git_submodules_r <- function(path, subdir) {
  subdir <- subdir %||% "."
  smfile <- file.path(path, ".gitmodules")
  if (!file.exists(smfile)) {
    return()
  }

  info <- parse_submodules(smfile)
  if (length(info) == 0) {
    return()
  }

  to_ignore <- in_r_build_ignore(
    info$path,
    file.path(path, subdir, ".Rbuildignore")
  )
  info <- info[!to_ignore, ]
  if (nrow(info) == 0) {
    return()
  }

  async_map(seq_len(nrow(info)), function(i) {
    async_update_submodule(
      info$url[i],
      file.path(path, info$path[i]),
      info$branch[i]
    )
  })$then(function() invisible())
}

update_git_submodules <- function(path) {
  synchronize(async_update_git_submodules(path))
}

async_update_git_submodules <- function(path) {
  smfile <- file.path(path, ".gitmodules")
  if (!file.exists(smfile)) {
    return()
  }

  info <- parse_submodules(smfile)
  if (nrow(info) == 0) {
    return()
  }

  async_map(seq_len(nrow(info)), function(i) {
    async_update_submodule(
      info$url[i],
      file.path(path, info$path[i]),
      info$branch[i]
    )
  })$then(function() invisible())
}

r_build_ignore_patterns <- c(
  "^\\.Rbuildignore$",
  "(^|/)\\.DS_Store$",
  "^\\.(RData|Rhistory)$",
  "~$",
  "\\.bak$",
  "\\.swp$",
  "(^|/)\\.#[^/]*$",
  "(^|/)#[^/]*#$",
  "^TITLE$",
  "^data/00Index$",
  "^inst/doc/00Index\\.dcf$",
  "^config\\.(cache|log|status)$",
  "(^|/)autom4te\\.cache$",
  "^src/.*\\.d$",
  "^src/Makedeps$",
  "^src/so_locations$",
  "^inst/doc/Rplots\\.(ps|pdf)$"
)

in_r_build_ignore <- function(paths, ignore_file) {
  ignore <- tryCatch(
    asNamespace("tools")$get_exclude_patterns(),
    error = function(e) r_build_ignore_patterns
  )

  if (file.exists(ignore_file)) {
    ignore <- c(ignore, readLines(ignore_file, warn = FALSE))
  }

  matches_ignores <- function(x) {
    any(vlapply(ignore, grepl, x, perl = TRUE, ignore.case = TRUE))
  }

  # We need to search for the paths as well as directories in the path, so
  # `^foo$` matches `foo/bar`
  should_ignore <- function(path) {
    any(vlapply(c(path, directories(path)), matches_ignores))
  }

  vlapply(paths, should_ignore)
}

directories <- function(paths) {
  dirs <- unique(dirname(paths))
  out <- dirs[dirs != "."]
  while (length(dirs) > 0 && any(dirs != ".")) {
    out <- unique(c(out, dirs[dirs != "."]))
    dirs <- unique(dirname(dirs))
  }
  sort(out)
}

withr::local_envvar(GITHUB_PAT = "FAIL")

test_that("parse_submodules", {
  sm <- test_path("fixtures/submodules.ini")
  expect_snapshot({
    parse_submodules(sm)
    parse_submodules(read_char(sm))
  })

  file.create(tmp <- tempfile())
  on.exit(unlink(tmp), add = TRUE)
  expect_snapshot({
    parse_submodules(tmp)
  })

  # invalid file
  sm2 <- test_path("fixtures/submodules2.ini")
  sm3 <- test_path("fixtures/submodules3.ini")
  expect_snapshot({
    parse_submodules(sm2)
    parse_submodules(sm3)
  })
})

test_that("git_download_repo with submodules", {
  skip_on_cran()
  if (Sys.which("git") == "") {
    skip("Needs git")
  }

  dir.create(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  output <- file.path(tmp, "v1")
  git_download_repo(
    fake_git$url("/pak-test.git"),
    output = output
  )

  # tag
  writeLines(
    c(
      "[submodule \"submod\"]",
      "\tpath = submod",
      paste0("\turl = ", fake_git$url("/submod")),
      "\tbranch = v2"
    ),
    file.path(output, ".gitmodules")
  )

  update_git_submodules(output)
  expect_snapshot(dir(tmp, recursive = TRUE))
  expect_snapshot(readLines(file.path(output, "submod", "README")))

  # HEAD
  unlink(file.path(output, "submod"), recursive = TRUE)
  writeLines(
    c(
      "[submodule \"submod\"]",
      "\tpath = submod",
      paste0("\turl = ", fake_git$url("/submod"))
    ),
    file.path(output, ".gitmodules")
  )

  update_git_submodules(output)
  # it will skip existing ones
  update_git_submodules(output)
  expect_snapshot(dir(tmp, recursive = TRUE))
  expect_snapshot(readLines(file.path(output, "submod", "README")))
})

test_that("git_download_repo R package with submodules", {
  skip_on_cran()
  if (Sys.which("git") == "") {
    skip("Needs git")
  }

  dir.create(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  output <- file.path(tmp, "v1")
  git_download_repo(
    fake_git$url("/pak-test.git"),
    output = output
  )

  # tag
  writeLines(
    c(
      "[submodule \"submod\"]",
      "\tpath = submod",
      paste0("\turl = ", fake_git$url("/submod")),
      "\tbranch = v2"
    ),
    file.path(output, ".gitmodules")
  )

  update_git_submodules_r(output, ".")
  expect_snapshot(dir(tmp, recursive = TRUE))
  expect_snapshot(readLines(file.path(output, "submod", "README")))

  # HEAD
  unlink(file.path(output, "submod"), recursive = TRUE)
  writeLines(
    c(
      "[submodule \"submod\"]",
      "\tpath = submod",
      paste0("\turl = ", fake_git$url("/submod"))
    ),
    file.path(output, ".gitmodules")
  )

  update_git_submodules_r(output, ".")
  # it will skip existing ones
  update_git_submodules_r(output, ".")
  expect_snapshot(dir(tmp, recursive = TRUE))
  expect_snapshot(readLines(file.path(output, "submod", "README")))

  # no submodule file
  unlink(file.path(output, "submod"), recursive = TRUE)
  unlink(file.path(output, ".gitmodules"))
  update_git_submodules_r(output, ".")
  expect_snapshot(dir(tmp, recursive = TRUE))

  # Empty submodule file
  unlink(file.path(output, "submod"), recursive = TRUE)
  writeLines(character(), file.path(output, ".gitmodules"))
  update_git_submodules_r(output, ".")
  expect_snapshot(dir(tmp, recursive = TRUE))
})

test_that("git_download_repo R package with ignored submodule", {
  skip_on_cran()
  if (Sys.which("git") == "") {
    skip("Needs git")
  }

  dir.create(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  output <- file.path(tmp, "v1")
  git_download_repo(
    fake_git$url("/pak-test.git"),
    ref = "build-ignore",
    output = output
  )

  update_git_submodules_r(output, ".")
  expect_snapshot(dir(tmp, recursive = TRUE, all.files = TRUE, no.. = TRUE))
})

test_that("directories", {
  dir.create(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  wd <- getwd()
  on.exit(setwd(wd), add = TRUE)
  setwd(tmp)

  mkdirp("a/b/c/d")
  expect_snapshot({
    directories("a")
    directories("a/b/c/d")
  })
})

get_dep_type_from_path <- function(paths, orig = NULL) {
  tps <- rep("prod", length(paths))
  tps[paths == "man/roxygen/meta.R"] <- "dev"
  tps[startsWith(paths, "tests/") | startsWith(paths, "test/")] <- "test"
  if (!is.null(orig)) {
    # for DESCRIPTION we detect the type from the file itself
    dsc <- basename(paths) == "DESCRIPTION"
    tps[dsc] <- orig[dsc]
  }
  tps
}

get_dep_type_from_description_field <- function(fields) {
  tps <- rep("dev", length(fields))
  tps[fields %in% c("Depends", "Imports", "LinkingTo")] <- "prod"
  tps[fields %in% c("Suggests", "Enhanced")] <- "test"
  tps[fields == "Config/Needs/coverage"] <- "test"
  tps
}

ghrepo <- local({
  ghrepo_update <- function(
    repo,
    subdir,
    release_org = "cran",
    source_repo = "https://cran.r-project.org",
    ignore_build_errors = TRUE,
    packages = NULL
  ) {
    subdir <- sub("/+$", "", subdir)
    if (endsWith(subdir, "src/contrib")) {
      subdir <- sub("src/contrib$", "", subdir)
    }

    cli::cli_h2("Collect information")
    source_pkgs <- get_cran_packages(source_repo)
    mirror_pkgs <- get_mirror_packages(repo, subdir)

    to_build <- get_outdated_packages(source_pkgs, mirror_pkgs, packages)
    if (length(to_build) == 0) {
      return(invisible(character()))
    }

    cli::cli_h2("Install packages")
    dir.create(lib <- tempfile("pkgdepends-lib-"))
    inst <- install_pkgs(
      c(to_build, if (ignore_build_errors) "*=?ignore-build-errors"),
      library = lib,
      source_repo = source_repo,
      repo = repo,
      subdir = subdir
    )
    inst <- keep_updated(inst, mirror_pkgs)
    # I don't think this can happen....
    if (nrow(inst) == 0) {
      return(invisible(character()))
    }

    cli::cli_h2("Build binary packages")
    inst <- build_pkgs(inst, lib)
    # drop the ones that failed to build
    inst <- inst[!is.na(inst$built_path), ]

    cli::cli_par()
    cli::cli_h2("Update package mirror at GH (if needed)")
    # make sure these packages are mirrored on GH
    for (i in seq_along(inst$package)) {
      ghmirror$update(inst$package[[i]], source_repo = source_repo)
    }

    if (nrow(inst) > 0) {
      cli::cli_par()
      cli::cli_h2("Upload new release assets to GH")
      upload_releases(inst, release_org)
    }

    if (nrow(inst) > 0) {
      cli::cli_par()
      cli::cli_h2("Update repository metadata at GH")
      # unfortunate name collision
      repo_slug <- repo
      rm(repo)
      contrib_url <- utils::contrib.url(repos = "", type = .Platform$pkgType)
      repo$update_gh(
        repo_slug,
        paste0(subdir, contrib_url),
        inst$built_path,
        release_org = release_org
      )
    }

    cli::cli_par()
    cli::cli_h2("Build report")
    print_report(inst, mirror_pkgs)

    invisible(inst)
  }

  # -----------------------------------------------------------------------
  # Internal functions

  parse_build_number <- function(x) {
    x <- basename(x)
    mch <- re_match(x, "_b(?<build>[0-9]+)_")
    build <- as.integer(mch$build) + 1L
    build[is.na(build)] <- 1L
    build
  }

  get_cran_packages <- function(repo) {
    proc <- cli::cli_process_start("Getting packages from {.url {repo}}.")
    r_source <- suppressMessages(pkgcache::cranlike_metadata_cache$new(
      platforms = "source",
      bioc = FALSE,
      cran_mirror = repo,
      repos = NULL,
      update_after = as.difftime(30, units = "mins")
    ))
    suppressMessages(r_source$update())
    res <- r_source$list()

    cli::cli_process_done(proc)
    res
  }

  get_mirror_repo_url <- function(repo, subdir) {
    subdir <- sub("/src/contrib/?$", "", subdir)
    paste0(
      "https://raw.githubusercontent.com/",
      repo,
      "/main/",
      subdir
    )
  }

  get_mirror_packages <- function(repo, subdir) {
    proc <- cli::cli_process_start(
      "Getting packages from {.emph {repo}/{subdir}}."
    )
    platform <- if (.Platform$pkgType == "source") {
      "source"
    } else {
      pkgcache::current_r_platform()
    }
    r_mirror <- suppressMessages(pkgcache::cranlike_metadata_cache$new(
      platforms = platform,
      bioc = FALSE,
      cran_mirror = get_mirror_repo_url(repo, subdir),
      repos = NULL,
      update_after = as.difftime(1, units = "mins")
    ))
    suppressMessages(r_mirror$update())
    res <- r_mirror$list()

    cli::cli_process_done(proc)
    res
  }

  get_outdated_packages <- function(source_pkgs, mirror_pkgs, extra) {
    # these are included already
    extra <- setdiff(extra, mirror_pkgs$package)

    # TODO: update for breaking changes in R as well
    to_build <- intersect(mirror_pkgs$package, source_pkgs$package)
    r_mirror_ver <- mirror_pkgs$version[match(to_build, mirror_pkgs$package)]
    r_source_ver <- source_pkgs$version[match(to_build, source_pkgs$package)]
    to_build <- to_build[package_version(r_mirror_ver) < r_source_ver]

    res <- c(to_build, extra)
    if (length(res) == 0) {
      cli::cli_alert_info("All packages are current, no updates are needed")
    } else {
      cli::cli_alert_info(
        "Need to build {length(res)} package{?s}: {.pkg {res}}."
      )
    }
    res
  }

  install_pkgs <- function(pkgs, library, source_repo, repo, subdir) {
    opt <- options(
      repos = c(
        RHUB = get_mirror_repo_url(repo, subdir),
        CRAN = source_repo
      )
    )
    on.exit(options(opt), add = TRUE)
    prop <- new_pkg_installation_proposal(
      pkgs,
      config = list(
        library = library,
        platforms = c(current_r_platform(), "source"),
        use_bioconductor = FALSE,
        metadata_update_after = as.difftime(5, units = "mins")
      )
    )
    prop$set_solve_policy("upgrade")
    prop$solve()
    prop$show_solution()
    prop$download()
    prop$install_sysreqs()
    inst <- prop$install()

    # Drop packages that were kept in a different library. These are
    # recommended packages currently
    was_inst <- inst$package[inst$type == "installed"]
    drop <- setdiff(was_inst, dir(library))
    if (length(drop) > 0) {
      inst <- inst[!inst$package %in% drop, , drop = FALSE]
    }
    inst
  }

  keep_updated <- function(inst, mirror_pkgs) {
    # we installed these packages, now we update packages that are not in
    # the mirror, or we just built a newer version for them.
    # TODO: update for breaking changes in R as well
    to_add <- setdiff(inst$package, mirror_pkgs$package)
    mirrored <- intersect(inst$package, mirror_pkgs$package)
    inst_ver <- inst$version[match(mirrored, inst$package)]
    mirror_ver <- mirror_pkgs$version[match(mirrored, mirror_pkgs$package)]
    to_update <- mirrored[package_version(inst_ver) > mirror_ver]
    todo <- c(to_add, to_update)
    inst <- inst[match(todo, inst$package), , drop = FALSE]

    # get build numbers
    inst$buildnum <- rep(1L, nrow(inst))
    inst$buildnum[todo %in% mirrored] <-
      parse_build_number(mirror_pkgs$target[match(
        todo[todo %in% mirrored],
        mirror_pkgs$package
      )])
    inst
  }

  build_pkgs <- function(inst, library) {
    files <- rep(NA_character_, nrow(inst))
    for (i in seq_along(inst$package)) {
      if (
        length(inst$build_error[[i]]) &&
          !identical(inst$build_error[[i]], FALSE)
      ) {
        cli::cli_alert_warning(
          "Failed to build package {.pkg {inst$package[[i]]}}."
        )
      } else {
        proc <- cli::cli_process_start(
          "Building binary for package {.pkg {inst$package[[i]]}}."
        )
        files[i] <- pkg_build(
          inst$package[[i]],
          library = library,
          build_number = inst$buildnum[[i]]
        )
        cli::cli_process_done(proc)
      }
    }
    inst$built_path <- files
    inst
  }

  upload_releases <- function(inst, release_org) {
    # upload packages to releases
    for (i in seq_along(inst$package)) {
      slug <- paste0(release_org, "/", inst$package[[i]])
      ver <- inst$version[[i]]
      rels <- try(ghr$list(slug))
      if (inherits(rels, "try-error")) {
        Sys.sleep(5)
        rels <- ghr$list(slug)
      }
      if (!ver %in% rels$tag_name) {
        cli::cli_alert_info("Creating GH release {slug} {ver}.")
        ghr$create(slug, ver)
      }
      cli::cli_alert_info("Adding release asset for {slug} {ver}.")
      tryCatch(
        ghr$add_asset(slug, inst$built_path[i], ver),
        error = function(err) {
          cli::cli_alert_info("Try adding release asset again.")
          Sys.sleep(60)
          ghr$add_asset(slug, inst$built_path[i], ver)
        }
      )
      Sys.sleep(5)
    }
  }

  print_report <- function(inst, mirror_pkgs) {
    pkg <- format(inst$package)
    oldver <- ifelse(
      inst$package %in% mirror_pkgs$package,
      mirror_pkgs$version[match(inst$package, mirror_pkgs$package)],
      ""
    )

    inst$build_time <- inst$build_time %||% NA_real_
    inst$build_time <- as.double(inst$build_time)
    inst$build_time <- as.difftime(inst$build_time, units = "secs")
    cols <- data_frame(
      package = inst$package,
      "old" = oldver,
      "-" = "->",
      "new" = inst$version,
      "build" = paste0("b", inst$buildnum),
      "build time" = vcapply(inst$build_time, format_time$pretty_dt)
    )

    print_table(cols)
  }

  print_table <- function(cols) {
    cols <- as.list(cols)
    for (i in seq_along(cols)) {
      if (names(cols)[i] == "-") {
        names(cols)[i] <- ""
      }
      cols[[i]] <- format(c(names(cols)[i], "", cols[[i]]))
      if (names(cols)[i] != "") {
        cols[[i]][2] <- strrep("-", nchar(cols[[i]][1]))
      }
    }
    writeLines(do.call(paste, cols))
  }

  # -----------------------------------------------------------------------

  structure(
    list(
      .internal = environment(),

      update = ghrepo_update
    )
  )
})

# -------------------------------------------------------------------------

#' Update a CRAN-like repository of binary packages at GitHub
#'
#' These functions are currently experimental.
#'
#' @details
#'
#' ## Update a CRAN-like repository of binary packages at GitHub
#'
#' `ghrepo$update()` updates a binary package mirror.
#'
#' ### Usage
#' ```
#' ghrepo$update(
#'   repo,
#'   subdir,
#'   release_org = "cran",
#'   source_repo = "https://cran.r-project.org",
#'   packages = NULL
#' )
#' ```
#'
#' ### Arguments
#'
#' * `repo`: GitHub slug, e.g. `r-hub/repos`.
#' * `subdir`: subdirectory in the GitHub repository, where the R package
#'   metadata should be updated. It must exist in the repository.
#'   If it does not have `PACKAGES*` files, then they will be created.
#' * `release_org`: GitHub organization or user name where the packages
#'   will be published as releases.
#' * `source_repo`: A CRAN-like repository, where source packages are
#'   taken from.
#' * `packages`: A character vector of package names to add to the binary
#'   repository, in addition to updating the ones that are already there.
#'
# -------------------------------------------------------------------------
#'
#' @name ghrepo
#' @keywords internal
#' @export

ghrepo

should_show_progress_bar <- function() {
  # Option takes precedence
  opt <- getOption("pkg.show_progress", NULL)
  if (!is.null(opt)) {
    return(isTRUE(opt))
  }

  # FALSE on CI, this should be in cli, probably
  if (Sys.getenv("CI", "") != "") {
    return(FALSE)
  }

  # FALSE in knitr, this should be in cli, probably
  if (isTRUE(getOption("knitr.in.progress"))) {
    return(FALSE)
  }

  # FALSE in testthat as well, this should NOT be in cli, maybe
  if (identical(Sys.getenv("TESTTHAT"), "true")) {
    return(FALSE)
  }

  # Otherwise trust cli
  cli::is_dynamic_tty()
}

test_that("should_show_progress_bar", {
  withr::local_options(pkg.show_progress = TRUE)
  expect_true(should_show_progress_bar())

  withr::local_options(pkg.show_progress = FALSE)
  expect_false(should_show_progress_bar())

  withr::local_options(pkg.show_progress = NULL)
  withr::local_envvar(CI = "true")
  expect_false(should_show_progress_bar())

  withr::local_envvar(CI = NA_character_)
  withr::local_options(knitr.in.progress = TRUE)
  expect_false(should_show_progress_bar())

  withr::local_options(knitr.in.progress = NULL)
  withr::local_envvar(TESTTHAT = "true")
  expect_false(should_show_progress_bar())

  withr::local_envvar(TESTTHAT = NA_character_)
  mockery::stub(should_show_progress_bar, "cli::is_dynamic_tty", TRUE)
  expect_true(should_show_progress_bar())
})

# pkg_dep_types_*

    Code
      pkg_dep_types_hard()
    Output
      [1] "Depends"   "Imports"   "LinkingTo"

---

    Code
      pkg_dep_types_soft()
    Output
      [1] "Suggests" "Enhances"

---

    Code
      pkg_dep_types()
    Output
      [1] "Depends"   "Imports"   "LinkingTo" "Suggests"  "Enhances" 

# parse_deps

    Code
      parse_deps(character(), character())
    Output
      list()

---

    Code
      parse_deps("foo", "Imports")
    Output
      [[1]]
      # A data frame: 1 x 4
        type    package op    version
      * <chr>   <chr>   <chr> <chr>  
      1 Imports foo     ""    ""     
      

---

    Code
      parse_deps(c("foo", "foo"), c("Imports", "LinkingTo"))
    Output
      [[1]]
      # A data frame: 1 x 4
        type    package op    version
      * <chr>   <chr>   <chr> <chr>  
      1 Imports foo     ""    ""     
      
      [[2]]
      # A data frame: 1 x 4
        type      package op    version
      * <chr>     <chr>   <chr> <chr>  
      1 LinkingTo foo     ""    ""     
      

---

    Code
      parse_deps("foo (>= 1.0.0)", "Imports")
    Output
      [[1]]
      # A data frame: 1 x 4
        type    package op    version
      * <chr>   <chr>   <chr> <chr>  
      1 Imports foo     >=    1.0.0  
      

---

    Code
      parse_deps("foo (>= 1.0.0), bar", "Imports")
    Output
      [[1]]
      # A data frame: 2 x 4
        type    package op    version
      * <chr>   <chr>   <chr> <chr>  
      1 Imports foo     ">="  "1.0.0"
      2 Imports bar     ""    ""     
      

---

    Code
      parse_deps("foo (>= 1.0.0)\n bar", "Imports")
    Output
      [[1]]
      # A data frame: 1 x 4
        type    package op    version
      * <chr>   <chr>   <chr> <chr>  
      1 Imports <NA>    <NA>  <NA>   
      

---

    Code
      parse_deps("foo (>= \n 1.0.0)\n bar", "Imports")
    Output
      [[1]]
      # A data frame: 1 x 4
        type    package op    version
      * <chr>   <chr>   <chr> <chr>  
      1 Imports <NA>    <NA>  <NA>   
      

---

    Code
      parse_deps("R (== 4.2.1)", "Depends")
    Output
      [[1]]
      # A data frame: 1 x 4
        type    package op    version
      * <chr>   <chr>   <chr> <chr>  
      1 Depends R       ==    4.2.1  
      

---

    Code
      parse_deps("R (< 4.2.1)", "Depends")
    Output
      [[1]]
      # A data frame: 1 x 4
        type    package op    version
      * <chr>   <chr>   <chr> <chr>  
      1 Depends R       <     4.2.1  
      

---

    Code
      parse_deps("R (> 4.2.1)", "Depends")
    Output
      [[1]]
      # A data frame: 1 x 4
        type    package op    version
      * <chr>   <chr>   <chr> <chr>  
      1 Depends R       >     4.2.1  
      

---

    Code
      parse_deps("R (<= 4.2.1)", "Depends")
    Output
      [[1]]
      # A data frame: 1 x 4
        type    package op    version
      * <chr>   <chr>   <chr> <chr>  
      1 Depends R       <=    4.2.1  
      

---

    Code
      parse_deps("foo, stats", "Imports")
    Output
      [[1]]
      # A data frame: 1 x 4
        type    package op    version
      * <chr>   <chr>   <chr> <chr>  
      1 Imports foo     ""    ""     
      

# resolve_ref_deps

    Code
      resolve_ref_deps(dsc$get_deps(), NA_character_, NULL)
    Output
      # A data frame: 7 x 5
        ref        type     package    op    version
        <chr>      <chr>    <chr>      <chr> <chr>  
      1 covr       Suggests covr       ""    ""     
      2 jsonlite   Suggests jsonlite   ""    ""     
      3 testthat   Suggests testthat   ">="  "3.1.0"
      4 assertthat Imports  assertthat ""    ""     
      5 curl       Imports  curl       ""    ""     
      6 R6         Imports  R6         ""    ""     
      7 rlang      Imports  rlang      ">="  "1.0.0"

---

    Code
      resolve_ref_deps(dsc$get_deps(), dsc$get("Remotes"), dsc$get(
        extra_config_fields(dsc$fields())))
    Output
      # A data frame: 9 x 5
        ref           type                 package    op    version
        <chr>         <chr>                <chr>      <chr> <chr>  
      1 r-lib/covr    Suggests             covr       ""    ""     
      2 jsonlite      Suggests             jsonlite   ""    ""     
      3 testthat      Suggests             testthat   ">="  "3.1.0"
      4 assertthat    Imports              assertthat ""    ""     
      5 curl          Imports              curl       ""    ""     
      6 R6            Imports              R6         ""    ""     
      7 rlang         Imports              rlang      ">="  "1.0.0"
      8 pkgdown       Config/Needs/website pkgdown    ""    ""     
      9 r-lib/downlit Config/Needs/website downlit    ""    ""     

# resolve_ref_deps, package name from remote

    Code
      resolve_ref_deps(dsc$get_deps(), dsc$get("Remotes"), dsc$get(
        extra_config_fields(dsc$fields())))
    Output
      # A data frame: 2 x 5
        ref              type                 package op    version
        <chr>            <chr>                <chr>   <chr> <chr>  
      1 foo=r-lib/bar    Imports              foo     ""    ""     
      2 foobar=r-lib/baz Config/Needs/website foobar  ""    ""     

---

    Code
      resolve_ref_deps(dsc$get_deps(), dsc$get("Remotes"), NULL)
    Condition
      Error:
      ! ! Cannot determine package name for 1 package: "url::http://example.com".
      i Maybe you need to add a `<packagename>=` prefix?

# as_pkg_dependencies

    Code
      as_pkg_dependencies(TRUE)
    Output
      $direct
      [1] "Depends"   "Imports"   "LinkingTo" "Suggests" 
      
      $indirect
      [1] "Depends"   "Imports"   "LinkingTo"
      

---

    Code
      as_pkg_dependencies("all")
    Output
      $direct
      [1] "Depends"   "Imports"   "LinkingTo" "Suggests"  "Enhances" 
      
      $indirect
      [1] "Depends"   "Imports"   "LinkingTo"
      

---

    Code
      as_pkg_dependencies("hard")
    Output
      $direct
      [1] "Depends"   "Imports"   "LinkingTo"
      
      $indirect
      [1] "Depends"   "Imports"   "LinkingTo"
      

---

    Code
      as_pkg_dependencies(FALSE)
    Output
      $direct
      character(0)
      
      $indirect
      character(0)
      

---

    Code
      as_pkg_dependencies(NA)
    Output
      $direct
      [1] "Depends"   "Imports"   "LinkingTo"
      
      $indirect
      [1] "Depends"   "Imports"   "LinkingTo"
      

---

    Code
      as_pkg_dependencies(list(direct = "Imports", indirect = c("Imports", "Suggests")))
    Output
      $direct
      [1] "Imports"
      
      $indirect
      [1] "Imports"  "Suggests"
      

---

    Code
      as_pkg_dependencies(c("Depends", "Imports", "LinkingTo"))
    Output
      $direct
      [1] "Depends"   "Imports"   "LinkingTo"
      
      $indirect
      [1] "Depends"   "Imports"   "LinkingTo"
      

---

    Code
      as_pkg_dependencies(list(direct = "all", indirect = "hard"))
    Output
      $direct
      [1] "Depends"   "Imports"   "LinkingTo" "Suggests"  "Enhances" 
      
      $indirect
      [1] "Depends"   "Imports"   "LinkingTo"
      

---

    Code
      as_pkg_dependencies(list(direct = "all", indirect = "all"))
    Output
      $direct
      [1] "Depends"   "Imports"   "LinkingTo" "Suggests"  "Enhances" 
      
      $indirect
      [1] "Depends"   "Imports"   "LinkingTo" "Suggests"  "Enhances" 
      

---

    Code
      as_pkg_dependencies(list(direct = "all", indirect = c("hard", "soft")))
    Output
      $direct
      [1] "Depends"   "Imports"   "LinkingTo" "Suggests"  "Enhances" 
      
      $indirect
      [1] "Depends"   "Imports"   "LinkingTo" "Suggests"  "Enhances" 
      

# failure in predicate

    Code
      assert_that(assert())
    Condition
      Error:
      ! ! isTRUE(x = FALSE) is not true

# assert_error

    Code
      fn()
    Output
      <extraClass/assertError/rlib_error_3_0/rlib_error/error>
      Error in `fn()`:
      ! FALSE is not TRUE

# assertion returns invalid value

    Code
      assert_that(2 * 2)
    Condition
      Error:
      ! ! `assert_that()`: assertion must return a logical value.
      i it returned a number instead.

---

    Code
      assert_that(c(TRUE, FALSE))
    Condition
      Error:
      ! ! `assert_that()`: assertion must return a scalar.
      i it returned a vector of length 2.

---

    Code
      assert_that(NA)
    Condition
      Error:
      ! ! `assert_that()`: assertion must not return `NA`.

# default messages

    Code
      r_process()
    Output
      > fn <- function(x) assert_that(x == 1)
      > fn(2)
      [1m[33mError[39m[22m in `fn(2)`:
      [33m![39m [1m[22m`x` must equal `1`.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(x < 1)
      > fn(2)
      [1m[33mError[39m[22m in `fn(2)`:
      [33m![39m [1m[22m`x` must be less than `1`.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(x > 2)
      > fn(1)
      [1m[33mError[39m[22m in `fn(1)`:
      [33m![39m [1m[22m`x` must be greater than `2`.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(x >= 2)
      > fn(1)
      [1m[33mError[39m[22m in `fn(1)`:
      [33m![39m [1m[22m`x` must be greater than or equal to `2`.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(x <= 1)
      > fn(2)
      [1m[33mError[39m[22m in `fn(2)`:
      [33m![39m [1m[22m`x` must be less than or equal to `1`.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(x != 1)
      > fn(2)

---

    Code
      r_process()
    Output
      > fn <- function(x) assert_that(is.atomic(x))
      > fn(mtcars)
      [1m[33mError[39m[22m in `fn(mtcars)`:
      [33m![39m [1m[22m`x` must be an atomic vector.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.character(x))
      > fn(1:2)
      [1m[33mError[39m[22m in `fn(1:2)`:
      [33m![39m [1m[22m`x` must be a character vector.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.complex(x))
      > fn(1:5)
      [1m[33mError[39m[22m in `fn(1:5)`:
      [33m![39m [1m[22m`x` must be a complex vector.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.double(x))
      > fn(letters)
      [1m[33mError[39m[22m in `fn(letters)`:
      [33m![39m [1m[22m`x` must be a numeric vector.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.integer(x))
      > fn(letters)
      [1m[33mError[39m[22m in `fn(letters)`:
      [33m![39m [1m[22m`x` must be an integer vector.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.numeric(x))
      > fn(letters)
      [1m[33mError[39m[22m in `fn(letters)`:
      [33m![39m [1m[22m`x` must be a numeric or integer vector.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.raw(x))
      > fn(1:10)
      [1m[33mError[39m[22m in `fn(1:10)`:
      [33m![39m [1m[22m`x` must be a raw vector.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.vector(x))
      > fn(mtcars)
      [1m[33mError[39m[22m in `fn(mtcars)`:
      [33m![39m [1m[22m`x` must be an atomic vector without attributes.
      [90mType .Last.error to see the more details.[39m

---

    Code
      r_process()
    Output
      > fn <- function(x) assert_that(is.factor(x))
      > fn(letters)
      [1m[33mError[39m[22m in `fn(letters)`:
      [33m![39m [1m[22m`x` must be a factor.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.ordered(x))
      > fn(factor(letters))
      [1m[33mError[39m[22m in `fn(factor(letters))`:
      [33m![39m [1m[22m`x` must be an ordered factor.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.array(x))
      > fn(factor(letters))
      [1m[33mError[39m[22m in `fn(factor(letters))`:
      [33m![39m [1m[22m`x` must be an array.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.ordered(x))
      > fn(factor(letters))
      [1m[33mError[39m[22m in `fn(factor(letters))`:
      [33m![39m [1m[22m`x` must be an ordered factor.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.array(x))
      > fn(letters)
      [1m[33mError[39m[22m in `fn(letters)`:
      [33m![39m [1m[22m`x` must be an array.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.data.frame(x))
      > fn(1:10)
      [1m[33mError[39m[22m in `fn(1:10)`:
      [33m![39m [1m[22m`x` must be a data frame.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.list(x))
      > fn("foobar")
      [1m[33mError[39m[22m in `fn("foobar")`:
      [33m![39m [1m[22m`x` must be a list.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.matrix(x))
      > fn(letters)
      [1m[33mError[39m[22m in `fn(letters)`:
      [33m![39m [1m[22m`x` must be a matrix.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.null(x))
      > fn("not")
      [1m[33mError[39m[22m in `fn("not")`:
      [33m![39m [1m[22m`x` must be `NULL`.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.environment(x))
      > fn(list())
      [1m[33mError[39m[22m in `fn(list())`:
      [33m![39m [1m[22m`x` must be an environment.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.function(x))
      > fn("clearly noy")
      [1m[33mError[39m[22m in `fn("clearly noy")`:
      [33m![39m [1m[22m`x` must be a function.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.promitive(x))
      > fn(asesrt_that)
      Error in is.promitive(x) : could not find function "is.promitive"
      > fn <- function(x) assert_that(is.call(x))
      > fn("not")
      [1m[33mError[39m[22m in `fn("not")`:
      [33m![39m [1m[22m`x` must be a quoted call.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.expression(x))
      > fn(mtcars)
      [1m[33mError[39m[22m in `fn(mtcars)`:
      [33m![39m [1m[22m`x` must be an expression object.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.name(x))
      > fn("not really")
      [1m[33mError[39m[22m in `fn("not really")`:
      [33m![39m [1m[22m`x` must be a name.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.pairlist(x))
      > fn(1:10)
      [1m[33mError[39m[22m in `fn(1:10)`:
      [33m![39m [1m[22m`x` must be a pairlist.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.recursive(x))
      > fn(1)
      [1m[33mError[39m[22m in `fn(1)`:
      [33m![39m [1m[22m`x` must be a recursive object.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(is.symbol(x))
      > fn("almost")
      [1m[33mError[39m[22m in `fn("almost")`:
      [33m![39m [1m[22m`x` must be a name.
      [90mType .Last.error to see the more details.[39m

---

    Code
      r_process()
    Output
      > fn <- function(x, y) assert_that(x && y)
      > fn(is.integer(1L), is.integer("a"))
      [1m[33mError[39m[22m in `fn(is.integer(1L), is.integer("a"))`:
      [33m![39m [1m[22m`x` and `y` must both be true.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x, y) assert_that(x || y)
      > fn(is.integer("b"), is.integer("a"))
      [1m[33mError[39m[22m in `fn(is.integer("b"), is.integer("a"))`:
      [33m![39m [1m[22mOne of `x` and `y` must be true.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(any(x))
      > fn(rep(FALSE, 10))
      [1m[33mError[39m[22m in `fn(rep(FALSE, 10))`:
      [33m![39m [1m[22mAt least one of `x` must be true.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(all(x))
      > fn(c(FALSE, TRUE, TRUE))
      [1m[33mError[39m[22m in `fn(c(FALSE, TRUE, TRUE))`:
      [33m![39m [1m[22mAll of `x` must be true.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x) assert_that(file.exists(x))
      > fn("/file7e0fe1739e0")
      [1m[33mError[39m[22m in `fn("/file7e0fe1739e0")`:
      [33m![39m [1m[22mPath `x` must exist.
      [90mType .Last.error to see the more details.[39m
      > fn <- function(x, y) assert_that(identical(x, y))
      > fn(1, 1L)
      [1m[33mError[39m[22m in `fn(1, 1L)`:
      [33m![39m [1m[22m`x` must be identical to `y`.
      [90mType .Last.error to see the more details.[39m

# long call is truncated

    Code
      fail_default(call("==", strrep("-", 100), strrep("x", 100)))
    Output
      [1] "\"----------------------------------------------------------------------------------------------------\" == ... is not true"

# has_attr

    Code
      assert_that(has_attr(1L, "foobar"))
    Condition
      Error:
      ! ! `1L` must have attribute `foobar`.

---

    Code
      assert_that(1L %has_attr% "foobar")
    Condition
      Error:
      ! ! `1L` must have attribute `foobar`.

# custom message

    Code
      fn(1:10)
    Condition
      Error:
      ! ! `name` must be a name (character scalar).
      i It is an integer vector.

# create_lockfile

    Code
      lock$packages
    Output
         ref package version     type direct binary dependencies vignettes
      1 pkg1    pkg1   1.0.0 standard  FALSE  FALSE                  FALSE
      2 pkg2    pkg2   1.0.0 standard  FALSE  FALSE         pkg1     FALSE
      3 pkg3    pkg3   1.0.0 standard   TRUE  FALSE         pkg2     FALSE
        needscompilation metadata.RemoteType metadata.RemotePkgRef metadata.RemoteRef
      1            FALSE            standard                  pkg1               pkg1
      2            FALSE            standard                  pkg2               pkg2
      3            FALSE            standard                  pkg3               pkg3
           metadata.RemoteRepos metadata.RemotePkgPlatform metadata.RemoteSha
      1 http://127.0.0.1:<port>/                     source              1.0.0
      2 http://127.0.0.1:<port>/                     source              1.0.0
      3 http://127.0.0.1:<port>/                     source              1.0.0
                                                                                                                          sources
      1 http://127.0.0.1:<port>//src/contrib/pkg1_1.0.0.tar.gz, http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_1.0.0.tar.gz
      2 http://127.0.0.1:<port>//src/contrib/pkg2_1.0.0.tar.gz, http://127.0.0.1:<port>//src/contrib/Archive/pkg2/pkg2_1.0.0.tar.gz
      3 http://127.0.0.1:<port>//src/contrib/pkg3_1.0.0.tar.gz, http://127.0.0.1:<port>//src/contrib/Archive/pkg3/pkg3_1.0.0.tar.gz
                               target platform rversion directpkg
      1 src/contrib/pkg1_1.0.0.tar.gz   source        *     FALSE
      2 src/contrib/pkg2_1.0.0.tar.gz   source        *     FALSE
      3 src/contrib/pkg3_1.0.0.tar.gz   source        *      TRUE
                                                                  sha256 filesize
      1 <sha>      100
      2 <sha>      100
      3 <sha>      100
                          dep_types params install_args repotype
      1 Depends, Imports, LinkingTo   NULL                  cran
      2 Depends, Imports, LinkingTo   NULL                  cran
      3 Depends, Imports, LinkingTo   NULL                  cran

# sysreqs_platforms

    Code
      print(sysreqs_platforms(), n = Inf)
    Output
      # A data frame: 19 x 9
         name                     os    id                  distribution version version_match update_command    install_command                  query_command
         <chr>                    <chr> <chr>               <chr>        <chr>   <chr>         <chr>             <chr>                            <chr>        
       1 Ubuntu Linux             linux ubuntu              ubuntu       *       <NA>          apt-get -y update apt-get -y install               dpkg-query   
       2 Debian Linux             linux debian              debian       *       <NA>          apt-get -y update apt-get -y install               dpkg-query   
       3 CentOS Linux             linux centos              centos       *       major         <NA>              yum install -y                   rpm          
       4 Rocky Linux              linux rocky               rockylinux   *       major         <NA>              dnf install -y                   rpm          
       5 Rocky Linux              linux rockylinux          rockylinux   *       major         <NA>              dnf install -y                   rpm          
       6 AlmaLinux                linux almalinux           rockylinux   *       major         <NA>              dnf install -y                   rpm          
       7 Red Hat Enterprise Linux linux rhel                redhat       6       major         <NA>              yum install -y                   rpm          
       8 Red Hat Enterprise Linux linux rhel                redhat       7       major         <NA>              yum install -y                   rpm          
       9 Red Hat Enterprise Linux linux rhel                redhat       *       major         <NA>              dnf install -y                   rpm          
      10 Red Hat Enterprise Linux linux redhat              redhat       6       major         <NA>              yum install -y                   rpm          
      11 Red Hat Enterprise Linux linux redhat              redhat       7       major         <NA>              yum install -y                   rpm          
      12 Red Hat Enterprise Linux linux redhat              redhat       *       major         <NA>              dnf install -y                   rpm          
      13 Fedora Linux             linux fedora              fedora       *       <NA>          <NA>              dnf install -y                   rpm          
      14 openSUSE Linux           linux opensuse            opensuse     *       <NA>          <NA>              zypper --non-interactive install rpm          
      15 openSUSE Linux           linux opensuse-leap       opensuse     *       <NA>          <NA>              zypper --non-interactive install rpm          
      16 openSUSE Linux           linux opensuse-tumbleweed opensuse     *       <NA>          <NA>              zypper --non-interactive install rpm          
      17 SUSE Linux Enterprise    linux sles                sle          *       <NA>          <NA>              zypper --non-interactive install rpm          
      18 SUSE Linux Enterprise    linux sle                 sle          *       <NA>          <NA>              zypper --non-interactive install rpm          
      19 Alpine Linux             linux alpine              alpine       *       minor         <NA>              apk add --no-cache               apk          

# sysreqs_db_list

    Code
      lst$packages
    Output
      [[1]]
      [1] "libcairo2-dev"
      
      [[2]]
      [1] "chromium"
      
      [[3]]
      [1] "libcurl4-openssl-dev"
      
    Code
      lst$pre_install
    Output
      [[1]]
      NULL
      
      [[2]]
      [1] "apt-get install -y software-properties-common"
      [2] "add-apt-repository -y ppa:xtradeb/apps"       
      [3] "apt-get update"                               
      
      [[3]]
      NULL
      
    Code
      lst$post_install
    Output
      [[1]]
      NULL
      
      [[2]]
      NULL
      
      [[3]]
      NULL
      

# sysreqs_db_match

    Code
      res
    Output
      [[1]]
      # A data frame: 2 x 5
        spec             sysreq  packages  pre_install post_install
        <chr>            <chr>   <list>    <list>      <list>      
      1 java and libcurl java    <chr [1]> <NULL>      <chr [1]>   
      2 java and libcurl libcurl <chr [1]> <NULL>      <NULL>      
      
      [[2]]
      # A data frame: 1 x 5
        spec                  sysreq  packages  pre_install post_install
        <chr>                 <chr>   <list>    <list>      <list>      
      1 openssl would be good openssl <chr [1]> <NULL>      <NULL>      
      
    Code
      lapply(res, "[[", "packages")
    Output
      [[1]]
      [[1]][[1]]
      [1] "default-jdk"
      
      [[1]][[2]]
      [1] "libcurl4-openssl-dev"
      
      
      [[2]]
      [[2]][[1]]
      [1] "libssl-dev"
      
      
    Code
      lapply(res, "[[", "post_install")
    Output
      [[1]]
      [[1]][[1]]
      [1] "R CMD javareconf"
      
      [[1]][[2]]
      NULL
      
      
      [[2]]
      [[2]][[1]]
      NULL
      
      

# sysreqs_install_plan

    Code
      res
    Output
      $os
      [1] "linux"
      
      $distribution
      [1] "ubuntu"
      
      $version
      [1] "22.04"
      
      $pre_install
      [1] "apt-get -y update"
      
      $install_scripts
      [1] "apt-get -y install libcurl4-openssl-dev libssl-dev"
      
      $post_install
      character(0)
      
      $packages
      # A data frame: 2 x 5
        sysreq  packages  pre_install system_packages post_install
        <chr>   <list>    <list>      <list>          <list>      
      1 libcurl <chr [1]> <chr [0]>   <chr [1]>       <chr [0]>   
      2 openssl <chr [1]> <chr [0]>   <chr [1]>       <chr [0]>   
      
    Code
      res$packages$packages
    Output
      [[1]]
      [1] "curl"
      
      [[2]]
      [1] "curl"
      
    Code
      res$packages$system_packages
    Output
      [[1]]
      [1] "libcurl4-openssl-dev"
      
      [[2]]
      [1] "libssl-dev"
      

---

    Code
      res2
    Output
      $os
      [1] "linux"
      
      $distribution
      [1] "centos"
      
      $version
      [1] "7"
      
      $pre_install
      character(0)
      
      $install_scripts
      [1] "yum install -y make java-1.8.0-openjdk-devel"
      
      $post_install
      [1] "R CMD javareconf"
      
      $packages
      # A data frame: 2 x 5
        sysreq  packages  pre_install system_packages post_install
        <chr>   <list>    <list>      <list>          <list>      
      1 gnumake <chr [1]> <chr [0]>   <chr [1]>       <chr [0]>   
      2 java    <chr [1]> <chr [0]>   <chr [1]>       <chr [1]>   
      
    Code
      res2$packages$packages
    Output
      [[1]]
      [1] "rJava"
      
      [[2]]
      [1] "rJava"
      
    Code
      res2$packages$system_packages
    Output
      [[1]]
      [1] "make"
      
      [[2]]
      [1] "java-1.8.0-openjdk-devel"
      

# sysreqs_check_installed

    Code
      res
    Output
      system package       installed required by
      --------------       --        -----------
      gsfonts              v         magick     
      libcurl4-openssl-dev v         curl       
      libmagick++-dev      x         magick     
      libssl-dev           x         curl       

---

    Code
      res[]
    Output
      # A data frame: 4 x 5
        system_package       installed packages  pre_install post_install
        <chr>                <lgl>     <list>    <list>      <list>      
      1 gsfonts              TRUE      <chr [1]> <NULL>      <NULL>      
      2 libcurl4-openssl-dev TRUE      <chr [1]> <NULL>      <NULL>      
      3 libmagick++-dev      FALSE     <chr [1]> <NULL>      <NULL>      
      4 libssl-dev           FALSE     <chr [1]> <NULL>      <NULL>      

---

    Code
      sysreqs_check_installed()
    Output
      system package installed required by
      -------------- --        -----------

# async_parse_installed

    Code
      synchronize(async_parse_installed(library = .libPaths()[1], packages = c("foo",
        "bar", "baz")))
    Condition
      Warning in `async_parse_installed()`:
      Ignored 1 package that is not installed: baz.
    Output
      # A data frame: 2 x 1
        Package
        <chr>  
      1 foo    
      2 bar    

# parse_sysreqs_platform

    Code
      parse_sysreqs_platform("x86_64-pc-linux-gnu-ubuntu-22.04")
    Output
      # A data frame: 1 x 5
        cpu    vendor os    distribution version
        <chr>  <chr>  <chr> <chr>        <chr>  
      1 x86_64 pc     linux ubuntu       22.04  
    Code
      parse_sysreqs_platform("x86_64-pc-linux-musl-alpine-3.14.1")
    Output
      # A data frame: 1 x 5
        cpu    vendor os    distribution version
        <chr>  <chr>  <chr> <chr>        <chr>  
      1 x86_64 pc     linux alpine       3.14.1 
    Code
      parse_sysreqs_platform("x86_64-pc-linux-ubuntu-22.04")
    Output
      # A data frame: 1 x 5
        cpu    vendor os    distribution version
        <chr>  <chr>  <chr> <chr>        <chr>  
      1 x86_64 pc     linux ubuntu       22.04  
    Code
      parse_sysreqs_platform("ubuntu-22.04")
    Output
      # A data frame: 1 x 5
        cpu   vendor os    distribution version
        <chr> <chr>  <chr> <chr>        <chr>  
      1 <NA>  <NA>   linux ubuntu       22.04  
    Code
      parse_sysreqs_platform("aarch64-apple-darwin20")
    Output
      # A data frame: 1 x 5
        cpu     vendor os       distribution version
        <chr>   <chr>  <chr>    <chr>        <chr>  
      1 aarch64 apple  darwin20 <NA>         <NA>   
    Code
      parse_sysreqs_platform("i386+x86_64-w64-mingw32")
    Output
      # A data frame: 1 x 5
        cpu         vendor os      distribution version
        <chr>       <chr>  <chr>   <chr>        <chr>  
      1 i386+x86_64 w64    mingw32 <NA>         <NA>   
    Code
      parse_sysreqs_platform("x86_64-w64-mingw32")
    Output
      # A data frame: 1 x 5
        cpu    vendor os      distribution version
        <chr>  <chr>  <chr>   <chr>        <chr>  
      1 x86_64 w64    mingw32 <NA>         <NA>   
    Code
      parse_sysreqs_platform("ubuntu")
    Output
      # A data frame: 1 x 5
        cpu   vendor os    distribution version
        <chr> <chr>  <chr> <chr>        <chr>  
      1 <NA>  <NA>   linux ubuntu       <NA>   
    Code
      parse_sysreqs_platform("x86_64-px-linux-gnu-ubuntu-bar-baz")
    Output
      # A data frame: 1 x 5
        cpu    vendor os    distribution version
        <chr>  <chr>  <chr> <chr>        <chr>  
      1 x86_64 px     linux ubuntu       bar-baz
    Code
      parse_sysreqs_platform("ubuntu-22.04-libc++")
    Output
      # A data frame: 1 x 5
        cpu   vendor os    distribution version     
        <chr> <chr>  <chr> <chr>        <chr>       
      1 <NA>  <NA>   linux ubuntu       22.04-libc++

# query, post_install

    Code
      srq <- sysreqs_resolve("java", "ubuntu-22.04")
      srq$total <- 1 / 3
      srq
    Output
      $os
      [1] "linux"
      
      $distribution
      [1] "ubuntu"
      
      $version
      [1] "22.04"
      
      $url
      [1] "<server>/__api__/repos/1/sysreqs?distribution=ubuntu&release=22.04"
      
      $total
      [1] 0.3333333
      
      $pre_install
      character(0)
      
      $install_scripts
      [1] "apt-get install -y default-jdk"
      
      $post_install
      [1] "R CMD javareconf"
      
      attr(,"class")
      [1] "pkg_sysreqs_result" "list"              

# pre_install

    Code
      srq <- sysreqs_resolve("this needs geos please", "ubuntu-16.04")
      srq$total <- 1 / 3
      srq
    Output
      $os
      [1] "linux"
      
      $distribution
      [1] "ubuntu"
      
      $version
      [1] "16.04"
      
      $url
      [1] "<server>/__api__/repos/1/sysreqs?distribution=ubuntu&release=16.04"
      
      $total
      [1] 0.3333333
      
      $pre_install
      [1] "apt-get install -y software-properties-common"
      [2] "add-apt-repository -y ppa:ubuntugis/ppa"      
      [3] "apt-get update"                               
      
      $install_scripts
      [1] "apt-get install -y libgeos-dev"
      
      $post_install
      character(0)
      
      attr(,"class")
      [1] "pkg_sysreqs_result" "list"              

# multiple sysreqs

    Code
      srq <- sysreqs_resolve("java and also libcurl", "ubuntu-22.04")
      srq$total <- 1 / 3
      srq
    Output
      $os
      [1] "linux"
      
      $distribution
      [1] "ubuntu"
      
      $version
      [1] "22.04"
      
      $url
      [1] "<server>/__api__/repos/1/sysreqs?distribution=ubuntu&release=22.04"
      
      $total
      [1] 0.3333333
      
      $pre_install
      character(0)
      
      $install_scripts
      [1] "apt-get install -y default-jdk"         
      [2] "apt-get install -y libcurl4-openssl-dev"
      
      $post_install
      [1] "R CMD javareconf"
      
      attr(,"class")
      [1] "pkg_sysreqs_result" "list"              

# error, unknown os

    Code
      sysreqs_resolve("java", "foobar-11")
    Condition
      Error:
      ! ! Failed to look up system requirements for OS foobar-11.
      i HTTP error 400 for <<server>/__api__/repos/1/sysreqs?distribution=foobar&release=11>.
      i Response: "{\"code\":14,\"error\":\"Unsupported system\",\"payload\":null}".

# sysreqs_install

    Code
      sysreqs_install(srq)
    Message
      i Installing system requirements
      i Executing `sh -c echo apt-get install -y libssl-dev libcurl4-openssl-dev`

---

    Code
      sysreqs_install(srq)
    Message
      i Installing system requirements
      i Executing `sh -c echo apt-get install -y default-jdk`
      i Executing `sh -c echo apt-get install -y libcurl4-openssl-dev`
      i Executing `sh -c echo R CMD javareconf`

---

    Code
      sysreqs_install(srq)
    Message
      i Installing system requirements
      i Executing `sh -c echo apt-get install -y default-jdk`
      apt-get install -y default-jdk
      i Executing `sh -c echo apt-get install -y libcurl4-openssl-dev`
      apt-get install -y libcurl4-openssl-dev
      i Executing `sh -c echo R CMD javareconf`
      R CMD javareconf

# compact_cmds

    Code
      compact_cmds(character())
    Output
      character(0)
    Code
      compact_cmds(c("apt-get install -y libssl-dev"))
    Output
      [1] "apt-get install -y libssl-dev"
    Code
      compact_cmds(c("apt-get install -y libssl-dev",
        "apt-get install -y libcurl4-openssl-dev"))
    Output
      [1] "apt-get install -y libssl-dev libcurl4-openssl-dev"

# highlight_sysreqs

    Code
      highlight_sysreqs(sq)
    Output
      [1] ""                                                          
      [2] " + v libfontconfig1-dev, v libfreetype6-dev"               
      [3] " + v libfreetype6-dev, x libfribidi-dev, x libharfbuzz-dev"
      [4] ""                                                          
    Code
      highlight_sysreqs(sq2)
    Output
      [1] ""                                                    
      [2] " + libfontconfig1-dev, libfreetype6-dev"             
      [3] " + libfreetype6-dev, libfribidi-dev, libharfbuzz-dev"
      [4] ""                                                    

---

    Code
      highlight_sysreqs(sq)
    Output
      [1] ""                                                                                                                                                                          
      [2] "\033[90m + \033[39m\033[36m\033[32m✔\033[36m libfontconfig1-dev\033[39m, \033[36m\033[32m✔\033[36m libfreetype6-dev\033[39m"                                               
      [3] "\033[90m + \033[39m\033[36m\033[32m✔\033[36m libfreetype6-dev\033[39m, \033[36m\033[31m✖\033[36m libfribidi-dev\033[39m, \033[36m\033[31m✖\033[36m libharfbuzz-dev\033[39m"
      [4] ""                                                                                                                                                                          
    Code
      highlight_sysreqs(sq2)
    Output
      [1] ""                                                                                                                    
      [2] "\033[90m + \033[39m\033[36mlibfontconfig1-dev\033[39m, \033[36mlibfreetype6-dev\033[39m"                             
      [3] "\033[90m + \033[39m\033[36mlibfreetype6-dev\033[39m, \033[36mlibfribidi-dev\033[39m, \033[36mlibharfbuzz-dev\033[39m"
      [4] ""                                                                                                                    

---

    Code
      highlight_sysreqs(sq)
    Output
      [1] ""                                                                                                                                                                          
      [2] "\033[90m + \033[39m\033[36mchrome (installer)\033[39m"                                                                                                                     
      [3] "\033[90m + \033[39m\033[36m\033[32m✔\033[36m libfreetype6-dev\033[39m, \033[36m\033[31m✖\033[36m libfribidi-dev\033[39m, \033[36m\033[31m✖\033[36m libharfbuzz-dev\033[39m"
      [4] ""                                                                                                                                                                          

# resolve_remote

    Code
      as.list(res[, intersect(names(res), cols)])
    Output
      $package
      [1] "crayon"
      
      $version
      [1] "1.0.0"
      
      $needscompilation
      [1] FALSE
      
      $repodir
      [1] "src/contrib"
      
      $rversion
      [1] "*"
      
      $platform
      [1] "source"
      
      $priority
      [1] NA
      
      $ref
      [1] "crayon"
      
      $type
      [1] "standard"
      
      $status
      [1] "OK"
      
      $target
      [1] "src/contrib/crayon_1.0.0.tar.gz"
      
      $mirror
      [1] "http://127.0.0.1:<port>/"
      
      $sources
      $sources[[1]]
      [1] "http://127.0.0.1:<port>//src/contrib/crayon_1.0.0.tar.gz"               
      [2] "http://127.0.0.1:<port>//src/contrib/Archive/crayon/crayon_1.0.0.tar.gz"
      
      
      $sysreqs
      [1] NA
      
      $deps
      $deps[[1]]
      # A data frame: 0 x 5
      # ... with 5 variables: ref <chr>, type <chr>, package <chr>, op <chr>,
      #   version <chr>
      
      
      $built
      [1] NA
      
      $repotype
      [1] "cran"
      
      $direct
      [1] TRUE
      
      $params
      $params[[1]]
      character(0)
      
      
      $metadata
      $metadata[[1]]
                     RemoteType              RemotePkgRef                 RemoteRef 
                     "standard"                  "crayon"                  "crayon" 
                    RemoteRepos         RemotePkgPlatform                 RemoteSha 
      "http://127.0.0.1:<port>/"                  "source"                   "1.0.0" 
      
      

---

    Code
      as.list(res[, intersect(names(res), cols)])
    Output
      $ref
      [1] "thispackagecannotexistforsure"
      
      $type
      [1] "standard"
      
      $direct
      [1] TRUE
      
      $status
      [1] "FAILED"
      
      $package
      [1] "thispackagecannotexistforsure"
      
      $version
      [1] NA
      
      $needscompilation
      [1] TRUE
      
      $priority
      [1] NA
      
      $built
      [1] NA
      
      $platform
      [1] "source"
      
      $rversion
      [1] "*"
      
      $repotype
      [1] NA
      
      $repodir
      [1] "src/contrib"
      
      $target
      [1] "src/contrib/thispackagecannotexistforsure_NA.tar.gz"
      
      $deps
      $deps[[1]]
      # A data frame: 0 x 5
      # ... with 5 variables: ref <chr>, type <chr>, package <chr>, op <chr>,
      #   version <chr>
      
      
      $mirror
      [1] NA
      
      $sources
      $sources[[1]]
      [1] NA
      
      
      $metadata
      $metadata[[1]]
      list()
      
      
      $params
      $params[[1]]
      character(0)
      
      
      $sysreqs
      [1] NA
      

# resolve_remote

    Code
      res
    Output
      $ref
      [1] "local::fixtures/foobar_1.0.0.tar.gz"
      
      $type
      [1] "local"
      
      $direct
      [1] TRUE
      
      $status
      [1] "OK"
      
      $package
      [1] "foobar"
      
      $version
      [1] "1.0.0"
      
      $license
      [1] "MIT + file LICENSE"
      
      $needscompilation
      [1] FALSE
      
      $md5sum
      [1] NA
      
      $built
      [1] NA
      
      $platform
      [1] "source"
      
      $rversion
      [1] "*"
      
      $deps
      $deps[[1]]
      # A data frame: 1 x 5
        ref      type     package  op    version
        <chr>    <chr>    <chr>    <chr> <chr>  
      1 testthat Suggests testthat ""    ""     
      
      
      $sources
      [1] "foobar_1.0.0.tar.gz"
      
      $remote
      $remote[[1]]
      $package
      [1] NA
      
      $path
      [1] "fixtures/foobar_1.0.0.tar.gz"
      
      $ref
      [1] "local::fixtures/foobar_1.0.0.tar.gz"
      
      $type
      [1] "local"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_local" "remote_ref"       "list"            
      
      
      $unknown_deps
      character(0)
      
      $extra
      $extra[[1]]
      $extra[[1]]$description
      Package: foobar
      Title: Foobar Package
      Version: 1.0.0
      Author: Gabor Csardi
      Maintainer: Gabor Csardi <csardi.gabor@gmail.com>
      Description: Just a package to use in test cases.
      License: MIT + file LICENSE
      URL: https://github.com/gaborcsardi/foobar
      BugReports: https://github.com/gaborcsardi/foobar/issues
      Suggests:
          testthat
      LazyData: true
      NeedsCompilation: no
      Packaged: 2017-12-25 19:18:10 UTC; gaborcsardi
      RoxygenNote: 6.0.1
      
      
      
      $metadata
      $metadata$RemotePkgRef
      [1] "local::fixtures/foobar_1.0.0.tar.gz"
      
      $metadata$RemoteType
      [1] "local"
      
      
      $params
      $params[[1]]
      character(0)
      
      
      $sysreqs
      [1] ""
      

---

    Code
      res
    Output
      $ref
      [1] "local::foobar_1.0.0.tar.gz"
      
      $type
      [1] "local"
      
      $direct
      [1] TRUE
      
      $status
      [1] "OK"
      
      $package
      [1] "foobar"
      
      $version
      [1] "1.0.0"
      
      $license
      [1] "MIT + file LICENSE"
      
      $needscompilation
      [1] FALSE
      
      $md5sum
      [1] NA
      
      $built
      [1] NA
      
      $platform
      [1] "source"
      
      $rversion
      [1] "*"
      
      $deps
      $deps[[1]]
      # A data frame: 1 x 5
        ref      type     package  op    version
        <chr>    <chr>    <chr>    <chr> <chr>  
      1 testthat Suggests testthat ""    ""     
      
      
      $sources
      [1] "foobar_1.0.0.tar.gz"
      
      $remote
      $remote[[1]]
      $package
      [1] NA
      
      $path
      [1] "foobar_1.0.0.tar.gz"
      
      $ref
      [1] "local::foobar_1.0.0.tar.gz"
      
      $type
      [1] "local"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_local" "remote_ref"       "list"            
      
      
      $unknown_deps
      character(0)
      
      $extra
      $extra[[1]]
      $extra[[1]]$description
      Package: foobar
      Title: Foobar Package
      Version: 1.0.0
      Author: Gabor Csardi
      Maintainer: Gabor Csardi <csardi.gabor@gmail.com>
      Description: Just a package to use in test cases.
      License: MIT + file LICENSE
      URL: https://github.com/gaborcsardi/foobar
      BugReports: https://github.com/gaborcsardi/foobar/issues
      Suggests:
          testthat
      LazyData: true
      NeedsCompilation: no
      Packaged: 2017-12-25 19:18:10 UTC; gaborcsardi
      RoxygenNote: 6.0.1
      
      
      
      $metadata
      $metadata$RemotePkgRef
      [1] "local::foobar_1.0.0.tar.gz"
      
      $metadata$RemoteType
      [1] "local"
      
      
      $params
      $params[[1]]
      character(0)
      
      
      $sysreqs
      [1] ""
      

# download_remote error

    Code
      invisible(r$download_resolution())
    Message
      i Getting 1 pkg with unknown size
      x Failed to download foobar 1.0.0 (source)

# satisfy

    Code
      # different package name, FALSE
      satisfy_remote_local(list(package = "foo"), list(package = "bar"))
    Output
      [1] FALSE
      attr(,"reason")
      [1] "Package names differ"
    Code
      # different type is bad
      satisfy_remote_local(list(package = "foo", remote = list(list(package = NA_character_)),
      path = "/tmp/foo"), list(type = "notgood", package = "foo", path = "/tmp/foo"))
    Output
      [1] FALSE
      attr(,"reason")
      [1] "Package source mismatch"
    Code
      # any package name is good, good
      satisfy_remote_local(list(package = "foo", remote = list(list(package = NA_character_,
        path = "/tmp/foo"))), list(type = "local", package = "foo", remote = list(
        list(path = "/tmp/foo"))))
    Output
      [1] TRUE
    Code
      # requested a different package, bad
      satisfy_remote_local(list(package = "foo", remote = list(list(package = "bar",
        path = "/tmp/foo"))), list(type = "local", package = "foo", remote = list(
        list(package = "bar", path = "/tmp/foo"))))
    Output
      [1] FALSE
      attr(,"reason")
      [1] "Requested package bar but got foo"
    Code
      # different paths are bad
      satisfy_remote_local(list(package = "foo", remote = list(list(package = "foo",
        path = "/tmp/foo"))), list(type = "local", package = "foo", remote = list(
        list(package = "foo", path = "/tmp/foo2"))))
    Output
      [1] FALSE
      attr(,"reason")
      [1] "Paths differ"

# installed preferred over download

    Code
      sol[, c("type", "package", "version")]
    Output
      # A data frame: 2 x 3
        type      package version
      * <chr>     <chr>   <chr>  
      1 installed pkg1    1.0.0  
      2 installed pkg2    1.0.0  

---

    Code
      sol[, c("type", "package", "version")]
    Output
      # A data frame: 2 x 3
        type      package version
      * <chr>     <chr>   <chr>  
      1 installed pkg1    1.0.0  
      2 installed pkg2    1.0.0  

---

    Code
      sol[, c("type", "package", "version")]
    Output
      # A data frame: 2 x 3
        type     package version
      * <chr>    <chr>   <chr>  
      1 standard pkg1    1.0.0  
      2 standard pkg2    1.0.0  

# resolve_remote

    Code
      as.list(res[, intersect(names(res), cols)])
    Output
      $package
      [1] "Biobase"
      
      $version
      [1] "1.2.3"
      
      $needscompilation
      [1] FALSE
      
      $repodir
      [1] "src/contrib"
      
      $rversion
      [1] "*"
      
      $platform
      [1] "source"
      
      $priority
      [1] NA
      
      $ref
      [1] "Biobase"
      
      $type
      [1] "standard"
      
      $status
      [1] "OK"
      
      $target
      [1] "src/contrib/Biobase_1.2.3.tar.gz"
      
      $mirror
      [1] "http://127.0.0.1:<port>//packages/<bioc-version>/bioc"
      
      $sources
      $sources[[1]]
      [1] "http://127.0.0.1:<port>//packages/<bioc-version>/bioc/src/contrib/Biobase_1.2.3.tar.gz"
      
      
      $sysreqs
      [1] NA
      
      $deps
      $deps[[1]]
      # A data frame: 9 x 5
        ref          type     package      op    version 
        <chr>        <chr>    <chr>        <chr> <chr>   
      1 ALL          suggests ALL          ""    ""      
      2 BiocGenerics depends  BiocGenerics ">="  "0.27.1"
      3 R            depends  R            ">="  "2.10"  
      4 RUnit        suggests RUnit        ""    ""      
      5 golubEsets   suggests golubEsets   ""    ""      
      6 methods      imports  methods      ""    ""      
      7 tkWidgets    suggests tkWidgets    ""    ""      
      8 tools        suggests tools        ""    ""      
      9 utils        depends  utils        ""    ""      
      
      
      $built
      [1] NA
      
      $repotype
      [1] "bioc"
      
      $direct
      [1] TRUE
      
      $params
      $params[[1]]
      character(0)
      
      
      $metadata
      $metadata[[1]]
                                        RemoteType 
                                        "standard" 
                                      RemotePkgRef 
                                         "Biobase" 
                                         RemoteRef 
                                         "Biobase" 
                                       RemoteRepos 
      "http://127.0.0.1:<port>//packages/<bioc-version>/bioc" 
                                 RemotePkgPlatform 
                                          "source" 
                                         RemoteSha 
                                           "1.2.3" 
      
      

# can package a tree

    Code
      install_package_plan(plan, lib = lib)
    Message
      i Packaging foo 0.0.0.9000
      v Packaged foo 0.0.0.9000
      i Building foo 0.0.0.9000
      v Built foo 0.0.0.9000
      v Installed foo 0.0.0.9000 (local)
      v Summary:

# can package a compressed tree

    Code
      install_package_plan(plan, lib = lib)
    Message
      i Packaging foo 0.0.0.9000
      v Packaged foo 0.0.0.9000
      i Building foo 0.0.0.9000
      v Built foo 0.0.0.9000
      v Installed foo 0.0.0.9000 (local)
      v Summary:

# can package a source package

    Code
      install_package_plan(plan, lib = lib)
    Message
      i Building foo 0.0.0.9000
      v Built foo 0.0.0.9000
      v Installed foo 0.0.0.9000 (local)
      v Summary:

# add_recursive_dependencies

    Code
      add_recursive_dependencies(plan)
    Output
      # A data frame: 3 x 6
        package type  binary dependencies file  needscompilation
        <chr>   <chr> <lgl>  <list>       <chr> <lgl>           
      1 p1      cran  FALSE  <chr [2]>    <NA>  FALSE           
      2 p2      cran  TRUE   <chr [1]>    <NA>  FALSE           
      3 p3      cran  FALSE  <chr [0]>    <NA>  FALSE           

---

    Code
      add_recursive_dependencies(plan)$dependencies
    Output
      [[1]]
      [1] "p2" "p3"
      
      [[2]]
      [1] "p3"
      
      [[3]]
      character(0)
      

# ignore-build-errors parameter

    Code
      suppressMessages(inst$solve())
      suppressMessages(inst$download())
      inst$install()
    Message
      i Packaging badbuild 1.0.0
      v Packaged badbuild 1.0.0
      i Building badbuild 1.0.0
      x Failed to build badbuild 1.0.0
    Condition
      Error:
      ! ! Failed to build source package badbuild.

---

    Code
      suppressMessages(inst$solve())
      suppressMessages(inst$download())
      inst$install()
    Message
      i Packaging badbuild 1.0.0
      v Packaged badbuild 1.0.0
      i Building badbuild 1.0.0
      ! Failed to build badbuild 1.0.0
      i Packaging goodbuild 1.0.0
      v Packaged goodbuild 1.0.0
      i Building goodbuild 1.0.0
      ! Failed to build goodbuild 1.0.0
      v Summary:   2 new

# error

    Code
      res$error[[1]]
    Output
      <async_rejected/rlib_error_3_0/rlib_error/error>
      Error in `value[[3L]](cond)`:
      ! pkgdepends resolution error for foo::bar.
      Caused by error: 
      ! foobar

---

    Code
      res$error[[2]]
    Output
      <async_rejected/rlib_error_3_0/rlib_error/error>
      Error in `value[[3L]](cond)`:
      ! pkgdepends resolution error for foo::bar2.
      Caused by error: 
      ! foobar

# reinstall, standard

    Code
      # pkg1 is already installed
      p$get_solution()
    Output
      <pkg_solution>
      + result: OK
      + refs:
        - pkg1
      + constraints (1):
        - select pkg1 exactly once
      + solution:
        - installed::.../pkg1

---

    Code
      # request a reinstall
      p$get_solution()
    Output
      <pkg_solution>
      + result: OK
      + refs:
        - pkg3
      + constraints (6):
        - select pkg3 exactly once
        - select pkg1 at most once
        - select pkg2 at most once
        - pkg2 depends on pkg1: version pkg1 1.0.0, installed::.../pkg1 1.0.0
        - pkg3 depends on pkg2: version pkg2 1.0.0, installed::.../pkg2 1.0.0
        - installed::.../pkg1 1.0.0
      + solution:
        - installed::.../pkg1
        - installed::.../pkg2
        - pkg3

---

    Code
      # request a reinstall of a dependency
      p$get_solution()
    Output
      <pkg_solution>
      + result: OK
      + refs:
        - pkg3
      + constraints (7):
        - select pkg3 exactly once
        - select pkg1 at most once
        - select pkg2 at most once
        - pkg2 depends on pkg1: version pkg1 1.0.0
        - pkg3 depends on pkg2: version pkg2 1.0.0, installed::.../pkg2 1.0.0
        - installed::.../pkg2 depends on pkg1: version pkg1 1.0.0
        - installed::.../pkg2 1.0.0
      + solution:
        - installed::.../pkg2
        - installed::.../pkg3
        - pkg1

---

    Code
      # one reinstall, one not
      p$get_solution()
    Output
      <pkg_solution>
      + result: OK
      + refs:
        - pkg1
        - pkg3
      + constraints (7):
        - select pkg3 exactly once
        - select pkg1 exactly once
        - select pkg2 at most once
        - pkg2 depends on pkg1: version pkg1 1.0.0, installed::.../pkg1 1.0.0
        - pkg3 depends on pkg2: version pkg2 1.0.0, installed::.../pkg2 1.0.0
        - installed::.../pkg1 1.0.0
        - installed::.../pkg2 1.0.0
      + solution:
        - installed::.../pkg1
        - installed::.../pkg2
        - installed::.../pkg3

# reinstsll from URL

    Code
      # reinstall from direct URL
      p$get_solution()
    Output
      <pkg_solution>
      + result: OK
      + refs:
        - url::http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz
      + constraints (2):
        - select pkg1 exactly once
        - `url::http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz` is not satisfied by `installed::.../pkg1`
      + solution:
        - url::http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz

---

    Code
      # reinstall from URL, extra parameters
      p$get_solution()
    Output
      <pkg_solution>
      + result: OK
      + refs:
        - url::http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz
      + constraints (1):
        - select pkg1 exactly once
      + solution:
        - url::http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz

# source

    Code
      sol[, c("package", "platform")]
    Output
      # A data frame: 2 x 2
        package platform               
      * <chr>   <chr>                  
      1 pkg     source                 
      2 pkg2    i386+x86_64-w64-mingw32

# source for dependency

    Code
      sol[, c("package", "platform")]
    Output
      # A data frame: 2 x 2
        package platform               
      * <chr>   <chr>                  
      1 pkg     i386+x86_64-w64-mingw32
      2 pkg2    source                 

# parse_remote_git

    Code
      parse_remote_git("git::https://github.com/r-lib/cli")
    Output
      [[1]]
      [[1]]$package
      [1] "cli"
      
      [[1]]$protocol
      [1] "https"
      
      [[1]]$host
      [1] "github.com"
      
      [[1]]$path
      [1] "/r-lib/"
      
      [[1]]$repo
      [1] "cli"
      
      [[1]]$commitish
      [1] "HEAD"
      
      [[1]]$ref
      [1] "git::https://github.com/r-lib/cli"
      
      [[1]]$type
      [1] "git"
      
      [[1]]$dotgit
      [1] ""
      
      [[1]]$url
      [1] "https://github.com/r-lib/cli"
      
      

---

    Code
      parse_remote_git("git::https://github.com/r-lib/cli.git")
    Output
      [[1]]
      [[1]]$package
      [1] "cli"
      
      [[1]]$protocol
      [1] "https"
      
      [[1]]$host
      [1] "github.com"
      
      [[1]]$path
      [1] "/r-lib/"
      
      [[1]]$repo
      [1] "cli"
      
      [[1]]$commitish
      [1] "HEAD"
      
      [[1]]$ref
      [1] "git::https://github.com/r-lib/cli.git"
      
      [[1]]$type
      [1] "git"
      
      [[1]]$dotgit
      [1] ".git"
      
      [[1]]$url
      [1] "https://github.com/r-lib/cli.git"
      
      

---

    Code
      parse_remote_git("git::https://github.com/r-lib/cli@branch")
    Output
      [[1]]
      [[1]]$package
      [1] "cli"
      
      [[1]]$protocol
      [1] "https"
      
      [[1]]$host
      [1] "github.com"
      
      [[1]]$path
      [1] "/r-lib/"
      
      [[1]]$repo
      [1] "cli"
      
      [[1]]$commitish
      [1] "branch"
      
      [[1]]$ref
      [1] "git::https://github.com/r-lib/cli@branch"
      
      [[1]]$type
      [1] "git"
      
      [[1]]$dotgit
      [1] ""
      
      [[1]]$url
      [1] "https://github.com/r-lib/cli"
      
      

---

    Code
      parse_remote_git("git::https://github.com/r-lib/cli.git@branch")
    Output
      [[1]]
      [[1]]$package
      [1] "cli"
      
      [[1]]$protocol
      [1] "https"
      
      [[1]]$host
      [1] "github.com"
      
      [[1]]$path
      [1] "/r-lib/"
      
      [[1]]$repo
      [1] "cli"
      
      [[1]]$commitish
      [1] "branch"
      
      [[1]]$ref
      [1] "git::https://github.com/r-lib/cli.git@branch"
      
      [[1]]$type
      [1] "git"
      
      [[1]]$dotgit
      [1] ".git"
      
      [[1]]$url
      [1] "https://github.com/r-lib/cli.git"
      
      

---

    Code
      parse_remote_git("pkg=git::https://github.com/r-lib/cli")
    Output
      [[1]]
      [[1]]$package
      [1] "pkg"
      
      [[1]]$protocol
      [1] "https"
      
      [[1]]$host
      [1] "github.com"
      
      [[1]]$path
      [1] "/r-lib/"
      
      [[1]]$repo
      [1] "cli"
      
      [[1]]$commitish
      [1] "HEAD"
      
      [[1]]$ref
      [1] "pkg=git::https://github.com/r-lib/cli"
      
      [[1]]$type
      [1] "git"
      
      [[1]]$dotgit
      [1] ""
      
      [[1]]$url
      [1] "https://github.com/r-lib/cli"
      
      

---

    Code
      parse_remote_git("pkg=git::https://github.com/r-lib/cli.git")
    Output
      [[1]]
      [[1]]$package
      [1] "pkg"
      
      [[1]]$protocol
      [1] "https"
      
      [[1]]$host
      [1] "github.com"
      
      [[1]]$path
      [1] "/r-lib/"
      
      [[1]]$repo
      [1] "cli"
      
      [[1]]$commitish
      [1] "HEAD"
      
      [[1]]$ref
      [1] "pkg=git::https://github.com/r-lib/cli.git"
      
      [[1]]$type
      [1] "git"
      
      [[1]]$dotgit
      [1] ".git"
      
      [[1]]$url
      [1] "https://github.com/r-lib/cli.git"
      
      

---

    Code
      parse_remote_git("pkg=git::https://github.com/r-lib/cli@branch")
    Output
      [[1]]
      [[1]]$package
      [1] "pkg"
      
      [[1]]$protocol
      [1] "https"
      
      [[1]]$host
      [1] "github.com"
      
      [[1]]$path
      [1] "/r-lib/"
      
      [[1]]$repo
      [1] "cli"
      
      [[1]]$commitish
      [1] "branch"
      
      [[1]]$ref
      [1] "pkg=git::https://github.com/r-lib/cli@branch"
      
      [[1]]$type
      [1] "git"
      
      [[1]]$dotgit
      [1] ""
      
      [[1]]$url
      [1] "https://github.com/r-lib/cli"
      
      

---

    Code
      parse_remote_git("pkg=git::https://github.com/r-lib/cli.git@branch")
    Output
      [[1]]
      [[1]]$package
      [1] "pkg"
      
      [[1]]$protocol
      [1] "https"
      
      [[1]]$host
      [1] "github.com"
      
      [[1]]$path
      [1] "/r-lib/"
      
      [[1]]$repo
      [1] "cli"
      
      [[1]]$commitish
      [1] "branch"
      
      [[1]]$ref
      [1] "pkg=git::https://github.com/r-lib/cli.git@branch"
      
      [[1]]$type
      [1] "git"
      
      [[1]]$dotgit
      [1] ".git"
      
      [[1]]$url
      [1] "https://github.com/r-lib/cli.git"
      
      

---

    Code
      parse_pkg_refs(c("git::https://github.com/cran/falsy.git",
        "git::https://github.com/cran/falsy2.git"))
    Output
      [[1]]
      $package
      [1] "falsy"
      
      $protocol
      [1] "https"
      
      $host
      [1] "github.com"
      
      $path
      [1] "/cran/"
      
      $repo
      [1] "falsy"
      
      $commitish
      [1] "HEAD"
      
      $ref
      [1] "git::https://github.com/cran/falsy.git"
      
      $type
      [1] "git"
      
      $dotgit
      [1] ".git"
      
      $url
      [1] "https://github.com/cran/falsy.git"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_git" "remote_ref"     "list"          
      
      [[2]]
      $package
      [1] "falsy2"
      
      $protocol
      [1] "https"
      
      $host
      [1] "github.com"
      
      $path
      [1] "/cran/"
      
      $repo
      [1] "falsy2"
      
      $commitish
      [1] "HEAD"
      
      $ref
      [1] "git::https://github.com/cran/falsy2.git"
      
      $type
      [1] "git"
      
      $dotgit
      [1] ".git"
      
      $url
      [1] "https://github.com/cran/falsy2.git"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_git" "remote_ref"     "list"          
      

# download_remote_git

    Code
      dir(file.path(prop$get_downloads()$fulltarget_tree, "cli"))
    Output
       [1] "DESCRIPTION"  "LICENSE"      "LICENSE.md"   "LICENSE.note" "Makefile"    
       [6] "NAMESPACE"    "NEWS.md"      "R"            "README.Rmd"   "README.md"   
      [11] "_pkgdown.yml" "cli.Rproj"    "codecov.yml"  "inst"         "man"         
      [16] "src"          "tests"        "tools"        "vignettes"   

# parse_pkg_refs error on unknown type

    Code
      parse_pkg_refs(c("notgood::pkg", "good", "my_package"))
    Condition
      Error:
      ! ! Cannot parse package: my_package.
      i See `?pkgdepends::pkg_refs()` for supported package sources.

# explicit package names

    Code
      parse_pkg_ref("package=user/notpackage")
    Output
      $package
      [1] "package"
      
      $username
      [1] "user"
      
      $repo
      [1] "notpackage"
      
      $subdir
      [1] ""
      
      $commitish
      [1] ""
      
      $pull
      [1] ""
      
      $release
      [1] ""
      
      $ref
      [1] "package=user/notpackage"
      
      $type
      [1] "github"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_github" "remote_ref"        "list"             
    Code
      parse_pkg_ref("package=user/notpackage@tag")
    Output
      $package
      [1] "package"
      
      $username
      [1] "user"
      
      $repo
      [1] "notpackage"
      
      $subdir
      [1] ""
      
      $commitish
      [1] "tag"
      
      $pull
      [1] ""
      
      $release
      [1] ""
      
      $ref
      [1] "package=user/notpackage@tag"
      
      $type
      [1] "github"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_github" "remote_ref"        "list"             
    Code
      parse_pkg_ref("package=github::user/notpackage@tag")
    Output
      $package
      [1] "package"
      
      $username
      [1] "user"
      
      $repo
      [1] "notpackage"
      
      $subdir
      [1] ""
      
      $commitish
      [1] "tag"
      
      $pull
      [1] ""
      
      $release
      [1] ""
      
      $ref
      [1] "package=github::user/notpackage@tag"
      
      $type
      [1] "github"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_github" "remote_ref"        "list"             
    Code
      parse_pkg_ref("package=local::/abs/path")
    Output
      $package
      [1] "package"
      
      $path
      [1] "/abs/path"
      
      $ref
      [1] "package=local::/abs/path"
      
      $type
      [1] "local"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_local" "remote_ref"       "list"            
    Code
      parse_pkg_ref("package=local::rel/path")
    Output
      $package
      [1] "package"
      
      $path
      [1] "rel/path"
      
      $ref
      [1] "package=local::rel/path"
      
      $type
      [1] "local"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_local" "remote_ref"       "list"            
    Code
      parse_pkg_ref("package=local::~/home/path")
    Output
      $package
      [1] "package"
      
      $path
      [1] "~/home/path"
      
      $ref
      [1] "package=local::~/home/path"
      
      $type
      [1] "local"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_local" "remote_ref"       "list"            
    Code
      parse_pkg_ref("package=/abs/path")
    Output
      $package
      [1] "package"
      
      $path
      [1] "/abs/path"
      
      $ref
      [1] "package=local::/abs/path"
      
      $type
      [1] "local"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_local" "remote_ref"       "list"            
    Code
      parse_pkg_ref("package=./rel/path")
    Output
      $package
      [1] "package"
      
      $path
      [1] "./rel/path"
      
      $ref
      [1] "package=local::./rel/path"
      
      $type
      [1] "local"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_local" "remote_ref"       "list"            
    Code
      parse_pkg_ref("package=~/home/path")
    Output
      $package
      [1] "package"
      
      $path
      [1] "~/home/path"
      
      $ref
      [1] "package=local::~/home/path"
      
      $type
      [1] "local"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_local" "remote_ref"       "list"            
    Code
      parse_pkg_ref("package=url::https://example.com/p1.tar.gz")
    Output
      $package
      [1] "package"
      
      $url
      [1] "https://example.com/p1.tar.gz"
      
      $ref
      [1] "package=url::https://example.com/p1.tar.gz"
      
      $type
      [1] "url"
      
      $hash
      [1] "abc44e36b1d7392a347beee661d5342b"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_url" "remote_ref"     "list"          

# gitlab

    Code
      parse_pkg_ref("gitlab::user/project")
    Output
      $package
      [1] "project"
      
      $protocol
      [1] "https"
      
      $host
      [1] "gitlab.com"
      
      $projectpath
      [1] "user"
      
      $project
      [1] "project"
      
      $subdir
      [1] ""
      
      $commitish
      [1] "HEAD"
      
      $ref
      [1] "gitlab::user/project"
      
      $path
      [1] "/user/project"
      
      $dotgit
      [1] ""
      
      $url
      [1] "https://gitlab.com/user/project.git"
      
      $type
      [1] "gitlab"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_gitlab" "remote_ref"        "list"             
    Code
      parse_pkg_ref("gitlab::user/project@ref")
    Output
      $package
      [1] "project"
      
      $protocol
      [1] "https"
      
      $host
      [1] "gitlab.com"
      
      $projectpath
      [1] "user"
      
      $project
      [1] "project"
      
      $subdir
      [1] ""
      
      $commitish
      [1] "ref"
      
      $ref
      [1] "gitlab::user/project@ref"
      
      $path
      [1] "/user/project"
      
      $dotgit
      [1] ""
      
      $url
      [1] "https://gitlab.com/user/project.git"
      
      $type
      [1] "gitlab"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_gitlab" "remote_ref"        "list"             
    Code
      parse_pkg_ref("gitlab::user/project/-/sub/dir")
    Output
      $package
      [1] "project"
      
      $protocol
      [1] "https"
      
      $host
      [1] "gitlab.com"
      
      $projectpath
      [1] "user"
      
      $project
      [1] "project"
      
      $subdir
      [1] "sub/dir"
      
      $commitish
      [1] "HEAD"
      
      $ref
      [1] "gitlab::user/project/-/sub/dir"
      
      $path
      [1] "/user/project"
      
      $dotgit
      [1] ""
      
      $url
      [1] "https://gitlab.com/user/project.git"
      
      $type
      [1] "gitlab"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_gitlab" "remote_ref"        "list"             
    Code
      parse_pkg_ref("gitlab::user/project/-/sub/dir@ref")
    Output
      $package
      [1] "project"
      
      $protocol
      [1] "https"
      
      $host
      [1] "gitlab.com"
      
      $projectpath
      [1] "user"
      
      $project
      [1] "project"
      
      $subdir
      [1] "sub/dir"
      
      $commitish
      [1] "ref"
      
      $ref
      [1] "gitlab::user/project/-/sub/dir@ref"
      
      $path
      [1] "/user/project"
      
      $dotgit
      [1] ""
      
      $url
      [1] "https://gitlab.com/user/project.git"
      
      $type
      [1] "gitlab"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_gitlab" "remote_ref"        "list"             
    Code
      parse_pkg_ref("gitlab::group/subgroup/project")
    Output
      $package
      [1] "project"
      
      $protocol
      [1] "https"
      
      $host
      [1] "gitlab.com"
      
      $projectpath
      [1] "group/subgroup"
      
      $project
      [1] "project"
      
      $subdir
      [1] ""
      
      $commitish
      [1] "HEAD"
      
      $ref
      [1] "gitlab::group/subgroup/project"
      
      $path
      [1] "/group/subgroup/project"
      
      $dotgit
      [1] ""
      
      $url
      [1] "https://gitlab.com/group/subgroup/project.git"
      
      $type
      [1] "gitlab"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_gitlab" "remote_ref"        "list"             
    Code
      parse_pkg_ref("gitlab::group/subgroup/project@ref")
    Output
      $package
      [1] "project"
      
      $protocol
      [1] "https"
      
      $host
      [1] "gitlab.com"
      
      $projectpath
      [1] "group/subgroup"
      
      $project
      [1] "project"
      
      $subdir
      [1] ""
      
      $commitish
      [1] "ref"
      
      $ref
      [1] "gitlab::group/subgroup/project@ref"
      
      $path
      [1] "/group/subgroup/project"
      
      $dotgit
      [1] ""
      
      $url
      [1] "https://gitlab.com/group/subgroup/project.git"
      
      $type
      [1] "gitlab"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_gitlab" "remote_ref"        "list"             
    Code
      parse_pkg_ref("gitlab::group/subgroup/project/-/sub/dir")
    Output
      $package
      [1] "project"
      
      $protocol
      [1] "https"
      
      $host
      [1] "gitlab.com"
      
      $projectpath
      [1] "group/subgroup"
      
      $project
      [1] "project"
      
      $subdir
      [1] "sub/dir"
      
      $commitish
      [1] "HEAD"
      
      $ref
      [1] "gitlab::group/subgroup/project/-/sub/dir"
      
      $path
      [1] "/group/subgroup/project"
      
      $dotgit
      [1] ""
      
      $url
      [1] "https://gitlab.com/group/subgroup/project.git"
      
      $type
      [1] "gitlab"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_gitlab" "remote_ref"        "list"             
    Code
      parse_pkg_ref("gitlab::group/subgroup/project/-/sub/dir@ref")
    Output
      $package
      [1] "project"
      
      $protocol
      [1] "https"
      
      $host
      [1] "gitlab.com"
      
      $projectpath
      [1] "group/subgroup"
      
      $project
      [1] "project"
      
      $subdir
      [1] "sub/dir"
      
      $commitish
      [1] "ref"
      
      $ref
      [1] "gitlab::group/subgroup/project/-/sub/dir@ref"
      
      $path
      [1] "/group/subgroup/project"
      
      $dotgit
      [1] ""
      
      $url
      [1] "https://gitlab.com/group/subgroup/project.git"
      
      $type
      [1] "gitlab"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_gitlab" "remote_ref"        "list"             
    Code
      parse_pkg_ref("gitlab::https://acme.co/group/subgroup/project/-/sub/dir@ref")
    Output
      $package
      [1] "project"
      
      $protocol
      [1] "https"
      
      $host
      [1] "acme.co"
      
      $projectpath
      [1] "group/subgroup"
      
      $project
      [1] "project"
      
      $subdir
      [1] "sub/dir"
      
      $commitish
      [1] "ref"
      
      $ref
      [1] "gitlab::https://acme.co/group/subgroup/project/-/sub/dir@ref"
      
      $path
      [1] "/group/subgroup/project"
      
      $dotgit
      [1] ""
      
      $url
      [1] "https://acme.co/group/subgroup/project.git"
      
      $type
      [1] "gitlab"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_gitlab" "remote_ref"        "list"             
    Code
      parse_pkg_ref("pkg=gitlab::user/project")
    Output
      $package
      [1] "pkg"
      
      $protocol
      [1] "https"
      
      $host
      [1] "gitlab.com"
      
      $projectpath
      [1] "user"
      
      $project
      [1] "project"
      
      $subdir
      [1] ""
      
      $commitish
      [1] "HEAD"
      
      $ref
      [1] "pkg=gitlab::user/project"
      
      $path
      [1] "/user/project"
      
      $dotgit
      [1] ""
      
      $url
      [1] "https://gitlab.com/user/project.git"
      
      $type
      [1] "gitlab"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_gitlab" "remote_ref"        "list"             
    Code
      parse_pkg_ref("pkg=gitlab::user/project@ref")
    Output
      $package
      [1] "pkg"
      
      $protocol
      [1] "https"
      
      $host
      [1] "gitlab.com"
      
      $projectpath
      [1] "user"
      
      $project
      [1] "project"
      
      $subdir
      [1] ""
      
      $commitish
      [1] "ref"
      
      $ref
      [1] "pkg=gitlab::user/project@ref"
      
      $path
      [1] "/user/project"
      
      $dotgit
      [1] ""
      
      $url
      [1] "https://gitlab.com/user/project.git"
      
      $type
      [1] "gitlab"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_gitlab" "remote_ref"        "list"             
    Code
      parse_pkg_ref("pkg=gitlab::user/project/-/sub/dir")
    Output
      $package
      [1] "pkg"
      
      $protocol
      [1] "https"
      
      $host
      [1] "gitlab.com"
      
      $projectpath
      [1] "user"
      
      $project
      [1] "project"
      
      $subdir
      [1] "sub/dir"
      
      $commitish
      [1] "HEAD"
      
      $ref
      [1] "pkg=gitlab::user/project/-/sub/dir"
      
      $path
      [1] "/user/project"
      
      $dotgit
      [1] ""
      
      $url
      [1] "https://gitlab.com/user/project.git"
      
      $type
      [1] "gitlab"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_gitlab" "remote_ref"        "list"             
    Code
      parse_pkg_ref("pkg=gitlab::user/project/-/sub/dir@ref")
    Output
      $package
      [1] "pkg"
      
      $protocol
      [1] "https"
      
      $host
      [1] "gitlab.com"
      
      $projectpath
      [1] "user"
      
      $project
      [1] "project"
      
      $subdir
      [1] "sub/dir"
      
      $commitish
      [1] "ref"
      
      $ref
      [1] "pkg=gitlab::user/project/-/sub/dir@ref"
      
      $path
      [1] "/user/project"
      
      $dotgit
      [1] ""
      
      $url
      [1] "https://gitlab.com/user/project.git"
      
      $type
      [1] "gitlab"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_gitlab" "remote_ref"        "list"             
    Code
      parse_pkg_ref("pkg=gitlab::group/subgroup/project")
    Output
      $package
      [1] "pkg"
      
      $protocol
      [1] "https"
      
      $host
      [1] "gitlab.com"
      
      $projectpath
      [1] "group/subgroup"
      
      $project
      [1] "project"
      
      $subdir
      [1] ""
      
      $commitish
      [1] "HEAD"
      
      $ref
      [1] "pkg=gitlab::group/subgroup/project"
      
      $path
      [1] "/group/subgroup/project"
      
      $dotgit
      [1] ""
      
      $url
      [1] "https://gitlab.com/group/subgroup/project.git"
      
      $type
      [1] "gitlab"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_gitlab" "remote_ref"        "list"             
    Code
      parse_pkg_ref("pkg=gitlab::group/subgroup/project@ref")
    Output
      $package
      [1] "pkg"
      
      $protocol
      [1] "https"
      
      $host
      [1] "gitlab.com"
      
      $projectpath
      [1] "group/subgroup"
      
      $project
      [1] "project"
      
      $subdir
      [1] ""
      
      $commitish
      [1] "ref"
      
      $ref
      [1] "pkg=gitlab::group/subgroup/project@ref"
      
      $path
      [1] "/group/subgroup/project"
      
      $dotgit
      [1] ""
      
      $url
      [1] "https://gitlab.com/group/subgroup/project.git"
      
      $type
      [1] "gitlab"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_gitlab" "remote_ref"        "list"             
    Code
      parse_pkg_ref("pkg=gitlab::group/subgroup/project/-/sub/dir")
    Output
      $package
      [1] "pkg"
      
      $protocol
      [1] "https"
      
      $host
      [1] "gitlab.com"
      
      $projectpath
      [1] "group/subgroup"
      
      $project
      [1] "project"
      
      $subdir
      [1] "sub/dir"
      
      $commitish
      [1] "HEAD"
      
      $ref
      [1] "pkg=gitlab::group/subgroup/project/-/sub/dir"
      
      $path
      [1] "/group/subgroup/project"
      
      $dotgit
      [1] ""
      
      $url
      [1] "https://gitlab.com/group/subgroup/project.git"
      
      $type
      [1] "gitlab"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_gitlab" "remote_ref"        "list"             
    Code
      parse_pkg_ref("pkg=gitlab::group/subgroup/project/-/sub/dir@ref")
    Output
      $package
      [1] "pkg"
      
      $protocol
      [1] "https"
      
      $host
      [1] "gitlab.com"
      
      $projectpath
      [1] "group/subgroup"
      
      $project
      [1] "project"
      
      $subdir
      [1] "sub/dir"
      
      $commitish
      [1] "ref"
      
      $ref
      [1] "pkg=gitlab::group/subgroup/project/-/sub/dir@ref"
      
      $path
      [1] "/group/subgroup/project"
      
      $dotgit
      [1] ""
      
      $url
      [1] "https://gitlab.com/group/subgroup/project.git"
      
      $type
      [1] "gitlab"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_gitlab" "remote_ref"        "list"             
    Code
      parse_pkg_ref(
        "pkg=gitlab::https://acme.co/group/subgroup/project/-/sub/dir@ref")
    Output
      $package
      [1] "pkg"
      
      $protocol
      [1] "https"
      
      $host
      [1] "acme.co"
      
      $projectpath
      [1] "group/subgroup"
      
      $project
      [1] "project"
      
      $subdir
      [1] "sub/dir"
      
      $commitish
      [1] "ref"
      
      $ref
      [1] "pkg=gitlab::https://acme.co/group/subgroup/project/-/sub/dir@ref"
      
      $path
      [1] "/group/subgroup/project"
      
      $dotgit
      [1] ""
      
      $url
      [1] "https://acme.co/group/subgroup/project.git"
      
      $type
      [1] "gitlab"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_gitlab" "remote_ref"        "list"             

# uncompress error

    Code
      install_package_plan(plan, lib = tempdir())
    Message
      i Packaging foo 
      x Failed to uncompress foo 
    Condition
      Error:
      ! ! Failed to uncompress foo from .../foo-t.zip'.

# print.pkg_name_check

    Hash
      01ca05e7e503765b871858b0a893ae85
    Code
      {
          list(pkg_name_check("tools"), pkg_name_check("tools", dictionaries = "urban"))
      }

# format.pkg_name_basics

    Hash
      aa4b0a131ad85ee937ac2fa09743c560
    Code
      {
          list(sy(async_pnc_basics("pwr")), sy(async_pnc_basics(paste0("s", 
              "h", "i", "t"))))
      }

# format.pkg_name_check_wikipedia

    Hash
      1e01504bf366a3d1d8a30a1bbfe42923
    Code
      {
          list(sy(async_wikipedia_get("cli")), sy(async_wikipedia_get("surely-not-this")), 
              sy(async_wikipedia_get("GNU R")))
      }

# async_wiktionary_get

    Hash
      976e3d01a855e39ebd71c60482a1cac9
    Code
      {
          sy(async_wiktionary_get_query("foobar"))
      }

# wiktionary_get_process

    Hash
      976e3d01a855e39ebd71c60482a1cac9
    Code
      {
          sy(async_wiktionary_get_query("foobar"))
      }

# format.pkg_name_check_wiktionary

    Hash
      976e3d01a855e39ebd71c60482a1cac9
    Code
      {
          sy(async_wiktionary_get_query("foobar"))
      }

---

    Hash
      936aece313cdae39b02cde924e52eeb9
    Code
      {
          sy(async_wiktionary_get_query("not-at-all-sdfsdfsdf"))
      }

# async_profanity_get

    Hash
      51e4c0280615fb2fc232bea02081489b
    Code
      {
          sy(async_profanity_get_query("nope"))
      }

# profanity_get_process

    Hash
      51e4c0280615fb2fc232bea02081489b
    Code
      {
          sy(async_profanity_get_query("nope"))
      }

# async_urban_get

    Hash
      13205944297f7f4f50e756b13397c0e5
    Code
      {
          sy(async_urban_get_query("tool"))
      }

# urban_get_process

    Hash
      13205944297f7f4f50e756b13397c0e5
    Code
      {
          sy(async_urban_get_query("tool"))
      }

# format.pkg_name_check_urban

    Hash
      0468c02ed8cc959d7380ea172a55e336
    Code
      {
          sy(async_urban_get("tool"))
      }

---

    Hash
      c6ad8510dbbba26e1ee8e9e21bcee98f
    Code
      {
          sy(async_urban_get("not-this-one-asdfsf"))
      }

# async_pnc_bioc_web

    Hash
      4e90ad0d1fde0217f9683ea1f804af97
    Code
      {
          sy(async_pnc_bioc_query("all"))
      }

# pnc_bioc_process

    Hash
      4e90ad0d1fde0217f9683ea1f804af97
    Code
      {
          sy(async_pnc_bioc_query("all"))
      }

# vignettes can be turned on and off

    Code
      inst$solve()
      inst$download()
    Message
      i Getting 1 pkg with unknown size
      v Got pkgdependstest 1.0.0 (source)
    Code
      inst$install()
    Message
      i Packaging pkgdependstest 1.0.0
      v Packaged pkgdependstest 1.0.0
      i Building pkgdependstest 1.0.0
      v Built pkgdependstest 1.0.0
      v Installed pkgdependstest 1.0.0 (local)
      v Summary:   1 new

---

    Code
      inst2$solve()
      inst2$download()
    Message
      i Getting 1 pkg with unknown size
      v Got pkgdependstest 1.0.0 (source)
    Code
      inst2$install()
    Message
      i Packaging pkgdependstest 1.0.0
      v Packaged pkgdependstest 1.0.0
      i Building pkgdependstest 1.0.0
      v Built pkgdependstest 1.0.0
      v Installed pkgdependstest 1.0.0 (local)
      v Summary:   1 new

# dependencies

    Code
      p$draw()
    Output
      pkg 1.0.0 [new][bld][dl] (<size>)
      +-pd 1.0.0 [new][bld][dl] (<size>)
      | +-pdd 1.0.0 [new][bld][dl] (<size>)
      | +-pdi 1.0.0 [new][bld][dl] (<size>)
      | \-pdl 1.0.0 [new][bld][dl] (<size>)
      +-pi 1.0.0 [new][bld][dl] (<size>)
      | +-pid 1.0.0 [new][bld][dl] (<size>)
      | +-pii 1.0.0 [new][bld][dl] (<size>)
      | \-pil 1.0.0 [new][bld][dl] (<size>)
      \-pl 1.0.0 [new][bld][dl] (<size>)
        +-pld 1.0.0 [new][bld][dl] (<size>)
        +-pli 1.0.0 [new][bld][dl] (<size>)
        \-pll 1.0.0 [new][bld][dl] (<size>)
      
      Key:  [new] new | [dl] download | [bld] build

---

    Code
      p$draw()
    Output
      pkg 1.0.0 [new][bld][dl] (<size>)
      +-pd 1.0.0 [new][bld][dl] (<size>)
      | +-pdd 1.0.0 [new][bld][dl] (<size>)
      | +-pdi 1.0.0 [new][bld][dl] (<size>)
      | \-pdl 1.0.0 [new][bld][dl] (<size>)
      +-pi 1.0.0 [new][bld][dl] (<size>)
      | +-pid 1.0.0 [new][bld][dl] (<size>)
      | +-pii 1.0.0 [new][bld][dl] (<size>)
      | \-pil 1.0.0 [new][bld][dl] (<size>)
      +-pl 1.0.0 [new][bld][dl] (<size>)
      | +-pld 1.0.0 [new][bld][dl] (<size>)
      | +-pli 1.0.0 [new][bld][dl] (<size>)
      | \-pll 1.0.0 [new][bld][dl] (<size>)
      \-ps 1.0.0 [new][bld][dl] (<size>)
        +-psd 1.0.0 [new][bld][dl] (<size>)
        +-psi 1.0.0 [new][bld][dl] (<size>)
        \-psl 1.0.0 [new][bld][dl] (<size>)
      
      Key:  [new] new | [dl] download | [bld] build

---

    Code
      p$draw()
    Output
      pkg 1.0.0 [new][bld][dl] (<size>)
      \-pl 1.0.0 [new][bld][dl] (<size>)
        +-pld 1.0.0 [new][bld][dl] (<size>)
        +-pli 1.0.0 [new][bld][dl] (<size>)
        \-pll 1.0.0 [new][bld][dl] (<size>)
      
      Key:  [new] new | [dl] download | [bld] build

---

    Code
      plan$package
    Output
      [1] "pkg" "pl"  "pld" "pli" "pll"
    Code
      plan$dependencies
    Output
      [[1]]
      [1] "pl"
      
      [[2]]
      [1] "pld" "pli" "pll"
      
      [[3]]
      character(0)
      
      [[4]]
      character(0)
      
      [[5]]
      character(0)
      

# resolve_remote

    Code
      res[, c("ref", "sources")]
    Output
      # A data frame: 10 x 2
         ref                         sources                                          
       * <chr>                       <chr>                                            
       1 crayon=github::r-lib/crayon /repos/r-lib/crayon/zipball/bdd9a1bcf062396790c3~
       2 crayon=r-lib/crayon         /repos/r-lib/crayon/zipball/bdd9a1bcf062396790c3~
       3 gaborcsardi/secret          /repos/gaborcsardi/secret/zipball/7f9fb08e26015e~
       4 gaborcsardi/secret@x        /repos/gaborcsardi/secret/zipball/7f9fb08e26015e~
       5 github::r-lib/crayon        /repos/r-lib/crayon/zipball/bdd9a1bcf062396790c3~
       6 r-lib/crayon                /repos/r-lib/crayon/zipball/bdd9a1bcf062396790c3~
       7 r-lib/crayon#79             /repos/r-lib/crayon/zipball/9d93692f8f7c1d6b2308~
       8 r-lib/crayon@b5221ab0246050 /repos/r-lib/crayon/zipball/b5221ab024605019800d~
       9 r-lib/pkgconfig#7           /repos/r-lib/pkgconfig/zipball/c9be9cde5e91ad771~
      10 wesm/feather/R              /repos/wesm/feather/zipball/ec40c1eae1ac83b86fc4~

---

    Code
      res$metadata
    Output
      [[1]]
                                                              RemoteType 
                                                                "github" 
                                                              RemoteHost 
                                                      "127.0.0.1:<port>/" 
                                                              RemoteRepo 
                                                                "crayon" 
                                                          RemoteUsername 
                                                                 "r-lib" 
                                                            RemotePkgRef 
                                           "crayon=github::r-lib/crayon" 
                                                               RemoteRef 
                                                                  "HEAD" 
                                                               RemoteSha 
      "bdd9a1bcf062396790c341cf1dba563eb0277f2ca0a6d524bc3da98a9a6f2975" 
                                                              GithubRepo 
                                                                "crayon" 
                                                          GithubUsername 
                                                                 "r-lib" 
                                                               GithubRef 
                                                                  "HEAD" 
                                                              GithubSHA1 
      "bdd9a1bcf062396790c341cf1dba563eb0277f2ca0a6d524bc3da98a9a6f2975" 
      
      [[2]]
                                                              RemoteType 
                                                                "github" 
                                                              RemoteHost 
                                                      "127.0.0.1:<port>/" 
                                                              RemoteRepo 
                                                                "crayon" 
                                                          RemoteUsername 
                                                                 "r-lib" 
                                                            RemotePkgRef 
                                                   "crayon=r-lib/crayon" 
                                                               RemoteRef 
                                                                  "HEAD" 
                                                               RemoteSha 
      "bdd9a1bcf062396790c341cf1dba563eb0277f2ca0a6d524bc3da98a9a6f2975" 
                                                              GithubRepo 
                                                                "crayon" 
                                                          GithubUsername 
                                                                 "r-lib" 
                                                               GithubRef 
                                                                  "HEAD" 
                                                              GithubSHA1 
      "bdd9a1bcf062396790c341cf1dba563eb0277f2ca0a6d524bc3da98a9a6f2975" 
      
      [[3]]
                                                              RemoteType 
                                                                "github" 
                                                              RemoteHost 
                                                      "127.0.0.1:<port>/" 
                                                              RemoteRepo 
                                                                "secret" 
                                                          RemoteUsername 
                                                           "gaborcsardi" 
                                                            RemotePkgRef 
                                                    "gaborcsardi/secret" 
                                                               RemoteRef 
                                                                  "HEAD" 
                                                               RemoteSha 
      "7f9fb08e26015e05529cd4d7fc2a7edbd88c783d456ff83a96dcc58ace1d3ea5" 
                                                              GithubRepo 
                                                                "secret" 
                                                          GithubUsername 
                                                           "gaborcsardi" 
                                                               GithubRef 
                                                                  "HEAD" 
                                                              GithubSHA1 
      "7f9fb08e26015e05529cd4d7fc2a7edbd88c783d456ff83a96dcc58ace1d3ea5" 
      
      [[4]]
                                                              RemoteType 
                                                                "github" 
                                                              RemoteHost 
                                                      "127.0.0.1:<port>/" 
                                                              RemoteRepo 
                                                                "secret" 
                                                          RemoteUsername 
                                                           "gaborcsardi" 
                                                            RemotePkgRef 
                                                  "gaborcsardi/secret@x" 
                                                               RemoteRef 
                                                                     "x" 
                                                               RemoteSha 
      "7f9fb08e26015e05529cd4d7fc2a7edbd88c783d456ff83a96dcc58ace1d3ea5" 
                                                              GithubRepo 
                                                                "secret" 
                                                          GithubUsername 
                                                           "gaborcsardi" 
                                                               GithubRef 
                                                                     "x" 
                                                              GithubSHA1 
      "7f9fb08e26015e05529cd4d7fc2a7edbd88c783d456ff83a96dcc58ace1d3ea5" 
      
      [[5]]
                                                              RemoteType 
                                                                "github" 
                                                              RemoteHost 
                                                      "127.0.0.1:<port>/" 
                                                              RemoteRepo 
                                                                "crayon" 
                                                          RemoteUsername 
                                                                 "r-lib" 
                                                            RemotePkgRef 
                                                  "github::r-lib/crayon" 
                                                               RemoteRef 
                                                                  "HEAD" 
                                                               RemoteSha 
      "bdd9a1bcf062396790c341cf1dba563eb0277f2ca0a6d524bc3da98a9a6f2975" 
                                                              GithubRepo 
                                                                "crayon" 
                                                          GithubUsername 
                                                                 "r-lib" 
                                                               GithubRef 
                                                                  "HEAD" 
                                                              GithubSHA1 
      "bdd9a1bcf062396790c341cf1dba563eb0277f2ca0a6d524bc3da98a9a6f2975" 
      
      [[6]]
                                                              RemoteType 
                                                                "github" 
                                                              RemoteHost 
                                                      "127.0.0.1:<port>/" 
                                                              RemoteRepo 
                                                                "crayon" 
                                                          RemoteUsername 
                                                                 "r-lib" 
                                                            RemotePkgRef 
                                                          "r-lib/crayon" 
                                                               RemoteRef 
                                                                  "HEAD" 
                                                               RemoteSha 
      "bdd9a1bcf062396790c341cf1dba563eb0277f2ca0a6d524bc3da98a9a6f2975" 
                                                              GithubRepo 
                                                                "crayon" 
                                                          GithubUsername 
                                                                 "r-lib" 
                                                               GithubRef 
                                                                  "HEAD" 
                                                              GithubSHA1 
      "bdd9a1bcf062396790c341cf1dba563eb0277f2ca0a6d524bc3da98a9a6f2975" 
      
      [[7]]
                                                              RemoteType 
                                                                "github" 
                                                              RemoteHost 
                                                      "127.0.0.1:<port>/" 
                                                              RemoteRepo 
                                                                "crayon" 
                                                          RemoteUsername 
                                                                 "r-lib" 
                                                            RemotePkgRef 
                                                       "r-lib/crayon#79" 
                                                              RemotePull 
                                                                    "79" 
                                                               RemoteSha 
      "9d93692f8f7c1d6b2308d0c4aa83cdc2d99ec1fd0097cede1d9aa1301247cb01" 
                                                              GithubRepo 
                                                                "crayon" 
                                                          GithubUsername 
                                                                 "r-lib" 
                                                              GitHubPull 
                                                                    "79" 
                                                              GithubSHA1 
      "9d93692f8f7c1d6b2308d0c4aa83cdc2d99ec1fd0097cede1d9aa1301247cb01" 
      
      [[8]]
                                                              RemoteType 
                                                                "github" 
                                                              RemoteHost 
                                                      "127.0.0.1:<port>/" 
                                                              RemoteRepo 
                                                                "crayon" 
                                                          RemoteUsername 
                                                                 "r-lib" 
                                                            RemotePkgRef 
                                           "r-lib/crayon@b5221ab0246050" 
                                                               RemoteRef 
                                                        "b5221ab0246050" 
                                                               RemoteSha 
      "b5221ab024605019800ddea474f7a0981a4d53f719f5af2b1af627b34e0760b2" 
                                                              GithubRepo 
                                                                "crayon" 
                                                          GithubUsername 
                                                                 "r-lib" 
                                                               GithubRef 
                                                        "b5221ab0246050" 
                                                              GithubSHA1 
      "b5221ab024605019800ddea474f7a0981a4d53f719f5af2b1af627b34e0760b2" 
      
      [[9]]
                                                              RemoteType 
                                                                "github" 
                                                              RemoteHost 
                                                      "127.0.0.1:<port>/" 
                                                              RemoteRepo 
                                                             "pkgconfig" 
                                                          RemoteUsername 
                                                                 "r-lib" 
                                                            RemotePkgRef 
                                                     "r-lib/pkgconfig#7" 
                                                              RemotePull 
                                                                     "7" 
                                                               RemoteSha 
      "c9be9cde5e91ad771d1b5150781e6e8d32a7be0e9ab227bdf45cb41ad513004c" 
                                                              GithubRepo 
                                                             "pkgconfig" 
                                                          GithubUsername 
                                                                 "r-lib" 
                                                              GitHubPull 
                                                                     "7" 
                                                              GithubSHA1 
      "c9be9cde5e91ad771d1b5150781e6e8d32a7be0e9ab227bdf45cb41ad513004c" 
      
      [[10]]
                                                              RemoteType 
                                                                "github" 
                                                              RemoteHost 
                                                      "127.0.0.1:<port>/" 
                                                              RemoteRepo 
                                                               "feather" 
                                                          RemoteUsername 
                                                                  "wesm" 
                                                            RemotePkgRef 
                                                        "wesm/feather/R" 
                                                               RemoteRef 
                                                                  "HEAD" 
                                                               RemoteSha 
      "ec40c1eae1ac83b86fc41bb2f5cd916152d19015649c3d209f2c08115dd993b1" 
                                                            RemoteSubdir 
                                                                     "R" 
                                                              GithubRepo 
                                                               "feather" 
                                                          GithubUsername 
                                                                  "wesm" 
                                                               GithubRef 
                                                                  "HEAD" 
                                                              GithubSHA1 
      "ec40c1eae1ac83b86fc41bb2f5cd916152d19015649c3d209f2c08115dd993b1" 
                                                            GithubSubdir 
                                                                     "R" 
      

# new_pkg_download_proposal

    Code
      dl
    Output
      <pkg_download_proposal>
      + refs:
        - pkg3
      (use `$resolve()` to resolve dependencies)

---

    Code
      dl$get_refs()
    Output
      [1] "pkg3"

---

    Code
      sort(dl$get_config()$list())
    Output
       [1] "build_vignettes"           "cache_dir"                
       [3] "cran_mirror"               "dependencies"             
       [5] "git_submodules"            "goal"                     
       [7] "include_linkingto"         "library"                  
       [9] "metadata_cache_dir"        "metadata_update_after"    
      [11] "package_cache_dir"         "platforms"                
      [13] "r_versions"                "sysreqs"                  
      [15] "sysreqs_db_update"         "sysreqs_db_update_timeout"
      [17] "sysreqs_dry_run"           "sysreqs_lookup_system"    
      [19] "sysreqs_platform"          "sysreqs_rspm_repo_id"     
      [21] "sysreqs_rspm_url"          "sysreqs_sudo"             
      [23] "sysreqs_update"            "sysreqs_verbose"          
      [25] "use_bioconductor"          "windows_archs"            

---

    Code
      dl
    Output
      <pkg_download_proposal>
      + refs:
        - pkg3
      + has resolution (+2 dependencies)
      (use `$download()` to download packages)
      (use `$get_resolution()` to see resolution results)

---

    Code
      dl$get_resolution()[, c("ref", "type", "directpkg", "package", "error")]
    Output
      # A data frame: 3 x 5
        ref   type     directpkg package error     
        <chr> <chr>    <lgl>     <chr>   <list>    
      1 pkg1  standard FALSE     pkg1    <list [0]>
      2 pkg2  standard FALSE     pkg2    <list [0]>
      3 pkg3  standard TRUE      pkg3    <list [0]>

# async_resolve

    Code
      synchronize(dl$async_resolve()$then(function() "done"))
    Output
      [1] "done"

---

    Code
      dl$get_resolution()[, c("ref", "type", "directpkg", "package", "error")]
    Output
      # A data frame: 3 x 5
        ref   type     directpkg package error     
        <chr> <chr>    <lgl>     <chr>   <list>    
      1 pkg1  standard FALSE     pkg1    <list [0]>
      2 pkg2  standard FALSE     pkg2    <list [0]>
      3 pkg3  standard TRUE      pkg3    <list [0]>

# download

    Code
      dir(tmp, recursive = TRUE, pattern = "[.]tar[.]gz$")
    Output
      [1] "src/contrib/pkg1_1.0.0.tar.gz" "src/contrib/pkg2_1.0.0.tar.gz"
      [3] "src/contrib/pkg3_1.0.0.tar.gz"

# async_download

    Code
      dir(tmp, recursive = TRUE, pattern = "[.]tar[.]gz$")
    Output
      [1] "src/contrib/pkg1_1.0.0.tar.gz" "src/contrib/pkg2_1.0.0.tar.gz"
      [3] "src/contrib/pkg3_1.0.0.tar.gz"

---

    Code
      dl$get_downloads()$target
    Output
      [1] "src/contrib/pkg1_1.0.0.tar.gz" "src/contrib/pkg2_1.0.0.tar.gz"
      [3] "src/contrib/pkg3_1.0.0.tar.gz"

# type_github_get_data, sha, description

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak")))
    Output
      $sha
      [1] "111ef906acb58fe406370f7bc0a72cac55dbbb231ea687494c25742ca521255a"
      
      $description
      Package: pak
      Version: 1.0.0
      

---

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak@HEAD")))
    Output
      $sha
      [1] "111ef906acb58fe406370f7bc0a72cac55dbbb231ea687494c25742ca521255a"
      
      $description
      Package: pak
      Version: 1.0.0
      

---

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak@v0.1.2")))
    Output
      $sha
      [1] "a503fe843f11c279864f29d58137f8de319d115b239ce48ccc15406306019480"
      
      $description
      Package: pak
      Version: 1.0.0
      

---

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak@*release")))
    Output
      $sha
      [1] "b001d6ddeab1589ad367b62baabbeeb2af3b0ebac2e61d239df660c1d63e3232"
      
      $description
      Package: pak
      Version: 1.0.0
      

---

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak@e65de1e9630d")))
    Output
      $sha
      [1] "e65de1e9630dbfcaf1044718b742bf806486b107239ce48ccc15406306019480"
      
      $description
      Package: pak
      Version: 1.0.0
      

---

    Code
      synchronise(type_github_get_data(parse_pkg_ref(
        "r-lib/pak@e65de1e9630dbfcaf1044718b742bf806486b107")))
    Output
      $sha
      [1] "e65de1e9630dbfcaf1044718b742bf806486b107239ce48ccc15406306019480"
      
      $description
      Package: pak
      Version: 1.0.0
      

---

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak#90")))
    Output
      $sha
      [1] "b001d6ddeab1589ad367b62baabbeeb2af3b0ebac2e61d239df660c1d63e3232"
      
      $description
      Package: pak
      Version: 1.0.0
      

---

    Code
      synchronise(type_github_get_data(parse_pkg_ref(
        "wesm/feather/R@ec40c1eae1ac83b86fc41bb2f5cd916152d19015")))
    Output
      $sha
      [1] "ec40c1eae1ac83b86fc41bb2f5cd916152d19015649c3d209f2c08115dd993b1"
      
      $description
      Package: feather
      Version: 1.0.0
      

# type_github_get_data, no such user

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib-xxx-xxx/pak")))
    Condition
      Error:
      ! ! Can't find GitHub repo r-lib-xxx-xxx/pak.

---

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib-xxx-xxx/pak#90")))
    Condition
      Error:
      ! ! Can't find GitHub repo r-lib-xxx-xxx/pak.

# type_github_get_data, no such repo

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak-xxx-xxx")))
    Condition
      Error:
      ! ! Can't find GitHub repo r-lib/pak-xxx-xxx.

---

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak-xxx-xxx#90")))
    Condition
      Error:
      ! ! Can't find GitHub repo r-lib/pak-xxx-xxx.

# github_query, invalid PAT

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak-xxx-xxx")))
    Condition
      Error:
      ! ! Bad GitHub credentials, make sure that your GitHub token is valid.
      Caused by error in `stop(http_error(resp))`:
      ! Unauthorized (HTTP 401).

---

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak-xxx-xxx#90")))
    Condition
      Error:
      ! ! Bad GitHub credentials, make sure that your GitHub token is valid.
      Caused by error in `stop(http_error(resp))`:
      ! Unauthorized (HTTP 401).

# github_query, no internet

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak")))
    Condition
      Error:
      ! ! Cannot query GitHub, are you offline?

# github_query, access denied

    Code
      synchronise(type_github_get_data(parse_pkg_ref("gaborcsardi/secret-test")))
    Condition
      Error:
      ! ! Can't find GitHub repo gaborcsardi/secret-test.

# cannot find R package on GitHub, no DESCRIPTION

    Code
      synchronise(type_github_get_data(parse_pkg_ref("tidyverse/tidyverse.org")))
    Condition
      Error:
      ! ! Can't find R package in GitHub repo tidyverse/tidyverse.org

---

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib/cranyon/R")))
    Condition
      Error:
      ! ! Can't find GitHub repo r-lib/cranyon.

---

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib/crayon/R#79")))
    Condition
      Error:
      ! ! Can't find R package in GitHub repo r-lib/crayon in directory 'R'

# cannot parse DESCRIPTION on GH

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib/bad@main")))
    Condition
      Error:
      ! ! Can't parse DESCRIPTION file in GitHub repo r-lib/bad
      Caused by error in `read.dcf(con)`:
      ! Line starting 'this is not ...' is malformed!

---

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib/bad#100")))
    Condition
      Error:
      ! ! Can't parse DESCRIPTION file in GitHub repo r-lib/bad
      Caused by error in `read.dcf(con)`:
      ! Line starting 'this is not ...' is malformed!

---

    Code
      synchronize(type_github_get_data(parse_pkg_ref("r-lib/bad/bin@main")))
    Condition
      Error:
      ! ! Can't parse DESCRIPTION file in GitHub repo r-lib/bad, in directory `bin`

# http error

    Code
      synchronise(type_github_get_data(parse_pkg_ref("foo/bar")))
    Condition
      Error:
      ! ! GitHub HTTP error
      Caused by error in `stop(http_error(resp))`:
      ! Not Found (HTTP 404).

# no such PR error

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak#89")))
    Condition
      Error:
      ! ! Can't find PR #89 in GitHub repo r-lib/pak

# no such ref error

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib/pak@bad-ref-no-no-no")))
    Condition
      Error:
      ! ! Can't find reference @bad-ref-no-no-no in GitHub repo r-lib/pak.

# no release error

    Code
      synchronise(type_github_get_data(parse_pkg_ref("r-lib/bad@*release")))
    Condition
      Error:
      ! ! Can't find any release in GitHub repo r-lib/bad.

# builtin token messages once per session

    Code
      type_github_builtin_token()
    Message
      ! Using bundled GitHub PAT. Please add your own PAT using `gitcreds::gitcreds_set()`.
    Output
      [1] "builtin-token"

---

    Code
      type_github_builtin_token()
    Output
      [1] "builtin-token"

# CI specific token is picked up if set

    Code
      type_github_get_headers()
    Output
                                 Accept                      Content-Type 
       "application/vnd.github.v3+json" "application/json; charset=utf-8" 
                          Authorization                        User-Agent 
                       "token ci-token"                       "r-lib/pak" 

# builtin token is used if no other token is available

    Code
      type_github_get_headers()
    Output
                                 Accept                      Content-Type 
       "application/vnd.github.v3+json" "application/json; charset=utf-8" 
                          Authorization                        User-Agent 
                  "token builtin-token"                       "r-lib/pak" 

# recommended_packages

    Code
      recommended_packages()
    Output
       [1] "boot"       "class"      "cluster"    "codetools"  "foreign"   
       [6] "KernSmooth" "lattice"    "MASS"       "Matrix"     "mgcv"      
      [11] "nlme"       "nnet"       "rpart"      "spatial"    "survival"  

# vlapply

    Code
      vlapply(l, identity)
    Condition
      Error in `vapply()`:
      ! values must be length 1,
       but FUN(X[[1]]) result is length 0
    Code
      vlapply(1:5, identity)
    Condition
      Error in `vapply()`:
      ! values must be type 'logical',
       but FUN(X[[1]]) result is type 'integer'

# viapply

    Code
      viapply(c(a = 1L, b = 2L), function(x) 1)
    Condition
      Error in `vapply()`:
      ! values must be type 'integer',
       but FUN(X[[1]]) result is type 'double'

# vdapply

    Code
      vdapply(l, identity)
    Condition
      Error in `vapply()`:
      ! values must be length 1,
       but FUN(X[[1]]) result is length 0
    Code
      vdapply(letters, identity)
    Condition
      Error in `vapply()`:
      ! values must be type 'double',
       but FUN(X[[1]]) result is type 'character'

# cat0

    Code
      cat0("foo", "bar", "\n")
    Output
      foobar
    Code
      cat0("foo", "bar", "\n", sep = " ")
    Output
      foo bar 

# lapply_rows

    Code
      lapply_rows(mtcars[1:3, ], function(row) row)
    Output
      [[1]]
                mpg cyl disp  hp drat   wt  qsec vs am gear carb
      Mazda RX4  21   6  160 110  3.9 2.62 16.46  0  1    4    4
      
      [[2]]
                    mpg cyl disp  hp drat    wt  qsec vs am gear carb
      Mazda RX4 Wag  21   6  160 110  3.9 2.875 17.02  0  1    4    4
      
      [[3]]
                  mpg cyl disp hp drat   wt  qsec vs am gear carb
      Datsun 710 22.8   4  108 93 3.85 2.32 18.61  1  1    4    1
      

# rbind_expand

    Code
      rbind_expand(data_frame(), data_frame())
    Output
      data frame with 0 columns and 0 rows
    Code
      rbind_expand(data_frame(), data_frame(foo = 1:2))
    Output
      # A data frame: 2 x 1
          foo
        <int>
      1     1
      2     2
    Code
      rbind_expand(data_frame(bar = c("a", "b")), data_frame())
    Output
      # A data frame: 2 x 1
        bar  
        <chr>
      1 a    
      2 b    
    Code
      rbind_expand(data_frame(foo = 1:2), data_frame(foo = 3:4))
    Output
      # A data frame: 4 x 1
          foo
        <int>
      1     1
      2     2
      3     3
      4     4
    Code
      rbind_expand(data_frame(foo = 1:2), data_frame(bar = 3:4))
    Output
      # A data frame: 4 x 2
          foo   bar
        <int> <int>
      1     1    NA
      2     2    NA
      3    NA     3
      4    NA     4
    Code
      rbind_expand(data_frame(foo = list(1, 2), bar = letters[1:2]), data_frame(foo = list(
        3, 4), baz = list("x", "y")))
    Output
      # A data frame: 4 x 3
        foo       bar   baz      
        <list>    <chr> <list>   
      1 <dbl [1]> a     <lgl [1]>
      2 <dbl [1]> b     <lgl [1]>
      3 <dbl [1]> <NA>  <chr [1]>
      4 <dbl [1]> <NA>  <chr [1]>

# update_named_vector

    Code
      update_named_vector(1, c(a = 1))
    Condition
      Error:
      ! ! All elements in `old` must be named.
    Code
      update_named_vector(c(a = 1), 1)
    Condition
      Error:
      ! ! All elements in `new` must be named.

# once_per_session

    Code
      once_per_session(message("hello"))
    Message
      hello
    Code
      once_per_session(message("hello"))
      once_per_session(reset = TRUE)
      once_per_session(message("hello"))
    Message
      hello
    Code
      once_per_session(message("hello"))

# format_error_with_stdout

    Code
      format_error_with_stdout(err)
    Output
      [1] "! message (output not available)"

---

    Code
      format_error_with_stdout(err)
    Output
      [1] "! message, stdout + stderr:" ""                           
      [3] "OE> this is"                 "OE> the"                    
      [5] "OE> standard output"        

# last_stdout_lines

    Code
      last_stdout_lines(letters[1:3], "stdout + stderr")
    Output
      [1] ", stdout + stderr:" ""                   "E> a"              
      [4] "E> b"               "E> c"              

---

    Code
      last_stdout_lines(letters[1:11], "stdout + stderr")
    Output
       [1] ", stdout + stderr (last 10 lines):" ""                                  
       [3] "E> b"                               "E> c"                              
       [5] "E> d"                               "E> e"                              
       [7] "E> f"                               "E> g"                              
       [9] "E> h"                               "E> i"                              
      [11] "E> j"                               "E> k"                              

---

    Code
      last_stdout_lines(letters[1:11], "stdout + stderr")
    Output
       [1] ", stdout + stderr:" ""                   "E> a"              
       [4] "E> b"               "E> c"               "E> d"              
       [7] "E> e"               "E> f"               "E> g"              
      [10] "E> h"               "E> i"               "E> j"              
      [13] "E> k"              

# ansi_align_width [plain]

    Code
      ansi_align_width(c("foobar", cli::col_red("bar")))
    Output
      <cli_ansi_string>
      [1] foobar
      [2] bar   

# ansi_align_width [ansi]

    Code
      ansi_align_width(c("foobar", cli::col_red("bar")))
    Output
      <cli_ansi_string>
      [1] foobar                
      [2] [31mbar[39m   

# different versions required

    Code
      p$stop_for_solution_error()
    Condition
      Error:
      * r-lib/crayon: Conflicts with crayon
      * crayon: Conflicts with r-lib/crayon
      ! ! Could not solve package dependencies:

# direct CRAN conflicts with downstream GH dep

    Code
      p$stop_for_solution_error()
    Condition
      Error:
      ! ! Could not solve package dependencies:
      * r-lib/foo: Can't install dependency r-lib/crayon
      * r-lib/crayon: Conflicts with crayon

# no required version

    Code
      p$stop_for_solution_error()
    Condition
      Error:
      ! ! Could not solve package dependencies:
      * pkg1: Can't install dependency pkg2 (>= 2.0.0)

# failed resolution

    Code
      p1$stop_for_solution_error()
    Condition
      Error:
      ! ! Could not solve package dependencies:
      * SDF: Can't find package called SDF.

---

    Code
      p2$stop_for_solution_error()
    Condition
      Error:
      ! ! Could not solve package dependencies:
      * SDF/SDF: ! pkgdepends resolution error for SDF/SDF.
      Caused by error: 
      ! Can't find GitHub repo SDF/SDF.

# is_character errors

    Code
      r_process()
    Output
      > fn <- function(x) assert_that(is_character(x))
      > fn(1:2)
      [1m[33mError[39m[22m in `fn(1:2)`:
      [33m![39m [1m[22m`x` must be a character vector without `NA`, but it is an integer vector
      [90mType .Last.error to see the more details.[39m
      > fn(c("", NA_character_))
      [1m[33mError[39m[22m in `fn(c("", NA_character_))`:
      [33m![39m [1m[22m`x` must be a character vector without `NA`, but it has 1 `NA` value.
      [90mType .Last.error to see the more details.[39m
      > fn(rep(NA_character_, 5))
      [1m[33mError[39m[22m in `fn(rep(NA_character_, 5))`:
      [33m![39m [1m[22m`x` must be a character vector without `NA`, but it has 5 `NA` values.
      [90mType .Last.error to see the more details.[39m

# is_character errors, noninteractive

    Code
      r_process()
    Output
      > fn <- function(x) assert_that(is_character(x))
      > options(cli.unicode = FALSE)
      > fn(1:2)
      [1m[33mError[39m[22m in `fn(1:2)`:
      [33m![39m [1m[22m`x` must be a character vector without `NA`, but it is an integer vector
      ---
      Backtrace:
      [90m1. [39mglobal [1mfn[22m[38;5;178m([38;5;169m1[38;5;178m:[38;5;169m2[38;5;178m)[39m
      [90m2. | pkgdepends:::assert_that(is_character(x))[39m
      [90m3. | pkgdepends:::throw(assert_error(assertion, res, msg, call. = sys.call(-1), ...[39m
      Execution halted

---

    Code
      r_process()
    Output
      > fn <- function(x) assert_that(is_character(x))
      > options(cli.unicode = FALSE)
      > fn(c("", NA_character_))
      [1m[33mError[39m[22m in `fn(c("", NA_character_))`:
      [33m![39m [1m[22m`x` must be a character vector without `NA`, but it has 1 `NA` value.
      ---
      Backtrace:
      [90m1. [39mglobal [1mfn[22m[38;5;178m([39m[1mc[22m[33m([39m[38;5;37m""[39m, [38;5;169mNA_character_[39m[33m)[39m[38;5;178m)[39m
      [90m2. | pkgdepends:::assert_that(is_character(x))[39m
      [90m3. | pkgdepends:::throw(assert_error(assertion, res, msg, call. = sys.call(-1), ...[39m
      Execution halted

---

    Code
      r_process()
    Output
      > fn <- function(x) assert_that(is_character(x))
      > options(cli.unicode = FALSE)
      > fn(rep(NA_character_, 5))
      [1m[33mError[39m[22m in `fn(rep(NA_character_, 5))`:
      [33m![39m [1m[22m`x` must be a character vector without `NA`, but it has 5 `NA` values.
      ---
      Backtrace:
      [90m1. [39mglobal [1mfn[22m[38;5;178m([39m[1mrep[22m[33m([39m[38;5;169mNA_character_[39m, [38;5;169m5[39m[33m)[39m[38;5;178m)[39m
      [90m2. | pkgdepends:::assert_that(is_character(x))[39m
      [90m3. | pkgdepends:::throw(assert_error(assertion, res, msg, call. = sys.call(-1), ...[39m
      Execution halted

# is_string errors

    Code
      r_process()
    Output
      > fn <- function(x) assert_that(is_string(x))
      > fn(1:2)
      [1m[33mError[39m[22m in `fn(1:2)`:
      [33m![39m [1m[22m`x` must be a string (character scalar), but it is an integer vector.
      [90mType .Last.error to see the more details.[39m
      > fn("foo")
      > fn(NA_character_)
      [1m[33mError[39m[22m in `fn(NA_character_)`:
      [33m![39m [1m[22m`x` must not be `NA`.
      [90mType .Last.error to see the more details.[39m
      > fn(NULL)
      [1m[33mError[39m[22m in `fn(NULL)`:
      [33m![39m [1m[22m`x` must be a string (character scalar), but it is NULL.
      [90mType .Last.error to see the more details.[39m

# is_optional_string errors

    Code
      r_process()
    Output
      > fn <- function(x) assert_that(is_optional_string(x))
      > fn(1:2)
      [1m[33mError[39m[22m in `fn(1:2)`:
      [33m![39m [1m[22m`x` must be a path (character scalar), but it is an integer vector.
      [90mType .Last.error to see the more details.[39m
      > fn(c("foo", "bar"))
      [1m[33mError[39m[22m in `fn(c("foo", "bar"))`:
      [33m![39m [1m[22m`x` must be a path (character scalar), but it is a character vector.
      [90mType .Last.error to see the more details.[39m
      > fn(NA_character_)
      [1m[33mError[39m[22m in `fn(NA_character_)`:
      [33m![39m [1m[22m`x` must be a path (character scalar), but it is a character `NA`.
      [90mType .Last.error to see the more details.[39m

# is_flag errors

    Code
      r_process()
    Output
      > fn <- function(x) assert_that(is_flag(x))
      > fn(NA)
      [1m[33mError[39m[22m in `fn(NA)`:
      [33m![39m [1m[22m`x` must not be `NA`.
      [90mType .Last.error to see the more details.[39m
      > fn(1:10)
      [1m[33mError[39m[22m in `fn(1:10)`:
      [33m![39m [1m[22m`x` must be a flag (logical scalar), but it is an integer vector.
      [90mType .Last.error to see the more details.[39m

# is_path errors

    Code
      r_process()
    Output
      > fn <- function(x) assert_that(is_path(x))
      > fn(1:2)
      [1m[33mError[39m[22m in `fn(1:2)`:
      [33m![39m [1m[22m`x` must be a path (character scalar), but it is an integer vector.
      [90mType .Last.error to see the more details.[39m
      > fn(c("foo", "bar"))
      [1m[33mError[39m[22m in `fn(c("foo", "bar"))`:
      [33m![39m [1m[22m`x` must be a path (character scalar), but it is a character vector.
      [90mType .Last.error to see the more details.[39m
      > fn(NA_character_)
      [1m[33mError[39m[22m in `fn(NA_character_)`:
      [33m![39m [1m[22m`x` must be a path (character scalar), but it is a character `NA`.
      [90mType .Last.error to see the more details.[39m

# is_optional path errors

    Code
      r_process()
    Output
      > fn <- function(x) assert_that(is_optional_path(x))
      > fn(1:2)
      [1m[33mError[39m[22m in `fn(1:2)`:
      [33m![39m [1m[22m`x` must be a path (character scalar) or `NULL`, but it is an integer
      vector.
      [90mType .Last.error to see the more details.[39m
      > fn(c("foo", "bar"))
      [1m[33mError[39m[22m in `fn(c("foo", "bar"))`:
      [33m![39m [1m[22m`x` must be a path (character scalar) or `NULL`, but it is a character
      vector.
      [90mType .Last.error to see the more details.[39m
      > fn(NA_character_)
      [1m[33mError[39m[22m in `fn(NA_character_)`:
      [33m![39m [1m[22m`x` must be a path (character scalar) or `NULL`, but it is a character
      `NA`.
      [90mType .Last.error to see the more details.[39m

# all_named errors

    Code
      r_process()
    Output
      > fn <- function(x) assert_that(all_named(x))
      > fn(c(a = 1, 2))
      [1m[33mError[39m[22m in `fn(c(a = 1, 2))`:
      [33m![39m [1m[22mAll elements in `x` must be named.
      [90mType .Last.error to see the more details.[39m

# is_existing_file errors

    Code
      r_process()
    Output
      > fn <- function(x) assert_that(is_existing_file(x))
      > fn("file1500454880b58")
      [1m[33mError[39m[22m in `fn("file1500454880b58")`:
      [33m![39m [1m[22mPath [34mfile1500454880b58[39m (from `x`) does not exist
      [90mType .Last.error to see the more details.[39m
      > fn(".")
      [1m[33mError[39m[22m in `fn(".")`:
      [33m![39m [1m[22mFile [34m.[39m (from `x`) must be a regular file, not a directory
      [90mType .Last.error to see the more details.[39m

# is_platform_list errors

    Code
      r_process()
    Output
      > fn <- function(x) assert_that(is_platform_list(x))
      > fn(1:10)
      [1m[33mError[39m[22m in `fn(1:10)`:
      [33m![39m [1m[22m`x` must be a list of platforms, a non-empty character vector, but it is
      an integer vector
      [90mType .Last.error to see the more details.[39m
      > fn(character())
      [1m[33mError[39m[22m in `fn(character())`:
      [33m![39m [1m[22m`x` must be a list of platforms, a non-empry character vector
      [90mType .Last.error to see the more details.[39m
      > fn(NA_character_)
      [1m[33mError[39m[22m in `fn(NA_character_)`:
      [33m![39m [1m[22m`((.arg)`} must be a list of platforms, a character vector without
      missing values, but it has 1 missing value.
      [90mType .Last.error to see the more details.[39m

# is_dependencies errors

    Code
      r_process()
    Output
      > fn <- function(x) assert_that(is_dependencies(x))
      > options(cli.unicode = FALSE)
      > fn(c("Depends", NA))
      [1m[33mError[39m[22m in `fn(c("Depends", NA))`:
      [33m![39m [1m[22m`x` must be one of the following: `NA`, `TRUE`, `FALSE`, a character
      vector of dependency types, a named list with entries `direct` and `indirect`,
      both character vectors of dependency types.
      [36mi[39m valid dependency types are: [34m"Depends"[39m, [34m"Imports"[39m, [34m"LinkingTo"[39m, [34m"Suggests"[39m,
        [34m"Enhances"[39m, [34m"soft"[39m, [34m"hard"[39m, and [34m"all"[39m, and `config/needs/*` types
      [90mType .Last.error to see the more details.[39m

# is_r_version_list errors

    Code
      r_process()
    Output
      > fn <- function(x) assert_that(is_r_version_list(x))
      > fn(1:10)
      [1m[33mError[39m[22m in `fn(1:10)`:
      [33m![39m [1m[22m`x` must be a list of R versions, a character vector, but it is an
      integer vector.
      [90mType .Last.error to see the more details.[39m
      > fn(character())
      [1m[33mError[39m[22m in `fn(character())`:
      [33m![39m [1m[22m`x` must be a non-empty list of R versions, a character vector, but it
      is empty.
      [90mType .Last.error to see the more details.[39m
      > fn(c("1.2", "foo", "bar", "1.2.0"))
      [1m[33mError[39m[22m in `lapply(text, glue_cmd, .envir = .envir)`:
      [33m![39m [1m[22mCould not evaluate cli `{}` expression: `sum(is.na(cv))`.
      [1mCaused by error[22m in `eval(expr, envir = envir)`:
      [33m![39m object 'cv' not found
      [90mType .Last.error to see the more details.[39m

# is_difftime errors

    Code
      r_process()
    Output
      > fn <- function(x) assert_that(is_difftime(x))
      > fn(1)
      [1m[33mError[39m[22m in `fn(1)`:
      [33m![39m [1m[22m`x` must be a [34m<difftime>[39m object, but it is a number.
      [90mType .Last.error to see the more details.[39m

# is_count errors

    Code
      fn(letters)
    Condition
      Error:
      ! ! `x` must be a count, a non-negative integer scalar.
      i It is a character vector.

---

    Code
      fn(1:10)
    Condition
      Error:
      ! ! `x` must be a count, a non-negative integer scalar.
      i It is an integer vector.

---

    Code
      fn(-1)
    Condition
      Error:
      ! ! `x` must be at least 0.

---

    Code
      fn(0, min = 1)
    Condition
      Error:
      ! ! `x` must be at least 1.

---

    Code
      fn(NA_integer_)
    Condition
      Error:
      ! ! `x` must not be a missing value (`NA`).

# is_valid_package_name

    Code
      is_valid_package_name("foo")
    Output
      [1] TRUE
    Code
      is_valid_package_name("pkg")
    Output
      [1] FALSE
      attr(,"reason")
      [1] "Package name forbidden by CRAN."
    Code
      is_valid_package_name("fáf")
    Output
      [1] FALSE
      attr(,"reason")
      [1] "It can only contain ASCII characters."
    Code
      is_valid_package_name("foo-bar")
    Output
      [1] FALSE
      attr(,"reason")
      [1] "It can only contain letters, numbers and dot."
    Code
      is_valid_package_name("x")
    Output
      [1] FALSE
      attr(,"reason")
      [1] "It must have at least two characters."
    Code
      is_valid_package_name("1xyz")
    Output
      [1] FALSE
      attr(,"reason")
      [1] "It must start with a letter."
    Code
      is_valid_package_name("dotted.")
    Output
      [1] FALSE
      attr(,"reason")
      [1] "It must not end with a dot."

# url remote basics

    Code
      r$get_solution()
    Output
      <pkg_solution>
      + result: OK
      + refs:
        - url::http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz
      + constraints (1):
        - select pkg1 exactly once
      + solution:
        - url::http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz

# url to an tree (not a source package), from a zip file

    Code
      r$get_solution()
    Output
      <pkg_solution>
      + result: OK
      + refs:
        - url::http://127.0.0.1:<port>/repos/r-lib/pak/zipball/111ef906acb58fe406370f7bc0a72cac55dbbb231ea687494c25742ca521255a
      + constraints (1):
        - select pak exactly once
      + solution:
        - url::http://127.0.0.1:<port>/repos/r-lib/pak/zipball/111ef906acb58fe406370f7bc0a72cac55dbbb231ea687494c25742ca521255a

# satisfy

    Code
      satisfy_remote_url(list(package = "package"), list(package = "other"))
    Output
      [1] FALSE
      attr(,"reason")
      [1] "Package names differ"

---

    Code
      satisfy_remote_url(list(package = "package", params = list(c(reinstall = ""))),
      list(type = "installed", package = "package"))
    Output
      [1] FALSE
      attr(,"reason")
      [1] "Re-install requested"

---

    Code
      satisfy_remote_url(list(package = "package", metadata = list(list(RemoteEtag = "myetag"))),
      list(type = "installed", package = "package"))
    Output
      [1] FALSE
      attr(,"reason")
      [1] "Installed URL etag mismatch"

---

    Code
      satisfy_remote_url(list(package = "package", metadata = list(list(RemoteEtag = "myetag"))),
      list(type = "installed", package = "package", extra = list(list(remoteetag = "otheretag"))))
    Output
      [1] FALSE
      attr(,"reason")
      [1] "Installed URL etag mismatch"

---

    Code
      satisfy_remote_url(list(package = "package", metadata = list(list(RemoteEtag = "myetag"))),
      list(type = "installed", package = "package", extra = list(list(remoteetag = "myetag"))))
    Output
      [1] TRUE

---

    Code
      satisfy_remote_url(list(package = "package", ref = "https://myurl"), list(type = "url",
        package = "package", ref = "https://anotherurl"))
    Output
      [1] FALSE
      attr(,"reason")
      [1] "URL mismatch"

---

    Code
      satisfy_remote_url(list(package = "package", ref = "https://myurl"), list(type = "url",
        package = "package", ref = "https://myurl"))
    Output
      [1] TRUE

---

    Code
      satisfy_remote_url(list(package = "package", ref = "https://myurl"), list(type = "local",
        package = "package"))
    Output
      [1] FALSE
      attr(,"reason")
      [1] "Repo type mismatch"

# nocache

    Code
      r$get_solution()
    Output
      <pkg_solution>
      + result: OK
      + refs:
        - url::http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz
      + constraints (1):
        - select pkg1 exactly once
      + solution:
        - url::http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz

# bad archive file (multiple directories)

    Code
      get_pkg_dir_from_archive_dir(tmp)
    Condition
      Error:
      ! ! Package archive at '<tempdir>/<tempfile>' should contain exactly one directory.
      i It has 2 files/directories: 'pkg1' and 'pkg2'

# async_system_list_packages_dpkg_query

    Code
      ans
    Output
      # A data frame: 3 x 4
        status package      version      provides 
        <chr>  <chr>        <chr>        <list>   
      1 ii     adduser      3.118ubuntu5 <chr [0]>
      2 ii     apt          2.4.9        <chr [1]>
      3 ii     x11proto-dev 2021.5-1     <chr [4]>

# async_system_list_packages_rpm

    Code
      ans
    Output
      # A data frame: 7 x 4
        status package                  version                  provides  
      * <chr>  <chr>                    <chr>                    <list>    
      1 ii     crypto-policies          crypto-policies          <chr [1]> 
      2 ii     gpg-pubkey               gpg-pubkey               <chr [5]> 
      3 ii     libgcc                   libgcc                   <chr [16]>
      4 ii     ncurses-base             ncurses-base             <chr [0]> 
      5 ii     pcre2-syntax             pcre2-syntax             <chr [0]> 
      6 ii     python3-setuptools-wheel python3-setuptools-wheel <chr [7]> 
      7 ii     tzdata                   tzdata                   <chr [0]> 

# make_null_deps

    Code
      make_null_deps()
    Output
      # A data frame: 0 x 5
      # i 5 variables: ref <chr>, type <chr>, package <chr>, op <chr>, version <chr>

# parse_deps

    Code
      parse_deps("grid, stats", "Imports")
    Output
      [[1]]
      # A data frame: 0 x 4
      # i 4 variables: type <chr>, package <chr>, op <chr>, version <chr>
      

# parse_all_deps

    Code
      parse_all_deps(c(Imports = NA_character_))
    Output
      # A data frame: 0 x 5
      # i 5 variables: ref <chr>, type <chr>, package <chr>, op <chr>, version <chr>

---

    Code
      parse_all_deps(c(Imports = "grid"))
    Output
      # A data frame: 0 x 5
      # i 5 variables: ref <chr>, type <chr>, package <chr>, op <chr>, version <chr>

---

    Code
      parse_all_deps(c(Imports = "foo (>= 1.0.0), bar", Suggests = "baz, foobaz (>= 2.0.0)"))
    Output
      # A data frame: 4 x 5
        ref    type     package op    version
        <chr>  <chr>    <chr>   <chr> <chr>  
      1 foo    Imports  foo     ">="  "1.0.0"
      2 bar    Imports  bar     ""    ""     
      3 baz    Suggests baz     ""    ""     
      4 foobaz Suggests foobaz  ">="  "2.0.0"

# install_extracted_binary, add_metadata

    Code
      dir(pkg_cache)
    Output
      [1] "foo"

---

    Code
      d <- desc::desc(file.path(lib, "foo"))
      d$get_field("Package")
    Output
      [1] "foo"
    Code
      d$get_field("Foo")
    Output
      [1] "bar"
    Code
      d$get_field("Foobar")
    Output
      [1] "baz"

---

    Code
      rds$DESCRIPTION[c("Package", "Foo", "Foobar")]
    Output
      Package     Foo  Foobar 
        "foo"   "bar"   "baz" 

---

    Code
      d <- desc::desc(file.path(lib, "foo"))
      d$get_field("Package")
    Output
      [1] "foo"
    Code
      d$get_field("Another")
    Output
      [1] "field"

---

    Code
      rds$DESCRIPTION[c("Package", "Another")]
    Output
      Package Another 
        "foo" "field" 

---

    Code
      dir(pkg_cache)
    Output
      [1] "foo"

---

    Code
      d <- desc::desc(file.path(lib, "foo"))
      d$get_field("Package")
    Output
      [1] "foo"
    Code
      d$get_field("Foo")
    Output
      [1] "bar2"
    Code
      d$get_field("Foobar")
    Output
      [1] "baz2"

---

    Code
      rds$DESCRIPTION[c("Package", "Foo", "Foobar")]
    Output
      Package     Foo  Foobar 
        "foo"  "bar2"  "baz2" 

# add_metadata error

    Code
      add_metadata(tmp, c(foo = "bar"))
    Condition
      Error:
      ! ! Could not find 'DESCRIPTION' file when installing package into '<tempdir>/<tempfile>'.
      i This is an internal error in pkgdepends, please report an issue at <https://github.com/r-lib/pkgdepends/issues>.

# parse_remote_any

    Code
      plan$draw()
    Output
      any::pkg3 1.0.0 [new][bld][dl] (<size>)
      \-pkg2 1.0.0 [new][bld][dl] (<size>)
        \-pkg1 1.0.0 [new][bld][dl] (<size>)
      
      Key:  [new] new | [dl] download | [bld] build

---

    Code
      plan$draw()
    Output
      pkg3 1.0.0 
      \-pkg2 1.0.0 
        \-pkg1 1.0.0 
      

# dep_types

    Code
      as.list(res[c("ref", "dep_types")])
    Output
      $ref
      [1] "any::pkg3" "pkg1"      "pkg2"      "pkg3"     
      
      $dep_types
      $dep_types[[1]]
      [1] "Depends"   "Imports"   "LinkingTo"
      
      $dep_types[[2]]
      [1] "Depends"   "Imports"   "LinkingTo"
      
      $dep_types[[3]]
      [1] "Depends"   "Imports"   "LinkingTo"
      
      $dep_types[[4]]
      [1] "Depends"   "Imports"   "LinkingTo" "Suggests" 
      
      

---

    Code
      as.list(res[c("ref", "dep_types")])
    Output
      $ref
      [1] "any::pkg3"              "installed::<path>/pkg1" "installed::<path>/pkg2"
      [4] "installed::<path>/pkg3" "pkg1"                   "pkg2"                  
      [7] "pkg3"                  
      
      $dep_types
      $dep_types[[1]]
      [1] "Depends"   "Imports"   "LinkingTo"
      
      $dep_types[[2]]
      [1] "Depends"   "Imports"   "LinkingTo"
      
      $dep_types[[3]]
      [1] "Depends"   "Imports"   "LinkingTo"
      
      $dep_types[[4]]
      [1] "Depends"   "Imports"   "LinkingTo" "Suggests" 
      
      $dep_types[[5]]
      [1] "Depends"   "Imports"   "LinkingTo"
      
      $dep_types[[6]]
      [1] "Depends"   "Imports"   "LinkingTo"
      
      $dep_types[[7]]
      [1] "Depends"   "Imports"   "LinkingTo" "Suggests" 
      
      

# failure in non-needed package is ignored

    Code
      p$draw()
    Output
      needspak 1.0.0 [new][bld][dl] (<size>)
      \-r-lib/pak 1.0.0 [new][bld][cmp][dl] (unknown size)
      
      r-lib/pak 1.0.0 [new][bld][cmp][dl] (unknown size)
      
      Key:  [new] new | [dl] download | [bld] build | [cmp] compile

# q_library_0

    Code
      do("library(pkg, lib.loc = path)")
    Output
      # A data frame: 2 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name     code                        
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr>    <chr>                       
      1     2       1     1          1       28         1            1       1         29 dep-code library(pkg, lib.loc = path)
      2     1       1     1          1        7         1            1       1          8 fn-name  library                     
    Code
      do("library('pkg', lib.loc = path)")
    Output
      # A data frame: 2 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name     code                          
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr>    <chr>                         
      1     2       1     1          1       30         1            1       1         31 dep-code library('pkg', lib.loc = path)
      2     1       1     1          1        7         1            1       1          8 fn-name  library                       
    Code
      do("library(lib.loc = path, pkg)")
    Output
      # A data frame: 2 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name     code                        
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr>    <chr>                       
      1     2       1     1          1       28         1            1       1         29 dep-code library(lib.loc = path, pkg)
      2     1       1     1          1        7         1            1       1          8 fn-name  library                     
    Code
      do("require(lib.loc = path, character.only = TRUE, 'pkg')")
    Output
      # A data frame: 2 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name     code                                                 
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr>    <chr>                                                
      1     2       1     1          1       53         1            1       1         54 dep-code require(lib.loc = path, character.only = TRUE, 'pkg')
      2     1       1     1          1        7         1            1       1          8 fn-name  require                                              
    Code
      do("library(foo, require(bar))")
    Output
      # A data frame: 4 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name     code                      
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr>    <chr>                     
      1     2       1     1          1       26         1            1       1         27 dep-code library(foo, require(bar))
      2     1       1     1          1        7         1            1       1          8 fn-name  library                   
      3     2       1     2         14       25         1           14       1         26 dep-code require(bar)              
      4     1       1     2         14       20         1           14       1         21 fn-name  require                   

# q_module_import

    Code
      do("fixtures/scan/modules.R")
    Output
      # A data frame: 18 x 11
            id pattern match start_byte end_byte start_row start_column end_row end_column name     code                      
         <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr>    <chr>                     
       1     2       1     1         12       22         2            3       2         14 dep-code "import(\"A\")"           
       2     1       1     1         12       17         2            3       2          9 fn-name  "import"                  
       3     2       1     2         26       34         3            3       3         12 dep-code "import(B)"               
       4     1       1     2         26       31         3            3       3          9 fn-name  "import"                  
       5     2       1     3         38       55         4            3       4         21 dep-code "import(from = \"C\")"    
       6     1       1     3         38       43         4            3       4          9 fn-name  "import"                  
       7     2       1     4         59       82         5            3       5         27 dep-code "import(symbol, from = D)"
       8     1       1     4         59       64         5            3       5          9 fn-name  "import"                  
       9     2       1     5        167      177        10            1      10         12 dep-code "import(\"e\")"           
      10     1       1     5        167      172        10            1      10          7 fn-name  "import"                  
      11     2       1     6        179      187        11            1      11         10 dep-code "import(f)"               
      12     1       1     6        179      184        11            1      11          7 fn-name  "import"                  
      13     2       2     7        267      286        15            1      15         21 dep-code "modules::import(\"G\")"  
      14     3       2     7        267      273        15            1      15          8 ns-name  "modules"                 
      15     1       2     7        276      281        15           10      15         16 fn-name  "import"                  
      16     2       2     8        288      305        16            1      16         19 dep-code "modules::import(H)"      
      17     3       2     8        288      294        16            1      16          8 ns-name  "modules"                 
      18     1       2     8        297      302        16           10      16         16 fn-name  "import"                  
    Code
      do("fixtures/scan/modules-empty.R")
    Output
      # A data frame: 0 x 11
      # i 11 variables: id <int>, pattern <int>, match <int>, start_byte <int>, end_byte <int>, start_row <int>, start_column <int>, end_row <int>, end_column <int>, name <chr>, code <chr>

# q_colon

    Code
      do("x <- foo::bar()")
    Output
      # A data frame: 2 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name     code    
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr>    <chr>   
      1     2       1     1          6       13         1            6       1         14 dep-code foo::bar
      2     1       1     1          6        8         1            6       1          9 pkg-name foo     
    Code
      do("1 + 2 + foo:::bar")
    Output
      # A data frame: 2 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name     code     
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr>    <chr>    
      1     2       1     1          9       17         1            9       1         18 dep-code foo:::bar
      2     1       1     1          9       11         1            9       1         12 pkg-name foo      

# q_methods

    Code
      do("setClass('myclass')")
    Output
      # A data frame: 2 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name     code               
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr>    <chr>              
      1     2       1     1          1       19         1            1       1         20 dep-code setClass('myclass')
      2     1       1     1          1        8         1            1       1          9 fn-name  setClass           
    Code
      do("setGeneric('props', function(object) attributes(object))")
    Output
      # A data frame: 2 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name     code                                                    
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr>    <chr>                                                   
      1     2       1     1          1       56         1            1       1         57 dep-code setGeneric('props', function(object) attributes(object))
      2     1       1     1          1       10         1            1       1         11 fn-name  setGeneric                                              

# q_junit_reporter

    Code
      do("JunitReporter$new()")
    Output
      # A data frame: 3 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name        code               
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr>       <chr>              
      1     3       1     1          1       19         1            1       1         20 dep-code    JunitReporter$new()
      2     1       1     1          1       13         1            1       1         14 class-name  JunitReporter      
      3     2       1     1         15       17         1           15       1         18 method-name new                
    Code
      do("testthat::JunitReporter$new()")
    Output
      # A data frame: 4 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name        code                         
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr>       <chr>                        
      1     3       2     1          1       29         1            1       1         30 dep-code    testthat::JunitReporter$new()
      2     4       2     1          1        8         1            1       1          9 pkg-name    testthat                     
      3     1       2     1         11       23         1           11       1         24 class-name  JunitReporter                
      4     2       2     1         25       27         1           25       1         28 method-name new                          

# q_knitr_dev

    Code
      do("opts_chunk$set()")
    Output
      # A data frame: 3 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name        code            
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr>       <chr>           
      1     3       1     1          1       16         1            1       1         17 dep-code    opts_chunk$set()
      2     1       1     1          1       10         1            1       1         11 object-name opts_chunk      
      3     2       1     1         12       14         1           12       1         15 method-name set             
    Code
      do("knitr::opts_chunk$set()")
    Output
      # A data frame: 4 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name        code                   
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr>       <chr>                  
      1     3       2     1          1       23         1            1       1         24 dep-code    knitr::opts_chunk$set()
      2     4       2     1          1        5         1            1       1          6 pkg-name    knitr                  
      3     1       2     1          8       17         1            8       1         18 object-name opts_chunk             
      4     2       2     1         19       21         1           19       1         22 method-name set                    

# renv_dependencies_database

    Code
      renv_dependencies_database()
    Output
      $ggplot2
      $ggplot2$geom_hex
      [1] "hexbin"
      
      
      $testthat
      $testthat$JunitReporter
      [1] "xml2"
      
      

# q_database

    Code
      do("geom_hex()")
    Output
      # A data frame: 1 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name  code    
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr> <chr>   
      1     1       1     1          1        8         1            1       1          9 id    geom_hex
    Code
      do("ggplot2::geom_hex()")
    Output
      # A data frame: 1 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name  code    
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr> <chr>   
      1     1       1     1         10       17         1           10       1         18 id    geom_hex
    Code
      do("JunitReporter")
    Output
      # A data frame: 1 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name  code         
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr> <chr>        
      1     1       1     1          1       13         1            1       1         14 id    JunitReporter
    Code
      do("testthat::JunitReporter")
    Output
      # A data frame: 1 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name  code         
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr> <chr>        
      1     1       1     1         11       23         1           11       1         24 id    JunitReporter

# q_database #3

    Code
      do("geom_hex()")
    Output
      # A data frame: 1 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name  code    
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr> <chr>   
      1     1       1     1          1        8         1            1       1          9 id    geom_hex
    Code
      do("foopkg::foofun()")
    Output
      # A data frame: 1 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name  code  
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr> <chr> 
      1     1       1     1          9       14         1            9       1         15 id    foofun
    Code
      do("foofun")
    Output
      # A data frame: 1 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name  code  
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr> <chr> 
      1     1       1     1          1        6         1            1       1          7 id    foofun

# q_deps

    Code
      q_deps()
    Output
      [1] 1 2 3 4 5 6

# q_deps_rmd

    Code
      code_query(readLines(test_path("fixtures/scan/chunk-errors.Rmd")), query = q_deps_rmd(), language = "markdown")[["matched_captures"]]
    Output
      # A data frame: 14 x 11
            id pattern match start_byte end_byte start_row start_column end_row end_column name     code                                                                                      
         <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr>    <chr>                                                                                     
       1     4       2     1          5       16         2            4       2         16 inline   "Parse Errors"                                                                            
       2     4       2     2         19      105         4            1       5         14 inline   "renv should still be able to recover dependencies from the chunks without\nparse errors."
       3     2       1     3        111      113         7            4       7          7 header   "{r}"                                                                                     
       4     1       1     3        112      112         7            5       7          6 language "r"                                                                                       
       5     3       1     3        115      129         8            1       9          1 content  "library(dplyr)\n"                                                                        
       6     2       1     4        138      140        11            4      11          7 header   "{r}"                                                                                     
       7     1       1     4        139      139        11            5      11          6 language "r"                                                                                       
       8     3       1     4        142      169        12            1      13          1 content  "this chunk has parse errors\n"                                                           
       9     2       1     5        178      180        15            4      15          7 header   "{r}"                                                                                     
      10     1       1     5        179      179        15            5      15          6 language "r"                                                                                       
      11     3       1     5        182      204        16            1      17          1 content  "and so does this chunk\n"                                                                
      12     2       1     6        213      215        19            4      19          7 header   "{r}"                                                                                     
      13     1       1     6        214      214        19            5      19          6 language "r"                                                                                       
      14     3       1     6        217      245        20            1      20         30 content  "we forgot to close this chunk"                                                           

---

    Code
      code_query(readLines(test_path("fixtures/scan/inline-chunks.Rmd")), query = q_deps_rmd(), language = "markdown")[["matched_captures"]]
    Output
      # A data frame: 3 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name   code                                                                                                                           
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr>  <chr>                                                                                                                          
      1     4       2     1          5       17         2            4       2         17 inline Inline Chunks                                                                                                                  
      2     4       2     2         20      146         4            1       4        128 inline Users might request the use of packages with `r inline::chunks()`. Check that we handle `r multiple::calls()` on the same line.
      3     4       2     3        149      179         6            1       6         32 inline Also in `r separate::chunks()`.                                                                                                

# q_deps_rmd_inline

    Code
      code_query(readLines(test_path("fixtures/scan/inline-chunks.Rmd")), query = q_deps_rmd_inline(), language = "markdown-inline", ranges = code[, range_cols])[["matched_captures"]]
    Output
      # A data frame: 9 x 11
           id pattern match start_byte end_byte start_row start_column end_row end_column name  code                  
        <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr> <chr>                 
      1     3       1     1         65       84         4           46       4         66 code  `r inline::chunks()`  
      2     1       1     1         65       65         4           46       4         47 csd1  `                     
      3     2       1     1         84       84         4           65       4         66 csd2  `                     
      4     3       1     2        108      128         4           89       4        110 code  `r multiple::calls()` 
      5     1       1     2        108      108         4           89       4         90 csd1  `                     
      6     2       1     2        128      128         4          109       4        110 csd2  `                     
      7     3       1     3        157      178         6            9       6         31 code  `r separate::chunks()`
      8     1       1     3        157      157         6            9       6         10 csd1  `                     
      9     2       1     3        178      178         6           30       6         31 csd2  `                     

# q_deps_yaml_header

    Code
      print(n = Inf, code_query(readLines(test_path("fixtures/scan/header-shiny.Rmd")), query = q_deps_yaml_header(), language = "yaml")[["matched_captures"]])
    Output
      # A data frame: 12 x 11
            id pattern match start_byte end_byte start_row start_column end_row end_column name  code                   
         <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr> <chr>                  
       1     3       1     1         16       30         4            1       4         16 code  "'server': shiny"      
       2     1       1     1         16       23         4            1       4          9 key   "'server'"             
       3     2       1     1         26       30         4           11       4         16 value "shiny"                
       4     3       1     2         32       46         5            1       5         16 code  "server: 'shiny'"      
       5     1       1     2         32       37         5            1       5          7 key   "server"               
       6     2       1     2         40       46         5            9       5         16 value "'shiny'"              
       7     3       1     3         48       64         6            1       6         18 code  "\"server\": \"shiny\""
       8     1       1     3         48       55         6            1       6          9 key   "\"server\""           
       9     2       1     3         58       64         6           11       6         18 value "\"shiny\""            
      10     3       1     4         66       82         7            1       8          8 code  "server: |\n  shiny"   
      11     1       1     4         66       71         7            1       7          7 key   "server"               
      12     2       1     4         74       82         7            9       8          8 value "|\n  shiny"           
    Code
      print(n = Inf, code_query(readLines(test_path("fixtures/scan/header-shiny2.Rmd")), query = q_deps_yaml_header(), language = "yaml")[["matched_captures"]])
    Output
      # A data frame: 16 x 11
            id pattern match start_byte end_byte start_row start_column end_row end_column name  code                           
         <int>   <int> <int>      <int>    <int>     <int>        <int>   <int>      <int> <chr> <chr>                          
       1     3       2     1         16       36         4            1       5         14 code  "server:\n  type: shiny"       
       2     1       2     1         16       21         4            1       4          7 key   "server"                       
       3     4       2     1         26       29         5            3       5          7 key2  "type"                         
       4     2       2     1         32       36         5            9       5         14 value "shiny"                        
       5     3       2     2         38       62         6            1       7         16 code  "'server':\n  type: 'shiny'"   
       6     1       2     2         38       45         6            1       6          9 key   "'server'"                     
       7     4       2     2         50       53         7            3       7          7 key2  "type"                         
       8     2       2     2         56       62         7            9       7         16 value "'shiny'"                      
       9     3       2     3         64       88         8            1       9         18 code  "server:\n  'type': \"shiny\"" 
      10     1       2     3         64       69         8            1       8          7 key   "server"                       
      11     4       2     3         74       79         9            3       9          9 key2  "'type'"                       
      12     2       2     3         82       88         9           11       9         18 value "\"shiny\""                    
      13     3       2     4         90      116        10            1      12         10 code  "server:\n  type: >\n    shiny"
      14     1       2     4         90       95        10            1      10          7 key   "server"                       
      15     4       2     4        100      103        11            3      11          7 key2  "type"                         
      16     2       2     4        106      116        11            9      12         10 value ">\n    shiny"                 

# install_packages / works with source packages

    Code
      plan <- make_install_plan(paste0("local::", pkg, "?nocache"), lib = libpath)
    Message
      i Getting 1 pkg with unknown size
      v Got foo 0.0.0.9000 (source)
    Code
      install_package_plan(plan, lib = libpath)
    Message
      i Building foo 0.0.0.9000
      v Built foo 0.0.0.9000
      v Installed foo 0.0.0.9000 (local)
      v Summary:   1 new

# new_pkg_installation_plan

    Code
      plan
    Output
      <pkg_installation_plan>
      + refs:
        - pkg3
      + has solution
      (use `$update()` to update the plan for an updated library)
      (use `$show_solution()` to see the packages to install
      (use `$get_solution()` to see the full solution results)
      (use `$draw()` to draw the dependency tree)
      (use `$download()` to download packages)

---

    Code
      plan$resolve()
    Condition
      Error:
      ! ! Cannot resolve an installation plan, it is already resolved.
    Code
      plan$async_resolve()
    Condition
      Error:
      ! ! Cannot resolve an installation plan, it is already resolved.

---

    Code
      plan$set_solve_policy()
    Condition
      Error:
      ! ! Cannot solve an installation plan, it is already solved.
    Code
      plan$solve()
    Condition
      Error:
      ! ! Cannot solve an installation plan, it is already solved.

# sysreqs

    Code
      plan
    Output
      <pkg_installation_plan>
      + refs:
        - pkg3
      + has solution
      (use `$update()` to update the plan for an updated library)
      (use `$show_solution()` to see the packages to install
      (use `$get_solution()` to see the full solution results)
      (use `$draw()` to draw the dependency tree)
      (use `$download()` to download packages)

---

    Code
      plan$resolve()
    Condition
      Error:
      ! ! Cannot resolve an installation plan, it is already resolved.
    Code
      plan$async_resolve()
    Condition
      Error:
      ! ! Cannot resolve an installation plan, it is already resolved.

---

    Code
      plan$set_solve_policy()
    Condition
      Error:
      ! ! Cannot solve an installation plan, it is already solved.
    Code
      plan$solve()
    Condition
      Error:
      ! ! Cannot solve an installation plan, it is already solved.

# install_sysreqs

    Code
      plan$show_solution()
    Message
      + curl   1.0.0 [bld][dl] (<size>) + libcurl4-openssl-dev, libssl-dev
    Code
      plan$show_sysreqs()
    Message
      * libcurl4-openssl-dev  - curl
      * libssl-dev            - curl

# update_sysreqs

    Code
      plan$show_sysreqs()
    Message
      + libcurl4-openssl-dev  - curl

# update_sysreqs with old lock file

    Code
      plan$show_sysreqs()
    Message
        

# failed resolution

    Code
      as.list(res)
    Output
      $ref
      [1] "bioc::nosuchpackageinbiocforsure"
      
      $type
      [1] "bioc"
      
      $direct
      [1] TRUE
      
      $directpkg
      [1] FALSE
      
      $status
      [1] "FAILED"
      
      $package
      [1] "nosuchpackageinbiocforsure"
      
      $version
      [1] NA
      
      $license
      [1] NA
      
      $needscompilation
      [1] TRUE
      
      $priority
      [1] NA
      
      $md5sum
      [1] NA
      
      $sha256
      [1] NA
      
      $filesize
      [1] NA
      
      $built
      [1] NA
      
      $platform
      [1] "source"
      
      $rversion
      [1] "*"
      
      $repotype
      [1] NA
      
      $repodir
      [1] "src/contrib"
      
      $target
      [1] "src/contrib/nosuchpackageinbiocforsure_NA.tar.gz"
      
      $deps
      $deps[[1]]
      # A data frame: 0 x 5
      # i 5 variables: ref <chr>, type <chr>, package <chr>, op <chr>, version <chr>
      
      
      $mirror
      [1] NA
      
      $sources
      $sources[[1]]
      [1] NA
      
      
      $remote
      $remote[[1]]
      $package
      [1] "nosuchpackageinbiocforsure"
      
      $atleast
      [1] ""
      
      $version
      [1] ""
      
      $ref
      [1] "bioc::nosuchpackageinbiocforsure"
      
      $type
      [1] "bioc"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_bioc" "remote_ref"      "list"           
      
      
      $error
      $error[[1]]
      <error: Can't find package called bioc::nosuchpackageinbiocforsure.>
      
      
      $metadata
      $metadata[[1]]
      list()
      
      
      $extra
      $extra[[1]]
      list()
      
      
      $params
      $params[[1]]
      character(0)
      
      
      $sysreqs
      [1] NA
      
      $os_type
      [1] NA
      

# resolve

    Code
      res$error
    Output
      [[1]]
      list()
      
    Code
      res$package
    Output
      [1] "cli"
    Code
      res$version
    Output
      [1] "3.5.0.9000"
    Code
      res$metadata[[1]]
    Output
                                      RemoteType 
                                        "gitlab" 
                                       RemoteUrl 
        "https://gitlab.com/gaborcsardi/cli.git" 
                                    RemotePkgRef 
                       "gitlab::gaborcsardi/cli" 
                                       RemoteRef 
                                          "HEAD" 
                                       RemoteSha 
      "bc503509cddb65d0007aeb301936e27de76bc7d7" 
                                      RemoteHost 
                                    "gitlab.com" 
                                      RemoteRepo 
                                           "cli" 
                                  RemoteUsername 
                                   "gaborcsardi" 

---

    Code
      res$error
    Output
      [[1]]
      list()
      
    Code
      res$package
    Output
      [1] "filelock"
    Code
      res$version
    Output
      [1] "1.0.2"
    Code
      res$metadata[[1]]
    Output
                                      RemoteType 
                                        "gitlab" 
                                       RemoteUrl 
         "https://gitlab.com/r-hub/filelock.git" 
                                    RemotePkgRef 
             "gitlab::r-hub/filelock@cran-1-0-2" 
                                       RemoteRef 
                                    "cran-1-0-2" 
                                       RemoteSha 
      "dfc615b14dfc030df3f44bd377e752728f76df3d" 
                                      RemoteHost 
                                    "gitlab.com" 
                                      RemoteRepo 
                                      "filelock" 
                                  RemoteUsername 
                                         "r-hub" 

---

    Code
      res$error
    Output
      [[1]]
      list()
      
    Code
      res$package
    Output
      [1] "filelock"
    Code
      res$version
    Output
      [1] "1.0.2"
    Code
      res$metadata[[1]]
    Output
                                      RemoteType 
                                        "gitlab" 
                                       RemoteUrl 
         "https://gitlab.com/r-hub/filelock.git" 
                                    RemotePkgRef 
                 "gitlab::r-hub/filelock@v1.0.2" 
                                       RemoteRef 
                                        "v1.0.2" 
                                       RemoteSha 
      "9fdba75a62facaa3e818902f58891166e45eabe9" 
                                      RemoteHost 
                                    "gitlab.com" 
                                      RemoteRepo 
                                      "filelock" 
                                  RemoteUsername 
                                         "r-hub" 

---

    Code
      res$error
    Output
      [[1]]
      list()
      
    Code
      res$package
    Output
      [1] "feather"
    Code
      res$version
    Output
      [1] "0.3.5.9000"
    Code
      res$metadata[[1]]
    Output
                                        RemoteType 
                                          "gitlab" 
                                         RemoteUrl 
      "https://gitlab.com/gaborcsardi/feather.git" 
                                      RemotePkgRef 
                 "gitlab::gaborcsardi/feather/-/R" 
                                         RemoteRef 
                                            "HEAD" 
                                         RemoteSha 
        "3635466429b5ef7910da606267d0a339cf1c9821" 
                                        RemoteHost 
                                      "gitlab.com" 
                                        RemoteRepo 
                                         "feather" 
                                    RemoteUsername 
                                     "gaborcsardi" 
                                      RemoteSubdir 
                                               "R" 

# download

    Code
      dir(file.path(dl$fulltarget_tree, "feather", "R"))
    Output
       [1] "DESCRIPTION"      "NAMESPACE"        "NEWS.md"          "R"               
       [5] "README.md"        "cran-comments.md" "feather.Rproj"    "inst"            
       [9] "man"              "revdep"           "tests"           

---
title: "pkgdepends HOWTO"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{pkgdepends test}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r dev = "svglite"}
print("Hello there!")
```

# scan_deps

    Code
      scan_deps(project, root = project)[]
    Output
      # A data frame: 11 x 9
         path            ref       package   version type  code              start_row start_column start_byte
         <chr>           <chr>     <chr>     <chr>   <chr> <chr>                 <int>        <int>      <int>
       1 R/code.R        CD        CD        *       prod  CD::pkg                   4            1         26
       2 R/code.R        AB        AB        *       prod  library(AB)               1            1          1
       3 R/code.R        BC        BC        *       prod  require(BC)               2            1         13
       4 doc.qmd         pkgload   pkgload   *       prod  pkgload::load_all        12            1        174
       5 doc.qmd         rmarkdown rmarkdown *       prod  r                        11            5        140
       6 index.Rmarkdown ST        ST        *       prod  ST::fun                  10            1         97
       7 index.Rmarkdown RS        RS        *       prod  library(RS)               9            1         85
       8 index.Rmarkdown rmarkdown rmarkdown *       prod  r                         8            5         82
       9 index.Rmd       ST        ST        *       prod  ST::fun                  10            1         97
      10 index.Rmd       RS        RS        *       prod  library(RS)               9            1         85
      11 index.Rmd       rmarkdown rmarkdown *       prod  r                         8            5         82

---

    Code
      scan_deps(project, root = project)
    Output
      
      Dependencies:
      + AB        @ R/code.R
      + BC        @ R/code.R
      + CD        @ R/code.R
      + RS        @ index.Rmarkdown, index.Rmd
      + ST        @ index.Rmarkdown, index.Rmd
      + pkgload   @ doc.qmd
      + rmarkdown @ doc.qmd, index.Rmarkdown, index.Rmd

# scan_path_deps

    Code
      scan_path_deps(rfile)
    Output
      # A data frame: 3 x 9
        path                             ref   package version type  code        start_row start_column start_byte
        <chr>                            <chr> <chr>   <chr>   <chr> <chr>           <int>        <int>      <int>
      1 fixtures/scan/project-1/R/code.R CD    CD      *       prod  CD::pkg             4            1         26
      2 fixtures/scan/project-1/R/code.R AB    AB      *       prod  library(AB)         1            1          1
      3 fixtures/scan/project-1/R/code.R BC    BC      *       prod  require(BC)         2            1         13

---

    Code
      scan_path_deps(rfile)
    Output
      # A data frame: 3 x 9
        path                             ref   package version type  code        start_row start_column start_byte
        <chr>                            <chr> <chr>   <chr>   <chr> <chr>           <int>        <int>      <int>
      1 fixtures/scan/project-1/R/code.R CD    CD      *       prod  CD::pkg             4            1         26
      2 fixtures/scan/project-1/R/code.R AB    AB      *       prod  library(AB)         1            1          1
      3 fixtures/scan/project-1/R/code.R BC    BC      *       prod  require(BC)         2            1         13

# binary preferred over source

    Code
      sol[, c("package", "platform")]
    Output
      # A data frame: 2 x 2
        package platform               
      * <chr>   <chr>                  
      1 pkg     i386+x86_64-w64-mingw32
      2 pkg2    i386+x86_64-w64-mingw32

# but source is used if that version is required

    Code
      sol[, c("package", "version", "platform")]
    Output
      # A data frame: 2 x 3
        package version platform               
      * <chr>   <chr>   <chr>                  
      1 pkg     1.0.0   i386+x86_64-w64-mingw32
      2 pkg2    2.0.0   source                 

# can opt in into including LinkingTo

    Code
      sol[, c("package", "platform")]
    Output
      # A data frame: 2 x 2
        package platform               
      * <chr>   <chr>                  
      1 pkg     i386+x86_64-w64-mingw32
      2 pkg2    i386+x86_64-w64-mingw32

---

    Code
      sol[, c("package", "platform")]
    Output
      # A data frame: 1 x 2
        package platform               
      * <chr>   <chr>                  
      1 pkg     i386+x86_64-w64-mingw32

# git_list_refs

    Code
      git_list_refs(fake_git$url("/pak-test.git"))$refs
    Output
      # A data frame: 5 x 2
        ref                     hash                                    
        <chr>                   <chr>                                   
      1 HEAD                    3f3b0b4ee8a0ff4563073924e5fe069da67a6d8b
      2 refs/heads/build-ignore a9ffc55f59e0567ecdc67fb3f0333eca49be8d03
      3 refs/heads/main         3f3b0b4ee8a0ff4563073924e5fe069da67a6d8b
      4 refs/heads/subdir       cefdc0eebcd7f757efb9a80652fd8aaf1a87508e
      5 refs/tags/v1            cefdc0eebcd7f757efb9a80652fd8aaf1a87508e

---

    Code
      git_list_refs_v2(fake_git$url("/pak-test.git"), "refs/heads/")$refs
    Output
      # A data frame: 3 x 2
        ref                     hash                                    
        <chr>                   <chr>                                   
      1 refs/heads/build-ignore a9ffc55f59e0567ecdc67fb3f0333eca49be8d03
      2 refs/heads/main         3f3b0b4ee8a0ff4563073924e5fe069da67a6d8b
      3 refs/heads/subdir       cefdc0eebcd7f757efb9a80652fd8aaf1a87508e

# git_list_files

    Code
      git_list_files(fake_git$url("/pak-test.git"), "foobar")
    Condition
      Error:
      ! ! Unknown git ref: "foobar".

---

    Code
      git_list_files(fake_git$url("/pak-test.git"),
      "cefdc0eebcd7f757efb9a80652fd8aaf1a87508e")
    Output
      $ref
      [1] "cefdc0eebcd7f757efb9a80652fd8aaf1a87508e"
      
      $sha
      [1] "cefdc0eebcd7f757efb9a80652fd8aaf1a87508e"
      
      $commit
                                                          tree 
                    "aac34e05a0cb9852aa425757dc8480fa29c6c783" 
                                                        parent 
                    "cb8e35b36f8096ef604eff80c70259c64b34f125" 
                                                        author 
      "Gábor Csárdi <csardi.gabor@gmail.com> 1639165496 +0100" 
                                                     committer 
      "Gábor Csárdi <csardi.gabor@gmail.com> 1639165496 +0100" 
                                                       message 
                                                "Just bumping" 
      
      $tree
      [1] "aac34e05a0cb9852aa425757dc8480fa29c6c783"
      
      $files
      # A data frame: 24 x 4
         hash                                     type  mode   path                   
         <chr>                                    <chr> <chr>  <chr>                  
       1 61102a589aeef025b8c11015d5de4fd6f91b0fdb tree  40000  .github                
       2 0784f430fc5830e1b26ceb5fa9a22551fbe9a400 tree  40000  .github/actions        
       3 c3117c75084ec688a9bf82ce11789e0295f01ca3 tree  40000  .github/actions/parame~
       4 0acc01668acf00adc8c3921edea2526d70c67882 blob  100644 .github/actions/parame~
       5 23b72b52356f0a1ac89925f5431e44a0ab23bab6 tree  40000  .github/workflows      
       6 cddf411c750f674622bd53fe679a71698dadda4c blob  100644 .github/workflows/chec~
       7 0d9009c1027504fd192ec1270c72581c7fc83825 blob  100644 .gitignore             
       8 e69de29bb2d1d6434b8b29ae775ad8c2e48c5391 blob  100644 foo                    
       9 5ba9d25318eacd311bb9f73cf83f859b0bfb7235 tree  40000  subdir                 
      10 75210921adb3ebdf805d8b01691e2d4b3915d096 tree  40000  subdir/dotenv          
      # i 14 more rows
      
    Code
      git_list_files(fake_git$url("/pak-test.git"), "refs/tags/v1")
    Output
      $ref
      [1] "refs/tags/v1"
      
      $sha
      [1] "cefdc0eebcd7f757efb9a80652fd8aaf1a87508e"
      attr(,"protocol")
      [1] 2
      attr(,"filter")
      [1] TRUE
      
      $commit
                                                          tree 
                    "aac34e05a0cb9852aa425757dc8480fa29c6c783" 
                                                        parent 
                    "cb8e35b36f8096ef604eff80c70259c64b34f125" 
                                                        author 
      "Gábor Csárdi <csardi.gabor@gmail.com> 1639165496 +0100" 
                                                     committer 
      "Gábor Csárdi <csardi.gabor@gmail.com> 1639165496 +0100" 
                                                       message 
                                                "Just bumping" 
      
      $tree
      [1] "aac34e05a0cb9852aa425757dc8480fa29c6c783"
      
      $files
      # A data frame: 24 x 4
         hash                                     type  mode   path                   
         <chr>                                    <chr> <chr>  <chr>                  
       1 61102a589aeef025b8c11015d5de4fd6f91b0fdb tree  40000  .github                
       2 0784f430fc5830e1b26ceb5fa9a22551fbe9a400 tree  40000  .github/actions        
       3 c3117c75084ec688a9bf82ce11789e0295f01ca3 tree  40000  .github/actions/parame~
       4 0acc01668acf00adc8c3921edea2526d70c67882 blob  100644 .github/actions/parame~
       5 23b72b52356f0a1ac89925f5431e44a0ab23bab6 tree  40000  .github/workflows      
       6 cddf411c750f674622bd53fe679a71698dadda4c blob  100644 .github/workflows/chec~
       7 0d9009c1027504fd192ec1270c72581c7fc83825 blob  100644 .gitignore             
       8 e69de29bb2d1d6434b8b29ae775ad8c2e48c5391 blob  100644 foo                    
       9 5ba9d25318eacd311bb9f73cf83f859b0bfb7235 tree  40000  subdir                 
      10 75210921adb3ebdf805d8b01691e2d4b3915d096 tree  40000  subdir/dotenv          
      # i 14 more rows
      

# async_git_list_files_process

    Code
      sort(async_git_list_files_process(pack, ref = ref, sha = ref, url = "url")$
      files$path)
    Output
       [1] ".github"                              
       [2] ".github/actions"                      
       [3] ".github/actions/parameters"           
       [4] ".github/actions/parameters/action.yml"
       [5] ".github/workflows"                    
       [6] ".github/workflows/check-standard.yaml"
       [7] ".gitignore"                           
       [8] "foo"                                  
       [9] "subdir"                               
      [10] "subdir/dotenv"                        
      [11] "subdir/dotenv/.Rbuildignore"          
      [12] "subdir/dotenv/.gitignore"             
      [13] "subdir/dotenv/DESCRIPTION"            
      [14] "subdir/dotenv/LICENSE"                
      [15] "subdir/dotenv/NAMESPACE"              
      [16] "subdir/dotenv/NEWS.md"                
      [17] "subdir/dotenv/R"                      
      [18] "subdir/dotenv/R/dotenv-package.r"     
      [19] "subdir/dotenv/README.Rmd"             
      [20] "subdir/dotenv/README.md"              
      [21] "subdir/dotenv/man"                    
      [22] "subdir/dotenv/man/dotenv-package.Rd"  
      [23] "subdir/dotenv/man/load_dot_env.Rd"    
      [24] "wipe.R"                               

# git_download_file

    Code
      out <- git_download_file(fake_git$url("/pak-test.git"),
      "a1e2d6741374d1f32ec138ee2020eae36b859e99", tmp)
      out
    Output
      $type
      [1] "blob"
      
      $raw
        [1] 2e 64 79 6e 4c 69 62 73 28 29 0a 70 6b 67 6c 6f 61 64 3a 3a 6c 6f 61 64 5f
       [26] 61 6c 6c 28 29 0a 2e 64 79 6e 4c 69 62 73 28 29 0a 67 63 28 29 0a 2e 64 79
       [51] 6e 4c 69 62 73 28 29 0a 70 6b 67 6c 6f 61 64 3a 3a 6c 6f 61 64 5f 61 6c 6c
       [76] 28 29 0a 2e 64 79 6e 4c 69 62 73 28 29 0a 67 63 28 29 0a 2e 64 79 6e 4c 69
      [101] 62 73 28 29 0a
      
      $size
      [1] 105
      
      $packed_size
      [1] 44
      
      $hash
      [1] "a1e2d6741374d1f32ec138ee2020eae36b859e99"
      
    Code
      readLines(tmp)
    Output
      [1] ".dynLibs()"          "pkgload::load_all()" ".dynLibs()"         
      [4] "gc()"                ".dynLibs()"          "pkgload::load_all()"
      [7] ".dynLibs()"          "gc()"                ".dynLibs()"         

# git_fetch

    Code
      cat(pack[[1]]$object)
    Output
      tree aac34e05a0cb9852aa425757dc8480fa29c6c783
      parent cb8e35b36f8096ef604eff80c70259c64b34f125
      author Gábor Csárdi <csardi.gabor@gmail.com> 1639165496 +0100
      committer Gábor Csárdi <csardi.gabor@gmail.com> 1639165496 +0100
      
      Just bumping

# git_parse_message errors

    Code
      git_parse_message(raw(3))
    Condition
      Error:
      ! ! Invalid pkt-line at the enf of message from git.

---

    Code
      git_parse_message(charToRaw("foobvar"))
    Condition
      Error:
      ! ! Invalid pkt-len field in message from git, must be four hexa digits.

---

    Code
      git_parse_message(charToRaw("00bbnoteonugh"))
    Condition
      Error:
      ! ! Invalid pkt-payload in message from git.
      i Need 187 bytes, found 13.

# git_create_message_v1

    Code
      git_create_message_v1(character())
    Condition
      Error:
      ! ! Invalid git protocol (v1) message, must have at least one argument

# pkt_line

    Code
      pkt_line(raw(70000))
    Condition
      Error:
      ! ! packet line longer than 65516 bytes is not implemented yet.

# git_list_refs_v1

    Code
      git_list_refs_v1(fake_git$url("/pak-test.git"))$refs
    Output
      # A data frame: 5 x 2
        ref                     hash                                    
        <chr>                   <chr>                                   
      1 HEAD                    3f3b0b4ee8a0ff4563073924e5fe069da67a6d8b
      2 refs/heads/build-ignore a9ffc55f59e0567ecdc67fb3f0333eca49be8d03
      3 refs/heads/main         3f3b0b4ee8a0ff4563073924e5fe069da67a6d8b
      4 refs/heads/subdir       cefdc0eebcd7f757efb9a80652fd8aaf1a87508e
      5 refs/tags/v1            cefdc0eebcd7f757efb9a80652fd8aaf1a87508e

# git_list_refs_v1_process_1

    Code
      git_list_refs_v1_process_1(resp, fake_git$url("/pak-test.git"), "refs/tags/v1")$
        refs
    Output
      # A data frame: 1 x 2
        ref          hash                                    
      * <chr>        <chr>                                   
      1 refs/tags/v1 cefdc0eebcd7f757efb9a80652fd8aaf1a87508e

# async_git_list_refs_v2_process_1

    Code
      sy(async_git_list_refs_v2_process_1(resp, fake_git$url("/pak-test.git"),
      "refs/tags/v1"))$refs
    Output
      # A data frame: 1 x 2
        ref          hash                                    
      * <chr>        <chr>                                   
      1 refs/tags/v1 cefdc0eebcd7f757efb9a80652fd8aaf1a87508e

# async_git_list_refs_v2_process_2

    Code
      async_git_list_refs_v2_process_2(NULL, psd2, url, NULL)
    Condition
      Error:
      ! ! Invalid git protocol message from <http://localhost:3000/git/cli>.

---

    Code
      async_git_list_refs_v2_process_2(NULL, psd2, url, NULL)
    Condition
      Error:
      ! ! Only git protocol version 2 is supported, not version 10.

---

    Code
      async_git_list_refs_v2_process_2(NULL, psd2, url, NULL)
    Condition
      Error:
      ! ! Response from git server does not have a closing `flush-pkt`.

---

    Code
      async_git_list_refs_v2_process_3(list(list(type = "data-pkt")), character(),
      url)
    Condition
      Error:
      ! ! Response from git server does not have a closing `flush-pkt`.

# check_initial_response

    Code
      check_initial_response(list(), "http://localhost:3000/git/cli")
    Condition
      Error:
      ! ! Unexpected response from git server, no `data-pkt` line.

# git_unpack

    Code
      git_unpack(path)
    Output
      $cefdc0eebcd7f757efb9a80652fd8aaf1a87508e
      $cefdc0eebcd7f757efb9a80652fd8aaf1a87508e$type
      [1] "commit"
      
      $cefdc0eebcd7f757efb9a80652fd8aaf1a87508e$raw
        [1] 74 72 65 65 20 61 61 63 33 34 65 30 35 61 30 63 62 39 38 35 32 61 61 34 32
       [26] 35 37 35 37 64 63 38 34 38 30 66 61 32 39 63 36 63 37 38 33 0a 70 61 72 65
       [51] 6e 74 20 63 62 38 65 33 35 62 33 36 66 38 30 39 36 65 66 36 30 34 65 66 66
       [76] 38 30 63 37 30 32 35 39 63 36 34 62 33 34 66 31 32 35 0a 61 75 74 68 6f 72
      [101] 20 47 c3 a1 62 6f 72 20 43 73 c3 a1 72 64 69 20 3c 63 73 61 72 64 69 2e 67
      [126] 61 62 6f 72 40 67 6d 61 69 6c 2e 63 6f 6d 3e 20 31 36 33 39 31 36 35 34 39
      [151] 36 20 2b 30 31 30 30 0a 63 6f 6d 6d 69 74 74 65 72 20 47 c3 a1 62 6f 72 20
      [176] 43 73 c3 a1 72 64 69 20 3c 63 73 61 72 64 69 2e 67 61 62 6f 72 40 67 6d 61
      [201] 69 6c 2e 63 6f 6d 3e 20 31 36 33 39 31 36 35 34 39 36 20 2b 30 31 30 30 0a
      [226] 0a 4a 75 73 74 20 62 75 6d 70 69 6e 67 0a
      
      $cefdc0eebcd7f757efb9a80652fd8aaf1a87508e$size
      [1] 239
      
      $cefdc0eebcd7f757efb9a80652fd8aaf1a87508e$packed_size
      [1] 161
      
      $cefdc0eebcd7f757efb9a80652fd8aaf1a87508e$object
      [1] "tree aac34e05a0cb9852aa425757dc8480fa29c6c783\nparent cb8e35b36f8096ef604eff80c70259c64b34f125\nauthor Gábor Csárdi <csardi.gabor@gmail.com> 1639165496 +0100\ncommitter Gábor Csárdi <csardi.gabor@gmail.com> 1639165496 +0100\n\nJust bumping\n"
      
      $cefdc0eebcd7f757efb9a80652fd8aaf1a87508e$hash
      [1] "cefdc0eebcd7f757efb9a80652fd8aaf1a87508e"
      
      
      $aac34e05a0cb9852aa425757dc8480fa29c6c783
      $aac34e05a0cb9852aa425757dc8480fa29c6c783$type
      [1] "tree"
      
      $aac34e05a0cb9852aa425757dc8480fa29c6c783$raw
        [1] 34 30 30 30 30 20 2e 67 69 74 68 75 62 00 61 10 2a 58 9a ee f0 25 b8 c1 10
       [26] 15 d5 de 4f d6 f9 1b 0f db 31 30 30 36 34 34 20 2e 67 69 74 69 67 6e 6f 72
       [51] 65 00 0d 90 09 c1 02 75 04 fd 19 2e c1 27 0c 72 58 1c 7f c8 38 25 31 30 30
       [76] 36 34 34 20 66 6f 6f 00 e6 9d e2 9b b2 d1 d6 43 4b 8b 29 ae 77 5a d8 c2 e4
      [101] 8c 53 91 34 30 30 30 30 20 73 75 62 64 69 72 00 5b a9 d2 53 18 ea cd 31 1b
      [126] b9 f7 3c f8 3f 85 9b 0b fb 72 35 31 30 30 36 34 34 20 77 69 70 65 2e 52 00
      [151] a1 e2 d6 74 13 74 d1 f3 2e c1 38 ee 20 20 ea e3 6b 85 9e 99
      
      $aac34e05a0cb9852aa425757dc8480fa29c6c783$size
      [1] 170
      
      $aac34e05a0cb9852aa425757dc8480fa29c6c783$packed_size
      [1] 166
      
      $aac34e05a0cb9852aa425757dc8480fa29c6c783$object
      # A data frame: 5 x 4
        type  mode   path       hash                                    
        <chr> <chr>  <chr>      <chr>                                   
      1 tree  40000  .github    61102a589aeef025b8c11015d5de4fd6f91b0fdb
      2 blob  100644 .gitignore 0d9009c1027504fd192ec1270c72581c7fc83825
      3 blob  100644 foo        e69de29bb2d1d6434b8b29ae775ad8c2e48c5391
      4 tree  40000  subdir     5ba9d25318eacd311bb9f73cf83f859b0bfb7235
      5 blob  100644 wipe.R     a1e2d6741374d1f32ec138ee2020eae36b859e99
      
      $aac34e05a0cb9852aa425757dc8480fa29c6c783$hash
      [1] "aac34e05a0cb9852aa425757dc8480fa29c6c783"
      
      
      $`61102a589aeef025b8c11015d5de4fd6f91b0fdb`
      $`61102a589aeef025b8c11015d5de4fd6f91b0fdb`$type
      [1] "tree"
      
      $`61102a589aeef025b8c11015d5de4fd6f91b0fdb`$raw
       [1] 34 30 30 30 30 20 61 63 74 69 6f 6e 73 00 07 84 f4 30 fc 58 30 e1 b2 6c eb
      [26] 5f a9 a2 25 51 fb e9 a4 00 34 30 30 30 30 20 77 6f 72 6b 66 6c 6f 77 73 00
      [51] 23 b7 2b 52 35 6f 0a 1a c8 99 25 f5 43 1e 44 a0 ab 23 ba b6
      
      $`61102a589aeef025b8c11015d5de4fd6f91b0fdb`$size
      [1] 70
      
      $`61102a589aeef025b8c11015d5de4fd6f91b0fdb`$packed_size
      [1] 75
      
      $`61102a589aeef025b8c11015d5de4fd6f91b0fdb`$object
      # A data frame: 2 x 4
        type  mode  path      hash                                    
        <chr> <chr> <chr>     <chr>                                   
      1 tree  40000 actions   0784f430fc5830e1b26ceb5fa9a22551fbe9a400
      2 tree  40000 workflows 23b72b52356f0a1ac89925f5431e44a0ab23bab6
      
      $`61102a589aeef025b8c11015d5de4fd6f91b0fdb`$hash
      [1] "61102a589aeef025b8c11015d5de4fd6f91b0fdb"
      
      
      $`0784f430fc5830e1b26ceb5fa9a22551fbe9a400`
      $`0784f430fc5830e1b26ceb5fa9a22551fbe9a400`$type
      [1] "tree"
      
      $`0784f430fc5830e1b26ceb5fa9a22551fbe9a400`$raw
       [1] 34 30 30 30 30 20 70 61 72 61 6d 65 74 65 72 73 00 c3 11 7c 75 08 4e c6 88
      [26] a9 bf 82 ce 11 78 9e 02 95 f0 1c a3
      
      $`0784f430fc5830e1b26ceb5fa9a22551fbe9a400`$size
      [1] 37
      
      $`0784f430fc5830e1b26ceb5fa9a22551fbe9a400`$packed_size
      [1] 44
      
      $`0784f430fc5830e1b26ceb5fa9a22551fbe9a400`$object
      # A data frame: 1 x 4
        type  mode  path       hash                                    
        <chr> <chr> <chr>      <chr>                                   
      1 tree  40000 parameters c3117c75084ec688a9bf82ce11789e0295f01ca3
      
      $`0784f430fc5830e1b26ceb5fa9a22551fbe9a400`$hash
      [1] "0784f430fc5830e1b26ceb5fa9a22551fbe9a400"
      
      
      $c3117c75084ec688a9bf82ce11789e0295f01ca3
      $c3117c75084ec688a9bf82ce11789e0295f01ca3$type
      [1] "tree"
      
      $c3117c75084ec688a9bf82ce11789e0295f01ca3$raw
       [1] 31 30 30 36 34 34 20 61 63 74 69 6f 6e 2e 79 6d 6c 00 0a cc 01 66 8a cf 00
      [26] ad c8 c3 92 1e de a2 52 6d 70 c6 78 82
      
      $c3117c75084ec688a9bf82ce11789e0295f01ca3$size
      [1] 38
      
      $c3117c75084ec688a9bf82ce11789e0295f01ca3$packed_size
      [1] 47
      
      $c3117c75084ec688a9bf82ce11789e0295f01ca3$object
      # A data frame: 1 x 4
        type  mode   path       hash                                    
        <chr> <chr>  <chr>      <chr>                                   
      1 blob  100644 action.yml 0acc01668acf00adc8c3921edea2526d70c67882
      
      $c3117c75084ec688a9bf82ce11789e0295f01ca3$hash
      [1] "c3117c75084ec688a9bf82ce11789e0295f01ca3"
      
      
      $`23b72b52356f0a1ac89925f5431e44a0ab23bab6`
      $`23b72b52356f0a1ac89925f5431e44a0ab23bab6`$type
      [1] "tree"
      
      $`23b72b52356f0a1ac89925f5431e44a0ab23bab6`$raw
       [1] 31 30 30 36 34 34 20 63 68 65 63 6b 2d 73 74 61 6e 64 61 72 64 2e 79 61 6d
      [26] 6c 00 cd df 41 1c 75 0f 67 46 22 bd 53 fe 67 9a 71 69 8d ad da 4c
      
      $`23b72b52356f0a1ac89925f5431e44a0ab23bab6`$size
      [1] 47
      
      $`23b72b52356f0a1ac89925f5431e44a0ab23bab6`$packed_size
      [1] 56
      
      $`23b72b52356f0a1ac89925f5431e44a0ab23bab6`$object
      # A data frame: 1 x 4
        type  mode   path                hash                                    
        <chr> <chr>  <chr>               <chr>                                   
      1 blob  100644 check-standard.yaml cddf411c750f674622bd53fe679a71698dadda4c
      
      $`23b72b52356f0a1ac89925f5431e44a0ab23bab6`$hash
      [1] "23b72b52356f0a1ac89925f5431e44a0ab23bab6"
      
      
      $`5ba9d25318eacd311bb9f73cf83f859b0bfb7235`
      $`5ba9d25318eacd311bb9f73cf83f859b0bfb7235`$type
      [1] "tree"
      
      $`5ba9d25318eacd311bb9f73cf83f859b0bfb7235`$raw
       [1] 34 30 30 30 30 20 64 6f 74 65 6e 76 00 75 21 09 21 ad b3 eb df 80 5d 8b 01
      [26] 69 1e 2d 4b 39 15 d0 96
      
      $`5ba9d25318eacd311bb9f73cf83f859b0bfb7235`$size
      [1] 33
      
      $`5ba9d25318eacd311bb9f73cf83f859b0bfb7235`$packed_size
      [1] 40
      
      $`5ba9d25318eacd311bb9f73cf83f859b0bfb7235`$object
      # A data frame: 1 x 4
        type  mode  path   hash                                    
        <chr> <chr> <chr>  <chr>                                   
      1 tree  40000 dotenv 75210921adb3ebdf805d8b01691e2d4b3915d096
      
      $`5ba9d25318eacd311bb9f73cf83f859b0bfb7235`$hash
      [1] "5ba9d25318eacd311bb9f73cf83f859b0bfb7235"
      
      
      $`75210921adb3ebdf805d8b01691e2d4b3915d096`
      $`75210921adb3ebdf805d8b01691e2d4b3915d096`$type
      [1] "tree"
      
      $`75210921adb3ebdf805d8b01691e2d4b3915d096`$raw
        [1] 31 30 30 36 34 34 20 2e 52 62 75 69 6c 64 69 67 6e 6f 72 65 00 7a 1a 03 c7
       [26] 60 50 51 78 31 cc 70 db 25 e1 2f 03 7f ec 83 22 31 30 30 36 34 34 20 2e 67
       [51] 69 74 69 67 6e 6f 72 65 00 61 ff c7 cb 1f 6d 9a fb 3f 7a 57 22 2b 9a 4a 75
       [76] 7a 02 39 f3 31 30 30 36 34 34 20 44 45 53 43 52 49 50 54 49 4f 4e 00 21 80
      [101] 6f 7f 52 46 6f a0 18 a9 4c d6 9b 76 3c 4a 1c b1 0d 1d 31 30 30 36 34 34 20
      [126] 4c 49 43 45 4e 53 45 00 ce 4f 59 a4 6d a7 92 e7 47 fc 90 05 a7 95 8d 3d cd
      [151] ca 03 cb 31 30 30 36 34 34 20 4e 41 4d 45 53 50 41 43 45 00 bf d1 84 43 a5
      [176] 89 f7 c8 e0 9c a9 02 9e e4 df 97 98 c9 ab 00 31 30 30 36 34 34 20 4e 45 57
      [201] 53 2e 6d 64 00 a8 74 54 ad e2 64 a8 c4 07 fc 2f 42 b5 e6 0a 5a 03 07 ef 54
      [226] 34 30 30 30 30 20 52 00 9a 09 c0 54 0f 29 a3 f0 e0 cd 0b f9 ac d7 30 3e 1f
      [251] a9 bf f2 31 30 30 36 34 34 20 52 45 41 44 4d 45 2e 52 6d 64 00 9c 68 33 64
      [276] 87 70 ee db 4e c8 a1 90 02 f3 45 e8 16 d6 2d ae 31 30 30 36 34 34 20 52 45
      [301] 41 44 4d 45 2e 6d 64 00 12 14 f7 69 75 78 1d 3b 0c 51 7f 42 3b 12 80 37 c6
      [326] e4 50 ed 34 30 30 30 30 20 6d 61 6e 00 d4 ef 76 62 04 f3 08 7d 32 14 8d 71
      [351] 4d f6 64 ae 66 ee ce 82
      
      $`75210921adb3ebdf805d8b01691e2d4b3915d096`$size
      [1] 358
      
      $`75210921adb3ebdf805d8b01691e2d4b3915d096`$packed_size
      [1] 321
      
      $`75210921adb3ebdf805d8b01691e2d4b3915d096`$object
      # A data frame: 10 x 4
         type  mode   path          hash                                    
         <chr> <chr>  <chr>         <chr>                                   
       1 blob  100644 .Rbuildignore 7a1a03c76050517831cc70db25e12f037fec8322
       2 blob  100644 .gitignore    61ffc7cb1f6d9afb3f7a57222b9a4a757a0239f3
       3 blob  100644 DESCRIPTION   21806f7f52466fa018a94cd69b763c4a1cb10d1d
       4 blob  100644 LICENSE       ce4f59a46da792e747fc9005a7958d3dcdca03cb
       5 blob  100644 NAMESPACE     bfd18443a589f7c8e09ca9029ee4df9798c9ab00
       6 blob  100644 NEWS.md       a87454ade264a8c407fc2f42b5e60a5a0307ef54
       7 tree  40000  R             9a09c0540f29a3f0e0cd0bf9acd7303e1fa9bff2
       8 blob  100644 README.Rmd    9c6833648770eedb4ec8a19002f345e816d62dae
       9 blob  100644 README.md     1214f76975781d3b0c517f423b128037c6e450ed
      10 tree  40000  man           d4ef766204f3087d32148d714df664ae66eece82
      
      $`75210921adb3ebdf805d8b01691e2d4b3915d096`$hash
      [1] "75210921adb3ebdf805d8b01691e2d4b3915d096"
      
      
      $`9a09c0540f29a3f0e0cd0bf9acd7303e1fa9bff2`
      $`9a09c0540f29a3f0e0cd0bf9acd7303e1fa9bff2`$type
      [1] "tree"
      
      $`9a09c0540f29a3f0e0cd0bf9acd7303e1fa9bff2`$raw
       [1] 31 30 30 36 34 34 20 64 6f 74 65 6e 76 2d 70 61 63 6b 61 67 65 2e 72 00 8f
      [26] f5 0d 05 56 b5 68 d7 e2 ce af 54 1d a1 10 ff f4 8d 26 68
      
      $`9a09c0540f29a3f0e0cd0bf9acd7303e1fa9bff2`$size
      [1] 44
      
      $`9a09c0540f29a3f0e0cd0bf9acd7303e1fa9bff2`$packed_size
      [1] 53
      
      $`9a09c0540f29a3f0e0cd0bf9acd7303e1fa9bff2`$object
      # A data frame: 1 x 4
        type  mode   path             hash                                    
        <chr> <chr>  <chr>            <chr>                                   
      1 blob  100644 dotenv-package.r 8ff50d0556b568d7e2ceaf541da110fff48d2668
      
      $`9a09c0540f29a3f0e0cd0bf9acd7303e1fa9bff2`$hash
      [1] "9a09c0540f29a3f0e0cd0bf9acd7303e1fa9bff2"
      
      
      $d4ef766204f3087d32148d714df664ae66eece82
      $d4ef766204f3087d32148d714df664ae66eece82$type
      [1] "tree"
      
      $d4ef766204f3087d32148d714df664ae66eece82$raw
       [1] 31 30 30 36 34 34 20 64 6f 74 65 6e 76 2d 70 61 63 6b 61 67 65 2e 52 64 00
      [26] fd d3 00 28 69 44 e3 6a 1c 7f 4b 76 d0 31 f0 be a4 14 7d 47 31 30 30 36 34
      [51] 34 20 6c 6f 61 64 5f 64 6f 74 5f 65 6e 76 2e 52 64 00 dc 24 b6 89 8c bc 56
      [76] ee e6 82 ca 1c 32 73 5e 72 e8 19 e4 ab
      
      $d4ef766204f3087d32148d714df664ae66eece82$size
      [1] 88
      
      $d4ef766204f3087d32148d714df664ae66eece82$packed_size
      [1] 90
      
      $d4ef766204f3087d32148d714df664ae66eece82$object
      # A data frame: 2 x 4
        type  mode   path              hash                                    
        <chr> <chr>  <chr>             <chr>                                   
      1 blob  100644 dotenv-package.Rd fdd300286944e36a1c7f4b76d031f0bea4147d47
      2 blob  100644 load_dot_env.Rd   dc24b6898cbc56eee682ca1c32735e72e819e4ab
      
      $d4ef766204f3087d32148d714df664ae66eece82$hash
      [1] "d4ef766204f3087d32148d714df664ae66eece82"
      
      

# git_list_pack_index

    Code
      git_list_pack_index(path)
    Output
      $objects
                                             hash        crc offset
      1  cefdc0eebcd7f757efb9a80652fd8aaf1a87508e 3597296333     12
      2  aac34e05a0cb9852aa425757dc8480fa29c6c783  585558493    175
      3  61102a589aeef025b8c11015d5de4fd6f91b0fdb 3679009946    343
      4  0784f430fc5830e1b26ceb5fa9a22551fbe9a400  125079169    420
      5  c3117c75084ec688a9bf82ce11789e0295f01ca3  706758901    466
      6  23b72b52356f0a1ac89925f5431e44a0ab23bab6 3850203054    515
      7  5ba9d25318eacd311bb9f73cf83f859b0bfb7235 3361408054    573
      8  75210921adb3ebdf805d8b01691e2d4b3915d096 2355823347    615
      9  9a09c0540f29a3f0e0cd0bf9acd7303e1fa9bff2 2235782520    938
      10 d4ef766204f3087d32148d714df664ae66eece82   23708622    993
      
      $data_chksum
      [1] "775021fe2c99a033831f23b73d3a1176c9c4e435"
      
      $idx_chksum
      [1] "8037ebb73a48b7d4a7d18250fda6bb8370160158"
      

# git_unpack errors

    Code
      git_unpack(pack[1:30])
    Condition
      Error:
      ! ! Invalid packfile from git, too short.

---

    Code
      git_unpack(charToRaw("nope and some more so we have enough bytes"))
    Condition
      Error:
      ! ! Not a git packfile, it does not have a `PACK` header.

---

    Code
      git_unpack(pack2)
    Condition
      Error:
      ! ! Unexpected packfile version, must be version 2.

---

    Code
      git_unpack(pack2)
    Condition
      Error:
      ! ! Checksum mismatch in git packfile.

# parse_int32_nwb

    Code
      parse_int32_nwb(raw(3))
    Condition
      Error:
      ! ! Cannot parse integer, not raw or number of bytes is wrong.

# async_git_resolve_ref

    Code
      sy(async_git_resolve_ref(fake_git$url("/pak-test.git"), "main"))
    Output
      [1] "3f3b0b4ee8a0ff4563073924e5fe069da67a6d8b"
      attr(,"protocol")
      [1] 2
      attr(,"filter")
      [1] TRUE

---

    Code
      sy(async_git_resolve_ref(fake_git$url("/pak-test.git"), "v1"))
    Output
      [1] "cefdc0eebcd7f757efb9a80652fd8aaf1a87508e"
      attr(,"protocol")
      [1] 2
      attr(,"filter")
      [1] TRUE

---

    Code
      sy(async_git_resolve_ref(fake_git$url("/pak-test.git"), "3f3b0b4ee8a0ff"))
    Output
      [1] "3f3b0b4ee8a0ff4563073924e5fe069da67a6d8b"
      attr(,"protocol")
      [1] 2
      attr(,"filter")
      [1] TRUE

---

    Code
      sy(async_git_resolve_ref(fake_git$url("/pak-test.git"), "badcafe"))
    Condition
      Error:
      ! ! Unknown git ref: "badcafe".

---

    Code
      sy(async_git_resolve_ref(fake_git$url("/pak-test.git"), "badcafe"))
    Condition
      Error:
      ! ! Found multiple git refs with prefix "badcafe", it is ambiguous.
      i Matching git refs: "badcafe1" and "badcafe2".
      i Specify a longer prefix to choose a single git ref.

# git_fetch_process_v1

    Code
      git_fetch_process_v1(list(), url, "badcafe")
    Condition
      Error:
      ! ! Empty reply from git server (protocol v1) at <https://<auth>@example.com>.

---

    Code
      git_fetch_process_v1(list(list(type = "boo")), url, "badcafe")
    Condition
      Error:
      ! ! No PACK in git server response (protocol v1) from <https://<auth>@example.com>.

# git_download_repo

    Code
      dir(tmp, recursive = TRUE)
    Output
       [1] "v1/foo"                                
       [2] "v1/subdir/dotenv/DESCRIPTION"          
       [3] "v1/subdir/dotenv/LICENSE"              
       [4] "v1/subdir/dotenv/NAMESPACE"            
       [5] "v1/subdir/dotenv/NEWS.md"              
       [6] "v1/subdir/dotenv/R/dotenv-package.r"   
       [7] "v1/subdir/dotenv/README.Rmd"           
       [8] "v1/subdir/dotenv/README.md"            
       [9] "v1/subdir/dotenv/man/dotenv-package.Rd"
      [10] "v1/subdir/dotenv/man/load_dot_env.Rd"  
      [11] "v1/wipe.R"                             

# unpack_packfile_repo

    Code
      sort(dir(output, recursive = TRUE))
    Output
       [1] "foo"                                 "subdir/dotenv/DESCRIPTION"          
       [3] "subdir/dotenv/LICENSE"               "subdir/dotenv/NAMESPACE"            
       [5] "subdir/dotenv/NEWS.md"               "subdir/dotenv/R/dotenv-package.r"   
       [7] "subdir/dotenv/README.Rmd"            "subdir/dotenv/README.md"            
       [9] "subdir/dotenv/man/dotenv-package.Rd" "subdir/dotenv/man/load_dot_env.Rd"  
      [11] "wipe.R"                             

# resolve_remote

    Code
      snapshot(res, extra = "all")
    Output
      # A data frame: 1 x 31
        ref   type     direct directpkg status package version license
        <chr> <chr>    <lgl>  <lgl>     <chr>  <chr>   <chr>   <chr>  
      1 pkg1  standard TRUE   TRUE      OK     pkg1    1.0.0   <NA>   
        needscompilation priority md5sum sha256   filesize built platform rversion
        <lgl>            <chr>    <chr>  <chr>       <int> <chr> <chr>    <chr>   
      1 FALSE            <NA>     <md5>  <sha256>       42 <NA>  source   *       
        repotype repodir     target                        deps         
        <chr>    <chr>       <chr>                         <list>       
      1 cran     src/contrib src/contrib/pkg1_1.0.0.tar.gz <tbl [0 x 5]>
        mirror                  sources   remote         error      metadata 
        <chr>                   <list>    <list>         <list>     <list>   
      1 http://127.0.0.1:<port>/ <chr [2]> <rmt_rf_s [6]> <list [0]> <chr [6]>
        extra      dep_types params    sysreqs os_type cache_status
        <list>     <list>    <list>    <chr>   <chr>   <chr>       
      1 <list [0]> <chr [3]> <chr [0]> <NA>    <NA>    miss        
      + sources:
      http://127.0.0.1:<port>//src/contrib/pkg1_1.0.0.tar.gz, http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_1.0.0.tar.gz
      + remote:
      <remote_ref_standard/remote_ref/list> package: pkg1; atleast: ; version: ; ref: pkg1; type: standard; params: 
      + error:
      -
      + metadata:
      RemoteType: standard; RemotePkgRef: pkg1; RemoteRef: pkg1; RemoteRepos: http://127.0.0.1:<port>/; RemotePkgPlatform: source; RemoteSha: 1.0.0
      + dep_types:
      Depends, Imports, LinkingTo

# resolve_remote, multiple

    Code
      snapshot(res, extra = "all")
    Output
      # A data frame: 3 x 31
        ref        type     direct directpkg status package version license
        <chr>      <chr>    <lgl>  <lgl>     <chr>  <chr>   <chr>   <chr>  
      1 pkg2       standard FALSE  FALSE     OK     pkg2    1.0.0   <NA>   
      2 cran::pkg3 cran     TRUE   TRUE      OK     pkg3    1.0.0   <NA>   
      3 pkg1       standard TRUE   TRUE      OK     pkg1    1.0.0   <NA>   
        needscompilation priority md5sum sha256   filesize built platform rversion
        <lgl>            <chr>    <chr>  <chr>       <int> <chr> <chr>    <chr>   
      1 FALSE            <NA>     <md5>  <sha256>       42 <NA>  source   *       
      2 FALSE            <NA>     <md5>  <sha256>       42 <NA>  source   *       
      3 FALSE            <NA>     <md5>  <sha256>       42 <NA>  source   *       
        repotype repodir     target                        deps         
        <chr>    <chr>       <chr>                         <list>       
      1 cran     src/contrib src/contrib/pkg2_1.0.0.tar.gz <tbl [1 x 5]>
      2 cran     src/contrib src/contrib/pkg3_1.0.0.tar.gz <tbl [1 x 5]>
      3 cran     src/contrib src/contrib/pkg1_1.0.0.tar.gz <tbl [0 x 5]>
        mirror                  sources   remote         error      metadata 
        <chr>                   <list>    <list>         <list>     <list>   
      1 http://127.0.0.1:<port>/ <chr [2]> <rmt_rf_s [6]> <list [0]> <chr [6]>
      2 http://127.0.0.1:<port>/ <chr [2]> <rmt_rf_c [6]> <list [0]> <chr [6]>
      3 http://127.0.0.1:<port>/ <chr [2]> <rmt_rf_s [6]> <list [0]> <chr [6]>
        extra      dep_types params    sysreqs os_type cache_status
        <list>     <list>    <list>    <chr>   <chr>   <chr>       
      1 <list [0]> <chr [3]> <chr [0]> <NA>    <NA>    miss        
      2 <list [0]> <chr [3]> <chr [0]> <NA>    <NA>    miss        
      3 <list [0]> <chr [3]> <chr [0]> <NA>    <NA>    miss        
      + sources:
      http://127.0.0.1:<port>//src/contrib/pkg2_1.0.0.tar.gz, http://127.0.0.1:<port>//src/contrib/Archive/pkg2/pkg2_1.0.0.tar.gz
      http://127.0.0.1:<port>//src/contrib/pkg3_1.0.0.tar.gz, http://127.0.0.1:<port>//src/contrib/Archive/pkg3/pkg3_1.0.0.tar.gz
      http://127.0.0.1:<port>//src/contrib/pkg1_1.0.0.tar.gz, http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_1.0.0.tar.gz
      + remote:
      <remote_ref_standard/remote_ref/list> package: pkg2; atleast: ; version: ; ref: pkg2; type: standard; params: 
      <remote_ref_cran/remote_ref/list> package: pkg3; atleast: ; version: ; ref: cran::pkg3; type: cran; params: 
      <remote_ref_standard/remote_ref/list> package: pkg1; atleast: ; version: ; ref: pkg1; type: standard; params: 
      + error:
      -
      -
      -
      + metadata:
      RemoteType: standard; RemotePkgRef: pkg2; RemoteRef: pkg2; RemoteRepos: http://127.0.0.1:<port>/; RemotePkgPlatform: source; RemoteSha: 1.0.0
      RemoteType: cran; RemotePkgRef: cran::pkg3; RemoteRef: cran::pkg3; RemoteRepos: http://127.0.0.1:<port>/; RemotePkgPlatform: source; RemoteSha: 1.0.0
      RemoteType: standard; RemotePkgRef: pkg1; RemoteRef: pkg1; RemoteRepos: http://127.0.0.1:<port>/; RemotePkgPlatform: source; RemoteSha: 1.0.0
      + dep_types:
      Depends, Imports, LinkingTo
      Depends, Imports, LinkingTo
      Depends, Imports, LinkingTo

# dependencies

    Code
      snapshot(res, extra = "all")
    Output
      # A data frame: 3 x 31
        ref   type     direct directpkg status package version license
        <chr> <chr>    <lgl>  <lgl>     <chr>  <chr>   <chr>   <chr>  
      1 pkg1  standard FALSE  FALSE     OK     pkg1    1.0.0   <NA>   
      2 pkg2  standard FALSE  FALSE     OK     pkg2    1.0.0   <NA>   
      3 pkg3  standard TRUE   TRUE      OK     pkg3    1.0.0   <NA>   
        needscompilation priority md5sum sha256   filesize built platform rversion
        <lgl>            <chr>    <chr>  <chr>       <int> <chr> <chr>    <chr>   
      1 FALSE            <NA>     <md5>  <sha256>       42 <NA>  source   *       
      2 FALSE            <NA>     <md5>  <sha256>       42 <NA>  source   *       
      3 FALSE            <NA>     <md5>  <sha256>       42 <NA>  source   *       
        repotype repodir     target                        deps         
        <chr>    <chr>       <chr>                         <list>       
      1 cran     src/contrib src/contrib/pkg1_1.0.0.tar.gz <tbl [0 x 5]>
      2 cran     src/contrib src/contrib/pkg2_1.0.0.tar.gz <tbl [1 x 5]>
      3 cran     src/contrib src/contrib/pkg3_1.0.0.tar.gz <tbl [1 x 5]>
        mirror                  sources   remote         error      metadata 
        <chr>                   <list>    <list>         <list>     <list>   
      1 http://127.0.0.1:<port>/ <chr [2]> <rmt_rf_s [6]> <list [0]> <chr [6]>
      2 http://127.0.0.1:<port>/ <chr [2]> <rmt_rf_s [6]> <list [0]> <chr [6]>
      3 http://127.0.0.1:<port>/ <chr [2]> <rmt_rf_s [6]> <list [0]> <chr [6]>
        extra      dep_types params    sysreqs os_type cache_status
        <list>     <list>    <list>    <chr>   <chr>   <chr>       
      1 <list [0]> <chr [3]> <chr [0]> <NA>    <NA>    miss        
      2 <list [0]> <chr [3]> <chr [0]> <NA>    <NA>    miss        
      3 <list [0]> <chr [3]> <chr [0]> <NA>    <NA>    miss        
      + sources:
      http://127.0.0.1:<port>//src/contrib/pkg1_1.0.0.tar.gz, http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_1.0.0.tar.gz
      http://127.0.0.1:<port>//src/contrib/pkg2_1.0.0.tar.gz, http://127.0.0.1:<port>//src/contrib/Archive/pkg2/pkg2_1.0.0.tar.gz
      http://127.0.0.1:<port>//src/contrib/pkg3_1.0.0.tar.gz, http://127.0.0.1:<port>//src/contrib/Archive/pkg3/pkg3_1.0.0.tar.gz
      + remote:
      <remote_ref_standard/remote_ref/list> package: pkg1; atleast: ; version: ; ref: pkg1; type: standard; params: 
      <remote_ref_standard/remote_ref/list> package: pkg2; atleast: ; version: ; ref: pkg2; type: standard; params: 
      <remote_ref_standard/remote_ref/list> package: pkg3; atleast: ; version: ; ref: pkg3; type: standard; params: 
      + error:
      -
      -
      -
      + metadata:
      RemoteType: standard; RemotePkgRef: pkg1; RemoteRef: pkg1; RemoteRepos: http://127.0.0.1:<port>/; RemotePkgPlatform: source; RemoteSha: 1.0.0
      RemoteType: standard; RemotePkgRef: pkg2; RemoteRef: pkg2; RemoteRepos: http://127.0.0.1:<port>/; RemotePkgPlatform: source; RemoteSha: 1.0.0
      RemoteType: standard; RemotePkgRef: pkg3; RemoteRef: pkg3; RemoteRepos: http://127.0.0.1:<port>/; RemotePkgPlatform: source; RemoteSha: 1.0.0
      + dep_types:
      Depends, Imports, LinkingTo
      Depends, Imports, LinkingTo
      Depends, Imports, LinkingTo

# failed resolution

    Code
      snapshot(res, extra = "all")
    Output
      # A data frame: 1 x 31
        ref                      type  direct directpkg status package           
        <chr>                    <chr> <lgl>  <lgl>     <chr>  <chr>             
      1 cran::xxyyzzqwertyqwerty cran  TRUE   TRUE      FAILED xxyyzzqwertyqwerty
        version license needscompilation priority md5sum sha256 filesize built
        <chr>   <chr>   <lgl>            <chr>    <chr>  <chr>     <int> <chr>
      1 <NA>    <NA>    TRUE             <NA>     <NA>   <NA>         NA <NA> 
        platform rversion repotype repodir    
        <chr>    <chr>    <chr>    <chr>      
      1 source   *        <NA>     src/contrib
        target                                   deps          mirror sources  
        <chr>                                    <list>        <chr>  <list>   
      1 src/contrib/xxyyzzqwertyqwerty_NA.tar.gz <tbl [0 x 5]> <NA>   <chr [1]>
        remote         error   metadata   extra      dep_types params    sysreqs
        <list>         <list>  <list>     <list>     <list>    <list>    <chr>  
      1 <rmt_rf_c [6]> <error> <list [0]> <list [0]> <chr [3]> <chr [0]> <NA>   
        os_type cache_status
        <chr>   <chr>       
      1 <NA>    miss        
      + sources:
      NA
      + remote:
      <remote_ref_cran/remote_ref/list> package: xxyyzzqwertyqwerty; atleast: ; version: ; ref: cran::xxyyzzqwertyqwerty; type: cran; params: 
      + error:
      <error/condition> Can't find package called cran::xxyyzzqwertyqwerty.
      + metadata:
      -
      + dep_types:
      Depends, Imports, LinkingTo

# failed resolution, multiple

    Code
      snapshot(res, extra = "all")
    Output
      # A data frame: 2 x 31
        ref                      type  direct directpkg status package           
        <chr>                    <chr> <lgl>  <lgl>     <chr>  <chr>             
      1 cran::pkg1               cran  TRUE   TRUE      OK     pkg1              
      2 cran::xxyyzzqwertyqwerty cran  TRUE   TRUE      FAILED xxyyzzqwertyqwerty
        version license needscompilation priority md5sum sha256   filesize built
        <chr>   <chr>   <lgl>            <chr>    <chr>  <chr>       <int> <chr>
      1 1.0.0   <NA>    FALSE            <NA>     <md5>  <sha256>       42 <NA> 
      2 <NA>    <NA>    TRUE             <NA>     <NA>   <NA>           NA <NA> 
        platform rversion repotype repodir    
        <chr>    <chr>    <chr>    <chr>      
      1 source   *        cran     src/contrib
      2 source   *        <NA>     src/contrib
        target                                   deps          mirror                 
        <chr>                                    <list>        <chr>                  
      1 src/contrib/pkg1_1.0.0.tar.gz            <tbl [0 x 5]> http://127.0.0.1:<port>/
      2 src/contrib/xxyyzzqwertyqwerty_NA.tar.gz <tbl [0 x 5]> <NA>                   
        sources   remote         error      metadata   extra      dep_types params   
        <list>    <list>         <list>     <list>     <list>     <list>    <list>   
      1 <chr [2]> <rmt_rf_c [6]> <list [0]> <chr [6]>  <list [0]> <chr [3]> <chr [0]>
      2 <chr [1]> <rmt_rf_c [6]> <error>    <list [0]> <list [0]> <chr [3]> <chr [0]>
        sysreqs os_type cache_status
        <chr>   <chr>   <chr>       
      1 <NA>    <NA>    miss        
      2 <NA>    <NA>    miss        
      + sources:
      http://127.0.0.1:<port>//src/contrib/pkg1_1.0.0.tar.gz, http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_1.0.0.tar.gz
      NA
      + remote:
      <remote_ref_cran/remote_ref/list> package: pkg1; atleast: ; version: ; ref: cran::pkg1; type: cran; params: 
      <remote_ref_cran/remote_ref/list> package: xxyyzzqwertyqwerty; atleast: ; version: ; ref: cran::xxyyzzqwertyqwerty; type: cran; params: 
      + error:
      -
      <error/condition> Can't find package called cran::xxyyzzqwertyqwerty.
      + metadata:
      RemoteType: cran; RemotePkgRef: cran::pkg1; RemoteRef: cran::pkg1; RemoteRepos: http://127.0.0.1:<port>/; RemotePkgPlatform: source; RemoteSha: 1.0.0
      -
      + dep_types:
      Depends, Imports, LinkingTo
      Depends, Imports, LinkingTo

# resolve current version

    Code
      snapshot(res, extra = "all")
    Output
      # A data frame: 1 x 31
        ref                type  direct directpkg status package version license
        <chr>              <chr> <lgl>  <lgl>     <chr>  <chr>   <chr>   <chr>  
      1 cran::pkg1@current cran  TRUE   TRUE      OK     pkg1    1.0.0   <NA>   
        needscompilation priority md5sum sha256   filesize built platform rversion
        <lgl>            <chr>    <chr>  <chr>       <int> <chr> <chr>    <chr>   
      1 FALSE            <NA>     <md5>  <sha256>       42 <NA>  source   *       
        repotype repodir     target                        deps         
        <chr>    <chr>       <chr>                         <list>       
      1 cran     src/contrib src/contrib/pkg1_1.0.0.tar.gz <tbl [0 x 5]>
        mirror                  sources   remote         error      metadata 
        <chr>                   <list>    <list>         <list>     <list>   
      1 http://127.0.0.1:<port>/ <chr [2]> <rmt_rf_c [6]> <list [0]> <chr [6]>
        extra      dep_types params    sysreqs os_type cache_status
        <list>     <list>    <list>    <chr>   <chr>   <chr>       
      1 <list [0]> <chr [3]> <chr [0]> <NA>    <NA>    miss        
      + sources:
      http://127.0.0.1:<port>//src/contrib/pkg1_1.0.0.tar.gz, http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_1.0.0.tar.gz
      + remote:
      <remote_ref_cran/remote_ref/list> package: pkg1; atleast: ==; version: current; ref: cran::pkg1@current; type: cran; params: 
      + error:
      -
      + metadata:
      RemoteType: cran; RemotePkgRef: cran::pkg1@current; RemoteRef: cran::pkg1@current; RemoteRepos: http://127.0.0.1:<port>/; RemotePkgPlatform: source; RemoteSha: 1.0.0
      + dep_types:
      Depends, Imports, LinkingTo

# resolve an old version

    Code
      snapshot(res, extra = "all")
    Output
      # A data frame: 2 x 31
        ref        type     direct directpkg status package version license
        <chr>      <chr>    <lgl>  <lgl>     <chr>  <chr>   <chr>   <chr>  
      1 pkg1@0.9.0 standard TRUE   TRUE      OK     pkg1    0.9.0   <NA>   
      2 pkg1       standard FALSE  TRUE      OK     pkg1    1.0.0   <NA>   
        needscompilation priority md5sum sha256   filesize built platform rversion
        <lgl>            <chr>    <chr>  <chr>       <int> <chr> <chr>    <chr>   
      1 FALSE            <NA>     <NA>   <NA>           NA <NA>  source   *       
      2 FALSE            <NA>     <md5>  <sha256>       42 <NA>  source   *       
        repotype repodir     target                                   deps         
        <chr>    <chr>       <chr>                                    <list>       
      1 <NA>     src/contrib src/contrib/pkg1_0.9.0-<hash>.tar.gz <tbl [0 x 5]>
      2 cran     src/contrib src/contrib/pkg1_1.0.0.tar.gz            <tbl [0 x 5]>
        mirror                  sources   remote         error      metadata 
        <chr>                   <list>    <list>         <list>     <list>   
      1 <NA>                    <chr [1]> <rmt_rf_s [8]> <list [0]> <chr [8]>
      2 http://127.0.0.1:<port>/ <chr [2]> <rmt_rf_s [6]> <list [0]> <chr [6]>
        extra            dep_types params    sysreqs os_type cache_status
        <list>           <list>    <list>    <chr>   <chr>   <chr>       
      1 <named list [1]> <chr [3]> <chr [0]> ""      <NA>    miss        
      2 <list [0]>       <chr [3]> <chr [0]> <NA>    <NA>    miss        
      + sources:
      http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz
      http://127.0.0.1:<port>//src/contrib/pkg1_1.0.0.tar.gz, http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_1.0.0.tar.gz
      + remote:
      <remote_ref_standard/remote_ref/list> package: pkg1; atleast: ==; version: 0.9.0; ref: pkg1@0.9.0; type: standard; params: ; url: http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz; hash: <hash>
      <remote_ref_standard/remote_ref/list> package: pkg1; atleast: ; version: ; ref: pkg1; type: standard; params: 
      + error:
      -
      -
      + metadata:
      RemotePkgRef: pkg1@0.9.0; RemoteType: standard; RemoteEtag: "<etag>"; RemotePackaged: TRUE; RemoteRef: pkg1; RemoteRepos: http://127.0.0.1:<port>/; RemotePkgPlatform: source; RemoteSha: 0.9.0
      RemoteType: standard; RemotePkgRef: pkg1; RemoteRef: pkg1; RemoteRepos: http://127.0.0.1:<port>/; RemotePkgPlatform: source; RemoteSha: 1.0.0
      + dep_types:
      Depends, Imports, LinkingTo
      Depends, Imports, LinkingTo

---

    Code
      snapshot(res, extra = "all")
    Output
      # A data frame: 1 x 31
        ref        type     direct directpkg status package version license
        <chr>      <chr>    <lgl>  <lgl>     <chr>  <chr>   <chr>   <chr>  
      1 pkg1@1.0.0 standard TRUE   TRUE      OK     pkg1    1.0.0   <NA>   
        needscompilation priority md5sum sha256   filesize built platform rversion
        <lgl>            <chr>    <chr>  <chr>       <int> <chr> <chr>    <chr>   
      1 FALSE            <NA>     <md5>  <sha256>       42 <NA>  source   *       
        repotype repodir     target                        deps         
        <chr>    <chr>       <chr>                         <list>       
      1 cran     src/contrib src/contrib/pkg1_1.0.0.tar.gz <tbl [0 x 5]>
        mirror                  sources   remote         error      metadata 
        <chr>                   <list>    <list>         <list>     <list>   
      1 http://127.0.0.1:<port>/ <chr [2]> <rmt_rf_s [6]> <list [0]> <chr [6]>
        extra      dep_types params    sysreqs os_type cache_status
        <list>     <list>    <list>    <chr>   <chr>   <chr>       
      1 <list [0]> <chr [3]> <chr [0]> <NA>    <NA>    miss        
      + sources:
      http://127.0.0.1:<port>//src/contrib/pkg1_1.0.0.tar.gz, http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_1.0.0.tar.gz
      + remote:
      <remote_ref_standard/remote_ref/list> package: pkg1; atleast: ==; version: 1.0.0; ref: pkg1@1.0.0; type: standard; params: 
      + error:
      -
      + metadata:
      RemoteType: standard; RemotePkgRef: pkg1@1.0.0; RemoteRef: pkg1@1.0.0; RemoteRepos: http://127.0.0.1:<port>/; RemotePkgPlatform: source; RemoteSha: 1.0.0
      + dep_types:
      Depends, Imports, LinkingTo

# resolve a version range

    Code
      snapshot(res, extra = "all")
    Output
      # A data frame: 1 x 31
        ref          type     direct directpkg status package version license
        <chr>        <chr>    <lgl>  <lgl>     <chr>  <chr>   <chr>   <chr>  
      1 pkg1@>=0.9.0 standard TRUE   TRUE      FAILED pkg1    <NA>    <NA>   
        needscompilation priority md5sum sha256 filesize built platform rversion
        <lgl>            <chr>    <chr>  <chr>     <int> <chr> <chr>    <chr>   
      1 TRUE             <NA>     <NA>   <NA>         NA <NA>  source   *       
        repotype repodir     target                     deps          mirror sources  
        <chr>    <chr>       <chr>                      <list>        <chr>  <list>   
      1 <NA>     src/contrib src/contrib/pkg1_NA.tar.gz <tbl [0 x 5]> <NA>   <chr [1]>
        remote         error      metadata   extra      dep_types params    sysreqs
        <list>         <list>     <list>     <list>     <list>    <list>    <chr>  
      1 <rmt_rf_s [6]> <async_rj> <list [0]> <list [0]> <chr [3]> <chr [0]> <NA>   
        os_type cache_status
        <chr>   <chr>       
      1 <NA>    miss        
      + sources:
      NA
      + remote:
      <remote_ref_standard/remote_ref/list> package: pkg1; atleast: >=; version: 0.9.0; ref: pkg1@>=0.9.0; type: standard; params: 
      + error:
      <async_rejected/rlib_error_3_0/rlib_error/error/condition> ! pkgdepends resolution error for pkg1@>=0.9.0.
      Caused by error:
      ! Version ranges are not implemented yet.
      + metadata:
      -
      + dep_types:
      Depends, Imports, LinkingTo

# versioned cran

    Code
      prop$draw()
    Output
      local::<path>
      +-pkg1 1.0.0 [new][bld][dl] (<size>)
      \-pkg3 0.9.9 [new][bld][dl] (unknown size)
      
      Key:  [new] new | [dl] download | [bld] build

# resolve_remote_git

    Code
      as.list(res)
    Output
      $ref
      [1] "git::https://github.com/r-lib/cli@v3.6.0"
      
      $type
      [1] "git"
      
      $direct
      [1] TRUE
      
      $directpkg
      [1] TRUE
      
      $status
      [1] "OK"
      
      $package
      [1] "cli"
      
      $version
      [1] "3.6.0"
      
      $license
      [1] "MIT + file LICENSE"
      
      $needscompilation
      [1] TRUE
      
      $priority
      [1] NA
      
      $md5sum
      [1] NA
      
      $sha256
      [1] NA
      
      $filesize
      [1] NA
      
      $built
      [1] NA
      
      $platform
      [1] "source"
      
      $rversion
      [1] "*"
      
      $repotype
      [1] NA
      
      $repodir
      [1] "src/contrib"
      
      $target
      [1] "src/contrib/cli_3.6.0_git_4cc6312"
      
      $deps
      $deps[[1]]
      # A data frame: 32 x 5
         ref         type     package     op    version
         <chr>       <chr>    <chr>       <chr> <chr>  
       1 R           Depends  R           ">="  "3.4"  
       2 callr       Suggests callr       ""    ""     
       3 covr        Suggests covr        ""    ""     
       4 crayon      Suggests crayon      ""    ""     
       5 digest      Suggests digest      ""    ""     
       6 glue        Suggests glue        ">="  "1.6.0"
       7 htmltools   Suggests htmltools   ""    ""     
       8 htmlwidgets Suggests htmlwidgets ""    ""     
       9 knitr       Suggests knitr       ""    ""     
      10 mockery     Suggests mockery     ""    ""     
      # i 22 more rows
      
      
      $mirror
      [1] NA
      
      $sources
      $sources[[1]]
      [1] "https://github.com/r-lib/cli"
      
      
      $remote
      $remote[[1]]
      $package
      [1] "cli"
      
      $protocol
      [1] "https"
      
      $host
      [1] "github.com"
      
      $path
      [1] "/r-lib/"
      
      $repo
      [1] "cli"
      
      $commitish
      [1] "v3.6.0"
      
      $ref
      [1] "git::https://github.com/r-lib/cli@v3.6.0"
      
      $type
      [1] "git"
      
      $dotgit
      [1] ""
      
      $url
      [1] "https://github.com/r-lib/cli"
      
      $params
      character(0)
      
      attr(,"class")
      [1] "remote_ref_git" "remote_ref"     "list"          
      
      
      $error
      $error[[1]]
      list()
      
      
      $metadata
      $metadata[[1]]
                                      RemoteType 
                                           "git" 
                                       RemoteUrl 
                  "https://github.com/r-lib/cli" 
                                    RemotePkgRef 
      "git::https://github.com/r-lib/cli@v3.6.0" 
                                       RemoteRef 
                                        "v3.6.0" 
                                       RemoteSha 
      "4cc6312883f2842e948f34a816ce8e846f323a6e" 
      
      
      $extra
      $extra[[1]]
      $extra[[1]]$remotesha
      [1] "4cc6312883f2842e948f34a816ce8e846f323a6e"
      attr(,"protocol")
      [1] 2
      attr(,"filter")
      [1] TRUE
      
      
      
      $dep_types
      $dep_types[[1]]
      [1] "Depends"   "Imports"   "LinkingTo"
      
      
      $params
      $params[[1]]
      character(0)
      
      
      $sysreqs
      [1] ""
      
      $os_type
      [1] NA
      
      $cache_status
      [1] "miss"
      
      attr(,"metadata")
      named list()

# new_pkg_installation_proposal

    Code
      prop
    Output
      <pkg_installation_proposal>
      + refs:
        - pkg3
      + solution policy: lazy
      (use `$solve()` to solve dependencies)

---

    Code
      prop$get_refs()
    Output
      [1] "pkg3"

---

    Code
      sort(prop$get_config()$list())
    Output
       [1] "build_vignettes"           "cache_dir"                
       [3] "cran_mirror"               "dependencies"             
       [5] "git_submodules"            "goal"                     
       [7] "include_linkingto"         "library"                  
       [9] "metadata_cache_dir"        "metadata_update_after"    
      [11] "package_cache_dir"         "platforms"                
      [13] "r_versions"                "sysreqs"                  
      [15] "sysreqs_db_update"         "sysreqs_db_update_timeout"
      [17] "sysreqs_dry_run"           "sysreqs_lookup_system"    
      [19] "sysreqs_platform"          "sysreqs_rspm_repo_id"     
      [21] "sysreqs_rspm_url"          "sysreqs_sudo"             
      [23] "sysreqs_update"            "sysreqs_verbose"          
      [25] "use_bioconductor"          "windows_archs"            

---

    Code
      prop$get_resolution()[, c("ref", "type", "directpkg", "package", "error")]
    Output
      # A data frame: 3 x 5
        ref   type     directpkg package error     
        <chr> <chr>    <lgl>     <chr>   <list>    
      1 pkg1  standard FALSE     pkg1    <list [0]>
      2 pkg2  standard FALSE     pkg2    <list [0]>
      3 pkg3  standard TRUE      pkg3    <list [0]>

---

    Code
      prop$get_solve_policy()
    Output
      [1] "lazy"

---

    Code
      prop$get_solve_policy()
    Output
      [1] "upgrade"

---

    Code
      prop$get_solution()
    Output
      <pkg_solution>
      + result: OK
      + refs:
        - pkg3
      + constraints (5):
        - select pkg3 exactly once
        - select pkg1 at most once
        - select pkg2 at most once
        - pkg2 depends on pkg1: version pkg1 1.0.0
        - pkg3 depends on pkg2: version pkg2 1.0.0
      + solution:
        - pkg1
        - pkg2
        - pkg3

---

    Code
      prop$show_solution()
    Message
      + pkg1   1.0.0 [bld][dl] (<size>)
      + pkg2   1.0.0 [bld][dl] (<size>)
      + pkg3   1.0.0 [bld][dl] (<size>)

---

    Code
      prop$draw()
    Output
      pkg3 1.0.0 [new][bld][dl] (<size>)
      \-pkg2 1.0.0 [new][bld][dl] (<size>)
        \-pkg1 1.0.0 [new][bld][dl] (<size>)
      
      Key:  [new] new | [dl] download | [bld] build

# async_resolve

    Code
      suppressMessages(synchronize(prop$async_resolve()$then(function() "done")))
    Output
      [1] "done"

---

    Code
      prop$get_resolution()[, c("ref", "type", "directpkg", "package", "error")]
    Output
      # A data frame: 3 x 5
        ref   type     directpkg package error     
        <chr> <chr>    <lgl>     <chr>   <list>    
      1 pkg1  standard FALSE     pkg1    <list [0]>
      2 pkg2  standard FALSE     pkg2    <list [0]>
      3 pkg3  standard TRUE      pkg3    <list [0]>

# download

    Code
      prop$get_downloads()[, c("target", "error")]
    Output
      # A data frame: 3 x 2
        target                        error     
        <chr>                         <list>    
      1 src/contrib/pkg1_1.0.0.tar.gz <list [0]>
      2 src/contrib/pkg2_1.0.0.tar.gz <list [0]>
      3 src/contrib/pkg3_1.0.0.tar.gz <list [0]>

# async_download

    Code
      prop$get_downloads()[, c("target", "error")]
    Output
      # A data frame: 3 x 2
        target                        error     
        <chr>                         <list>    
      1 src/contrib/pkg1_1.0.0.tar.gz <list [0]>
      2 src/contrib/pkg2_1.0.0.tar.gz <list [0]>
      3 src/contrib/pkg3_1.0.0.tar.gz <list [0]>

# install

    Code
      pkgcache::parse_installed(lib)[, c("Package", "RemoteRepos")]
    Output
      # A data frame: 3 x 2
        Package RemoteRepos            
        <chr>   <chr>                  
      1 pkg1    http://127.0.0.1:<port>/
      2 pkg2    http://127.0.0.1:<port>/
      3 pkg3    http://127.0.0.1:<port>/

# get_install_plan

    Code
      prop$get_install_plan()[, c("package", "direct", "dependencies")]
    Output
      # A data frame: 3 x 3
        package direct dependencies
        <chr>   <lgl>  <I<list>>   
      1 pkg1    FALSE  <chr [0]>   
      2 pkg2    FALSE  <chr [1]>   
      3 pkg3    TRUE   <chr [1]>   

---

    Code
      prop$get_install_plan()[["dependencies"]]
    Output
      [[1]]
      character(0)
      
      [[2]]
      [1] "pkg1"
      
      [[3]]
      [1] "pkg2"
      

# show_sysreqs

    Code
      prop$show_sysreqs()
    Message
      * libcurl4-openssl-dev  - curl
      * libssl-dev            - curl

# install_sysreqs

    Code
      prop$install_sysreqs()
    Message
      i Installing system requirements
      i Executing `sh -c echo apt-get -y update`
      i Executing `sh -c echo apt-get -y install libcurl4-openssl-dev libssl-dev`

---

    Code
      prop$install_sysreqs()
    Output
      NULL

---

    Code
      prop$install_sysreqs()
    Output
      NULL

# write_str vectors

    Code
      tojson$write_str(NULL)
    Output
      [1] "{}"
    Code
      tojson$write_str(list())
    Output
      [1] "[]"
    Code
      tojson$write_str(integer())
    Output
      [1] "[]"
    Code
      tojson$write_str(double())
    Output
      [1] "[]"
    Code
      tojson$write_str(character())
    Output
      [1] "[]"
    Code
      tojson$write_str(logical())
    Output
      [1] "[]"

---

    Code
      tojson$write_str(complex())
    Condition
      Error:
      ! ! Cannot convert atomic complex vectors to JSON.
    Code
      tojson$write_str(raw())
    Condition
      Error:
      ! ! Cannot convert atomic raw vectors to JSON.

---

    Code
      tojson$write_str(list(1))
    Output
      [1] "[[1]]"
    Code
      tojson$write_str(1L)
    Output
      [1] "[1]"
    Code
      tojson$write_str(1)
    Output
      [1] "[1]"
    Code
      tojson$write_str("foo")
    Output
      [1] "[\"foo\"]"
    Code
      tojson$write_str(TRUE)
    Output
      [1] "[true]"

---

    Code
      tojson$write_str(list(1, 2))
    Output
      [1] "[[1],[2]]"
    Code
      tojson$write_str(1:2)
    Output
      [1] "[1,2]"
    Code
      tojson$write_str(c(1, 2))
    Output
      [1] "[1,2]"
    Code
      tojson$write_str(c("foo", "bar"))
    Output
      [1] "[\"foo\",\"bar\"]"
    Code
      tojson$write_str(c(TRUE, FALSE))
    Output
      [1] "[true,false]"

---

    Code
      tojson$write_str(NA_integer_)
    Output
      [1] "[null]"
    Code
      tojson$write_str(NA_real_)
    Output
      [1] "[null]"
    Code
      tojson$write_str(NA_character_)
    Output
      [1] "[null]"
    Code
      tojson$write_str(NA)
    Output
      [1] "[null]"

---

    Code
      tojson$write_str(c(1L, NA_integer_))
    Output
      [1] "[1,null]"
    Code
      tojson$write_str(c(1, NA_real_))
    Output
      [1] "[1,null]"
    Code
      tojson$write_str(c("foo", NA_character_))
    Output
      [1] "[\"foo\",null]"
    Code
      tojson$write_str(c(TRUE, NA))
    Output
      [1] "[true,null]"

---

    Code
      tojson$write_str(c(a = 1L, b = 2L))
    Output
      [1] "[1,2]"
    Code
      tojson$write_str(c(a = 1, b = 2))
    Output
      [1] "[1,2]"
    Code
      tojson$write_str(c(a = "foo", b = "bar"))
    Output
      [1] "[\"foo\",\"bar\"]"
    Code
      tojson$write_str(c(a = TRUE, b = FALSE))
    Output
      [1] "[true,false]"

# write_str character encoding and escaping

    Code
      cat(tojson$write_str("foo\"\\bar"))
    Output
      ["foo\"\\bar"]
    Code
      charToRaw(tojson$write_str(iconv(utf8, "UTF-8", "latin1")))
    Output
       [1] 5b 22 47 c3 a1 62 6f 72 22 5d

# lists

    Code
      cat(tojson$write_str(list(list(1, 2, 3), list(4, 5, 6), list(7, 8, 9))))
    Output
      [[[1],[2],[3]],[[4],[5],[6]],[[7],[8],[9]]]

---

    Code
      cat(tojson$write_str(list(a = 1, b = 2)))
    Output
      {"a":[1],"b":[2]}

---

    Code
      cat(tojson$write_str(list(a = list(a1 = 1, a2 = 2), b = list(b1 = 3, b2 = 4))))
    Output
      {"a":{"a1":[1],"a2":[2]},"b":{"b1":[3],"b2":[4]}}

---

    Code
      cat(tojson$write_str(list(a = list(1, a2 = 2), list(b1 = 3, 4))))
    Output
      {"a":{"1":[1],"a2":[2]},"2":{"b1":[3],"2":[4]}}

# auto-unbox

    Code
      tojson$write_str(list(1), opts = list(auto_unbox = TRUE))
    Output
      [1] "[1]"
    Code
      tojson$write_str(1L, opts = list(auto_unbox = TRUE))
    Output
      [1] "1"
    Code
      tojson$write_str(1, opts = list(auto_unbox = TRUE))
    Output
      [1] "1"
    Code
      tojson$write_str("foo", opts = list(auto_unbox = TRUE))
    Output
      [1] "\"foo\""
    Code
      tojson$write_str(TRUE, opts = list(auto_unbox = TRUE))
    Output
      [1] "true"

---

    Code
      tojson$write_str(NA_integer_, opts = list(auto_unbox = TRUE))
    Output
      [1] "null"
    Code
      tojson$write_str(NA_real_, opts = list(auto_unbox = TRUE))
    Output
      [1] "null"
    Code
      tojson$write_str(NA_character_, opts = list(auto_unbox = TRUE))
    Output
      [1] "null"
    Code
      tojson$write_str(NA, opts = list(auto_unbox = TRUE))
    Output
      [1] "null"

---

    Code
      cat(tojson$write_str(list(list(1, 2, 3), list(4, 5, 6), list(7, 8, 9)), opts = list(
        auto_unbox = TRUE)))
    Output
      [[1,2,3],[4,5,6],[7,8,9]]

---

    Code
      cat(tojson$write_str(list(a = 1, b = 2), opts = list(auto_unbox = TRUE)))
    Output
      {"a":1,"b":2}

---

    Code
      cat(tojson$write_str(list(a = list(a1 = 1, a2 = 2), b = list(b1 = 3, b2 = 4)),
      opts = list(auto_unbox = TRUE)))
    Output
      {"a":{"a1":1,"a2":2},"b":{"b1":3,"b2":4}}

---

    Code
      cat(tojson$write_str(list(list(1, 2:3, 4), list(4:5, 6, 7:8), list(9:10, 11, 12)),
      opts = list(auto_unbox = TRUE)))
    Output
      [[1,[2,3],4],[[4,5],6,[7,8]],[[9,10],11,12]]

---

    Code
      cat(tojson$write_str(list(a = 1:3, b = 4), opts = list(auto_unbox = TRUE)))
    Output
      {"a":[1,2,3],"b":4}

---

    Code
      cat(tojson$write_str(list(a = list(a1 = 1, a2 = 2:3), b = list(b1 = 4:6, b2 = 7)),
      opts = list(auto_unbox = TRUE)))
    Output
      {"a":{"a1":1,"a2":[2,3]},"b":{"b1":[4,5,6],"b2":7}}

# unbox

    Code
      tojson$write_str(list(tojson$unbox(1)))
    Output
      [1] "[1]"
    Code
      tojson$write_str(tojson$unbox(1L))
    Output
      [1] "1"
    Code
      tojson$write_str(tojson$unbox(1))
    Output
      [1] "1"
    Code
      tojson$write_str(tojson$unbox("foo"))
    Output
      [1] "\"foo\""
    Code
      tojson$write_str(tojson$unbox(TRUE))
    Output
      [1] "true"

---

    Code
      tojson$write_str(tojson$unbox(NA_integer_))
    Output
      [1] "null"
    Code
      tojson$write_str(tojson$unbox(NA_real_))
    Output
      [1] "null"
    Code
      tojson$write_str(tojson$unbox(NA_character_))
    Output
      [1] "null"
    Code
      tojson$write_str(tojson$unbox(NA))
    Output
      [1] "null"

---

    Code
      cat(tojson$write_str(list(list(1, tojson$unbox(2), 3), list(tojson$unbox(4), 5,
      6), list(7, 8, 9))))
    Output
      [[[1],2,[3]],[4,[5],[6]],[[7],[8],[9]]]

---

    Code
      cat(tojson$write_str(list(a = 1, b = tojson$unbox(2))))
    Output
      {"a":[1],"b":2}

---

    Code
      cat(tojson$write_str(list(a = list(a1 = tojson$unbox(1), a2 = 2), b = list(b1 = 3,
        b2 = tojson$unbox(4)))))
    Output
      {"a":{"a1":1,"a2":[2]},"b":{"b1":[3],"b2":4}}

---

    Code
      tojson$unbox(list())
    Condition
      Error:
      ! ! Can only unbox atomic scalar, not list.
    Code
      tojson$unbox(1:2)
    Condition
      Error:
      ! ! Cannot unbox vector of length 2.
    Code
      tojson$unbox(double(2))
    Condition
      Error:
      ! ! Cannot unbox vector of length 2.
    Code
      tojson$unbox(character(2))
    Condition
      Error:
      ! ! Cannot unbox vector of length 2.
    Code
      tojson$unbox(logical(2))
    Condition
      Error:
      ! ! Cannot unbox vector of length 2.

# pretty

    Code
      tojson$write_str(1:5, opts = list(pretty = TRUE))
    Output
      [1] "[1, 2, 3, 4, 5]"
    Code
      tojson$write_str(1:5 / 2, opts = list(pretty = TRUE))
    Output
      [1] "[0.5, 1, 1.5, 2, 2.5]"
    Code
      tojson$write_str(letters[1:5], opts = list(pretty = TRUE))
    Output
      [1] "[\"a\", \"b\", \"c\", \"d\", \"e\"]"
    Code
      tojson$write_str(1:5 %% 2 == 0, opts = list(pretty = TRUE))
    Output
      [1] "[false, true, false, true, false]"

---

    Code
      cat(tojson$write_str(list(list(1, tojson$unbox(2), 3), list(tojson$unbox(4), 5,
      6), list(7, 8, 9)), opts = list(pretty = TRUE)))
    Output
      [
        [
          [1],
          2,
          [3]
        ],
        [
          4,
          [5],
          [6]
        ],
        [
          [7],
          [8],
          [9]
        ]
      ]

---

    Code
      cat(tojson$write_str(list(a = 1, b = tojson$unbox(2)), opts = list(pretty = TRUE)))
    Output
      {
        "a": [1],
        "b": 2
      }

---

    Code
      cat(tojson$write_str(list(a = list(a1 = tojson$unbox(1), a2 = 2), b = list(b1 = 3,
        b2 = tojson$unbox(4))), opts = list(pretty = TRUE)))
    Output
      {
        "a": {
          "a1": 1,
          "a2": [2]
        },
        "b": {
          "b1": [3],
          "b2": 4
        }
      }

# data frames

    Code
      tojson$write_str(df)
    Output
      [1] "[{\"mpg\":21,\"cyl\":6,\"disp\":160},{\"mpg\":21,\"cyl\":6,\"disp\":160},{\"mpg\":22.8,\"cyl\":4,\"disp\":108}]"
    Code
      cat(tojson$write_str(df, opts = list(pretty = TRUE)))
    Output
      [
        {
          "mpg": 21,
          "cyl": 6,
          "disp": 160
        },
        {
          "mpg": 21,
          "cyl": 6,
          "disp": 160
        },
        {
          "mpg": 22.8,
          "cyl": 4,
          "disp": 108
        }
      ]

---

    Code
      tojson$write_str(df)
    Output
      [1] "[{\"cyl\":6,\"disp\":160},{\"mpg\":21,\"cyl\":6},{\"mpg\":22.8,\"cyl\":4,\"disp\":108}]"
    Code
      cat(tojson$write_str(df, opts = list(pretty = TRUE)))
    Output
      [
        {
          "cyl": 6,
          "disp": 160
        },
        {
          "mpg": 21,
          "cyl": 6
        },
        {
          "mpg": 22.8,
          "cyl": 4,
          "disp": 108
        }
      ]

---

    Code
      tojson$write_str(df)
    Output
      [1] "[{\"cyl\":6,\"disp\":160,\"list\":[1,2]},{\"mpg\":21,\"cyl\":6,\"list\":[5]},{\"mpg\":22.8,\"cyl\":4,\"disp\":108,\"list\":[\"a\",\"b\",\"c\"]}]"
    Code
      cat(tojson$write_str(df, opts = list(pretty = TRUE)))
    Output
      [
        {
          "cyl": 6,
          "disp": 160,
          "list": [1, 2]
        },
        {
          "mpg": 21,
          "cyl": 6,
          "list": [5]
        },
        {
          "mpg": 22.8,
          "cyl": 4,
          "disp": 108,
          "list": ["a", "b", "c"]
        }
      ]

# handle_event, process still running

    Code
      state <- start_task_build(state, task("build", pkgidx = 1))
    Message
      i Building R6 2.2.2

# handle_event, build process finished

    Code
      state <- start_task_build(state, task("build", pkgidx = 1))
    Message
      i Building R6 2.2.2

---

    Code
      repeat {
        events <- poll_workers(state)
        state <- handle_events(state, events)
        if (all(state$plan$build_done)) break
      }
    Message
      v Built R6 2.2.2

# handle event, build process finished, but failed

    Code
      state <- make_start_state(plan, list(foo = "bar"))

---

    Code
      state <- start_task_install(state, task("install", pkgidx = 1))

---

    Code
      repeat {
        events <- poll_workers(state)
        state <- handle_events(state, events)
        if (all(state$plan$build_done)) break
      }
    Message
      x Failed to install R6 2.2.2
    Condition
      Error:
      ! ! Failed to install binary package R6.

# handle_event, install process finished

    Code
      repeat {
        events <- poll_workers(state)
        state <- handle_events(state, events)
        if (done) {
          break
        }
        if (!proc$is_alive()) done <- TRUE
      }
    Message
      v Installed R6 2.2.2 

# handle event, install process finished, but failed

    Code
      done <- FALSE
      repeat {
        events <- poll_workers(state)
        state <- handle_events(state, events)
        if (done) {
          break
        }
        if (!proc$is_alive()) done <- TRUE
      }
    Message
      x Failed to install R6 2.2.2
    Condition
      Error:
      ! ! Failed to install binary package R6.

# start_task

    Code
      start_task(list(), task("foobar"))
    Condition
      Error:
      ! ! Unknown task: "foobar".
      i This is an internal error in pkgdepends, please report an issue at <https://github.com/r-lib/pkgdepends/issues>.

# stop_task

    Code
      stop_task(list(), list(task = task("foobar")))
    Condition
      Error:
      ! ! Unknown task: "foobar".
      i This is an internal error in pkgdepends, please report an issue at <https://github.com/r-lib/pkgdepends/issues>.

# deadlock detection

    Code
      install_package_plan(plan, lib = tempfile())
    Condition
      Error:
      ! ! Cannot select new package installation task.
      i 3 packages still waiting to install: p1, p2, and p3.
      i This is an internal error in pkgdepends, please report an issue at <https://github.com/r-lib/pkgdepends/issues>.

# install_args are passed

    Code
      install_package_plan(plan, lib = lib)
    Message
      i Building foo 
      v Built foo 
      v Installed foo  (local)
      v Summary:

# installed_note

    Code
      installed_note(list(type = "cran"))
    Output
      [1] ""
    Code
      installed_note(list(type = "bioc"))
    Output
      [1] "(Bioconductor)"
    Code
      installed_note(list(type = "standard"))
    Output
      [1] ""
    Code
      installed_note(list(type = "local"))
    Output
      [1] "(local)"
    Code
      installed_note(list(type = "github", metadata = list(list(RemoteUsername = "r-lib",
        RemoteRepo = "pak", RemoteSha = "5a4da54df42528545af8a64e83112be21273907c6dfa0f31a0982ca88db6527d"))))
    Output
      [1] "(github::r-lib/pak@5a4da54)"

# pkgplan_i_lp_dependencies

    Code
      lp$conds[[3]]
    Output
      $vars
      [1]  7 13 28
      
      $coef
      [1]  1 -1 -1
      
      $op
      [1] "<="
      
      $rhs
      [1] 0
      
      $type
      [1] "dependency"
      
      $note
      $note$wh
      [1] 7
      
      $note$ref
      [1] "rlang"
      
      $note$cand
      [1] 13 28
      
      $note$good_cand
      [1] 13 28
      
      $note$txt
      [1] "hms depends on rlang: version rlang 0.4.0, rlang 0.4.0"
      
      $note$depop
      [1] ""
      
      $note$depver
      [1] ""
      
      

# windows_archs

    Code
      windows_archs()
    Output
      [1] "prefer-x64" "both"      

# env_decode_difftime

    Code
      env_decode_difftime("0s")
    Output
      Time difference of 0 secs
    Code
      env_decode_difftime("120s")
    Output
      Time difference of 120 secs
    Code
      env_decode_difftime("0m")
    Output
      Time difference of 0 mins
    Code
      env_decode_difftime("120m")
    Output
      Time difference of 120 mins
    Code
      env_decode_difftime("0h")
    Output
      Time difference of 0 hours
    Code
      env_decode_difftime("24h")
    Output
      Time difference of 24 hours
    Code
      env_decode_difftime("0d")
    Output
      Time difference of 0 days
    Code
      env_decode_difftime("1d")
    Output
      Time difference of 1 days

---

    Code
      env_decode_difftime("", "PKG_UPDATE")
    Condition
      Error:
      ! ! Invalid time interval specification in `PKG_UPDATE` environment variable: ""
      i It must have the form `<number><unit>`.
      i The unit must be a single letter: `s` (seconds), `m` (minutes), `h` (hours) or `d` (days).
      i Examples: 60s, 2h, 1d.

---

    Code
      env_decode_difftime("123", "PKG_UPDATE")
    Condition
      Error:
      ! ! Invalid time interval specification in `PKG_UPDATE` environment variable: "123"
      i It must have the form `<number><unit>`.
      i The unit must be a single letter: `s` (seconds), `m` (minutes), `h` (hours) or `d` (days).
      i Examples: 60s, 2h, 1d.

---

    Code
      env_decode_difftime("1k", "PKG_UPDATE")
    Condition
      Error:
      ! ! Invalid time interval specification in `PKG_UPDATE` environment variable: "1k"
      i It must have the form `<number><unit>`.
      i The unit must be a single letter: `s` (seconds), `m` (minutes), `h` (hours) or `d` (days).
      i Examples: 60s, 2h, 1d.

---

    Code
      env_decode_difftime("k1k", "PKG_UPDATE")
    Condition
      Error:
      ! ! Invalid time interval specification in `PKG_UPDATE` environment variable: "k1k"
      i It must have the form `<number><unit>`.
      i The unit must be a single letter: `s` (seconds), `m` (minutes), `h` (hours) or `d` (days).
      i Examples: 60s, 2h, 1d.

# current_config

    Code
      sort(current_config()$list())
    Output
       [1] "build_vignettes"           "cache_dir"                
       [3] "cran_mirror"               "dependencies"             
       [5] "git_submodules"            "goal"                     
       [7] "include_linkingto"         "library"                  
       [9] "metadata_cache_dir"        "metadata_update_after"    
      [11] "package_cache_dir"         "platforms"                
      [13] "r_versions"                "sysreqs"                  
      [15] "sysreqs_db_update"         "sysreqs_db_update_timeout"
      [17] "sysreqs_dry_run"           "sysreqs_lookup_system"    
      [19] "sysreqs_platform"          "sysreqs_rspm_repo_id"     
      [21] "sysreqs_rspm_url"          "sysreqs_sudo"             
      [23] "sysreqs_update"            "sysreqs_verbose"          
      [25] "use_bioconductor"          "windows_archs"            

# scan_deps errors

    Code
      scan_deps(1:10)
    Condition
      Error:
      ! ! `path` must be a character vector without `NA`, but it is an integer vector
    Code
      scan_deps(root = mtcars)
    Condition
      Error:
      ! ! `root` must be a string (character scalar), but it is a data frame.
    Code
      scan_deps(root = "sdfssdfsdf")
    Condition
      Error:
      ! ! Project root 'sdfssdfsdf' does not exist.
    Code
      scan_deps("sdfssdfsdf")
    Condition
      Error:
      ! ! Path does not exist: 'sdfssdfsdf'.
    Code
      scan_deps(c("sdfssdfsdf-1", "sdfssdfsdf-2"))
    Condition
      Error:
      ! ! Paths do not exist: 'sdfssdfsdf-1' and 'sdfssdfsdf-2'.
    Code
      scan_deps("foo", "bar")
    Condition
      Error:
      ! ! 'foo' is outside of project root 'bar'.
    Code
      scan_deps(c("foo", "foobar"), "bar")
    Condition
      Error:
      ! ! 'foo' and 'foobar' are outside of project root 'bar'.

# get_deps_cache_path

    Code
      writeLines(get_deps_cache_path())
    Output
      <tempdir>/<tempfile>/R/pkgcache/deps/2
    Code
      writeLines(get_deps_cache_path("badcafe"))
    Output
      <tempdir>/<tempfile>/R/pkgcache/deps/2/ba/badcafe

# clear_deps_cache

    Code
      dir(tmp, recursive = TRUE)
    Output
      [1] "R/pkgcache/deps/2/ba/badcafe"

---

    Code
      dir(tmp, recursive = TRUE)
    Output
      character(0)

# re_r_dep

    Code
      re_r_dep()
    Output
      [1] "library|require|loadNamespace|::|setClass|setGeneric|pkg_attach|p_load|module|import|box::|tar_option_set|glue|ggsave|set_engine|opts_chunk|geom_hex|JunitReporter|geom_hex|JunitReporter"

# scan_deps_df

    Code
      scan_deps_df()
    Output
      # A data frame: 0 x 9
      # i 9 variables: path <chr>, ref <chr>, package <chr>, version <chr>,
      #   type <chr>, code <chr>, start_row <int>, start_column <int>,
      #   start_byte <int>

# scan_path_deps_do

    Code
      scan_path_deps_do(readLines(rfile), basename(rfile))
    Output
      # A data frame: 3 x 9
        path   ref   package version type  code      start_row start_column start_byte
        <chr>  <chr> <chr>   <chr>   <chr> <chr>         <int>        <int>      <int>
      1 code.R CD    CD      *       prod  CD::pkg           4            1         26
      2 code.R AB    AB      *       prod  library(~         1            1          1
      3 code.R BC    BC      *       prod  require(~         2            1         13
    Code
      scan_path_deps_do(readLines(nsfile), basename(nsfile))
    Output
      # A data frame: 2 x 9
        path      ref   package version type  code  start_row start_column start_byte
        <chr>     <chr> <chr>   <chr>   <chr> <chr>     <int>        <int>      <int>
      1 NAMESPACE stats stats   *       prod  stats         1            1          1
      2 NAMESPACE utils utils   *       prod  utils         1            1          1

---

    Code
      scan_path_deps_do("code", "foo.unknown")
    Condition
      Error:
      ! ! Cannot parse .unknown file for dependencies, internal error

# scan_path_deps_do_r

    Code
      scan_path_deps_do_r(readLines(rfile), rfile)
    Output
      # A data frame: 3 x 9
        path                             ref   package version type  code        start_row start_column start_byte
        <chr>                            <chr> <chr>   <chr>   <chr> <chr>           <int>        <int>      <int>
      1 fixtures/scan/project-1/R/code.R CD    CD      *       prod  CD::pkg             4            1         26
      2 fixtures/scan/project-1/R/code.R AB    AB      *       prod  library(AB)         1            1          1
      3 fixtures/scan/project-1/R/code.R BC    BC      *       prod  require(BC)         2            1         13

# scan_path_deps_do_fn_hits

    Code
      scan_path_deps_do_r(readLines(rfile), rfile)
    Output
      # A data frame: 2 x 9
        path                    ref     package version type  code                                                               start_row start_column start_byte
        <chr>                   <chr>   <chr>   <chr>   <chr> <chr>                                                                  <int>        <int>      <int>
      1 fixtures/scan/methods.R methods methods *       prod  "setClass(\"track\", slots = c(x = \"numeric\", y = \"numeric\"))"         2           10         43
      2 fixtures/scan/methods.R methods methods *       prod  "setGeneric(\"plot\")"                                                     6            1        175

# scan_path_deps_do_jr_hits

    Code
      scan_path_deps_do_r(readLines(rfile), rfile)
    Output
      # A data frame: 6 x 9
        path                  ref      package  version type  code                          start_row start_column start_byte
        <chr>                 <chr>    <chr>    <chr>   <chr> <chr>                             <int>        <int>      <int>
      1 fixtures/scan/junit.R testthat testthat *       prod  testthat::JunitReporter               1            8          8
      2 fixtures/scan/junit.R testthat testthat *       prod  library(testthat)                     3            1         39
      3 fixtures/scan/junit.R xml2     xml2     *       prod  testthat::JunitReporter$new()         1            8          8
      4 fixtures/scan/junit.R xml2     xml2     *       prod  JunitReporter$new()                   5            9         66
      5 fixtures/scan/junit.R xml2     xml2     *       prod  JunitReporter                         1           18         18
      6 fixtures/scan/junit.R xml2     xml2     *       prod  JunitReporter                         5            9         66

# scan_pat_deps_do_ragg_hits

    Code
      scan_path_deps_do_rmd(readLines(rfile), rfile)
    Output
      # A data frame: 4 x 9
        path                    ref       package   version type  code                                        start_row start_column start_byte
        <chr>                   <chr>     <chr>     <chr>   <chr> <chr>                                           <int>        <int>      <int>
      1 fixtures/scan/knitr.Rmd knitr     knitr     *       prod  "knitr::opts_chunk"                                 3            1          9
      2 fixtures/scan/knitr.Rmd knitr     knitr     *       prod  "knitr::opts_chunk"                                 7            1         61
      3 fixtures/scan/knitr.Rmd ragg      ragg      *       prod  "knitr::opts_chunk$set(dev = \"ragg_png\")"         3            1          9
      4 fixtures/scan/knitr.Rmd rmarkdown rmarkdown *       prod  "r"                                                 2            5          6

---

    Code
      scan_path_deps_do_rmd(readLines(rfile), rfile)
    Output
      # A data frame: 2 x 9
        path                     ref       package   version type  code              start_row start_column start_byte
        <chr>                    <chr>     <chr>     <chr>   <chr> <chr>                 <int>        <int>      <int>
      1 fixtures/scan/noragg.Rmd knitr     knitr     *       prod  knitr::opts_chunk         2            1          8
      2 fixtures/scan/noragg.Rmd rmarkdown rmarkdown *       prod  r                         1            5          5

# safe_parse_pkg_from_call

    Code
      safe_parse_pkg_from_call(NA_character_, "library", "library(qwe)")
    Output
      [1] "qwe"

# parse_pkg_from_call

    Code
      parse_pkg_from_call(NA_character_, "library", "library(qwe)")
    Output
      [1] "qwe"
    Code
      parse_pkg_from_call("base", "loadNamespace", "loadNamespace('q1')")
    Output
      [1] "q1"
    Code
      parse_pkg_from_call("base", "requireNamespace", "requireNamespace('q1')")
    Output
      [1] "q1"
    Code
      parse_pkg_from_call(NA_character_, "pkg_attach", "pkg_attach('foobar')")
    Output
      [1] "foobar"
    Code
      parse_pkg_from_call(NA_character_, "pkg_attach2", "pkg_attach2('foobar')")
    Output
      [1] "foobar"
    Code
      parse_pkg_from_call("pacman", "p_load", "p_load('p1')")
    Output
      [1] "p1"
    Code
      parse_pkg_from_call(NA_character_, "import", "import(x1)")
    Output
      [1] "x1"
    Code
      parse_pkg_from_call(NA_character_, "module", "module({import('x2')})")
    Output
      [1] "x2"
    Code
      parse_pkg_from_call("import", "from", "import::from(dplyr)")
    Output
      [1] "dplyr"
    Code
      parse_pkg_from_call("import", "into",
        "import::into('operators', .from = 'dplyr')")
    Output
      [1] "dplyr"
    Code
      parse_pkg_from_call("import", "here", "import::here('dplyr')")
    Output
      [1] "dplyr"
    Code
      parse_pkg_from_call("box", "use", "box::use(dplyr[filter, select])")
    Output
      [1] "dplyr"
    Code
      parse_pkg_from_call("targets", "tar_option_set",
        "tar_option_set(packages = c('p1', 'p2'))")
    Output
      [1] "p1" "p2"
    Code
      parse_pkg_from_call("glue", "glue", "glue::glue('blah {library(x5)} blah')")
    Output
      [1] "x5"
    Code
      parse_pkg_from_call(NA_character_, "ggsave", "ggsave(filename = 'foo.svg')")
    Output
      [1] "svglite"
    Code
      parse_pkg_from_call(NA_character_, "set_engine", "set_engine(engine = 'spark')")
    Output
      [1] "sparklyr"
    Code
      parse_pkg_from_call("R6", "R6Class",
        "R6::R6Class('foobar', inherit = JunitReporter)")
    Output
      [1] "xml2"
    Code
      parse_pkg_from_call("testthat", "test_dir",
        "testthat::test_dir(reporter = 'junit')")
    Output
      [1] "xml2"

# parse_pkg_from_call_library

    Code
      ppcl("library", "library(qqq)")
    Output
      [1] "qqq"
    Code
      ppcl("library", "library('qqq')")
    Output
      [1] "qqq"
    Code
      ppcl("library", "library(qqq)", ns = "base")
    Output
      [1] "qqq"
    Code
      ppcl("require", "require(qqq)")
    Output
      [1] "qqq"
    Code
      ppcl("require", "require('qqq')")
    Output
      [1] "qqq"
    Code
      ppcl("require", "require('qqq')", ns = "base")
    Output
      [1] "qqq"

# dependencies_eval

    Code
      dependencies_eval(quote({
        1:10
        c(10:1)[1:3]
      }))
    Output
      [1] 10  9  8

# scan_path_deps_do_rmd

    Code
      scan_path_deps_do_rmd(readLines(path), "chunk-errors.Rmd")
    Output
      # A data frame: 2 x 9
        path             ref       package   version type  code           start_row start_column start_byte
        <chr>            <chr>     <chr>     <chr>   <chr> <chr>              <int>        <int>      <int>
      1 chunk-errors.Rmd dplyr     dplyr     *       prod  library(dplyr)         8            1        115
      2 chunk-errors.Rmd rmarkdown rmarkdown *       prod  r                      7            5        112

# scan_path_deps_do_rmd #2

    Code
      scan_path_deps_do_rmd(readLines(path), "inline-chunks.Rmd")
    Output
      # A data frame: 3 x 9
        path              ref      package  version type  code             start_row start_column start_byte
        <chr>             <chr>    <chr>    <chr>   <chr> <chr>                <int>        <int>      <int>
      1 inline-chunks.Rmd inline   inline   *       prod  inline::chunks           4           49         68
      2 inline-chunks.Rmd multiple multiple *       prod  multiple::calls          4           92        111
      3 inline-chunks.Rmd separate separate *       prod  separate::chunks         6           12        160

# scan_path_deps_do_rmd #3

    Code
      scan_path_deps_do_rmd(readLines(path), "nothing.Rmd")
    Output
      # A data frame: 1 x 9
        path        ref       package   version type  code  start_row start_column start_byte
        <chr>       <chr>     <chr>     <chr>   <chr> <chr>     <int>        <int>      <int>
      1 nothing.Rmd rmarkdown rmarkdown *       prod  r             7            5         90

# scan_path_deps_do_header_hits

    Code
      scan_path_deps_do_rmd(readLines(path), basename(path))
    Output
      # A data frame: 2 x 9
        path       ref   package version type  code        start_row start_column start_byte
        <chr>      <chr> <chr>   <chr>   <chr> <chr>           <int>        <int>      <int>
      1 header.Rmd p1    p1      *       prod  p1::fun             4           14         32
      2 header.Rmd p2    p2      *       prod  library(p2)         7           14         81

# scan_path_deps_do_header_shiny_hits

    Code
      scan_path_deps_do_rmd(readLines(path), basename(path))
    Output
      # A data frame: 4 x 9
        path             ref   package version type  code         start_row start_column start_byte
        <chr>            <chr> <chr>   <chr>   <chr> <chr>            <int>        <int>      <int>
      1 header-shiny.Rmd shiny shiny   *       prod  "shiny"              4           11         26
      2 header-shiny.Rmd shiny shiny   *       prod  "'shiny'"            5            9         40
      3 header-shiny.Rmd shiny shiny   *       prod  "\"shiny\""          6           11         58
      4 header-shiny.Rmd shiny shiny   *       prod  "|\n  shiny"         7            9         74

---

    Code
      scan_path_deps_do_rmd(readLines(path), basename(path))
    Output
      # A data frame: 4 x 9
        path              ref   package version type  code           start_row start_column start_byte
        <chr>             <chr> <chr>   <chr>   <chr> <chr>              <int>        <int>      <int>
      1 header-shiny2.Rmd shiny shiny   *       prod  "shiny"                5            9         32
      2 header-shiny2.Rmd shiny shiny   *       prod  "'shiny'"              7            9         56
      3 header-shiny2.Rmd shiny shiny   *       prod  "\"shiny\""            9           11         82
      4 header-shiny2.Rmd shiny shiny   *       prod  ">\n    shiny"        11            9        106

# scan_path_deps_do_header_bslib_hits

    Code
      scan_path_deps_do_rmd(readLines(path), basename(path))
    Output
      # A data frame: 1 x 9
        path             ref   package version type  code                                                              start_row start_column start_byte
        <chr>            <chr> <chr>   <chr>   <chr> <chr>                                                                 <int>        <int>      <int>
      1 header-bslib.Rmd bslib bslib   *       prod  "output:\n  html_document:\n    toc: true\n    theme: some theme"         4            1         16

# scan_path_deps_do_dsc

    Code
      print(scan_path_deps_do_dsc(readLines(path), basename(path)), n = Inf)
    Output
      # A data frame: 42 x 9
         path        ref                    package      version      type  code                   start_row start_column start_byte
         <chr>       <chr>                  <chr>        <chr>        <chr> <chr>                      <int>        <int>      <int>
       1 DESCRIPTION callr                  callr        >=3.3.1      prod  callr                          1            1          1
       2 DESCRIPTION r-lib/cli              cli          >=3.6.0      prod  r-lib/cli                      1            1          1
       3 DESCRIPTION curl                   curl         *            prod  curl                           1            1          1
       4 DESCRIPTION desc                   desc         >=1.4.3      prod  desc                           1            1          1
       5 DESCRIPTION filelock               filelock     >=1.0.2      prod  filelock                       1            1          1
       6 DESCRIPTION jsonlite               jsonlite     *            prod  jsonlite                       1            1          1
       7 DESCRIPTION lpSolve                lpSolve      *            prod  lpSolve                        1            1          1
       8 DESCRIPTION pkgbuild               pkgbuild     >=1.0.2      prod  pkgbuild                       1            1          1
       9 DESCRIPTION pkgcache               pkgcache     >=2.2.0      prod  pkgcache                       1            1          1
      10 DESCRIPTION processx               processx     >=3.4.2      prod  processx                       1            1          1
      11 DESCRIPTION ps                     ps           *            prod  ps                             1            1          1
      12 DESCRIPTION R6                     R6           *            prod  R6                             1            1          1
      13 DESCRIPTION zip                    zip          >=2.3.0      prod  zip                            1            1          1
      14 DESCRIPTION asciicast              asciicast    >=2.2.0.9000 test  asciicast                      1            1          1
      15 DESCRIPTION codetools              codetools    *            test  codetools                      1            1          1
      16 DESCRIPTION covr                   covr         *            test  covr                           1            1          1
      17 DESCRIPTION debugme                debugme      *            test  debugme                        1            1          1
      18 DESCRIPTION fansi                  fansi        *            test  fansi                          1            1          1
      19 DESCRIPTION fs                     fs           *            test  fs                             1            1          1
      20 DESCRIPTION gh                     gh           *            test  gh                             1            1          1
      21 DESCRIPTION gitcreds               gitcreds     *            test  gitcreds                       1            1          1
      22 DESCRIPTION glue                   glue         *            test  glue                           1            1          1
      23 DESCRIPTION htmlwidgets            htmlwidgets  *            test  htmlwidgets                    1            1          1
      24 DESCRIPTION mockery                mockery      *            test  mockery                        1            1          1
      25 DESCRIPTION pak                    pak          *            test  pak                            1            1          1
      26 DESCRIPTION pingr                  pingr        >=2.0.0      test  pingr                          1            1          1
      27 DESCRIPTION rmarkdown              rmarkdown    *            test  rmarkdown                      1            1          1
      28 DESCRIPTION rstudioapi             rstudioapi   *            test  rstudioapi                     1            1          1
      29 DESCRIPTION spelling               spelling     *            test  spelling                       1            1          1
      30 DESCRIPTION svglite                svglite      *            test  svglite                        1            1          1
      31 DESCRIPTION testthat               testthat     >=3.2.0      test  testthat                       1            1          1
      32 DESCRIPTION tibble                 tibble       *            test  tibble                         1            1          1
      33 DESCRIPTION webfakes               webfakes     >=1.1.5.9000 test  webfakes                       1            1          1
      34 DESCRIPTION withr                  withr        >=2.1.1      test  withr                          1            1          1
      35 DESCRIPTION gh                     gh           *            dev   gh                             1            1          1
      36 DESCRIPTION pkgsearch              pkgsearch    *            dev   pkgsearch                      1            1          1
      37 DESCRIPTION withr                  withr        >=2.1.1      dev   withr                          1            1          1
      38 DESCRIPTION r-lib/asciicast        asciicast    *            test  r-lib/asciicast                1            1          1
      39 DESCRIPTION covr                   covr         *            test  covr                           1            1          1
      40 DESCRIPTION r-lib/asciicast        asciicast    *            dev   r-lib/asciicast                1            1          1
      41 DESCRIPTION pkgdown                pkgdown      >=2.0.2      dev   pkgdown                        1            1          1
      42 DESCRIPTION tidyverse/tidytemplate tidytemplate *            dev   tidyverse/tidytemplate         1            1          1

# scan_path_deps_do_namespace

    Code
      print(scan_path_deps_do_namespace(readBin(path, "raw", 10000), path), n = Inf)
    Output
      # A data frame: 2 x 9
        path                    ref   package version type  code  start_row start_column start_byte
        <chr>                   <chr> <chr>   <chr>   <chr> <chr>     <int>        <int>      <int>
      1 fixtures/scan/NAMESPACE stats stats   *       prod  stats         1            1          1
      2 fixtures/scan/NAMESPACE utils utils   *       prod  utils         1            1          1

# scan_path_deps_do_{bookdown,pkgdown,quarto,renv_lock,rsconnect}

    Code
      scan_deps()[]
    Output
      # A data frame: 4 x 9
        path                   ref       package   version type  code  start_row start_column start_byte
        <chr>                  <chr>     <chr>     <chr>   <chr> <chr>     <int>        <int>      <int>
      1 bookdown/_bookdown.yml bookdown  bookdown  *       prod  <NA>          1            1          1
      2 _pkgdown.yml           pkgdown   pkgdown   *       prod  <NA>          1            1          1
      3 renv.lock              renv      renv      *       prod  <NA>          1            1          1
      4 rsconnect              rsconnect rsconnect *       prod  <NA>          1            1          1

# .Rproj file

    Code
      scan_deps(project, root = project)[]
    Output
      # A data frame: 2 x 9
        path             ref      package  version type  code                    start_row start_column start_byte
        <chr>            <chr>    <chr>    <chr>   <chr> <chr>                       <int>        <int>      <int>
      1 pkgdepends.Rproj devtools devtools *       prod  PackageUseDevtools: Yes         1            1          1
      2 pkgdepends.Rproj roxygen2 roxygen2 *       prod  PackageUseDevtools: Yes         1            1          1

# scan_path_deps_do_rnw_ranges

    Code
      scan_path_deps_do_rnw_ranges(code)
    Output
      [[1]]
      [1] 11 13
      
      [[2]]
      [1] 15 17
      
      [[3]]
      [1] 19 21
      
      [[4]]
      [1] 22 25
      
      [[5]]
      [1] 27 29
      

# scan_path_deps_do_rnw_parse_chunk_header

    Code
      scan_path_deps_do_rnw_parse_chunk_header("")
    Output
      named list()
    Code
      scan_path_deps_do_rnw_parse_chunk_header("name")
    Output
      $label
      [1] "name"
      
    Code
      scan_path_deps_do_rnw_parse_chunk_header(
        "name, foo = 1, bar = TRUE, this = that")
    Output
      $label
      [1] "name"
      
      $foo
      [1] 1
      
      $bar
      [1] TRUE
      
      $this
      that
      

# .Rnw file

    Code
      scan_path_deps_do_rnw(readLines(path), basename(path))
    Output
      # A data frame: 3 x 9
        path     ref      package  version type  code           start_row start_column start_byte
        <chr>    <chr>    <chr>    <chr>   <chr> <chr>              <int>        <int>      <int>
      1 test.Rnw cli      cli      *       prod  library(cli)           1            1          1
      2 test.Rnw filelock filelock *       prod  filelock::lock         1            1          1
      3 test.Rnw pak      pak      *       prod  library(pak)           2            1         43
    Code
      scan_path_deps_do(readLines(path), basename(path))
    Output
      # A data frame: 3 x 9
        path     ref      package  version type  code           start_row start_column start_byte
        <chr>    <chr>    <chr>    <chr>   <chr> <chr>              <int>        <int>      <int>
      1 test.Rnw cli      cli      *       prod  library(cli)           1            1          1
      2 test.Rnw filelock filelock *       prod  filelock::lock         1            1          1
      3 test.Rnw pak      pak      *       prod  library(pak)           2            1         43

# Ignored chunks in .Rnw file

    Code
      scan_path_deps_do(readBin(path, "raw", file.size(path)), basename(path))
    Output
      # A data frame: 6 x 9
        path            ref           package       version type  code                   start_row start_column start_byte
        <chr>           <chr>         <chr>         <chr>   <chr> <chr>                      <int>        <int>      <int>
      1 ignore-test.Rnw good          good          *       prod  library(good)                  1            1          1
      2 ignore-test.Rnw butthisisgood butthisisgood *       prod  library(butthisisgood)         1            1          1
      3 ignore-test.Rnw butthisisgood butthisisgood *       prod  library(butthisisgood)         1            1          1
      4 ignore-test.Rnw good          good          *       prod  library(good)                  1            1          1
      5 ignore-test.Rnw good          good          *       prod  library(good)                  1            1          1
      6 ignore-test.Rnw good          good          *       prod  library(good)                  1            1          1

# IPython notebook

    Code
      scan_path_deps_do(readLines(path), basename(path))
    Output
      # A data frame: 4 x 9
        path           ref      package  version type  code            start_row start_column start_byte
        <chr>          <chr>    <chr>    <chr>   <chr> <chr>               <int>        <int>      <int>
      1 notebook.ipynb IRkernel IRkernel *       prod  <NA>                    1            1          1
      2 notebook.ipynb MASS     MASS     *       prod  library(MASS)           1            1          1
      3 notebook.ipynb stats    stats    *       prod  stats::setNames         1            1          1
      4 notebook.ipynb cli      cli      *       prod  cli::cli_text           2            1         19

# new_pkg_deps

    Code
      deps
    Output
      <pkg_dependencies>
      + refs:
        - pkg3
      (use `$resolve()` to resolve dependencies)
      (use `$solve()` to solve dependencies)

---

    Code
      deps$get_refs()
    Output
      [1] "pkg3"

---

    Code
      sort(deps$get_config()$list())
    Output
       [1] "build_vignettes"           "cache_dir"                
       [3] "cran_mirror"               "dependencies"             
       [5] "git_submodules"            "goal"                     
       [7] "include_linkingto"         "library"                  
       [9] "metadata_cache_dir"        "metadata_update_after"    
      [11] "package_cache_dir"         "platforms"                
      [13] "r_versions"                "sysreqs"                  
      [15] "sysreqs_db_update"         "sysreqs_db_update_timeout"
      [17] "sysreqs_dry_run"           "sysreqs_lookup_system"    
      [19] "sysreqs_platform"          "sysreqs_rspm_repo_id"     
      [21] "sysreqs_rspm_url"          "sysreqs_sudo"             
      [23] "sysreqs_update"            "sysreqs_verbose"          
      [25] "use_bioconductor"          "windows_archs"            

---

    Code
      deps
    Output
      <pkg_dependencies>
      + refs:
        - pkg3
      + has resolution (+2 dependencies)
      (use `$get_resolution()` to see resolution results)
      (use `$solve()` to solve dependencies)

---

    Code
      deps
    Output
      <pkg_dependencies>
      + refs:
        - pkg3
      + has resolution (+2 dependencies)
      + has solution
      (use `$get_resolution()` to see resolution results)
      (use `$show_solution()` to see the dependencies
      (use `$get_solution()` to see the full solution results)
      (use `$draw()` to draw the dependency tree)

---

    Code
      deps$draw()
    Output
      pkg3 1.0.0 [new][bld][dl] (<size>)
      \-pkg2 1.0.0 [new][bld][dl] (<size>)
        \-pkg1 1.0.0 [new][bld][dl] (<size>)
      
      Key:  [new] new | [dl] download | [bld] build

---

    Code
      deps$get_solution()
    Output
      <pkg_solution>
      + result: OK
      + refs:
        - pkg3
      + constraints (5):
        - select pkg3 exactly once
        - select pkg1 at most once
        - select pkg2 at most once
        - pkg2 depends on pkg1: version pkg1 1.0.0
        - pkg3 depends on pkg2: version pkg2 1.0.0
      + solution:
        - pkg1
        - pkg2
        - pkg3

# async

    Code
      synchronize(deps$async_resolve()$then(function() "done"))
    Output
      [1] "done"

---

    Code
      deps$get_resolution()[, c("ref", "type", "directpkg", "package", "error")]
    Output
      # A data frame: 3 x 5
        ref   type     directpkg package error     
        <chr> <chr>    <lgl>     <chr>   <list>    
      1 pkg1  standard FALSE     pkg1    <list [0]>
      2 pkg2  standard FALSE     pkg2    <list [0]>
      3 pkg3  standard TRUE      pkg3    <list [0]>

# solve policy

    Code
      deps
    Output
      <pkg_dependencies>
      + refs:
        - pkg3
      + has resolution (+2 dependencies)
      + has solution
      (use `$get_resolution()` to see resolution results)
      (use `$show_solution()` to see the dependencies
      (use `$get_solution()` to see the full solution results)
      (use `$draw()` to draw the dependency tree)

# errors

    Code
      deps$stop_for_solution_error()
    Condition
      Error:
      ! ! Could not solve package dependencies:
      * needsfuturama: Can't install dependency futurama
      * futurama: Needs R >= 3000.0

# parse_submodules

    Code
      parse_submodules(sm)
    Output
                              submodule                            path
      1 dependencies/fast_double_parser dependencies/fast_double_parser
      2                dependencies/fmt                dependencies/fmt
      3         dependencies/boost_math         dependencies/boost_math
      4              dependencies/eigen              dependencies/eigen
                                                  url branch
      1 https://github.com/lemire/fast_double_parser/   <NA>
      2                https://github.com/fmtlib/fmt/   <NA>
      3              https://github.com/boostorg/math   <NA>
      4             https://gitlab.com/libeigen/eigen   <NA>
    Code
      parse_submodules(read_char(sm))
    Output
                              submodule                            path
      1 dependencies/fast_double_parser dependencies/fast_double_parser
      2                dependencies/fmt                dependencies/fmt
      3         dependencies/boost_math         dependencies/boost_math
      4              dependencies/eigen              dependencies/eigen
                                                  url branch
      1 https://github.com/lemire/fast_double_parser/   <NA>
      2                https://github.com/fmtlib/fmt/   <NA>
      3              https://github.com/boostorg/math   <NA>
      4             https://gitlab.com/libeigen/eigen   <NA>

---

    Code
      parse_submodules(tmp)
    Output
      list()

---

    Code
      parse_submodules(sm2)
    Condition
      Warning:
      Invalid submodule definition, skipping submodule installation
    Output
      list()
    Code
      parse_submodules(sm3)
    Condition
      Warning:
      Invalid submodule definition, skipping submodule installation
    Output
      list()

# git_download_repo with submodules

    Code
      dir(tmp, recursive = TRUE)
    Output
      [1] "v1/DESCRIPTION"   "v1/NAMESPACE"     "v1/R/foo.R"       "v1/README.md"    
      [5] "v1/submod/README" "v1/wipe.R"       

---

    Code
      readLines(file.path(output, "submod", "README"))
    Output
      [1] "A git submodule" "Another commit" 

---

    Code
      dir(tmp, recursive = TRUE)
    Output
      [1] "v1/DESCRIPTION"   "v1/NAMESPACE"     "v1/R/foo.R"       "v1/README.md"    
      [5] "v1/submod/README" "v1/wipe.R"       

---

    Code
      readLines(file.path(output, "submod", "README"))
    Output
      [1] "A git submodule" "Another commit"  "Third commit"   

# git_download_repo R package with submodules

    Code
      dir(tmp, recursive = TRUE)
    Output
      [1] "v1/DESCRIPTION"   "v1/NAMESPACE"     "v1/R/foo.R"       "v1/README.md"    
      [5] "v1/submod/README" "v1/wipe.R"       

---

    Code
      readLines(file.path(output, "submod", "README"))
    Output
      [1] "A git submodule" "Another commit" 

---

    Code
      dir(tmp, recursive = TRUE)
    Output
      [1] "v1/DESCRIPTION"   "v1/NAMESPACE"     "v1/R/foo.R"       "v1/README.md"    
      [5] "v1/submod/README" "v1/wipe.R"       

---

    Code
      readLines(file.path(output, "submod", "README"))
    Output
      [1] "A git submodule" "Another commit"  "Third commit"   

---

    Code
      dir(tmp, recursive = TRUE)
    Output
      [1] "v1/DESCRIPTION" "v1/NAMESPACE"   "v1/R/foo.R"     "v1/README.md"  
      [5] "v1/wipe.R"     

---

    Code
      dir(tmp, recursive = TRUE)
    Output
      [1] "v1/DESCRIPTION" "v1/NAMESPACE"   "v1/R/foo.R"     "v1/README.md"  
      [5] "v1/wipe.R"     

# git_download_repo R package with ignored submodule

    Code
      dir(tmp, recursive = TRUE, all.files = TRUE, no.. = TRUE)
    Output
       [1] "v1/.Rbuildignore"                        
       [2] "v1/.github/actions/parameters/action.yml"
       [3] "v1/.github/workflows/inputtest.yml"      
       [4] "v1/.gitignore"                           
       [5] "v1/.gitmodules"                          
       [6] "v1/DESCRIPTION"                          
       [7] "v1/NAMESPACE"                            
       [8] "v1/R/foo.R"                              
       [9] "v1/README.md"                            
      [10] "v1/wipe.R"                               

# directories

    Code
      directories("a")
    Output
      character(0)
    Code
      directories("a/b/c/d")
    Output
      [1] "a"     "a/b"   "a/b/c"

# parse_remote_deps

    Code
      parse_remote_deps("deps::/foo/bar")
    Output
      [[1]]
      [[1]]$path
      [1] "/foo/bar"
      
      [[1]]$ref
      [1] "deps::/foo/bar"
      
      [[1]]$type
      [1] "deps"
      
      

---

    Code
      parse_remote_deps("deps::foo/bar")
    Output
      [[1]]
      [[1]]$path
      [1] "foo/bar"
      
      [[1]]$ref
      [1] "deps::foo/bar"
      
      [[1]]$type
      [1] "deps"
      
      

---

    Code
      parse_remote_deps("deps::~foo/bar")
    Output
      [[1]]
      [[1]]$path
      [1] "~foo/bar"
      
      [[1]]$ref
      [1] "deps::~foo/bar"
      
      [[1]]$type
      [1] "deps"
      
      

# resolve_remote_deps

    Code
      prop$get_resolution()[, c("ref", "package", "version")]
    Output
      # A data frame: 5 x 3
        ref      package        version
        <chr>    <chr>          <chr>  
      1 deps::.  mypackage-deps 1.0.0  
      2 local::. mypackage      1.0.0  
      3 pkg1     pkg1           1.0.0  
      4 pkg2     pkg2           1.0.0  
      5 pkg3     pkg3           1.0.0  

---

    Code
      prop$get_solution()
    Output
      <pkg_solution>
      + result: OK
      + refs:
        - deps::.
      + constraints (11):
        - select mypackage-deps exactly once
        - select mypackage at most once
        - select pkg1 at most once
        - select pkg2 at most once
        - select pkg3 at most once
        - deps::. depends on pkg2: version pkg2 1.0.0
        - deps::. depends on pkg3: version pkg3 1.0.0
        - local::. depends on pkg2: version pkg2 1.0.0
        - local::. depends on pkg3: version pkg3 1.0.0
        - pkg2 depends on pkg1: version pkg1 1.0.0
        ...
      + solution:
        - deps::.
        - pkg1
        - pkg2
        - pkg3

# match

    Code
      sr1
    Output
      $os
      [1] "linux"
      
      $distribution
      [1] "ubuntu"
      
      $version
      [1] "22.04"
      
      $url
      [1] NA
      
      $pre_install
      [1] "apt-get -y update"
      
      $install_scripts
      [1] "apt-get -y install default-jdk libcurl4-openssl-dev"
      
      $post_install
      [1] "R CMD javareconf"
      
      $packages
      [1] "default-jdk"          "libcurl4-openssl-dev"
      

---

    Code
      sr2
    Output
      $os
      [1] "linux"
      
      $distribution
      [1] "debian"
      
      $version
      [1] "unstable"
      
      $url
      [1] NA
      
      $pre_install
      [1] "apt-get -y update"
      
      $install_scripts
      [1] "apt-get -y install default-jdk libcurl4-openssl-dev"
      
      $post_install
      [1] "R CMD javareconf"
      
      $packages
      [1] "default-jdk"          "libcurl4-openssl-dev"
      

# sysreqs2_command error

    Code
      sysreqs2_command("foobar-2023")
    Condition
      Error:
      ! ! Unknown OS. Don't know how to query or install system packages for foobar-2023.

# do not run update if nothing to do

    Code
      sr1
    Output
      $os
      [1] "linux"
      
      $distribution
      [1] "ubuntu"
      
      $version
      [1] "22.04"
      
      $url
      [1] NA
      
      $pre_install
      character(0)
      
      $install_scripts
      NULL
      
      $post_install
      NULL
      
      $packages
      NULL
      

# sysreqs_platforms

    Code
      sysreqs("ubuntu-20.04")
    Output
      [1] "apt-get -y install libxml2-dev"
    Code
      sysreqs("ubuntu-22.04")
    Output
      [1] "apt-get -y install libxml2-dev"
    Code
      sysreqs("ubuntu-24.04")
    Output
      [1] "apt-get -y install libxml2-dev"
    Code
      sysreqs("debian-11")
    Output
      [1] "apt-get -y install libxml2-dev"
    Code
      sysreqs("debian-12")
    Output
      [1] "apt-get -y install libxml2-dev"
    Code
      sysreqs("debian-unstable")
    Output
      [1] "apt-get -y install libxml2-dev"
    Code
      sysreqs("opensuse-15.4")
    Output
      [1] "zypper --non-interactive install libxml2-devel"
    Code
      sysreqs("opensuse-15.5")
    Output
      [1] "zypper --non-interactive install libxml2-devel"
    Code
      sysreqs("opensuse-15.6")
    Output
      [1] "zypper --non-interactive install libxml2-devel"
    Code
      sysreqs("opensuse-leap-15.4")
    Output
      [1] "zypper --non-interactive install libxml2-devel"
    Code
      sysreqs("opensuse-leap-15.5")
    Output
      [1] "zypper --non-interactive install libxml2-devel"
    Code
      sysreqs("opensuse-leap-15.6")
    Output
      [1] "zypper --non-interactive install libxml2-devel"
    Code
      sysreqs("opensuse-tumbleweed-20250509")
    Output
      [1] "zypper --non-interactive install libxml2-devel"
    Code
      sysreqs("centos-6")
    Output
      [1] "yum install -y libxml2-devel"
    Code
      sysreqs("centos-7")
    Output
      [1] "yum install -y libxml2-devel"
    Code
      sysreqs("centos-8")
    Output
      [1] "yum install -y libxml2-devel"
    Code
      sysreqs("rhel-7")
    Output
      [1] "yum install -y libxml2-devel"
    Code
      sysreqs("rhel-7.9")
    Output
      [1] "yum install -y libxml2-devel"
    Code
      sysreqs("rhel-8")
    Output
      [1] "dnf install -y libxml2-devel"
    Code
      sysreqs("rhel-8.10")
    Output
      [1] "dnf install -y libxml2-devel"
    Code
      sysreqs("rhel-9")
    Output
      [1] "dnf install -y libxml2-devel"
    Code
      sysreqs("rhel-9.6")
    Output
      [1] "dnf install -y libxml2-devel"
    Code
      sysreqs("fedora-39")
    Output
      [1] "dnf install -y libxml2-devel"
    Code
      sysreqs("fedora-40")
    Output
      [1] "dnf install -y libxml2-devel"
    Code
      sysreqs("fedora-42")
    Output
      [1] "dnf install -y libxml2-devel"
    Code
      sysreqs("fedora-42")
    Output
      [1] "dnf install -y libxml2-devel"
    Code
      sysreqs("fedora-43")
    Output
      [1] "dnf install -y libxml2-devel"
    Code
      sysreqs("sles-15.4")
    Output
      [1] "zypper --non-interactive install libxml2-devel"
    Code
      sysreqs("sles-15.5")
    Output
      [1] "zypper --non-interactive install libxml2-devel"
    Code
      sysreqs("sles-15.6")
    Output
      [1] "zypper --non-interactive install libxml2-devel"
    Code
      sysreqs("almalinux-8.10")
    Output
      [1] "dnf install -y libxml2-devel"
    Code
      sysreqs("almalinux-9.6")
    Output
      [1] "dnf install -y libxml2-devel"
    Code
      sysreqs("rocky-8.9")
    Output
      [1] "dnf install -y libxml2-devel"
    Code
      sysreqs("rocky-9.3")
    Output
      [1] "dnf install -y libxml2-devel"
    Code
      sysreqs("alpine-3.19")
    Output
      [1] "apk add --no-cache libxml2-dev"
    Code
      sysreqs("alpine-3.20")
    Output
      [1] "apk add --no-cache libxml2-dev"
    Code
      sysreqs("alpine-3.21")
    Output
      [1] "apk add --no-cache libxml2-dev"
    Code
      sysreqs("alpine-3.edge")
    Output
      [1] "apk add --no-cache libxml2-dev"
    Code
      sysreqs("redhat-8")
    Output
      [1] "dnf install -y libxml2-devel"
    Code
      sysreqs("redhat-8.10")
    Output
      [1] "dnf install -y libxml2-devel"
    Code
      sysreqs("rockylinux-8")
    Output
      [1] "dnf install -y libxml2-devel"
    Code
      sysreqs("rockylinux-8.9")
    Output
      [1] "dnf install -y libxml2-devel"

# canonize_sysreqs_platform

    Code
      canonize_sysreqs_platform("ubuntu-20.04")
    Output
      [1] "ubuntu-20.04"
    Code
      canonize_sysreqs_platform("ubuntu-22.04")
    Output
      [1] "ubuntu-22.04"
    Code
      canonize_sysreqs_platform("ubuntu-24.04")
    Output
      [1] "ubuntu-24.04"
    Code
      canonize_sysreqs_platform("debian-11")
    Output
      [1] "debian-11"
    Code
      canonize_sysreqs_platform("debian-12")
    Output
      [1] "debian-12"
    Code
      canonize_sysreqs_platform("debian-unstable")
    Output
      [1] "debian-unstable"
    Code
      canonize_sysreqs_platform("opensuse-15.4")
    Output
      [1] "opensuse-15.4"
    Code
      canonize_sysreqs_platform("opensuse-15.5")
    Output
      [1] "opensuse-15.5"
    Code
      canonize_sysreqs_platform("opensuse-15.6")
    Output
      [1] "opensuse-15.6"
    Code
      canonize_sysreqs_platform("opensuse-leap-15.4")
    Output
      [1] "opensuse-15.4"
    Code
      canonize_sysreqs_platform("opensuse-leap-15.5")
    Output
      [1] "opensuse-15.5"
    Code
      canonize_sysreqs_platform("opensuse-leap-15.6")
    Output
      [1] "opensuse-15.6"
    Code
      canonize_sysreqs_platform("opensuse-tumbleweed-20250509")
    Output
      [1] "opensuse-20250509"
    Code
      canonize_sysreqs_platform("centos-6")
    Output
      [1] "centos-6"
    Code
      canonize_sysreqs_platform("centos-7")
    Output
      [1] "centos-7"
    Code
      canonize_sysreqs_platform("centos-8")
    Output
      [1] "centos-8"
    Code
      canonize_sysreqs_platform("rhel-7")
    Output
      [1] "redhat-7"
    Code
      canonize_sysreqs_platform("rhel-7.9")
    Output
      [1] "redhat-7"
    Code
      canonize_sysreqs_platform("rhel-8")
    Output
      [1] "redhat-8"
    Code
      canonize_sysreqs_platform("rhel-8.10")
    Output
      [1] "redhat-8"
    Code
      canonize_sysreqs_platform("rhel-9")
    Output
      [1] "redhat-9"
    Code
      canonize_sysreqs_platform("rhel-9.6")
    Output
      [1] "redhat-9"
    Code
      canonize_sysreqs_platform("fedora-39")
    Output
      [1] "fedora-39"
    Code
      canonize_sysreqs_platform("fedora-40")
    Output
      [1] "fedora-40"
    Code
      canonize_sysreqs_platform("fedora-42")
    Output
      [1] "fedora-42"
    Code
      canonize_sysreqs_platform("fedora-42")
    Output
      [1] "fedora-42"
    Code
      canonize_sysreqs_platform("fedora-43")
    Output
      [1] "fedora-43"
    Code
      canonize_sysreqs_platform("sles-15.4")
    Output
      [1] "sle-15.4"
    Code
      canonize_sysreqs_platform("sles-15.5")
    Output
      [1] "sle-15.5"
    Code
      canonize_sysreqs_platform("sles-15.6")
    Output
      [1] "sle-15.6"
    Code
      canonize_sysreqs_platform("almalinux-8.10")
    Output
      [1] "rockylinux-8"
    Code
      canonize_sysreqs_platform("almalinux-9.6")
    Output
      [1] "rockylinux-9"
    Code
      canonize_sysreqs_platform("rocky-8.9")
    Output
      [1] "rockylinux-8"
    Code
      canonize_sysreqs_platform("rocky-9.3")
    Output
      [1] "rockylinux-9"
    Code
      canonize_sysreqs_platform("alpine-3.19")
    Output
      [1] "alpine-3.19"
    Code
      canonize_sysreqs_platform("alpine-3.20")
    Output
      [1] "alpine-3.20"
    Code
      canonize_sysreqs_platform("alpine-3.21")
    Output
      [1] "alpine-3.21"
    Code
      canonize_sysreqs_platform("alpine-3.19.1")
    Output
      [1] "alpine-3.19"
    Code
      canonize_sysreqs_platform("alpine-3.20.2")
    Output
      [1] "alpine-3.20"
    Code
      canonize_sysreqs_platform("alpine-3.21.3")
    Output
      [1] "alpine-3.21"
    Code
      canonize_sysreqs_platform("alpine-3.edge")
    Output
      [1] "alpine-3.edge"
    Code
      canonize_sysreqs_platform("redhat-8")
    Output
      [1] "redhat-8"
    Code
      canonize_sysreqs_platform("redhat-8.10")
    Output
      [1] "redhat-8"
    Code
      canonize_sysreqs_platform("rockylinux-8")
    Output
      [1] "rockylinux-8"
    Code
      canonize_sysreqs_platform("rockylinux-8.9")
    Output
      [1] "rockylinux-8"

# conflicting dependencies

    Code
      dsc
    Output
      * aa: Can't install dependency cran::cc
      * cran::cc: cran::cc conflicts with cc/cc, to be installed

# conflicting dependencies downstream

    Code
      dsc
    Output
      * a0: Can't install dependency bb
      * bb: Can't install dependency cc/cc
      * cc/cc: cc/cc conflicts with cran::cc, to be installed

# conflicting dependencies and installed packages

    Code
      p$get_solution()$failures
    Output
      * cran/dm@1.0.0: Can't install dependency tidyr
      * tidyr: Can't install dependency dplyr (>= 1.0.10)
      * dplyr: Conflicts with cran/dplyr@1.0.9

---

    Code
      p$get_solution()$failures
    Output
      * cran/dm@1.0.0: Can't install dependency tidyr
      * installed::<path>/tidyr: Can't install dependency dplyr (>= 1.0.10)
      * dplyr: Conflicts with cran/dplyr@1.0.9

---

    Code
      p$get_solution()$failures
    Output
      * local::<path>/<pkg>: dependency conflict

---

    Code
      p2$get_solution()$failures
    Output
      * local::<path>/<pkg>: dependency conflict

# R version dependencies

    Code
      p1$stop_for_solution_error()
    Condition
      Error:
      ! ! Could not solve package dependencies:
      * futurama: Needs R >= 3000.0

---

    Code
      p2$stop_for_solution_error()
    Condition
      Error:
      ! ! Could not solve package dependencies:
      * needsfuturama: Can't install dependency futurama
      * futurama: Needs R >= 3000.0

# find_package_root errors

    Code
      find_package_root("file1716b354e464b")
    Condition
      Error:
      ! ! Path does not exist: file1716b354e464b

---

    Code
      find_package_root("/")
    Condition
      Error:
      ! ! Could not find R package in `/` or its parent directories.

# draw_solution_tree [plain]

    Code
      r$draw_solution_tree()
    Output
      dplyr 1.0.10 [new][bld][dl] (<size>)
      +-R6 2.5.1 [new][bld][dl] (<size>)
      +-generics 0.1.3 [new][bld][dl] (<size>)
      +-glue 1.6.2 [new][bld][dl] (<size>)
      +-lifecycle 1.0.3 [new][bld][dl] (<size>)
      | +-cli 3.4.1 [new][bld][dl] (<size>)
      | +-glue
      | \-rlang 1.0.6 [new][bld][dl] (<size>)
      +-magrittr 2.0.3 [new][bld][dl] (<size>)
      +-pillar 1.8.1 [new][bld][dl] (<size>)
      | +-cli
      | +-fansi 1.0.3 [new][bld][dl] (<size>)
      | +-glue
      | +-lifecycle
      | +-rlang
      | +-utf8 1.2.2 [new][bld][dl] (<size>)
      | \-vctrs 0.4.2 [new][bld][dl] (<size>)
      |   +-cli
      |   +-glue
      |   \-rlang
      +-rlang
      +-tibble 3.1.8 [new][bld][dl] (<size>)
      | +-fansi
      | +-lifecycle
      | +-magrittr
      | +-pillar
      | +-pkgconfig 2.0.3 [new][bld][dl] (<size>)
      | +-rlang
      | \-vctrs
      +-tidyselect 1.2.0 [new][bld][dl] (<size>)
      | +-cli
      | +-glue
      | +-lifecycle
      | +-rlang
      | +-vctrs
      | \-withr 2.5.0 [new][bld][dl] (<size>)
      \-vctrs
      
      pkgconfig 2.0.3 [new][bld][dl] (<size>)
      
      Key:  [new] new | [dl] download | [bld] build

# draw_solution_tree [ansi]

    Code
      r$draw_solution_tree()
    Output
      [1m[3m[36mdplyr [39m[90m1.0.10[39m[23m[22m [32m[new][bld][dl] [39m[90m(<size>)[39m
      +-R6 [90m2.5.1[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      +-generics [90m0.1.3[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      +-glue [90m1.6.2[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      +-lifecycle [90m1.0.3[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      | +-cli [90m3.4.1[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      | +-glue
      | \-rlang [90m1.0.6[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      +-magrittr [90m2.0.3[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      +-pillar [90m1.8.1[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      | +-cli
      | +-fansi [90m1.0.3[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      | +-glue
      | +-lifecycle
      | +-rlang
      | +-utf8 [90m1.2.2[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      | \-vctrs [90m0.4.2[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      |   +-cli
      |   +-glue
      |   \-rlang
      +-rlang
      +-tibble [90m3.1.8[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      | +-fansi
      | +-lifecycle
      | +-magrittr
      | +-pillar
      | +-[1m[3m[36mpkgconfig [39m[90m2.0.3[39m[23m[22m [32m[new][bld][dl] [39m[90m(<size>)[39m
      | +-rlang
      | \-vctrs
      +-tidyselect [90m1.2.0[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      | +-cli
      | +-glue
      | +-lifecycle
      | +-rlang
      | +-vctrs
      | \-withr [90m2.5.0[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      \-vctrs
      
      [1m[3m[36mpkgconfig [39m[90m2.0.3[39m[23m[22m [32m[new][bld][dl] [39m[90m(<size>)[39m
      
      Key:  [32m[new][39m new | [32m[dl][39m download | [32m[bld][39m build

# draw_solution_tree [unicode]

    Code
      r$draw_solution_tree()
    Output
      dplyr 1.0.10 [new][bld][dl] (<size>)
      ├─R6 2.5.1 [new][bld][dl] (<size>)
      ├─generics 0.1.3 [new][bld][dl] (<size>)
      ├─glue 1.6.2 [new][bld][dl] (<size>)
      ├─lifecycle 1.0.3 [new][bld][dl] (<size>)
      │ ├─cli 3.4.1 [new][bld][dl] (<size>)
      │ ├─glue
      │ └─rlang 1.0.6 [new][bld][dl] (<size>)
      ├─magrittr 2.0.3 [new][bld][dl] (<size>)
      ├─pillar 1.8.1 [new][bld][dl] (<size>)
      │ ├─cli
      │ ├─fansi 1.0.3 [new][bld][dl] (<size>)
      │ ├─glue
      │ ├─lifecycle
      │ ├─rlang
      │ ├─utf8 1.2.2 [new][bld][dl] (<size>)
      │ └─vctrs 0.4.2 [new][bld][dl] (<size>)
      │   ├─cli
      │   ├─glue
      │   └─rlang
      ├─rlang
      ├─tibble 3.1.8 [new][bld][dl] (<size>)
      │ ├─fansi
      │ ├─lifecycle
      │ ├─magrittr
      │ ├─pillar
      │ ├─pkgconfig 2.0.3 [new][bld][dl] (<size>)
      │ ├─rlang
      │ └─vctrs
      ├─tidyselect 1.2.0 [new][bld][dl] (<size>)
      │ ├─cli
      │ ├─glue
      │ ├─lifecycle
      │ ├─rlang
      │ ├─vctrs
      │ └─withr 2.5.0 [new][bld][dl] (<size>)
      └─vctrs
      
      pkgconfig 2.0.3 [new][bld][dl] (<size>)
      
      Key:  [new] new | [dl] download | [bld] build

# draw_solution_tree [fancy]

    Code
      r$draw_solution_tree()
    Output
      [1m[3m[36mdplyr [39m[90m1.0.10[39m[23m[22m [32m[new][bld][dl] [39m[90m(<size>)[39m
      ├─R6 [90m2.5.1[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      ├─generics [90m0.1.3[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      ├─glue [90m1.6.2[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      ├─lifecycle [90m1.0.3[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      │ ├─cli [90m3.4.1[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      │ ├─glue
      │ └─rlang [90m1.0.6[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      ├─magrittr [90m2.0.3[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      ├─pillar [90m1.8.1[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      │ ├─cli
      │ ├─fansi [90m1.0.3[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      │ ├─glue
      │ ├─lifecycle
      │ ├─rlang
      │ ├─utf8 [90m1.2.2[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      │ └─vctrs [90m0.4.2[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      │   ├─cli
      │   ├─glue
      │   └─rlang
      ├─rlang
      ├─tibble [90m3.1.8[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      │ ├─fansi
      │ ├─lifecycle
      │ ├─magrittr
      │ ├─pillar
      │ ├─[1m[3m[36mpkgconfig [39m[90m2.0.3[39m[23m[22m [32m[new][bld][dl] [39m[90m(<size>)[39m
      │ ├─rlang
      │ └─vctrs
      ├─tidyselect [90m1.2.0[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      │ ├─cli
      │ ├─glue
      │ ├─lifecycle
      │ ├─rlang
      │ ├─vctrs
      │ └─withr [90m2.5.0[39m [32m[new][bld][dl] [39m[90m(<size>)[39m
      └─vctrs
      
      [1m[3m[36mpkgconfig [39m[90m2.0.3[39m[23m[22m [32m[new][bld][dl] [39m[90m(<size>)[39m
      
      Key:  [32m[new][39m new | [32m[dl][39m download | [32m[bld][39m build

# tree from lockfile

    Code
      plan$draw()
    Output
      pkg3 1.0.0 [new][bld][dl] (<size>)
      \-pkg2 1.0.0 [new][bld][dl] (<size>)
        \-pkg1 1.0.0 [new][bld][dl] (<size>)
      
      Key:  [new] new | [dl] download | [bld] build

# update

    Code
      plan2$draw()
    Output
      pkg1 0.9.0 -> 1.0.0 [upd][bld][dl] (<size>)
      
      Key:  [upd] update | [dl] download | [bld] build

---

    Code
      plan3$draw()
    Output
      pkg1 0.9.0 < 1.0.0 [old]
      
      Key:  [old] outdated

# no emoji

    Code
      emoji("rocket")
    Output
      [1] "[upd]"
    Code
      emoji("sparkles")
    Output
      [1] "[new]"
    Code
      emoji("hand")
    Output
      [1] "[old]"
    Code
      emoji("dl")
    Output
      [1] "[dl]"
    Code
      emoji("builder")
    Output
      [1] "[bld]"
    Code
      emoji("wrench")
    Output
      [1] "[cmp]"
    Code
      emoji("pkg")
    Output
      [1] "pkg"
    Code
      emoji("pkgs")
    Output
      [1] "pkgs"
    Code
      emoji("foobar")
    Output
      [1] ""

# emoji

    Code
      emoji("rocket")
    Output
      [1] "🚀"
    Code
      emoji("sparkles")
    Output
      [1] "✨"
    Code
      emoji("hand")
    Output
      [1] "✋"
    Code
      emoji("dl")
    Output
      [1] " ⬇"
    Code
      emoji("builder")
    Output
      [1] "👷"
    Code
      emoji("wrench")
    Output
      [1] "🔧"
    Code
      emoji("pkg")
    Output
      [1] "📦"
    Code
      emoji("pkgs")
    Output
      [1] "📦"
    Code
      emoji("foobar")
    Output
      [1] ""

# emo_builder

    Code
      emo_builder(5)
    Output
      [1] "👷" "👷" "👷" "👷" "👷"

---

    Code
      emo_builder(5)
    Output
      [1] "👷🏻" "👷🏼" "👷🏽" "👷🏾" "👷🏿"

---

    Code
      emo_builder(5)
    Output
      [1] "👷🏽‍♂️" "👷🏾‍♀️" "👷🏾‍♂️" "👷🏿‍♀️" "👷🏿‍♂️"

# local dep needs name

    Code
      p$stop_for_solution_error()
    Condition
      Error:
      ! ! Could not solve package dependencies:
      * local::./pkg1: ! pkgdepends resolution error for local::./pkg1.
      Caused by error:
      ! Cannot determine package name for 1 package: "local::./pkg2".
      i Maybe you need to add a `<packagename>=` prefix?

---

    Code
      p$get_solution()
    Output
      <pkg_solution>
      + result: OK
      + refs:
        - local::./pkg1
      + constraints (3):
        - select pkg1 exactly once
        - select pkg2 at most once
        - local::./pkg1 depends on pkg2=local::./pkg2: version pkg2=local::./pkg2 1.0.0
      + solution:
        - local::./pkg1
        - pkg2=local::./pkg2

# url dep needs name

    Code
      p$stop_for_solution_error()
    Condition
      Error:
      ! ! Could not solve package dependencies:
      * local::./pkg2: ! pkgdepends resolution error for local::./pkg2.
      Caused by error:
      ! Cannot determine package name for 1 package: "url::http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz".
      i Maybe you need to add a `<packagename>=` prefix?

---

    Code
      p$get_solution()
    Output
      <pkg_solution>
      + result: OK
      + refs:
        - local::./pkg2
      + constraints (3):
        - select pkg2 exactly once
        - select pkg1 at most once
        - local::./pkg2 depends on pkg1=url::http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz: version pkg1=url::http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz 0.9.0
      + solution:
        - local::./pkg2
        - pkg1=url::http://127.0.0.1:<port>//src/contrib/Archive/pkg1/pkg1_0.9.0.tar.gz

# async_pkg_name_check

    Code
      sy(async_pkg_name_check(11))
    Condition
      Error:
      ! ! `name` must be a string (character scalar), but it is a number.
    Code
      sy(async_pkg_name_check(c("a", "b")))
    Condition
      Error:
      ! ! `name` must be a string (character scalar), but it is a character vector.
    Code
      sy(async_pkg_name_check(NA_character_))
    Condition
      Error:
      ! ! `name` must not be `NA`.
    Code
      sy(async_pkg_name_check("x", 11))
    Condition
      Error:
      ! ! One of `is.null(dictionaries)` and `is_character(dictionaries)` must be true.
    Code
      sy(async_pkg_name_check("x", NA_character_))
    Condition
      Error:
      ! ! One of `is.null(dictionaries)` and `is_character(dictionaries)` must be true.

# print.pkg_name_check

    Code
      print(ans[[1]])
    Output
      +------------------------------------------------------------------------------+
      |                               --*-- tools --*--                              |
      +------------------------------------------------------------------------------+
      +------------------------------------------------------------------------------+
      | v  valid name      x  CRAN            v  Bioconductor    v  not a profanity  |
      +------------------------------------------------------------------------------+
      + Wikipedia -------------------------------------------------------------------+
      | Tool (from Tools) A tool is an object that can extend an individual's        |
      | ability to modify features of the surrounding environment or help them       |
      | accomplish a particular task. Although many animals use simple tools, only   |
      | human beings, whose use of stone tools dates back hundreds of millennia,     |
      | have been observed using tools to make other tools.                          |
      |                                                                              |
      | ...                                                                          |
      +------------------------------------------ https://en.wikipedia.org/wiki/Tool +
      + Wiktionary ------------------------------------------------------------------+
      | tools Noun: tools                                                            |
      | plural of tool                                                               |
      | Verb: tools                                                                  |
      | third-person singular simple present indicative form of tool                 |
      | Anagrams: loots, lotos, sloot, sotol, stool, tosol                           |
      +---------------------------------------- https://en.wiktionary.org/wiki/tools +
      +------------------------------------------------------------------------------+
      | Sentiment: :| (0)                                                            |
      +------------------------------------------------------------------------------+

---

    Code
      print(ans[[2]])
    Output
      +------------------------------------------------------------------------------+
      |                               --*-- tools --*--                              |
      +------------------------------------------------------------------------------+
      +------------------------------------------------------------------------------+
      | v  valid name      x  CRAN            v  Bioconductor    v  not a profanity  |
      +------------------------------------------------------------------------------+
      + Urban dictionary ------------------------------------------------------------+
      | slang for the [hypodermic] [needles] used to [inject] drugs.                 |
      +-------------------------------------------- http://tools.urbanup.com/1443484 +

# pnc_base

    Code
      pnc_base("base")
    Output
      $base
      [1] FALSE
      
      $package
      [1] "base"
      
      attr(,"class")
      [1] "pkg_name_check_base" "list"               

---

    Code
      pnc_base("Base")
    Output
      $base
      [1] FALSE
      
      $package
      [1] "base"
      
      attr(,"class")
      [1] "pkg_name_check_base" "list"               

---

    Code
      pnc_base("TOOLS")
    Output
      $base
      [1] FALSE
      
      $package
      [1] "tools"
      
      attr(,"class")
      [1] "pkg_name_check_base" "list"               

---

    Code
      pnc_base("definitely-not")
    Output
      $base
      [1] TRUE
      
      $package
      NULL
      
      attr(,"class")
      [1] "pkg_name_check_base" "list"               

# format.pkg_name_basics

    Code
      writeLines(format(bss[[1]]))
    Output
      +------------------------------------------------------------------------------+
      |                                --*-- pwr --*--                               |
      +------------------------------------------------------------------------------+
      +------------------------------------------------------------------------------+
      | v  valid name      x  CRAN            v  Bioconductor    v  not a profanity  |
      +------------------------------------------------------------------------------+
    Code
      writeLines(format(bss[[2]]))
    Output
      +------------------------------------------------------------------------------+
      |                               --*-- shit --*--                               |
      +------------------------------------------------------------------------------+
      +------------------------------------------------------------------------------+
      | v  valid name      v  CRAN            v  Bioconductor    x  profanity        |
      +------------------------------------------------------------------------------+

# async_wikipedia_get_query

    Code
      show_request(ret)
    Output
      POST application/x-www-form-urlencoded
      Query string: 
      Body: action=query&format=json&prop=extracts&titles=foobar&redirects=1&exintro=1&explaintext=1

# format.pkg_name_check_wikipedia

    Code
      writeLines(format(wpd[[1]]))
    Output
      + Wikipedia -------------------------------------------------------------------+
      | CLI (from Cli) CLI may refer to multiple articles, see link.                 |
      +------------------------------------------- https://en.wikipedia.org/wiki/CLI +
    Code
      writeLines(format(wpd[[2]]))
    Output
      + Wikipedia -------------------------------------------------------------------+
      | Surely-not-this No definition found                                          |
      +------------------------------------------------------------------------------+
    Code
      writeLines(format(wpd[[3]]))
    Output
      + Wikipedia -------------------------------------------------------------------+
      | R (programming language) (from GNU R) R is a programming language for        |
      | statistical computing and graphics supported by the R Core Team and the R    |
      | Foundation for Statistical Computing. Created by statisticians Ross Ihaka    |
      | and Robert Gentleman, R is used among data miners, bioinformaticians and     |
      | statisticians for data analysis and developing statistical software. The     |
      | core R language is augmented by a large number of extension packages         |
      | ...                                                                          |
      +---------------------- https://en.wikipedia.org/wiki/R_(programming_language) +

# async_wiktionary_get

    Code
      print(ans)
    Output
      + Wiktionary ------------------------------------------------------------------+
      | foobar Etymology: Phonetic spelling of FUBAR, which is either an acronym     |
      | for "Fucked up beyond all recognition", or derived from foo.                 |
      | Noun: foobar (plural foobars)                                                |
      | (slang) A serious mistake. (programming) A metasyntactic variable name, a    |
      | place holder for words; compare foo, bar.                                    |
      | foo foo fighter FUBAR                                                        |
      | ...                                                                          |
      +--------------------------------------- https://en.wiktionary.org/wiki/foobar +

# async_wiktionary_get_query

    Code
      show_request(ret)
    Output
      POST application/x-www-form-urlencoded
      Query string: 
      Body: action=query&format=json&prop=extracts&titles=foobar&redirects=1&explaintext=1

# format.pkg_name_check_wiktionary

    Code
      writeLines(format(ans))
    Output
      + Wiktionary ------------------------------------------------------------------+
      | foobar Etymology: Phonetic spelling of FUBAR, which is either an acronym     |
      | for "Fucked up beyond all recognition", or derived from foo.                 |
      | Noun: foobar (plural foobars)                                                |
      | (slang) A serious mistake. (programming) A metasyntactic variable name, a    |
      | place holder for words; compare foo, bar.                                    |
      | foo foo fighter FUBAR                                                        |
      | ...                                                                          |
      +--------------------------------------- https://en.wiktionary.org/wiki/foobar +
    Code
      writeLines(format(ans2))
    Output
      + Wiktionary ------------------------------------------------------------------+
      | not-at-all-sdfsdfsdf No English definition found                             |
      +------------------------------------------------------------------------------+

# async_profanity_get_query

    Code
      show_request(ret)
    Output
      GET 
      Query string: text=foobar
      Body: 

# format.pkg_name_check_sentiment

    Code
      writeLines(format(ans1))
    Output
      +------------------------------------------------------------------------------+
      | Sentiment: :D (3)                                                            |
      +------------------------------------------------------------------------------+
    Code
      writeLines(format(ans2))
    Output
      +------------------------------------------------------------------------------+
      | Sentiment: :| (0)                                                            |
      +------------------------------------------------------------------------------+

# async_urban_get

    Code
      ans
    Output
      + Urban dictionary ------------------------------------------------------------+
      | One who lacks the mental capacity to know he is [being used].  A fool.  A    |
      | [cretin].  Characterized by low intelligence and/or [self-steem].            |
      +----------------------------------------------- http://tool.urbanup.com/38616 +

# async_urban_get_query

    Code
      show_request(ret)
    Output
      GET 
      Query string: term=foobar
      Body: 

# urban_get_process

    Code
      ans
    Output
      + Urban dictionary ------------------------------------------------------------+
      | One who lacks the mental capacity to know he is [being used].  A fool.  A    |
      | [cretin].  Characterized by low intelligence and/or [self-steem].            |
      +----------------------------------------------- http://tool.urbanup.com/38616 +

# format.pkg_name_check_urban

    Code
      writeLines(format(ans))
    Output
      + Urban dictionary ------------------------------------------------------------+
      | One who lacks the mental capacity to know he is [being used].  A fool.  A    |
      | [cretin].  Characterized by low intelligence and/or [self-steem].            |
      +----------------------------------------------- http://tool.urbanup.com/38616 +

---

    Code
      writeLines(format(ans2))
    Output
      + Urban dictionary ------------------------------------------------------------+
      | No definition found.                                                         |
      +------------------------------------------------------------------------------+

---

    Code
      writeLines(format(ans))
    Output
      + Urban dictionary ------------------------------------------------------------+
      | One who lacks the mental capacity to know he is [being used].  A fool.  A    |
      | [cretin].  Characterized by low intelligence and/or [self-steem]. One who    |
      | lacks the mental capacity to know he is [being used].  A fool.  A [cretin].  |
      | Characterized by low intelligence and/or [self-steem]. One who lacks the     |
      | mental capacity to know he is [being used].  A fool.  A [cretin].            |
      | Characterized by low intelligence and/or [self-steem]. One who lacks the     |
      | ...                                                                          |
      +----------------------------------------------- http://tool.urbanup.com/38616 +

# async_pnc_bioc_web

    Code
      ans
    Output
      $bioc
      [1] FALSE
      
      $package
      [1] "ALL"
      
      attr(,"class")
      [1] "pkg_name_check_bioc" "list"               

# pnc_bioc_query

    Code
      ans
    Output
      $bioc
      [1] FALSE
      
      $package
      [1] "ALL"
      
      attr(,"class")
      [1] "pkg_name_check_bioc" "list"               

---

    Code
      ans2
    Output
      $bioc
      [1] FALSE
      
      $package
      [1] "agcdf"
      
      attr(,"class")
      [1] "pkg_name_check_bioc" "list"               

---

    Code
      ans3
    Output
      $bioc
      [1] TRUE
      
      $package
      NULL
      
      attr(,"class")
      [1] "pkg_name_check_bioc" "list"               

# pnc_bioc_parse

    Code
      pkg1
    Output
      [1] "a4"          "a4Base"      "a4Classif"   "a4Core"      "a4Preproc"  
      [6] "a4Reporting" "aCGH"        "abseqR"      "ag.db"      
      attr(,"na.action")
      [1] 1 2
      attr(,"class")
      [1] "omit"

---

    Code
      pkg2
    Output
       [1] "ABAData"       "ABAEnrichment" "ABSSeq"        "AGDEX"        
       [5] "AHPathbankDbs" "AIMS"          "ALDEx2"        "ALL"          
       [9] "ALLMLL"        "ALPS"          "AMARETTO"     
      attr(,"na.action")
      [1] 1 2
      attr(,"class")
      [1] "omit"

# pnc_bioc_parse_pgz

    Code
      pkg
    Output
      [1] "adme16cod.db" "ag.db"        "agcdf"       

# initial state

    Code
      invisible(pkgplan__create_progress_bar(what))
    Message
      i No downloads are needed

---

    Code
      invisible(pkgplan__create_progress_bar(what))
    Message
      i No downloads are needed, 1 pkg (10 kB) is cached

---

    Code
      invisible(pkgplan__create_progress_bar(what))
    Message
      i No downloads are needed, 2 pkgs (20 kB) are cached

---

    Code
      invisible(pkgplan__create_progress_bar(what))
    Message
      i No downloads are needed, 1 pkg is cached

---

    Code
      invisible(pkgplan__create_progress_bar(what))
    Message
      i Getting 2 pkgs (30 kB)

---

    Code
      invisible(pkgplan__create_progress_bar(what))
    Message
      i Getting 2 pkgs (30 kB), 1 cached

---

    Code
      invisible(pkgplan__create_progress_bar(what))
    Message
      i Getting 2 pkgs (30 kB), 1 (20 kB) cached

---

    Code
      invisible(pkgplan__create_progress_bar(what))
    Message
      i Getting 1 pkg (20 kB) and 1 pkg with unknown size

---

    Code
      invisible(pkgplan__create_progress_bar(what))
    Message
      i Getting 1 pkg (10 kB) and 1 pkg with unknown size, 1 (30 kB) cached

---

    Code
      invisible(pkgplan__create_progress_bar(what))
    Message
      i Getting 1 pkg (10 kB) and 1 pkg with unknown size, 1 cached

---

    Code
      invisible(pkgplan__create_progress_bar(what))
    Message
      i Getting 2 pkgs with unknown sizes

---

    Code
      invisible(pkgplan__create_progress_bar(what))
    Message
      i Getting 2 pkgs with unknown sizes, 1 cached

---

    Code
      invisible(pkgplan__create_progress_bar(what))
    Message
      i Getting 2 pkgs with unknown sizes, 1 (30 kB) cached

# data updates

    Code
      bar <- pkgplan__create_progress_bar(what = what)
    Message
      i Getting 2 pkgs (30 kB), 1 (20 kB) cached
    Code
      pkgplan__update_progress_bar(bar, 1L, "data", list(current = 5000, total = 10000))
    Output
      [1] TRUE
    Code
      pkgplan__update_progress_bar(bar, 4L, "data", list(current = 20000, total = 20000))
    Output
      [1] TRUE
    Code
      pkgplan__done_progress_bar(bar)
    Output
      NULL
    Code
      bar$what
    Output
      # A tibble: 4 x 9
        type      filesize package cache_status   idx current  need status skip 
        <chr>        <dbl> <chr>   <chr>        <int>   <dbl> <dbl> <chr>  <lgl>
      1 cran         10000 foo     miss             1    5000 10000 data   FALSE
      2 installed    10000 foo2    <NA>             2       0 10000 skip   TRUE 
      3 cran         20000 bar     hit              3       0 20000 skip   TRUE 
      4 cran         20000 bar2    miss             4   20000 20000 data   FALSE

# all finish messages for updates

    Code
      do(1L)
    Message
      i Getting 2 pkgs (30 kB), 1 (20 kB) cached
      v Got foo 1.0.0 (source) (7 B)
    Output
      # A tibble: 4 x 9
        type      filesize package cache_status   idx current  need status skip 
        <chr>        <dbl> <chr>   <chr>        <int>   <dbl> <dbl> <chr>  <lgl>
      1 cran             7 foo     miss             1       7 10000 got    FALSE
      2 installed    10000 foo2    <NA>             2       0 10000 skip   TRUE 
      3 cran         20000 bar     hit              3       0 20000 skip   TRUE 
      4 cran         20000 bar2    miss             4       0 20000 todo   FALSE

---

    Code
      do(1L, make_tempfile = FALSE)
    Message
      i Getting 2 pkgs (30 kB), 1 (20 kB) cached
      v Got foo 1.0.0 (source)
    Output
      # A tibble: 4 x 9
        type      filesize package cache_status   idx current  need status skip 
        <chr>        <dbl> <chr>   <chr>        <int>   <dbl> <dbl> <chr>  <lgl>
      1 cran         10000 foo     miss             1   10000 10000 got    FALSE
      2 installed    10000 foo2    <NA>             2       0 10000 skip   TRUE 
      3 cran         20000 bar     hit              3       0 20000 skip   TRUE 
      4 cran         20000 bar2    miss             4       0 20000 todo   FALSE

---

    Code
      do(1L, event = "error")
    Message
      i Getting 2 pkgs (30 kB), 1 (20 kB) cached
      x Failed to download foo 1.0.0 (source)
    Output
      # A tibble: 4 x 9
        type      filesize package cache_status   idx current  need status skip 
        <chr>        <dbl> <chr>   <chr>        <int>   <int> <dbl> <chr>  <lgl>
      1 cran         10000 foo     miss             1       0     0 error  FALSE
      2 installed    10000 foo2    <NA>             2       0 10000 skip   TRUE 
      3 cran         20000 bar     hit              3       0 20000 skip   TRUE 
      4 cran         20000 bar2    miss             4       0 20000 todo   FALSE

# parts are calculated properly

    Code
      bar <- pkgplan__create_progress_bar(what)
    Message
      i Getting 2 pkgs (30 kB), 1 (20 kB) cached

# verify_extracted_package / errors if archive DESCRIPTION is not in the root directory

    Code
      run(f2)
    Condition
      Error:
      ! ! '<tempdir>/<tempfile>/test2.tgz' is not a valid binary, it is missing test2/Meta/package.rds and test2/DESCRIPTION.

# verify_extracted_package / can handle multiple DESCRIPTION files

    Code
      run(f4)
    Condition
      Error:
      ! ! '<tempdir>/<tempfile>/test4.tgz' is not a valid binary, it is missing test4/DESCRIPTION.

# verify_extracted_package / fails if the binary does not contain package.rds

    Code
      run(f5)
    Condition
      Error:
      ! ! '<tempdir>/<tempfile>/test5.tgz' is not a valid binary, it is missing test5/Meta/package.rds.

# verify_extracted_package / fails if the DESCRIPTION file is empty

    Code
      run(f6)
    Condition
      Error:
      ! ! '<tempdir>/<tempfile>/test6.tgz' is not a valid binary, empty DESCRIPTION file at 'test6/DESCRIPTION'.

\renewcommand{\eval}{\Sexpr[stage=render,results=rd]{pkgdepends:::#1}}
\renewcommand{\evalatinstall}{\Sexpr[stage=install,results=rd]{pkgdepends:::#1}}
\renewcommand{\evaltop}{\Sexpr[stage=render,results=rd]{pkgdepends:::#1}}
\renewcommand{\evalatinstalltop}{\Sexpr[stage=install,results=rd]{pkgdepends:::#1}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sysreqsdocs.R
\name{sysreqs_db_update}
\alias{sysreqs_db_update}
\title{Update the cached copy of the system requirements database}
\usage{
sysreqs_db_update()
}
\description{
Update the cached copy of the system requirements database
}
\details{
If the the cached copy is recent, then no update is attempted. See the
\code{metadata_update_after} \eval{man_config_link("configuration option")}.
}
\seealso{
Other system requirements functions: 
\code{\link{sysreqs_check_installed}()},
\code{\link{sysreqs_db_list}()},
\code{\link{sysreqs_db_match}()},
\code{\link{sysreqs_install_plan}()},
\code{\link{sysreqs_is_supported}()},
\code{\link{sysreqs_list_system_packages}()},
\code{\link{sysreqs_platforms}()}
}
\concept{system requirements functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/git-protocol.R
\name{git-protocol}
\alias{git-protocol}
\title{git protocol notes, for developers}
\description{
Assumptions, they might be relaxed or checked for later:
\itemize{
\item The server must speak the smart protocol, version 1 or 2.
(Although I added some functions specifically for the dumb protocol,
but they are pretty limited.)
\item We use HTTP transport, not SSH.
\item The server should have the \code{shallow} capability.
\item The server should have the \code{filter} capability if protocol version 2.
\item Only SHA-1 hashing is supported.
}
}
\details{
Improvements needed:
\itemize{
\item DONE Tests. (Can always have more.)
\item DONE Use async HTTP.
\item DONE Support packfiles with deltas. (\code{ofs-delta} objects are still
not supported.)
\item DONE Optionally send authorization. Already possibly in the URL.
\item DONE Better error messages.
\item Better errors for non-existing user, repository, ref, PR, etc.
}

Optional improvements:
\itemize{
\item Support \code{ofs-delta} objects in packfiles. Not necessarily, unless we
send this capability, the server is not sending \code{ofs-delta} objects.
\item Make unpacking faster. It is not fast currently, with all the bit
arithmetic in R. But it is already faster than a \code{tar.gz} + uncompress
download from GitHub, so not really needed.
}
\subsection{Docs and other helpful links:}{
\itemize{
\item \url{https://github.com/git/git/blob/master/Documentation/gitprotocol-common.adoc}
\item \url{https://github.com/git/git/blob/master/Documentation/gitprotocol-pack.adoc}
\item \url{https://github.com/git/git/blob/master/Documentation/gitprotocol-v2.adoc}
\item \url{https://github.com/calebsander/git-internals/blob/part2/src/main.rs}
\item \url{https://dev.to/calebsander/git-internals-part-1-the-git-object-model-474m}
}
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/platform.R
\name{current_r_platform}
\alias{current_r_platform}
\alias{default_platforms}
\title{R platforms}
\usage{
current_r_platform()

default_platforms()
}
\value{
\code{current_r_platform()} returns a string, the name of the current
platform.

\code{default_platforms()} returns a character vector of platform names.
}
\description{
\code{default_platfoms()} returns the default platforms for the current R
session. These typically consist of the detected platform of the current
R session, and \code{"source"}, for source packages.
}
\details{
\code{current_r_platform()} detects the platform of the current R version.

By default pkgdepends works with source packages and binary packages for
the current platform. You can change this, see
\link[=pkgdepends-config]{'Configuration'}.

The following platform names can be configured and returned by
\code{current_r_platform()} and \code{default_platforms()}:
\itemize{
\item \code{"source"} for source packages,
\item A platform string like \code{R.version$platform}, but on Linux the name
and version of the distribution are also included. Examples:
\itemize{
\item \code{x86_64-apple-darwin17.0}: macOS High Sierra.
\item \code{aarch64-apple-darwin20}: macOS Big Sur on arm64.
\item \code{x86_64-w64-mingw32}: 64 bit Windows.
\item \code{i386-w64-mingw32}: 32 bit Windows.
\item \code{i386+x86_64-w64-mingw32}: 64 bit + 32 bit Windows.
\item \code{i386-pc-solaris2.10}: 32 bit Solaris. (Some broken 64 Solaris
builds might have the same platform string, unfortunately.)
\item \code{x86_64-pc-linux-gnu-debian-10}: Debian Linux 10 on x86_64.
\item \verb{x86_64-pc-linux-musl-alpine-3.14.1}: Alpine Linux.
\item \code{x86_64-pc-linux-gnu-unknown}: Unknown Linux Distribution on x86_64.
\item \code{s390x-ibm-linux-gnu-ubuntu-20.04}: Ubuntu Linux 20.04 on S390x.
\item \code{amd64-portbld-freebsd12.1}: FreeBSD 12.1 on x86_64.
}
}

In addition, the following platform names can be used to configure
pkgdepends:
\itemize{
\item \code{"macos"} for macOS binaries that are appropriate for the R versions
pkgdepends is working with (defaulting to the version of the current
session), as defined by CRAN binaries. E.g. on R 3.5.0 macOS binaries
are built for macOS El Capitan.
\item \code{"windows"} for Windows binaries for the default CRAN architecture.
This is currently Windows Vista for all supported R versions, but it
might change in the future. The actual binary packages in the
repository might support both 32 bit and 64 builds, or only one of
them. In practice 32-bit only packages are very rare. CRAN builds
before and including R 4.1 have both architectures, from R 4.2 they
are 64 bit only. \code{"windows"} is an alias to \code{i386+x86_64-w64-mingw32}
currently.
}
}
\examples{
current_r_platform()
default_platforms()
}
\concept{platform functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resolution.R
\name{pkg_resolution}
\alias{pkg_resolution}
\alias{pkg_resolution_result}
\title{Dependency resolution}
\description{
Collect information about dependencies of R packages, recursively.
}
\details{
\code{\link{pkg_deps}}, \code{\link{pkg_download_proposal}} and \code{\link{pkg_installation_proposal}}
all resolve their dependencies recursively, to obtain information about
all packages needed for the specified \link[=pkg_refs]{package references}.
\subsection{CRAN and Bioconductor packages}{

Resolution currently start by downloading the CRAN and Bioconductor
metadata, if it is out of date. For CRAN, we also download additional
metadata, that includes file sizes, SHA hashes, system requirements,
and "built" (for binary packages) and "packaged" time stamps. The extra
meta information is updated daily currently, so for some packages it
might be incorrect or missing.
}

\subsection{GitHub packages}{

For GitHub packages, we query their download URL to be able to
download the package later, and also download their \code{DESCRIPTION}
file, to learn about their dependencies.
}

\subsection{Local packages}{

From local package files we extract the \code{DESCRIPTION} file, to learn
about their dependencies.
}

\subsection{The \code{remotes} field in \code{DESCRIPTION}}{

We support the non-standard \code{Remotes} field in the package \code{DESCRIPTION}
file. This field may contain a list of package references for any of the
dependencies that are specified in one of the \code{Depends}, \code{Includes},
\code{Suggests} or \code{Enhances} fields. The syntax is a comma separated list of
\link[=pkg_refs]{package references}.
}

\subsection{The result}{

The result of the resolution is a data frame with information about the
packages and their dependencies.
\itemize{
\item \code{built}: the \code{Built} field from the \code{DESCRIPTION} file of binary
packages, for which this information is available.
\item \code{cache_status}: whether the package file is in the package cache.
It is \code{NA} for \verb{installed::} package refs.
\item \code{dep_types}: character vector of dependency types that were
considered for this package. (This is a list column.)
\item \code{deps}: dependencies of the package, in a data frame. See
"Package dependency tables" below.
\item \code{direct}: whether this package (ref, really) was directly specified,
or added as a dependency.
\item \code{error}: this is a list column that contains error objects for the
refs that pkgdepends failed to resolve.
\item \code{filesize}: the file size in bytes, or \code{NA} if this information is
not available.
\item \code{license}: license of the package, or \code{NA} if not available.
\item \code{md5sum}: MD5 checksum of the package file, if available, or \code{NA} if
not.
\item \code{metadata}: a named character vector. These fields will be (should be)
added to the installed \code{DESCRIPTION} file of the package.
\item \code{mirror}: URL of the CRAN(-like) mirror site where the metadata was
obtained from. It is NA for non-CRAN-like sources, e.g. local files,
installed packages, GitHub, etc.
\item \code{needscompilation}: whether the package needs compilation.
\item \code{package}: package name.
\item \code{priority}: this is \code{"base"} for base packages, \code{"recommended"} for
recommended packages, and \code{NA} otherwise.
\item \code{ref}: package reference.
\item \code{remote}: the parsed \code{remote_ref} objects, see \code{\link[=parse_pkg_refs]{parse_pkg_refs()}}.
This is a list column.
\item \code{repodir}: the directory where this package should be in a CRAN-like
repository.
\item \code{sha256}: SHA256 hash of the package file, if available, otherwise
\code{NA}.
\item \code{sources}: URLs where this package can be downloaded from. This is not
necessarily a URL that you can download with a HTTP client. E.g. for
\verb{local::} refs it is a path, and for \verb{git::} refs it is a URL for git.
It is a zero length vector for \verb{installed::} refs.
\item \code{status}: status of the dependency resolution, \code{"OK"} or \code{"FAILED"}.
\item \code{target}: path where this package should be saved in a CRAN-repository.
\item \code{type}: ref type.
\item \code{version}: package version.
}

Additional columns might be present. They are either used internally or
they are experimental. They might be removed or changed at any time.

All columns are of type character, except for \code{direct} (logical),
\code{needscompilation} (logical), \code{filesize} (integer), \code{deps} (list column, see
"Package dependency tables" below), \code{sources} (list of character vectors),
\code{remote} (list), \code{error} (list), \code{metadata} (list), \code{dep_types} (list).
\subsection{Package dependency tables}{

A package dependency tables in the \code{deps} list column have five columns
currently:
\itemize{
\item \code{ref}: the package ref of the dependency.
\item \code{type}: the dependency type, in all lowercase. I.e. \code{imports},
\code{suggests}, etc.
\item \code{package}: package name of the dependency.
\item \code{op}: operator for version requirements, e.g. \code{>=}.
\item \code{version}: version number, for version requirements.
}
}

}

\subsection{Resolution failures}{

The resolution process does not stop on error. Instead, failed
resolutions return and error object in the \code{error} column of the result
data frame.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repo.R
\name{repo}
\alias{repo}
\title{Query and manipulate CRAN-like repositories}
\description{
These functions are currently experimental.
}
\details{
\subsection{List packages in a repository}{

\code{repo$list()} lists packages in a repository. It reads the \code{PACKAGES}
file containing the repository metadata.
\subsection{Usage}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{repo_list(..., path = ".")
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item \code{...}: ignored currently.
\item \code{path}: path to repository. Must contain a \code{PACKAGES} file.
}
}

\subsection{Value}{

Data frame of package data, a data frame with at least the following
columns, possibly more if there are other entries in the metadata:
\itemize{
\item \code{Package}
\item \code{Version}
\item \code{Depends}
\item \code{Imports}
\item \code{Suggests}
\item \code{Enhances}
\item \code{LinkingTo}
\item \code{License}
\item \code{File}
\item \code{DownloadURL}
\item \code{OS}
\item \code{Arch}
\item \code{Built}
\item \code{Filesize}
\item \code{SHA256}
\item \code{RVersion}
\item \code{Platform}
\item \code{GraphicsAPIVersion}
\item \code{InternalsId}
\item \code{SystemRequirements}
}
}

}

\subsection{Delete packages from repository metadata}{

\code{repo$delete()} deletes matching packages from the repository
metadata.
\subsection{Description}{

All matching packages will be removed.
It does not delete the files themselves.
}

\subsection{Usage}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{repo$delete(package, ..., path = ".")
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item \code{package}: package name.
\item \code{...}: other fields to match, they must be named. Matching is
case insensitive.
\item \code{path}: path to repository. Must contain a \code{PACKAGES} file.
}
}

}

\subsection{Add a package to a repository}{
\subsection{Description}{

It does not check if any version of the package is already
in the repository. If you want to \emph{update} a package, use
\code{repo$update()}.
}

\subsection{Usage}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{repo$add(file, ..., path = ".")
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item `file Package file.
\item `... Ignored currently.
\item \verb{path Path to repository. Must contain a }PACKAGES` file.
}
}

}

\subsection{Update a package in a repository}{
\subsection{Description}{

Previous version of the same package are removed. In particular,
it removes packages with matching:
\itemize{
\item package name (\code{Package} field),
\item R version (\code{Rversion} field),
\item same OS (\code{OS} field), or no \code{OS} field,
\item same architecture (\code{Arch} field), or not \code{Arch} field.
}
}

\subsection{Usage}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{repo$update(file, ..., path = ".")
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item `file Package file.
\item `... Ignored currently.
\item \verb{path Path to repository. Must contain a }PACKAGES` file.
}
}

}

\subsection{Update a file in a package metadata, stored on GitHub}{
\subsection{Description}{
\enumerate{
\item Clones the GitHub repository.
\item Calls \code{repo_update()} with \code{file}, in the \code{subdir} directory.
\item Adds and commits changes.
\item Pushes the git repository to GitHub.
If the push fails, then it resets the git repository, pulls it from
GitHub and tries the update process again, until the push succeeds.
}

This function needs command line git installed.

It sets up a \code{cache} credential helper, so the \verb{git push} works
without interaction with the user.
}

\subsection{Usage}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{repo$update_gh(repo, subdir, files)
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item \code{repo}: GitHub slug, e.g. \code{r-hub/repos}.
\item \code{subdir}: subdirectory in the GitHub repository, where the R package
metadata should be updated. It must exist in the repository.
If it does not have \verb{PACKAGES*} files, then they will be created.
\item \code{files}: package files to add. The files will \emph{not} be added to the
repository, only to the metadata. They should be in the repository
already.
}
}

}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\docType{data}
\name{config}
\alias{config}
\title{Package configuration (internals)}
\format{
\code{config} is a list of functions with a closure. You can use it two ways.

For \emph{one} set of configuration values in a package, include this in
\code{.onLoad()}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{conf <- config$new("prefix")
conf$add("myenrry", "string", "default-value")
...
conf$lock()
}\if{html}{\out{</div>}}

For _multiple _ sets of configuration values (e.g. one per object),
include a function like this in the package:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{current_config <- function() \{
  conf <- config$new("prefix")
  conf$add("myentry", "string", "default-value")
  ...
  conf$lock()
  conf
\}
}\if{html}{\out{</div>}}

and then call \code{current_config()} every time you need to create a
new set of configuration values.
}
\usage{
config
}
\description{
Internal machinery for pkgdepends configuration. See \code{\link[=current_config]{current_config()}}
for how this machinery can be used to define the configuration of a
package.

This is a standalone file that can be copied into another package without
changes.
}
\section{Terminology}{
\subsection{Config entry (or entry)}{

A single configurable entity. A config entry has a name. This name is
always standardized by replacing dashes with underscores.

A config entry has a type: a string, a flag, or some
user defined type. It can also have a default value. Examples:
\itemize{
\item The URL of an API, which can be a string.
\item Path to a directory to use for caching, either a string, or NULL
(\code{string_or_null} type).
\item Number of processor cores to use, a positive integer (\code{count} type).
}
}

\subsection{Configuration (or config)}{

A set of configuration entries. \code{config$new()} returns a configuration.
A configuration can be extended by adding new entries to it, until it
is locked.
}

\subsection{R option (or option)}{

An option set with \code{\link[base:options]{base::options()}} and queried with \code{\link[base:options]{base::getOption()}}.
}

\subsection{Configuration prefix}{

A prefix that is typically package specific, and it is used for all
config entries of a package, to avoid environment variable and option
name clashes between packages.
}

\subsection{Config entry type}{

It defines the allowed values of configuration entries, and also how
a string from an environment variable is translated to the value of
the entry. See the list of builtin types below. You can also define new
types.
}
}

\section{Built-in types}{
\subsection{\code{string}}{

String scalar, \code{NA} is not allowed.
}

\subsection{\code{count}}{

Non-negative integer, \code{NA} is not allowed.
}

\subsection{\code{flag}}{

Logical scalar, \code{NA} is not allowed. For environment variables the
\code{TRUE} values are: \code{yes}, \code{true}, \code{1} and \code{on}, and the \code{FALSE} values
are: \code{no}, \code{false}, \code{0}, \code{off}. (All are case insensitive.)
}

\subsection{\code{string_or_null}}{

String or \code{NULL}. \code{NA} is not allowed. In environment variables the
string \code{NULL} means an R \code{NULL} value.
}

\subsection{\code{character}}{

Character vector without \code{NA} values. In environment variables the
entries are separated by a semicolon.
}

\subsection{\code{custom}}{

Custom type. An \code{env_decode} function must be specified for each config
entry of this type, otherwise an error is throw if the corresponding
environment variable is set.
}
}

\section{\code{config$onload_hook()}: pretty-printing configuration}{
Call \code{config$onload_hook()} from the package's \code{.onLoad()} function
if you want to define S3 methods to print the package configuration
nicer.
\subsection{Usage}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{config$onload_hook()
}\if{html}{\out{</div>}}
}
}

\section{\code{config$new()}: Create a new configuration}{
Typically \code{new()} is called outside of the functions of the
package, so the \code{config} object is created at install time.
\subsection{Usage}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{conf <- config$new(prefix = utils::packageName(parent.frame()))
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item \code{prefix}: prefix of the config entry names. For environment
variables it is converted to uppercase, and dots are converted
to underscores. The default prefix is the name of the calling package.
An underscore separator is used between the prefix and the entry
name for environment variable names. A dot separator is used for
R option names.
}
}

\subsection{Value}{

\code{new()} returns a \code{config} object, which is a list containing
the configuration data and functions (methods) to query and change it.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{conf <- config$new()
}\if{html}{\out{</div>}}
}
}

\section{Configuration methods}{
\subsection{\code{conf$add()}: add a new configuration entry}{
\subsection{Usage}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{conf$add(
  name,
  type = conf$types,
  default = NULL,
  check = type[1],
  env_decode = type[1]
)
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item \code{name}: name of the entry.
\item \code{type}: type of the entry, string.
\item \code{default}: default value of config entry. If this is a function,
then it is called to generate the default, at the time the
config entry's value is queried.
\item \code{check}: if it is a function, it is called to check the value of
the entry. The function should return \code{TRUE} for successful and
\code{FALSE} for unsuccessful checks. If \code{NULL}, then no check is
performed. It can also be the name of a type, then the default
check for that type is performed, if any.
\item \code{env_decode}: if it is a function, then it is used to decode the
value of the entry from an environment variable, i.e. a string.
It can also be the name of a type, then the \code{env_decode()}
function of that type is used for decoding.
}
}

\subsection{Value}{

The configuration, invisibly.
}

}

\subsection{\code{conf$get()}: query the value of a configuration entry}{
\subsection{Usage}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{conf$get(name)
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item \code{name}: name of the entry.
}
}

\subsection{Value}{

Value of the entry.
\itemize{
\item If the entry was set via \code{conf$set()} or \code{conf$update()}, then
that value is returned.
\item Otherwise, if the entry is set via an R option (see
\code{\link[base:options]{base::getOption()}}), then that value is returned. The config
prefix is used to get the option name, with a dot separator.
I.e. for an entry called 'foo', and prefix 'pkg', the \code{pkg.foo}
option is used. If the entry has a check function, that is called
before returning.
\item Otherwise, if the entry is set via an environment variable, then
that value is returned. The config prefix is used to get the name
of the environment variable. In addition, dots are replaced with
underscores and the name is converted to uppercase. I.e. for an
entry called 'foo.bar', and prefix 'pkg', the 'PKG_FOO_BAR`
environment variable is used.
\item Otherwise, if the entry has a default value, that is returned.
The default value might be a function, in which case that function
is called to produce a default value. If the entry has a check
function, that is called before returning.
}
}

}

\subsection{\code{conf$set()}: set the value of a configuration entry}{

If you \verb{$set()} a config entry, then the \code{value} used in \verb{$set()}
will be returned by \verb{$get()}, without consulting R options or
environment variables.
\subsection{Usage}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{conf$set(name, value)
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item \code{name}: name of the entry.
\item \code{value}: value of the entry. If the entry has a check function,
it is called here.
}
}

\subsection{Value}{

The configuration, invisibly.
}

}

\subsection{\verb{$conf$unset()}: unset a configuration entry}{

Note that this function does \emph{not} unset R options and environment
variables. It merely removes a value that was assigned by \verb{$set()}
or \verb{$update()}.
\subsection{Usage}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{conf$unset(name)
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item \code{name}: name of the entry.
}
}

\subsection{Value}{

The configuration, invisibly.
}

}

\subsection{\code{cond$update()}: update the values of configuration entries}{

You can use this method to set()\code{and/or}$unset()\verb{multiple config entries.}NULL\verb{values in}new` will unset the corresponding entry.
\subsection{Usage}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{conf$update(new)
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item \code{new}: named list, where names are config entry names and values
are the corresponding config entry values. \code{NULL} values will
unset the config entry. The list is processed sequentially, so
for duplicates the latest values will be in effect. (But every
value is checked with its check function, if any.)
}
}

\subsection{Value}{

The configuration, invisibly.
}

}

\subsection{\code{conf$list()}: query the names of all config entries}{

Note that their order is non-deterministic currently.
\subsection{Usage}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{conf$list()
}\if{html}{\out{</div>}}
}

\subsection{Value}{

Character vector.
}

}

\subsection{\code{conf$exists()}: check if a config entry exists}{
\subsection{Usage}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{conf$exists(name)
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item \code{name}: name of the entry.
}
}

\subsection{Value}{

Logical scalar, \code{TRUE} if the entry exists in the configuration,
\code{FALSE} otherwise. Note that \code{TRUE} does not mean that the value
of the entry was set.
}

}

\subsection{\code{conf$lock()}: lock the configuration}{

If a configuration is locked, then no more entries can be added to
it.
\subsection{Usage}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{conf$lock()
}\if{html}{\out{</div>}}
}

\subsection{Value}{

The configuration, invisibly.
}

}

\subsection{\code{conf$fix()}: fix the value of a single configuration entry}{

The currently \verb{$set()} value is fixed and cannot be changed any
more. You can fix an entry without setting its value, in which case
R options and environment variables can still be used to update it.
\subsection{Usage}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{conf$unset(name)
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item \code{name}: name of the entry.
}
}

\subsection{Value}{

The configuration, invisibly.
}

}

\subsection{\code{conf$add_type()}: add a new config entry type to a configuration}{

After this the new type can be used in \code{conf$add()} calls.
\subsection{Usage}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{conf$add_type(type_name, check, env_decode)
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item \code{type_name}: type name.
\item \code{check}: default check function of the type. Use \code{NULL} for no
checks.
\item \code{env_decode}: function to use to decode a value of the type from
an environment variable (i.e. string).
}
}

\subsection{Value}{

The configuration, invisibly.
}

}
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type-installed.R
\name{lib_status}
\alias{lib_status}
\title{Status of packages in a library}
\usage{
lib_status(library = .libPaths()[1], packages = NULL)
}
\arguments{
\item{library}{Path to library.}

\item{packages}{If not \code{NULL}, then only these packages are shown.}
}
\value{
Data frame that contains data about the packages
installed in the library.

It has always has columns:
\itemize{
\item \code{biocviews}: the corresponding field from \code{DESCRIPTION}, it must be
present for all Bioconductor packages, other packages typically don't
have it.
\item \code{built}: the \code{Built} field from \code{DESCRIPTION}.
\item \code{depends}, \code{suggests}, \code{Imports}, \code{linkingto}, \code{enhances}: the corresponding
fields from the \code{DESCRIPTION} files.
\item \code{deps}: A list or data frames, the dependencies of the package. It has
columns: \code{ref}, \code{type} (dependency type in lowercase), \code{package}
(dependent package, or \code{R}), \code{op} and \code{version}, for last two are for
version requirement. \code{op} can be \code{>=}, \code{>}, \code{==} or \code{<=}, although the
only the first one is common in practice.
\item \code{library}: path to the package library containing the package.
\item \code{license}: from \code{DESCRIPTION}.
\item \code{md5sum}: from \code{DESCTIPTION}, typically \code{NA}, except on Windows.
\item \code{needscompilation}: from \code{DESCRIPTION}, this column is logical.
\item \code{package}: package name.
\item \code{platform}: from the \code{Built} field in \code{DESCRIPTION}, the current platform
if missing from \code{DESCRIPTION}.
\item \code{priority}: from \code{DESCRIPTION}, usually \code{base}, \code{recommended}, or missing.
\item \code{ref}: the corresponding \verb{installed::*} package reference.
\item \code{repository}: from \code{DESCRIPTION}. For packages from a CRAN repository this
is \code{CRAN}, some other repositories, e.g. R-universe adds the repository
URL here.
\item \code{repotype}: \code{cran}, \code{bioc} or missing.
\item \code{rversion}: from the \code{Built} field. If no such field, then the current
R version.
\item \code{sysreqs}: the \code{SystemRequirements} field from \code{DESCRIPTION}.
\item \code{title}: package title.
\item \code{type}: always \code{installed}.
\item \code{version}: package version (as string).
}

Most of these columns are unchanged from \code{DESCRIPTION}, but
pkgdepends also adds a couple.
\subsection{Notes:}{
\itemize{
\item In addition, it also has all \verb{remote*} and \verb{config/needs/*} entries from
the \code{DESCRIPTION} files. (Case insensitive.)
\item All columns are of type \code{character}, except for \code{needscompilation}, which
is logical and \code{deps}, which is a list columns.
\item If an entry is missing for a package, it is set to \code{NA}.
\item Note that column names are lowercase, even if the corresponding entries
are not in \code{DESCRIPTION}.
\item The order of the columns is not deterministic, so don't assume any order.
\item Additional columns might be present, these are internal for
pkgdepends and should not be used in user code.
}
}
}
\description{
Query data of all packages in a package library.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sysreqsdocs.R
\name{sysreqs_is_supported}
\alias{sysreqs_is_supported}
\title{Check if a platform has system requirements support}
\usage{
sysreqs_is_supported(sysreqs_platform = NULL)
}
\arguments{
\item{sysreqs_platform}{System requirements platform. If \code{NULL}, then the
\code{sysreqs_platform} \eval{man_config_link("configuration option")}
is used, which defaults to the current platform. Set this option if
\eval{.packageName} does not detect your platform correctly.}
}
\value{
Logical scalar.
}
\description{
Check if a platform has system requirements support
}
\examples{
\dontshow{if (Sys.getenv("IN_PKGDOWN") == "true") (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
sysreqs_is_supported()
\dontshow{\}) # examplesIf}
}
\seealso{
The \code{sysreqs_platform}
\eval{man_config_link("configuration option")}.

Other system requirements functions: 
\code{\link{sysreqs_check_installed}()},
\code{\link{sysreqs_db_list}()},
\code{\link{sysreqs_db_match}()},
\code{\link{sysreqs_db_update}()},
\code{\link{sysreqs_install_plan}()},
\code{\link{sysreqs_list_system_packages}()},
\code{\link{sysreqs_platforms}()}
}
\concept{system requirements functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sysreqs.R
\name{sysreqs_install_plan}
\alias{sysreqs_install_plan}
\title{Create an installation plan for system requirements}
\usage{
sysreqs_install_plan(refs, upgrade = TRUE, config = list())
}
\arguments{
\item{refs}{Packages to install.}

\item{upgrade}{If \code{TRUE}, pkgdepends will choose the latest available
versions of packages, instead of preferring binary packages over
source packages.}

\item{config}{Configuration options. See
\link[=pkgdepends-config]{'Configuration'}. If it does not include
\code{library}, then a temporary library is used, which is equivalent to
not assuming any preinstalled packages. Pass \code{sysreqs_platform} here
if you want a different platform than the one R is running on.}
}
\value{
List with entries:
\itemize{
\item \code{os}: character string. Operating system.
\item \code{distribution}: character string. Linux distribution, \code{NA} if the
OS is not Linux.
\item \code{version}: character string. Distribution version, \code{NA} is the OS
is not Linux.
\item \code{pre_install}: character vector. Commands to run before the
installation of system packages.
\item \code{install_scripts}: character vector. Commands to run to install the
system packages.
\item \code{post_install}: character vector. Commands to run after the
installation of system packages.
\item \code{packages}: data frame. Information about the system packages that
are needed. It has columns:
\itemize{
\item \code{sysreq}: string, cross-platform name of the system requirement.
\item \code{packages}: list column of character vectors. The names of the R
packages that have this system requirement.
\item \code{pre_install}: list column of character vectors. Commands run
before the package installation for this system requirement.
\item \code{system_packages}: list column of character vectors. Names of
system packages to install.
\item \code{post_install}: list column of character vectors. Commands run
after the package installation for this system requirement.
}
}
}
\description{
This function uses \code{\link[=new_pkg_installation_proposal]{new_pkg_installation_proposal()}} and its methods
to create an installation plan for one or more packages, and then print
their system requirements.
}
\examples{
\dontshow{if (!pkgdepends:::is_rcmd_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
sysreqs_install_plan(
  "tidyverse",
  config = list(sysreqs_platform = "ubuntu-22.04")
)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=new_pkg_installation_proposal]{new_pkg_installation_proposal()}} to actually install
packages, and potentially system requirements.

Other system requirements functions: 
\code{\link{sysreqs_check_installed}()},
\code{\link{sysreqs_db_list}()},
\code{\link{sysreqs_db_match}()},
\code{\link{sysreqs_db_update}()},
\code{\link{sysreqs_is_supported}()},
\code{\link{sysreqs_list_system_packages}()},
\code{\link{sysreqs_platforms}()}
}
\concept{system requirements functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sysreqsdocs.R
\name{sysreqs_list_system_packages}
\alias{sysreqs_list_system_packages}
\title{List installed system packages}
\usage{
sysreqs_list_system_packages()
}
\value{
Data frame with columns:
\itemize{
\item \code{status}. two or three characters, the notation of \code{dpkg} on Debian
based systems. \code{"ii"} means the package is correctly installed.
On \code{RPM} based systems it is always \code{"ii"} currently.
\item \code{package}: name of the system package.
\item \code{version}: installed version of the system package.
\item \code{capabilities}: list column of character vectors, the capabilities
provided by the package.
}
}
\description{
List installed system packages
}
\details{
This function uses the \code{sysreqs_platform} configuration option,
see \eval{man_config_link("Configuration")}. Set this if
pkgdepends does not detect your platform correctly.
}
\examples{
\dontshow{if (Sys.getenv("IN_PKGDOWN") == "true" && Sys.info()[["sysname"]] == "Linux") (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
sysreqs_list_system_packages()[1:10,]
\dontshow{\}) # examplesIf}
}
\seealso{
Other system requirements functions: 
\code{\link{sysreqs_check_installed}()},
\code{\link{sysreqs_db_list}()},
\code{\link{sysreqs_db_match}()},
\code{\link{sysreqs_db_update}()},
\code{\link{sysreqs_install_plan}()},
\code{\link{sysreqs_is_supported}()},
\code{\link{sysreqs_platforms}()}
}
\concept{system requirements functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dep-utils.R
\name{as_pkg_dependencies}
\alias{as_pkg_dependencies}
\title{Shorthands for dependency specifications}
\usage{
as_pkg_dependencies(deps)
}
\arguments{
\item{deps}{See below.}
}
\value{
A named list with two character vectors: \code{direct}, \code{indirect},
the dependency types to use for direct installations and dependent
packages.
}
\description{
Shorthands for dependency specifications
}
\details{
R packages may have various types of dependencies, see
\href{https://cran.r-project.org/doc/manuals/R-exts.html}{Writing R Extensions}.

pkgdepends groups dependencies into three groups:
\itemize{
\item hard dependencies: "Depends", "Imports", and "LinkingTo",
\item soft dependencies: "Suggests" and "Enhances",
\item extra dependencies, see below.
}

pkgdepends supports concise ways of specifying which types of
dependencies of a package should be installed.
It is similar to how \code{\link[utils:install.packages]{utils::install.packages()}} interprets its
\code{dependencies} argument.

You typically use one of these values:
\itemize{
\item \code{NA} or \code{"hard"} to install a package and its required dependencies,
\item \code{TRUE} to install all required dependencies, plus optional and development
dependencies.
}

If you need more flexibility, the full description of possible values for
the \code{deps} argument are:
\itemize{
\item \code{TRUE}: This means all hard dependencies plus \code{Suggests} for
direct installations, and hard dependencies only for dependent
packages.
\item \code{FALSE}: no dependencies are installed at all.
\item \code{NA} (any atomic type, so \code{NA_character_}, etc. as well): only hard
dependencies are installed.
See \code{\link[=pkg_dep_types_hard]{pkg_dep_types_hard()}}.
\item If a list with two entries named \code{direct} and \code{indirect}, it is taken
as the requested dependency types, for direct installations and
dependent packages.
\item If a character vector, then it is taken as the dependency types
for direct installations, and the hard dependencies are
used for the dependent packages.
}

If \code{"hard"} is included in the value or a list element, then it is replaced
by the hard dependency types.
If \code{"soft"} or \code{"all"} is included, then it is replaced by all
hard and soft dependency.
\subsection{Extra dependencies}{

pkgdepends supports extra dependency types for direct
installations not from CRAN-like repositories.
These are specified with a \verb{Config/Needs/} prefix in the \code{DESCRIPTION}
and they can contain package references, separated by commas.
For example you can specify packages that are only needed for the
pkgdown website of the package:

\if{html}{\out{<div class="sourceCode">}}\preformatted{Config/Needs/website: r-lib/pkgdown
}\if{html}{\out{</div>}}

To use these dependency types, you need to specify them in the
\code{deps} argument to
pkgdepends functions.

Note that \verb{Config/Needs/*} fields are currently \emph{not} used from CRAN
packages, and packages in CRAN-like repositories in general.

Usually you specify that a \verb{Config/Needs/*} dependency type should be
installed together with \code{"hard"} or \code{"all"}, to install all hard or
soft dependencies as well.
}
}
\seealso{
Other package dependency utilities: 
\code{\link{pkg_dep_types_hard}()}
}
\concept{package dependency utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dep-utils.R
\name{pkg_dep_types_hard}
\alias{pkg_dep_types_hard}
\alias{pkg_dep_types_soft}
\alias{pkg_dep_types}
\title{Possible package dependency types}
\usage{
pkg_dep_types_hard()

pkg_dep_types_soft()

pkg_dep_types()
}
\value{
A string vector of dependency types, capitalized.
}
\description{
Hard dependencies are needed for a package to load, soft dependencies
are optional.
}
\seealso{
Other package dependency utilities: 
\code{\link{as_pkg_dependencies}()}
}
\concept{package dependency utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan-deps.R
\name{scan_deps}
\alias{scan_deps}
\title{Scan R code for dependent packages}
\usage{
scan_deps(path = NULL, root = NULL)
}
\arguments{
\item{path}{Files and/or directories to scan. Defaults to the current
project, detected by finding the first parent directory of the current
working directory, that contains a file or directory called
DESCRIPTION, .git, .Rproj.user, renv.lock, or renv.
(Note that this is different from \code{renv::dependencies()}, which only
scans the current working directory by default!)

If \code{path} is not \code{NULL}, then only the specified files and directories
are scanned, the directories recursively. In this case the \code{root}
argument is used as the project root, to find \code{.gitignore} and
\code{.renvignore} files. All entries of \code{path} must be within the \code{root},
the project root.}

\item{root}{The root directory of the project. It is used to find the
\code{.gitignore} and \code{.renvignore} files. By default the same algorithm
is used to detect this as for \code{path}. If \code{path} is specified and it is
not within the detected or specified \code{root}, \code{scan_path()} throws an
error.}
}
\value{
Data frame with columns:
\itemize{
\item \code{path}: Path to the file in which the dependencies was found.
\item \code{package}: Detected package dependency. Typically a package name,
but it can also be a package reference, e.g. a package from GitHub.
\item \code{type}: Dependency type. It is \code{"prod"}, \code{"test"} or \code{"dev"}. See
'Dependency types' below.
\item \code{code}: The piece of code the dependency was extracted from.
\item \code{start_row}: Start row of the code the dependency was extracted
from.
\item \code{start_column}: Start column of the code the dependency was
extracted from.
\item \code{start_byte}: Start byte of the code the dependency was extracted
from.
}

Note the data frame may contain the same package multiple times, if it
was detected multiple times, e.g. multiple \code{library()} calls load the
same package.
}
\description{
Scan all R files of a project or directory for packages used within
them. It parses R code to find \code{library(package)}, \code{package::func()},
and similar calls that imply package dependencies. See details below.
}
\section{Detected dependencies}{
\code{scan_deps()} detects package dependencies from these R expressions:
\itemize{
\item \code{library()}, \code{require()}, \code{loadNamespace()} and \code{requireNamespace}
calls.
\item \code{::} and \code{:::} operators.
\item Any of the calls in this list in R code from R markdown or quarto
\code{R} and \code{Rscript} (case insensitive) code blocks or inline R code.
\item A dependency on the methods package is inferred from finding
\code{setClass()} and/or \code{setGeneric()} calls.
\item \code{xfun::pkg_attach()} and \code{xfun::pkg_attach2()} calls.
\item \code{pacman::p_load()} calls.
\item \code{modules::import()} and \code{modules::module()} calls.
\item \code{import::from()}, \code{import::here()} and \code{import::into()} calls.
\item \code{box::use()} calls.
\item \code{targets::tar_option_set(packages = ...)} calls.
\item Any of the calls in this list in R code from \code{glue::glue()} strings.
\item A dependency on the svglite package is inferred from
\code{ggplot2::ggsave()} calls saving \code{.svg} files.
\item Dependencies from \code{parsnip::set_engine()} calls, the default engine
to package mapping is:
\itemize{
\item \code{"glm"} -> stats,
\item \code{"glmnet"} -> glmnet,
\item \code{"keras"} -> keras,
\item \code{"kknn"} -> kknn,
\item \code{"nnet"} -> nnet,
\item \code{"rpart"} -> rpart,
\item \code{"spark"} -> sparklyr,
\item \code{"stan"} -> rstanarm.
You can override the default mapping by setting the
\code{renv.parsnip.engines} option to a named list.
}
\item A dependency on the xml2 package is inferred from using the
"Junit" reporter (\code{JunitReporter}) from the testthat package.
\item A dependency on the ragg package is inferred from setting the default
knitr device (\code{dev} option) to \code{"ragg_png"}.
\item A dependency on the hexbin package is inferred from using
\code{ggplot2::geom_hex()}.
\item A custom symbol name to package name mapping can be defined in the
\code{renv.dependencies.database} option. This must be a named list of
named lists, where the outer names are package names, the inner names
are function or object names, and the values are package names. E.g.

\if{html}{\out{<div class="sourceCode">}}\preformatted{options(renv.dependencies.database = list(
  ggplot2 = list(geom_hex = "hexbin"),
  testthat = list(JunitReporter = "xml2")
))
}\if{html}{\out{</div>}}
}
}

\section{Dependency types}{
\code{scan_deps()} classifies package dependencies into three groups, based
on which files they were found:
\itemize{
\item Production dependencies: \code{"prod"}.
\item Test dependencies: \code{"test"}.
\item Development dependencies: \code{"dev"}.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz-pkgdepends-config.R
\name{pkg_config}
\alias{pkg_config}
\alias{current_config}
\alias{pkgdepends-config}
\alias{pkgdepends_config}
\title{pkgdepends configuration}
\usage{
current_config()
}
\description{
Configuration entries for several pkgdepends classes.
}
\details{
pkgdepends configuration is set from several source. They are, in the
order of preference:
\itemize{
\item Function arguments, e.g. the \code{config} argument of
\code{\link[=new_pkg_installation_proposal]{new_pkg_installation_proposal()}}.
\item Global options, set via \code{\link[=options]{options()}}. The name of the global option
is the \code{pkg.} prefix plus the name of the pkgdepends configuration
entry. E.g. \code{pkg.platforms}.
\item Environment variables. The name of the environment variable is the
\code{PKG_} prefix, plus the name of the pkgdepends configuration entry, in
uppercase. E.g. \code{PKG_PLATFORMS}.
\item Default values.
}

Not all classes use all entries. E.g. a \code{\link{pkg_download_proposal}} is not
concerned about package libraries, so it'll ignore the \code{library}
configuration entry.

Call \code{current_config()} to print the current configuration.
}
\section{Configuration entries}{
\itemize{
\item \code{build_vignettes}: Whether to build vignettes for package trees.
This is only used if the package is obtained from a package tree,
and not from a source (or binary) package archive. By default
vignettes are not built in this case. If you set this to \code{TRUE},
then you need to make sure that the vignette builder packages are
available, as these are not installed by default currently.
\item \code{cache_dir}: Directory to download the packages to. Defaults to a temporary
directory within the R session temporary directory, see
\code{\link[base:tempfile]{base::tempdir()}}.
\item \code{cran_mirror}: CRAN mirror to use. Defaults to the \code{repos} option
(see \code{\link[base:options]{base::options()}}), if that's not set then
\verb{https://cran.rstudio.com}.
\item \code{dependencies}: Dependencies to consider or download or install.
Defaults to the hard dependencies, see
\code{\link[=pkg_dep_types_hard]{pkg_dep_types_hard()}}. The following values are
supported in the \code{PKG_DEPENDENCIES} environment variable:
\code{"TRUE"}, \code{"FALSE"}, \code{"NA"}, or a semicolon separated list of
dependency types. See \code{\link[=as_pkg_dependencies]{as_pkg_dependencies()}} for
details.
\item \code{git_submodules}: Whether or not to update submodules in git repositories. This
affects \verb{git::} and \verb{gitlab::} package sources only.
If the R package is in a subdirectory then only the submodules
within that directory are updated. If a submodule appears in
\code{.Rbuildignore}, then it is skipped.
\item \code{include_linkingto}: Whether to always include \code{LinkingTo} dependencies in the solution
of and installation, even if they are needed because the packages
are installed from binaries. This is sometimes useful, see e.g.
\url{https://github.com/r-lib/pak/issues/485} for an example use case.
\item \code{library}: Package library to install packages to. It is also used for
already installed packages when considering dependencies in
\link[=pkg_deps]{dependency lookup} or
\link[=pkg_installation_proposal]{package installation}. Defaults to the
first path in \code{\link[=.libPaths]{.libPaths()}}.
\item \code{metadata_cache_dir}: Location of metadata replica of
\code{\link[pkgcache:cranlike_metadata_cache]{pkgcache::cranlike_metadata_cache}}. Defaults to a temporary
directory within the R session temporary directory, see
\code{\link[base:tempfile]{base::tempdir()}}.
\item \code{metadata_update_after}: A time interval as a \link{difftime} object. pkgdepends will update the
metadata cache if it is older than this. The default is one day.
The \code{PKG_METADATA_UPDATE_AFTER} environment variable may be set
in seconds (\code{s} suffix), minutes (\code{m} suffix), hours (\code{h} suffix),
or days (\code{d} suffix). E.g: \verb{1d} means one day.
\item \code{package_cache_dir}: Package cache location of \code{\link[pkgcache:package_cache]{pkgcache::package_cache}}. The default
is the pkgcache default.
\item \code{platforms}: Character vector of platforms to \emph{download} or \emph{install} packages
for. See \code{\link[=default_platforms]{default_platforms()}} for possible platform
names. Defaults to the platform of the current R session, plus
\code{"source"}.
\item \code{r_versions}: Character vector, R versions to download or install
packages for. It defaults to the current R version.
\item \code{sysreqs}: Whether to automatically look up and install system requirements.
If \code{TRUE}, then \verb{r pak_or_pkgdepends()} will try to install required
system packages. If \code{FALSE}, then system requirements are still
printed (including OS packages on supported platforms), but they
are not installed.
By default it is \code{TRUE} on supported platforms,
if the current user is the root user or password-less \code{sudo} is
configured for the current user.
\item \code{sysreqs_db_update}: Whether to try to update the system requirements database from
GitHub. If the update fails, then the cached or the build-in
database if used. Defaults to TRUE.
\item \code{sysreqs_db_update_timeout}: Timeout for the system requirements database update.
Defaults to five seconds, except if the \code{CI} environment variable
is set, then it is one minute.
\item \code{sysreqs_dry_run}: If \code{TRUE}, then pkgdepends only prints the system commands to
install system requirements, but does not execute them.
\item \code{sysreqs_platform}: The platform to use for system requirements lookup. On Linux, where
system requirements are currently supported, it must be a string
containing the distribution name and release, separated by a dash.
E.g.: \code{"ubuntu-22.04"}, or \code{"rhel-9"}.
\item \code{sysreqs_rspm_repo_id}: Posit Package Manager (formerly RStudio Package Manager) repository
id to use for CRAN system requirements lookup. Defaults to the
\code{RSPM_REPO_ID} environment variable, if set. If not set, then it
defaults to \code{1}.
\item \code{sysreqs_rspm_url}: Root URL of Posit Package Manager (formerly RStudio Package
Manager) for system requirements lookup. By default the \code{RSPM_ROOT}
environment variable is used, if set. If not set,
it defaults to \verb{https://packagemanager.posit.co}.
\item \code{sysreqs_sudo}: Whether to use \code{sudo} to install system requirements,
on Unix. By default it is \code{TRUE} on Linux if the effective user id
of the current process is not the \code{root} user.
\item \code{sysreqs_update}: Whether to try to update system packages that are already installed.
It defaults to \code{TRUE} on CI systems: if the \code{CI} environment
variable is set to \code{true}.
\item \code{sysreqs_verbose}: Whether to echo the output of system requirements installation.
Defaults to \code{TRUE} if the \code{CI} environment variable is set.
\item \code{use_bioconductor}: Whether to automatically use the Bioconductor repositories.
Defaults to \code{TRUE}.
\item \code{windows_archs}: Character scalar specifying which architectures
to download/install for on Windows. Its possible values are:
\itemize{
\item \code{"prefer-x64"}: Generally prefer x64 binaries. If the current R
session is \code{x64}, then we download/install x64 packages.
(These packages might still be multi-architecture binaries!)
If the current R session is \code{i386}, then we download/install
packages for both architectures. This might mean compiling
packages from source if the binary packages are for \code{x64} only,
like the CRAN Windows binaries for R 4.2.x currently.
\code{"prefer-x64"} is the default for R 4.2.0 and later.
\item \code{"both"}: Always download/install packages for both \code{i386} and
\code{x64} architectures. This might need compilation from source
if the available binaries are for \code{x64} only, like the CRAN
Windows binaries for R 4.2.x currently. \code{"both"} is the default
for R 4.2.0 and earlier.
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkg-installation.R
\name{new_pkg_installation_proposal}
\alias{new_pkg_installation_proposal}
\alias{pkg_installation_proposal}
\title{R6 class for package download and installation.}
\usage{
new_pkg_installation_proposal(refs, config = list(), ...)
}
\arguments{
\item{refs}{Package names or references. See
\link[=pkg_refs]{'Package references'} for the syntax.}

\item{config}{Configuration options, a named list. See
\link[=pkgdepends-config]{'Configuration'}. If it does not include \code{library}, then
\code{.libPaths()[1]} is added as \code{library}.}

\item{...}{Additional arguments, passed to
\href{#method-new}{\code{pkg_installation_proposal$new()}}.}
}
\value{
\code{new_pkg_installation_proposal()} returns a new
\code{pkg_installation_proposal} object.
}
\description{
Download and install R packages, with their dependencies, from various
sources.
}
\details{
\code{new_pkg_installation_proposal()} creates a new object from the
\code{pkg_installation_proposal} class. The advantage of
\code{new_pkg_installation_proposal()} compared to using the
\link{pkg_installation_proposal} constructor directly is that it avoids
making pkgdepends a build time dependency.

Typical workflow to install a set of packages:
\enumerate{
\item Create a \code{pkg_installation_proposal} object with
\code{new_pkg_installation_proposal()}.
\item Resolve all possible dependencies with
\href{#method-resolve}{\code{pkg_installation_proposal$resolve()}}.
\item Solve the package dependencies, to get an installation plan, with
\href{#method-solve}{\code{pkg_installation_proposal$solve()}}.
\item Download all files with
\href{#method-download}{\code{pkg_installation_proposal$download()}}.
\item Install the downloaded files with
\href{#methods-install}{\code{pkg_installation_proposal$install()}}.
}
}
\examples{
\dontrun{
pdi <- new_pkg_installation_proposal(
  "pak",
  config = list(library = tempfile())
)
pdi

pdi$resolve()
pdi

pdi$solve()
pdi

pdi$download()
pdi
}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
pdi <- new_pkg_installation_proposal(
  "r-lib/pkgdepends",
  config = list(library = tempfile()))
pdi
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
pdi <- new_pkg_installation_proposal("r-lib/pkgdepends")
pdi$get_refs()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
pdi <- new_pkg_installation_proposal(
  "pak",
  config = list(library = tempfile())
)
pdi$get_config()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
pdi <- new_pkg_installation_proposal(
  "r-lib/pkgdepends",
  config = list(library = tempfile())
)
pdi$get_solve_policy()
pdi$set_solve_policy("upgrade")
pdi$get_solve_policy()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
pdi <- new_pkg_installation_proposal(
  "r-lib/pkgdepends",
  config = list(library = tempfile())
)
pdi$get_solve_policy()
pdi$set_solve_policy("upgrade")
pdi$get_solve_policy()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method print
pdi <- new_pkg_installation_proposal(
  "pak",
  config = list(library = tempfile())
)
pdi

pdi$resolve()
pdi

pdi$solve()
pdi

pdi$download()
pdi
\dontshow{\}) # examplesIf}

## ------------------------------------------------
## Method `pkg_installation_proposal$resolve`
## ------------------------------------------------

\dontrun{
pdi <- new_pkg_installation_proposal(
  "pak",
  config = list(library = tempfile())
)

pdi$resolve()
pdi$get_resolution()
}

## ------------------------------------------------
## Method `pkg_installation_proposal$get_resolution`
## ------------------------------------------------

\dontrun{
pdi <- new_pkg_installation_proposal(
  "r-lib/pkgdepends",
  config = list(library = tempfile())
)
pdi$resolve()
pdi$get_resolution()
}

## ------------------------------------------------
## Method `pkg_installation_proposal$solve`
## ------------------------------------------------

\dontrun{
pdi <- new_pkg_installation_proposal(
  "r-lib/pkgdepends",
  config = list(library = tempfile())
)
pdi$resolve()
pdi$solve()
pdi$get_solution()
}

## ------------------------------------------------
## Method `pkg_installation_proposal$get_solution`
## ------------------------------------------------

\dontrun{
pdi <- new_pkg_installation_proposal(
  "r-lib/pkgdepends",
  config = list(library = tempfile())
)
pdi$resolve()
pdi$solve()
pdi$get_solution()
}

## ------------------------------------------------
## Method `pkg_installation_proposal$show_solution`
## ------------------------------------------------

\dontrun{
pdi <- new_pkg_installation_proposal(
  "r-lib/pkgdepends",
  config = list(library = tempfile())
)
pdi$resolve()
pdi$solve()
pdi$get_solution()
pdi$show_solution()
}

## ------------------------------------------------
## Method `pkg_installation_proposal$stop_for_solution_error`
## ------------------------------------------------

\dontrun{
# This is an error, because the packages conflict:
pdi <- new_pkg_installation_proposal(
  c("r-lib/pak", "cran::pak"),
  config = list(library = tempfile())
)
pdi$resolve()
pdi$solve()
pdi
# This fails:
# pdi$stop_for_solution_error()
}

## ------------------------------------------------
## Method `pkg_installation_proposal$draw`
## ------------------------------------------------

\dontrun{
pdi <- new_pkg_installation_proposal(
  "pak",
  config = list(library = tempfile())
)
pdi$resolve()
pdi$solve()
pdi$draw()
}

## ------------------------------------------------
## Method `pkg_installation_proposal$download`
## ------------------------------------------------

\dontrun{
pdi <- new_pkg_installation_proposal(
  c("r-lib/pak", "cran::pak"),
  config = list(library = tempfile())
)
pdi$resolve()
pdi$solve()
pdi$download()
pdi$get_downloads()
}

## ------------------------------------------------
## Method `pkg_installation_proposal$get_downloads`
## ------------------------------------------------

\dontrun{
pdi <- new_pkg_installation_proposal(
  c("r-lib/pak", "cran::pak"),
  config = list(library = tempfile())
)
pdi$resolve()
pdi$solve()
pdi$download()
pdi$get_downloads()
}

## ------------------------------------------------
## Method `pkg_installation_proposal$get_install_plan`
## ------------------------------------------------

\dontrun{
pdi <- new_pkg_installation_proposal(
  "pak",
  config = list(library = tempfile())
)
pdi$resolve()
pdi$solve()
pdi$download()
pdi$get_install_plan()
}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-pkg_installation_proposal-new}{\code{pkg_installation_proposal$new()}}
\item \href{#method-pkg_installation_proposal-get_refs}{\code{pkg_installation_proposal$get_refs()}}
\item \href{#method-pkg_installation_proposal-get_config}{\code{pkg_installation_proposal$get_config()}}
\item \href{#method-pkg_installation_proposal-resolve}{\code{pkg_installation_proposal$resolve()}}
\item \href{#method-pkg_installation_proposal-async_resolve}{\code{pkg_installation_proposal$async_resolve()}}
\item \href{#method-pkg_installation_proposal-get_resolution}{\code{pkg_installation_proposal$get_resolution()}}
\item \href{#method-pkg_installation_proposal-get_solve_policy}{\code{pkg_installation_proposal$get_solve_policy()}}
\item \href{#method-pkg_installation_proposal-set_solve_policy}{\code{pkg_installation_proposal$set_solve_policy()}}
\item \href{#method-pkg_installation_proposal-solve}{\code{pkg_installation_proposal$solve()}}
\item \href{#method-pkg_installation_proposal-get_solution}{\code{pkg_installation_proposal$get_solution()}}
\item \href{#method-pkg_installation_proposal-show_solution}{\code{pkg_installation_proposal$show_solution()}}
\item \href{#method-pkg_installation_proposal-get_sysreqs}{\code{pkg_installation_proposal$get_sysreqs()}}
\item \href{#method-pkg_installation_proposal-show_sysreqs}{\code{pkg_installation_proposal$show_sysreqs()}}
\item \href{#method-pkg_installation_proposal-stop_for_solution_error}{\code{pkg_installation_proposal$stop_for_solution_error()}}
\item \href{#method-pkg_installation_proposal-create_lockfile}{\code{pkg_installation_proposal$create_lockfile()}}
\item \href{#method-pkg_installation_proposal-draw}{\code{pkg_installation_proposal$draw()}}
\item \href{#method-pkg_installation_proposal-download}{\code{pkg_installation_proposal$download()}}
\item \href{#method-pkg_installation_proposal-async_download}{\code{pkg_installation_proposal$async_download()}}
\item \href{#method-pkg_installation_proposal-get_downloads}{\code{pkg_installation_proposal$get_downloads()}}
\item \href{#method-pkg_installation_proposal-stop_for_download_error}{\code{pkg_installation_proposal$stop_for_download_error()}}
\item \href{#method-pkg_installation_proposal-install}{\code{pkg_installation_proposal$install()}}
\item \href{#method-pkg_installation_proposal-install_sysreqs}{\code{pkg_installation_proposal$install_sysreqs()}}
\item \href{#method-pkg_installation_proposal-get_install_plan}{\code{pkg_installation_proposal$get_install_plan()}}
\item \href{#method-pkg_installation_proposal-format}{\code{pkg_installation_proposal$format()}}
\item \href{#method-pkg_installation_proposal-print}{\code{pkg_installation_proposal$print()}}
\item \href{#method-pkg_installation_proposal-clone}{\code{pkg_installation_proposal$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-new"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-new}{}}}
\subsection{Method \code{new()}}{
Create a new \code{pkg_installation_proposal} object. Consider using
\code{new_pkg_installation_proposal()} instead of calling the constructor
directly.

The returned object can be used to look up (recursive) dependencies
of R packages from various sources, and then download and install
the package files.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$new(
  refs,
  config = list(),
  policy = c("lazy", "upgrade"),
  remote_types = NULL
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{refs}}{Package names or references. See
\link[=pkg_refs]{'Package references'} for the syntax.}

\item{\code{config}}{Configuration options, a named list. See
\link[=pkgdepends-config]{'Configuration'}. It needs to include the package
library to install to, in \code{library}.}

\item{\code{policy}}{Solution policy. See \link[=pkg_solution]{'The dependency solver'}.}

\item{\code{remote_types}}{Custom remote ref types, this is for advanced
use, and experimental currently.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-get_refs"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-get_refs}{}}}
\subsection{Method \code{get_refs()}}{
The package refs that were used to create the
\code{pkg_installation_proposal} object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$get_refs()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A character vector of package refs that were used to create the
\code{pkg_installation_proposal} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-get_config"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-get_config}{}}}
\subsection{Method \code{get_config()}}{
Configuration options for the \code{pkg_installation_proposal} object. See
\link[=pkgdepends-config]{'Configuration'} for details.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$get_config()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Named list. See \link[=pkgdepends-config]{'Configuration'} for the configuration
options.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-resolve"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-resolve}{}}}
\subsection{Method \code{resolve()}}{
Resolve the dependencies of the specified package references. This
usually means downloading metadata from CRAN and Bioconductor,
unless already cached, and also from GitHub if GitHub refs were
included, either directly or indirectly. See
\link[=pkg_resolution]{'Dependency resolution'} for details.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$resolve()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
The \code{pkg_installation_proposal} object, invisibly.
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
pdi <- new_pkg_installation_proposal(
  "pak",
  config = list(library = tempfile())
)

pdi$resolve()
pdi$get_resolution()
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-async_resolve"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-async_resolve}{}}}
\subsection{Method \code{async_resolve()}}{
The same as \href{#method-resolve}{\code{resolve()}}, but asynchronous. This
method is for advanced use.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$async_resolve()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A deferred value.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-get_resolution"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-get_resolution}{}}}
\subsection{Method \code{get_resolution()}}{
Query the result of the dependency resolution. This method can be
called after \href{#method-resolve}{\code{resolve()}} has completed.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$get_resolution()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A \link{pkg_resolution_result} object, which is also a data frame. See
\link[=pkg_resolution]{'Dependency resolution'} for its columns.
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
pdi <- new_pkg_installation_proposal(
  "r-lib/pkgdepends",
  config = list(library = tempfile())
)
pdi$resolve()
pdi$get_resolution()
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-get_solve_policy"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-get_solve_policy}{}}}
\subsection{Method \code{get_solve_policy()}}{
Returns the current policy of the dependency solver.
See \link[=pkg_solution]{'The dependency solver'} for details.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$get_solve_policy()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A character vector of length one.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-set_solve_policy"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-set_solve_policy}{}}}
\subsection{Method \code{set_solve_policy()}}{
Set the current policy of the dependency solver.
If the object already contains a solution and the new policy is
different than the old policy, then the solution is deleted.
See \link[=pkg_solution]{'The dependency solver'} for details.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$set_solve_policy(policy = c("lazy", "upgrade"))}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{policy}}{Policy to set.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-solve"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-solve}{}}}
\subsection{Method \code{solve()}}{
Solve the package dependencies. Out of the resolved dependencies,
it works out a set of packages, that can be installed together to
create a functional installation. The set includes all directly
specified packages, and all required (or suggested, depending on
the configuration) packages as well. It includes every package at
most once. See \link[=pkg_solution]{'The dependency solver'} for details.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$solve()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
The \code{pkg_installation_proposal} object itself, invisibly.
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
pdi <- new_pkg_installation_proposal(
  "r-lib/pkgdepends",
  config = list(library = tempfile())
)
pdi$resolve()
pdi$solve()
pdi$get_solution()
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-get_solution"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-get_solution}{}}}
\subsection{Method \code{get_solution()}}{
Returns the solution of the package dependencies.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$get_solution()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A \link{pkg_solution_result} object, which is a list. See
\link{pkg_solution_result} for details.
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
pdi <- new_pkg_installation_proposal(
  "r-lib/pkgdepends",
  config = list(library = tempfile())
)
pdi$resolve()
pdi$solve()
pdi$get_solution()
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-show_solution"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-show_solution}{}}}
\subsection{Method \code{show_solution()}}{
Show the solution on the screen.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$show_solution(key = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{key}}{Whether to show the key to the package list annotation.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A \link{pkg_solution_result} object, which is a list. See
\link{pkg_solution_result} for details.
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
pdi <- new_pkg_installation_proposal(
  "r-lib/pkgdepends",
  config = list(library = tempfile())
)
pdi$resolve()
pdi$solve()
pdi$get_solution()
pdi$show_solution()
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-get_sysreqs"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-get_sysreqs}{}}}
\subsection{Method \code{get_sysreqs()}}{
Query and categorize system requirements.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$get_sysreqs()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-show_sysreqs"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-show_sysreqs}{}}}
\subsection{Method \code{show_sysreqs()}}{
Show system requirements for the packages in the solution.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$show_sysreqs()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-stop_for_solution_error"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-stop_for_solution_error}{}}}
\subsection{Method \code{stop_for_solution_error()}}{
Error if the dependency solver failed to find a consistent set of
packages that can be installed together.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$stop_for_solution_error()}\if{html}{\out{</div>}}
}

\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
# This is an error, because the packages conflict:
pdi <- new_pkg_installation_proposal(
  c("r-lib/pak", "cran::pak"),
  config = list(library = tempfile())
)
pdi$resolve()
pdi$solve()
pdi
# This fails:
# pdi$stop_for_solution_error()
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-create_lockfile"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-create_lockfile}{}}}
\subsection{Method \code{create_lockfile()}}{
Create a lock file that contains the information to perform
the installation later, possibly in another R session.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$create_lockfile(path = "pkg.lock", version = 1)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{Name of the lock file. The default is \code{pkg.lock} in the
current working directory.}

\item{\code{version}}{Only version 1 is supported currently.}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Note, since the URLs of CRAN and most CRAN-like repositories change
over time, in practice you cannot perform the plan of the lock file
\emph{much} later. For example, binary packages of older package version
are removed, and won't be found.

Similarly, for \verb{url::} remote types, the URL might hold an updated
version of the package, compared to when the lock file was created.
Should this happen, pkgdepends prints a warning, but it will try
to continue the installation. The installation might fail if the
updated package has different (e.g. new) dependencies.

Currently the intended use case of lock files in on CI systems, to
facilitate caching. The (hash of the) lock file provides a good key
for caching systems.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-draw"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-draw}{}}}
\subsection{Method \code{draw()}}{
Draw a tree of package dependencies. It returns a \code{tree} object, see
\code{\link[cli:tree]{cli::tree()}}. Printing this object prints the dependency tree to the
screen.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$draw()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A \code{tree} object from the cli package, see \code{\link[cli:tree]{cli::tree()}}.
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
pdi <- new_pkg_installation_proposal(
  "pak",
  config = list(library = tempfile())
)
pdi$resolve()
pdi$solve()
pdi$draw()
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-download"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-download}{}}}
\subsection{Method \code{download()}}{
Download all packages that are part of the solution. It uses the
package cache in the pkgcache package by default, to avoid downloads
if possible.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$download()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
The \code{pkg_installation_proposal} object itself, invisibly.
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
pdi <- new_pkg_installation_proposal(
  c("r-lib/pak", "cran::pak"),
  config = list(library = tempfile())
)
pdi$resolve()
pdi$solve()
pdi$download()
pdi$get_downloads()
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-async_download"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-async_download}{}}}
\subsection{Method \code{async_download()}}{
The same as \href{#method-download}{\code{download()}}, but asynchronous.
This method is for advanced use.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$async_download()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A deferred value.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-get_downloads"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-get_downloads}{}}}
\subsection{Method \code{get_downloads()}}{
Returns the summary of the package downloads.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$get_downloads()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A \link{pkg_download_result} object, which is a list. See
\link{pkg_download_result} for details.
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
pdi <- new_pkg_installation_proposal(
  c("r-lib/pak", "cran::pak"),
  config = list(library = tempfile())
)
pdi$resolve()
pdi$solve()
pdi$download()
pdi$get_downloads()
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-stop_for_download_error"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-stop_for_download_error}{}}}
\subsection{Method \code{stop_for_download_error()}}{
Throw and error if the some of the downloads have failed for the
most recent
\href{#method-download}{\code{pkg_installation_proposal$download()}} call.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$stop_for_download_error()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-install"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-install}{}}}
\subsection{Method \code{install()}}{
Install the downloaded packages. It calls \code{\link[=install_package_plan]{install_package_plan()}}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$install()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
The return value of \code{\link[=install_package_plan]{install_package_plan()}}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-install_sysreqs"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-install_sysreqs}{}}}
\subsection{Method \code{install_sysreqs()}}{
Install system requirements. It does nothing if system requirements
are turned off.
Create an installation plan for the downloaded packages.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$install_sysreqs()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-get_install_plan"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-get_install_plan}{}}}
\subsection{Method \code{get_install_plan()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$get_install_plan()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
An installation plan, see \link[=install_plans]{'Installation plans'} for
the format.
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
pdi <- new_pkg_installation_proposal(
  "pak",
  config = list(library = tempfile())
)
pdi$resolve()
pdi$solve()
pdi$download()
pdi$get_install_plan()
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-format"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-format}{}}}
\subsection{Method \code{format()}}{
Format a \code{pkg_installation_proposal} object, typically for printing.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$format(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{not used currently.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A character vector, each element should be a line in the printout.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-print"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-print}{}}}
\subsection{Method \code{print()}}{
Prints a \code{pkg_installation_proposal} object to the screen.

The printout includes:
\itemize{
\item The package refs.
\item The policy of the dependency solver.
\item Whether the object has the solved dependencies.
\item Whether the solution had errors.
\item Whether the object has downloads.
\item Whether the downloads had errors.
\item Advice on which methods to call next.
}

See the example below.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{not used currently.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The \code{pkg_installation_proposal} object itself, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_installation_proposal-clone"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_installation_proposal-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_installation_proposal$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/git-protocol.R
\name{git_list_refs}
\alias{git_list_refs}
\title{List references in a remote git repository}
\usage{
git_list_refs(url, prefixes = NULL)
}
\arguments{
\item{url}{Repository URL, e.g. \verb{https://github.com/r-lib/pak.git}.
It might include authentication information, e.g. a GitHub token.}

\item{prefixes}{If not \code{NULL}, then only references of which one of
\code{prefixes} is a prefix, are listed.}
}
\value{
A list with entries: \code{refs} and \code{caps}.
\code{caps} is a character vector of capabilities advertised by the server.
\code{refs} is a data frame of git refs, it has columns \code{ref} and \code{hash}.
}
\description{
List references in a remote git repository
}
\details{
\itemize{
\item Branches have references named \verb{refs/heads/<branch>}, e.g.
\code{refs/heads/main}.
\item Tags have references named \verb{refs/tags/<tag>}, e.g. \code{refs/tags/v1.0.2}.
\item On GitHub pull requests have references named
\verb{refs/pull/<pr-number>/head}, e.g. \code{refs/pull/37/head}.
For open pull requests there should be a \verb{refs/pull/<pr-number>/merge}
reference as well, which is the branch after the pull request has been
merged.
\item There is a special reference called \code{HEAD}. This points to the default
branch on GitHub.
}
}
\section{Examples}{


\if{html}{\out{<div class="sourceCode r">}}\preformatted{git_list_refs("https://github.com/r-lib/filelock.git")
#> $refs
#>                                       ref
#> 1                                    HEAD
#> 2                   refs/heads/cran-1-0-2
#> 3  refs/heads/fix/testthat-error-parallel
#> 4                     refs/heads/gh-pages
#> 5                         refs/heads/main
#> 6                  refs/heads/refcounting
#> 7                       refs/pull/14/head
#> 8                       refs/pull/37/head
#> 9                      refs/pull/37/merge
#> 10                      refs/pull/38/head
#> 11                      refs/pull/40/head
#> 12                       refs/tags/v1.0.2
#>                                        hash
#> 1  0331c49c0b17845468b5f3e2da75680475bc9989
#> 2  dfc615b14dfc030df3f44bd377e752728f76df3d
#> 3  6f4b4bfcc1d3442d63496c7640984c0d9570a05a
#> 4  2956fbdcbab65b75e3347048a9dc18b15fe377d2
#> 5  0331c49c0b17845468b5f3e2da75680475bc9989
#> 6  48e5bbe0e031897c7f0e194edbf6897f813fbaee
#> 7  48e5bbe0e031897c7f0e194edbf6897f813fbaee
#> 8  6f4b4bfcc1d3442d63496c7640984c0d9570a05a
#> 9  74b253543f6d2e7213a8174ebbe8e397b9cc866b
#> 10 c7e58188a55844e6e75cd38cf81388c3452b1259
#> 11 10421f0b7117244002d377d7a9699c2742abdfcc
#> 12 9fdba75a62facaa3e818902f58891166e45eabe9
#> 
#> $caps
#>  [1] "multi_ack"                      "thin-pack"                     
#>  [3] "side-band"                      "side-band-64k"                 
#>  [5] "ofs-delta"                      "shallow"                       
#>  [7] "deepen-since"                   "deepen-not"                    
#>  [9] "deepen-relative"                "no-progress"                   
#> [11] "include-tag"                    "multi_ack_detailed"            
#> [13] "allow-tip-sha1-in-want"         "allow-reachable-sha1-in-want"  
#> [15] "no-done"                        "symref=HEAD:refs/heads/main"   
#> [17] "filter"                         "object-format=sha1"            
#> [19] "agent=git/github-g8119c27a9d8d"
}\if{html}{\out{</div>}}

List only references having a certain prefix:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{git_list_refs("https://github.com/r-lib/filelock.git", "refs/heads/main")
#> $refs
#>               ref                                     hash
#> 1 refs/heads/main 211dde89c4974b295aa02c5c77b5570ead0fc458
#> 
#> $caps
#> [1] "version 2"                          "agent=git/github-1c5ee477c782"     
#> [3] "ls-refs=unborn"                     "fetch=shallow wait-for-done filter"
#> [5] "server-option"                      "object-format=sha1"
}\if{html}{\out{</div>}}

Various services:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{git_list_refs("https://gitlab.com/Linaro/tuxmake.git", "HEAD")
#> $refs
#>    ref                                     hash
#> 1 HEAD 9441c1f6ae68148882e9d15ab2d6b3008f4c2c19
#> 
#> $caps
#> [1] "version 2"                          "agent=git/2.40.0.gl1"              
#> [3] "ls-refs=unborn"                     "fetch=shallow wait-for-done filter"
#> [5] "server-option"                      "object-format=sha1"                
#> [7] "object-info"
git_list_refs("https://bitbucket.org/gaborcsardi/cli.git")
#> $refs
#>                                        ref
#> 1                                     HEAD
#> 2                           refs/heads/ask
#> 3                    refs/heads/cran-3.0.0
#> 4                   refs/heads/cran/v3.0.1
#> 5      refs/heads/feature/cli-capabilities
#> 6   refs/heads/feature/cli-condition-width
#> 7  refs/heads/feature/format-error-options
#> 8           refs/heads/feature/rstudio-job
#> 9          refs/heads/feature/span-bullets
#> 10                refs/heads/feature/table
#> 11            refs/heads/fix/pkgdown-2.0.0
#> 12                     refs/heads/gh-pages
#> 13                         refs/heads/main
#> 14    refs/heads/use-glue-literal-argument
#> 15                        refs/tags/v1.0.1
#> 16                        refs/tags/v1.1.0
#> 17                    refs/tags/v1.1.0-pre
#> 18                        refs/tags/v2.0.0
#> 19                        refs/tags/v2.0.1
#> 20                        refs/tags/v2.0.2
#> 21                        refs/tags/v2.1.0
#> 22                        refs/tags/v2.2.0
#> 23                        refs/tags/v2.3.0
#> 24                        refs/tags/v2.3.1
#> 25                        refs/tags/v2.4.0
#> 26                        refs/tags/v2.5.0
#> 27                        refs/tags/v3.0.0
#> 28                        refs/tags/v3.0.1
#> 29                        refs/tags/v3.1.0
#>                                        hash
#> 1  928ff3e6a52a08bbabadc57027c43de13b0d7601
#> 2  736e5824c082e94fc1acc68fb604fa66b19d1826
#> 3  75f3f185c116ff69f3abd56bf89fa9032a9eb965
#> 4  722cd638ebe4529db20442962cd2d7dc56c694d7
#> 5  b6ebff6771ff9ee8e5b19f8b6c20e142182ab866
#> 6  ee931427a9ee51bdfd837bd03c22d2d60ad81946
#> 7  28d4efb9b8875e933d682285b65ab05c61f6aef2
#> 8  8ddc6e68b970f14d45af6109108b17352dc8a4e7
#> 9  bfaabf3658c229d406591b4436991e37720476cf
#> 10 0d4089fc2f6cf8c7364f02698cf257363e877660
#> 11 be878b57b6f9978953916b0a1d7272f104e27309
#> 12 928ff3e6a52a08bbabadc57027c43de13b0d7601
#> 13 f6794c4a2d845b6706a41c58f63b4acbbc663968
#> 14 5c2904c05e8f809d0136e2253e2e6ea4b7743992
#> 15 6e0ffb0de1b54953583e01068a8052ae4cbe72aa
#> 16 d4f0e8618aba3475974394075f3f5384695d1d28
#> 17 57fb26d1c4582fc73a3a8531aa534948f88e96bc
#> 18 9a133422f200f7aa0cdc78b28eb42770cdfe4b35
#> 19 31aaccbdb3904c5b681175483a95688940345725
#> 20 19f3b864f4ee04c31627a3246dbea44bc07b7967
#> 21 a914e326aea8a9ec7d07f7accd64bb624065ba8a
#> 22 e0d0668030b482c1bda1fce66db63f402c4e9c5a
#> 23 e4c9180c5b53f68c3f938473c0d5871ceefe9991
#> 24 5c0c8a88549050e156ea4469adf536e1c0ac0881
#> 25 251c84f298b46e0529dd992b9ec86da4aaaf4fa8
#> 26 17fd0dc202a51aa82a7704786712d9f77c690c4d
#> 27 df963095361d77322676260e60c549d8807ded1d
#> 28 525e785e8b34e889707bbbe354287bd4a60e8fcb
#> 29 e40cfd4eedad58c22c3e7bcfdba6f23f11a79e13
#> 
#> $caps
#>  [1] "multi_ack"                       "thin-pack"                      
#>  [3] "side-band"                       "side-band-64k"                  
#>  [5] "ofs-delta"                       "shallow"                        
#>  [7] "deepen-since"                    "deepen-not"                     
#>  [9] "deepen-relative"                 "no-progress"                    
#> [11] "include-tag"                     "multi_ack_detailed"             
#> [13] "allow-tip-sha1-in-want"          "allow-reachable-sha1-in-want"   
#> [15] "no-done"                         "symref=HEAD:refs/heads/gh-pages"
#> [17] "object-format=sha1"              "agent=git/2.32.0"
git_list_refs("https://git.savannah.nongnu.org/git/administration/savane.git")
#> $refs
#>                         ref                                     hash
#> 1                      HEAD bdfce5e39c3ebb36b748925dfc8a0cde483fda6a
#> 2           refs/heads/i18n e5f2c309b038789a82ab5a54d723a3564f1bde58
#> 3         refs/heads/master bdfce5e39c3ebb36b748925dfc8a0cde483fda6a
#> 4    refs/tags/release-3.10 e1db1a011effeb7d16a5af9c15b91d8998579ad7
#> 5 refs/tags/release-3.10^\{\} d16b7e7e849d5d63794b1c48847e8a57250ce2bd
#> 6    refs/tags/release-3.11 7d5723e022a7144c077501ecff5ac23984effe8d
#> 7 refs/tags/release-3.11^\{\} af29b8a99fe4b4e7deefb6bcbf9a6d98f43228bf
#> 
#> $caps
#>  [1] "multi_ack"                     "thin-pack"                    
#>  [3] "side-band"                     "side-band-64k"                
#>  [5] "ofs-delta"                     "shallow"                      
#>  [7] "deepen-since"                  "deepen-not"                   
#>  [9] "deepen-relative"               "no-progress"                  
#> [11] "include-tag"                   "multi_ack_detailed"           
#> [13] "no-done"                       "symref=HEAD:refs/heads/master"
#> [15] "agent=git/2.17.1"
}\if{html}{\out{</div>}}
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/valid-name.R
\name{is_valid_package_name}
\alias{is_valid_package_name}
\title{Check whether a package name is valid}
\usage{
is_valid_package_name(nm)
}
\arguments{
\item{nm}{Potential package name, string of length 1.}
}
\value{
Logical flag. If \code{FALSE}, then the \code{reason} attribute
contains a character string, the explanation why the package name
is invalid. See examples below.
}
\description{
Check whether a package name is valid
}
\examples{
is_valid_package_name("pak")
is_valid_package_name("pkg")
is_valid_package_name("pak\u00e1ge")
is_valid_package_name("good-package")
is_valid_package_name("x")
is_valid_package_name("1stpackage")
is_valid_package_name("dots.")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sysreqsdocs.R
\name{sysreqs_platforms}
\alias{sysreqs_platforms}
\title{List platforms with system requirements support}
\usage{
sysreqs_platforms()
}
\value{
Data frame with columns:
\itemize{
\item \code{name}: human readable OS name.
\item \code{os}: OS name, e.g. \code{linux}.
\item \code{distribution}: OS id, e.g. \code{ubuntu} or \code{redhat}.
\item \code{version}: distribution version. A star means that all versions are
supported, that are also supported by the vendor.
\item \code{update_command}: command to run to update the system package metadata.
\item \code{install_command}: command to run to install packages.
\item \code{query_command}: name of the tool to use to query system package
information.
}
}
\description{
List platforms with system requirements support
}
\examples{
\dontshow{if (Sys.getenv("IN_PKGDOWN") == "true") (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
sysreqs_platforms()
\dontshow{\}) # examplesIf}
}
\seealso{
Other system requirements functions: 
\code{\link{sysreqs_check_installed}()},
\code{\link{sysreqs_db_list}()},
\code{\link{sysreqs_db_match}()},
\code{\link{sysreqs_db_update}()},
\code{\link{sysreqs_install_plan}()},
\code{\link{sysreqs_is_supported}()},
\code{\link{sysreqs_list_system_packages}()}
}
\concept{system requirements functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gh-releases.R
\name{ghr}
\alias{ghr}
\title{GitHub Releases}
\description{
Functions to query and manipulate GitHub releases. These functions are
currently experimental.
}
\details{
\subsection{List releases}{
\subsection{Description}{

\code{ghr$list()} lists the last 100 releases for a GitHub repository.
\code{ghr$async_list()} is the async version of \code{ghr$list()}.
}

\subsection{Usage}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{ghr$list(repo)
ghr$async_list(repo)
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item \code{repo}: repository slug, e.g. \code{"cran/cli"}.
}
}

\subsection{Value}{

Data frame with columns:
\itemize{
\item \code{id}: release id,
\item \code{name}: release name, usually the version number, possibly with
a \code{v} prefix: \verb{3.6.1} or \code{v3.6.1}, but can be different.
\item \code{tag_name}: usually the same as \code{name}.
\item \code{created_at}: \code{POSIXct} vector.
}
}

}

\subsection{Get information about a release}{
\subsection{Description}{

\code{ghr$get()} downloads information about a release, including
release assets.

\code{ghr$async_get} is the async version of \code{ghr$get}.
}

\subsection{Usage}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{ghr$get(repo, tag)
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item \code{repo}: repository slug, e.g. \code{"cran/cli"}.
\item \code{tag}: tag to get.
}
}

\subsection{Value}{

Named list, see
\url{https://docs.github.com/en/rest/releases/releases#get-a-release}
for the entries.
}

}

\subsection{List assets of a release}{
\subsection{Description}{

\code{ghr$list_assets()} lists the last 100 assets of a release.

\code{ghr$async_list_assets()} is the async version of \code{ghr$list_assets()}
}

\subsection{Usage}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{ghr$list_assets(repo, tag)
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item \code{repo}: repository slug, e.g. \code{"cran/cli"}.
\item \code{tag}: tag to query.
}
}

\subsection{Value}{

Data frame with columns:
\itemize{
\item \code{id}: asset id,
\item \code{name}: file name of the asset,
\item \code{download_url}: download URL,
\item \code{size}: size in bytes,
\item \code{created_at}: \code{POSIXct} vector,
\item \code{updated_at}: \code{POSXct} vector,
\item \code{content_type}: content type of asset.
}
}

}

\subsection{Add a release asset}{
\subsection{Description}{

\code{ghr$add_asset()} adds an asset to a GitHub release.

\code{ghr$async_add_asset()} is the async version of \code{ghr$add_asset()}.
}

\subsection{Usage}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{ghr\%add_asset(repo, file, tag, name = basename(file))
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item \code{repo}: repository slug, e.g. \code{cran/cli}.
\item \code{file}: path to file to upload as an asset.
\item \code{tag}: tag name to add the asset to. It must exist on GitHub.
\item \code{name}: file name of the asset in the release.
}
}

\subsection{Details}{

If an asset with the same name already exists, then that will be deleted
first.
}

\subsection{Value}{

Response from GitHub as a named list. See
\url{https://docs.github.com/en/rest/releases/assets#upload-a-release-asset}
for the structure.
}

}

\subsection{Delete a release asset}{
\subsection{Description}{

\code{ghr$delete_asset()} deleted a release asset.

\code{ghr$async_delete_asset()} is an async version of \code{ghr$delete_asset()}.
}

\subsection{Usage}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{ghr$delete_asset(repo, tag, name)
ghr$async_delete_asset(repo, tag, name)
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item \code{repo}: repository slug, e.g. \code{cran/cli}.
\item \code{tag}: tag name to create a release for. It must exist on GitHub.
\item \code{name}: name of the asset.
}
}

\subsection{Value}{

A list with entries:
\itemize{
\item \code{release}: a list with the data about the release, before the
deletion. It has the same format as the return value of \code{ghr$get()}.
\item \code{deleted}: \code{TRUE} if the asset was deleted. \code{FALSE} if the asset
did not exist.
}
}

}

\subsection{Create a GitHub release}{
\subsection{Description}{

\code{ghr$create()} creates a GitHub release from a tag.

\code{ghr$async_create()} is an async version of \code{ghr$create()}.
}

\subsection{Usage}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{ghr$create(
  repo,
  tag,
  description = "",
  draft = FALSE,
  prerelease = FALSE,
  generage_release_notes = FALSE
)
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item \code{repo}: repository slug, e.g. \code{cran/cli}.
\item \code{tag}: tag name to create a release for. It must exist on GitHub.
\item \code{description}: release description.
\item \code{draft}: whether to create a draft release.
\item \code{prerelease}: whether to create a prerelease.
\item \code{generate_release_notes}: whether to auto-generate release notes.
}
}

\subsection{Value}{

Response from GitHub as a named list. See
\url{https://docs.github.com/en/rest/releases/releases#create-a-release}
for the structure.
}

}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install-plan.R
\name{install_plans}
\alias{install_plans}
\title{Installation plans}
\description{
An installation plan contains all data that is needed to install a
set of package files. It is usually created from an
\link[=pkg_installation_proposal]{installation proposal} with
\link[=pkg_solution]{solving} the dependencies and \link[=pkg_downloads]{downloading}
the package files.
}
\details{
It is also possible to create an installation plan a different way. An
installation plan object must be a data frame, with at least the
following columns:
\itemize{
\item \code{package}: The name of the package.
\item \code{type}: The type of the \link[=pkg_refs]{package reference}.
\item \code{binary}: Whether the package is a binary package.
\item \code{file}: Full path to the package file or directory.
\item \code{dependencies}: A list column that lists the names of the dependent
packages for each package.
\item \code{needscompilation}: Whether the package needs compilation. This should
be \code{FALSE} for binary packages.
}

For installation plans created via \link{pkg_installation_proposal}, the plan
contains all columns from \code{\link{pkg_download_result}}
objects, and some additional ones:
\itemize{
\item \code{library}: the library the package is supposed to be installed to.
\item \code{direct}: whether the package was directly requested or it is
installed as a dependency.
\item vignettes: whether the vignettes need to be (re)built.
\item \code{packaged}: whether \verb{R CMD build} was already called for the package.
}
}
\examples{
\dontrun{
pdi <- new_pkg_installation_proposal(
  "pak",
  config = list(library = tempfile())
)
pdi$resolve()
pdi$solve()
pdi$download()
pdi$get_install_plan()
}

}
\seealso{
\link{pkg_installation_proposal} to create install plans,
\code{\link[=install_package_plan]{install_package_plan()}} to install plans from any source.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse-remotes.R
\name{pkg_refs}
\alias{pkg_refs}
\title{Package references}
\description{
A package reference (ref) specifies a location from which an R package
can be obtained from. The full syntax of a reference is \code{type::ref}, but
\code{type} can be often omitted, the common ref types have shortcuts.
}
\section{Package references}{
Many pkgdepends and pak functions take package names as arguments.
E.g. \code{pak::pkg_install()} takes the names of the packages to install and
\code{pak::pkg_deps_tree()} takes the names of the packages to draw dependency
trees for.

Most of these functions can also take more generic \emph{package references}
instead of package names.
A package reference also tells pkgdepends where to find the
package source.

To specify a package source, use its name as a prefix, with a \code{::}
separator.
E.g. \code{cran::mypkg} means the \code{mypkg} package from CRAN.

A package name is a special package reference that implicitly specifies
the configured CRAN(-like) repositories as the package source
(we call this the \code{standard} package source).
So \code{mypkg} is equivalent to \code{standard::mypkg} and pkgdepends
looks for mypkg in any of the configured CRAN-like repositories.
If you did not explicitly specify any CRAN-like repositories (e.g. with
\code{options("repos")}), then pkgdepends uses the CRAN and
Bioconductor repositories by default.

This is the list of the currently supported package sources.
We will discuss each in detail below.
\itemize{
\item \code{cran}: a CRAN package.
\item \code{bioc}: a Bioconductor package.
\item \code{standard}: a package from a configured CRAN-like repository.
\item \code{github}: a package from GitHub.
\item \code{gitlab}: a package from GitLab.
\item \code{git}: a package in a Git repository.
\item \code{local}: a local package file or directory.
\item \code{url}: an URL that points to a package archive.
\item \code{installed}: an installed package.
\item \code{deps}: the dependencies of a local package file or directory.
\item \code{any}: a special reference type that accepts a package from any source.
See below.
\item \code{param}: a special reference to change how other references are downloaded
or installed. See "Parameters" below.
}
\subsection{Shorthands}{

To save typing, you do not always need to fully specify the package source
in a package reference.
You have seen before that a package name implicitly has a \code{standard}
package source.
Here are the complete rules for such shorthands, in the order they are
applied:

If the package reference is
\itemize{
\item a valid package name, or a package name with an \code{@}
version specification, the \code{standard} package source is used. E.g. \code{pkg} is
equivalent to \code{standard::pkg} and \verb{pkg@1.0} is equivalent to
\verb{standard::pkg@1.0}.
\item a valid \code{github} ref type without the \verb{github::} prefix,
then \code{github} is used. E.g. \code{user/repo} is equivalent to
\code{github::user/repo} and \code{user/repo@tag} is equivalent to
\code{github::user/repo@tag}, etc.
\item a GitHub URL (see below) without the \verb{github::} prefix,
then \code{github} is used.
\item a path that starts with \code{.} or \code{/} or \verb{\\} or \code{~}, then
\code{local} is used. (pkgdepends does not check if the path
exists.)
\item of the form \verb{<package-name>=?<parameters>},
then it will be the special \code{param} type. See "Parameters" below.
}

If the package reference does not have an explicit package source,
and the package source cannot be determined from these rules, then
pkgdepends throws an error.
}

\subsection{Package names}{

When pkgdepends is looking up the dependencies of a package,
it needs to be able to determine the name of the dependency from the
package reference.
This is sometimes not easy for dependencies in \code{Remotes} (or similar) fields.
\itemize{
\item For \verb{github::} and \verb{gitlab::} dependencies pkgdepends assumes
that the package name is the same as the name of the repository. If this
does not hold, then you need to specify the package name explicitly, using
a \verb{<package>=} prefix. E.g. \code{pins=rstudio/pins-r}. To specify both the
package source type and the package name at the same time, write it like this:
\code{pins=github::rstudio/pins-r}.
\item For \verb{git::} dependencies, pkgdepends assumes that the package
name is the same as the last component of the repository. If this does not
hold, then you need to specify the package name explicitly, using a
\verb{<package>=} prefix. E.g. \verb{pins=git::https://github.com/rstudio/pins-r}.
\item For \verb{local::} dependencies, you always need to specify the package name
explicitly. E.g. \verb{pins=local::~/works/pins}.
\item For \verb{url::} dependencies, you always need to specify the package name
explicitly. E.g. \verb{ggplot2=url::https://cloud.r-project.org/src/contrib/...}.
}
}

\subsection{Parameters}{

Package references may have optional parameters, added after a question
mark.
Different parameters are separated by an ampersand (\code{&}) character.
(This is very similar to how HTTP URLs take query parameters.)

Parameters may be flags that turn on some behavior, or they can have a
string value, assigned with an equal sign (\code{=}).
If no value is assigned, then a \code{true} value is assumed. For example,
these two package references are equivalent:

\if{html}{\out{<div class="sourceCode">}}\preformatted{cran::testthat?source&nocache
cran::testthat?source=true&nocache=true
}\if{html}{\out{</div>}}
\subsection{Parameters for downstream packages}{

pkgdepends allows specifying parameters for downstream packages,
using the \verb{<package>=?<params>} special package reference, where \code{package} is
the name of the package, and \verb{<params>} are the parameters, as above.
This is useful if you want to add a parameter to a downstream dependency.

For example, to install ggplot2, and always reinstall its cli package
dependency, you could use the \code{ggplot2} and \code{cli=?reinstall} package
references.
The latter tells pkgdepends to always reinstall cli, even if
it is already installed.
}

\subsection{Currently supported parameters}{
\itemize{
\item \code{ignore} is a flag parameter. If specified, the package is ignored.
This usually makes sense in the \code{packagename=?ignore} form, to ignore a
downstream soft dependency. If all versions of a hard dependency are
ignored that will lead to a solution error.
\item \code{ignore-before-r} is a version number parameter. The package will be
ignored on R versions that are older than the specified one. E.g.
\verb{Matrix=?ignore-before-r=4.1.2} will ignore the Matrix package on R versions
that are older than 4.1.2. This parameter really only makes sense in the
\code{packgename=?ignore-before-r} form.
\item \code{ignore-unavailable} is a flag. It can only be specified for soft
dependencies. If specified and the package is not available, it will be
ignored. This parameter really only makes sense in the
\code{packagename=?ignore-unavailable} form.
\item \code{source} is a flag parameter. If specified, then a source R package
is requested from a CRAN-like repository. For package installations,
\code{source} always triggers a re-install. In other words, \code{source} implies the
\code{reinstall} parameter. This parameter is supported for \verb{bioc::}, \verb{cran::}
and \verb{standard::} remote types, and it is ignored for others.
\item \code{reinstall} requests a re-install for package installations. It is
supported by the \verb{bioc::}, \verb{cran::}, \verb{git::}, \verb{github::}, \verb{gitlab::},
\verb{local::}, \verb{standard::}, and \verb{url::} remote types.
\item \code{nocache} will ignore the package cache. It will always download the
package file, and it will not add the downloaded (and built) package(s)
to the package cache. It is supported by the \verb{bioc::}, \verb{cran::}, \verb{git::},
\verb{github::}, \verb{gitlab::}, \verb{standard::} and \verb{url::} remote types.
}
}

}

\subsection{Package source details}{
\subsection{CRAN packages (\verb{cran::})}{

A package from CRAN. Full syntax:

\if{html}{\out{<div class="sourceCode">}}\preformatted{[cran::]<package>[@[>=]<version> | @current | @last]
}\if{html}{\out{</div>}}
\itemize{
\item \verb{<package>} is a valid package name.
\item \verb{<version>} is a version or a version requirement.
}

Examples:

\if{html}{\out{<div class="sourceCode">}}\preformatted{forecast
forecast@8.8
forecast@>=8.8
cran::forecast
forecast@last
forecast@current
}\if{html}{\out{</div>}}
}

\subsection{Bioconductor packages (\verb{bioc::})}{

A package from Bioconductor. The syntax is the same as for CRAN packages,
except for the prefix.

\if{html}{\out{<div class="sourceCode">}}\preformatted{[bioc::]<package>[@[>=]<version> | @current | @last]
}\if{html}{\out{</div>}}
}

\subsection{Standard packages (\verb{standard::})}{

These are packages either from CRAN or Bioconductor, the full syntax
is the same as for CRAN packages, except for the prefix:

\if{html}{\out{<div class="sourceCode">}}\preformatted{[standard::]<package>[@[>=]<version> | current | last]
}\if{html}{\out{</div>}}
}

\subsection{GitHub packages (\verb{github::})}{

Packages from a GitHub repository. Full syntax:

\if{html}{\out{<div class="sourceCode">}}\preformatted{[<package>=][github::]<username>/<repository>[/<subdir>][<detail>]
}\if{html}{\out{</div>}}
\itemize{
\item \verb{<package>} is the name of the package. If this is missing, then
the name of the repository is used.
\item \verb{<username>} is a GitHub username or organization name.
\item \verb{<repository>} is the name of the repository.
\item \verb{<subdir>} specifies an optional subdirectory, if the package is within a
subdirectory in the repository.
\item \verb{<detail>} specifies a certain version of the package, see below.
}

\verb{<detail>} may specify:
\itemize{
\item a Git branch, tag or (prefix of) a commit hash: \verb{@<commitish>};
\item a pull request: \verb{#<pull-request>}; or
\item the latest release: \verb{@*release}.
}

If \verb{<detail>} is missing, then the latest commit of the \emph{default}
branch is used.

Examples:

\if{html}{\out{<div class="sourceCode">}}\preformatted{r-lib/crayon
github::r-lib/crayon
r-lib/crayon@84be6207
r-lib/crayon@branch
r-lib/crayon#41
r-lib/crayon@release
}\if{html}{\out{</div>}}

For convenience, GitHub HTTP URLs can also be used to specify a
package from GitHub. Examples:

\if{html}{\out{<div class="sourceCode">}}\preformatted{https://github.com/r-lib/withr
# A branch:
https://github.com/r-lib/withr/tree/ghactions
# A tag:
https://github.com/r-lib/withr/tree/v2.1.1
# A commit:
https://github.com/r-lib/withr/commit/8fbcb548e316
# A pull request:
https://github.com/r-lib/withr/pull/76
# A release:
https://github.com/r-lib/withr/releases/tag/v2.1.0
}\if{html}{\out{</div>}}

A GitHub remote string can also be used instead of a URL, for example:
\code{git@github.com:r-lib/pak.git}
}

\subsection{GitLab packages (\verb{gitlab::})}{

Packages from a GitLab repository. Full syntax:

\if{html}{\out{<div class="sourceCode">}}\preformatted{[<package>=][gitlab::]<project-path>/<repository>[/-/<subdir>][<detail>]
}\if{html}{\out{</div>}}
\itemize{
\item \verb{<package>} is the name of the package. If this is missing, then
the name of the repository is used.
\item \verb{<project-path>} is typically the GitLab user or group name, but
it may contain subgroups.
\item \verb{<repository>} is the name of the repository, or the project in GitLab
terminology. GitLab
\href{https://docs.gitlab.com/ee/user/group/subgroups/}{subgroups} are
fully supported.
\item \verb{<subdir>} specifies an optional subdirectory, if the package is within a
subdirectory in the repository. Note that for GitLab, this must come
after a \verb{/-} prefix, to be able to distinguish it from subgroups.
\item \verb{<detail>} may specify a Git branch, tag or (prefix of) a commit hash.
}

If \verb{<detail>} is missing, then the latest commit of the \emph{default}
branch is used.

\verb{gitlab::} supports Git submodules, see the \code{git-submodules} configuration
entry.

Examples:

\if{html}{\out{<div class="sourceCode">}}\preformatted{gitlab::gaborcsardi/cli
gitlab::r-hub/filelock@main
gitlab::group/subgroup/subsubgroup/project/-/subdir@ref
}\if{html}{\out{</div>}}
}

\subsection{Packages in Git repositories (\verb{git::})}{

Full syntax:

\if{html}{\out{<div class="sourceCode">}}\preformatted{[<package>=]git::https?://<host>[<detail>]
}\if{html}{\out{</div>}}
\itemize{
\item \verb{<package>} is the name of the package. If this is missing, then
the last component of the \verb{<host>} is used.
\item \verb{<host>} is the host name and path of the Git repository. Some Git
repositories need the \code{.git} suffix here, others are more forgiving.
\item \verb{<detail>} specifies a certain version of the package:
a Git branch, tag or (prefix of) a commit hash: \verb{@<commitish>}.
}

If \verb{<detail>} is missing, then the latest commit of the \emph{default}
branch is used.

\verb{git::} supports Git submodules, see the \code{git-submodules} configuration
entry.

Examples:

\if{html}{\out{<div class="sourceCode">}}\preformatted{git::https://github.com/r-lib/crayon
git::https://github.com/r-lib/crayon.git
git::https://github.com/r-lib/crayon.git@84be6207
git::https://github.com/r-lib/crayon.git@branch
git::https://gitlab.com/gaborcsardi/cli.git
}\if{html}{\out{</div>}}

Note that pkgdepends has a built-in Git client, and does
\strong{not} require a system Git installation.

If the system has Git installed, then pkgdepends will use the
credentials stored in the configured Git credential store, automatically,
via the gitcreds package.
}

\subsection{Local packages (\verb{local::})}{

A path that refers to a package file built with \verb{R CMD build}, or a
directory that contains a package. Full syntax:

\if{html}{\out{<div class="sourceCode">}}\preformatted{local::<path>
}\if{html}{\out{</div>}}

For brevity, you can omit the \verb{local::} prefix, if you specify an
absolute path, a path from the user's home directory, starting with \code{~},
or a relative path starting with \verb{./} or \verb{.\\\\}.

A single dot (\code{"."}) is considered to be a local package in the current
working directory.

Examples:

\if{html}{\out{<div class="sourceCode">}}\preformatted{local::/foo/bar/package_1.0.0.tar.gz
local::/foo/bar/pkg
local::.
/absolute/path/package_1.0.0.tar.gz
~/path/from/home
./relative/path
.
}\if{html}{\out{</div>}}

If you specify a local package in a dependency (i.e. in \code{DESCRIPTION}), then
you also need to specify the name of the package, see "Package names"
above.
}

\subsection{URLs (\verb{url::})}{

You can use \verb{url::} to refer to URLs that hold R package archives
(i.e. properly built with \verb{R CMD build}), or compressed directories
of package trees (i.e. not built with \verb{R CMD build}). pkgdepends will
figure out if it needs to run \verb{R CMD build} on the package first.

This remote type supports \code{.tar.gz} and \code{.zip} files.

Note that URLs are not ideal remote types, because pkgdepends needs to
download the package file to resolve its dependencies. When this happens,
it puts the package file in the cache, so no further downloads are
needed when installing the package later.

Examples:

\if{html}{\out{<div class="sourceCode">}}\preformatted{url::https://cloud.r-project.org/src/contrib/Archive/cli/cli_1.0.0.tar.gz
url::https://github.com/tidyverse/stringr/archive/HEAD.zip
}\if{html}{\out{</div>}}

If you specify a package from a URL in a dependency (i.e. in \code{DESCRIPTION}),
then you also need to specify the name of the package, see "Package names"
above.
}

\subsection{Installed packages (\verb{installed::})}{

This is usually used internally, but can also be used directly.
Full syntax:

\if{html}{\out{<div class="sourceCode">}}\preformatted{installed::<path>/<package>
}\if{html}{\out{</div>}}
\itemize{
\item \verb{<path>} is the library the package is installed to.
\item \verb{<package>} is the package name.
}

Example:

\if{html}{\out{<div class="sourceCode">}}\preformatted{installed::~/R/3.6/crayon
}\if{html}{\out{</div>}}
}

\subsection{Package dependencies (\verb{deps::})}{

Usually used internally, it specifies the dependencies of a local
package. It can be used to download or install the dependencies of a
package, without downloading or installing the package itself.
Full syntax:

\if{html}{\out{<div class="sourceCode">}}\preformatted{deps::<path>
}\if{html}{\out{</div>}}

Examples:

\if{html}{\out{<div class="sourceCode">}}\preformatted{deps::/foo/bar/package_1.0.0.tar.gz
deps::/foo/bar/pkg
deps::.
}\if{html}{\out{</div>}}
}

\subsection{\verb{any::} packages}{

Sometimes you need to install additional packages, but you don't mind
where they are installed from. Here is an example. You want to install
cli from GitHub, from \code{r-lib/cli}. You also want to install glue, and
you don't mind which version of glue is installed, as long as it is
compatible with the requested cli version. If cli specifies the
development version of glue, then that is fine. If cli is fine with the
CRAN version of glue, that's OK, too. If a future version of cli does
not depend on glue, you still want glue installed, from CRAN. The \verb{any::}
reference type does exactly this.

In our example you might write

\if{html}{\out{<div class="sourceCode r">}}\preformatted{pak::pkg_install(c("glue", "r-lib/cli"))
}\if{html}{\out{</div>}}

first, but this will fail if \code{rlib/cli} requests (say) \code{tidyverse/glue},
because in \code{pkg_install()} \code{"glue"} is interpreted as \code{"standard::glue"},
creating a conflict with \code{tidyverse/glue}. On the other hand

\if{html}{\out{<div class="sourceCode r">}}\preformatted{pak::pkg_install(c("any::glue", "r-lib/cli"))
}\if{html}{\out{</div>}}

works, independently of which glue version is requested by cli.
}

\subsection{Parameter refs (\verb{param::})}{

See "Parameters" above.
}

}

\subsection{The \code{Remotes} field}{

In the \code{DESCRIPTION} file of an R package you can mark any regular
dependency defined in the \code{Depends}, \code{Imports}, \code{Suggests} or \code{Enhances}
fields as being installed from a non-standard package source by adding
a package reference to a \code{Remotes} entry.
pkgdepends will download and install the package from the
specified location, instead of a CRAN-like repository.

The remote dependencies specified in \code{Remotes} are a comma separated
list of package sources:

\if{html}{\out{<div class="sourceCode">}}\preformatted{Remotes: <pkg-source-1>, <pkg-source-2>, [ ... ]
}\if{html}{\out{</div>}}

Note that you will still need to add the package to one of the regular
dependency fields, i.e. \code{Imports}, \code{Suggests}, etc. Here is a concrete
example that specifies the \code{r-lib/glue} package:

\if{html}{\out{<div class="sourceCode">}}\preformatted{Imports: glue
Remotes: r-lib/glue,
  r-lib/httr@v0.4,
  klutometis/roxygen#142,
  r-lib/testthat@c67018fa4970
}\if{html}{\out{</div>}}

The CRAN and Bioconductor repositories do not support the \code{Remotes}
field, so you need to remove this field, before submitting your package
to either of them.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sysreqsdocs.R
\name{sysreqs_db_list}
\alias{sysreqs_db_list}
\title{List contents of the system requirements DB, for a platform}
\usage{
sysreqs_db_list(sysreqs_platform = NULL)
}
\arguments{
\item{sysreqs_platform}{System requirements platform. If \code{NULL}, then the
\code{sysreqs_platform} \eval{man_config_link("configuration option")}
is used, which defaults to the current platform. Set this option if
\eval{.packageName} does not detect your platform correctly.}
}
\value{
Data frame with columns:
\itemize{
\item \code{name}: cross platform system dependency name in the database.
\item \code{patterns}: one or more regular expressions to match to
\code{SystemRequirements} fields.
\item \code{packages}: one or more system package names to install.
\item \code{pre_install}: command(s) to run before installing the packages.
\item \code{post_install}:: command(s) to run after installing the packages.
}
}
\description{
It also tries to update the system dependency database, if it is
outdated. (I.e. older than allowed in the \code{metadata_update_after}
\eval{man_config_link("configuration option")}.
}
\examples{
\dontshow{if (Sys.getenv("IN_PKGDOWN") == "true") (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
sysreqs_db_list(sysreqs_platform = "ubuntu-22.04")
\dontshow{\}) # examplesIf}
}
\seealso{
Other system requirements functions: 
\code{\link{sysreqs_check_installed}()},
\code{\link{sysreqs_db_match}()},
\code{\link{sysreqs_db_update}()},
\code{\link{sysreqs_install_plan}()},
\code{\link{sysreqs_is_supported}()},
\code{\link{sysreqs_list_system_packages}()},
\code{\link{sysreqs_platforms}()}
}
\concept{system requirements functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgdepends.R
\docType{package}
\name{pkgdepends-package}
\alias{pkgdepends}
\alias{pkgdepends-package}
\title{pkgdepends: Package Dependency Resolution and Downloads}
\description{
pkgdepends is a toolkit for package dependencies, downloads and
installations, to be used in other packages. If you are looking for a
package manager, see \href{https://github.com/r-lib/pak}{pak}.
}
\section{Features}{
\itemize{
\item Look up package dependencies recursively.
\item Visualize package dependencies.
\item Download packages and their dependencies.
\item Install downloaded packages.
\item Includes a dependency solver to find a consistent set of dependencies.
\item Supports CRAN and Bioconductor packages automatically.
\item Supports packages on GitHub and GitLab.
\item Supports packages in git repositories.
\item Supports package bundles or files on the web.
\item Supports local package file and trees.
\item Supports the \code{Remotes} entry in the \code{DESCRIPTION} file.
\item Caches metadata and downloaded packages via
\href{https://github.com/r-lib/pkgcache}{pkgcache}
\item Performs all downloads and HTTP queries concurrently.
\item Builds and installs packages in parallel.
}
}

\section{Install}{
Install the package with:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{install.packages("pkgdepends")
}\if{html}{\out{</div>}}

If you need the development version, install it with

\if{html}{\out{<div class="sourceCode r">}}\preformatted{pak::pak("r-lib/pkgdepends")
}\if{html}{\out{</div>}}
}

\section{Usage}{
\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(pkgdepends)
}\if{html}{\out{</div>}}
\subsection{Package references}{

A package reference (ref) specifies a location from which an R package
can be obtained from. Examples:

\if{html}{\out{<div class="sourceCode">}}\preformatted{devtools
cran::devtools
bioc::Biobase
r-lib/pkgdepends
https://github.com/r-lib/pkgdepends
local::~/works/shiny
}\if{html}{\out{</div>}}

See \link[=pkg_refs]{“Package references”} for details.
}

\subsection{Package dependencies}{

Dependencies of the development version of the cli package:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{pd <- new_pkg_deps("r-lib/pkgcache")
pd$solve()
pd$draw()
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast"
style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000">
}}
\if{html}{\out{
<pre>
## <span style="color: #859900;">✔</span> Loading metadata database ... done                                    
## <span style="font-weight: bold;font-style: italic;color: #2AA198;">r-lib/pkgcache </span><span style="font-weight: bold;font-style: italic;color: #525252;">2.1.0.9000</span> <span style="color: #859900;">✨👷🏽🔧 </span>                                         
## ├─callr <span style="color: #525252;">3.7.3</span> <span style="color: #859900;">✨ ⬇ </span><span style="color: #525252;">(431.00 kB)</span>                                           
## │ ├─processx <span style="color: #525252;">3.8.1</span> <span style="color: #859900;">✨ ⬇ </span><span style="color: #525252;">(316.20 kB)</span>                                      
## │ │ ├─ps <span style="color: #525252;">1.7.5</span> <span style="color: #859900;">✨ ⬇ </span><span style="color: #525252;">(313.92 kB)</span>                                          
## │ │ └─R6 <span style="color: #525252;">2.5.1</span> <span style="color: #859900;">✨</span>                                                        
## │ └─R6                                                                  
## ├─cli <span style="color: #525252;">3.6.1</span> <span style="color: #859900;">✨ ⬇ </span><span style="color: #525252;">(1.38 MB)</span>                                               
## ├─curl <span style="color: #525252;">5.0.0</span> <span style="color: #859900;">✨ ⬇ </span><span style="color: #525252;">(777.64 kB)</span>                                            
## ├─filelock <span style="color: #525252;">1.0.2</span> <span style="color: #859900;">✨ ⬇ </span><span style="color: #525252;">(29.58 kB)</span>                                         
## ├─jsonlite <span style="color: #525252;">1.8.4</span> <span style="color: #859900;">✨ ⬇ </span><span style="color: #525252;">(1.13 MB)</span>                                          
## ├─prettyunits <span style="color: #525252;">1.1.1</span> <span style="color: #859900;">✨ ⬇ </span><span style="color: #525252;">(35.23 kB)</span>                                      
## ├─processx                                                              
## ├─R6                                                                    
## └─rappdirs <span style="color: #525252;">0.3.3</span> <span style="color: #859900;">✨ ⬇ </span><span style="color: #525252;">(47.50 kB)</span>                                         
##                                                                         
## Key:  <span style="color: #859900;">✨</span> new | <span style="color: #859900;"> ⬇</span> download | <span style="color: #859900;">👷🏽 </span>build | 🔧<span style="color: #859900;"> </span>compile                        
</pre>
}}
\if{html}{\out{
</div>
}}


See the \code{\link{pkg_deps}} class for details.
}

\subsection{Package downloads}{

Downloading all dependencies of a package:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{pdl <- new_pkg_download_proposal("r-lib/cli")
pdl$resolve()
pdl$download()
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast"
style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000">
}}
\if{html}{\out{
<pre>
## <span style="color: #2AA198;">ℹ</span> No downloads are needed, 1 pkg is cached                              
##                                                                         
</pre>
}}
\if{html}{\out{
</div>
}}


See the \code{\link{pkg_download_proposal}} class for
details.
}

\subsection{Package installation}{

Installing or updating a set of package:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{lib <- tempfile()
dir.create(lib)
pdi <- new_pkg_installation_proposal(
  "r-lib/cli",
  config = list(library = lib)
)
pdi$solve()
pdi$download()
pdi$install()
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast"
style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000">
}}
\if{html}{\out{
<pre>
## <span style="color: #2AA198;">ℹ</span> No downloads are needed, 1 pkg is cached                              
## <span style="color: #859900;">✔</span> Installed <span style="color: #268BD2;">cli</span> 3.6.1.9000 (github::r-lib/cli@c37f34b) <span style="color: #a3a3a3;">(36ms)</span>           
## <span style="color: #859900;">✔</span> Summary:  ✨ 1 new  in 36ms                                            
</pre>
}}
\if{html}{\out{
</div>
}}

}

\subsection{Dependency resolution}{

\code{\link{pkg_deps}},
\code{\link{pkg_download_proposal}} and
\code{\link{pkg_installation_proposal}} all resolve
their dependencies recursively, to obtain information about all packages
needed for the specified \link[=pkg_refs]{package references}. See
\link[=pkg_resolution]{“Dependency resolution”} for details.
}

\subsection{The dependency solver}{

The dependency solver takes the resolution information, and works out
the exact versions of each package that must be installed, such that
version and other requirements are satisfied. See \link[=pkg_solution]{“The dependency solver”} for details.
}

\subsection{Installation plans}{

\code{\link{pkg_installation_proposal}} can create
installation plans, and then also install them. It is also possible to
import installation plans that were created by other tools. See
\link[=install_plans]{“Installation plans”} for details.
}

\subsection{Configuration}{

The details of \code{\link{pkg_deps}},
\code{\link{pkg_download_proposal}} and
\code{\link{pkg_installation_proposal}} can be tuned
with a list of configuration options. See
\link[=pkgdepends-config]{“Configuration”} for details.
}
}

\section{Related}{
\itemize{
\item \href{https://github.com/r-lib/pak}{pak} – R package manager
\item \href{https://github.com/r-lib/pkgcache}{pkgcache} – Metadata and package
cache
\item \href{https://github.com/r-lib/devtools}{devtools} – Tools for R package
developers
}
\subsection{Code of Conduct}{

Please note that the pkgdepends project is released with a \href{https://r-lib.github.io/pkgdepends/dev/CODE_OF_CONDUCT.html}{Contributor Code of Conduct}.
By contributing to this project, you agree to abide by its terms.
}
}

\section{License}{
MIT (c) RStudio
}

\seealso{
Useful links:
\itemize{
  \item \url{https://r-lib.github.io/pkgdepends/}
  \item \url{https://github.com/r-lib/pkgdepends}
  \item Report bugs at \url{https://github.com/r-lib/pkgdepends/issues}
}

}
\author{
\strong{Maintainer}: Gábor Csárdi \email{csardi.gabor@gmail.com}

Other contributors:
\itemize{
  \item Posit Software, PBC (\href{https://ror.org/03wc8by49}{ROR}) [copyright holder, funder]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/name-check.R
\name{pkg_name_check}
\alias{pkg_name_check}
\title{Check if an R package name is available.}
\usage{
pkg_name_check(name, dictionaries = NULL)
}
\arguments{
\item{name}{Package name candidate.}

\item{dictionaries}{Character vector, the dictionaries to query.
Available dictionaries:
* \code{wikipedia}
* \code{wiktionary},
* \code{sentiment} (\url{https://github.com/fnielsen/afinn}),
* \code{urban} (Urban Dictionary).
If \code{NULL} (by default), the Urban Dictionary is omitted, as it
is often offensive.}
}
\value{
\code{pkg_name_check} object with a custom print method.
}
\description{
Additionally, look up the candidate name in a number of
dictionaries, to make sure that it does not have a negative
meaning.
}
\details{
\subsection{Valid package name check}{

Check the validity of \code{name} as a package name. See 'Writing R
Extensions' for the allowed package names. Also checked against a list
of names that are known to cause problems.
}

\subsection{CRAN checks}{

Check \code{name} against the names of all past and current packages on
CRAN, including base and recommended packages.
}

\subsection{Bioconductor checks}{

Check \code{name} against all past and current Bioconductor packages.
}

\subsection{Profanity check}{

Check \code{name} with \url{https://www.purgomalum.com/service/containsprofanity}
to make sure it is not a profanity.
}

\subsection{Dictionaries}{

See the \code{dictionaries} argument.
}
}
\examples{
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
pkg_name_check("cli")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkg-dependencies.R
\name{new_pkg_deps}
\alias{new_pkg_deps}
\alias{pkg_deps}
\title{R6 class for package dependency lookup}
\usage{
new_pkg_deps(refs, ...)
}
\arguments{
\item{refs}{Package names or references. See
\link[=pkg_refs]{'Package references'} for the syntax.}

\item{...}{Additional arguments, passed to
\href{#method-new}{\code{pkg_deps$new()}}.}
}
\value{
\code{new_pkg_deps()} returns a new \code{pkg_deps} object.
}
\description{
Look up dependencies of R packages from various sources.
}
\details{
\code{new_pkg_deps()} creates a new object from the \code{pkg_deps} class.
The advantage of \code{new_pkg_deps()} compared to using the \link{pkg_deps}
constructor directly is that it avoids making pkgdepends a build time
dependency.

The usual steps to query package dependencies are:
\enumerate{
\item Create a \code{pkg_deps} object with \code{new_pkg_deps()}.
\item Resolve all possible dependencies with
\href{#method-resolve}{\code{pkg_deps$resolve()}}.
\item Solve the dependencies, to obtain a subset of all possible
dependencies that can be installed together, with
\href{#method-solve}{\code{pkg_deps$solve()}}.
\item Call \href{#method-get-solution}{\code{pkg_deps$get_solution()}} to list the
result of the dependency solver.
}
}
\examples{
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method initialize()
pd <- pkg_deps$new("r-lib/pkgdepends")
pd
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method get_refs()
pd <- new_pkg_deps(c("pak", "jsonlite"))
pd$get_refs()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method get_config()
pd <- new_pkg_deps("pak")
pd$get_config()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method resolve()
pd <- new_pkg_deps("pak")
pd$resolve()
pd$get_resolution()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method get_resolution()
pd <- new_pkg_deps("r-lib/pkgdepends")
pd$resolve()
pd$get_resolution()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method get_solve_policy()
pdi <- new_pkg_deps("r-lib/pkgdepends")
pdi$get_solve_policy()
pdi$set_solve_policy("upgrade")
pdi$get_solve_policy()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method set_solve_policy()
pdi <- new_pkg_deps("r-lib/pkgdepends")
pdi$get_solve_policy()
pdi$set_solve_policy("upgrade")
pdi$get_solve_policy()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method solve()
pd <- new_pkg_deps("r-lib/pkgdepends")
pd$resolve()
pd$solve()
pd$get_solution()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method get_solution()
pd <- new_pkg_deps("pkgload")
pd$resolve()
pd$solve()
pd$get_solution()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method stop_for_solution_error()
# This is an error, because the packages conflict:
pd <- new_pkg_deps(
  c("r-lib/pak", "cran::pak"),
  config = list(library = tempfile())
)
pd$resolve()
pd$solve()
pd
# This fails:
# pd$stop_for_solution_error()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method draw()
pd <- new_pkg_deps("pkgload")
pd$solve()
pd$draw()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method print()
pd <- new_pkg_deps("r-lib/pkgdepends")
pd

pd$resolve()
pd

pd$solve()
pd
\dontshow{\}) # examplesIf}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-pkg_deps-new}{\code{pkg_deps$new()}}
\item \href{#method-pkg_deps-get_refs}{\code{pkg_deps$get_refs()}}
\item \href{#method-pkg_deps-get_config}{\code{pkg_deps$get_config()}}
\item \href{#method-pkg_deps-resolve}{\code{pkg_deps$resolve()}}
\item \href{#method-pkg_deps-async_resolve}{\code{pkg_deps$async_resolve()}}
\item \href{#method-pkg_deps-get_resolution}{\code{pkg_deps$get_resolution()}}
\item \href{#method-pkg_deps-get_solve_policy}{\code{pkg_deps$get_solve_policy()}}
\item \href{#method-pkg_deps-set_solve_policy}{\code{pkg_deps$set_solve_policy()}}
\item \href{#method-pkg_deps-solve}{\code{pkg_deps$solve()}}
\item \href{#method-pkg_deps-get_solution}{\code{pkg_deps$get_solution()}}
\item \href{#method-pkg_deps-stop_for_solution_error}{\code{pkg_deps$stop_for_solution_error()}}
\item \href{#method-pkg_deps-draw}{\code{pkg_deps$draw()}}
\item \href{#method-pkg_deps-format}{\code{pkg_deps$format()}}
\item \href{#method-pkg_deps-print}{\code{pkg_deps$print()}}
\item \href{#method-pkg_deps-clone}{\code{pkg_deps$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_deps-new"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_deps-new}{}}}
\subsection{Method \code{new()}}{
Create a new \code{pkg_deps} object. Consider using \code{new_pkg_deps()}
instead of calling the constructor directly.

The returned object can be used to look up (recursive) dependencies
of R packages from various sources. To perform the actual lookup,
you'll need to call the \href{#method-resolve}{\code{resolve()}} method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_deps$new(
  refs,
  config = list(),
  policy = c("lazy", "upgrade"),
  remote_types = NULL
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{refs}}{Package names or references. See
\link[=pkg_refs]{'Package references'} for the syntax.}

\item{\code{config}}{Configuration options, a named list. See
\link[=pkgdepends-config]{'Configuration'}.}

\item{\code{policy}}{Solution policy. See \link[=pkg_solution]{'The dependency solver'}.}

\item{\code{remote_types}}{Custom remote ref types, this is for advanced
use, and experimental currently.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{pkg_deps} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_deps-get_refs"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_deps-get_refs}{}}}
\subsection{Method \code{get_refs()}}{
The package refs that were used to create the \code{pkg_deps} object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_deps$get_refs()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A character vector of package refs that were used to create the
\code{pkg_deps} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_deps-get_config"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_deps-get_config}{}}}
\subsection{Method \code{get_config()}}{
Configuration options for the \code{pkg_deps} object. See
\link[=pkgdepends-config]{'Configuration'} for details.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_deps$get_config()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
See \link[=pkgdepends-config]{'Configuration'} for the configuration entries.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_deps-resolve"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_deps-resolve}{}}}
\subsection{Method \code{resolve()}}{
Resolve the dependencies of the specified package references. This
usually means downloading metadata from CRAN and Bioconductor, unless
already cached, and also from GitHub if GitHub refs were included,
either directly or indirectly. See
\link[=pkg_resolution]{'Dependency resolution'} for details.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_deps$resolve()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
The \code{pkg_deps} object itself, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_deps-async_resolve"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_deps-async_resolve}{}}}
\subsection{Method \code{async_resolve()}}{
The same as \href{#method-resolve}{\code{resolve()}}, but asynchronous.
This method is for advanced use.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_deps$async_resolve()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A deferred value.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_deps-get_resolution"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_deps-get_resolution}{}}}
\subsection{Method \code{get_resolution()}}{
Query the result of the dependency resolution. This method can be
called after \href{#method-resolve}{\code{resolve()}} has completed.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_deps$get_resolution()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A \link{pkg_resolution_result} object, which is also a data frame. See
\link[=pkg_resolution]{'Dependency resolution'} for its columns.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_deps-get_solve_policy"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_deps-get_solve_policy}{}}}
\subsection{Method \code{get_solve_policy()}}{
Returns the current policy of the dependency solver.
See \link[=pkg_solution]{'The dependency solver'} for details.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_deps$get_solve_policy()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A character vector of length one.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_deps-set_solve_policy"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_deps-set_solve_policy}{}}}
\subsection{Method \code{set_solve_policy()}}{
Set the current policy of the dependency solver.
If the object already contains a solution and the new policy is
different than the old policy, then the solution is deleted.
See \link[=pkg_solution]{'The dependency solver'} for details.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_deps$set_solve_policy(policy = c("lazy", "upgrade"))}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{policy}}{Policy to set.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_deps-solve"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_deps-solve}{}}}
\subsection{Method \code{solve()}}{
Solve the package dependencies. Out of the resolved dependencies, it
works out a set of packages, that can be installed together to
create a functional installation. The set includes all directly
specified packages, and all required (or suggested, depending on
the configuration) packages as well. It includes every package at
most once. See \link[=pkg_solution]{'The dependency solver'} for details.

\code{solve()} calls \href{#method-resolve}{\code{resolve()}} automatically, if it
hasn't been called yet.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_deps$solve()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
The \code{pkg_deps} object itself, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_deps-get_solution"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_deps-get_solution}{}}}
\subsection{Method \code{get_solution()}}{
Returns the solution of the package dependencies.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_deps$get_solution()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A \link{pkg_solution_result} object, which is a list. See
\link{pkg_solution_result} for details.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_deps-stop_for_solution_error"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_deps-stop_for_solution_error}{}}}
\subsection{Method \code{stop_for_solution_error()}}{
Error if the dependency solver failed to find a consistent set of
packages that can be installed together.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_deps$stop_for_solution_error()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_deps-draw"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_deps-draw}{}}}
\subsection{Method \code{draw()}}{
Draw a tree of package dependencies. It returns a \code{tree} object, see
\code{\link[cli:tree]{cli::tree()}}. Printing this object prints the dependency tree to the
screen.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_deps$draw()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A \code{tree} object from the cli package, see \code{\link[cli:tree]{cli::tree()}}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_deps-format"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_deps-format}{}}}
\subsection{Method \code{format()}}{
Format a \code{pkg_deps} object, typically for printing.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_deps$format(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Not used currently.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A character vector, each element should be a line in the printout.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_deps-print"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_deps-print}{}}}
\subsection{Method \code{print()}}{
Prints a \code{pkg_deps} object to the screen. The printout includes:
\itemize{
\item The package refs.
\item Whether the object has the resolved dependencies.
\item Whether the resolution had errors.
\item Whether the object has the solved dependencies.
\item Whether the solution had errors.
\item Advice on which methods to call next.
}

See the example below.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_deps$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{not used currently.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The \code{pkg_deps} object itself, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_deps-clone"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_deps-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_deps$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand % Please edit documentation in R/pkg-lockfile.R \name{new_pkg_installation_plan} \alias{new_pkg_installation_plan} \alias{pkg_installation_plan} \title{R6 class for installation from a lock file} \usage{ new_pkg_installation_plan(lockfile = "pkg.lock", config = list(), ...) } \arguments{ \item{lockfile}{Path to the lock file to use.} \item{config}{Configuration options, a named list. See \link[=pkgdepends-config]{'Configuration'}. If it does not include \code{library}, then \code{.libPaths()[1]} is added as \code{library}.} \item{...}{Additional arguments, passed to \href{#method-new}{\code{pkg_installation_plan$new()}}.} } \value{ \code{new_pkg_installation_plan()} returns a \code{pkg_installation_plan} object. } \description{ An installation plan is similar to an installation proposal (i.e. \link{pkg_installation_proposal}), but it already contains the solved dependencies, complete with download URLs. } \details{ Typically you create a \code{pkg_installation_plan} object with \code{new_pkg_installation_plan()} and then call its \verb{$download()} method to download the packages and then its \verb{$install()} method to install them. } \section{Super class}{ \code{\link[pkgdepends:pkg_installation_proposal]{pkgdepends::pkg_installation_proposal}} -> \code{pkg_installation_plan} } \section{Methods}{ \subsection{Public methods}{ \itemize{ \item \href{#method-pkg_installation_plan-new}{\code{pkg_installation_plan$new()}} \item \href{#method-pkg_installation_plan-resolve}{\code{pkg_installation_plan$resolve()}} \item \href{#method-pkg_installation_plan-async_resolve}{\code{pkg_installation_plan$async_resolve()}} \item \href{#method-pkg_installation_plan-get_solve_policy}{\code{pkg_installation_plan$get_solve_policy()}} \item \href{#method-pkg_installation_plan-set_solve_policy}{\code{pkg_installation_plan$set_solve_policy()}} \item \href{#method-pkg_installation_plan-solve}{\code{pkg_installation_plan$solve()}} \item \href{#method-pkg_installation_plan-update}{\code{pkg_installation_plan$update()}} \item \href{#method-pkg_installation_plan-update_sysreqs}{\code{pkg_installation_plan$update_sysreqs()}} \item \href{#method-pkg_installation_plan-format}{\code{pkg_installation_plan$format()}} \item \href{#method-pkg_installation_plan-clone}{\code{pkg_installation_plan$clone()}} } } \if{html}{\out{
Inherited methods

- `pkgdepends::pkg_installation_proposal$async_download()`
- `pkgdepends::pkg_installation_proposal$create_lockfile()`
- `pkgdepends::pkg_installation_proposal$download()`
- `pkgdepends::pkg_installation_proposal$draw()`
- `pkgdepends::pkg_installation_proposal$get_config()`
- `pkgdepends::pkg_installation_proposal$get_downloads()`
- `pkgdepends::pkg_installation_proposal$get_install_plan()`
- `pkgdepends::pkg_installation_proposal$get_refs()`
- `pkgdepends::pkg_installation_proposal$get_resolution()`
- `pkgdepends::pkg_installation_proposal$get_solution()`
- `pkgdepends::pkg_installation_proposal$get_sysreqs()`
- `pkgdepends::pkg_installation_proposal$install()`
- `pkgdepends::pkg_installation_proposal$install_sysreqs()`
- `pkgdepends::pkg_installation_proposal$print()`
- `pkgdepends::pkg_installation_proposal$show_solution()`
- `pkgdepends::pkg_installation_proposal$show_sysreqs()`
- `pkgdepends::pkg_installation_proposal$stop_for_download_error()`
- `pkgdepends::pkg_installation_proposal$stop_for_solution_error()`

}} \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-pkg_installation_plan-new}{}}} \subsection{Method \code{new()}}{ Create a new \code{pkg_installation_plan} object. Consider using \code{new_pkg_installation_plan()} instead of calling the constructor directly. The returned object can be used to download and install packages, according to the plan. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{pkg_installation_plan$new( lockfile = "pkg.lock", config = list(), remote_types = NULL )}\if{html}{\out{
}} } \subsection{Arguments}{ \if{html}{\out{
}} \describe{ \item{\code{lockfile}}{Path to the lock file to use.} \item{\code{config}}{Configuration options. See \link[=pkgdepends-config]{'Configuration'}. It needs to include the package library to install to, in \code{library}.} \item{\code{remote_types}}{Custom remote ref types, this is for advanced use, and experimental currently.} } \if{html}{\out{
}} } } \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-pkg_installation_plan-resolve}{}}} \subsection{Method \code{resolve()}}{ This function is implemented for installation plans, and will error. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{pkg_installation_plan$resolve()}\if{html}{\out{
}} } } \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-pkg_installation_plan-async_resolve}{}}} \subsection{Method \code{async_resolve()}}{ This function is implemented for installation plans, and will error. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{pkg_installation_plan$async_resolve()}\if{html}{\out{
}} } } \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-pkg_installation_plan-get_solve_policy}{}}} \subsection{Method \code{get_solve_policy()}}{ Installation plans are already solved, and this method will return \code{NA_character_}, always. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{pkg_installation_plan$get_solve_policy()}\if{html}{\out{
}} } } \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-pkg_installation_plan-set_solve_policy}{}}} \subsection{Method \code{set_solve_policy()}}{ This function is implemented for installation plans, and will error. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{pkg_installation_plan$set_solve_policy()}\if{html}{\out{
}} } } \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-pkg_installation_plan-solve}{}}} \subsection{Method \code{solve()}}{ This function is implemented for installation plans, and will error. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{pkg_installation_plan$solve()}\if{html}{\out{
}} } } \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-pkg_installation_plan-update}{}}} \subsection{Method \code{update()}}{ Update the plan to the current state of the library. If the library has not changed since the plan was created, then it does nothing. If new packages have been installed, then it might not be necessary to download and install all packages in the plan. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{pkg_installation_plan$update()}\if{html}{\out{
}} } \subsection{Details}{ This operation is different than creating a new proposal with the updated library, because it uses the the packages and package versions of the original plan. E.g. if the library has a newer version of a package, then \verb{$update()} will downgrade it to the version in the plan. } } \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-pkg_installation_plan-update_sysreqs}{}}} \subsection{Method \code{update_sysreqs()}}{ Update information about installed and missing system requirements. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{pkg_installation_plan$update_sysreqs()}\if{html}{\out{
}} } } \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-pkg_installation_plan-format}{}}} \subsection{Method \code{format()}}{ Format a \code{pkg_installation_plan} object, typically for printing. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{pkg_installation_plan$format(...)}\if{html}{\out{
}} } \subsection{Arguments}{ \if{html}{\out{
}} \describe{ \item{\code{...}}{not used currently.} } \if{html}{\out{
}} } \subsection{Returns}{ A character vector, each element should be a line in the printout. } } \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-pkg_installation_plan-clone}{}}} \subsection{Method \code{clone()}}{ The objects of this class are cloneable with this method. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{pkg_installation_plan$clone(deep = FALSE)}\if{html}{\out{
}} } \subsection{Arguments}{ \if{html}{\out{
}} \describe{ \item{\code{deep}}{Whether to make a deep clone.} } \if{html}{\out{
}} } } }

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/git-protocol.R
\name{git_download_file}
\alias{git_download_file}
\title{Download a blob for a remote git repository}
\usage{
git_download_file(url, sha, output = sha)
}
\arguments{
\item{url}{Repository URL, e.g. \verb{https://github.com/r-lib/pak.git}.
It might include authentication information, e.g. a GitHub token.}

\item{sha}{SHA hash of the blob.}

\item{output}{Path where the blob will be written. It's directory is
created if it does not exist.}
}
\value{
A list that corresponds to a git packfile entry, invisibly.
It has entries: \code{type} (always \code{"blob"}), \code{object} (raw object, the
blob itself), \code{size}, \code{packed_size}, \code{hash}.
}
\description{
Download a blob for a remote git repository
}
\section{Examples}{

Download a \code{DESCRIPTION} file from GitHub:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{pak_repo <- "https://github.com/r-lib/pak.git"
fls <- git_list_files(pak_repo, "HEAD")
git_download_file(
  pak_repo,
  fls$files$hash[fls$files$path == "DESCRIPTION"],
  output = tmp <- tempfile()
)
readLines(tmp)[1:5]
#> [1] "Package: pak"                                                                  
#> [2] "Version: 0.3.1.9999"                                                           
#> [3] "Title: Another Approach to Package Installation"                               
#> [4] "Description: The goal of 'pak' is to make package installation faster and"     
#> [5] "    more reliable. In particular, it performs all HTTP operations in parallel,"
}\if{html}{\out{</div>}}
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gh-repo.R
\name{ghrepo}
\alias{ghrepo}
\title{Update a CRAN-like repository of binary packages at GitHub}
\description{
These functions are currently experimental.
}
\details{
\subsection{Update a CRAN-like repository of binary packages at GitHub}{

\code{ghrepo$update()} updates a binary package mirror.
\subsection{Usage}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{ghrepo$update(
  repo,
  subdir,
  release_org = "cran",
  source_repo = "https://cran.r-project.org",
  packages = NULL
)
}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\itemize{
\item \code{repo}: GitHub slug, e.g. \code{r-hub/repos}.
\item \code{subdir}: subdirectory in the GitHub repository, where the R package
metadata should be updated. It must exist in the repository.
If it does not have \verb{PACKAGES*} files, then they will be created.
\item \code{release_org}: GitHub organization or user name where the packages
will be published as releases.
\item \code{source_repo}: A CRAN-like repository, where source packages are
taken from.
\item \code{packages}: A character vector of package names to add to the binary
repository, in addition to updating the ones that are already there.
}
}

}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/builder.R
\name{pkg_build}
\alias{pkg_build}
\title{Create a binary package from an installed package}
\usage{
pkg_build(
  pkg,
  library = .libPaths()[1],
  flavor = Sys.getenv("PKG_BUILD_FLAVOR"),
  build_number = 1L
)
}
\arguments{
\item{pkg}{Package name.}

\item{library}{Library path.}

\item{flavor}{Platform flavor. Defaults to the \code{PKG_BUILD_FLAVOR}
environment variable. If not \code{NULL} or an empty string, then it is
appended to the platform string with a dash.}

\item{build_number}{An integer number that is added to the file name,
after the version number, to be able to have multiple builds for the
same package version.}
}
\value{
Path to the built package.
}
\description{
The built package will be in the current working directory.
}
\details{
This function is currently experimental.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/solve.R
\name{pkg_solution}
\alias{pkg_solution}
\alias{pkg_solution_result}
\title{The dependency solver}
\description{
The dependency solver takes the resolution information, and works out
the exact versions of each package that must be installed, such that
version and other requirements are satisfied.
}
\details{
\subsection{Solution policies}{

The dependency solver currently supports two policies: \code{lazy} and
\code{upgrade}. The \code{lazy} policy prefers to minimize installation time,
and it does not perform package upgrades, unless version requirements
require them. The \code{upgrade} policy prefers to update all package to
their latest possible versions, but it still considers that version
requirements.
}

\subsection{The integer problem}{

Solving the package dependencies requires solving an integer linear
problem (ILP). This subsection briefly describes how the problem is
represented as an integer problem, and what the solution policies
exactly mean.

Every row of the package resolution is a candidate for the dependency
solver. In the integer problem, every candidate corresponds to a binary
variable. This is 1 if that candidate is selected as part of the
solution, and 0 otherwise.

The objective of the ILP minimization is defined differently for
different solution policies. The ILP conditions are the same.
\enumerate{
\item For the \code{lazy} policy, \verb{installed::} packaged get 0 points, binary
packages 1 point, sources packages 5 points.
\item For the 'upgrade' policy, we rank all candidates for a given package
according to their version numbers, and assign more points to older
versions. Points are assigned by 100 and candidates with equal
versions get equal points. We still prefer installed packages to
binaries to source packages, so also add 0 point for already
installed candidates, 1 extra points for binaries and 5 points for
source packages.
\item For directly specified refs, we aim to install each package exactly
once. So for these we require that the variables corresponding to
the same package sum up to 1.
\item For non-direct refs (i.e. dependencies), we require that the
variables corresponding to the same package sum up to at most one.
Since every candidate has at least 1 point in the objective function
of the minimization problem, non-needed dependencies will be
omitted.
\item For direct refs, we require that their candidates satisfy their
references. What this means exactly depends on the ref types. E.g.
for CRAN packages, it means that a CRAN candidate must be selected.
For a standard ref, a GitHub candidate is OK as well.
\item We rule out candidates for which the dependency resolution failed.
\item We go over all the dependency requirements and rule out packages
that do not meet them. For every package \code{A}, that requires
package \code{B}, we select the \verb{B(i, i=1..k)} candidates of \code{B} that
satisfy \code{A}'s requirements and add a \code{A - B(1) - ... - B(k) <= 0}
rule. To satisfy this rule, either we cannot install \code{A}, or if \code{A}
is installed, then one of the good \code{B} candidates must be installed
as well.
\item We rule out non-installed CRAN and Bioconductor candidates for
packages that have an already installed candidate with the same exact
version.
\item We also rule out source CRAN and Bioconductor candidates for
packages that have a binary candidate with the same exact version.
}
}

\subsection{Explaining why the solver failed}{

To be able to explain why a solution attempt failed, we also add a dummy
variable for each directly required package. This dummy variable has a
very large objective value, and it is only selected if there is no
way to install the directly required package.

After a failed solution, we look the dummy variables that were selected,
to see which directly required package failed to solve. Then we check
which rule(s) ruled out the installation of these packages, and their
dependencies, recursively.
}

\subsection{The result}{

The result of the solution is a \code{pkg_solution_result} object. It is a
named list with entries:
\itemize{
\item \code{status}: Status of the solution attempt, \code{"OK"} or \code{"FAILED"}.
\item \code{data}: The selected candidates. This is very similar to a
\link{pkg_resolution_result} object, but it has two extra columns:
\itemize{
\item \code{lib_status}: status of the package in the library, after the
installation. Possible values: \code{new} (will be newly installed),
\code{current} (up to date, not installed), \code{update} (will be updated),
\code{no-update} (could update, but will not).
\item \code{old_version}: The old (current) version of the package in the
library, or \code{NA} if the package is currently not installed.
}
\item \code{problem}: The ILP problem. The exact representation is an
implementation detail, but it does have an informative print method.
\item \code{solution}: The return value of the internal solver.
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sysreqsdocs.R
\name{sysreqs_db_match}
\alias{sysreqs_db_match}
\title{Match system requirement descriptions to the database}
\usage{
sysreqs_db_match(specs, sysreqs_platform = NULL)
}
\arguments{
\item{specs}{Character vector of system requirements descriptions.}

\item{sysreqs_platform}{System requirements platform. If \code{NULL}, then the
\code{sysreqs_platform} \eval{man_config_link("configuration option")}
is used, which defaults to the current platform. Set this option if
\eval{.packageName} does not detect your platform correctly.}
}
\value{
Data frame with columns:
\itemize{
\item \code{spec}: the input \code{specs}.
\item \code{sysreq}:  name of the system library or tool.
\item \code{packages}: system packages, list column of character vectors.
Rarely it can be an empty string, e.g. if a \code{pre_install} script
performs the installation.
\item \code{pre_install}: list column of character vectors. Shell script(s) to
run before the installation.
\item \code{post_install}: list column of character vectors. Shell script(s) to
run after the installation.
}
}
\description{
In the usual workflow pkgdepends matches the
\code{SystemRequirements} fields of the \code{DESCRIPTION} files to the database.
}
\details{
The \code{sysreqs_db_match()} function lets you match any string, and it is
mainly useful for debugging.
}
\examples{
\dontshow{if (Sys.getenv("IN_PKGDOWN") == "true") (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
sysreqs_db_match(
  c("Needs libcurl", "Java, libssl"),
  sysreqs_platform = "ubuntu-22.04"
)
\dontshow{\}) # examplesIf}
}
\seealso{
Other system requirements functions: 
\code{\link{sysreqs_check_installed}()},
\code{\link{sysreqs_db_list}()},
\code{\link{sysreqs_db_update}()},
\code{\link{sysreqs_install_plan}()},
\code{\link{sysreqs_is_supported}()},
\code{\link{sysreqs_list_system_packages}()},
\code{\link{sysreqs_platforms}()}
}
\concept{system requirements functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse-remotes.R
\name{parse_pkg_refs}
\alias{parse_pkg_refs}
\alias{parse_pkg_ref}
\title{Parse package location references}
\usage{
parse_pkg_refs(refs, remote_types = NULL, ...)

parse_pkg_ref(ref, remote_types = NULL, ...)
}
\arguments{
\item{refs}{Character vector of references.}

\item{remote_types}{Custom remote types can be added here, this is
for advanced use, and experimental currently.}

\item{...}{Additional arguments are passed to the individual parser
functions.}

\item{ref}{A package reference, like \code{refs}, but a length one vector,
for convenience.}
}
\value{
\code{parse_pkg_refs()} returns a list of parsed references.
\code{parse_pkg_ref()} returns one parsed reference. A parsed reference is
a list, with at least elements:
\itemize{
\item \code{ref}: The original reference string.
\item \code{type}: The reference type.
\item \code{package}: The package name.
It typically contains additional data, specific to the various
reference types. See \link{pkg_refs} for details.
The parsed reference always has class \verb{remote_ref_<type>} and
\code{remote_ref}.
}
}
\description{
See \link{pkg_refs} for more about supported package references.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install-plan.R
\name{install_package_plan}
\alias{install_package_plan}
\title{Perform a package installation plan}
\usage{
install_package_plan(
  plan,
  lib = .libPaths()[[1]],
  num_workers = 1,
  cache = NULL
)
}
\arguments{
\item{plan}{Package plan object, a data frame, see
\link[=install_plans]{'Installation plans'} for the format.}

\item{lib}{Library directory to install to.}

\item{num_workers}{Number of worker processes to use.}

\item{cache}{Package cache to use, or \code{NULL}.}
}
\value{
Information about the installation process.
}
\description{
See \link[=install_plans]{'Installation plans'} for the details and the format.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkg-downloads.R
\name{new_pkg_download_proposal}
\alias{new_pkg_download_proposal}
\alias{pkg_download_proposal}
\title{R6 class for package downloads}
\usage{
new_pkg_download_proposal(refs, ...)
}
\arguments{
\item{refs}{Package names or references. See
\link[=pkg_refs]{'Package references'} for the syntax.}

\item{...}{Additional arguments, passed to
\href{#method-new}{\code{pkg_download_proposal$new()}}.}
}
\value{
\code{new_pkg_download_proposal()} returns a new
\code{pkg_download_proposal} object.
}
\description{
Download packages with their dependencies, from various sources.
}
\details{
\code{new_pkg_download_proposal()} creates a new object from the
\code{pkg_download_proposal} class, that can be used to look up and download
R packages and their dependencies. The advantage of
\code{new_pkg_download_proposal()} compared to using the
\link{pkg_download_proposal} constructor directly is that it avoids making
pkgdepends a build time dependency.

Typical workflow to download a set of packages:
\enumerate{
\item Create a \code{pkg_download_proposal} object with
\code{new_pkg_download_proposal()}.
\item Resolve all possible dependencies with
\href{#method-resolve}{\code{pkg_download_proposal$resolve()}}.
\item Download all files with
\href{#method-download}{\code{pkg_download_proposal$download()}}.
\item Get the data about the packages and downloads with
\href{#method-get-downloads}{\code{pkg_download_proposal$get_downloads()}}.
}
}
\examples{
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method get_refs()
pdl <- new_pkg_download_proposal(c("pak", "jsonlite"))
pdl$get_refs()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method get_config()
pdl <- new_pkg_download_proposal("pak")
pdl$get_config()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method resolve()
pdl <- new_pkg_download_proposal("pak")
pdl$resolve()
pdl$get_resolution()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method get_resolution()
pdl <- new_pkg_download_proposal("r-lib/pkgdepends")
pdl$resolve()
pdl$get_resolution()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method download()
pdl <- new_pkg_download_proposal("r-lib/pkgdepends")
pdl$resolve()
pdl$download()
pdl$get_downloads()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method get_downloads()
pdl <- new_pkg_download_proposal("pkgload")
pdl$resolve()
pdl$download()
pdl$get_downloads()
\dontshow{\}) # examplesIf}
\dontshow{if (pkgdepends:::is_online()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Method print()
pdl <- new_pkg_download_proposal("r-lib/pkgdepends")
pdl

pdl$resolve()
pdl

pdl$download()
pdl
\dontshow{\}) # examplesIf}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-pkg_download_proposal-new}{\code{pkg_download_proposal$new()}}
\item \href{#method-pkg_download_proposal-get_refs}{\code{pkg_download_proposal$get_refs()}}
\item \href{#method-pkg_download_proposal-get_config}{\code{pkg_download_proposal$get_config()}}
\item \href{#method-pkg_download_proposal-resolve}{\code{pkg_download_proposal$resolve()}}
\item \href{#method-pkg_download_proposal-async_resolve}{\code{pkg_download_proposal$async_resolve()}}
\item \href{#method-pkg_download_proposal-get_resolution}{\code{pkg_download_proposal$get_resolution()}}
\item \href{#method-pkg_download_proposal-download}{\code{pkg_download_proposal$download()}}
\item \href{#method-pkg_download_proposal-async_download}{\code{pkg_download_proposal$async_download()}}
\item \href{#method-pkg_download_proposal-get_downloads}{\code{pkg_download_proposal$get_downloads()}}
\item \href{#method-pkg_download_proposal-stop_for_download_error}{\code{pkg_download_proposal$stop_for_download_error()}}
\item \href{#method-pkg_download_proposal-format}{\code{pkg_download_proposal$format()}}
\item \href{#method-pkg_download_proposal-print}{\code{pkg_download_proposal$print()}}
\item \href{#method-pkg_download_proposal-clone}{\code{pkg_download_proposal$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_download_proposal-new"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_download_proposal-new}{}}}
\subsection{Method \code{new()}}{
Create a new \code{pkg_download_proposal} object. Consider using
\code{new_pkg_download_proposal()} instead of calling the constructor
directly.

The returned object can be used to look up (recursive) dependencies
of R packages from various sources, and then to download the
package files.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_download_proposal$new(refs, config = list(), remote_types = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{refs}}{Package names or references. See
\link[=pkg_refs]{'Package references'} for the syntax.}

\item{\code{config}}{Configuration options, a named list. See
\link[=pkgdepends-config]{'Configuration'}.}

\item{\code{remote_types}}{Custom remote ref types, this is for advanced
use, and experimental currently.
\subsection{Examples}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{pdl <- pkg_download_proposal$new("r-lib/pkgdepends")
pdl
}\if{html}{\out{</div>}}
}}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_download_proposal-get_refs"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_download_proposal-get_refs}{}}}
\subsection{Method \code{get_refs()}}{
The package refs that were used to create the
\code{pkg_download_proposal} object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_download_proposal$get_refs()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A character vector of package refs that were used to create the
\code{pkg_download_proposal} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_download_proposal-get_config"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_download_proposal-get_config}{}}}
\subsection{Method \code{get_config()}}{
Configuration options for the \code{pkg_download_proposal} object. See
\link[=pkgdepends-config]{'Configuration'} for details.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_download_proposal$get_config()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Named list. See \link[=pkgdepends-config]{'Configuration'} for the configuration
options.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_download_proposal-resolve"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_download_proposal-resolve}{}}}
\subsection{Method \code{resolve()}}{
Resolve the dependencies of the specified package references. This
usually means downloading metadata from CRAN and Bioconductor,
unless already cached, and also from GitHub if GitHub refs were
included, either directly or indirectly. See
\link[=pkg_resolution]{'Dependency resolution'} for details.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_download_proposal$resolve()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
The \code{pkg_download_proposal} object itself, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_download_proposal-async_resolve"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_download_proposal-async_resolve}{}}}
\subsection{Method \code{async_resolve()}}{
The same as \href{#method-resolve}{\code{resolve()}}, but asynchronous.
This method is for advanced use.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_download_proposal$async_resolve()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A deferred value.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_download_proposal-get_resolution"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_download_proposal-get_resolution}{}}}
\subsection{Method \code{get_resolution()}}{
Query the result of the dependency resolution. This method can be
called after \href{#method-resolve}{\code{resolve()}} has completed.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_download_proposal$get_resolution()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A \link{pkg_resolution_result} object, which is also a data frame. See
\link[=pkg_resolution]{'Dependency resolution'} for its columns.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_download_proposal-download"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_download_proposal-download}{}}}
\subsection{Method \code{download()}}{
Download all resolved packages. It uses the package cache in the
pkgcache package by default, to avoid downloads if possible.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_download_proposal$download()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
The \code{pkg_download_proposal} object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_download_proposal-async_download"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_download_proposal-async_download}{}}}
\subsection{Method \code{async_download()}}{
The same as \href{#method-download}{\code{download()}}, but asynchronous.
This method is for advanced use.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_download_proposal$async_download()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A deferred value.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_download_proposal-get_downloads"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_download_proposal-get_downloads}{}}}
\subsection{Method \code{get_downloads()}}{
Returns the summary of the package downloads.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_download_proposal$get_downloads()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A \link{pkg_download_result} object, which is a list. See
\link{pkg_download_result} for details.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_download_proposal-stop_for_download_error"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_download_proposal-stop_for_download_error}{}}}
\subsection{Method \code{stop_for_download_error()}}{
Throw and error if the some of the downloads have failed for the
most recent
\href{#method-download}{\code{pkg_download_proposal$download()}} call.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_download_proposal$stop_for_download_error()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_download_proposal-format"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_download_proposal-format}{}}}
\subsection{Method \code{format()}}{
Format a \code{pkg_download_proposal} object, typically for printing.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_download_proposal$format(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{not used currently.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Nothing.
A character vector, each element should be a line in the printout.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_download_proposal-print"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_download_proposal-print}{}}}
\subsection{Method \code{print()}}{
Prints a \code{pkg_download_proposal} object to the screen. The printout
includes:
\itemize{
\item The package refs.
\item Whether the object has the resolved dependencies.
\item Whether the resolution had errors.
\item Whether the downloads were completed.
\item Whether the downloads had errors.
\item Advice on which methods to call next.
}

See the example below.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_download_proposal$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{not used currently.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The \code{pkg_download_proposal} object itself, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-pkg_download_proposal-clone"></a>}}
\if{latex}{\out{\hypertarget{method-pkg_download_proposal-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{pkg_download_proposal$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rx.R
\name{pkg_rx}
\alias{pkg_rx}
\title{A set of handy regular expressions related to R packages}
\usage{
pkg_rx()
}
\value{
A named list of strings.
}
\description{
If you use these in R, make sure you specify \code{perl = TRUE}, see
\code{\link[base:grep]{base::grep()}}.
}
\details{
Currently included:
\itemize{
\item \code{pkg_name}: A valid package name.
\item \code{type_cran}: A \verb{cran::} package reference.
\item \code{type_bioc}: A \verb{bioc::} package reference.
\item \code{type_standard}: A \verb{standard::} package reference.
\item \code{type_github}: A \verb{github::} package reference.
\item \code{type_git}: A \verb{git::} package reference.
\item \code{type_local}: A \verb{local::} package reference.
\item \code{type_deps}: A \verb{deps::} package reference.
\item \code{type_installed}: An \verb{installed::} package reference.
\item \code{github_username}: A GitHub username.
\item \code{github_repo}: A GitHub repository name.
\item \code{github_url}: A GitHub URL.
}
}
\examples{
pkg_rx()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download.R
\name{pkg_downloads}
\alias{pkg_downloads}
\alias{pkg_download_result}
\title{Package downloads}
\description{
The \code{\link{pkg_download_proposal}} and \code{\link{pkg_installation_proposal}} classes
both have download methods, to downloads package files into a
configured directory (see \link[=pkgdepends-config]{'Configuration'}).
}
\details{
They return a \code{pkg_download_result} object, which is a data frame,
that adds extra columns to \code{\link{pkg_resolution_result}} (for
\code{\link{pkg_download_proposal}}) or \code{\link{pkg_solution_result}}
(for \code{\link{pkg_installation_proposal}}):
\itemize{
\item \code{built}: the \code{Built} field from the \code{DESCRIPTION} file of binary
packages, for which this information is available.
\item \code{cache_status}: whether the package file is in the package cache.
It is \code{NA} for \verb{installed::} package refs.
\item \code{dep_types}: character vector of dependency types that were
considered for this package. (This is a list column.)
\item \code{deps}: dependencies of the package, in a data frame. See
"Package dependency tables" below.
\item \code{direct}: whether this package (ref, really) was directly specified,
or added as a dependency.
\item \code{error}: this is a list column that contains error objects for the
refs that pkgdepends failed to resolve.
\item \code{filesize}: the file size in bytes, or \code{NA} if this information is
not available.
\item \code{license}: license of the package, or \code{NA} if not available.
\item \code{md5sum}: MD5 checksum of the package file, if available, or \code{NA} if
not.
\item \code{metadata}: a named character vector. These fields will be (should be)
added to the installed \code{DESCRIPTION} file of the package.
\item \code{mirror}: URL of the CRAN(-like) mirror site where the metadata was
obtained from. It is NA for non-CRAN-like sources, e.g. local files,
installed packages, GitHub, etc.
\item \code{needscompilation}: whether the package needs compilation.
\item \code{package}: package name.
\item \code{priority}: this is \code{"base"} for base packages, \code{"recommended"} for
recommended packages, and \code{NA} otherwise.
\item \code{ref}: package reference.
\item \code{remote}: the parsed \code{remote_ref} objects, see \code{\link[=parse_pkg_refs]{parse_pkg_refs()}}.
This is a list column.
\item \code{repodir}: the directory where this package should be in a CRAN-like
repository.
\item \code{sha256}: SHA256 hash of the package file, if available, otherwise
\code{NA}.
\item \code{sources}: URLs where this package can be downloaded from. This is not
necessarily a URL that you can download with a HTTP client. E.g. for
\verb{local::} refs it is a path, and for \verb{git::} refs it is a URL for git.
It is a zero length vector for \verb{installed::} refs.
\item \code{status}: status of the dependency resolution, \code{"OK"} or \code{"FAILED"}.
\item \code{target}: path where this package should be saved in a CRAN-repository.
\item \code{type}: ref type.
\item \code{version}: package version.
\item \code{fulltarget}: absolute path to the downloaded file. At most one of
\code{fulltarget} and \code{fulltarget_tree} must exist on the disk.
\item \code{fulltarget_tree}: absolute path to a package tree directory. At most
one of \code{fulltarget} and \code{fulltarget_tree} must exist on the disk.
\item \code{download_status}: \code{"Had"} or \code{"Got"}, depending on whether the file
was obtained from the cache.
\item \code{download_error}: error object for failed downloads.
\item \code{file_size}: Size of the file, or \code{NA}. For \verb{installed::} refs, it is
\code{NA}, and it is also \code{NA} for refs that created \code{fulltarget_tree}
instead of \code{fulltarget}.
}

\code{fulltarget}, if it exists, contains a packaged (via \verb{R CMD build})
source R package. If \code{fulltarget_tree} exists, it is a package tree
directory, that still needs an \verb{R CMD build} call.

Additional columns might be present. They are either used internally or
they are experimental. They might be removed or changed at any time.

All columns are of type character, except for \code{direct} (logical),
\code{needscompilation} (logical), \code{filesize} (integer), \code{deps} (list column, see
"Package dependency tables" below), \code{sources} (list of character vectors),
\code{remote} (list), \code{error} (list), \code{metadata} (list), \code{dep_types} (list).
\subsection{Package dependency tables}{

A package dependency tables in the \code{deps} list column have five columns
currently:
\itemize{
\item \code{ref}: the package ref of the dependency.
\item \code{type}: the dependency type, in all lowercase. I.e. \code{imports},
\code{suggests}, etc.
\item \code{package}: package name of the dependency.
\item \code{op}: operator for version requirements, e.g. \code{>=}.
\item \code{version}: version number, for version requirements.
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sysreqsdocs.R
\name{sysreqs_check_installed}
\alias{sysreqs_check_installed}
\alias{sysreqs_fix_installed}
\title{Check if installed packages have all their system requirements}
\usage{
sysreqs_check_installed(packages = NULL, library = .libPaths()[1])
sysreqs_fix_installed(packages = NULL, library = .libPaths()[1])
}
\arguments{
\item{packages}{If not \code{NULL}, then only these packages are checked.
If a package in \code{packages} is not installed, then
pkgdepends throws a warning.}

\item{library}{Library or libraries to check.}
}
\value{
Data frame with a custom print and format method, and a
\code{pkg_sysreqs_check_result} class. Its columns are:
\itemize{
\item \code{system_package}: string, name of the required system package.
\item \code{installed}: logical, whether the system package is correctly
installed.
\item \code{packages}: list column of character vectors. The names of the
installed R packages that need this system package.
\item \code{pre_install}: list column of character vectors. Commands to run
before the installation of the the system package.
\item \code{post_install}: list column of character vectors. Commands to run
after the installation of the system package.
}

The data frame also have two attributes with additional data:
\itemize{
\item \code{sysreqs_records}: the raw system requirements records, and
\item \code{system_packages}: the list of the installed system packages.
}

\code{sysreqs_fix_packages()} returns the same value, but invisibly.
}
\description{
\code{sysreqs_check_installed()} checks if the system requirements of all
packages (or a subset of packages) are installed.

\code{sysreqs_fix_installed()} installs the missing system packages.
}
\details{
These functions use the \code{sysreqs_platform} configuration option,
see \eval{man_config_link("Configuration")}. Set this if
pkgdepends does not detect your platform correctly.
}
\examples{
\dontshow{if (Sys.getenv("IN_PKGDOWN") == "true" && Sys.info()[["sysname"]] == "Linux") (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# This only works on supported platforms
sysreqs_check_installed()
\dontshow{\}) # examplesIf}
}
\seealso{
Other system requirements functions: 
\code{\link{sysreqs_db_list}()},
\code{\link{sysreqs_db_match}()},
\code{\link{sysreqs_db_update}()},
\code{\link{sysreqs_install_plan}()},
\code{\link{sysreqs_is_supported}()},
\code{\link{sysreqs_list_system_packages}()},
\code{\link{sysreqs_platforms}()}
}
\concept{system requirements functions}
